package fuegocore.notify;

import java.net.*;
import java.util.*;

import fuegocore.message.*;
import fuegocore.message.protocol.*;
import fuegocore.message.service.*;
import fuegocore.notify.core.channel.*;
import fuegocore.notify.core.server.*;
import fuegocore.notify.core.server.engine.*;
import fuegocore.notify.core.server.handlers.*;
import fuegocore.notify.util.*;

/**
 * This class starts the access server, and the resolution server
 * according to the configuration information.  Currently both listen
 * on the same port, using different rpc identifiers (rpc and rpc_ec).
 * Communication between these components uses local object passing
 * if they are in the same VM, if the event channel is on a different VM
 * Fuego RPC is used.
 */
public class StartServer {
  static int current_port = 80;
  static int admin_port = 8083;
  static String server = null;
  static String ecserver = null;
  static String ec_config = null;

  /** master event channel table, built from configuration information */
  static Hashtable event_channels;

  public static void main(String[] args) {
    int v = 0;
    try {
      v = (int) ( (new Integer(args[0])).intValue());
    }
    catch (Exception e) {
      System.out.println("Exception:" + e);
    }

    initServer("servers.cfg", v);

    System.out.print("Deploying Access Server and Resolution Server:");

    // Set access server identifier
    fuegocore.notify.profile.CommProfile.access_server_id = server;
    fuegocore.notify.profile.CommProfile.event_server_id = ecserver;

    // add all event types (channels) + their URI to the access server lookup table

    // Creater router part
    FuegoRoutingEngine fuego = new FuegoRoutingEngine();
    EventServer es = EventServer.create(fuego);

    es.getRoutingEngine().getRouterManager().addEventChannels( (Hashtable)
        event_channels.clone());

    if (es.getRoutingEngine().getRouterManager().event_channels == null) {
      System.out.println("Fatal error: Event channel lookup table was null");
      System.exit(0);
    }
    try {

      MessageTransportService mts = MessageTransportService.init();
      MessageService ms = mts.getService();
      System.out.println("Starting access server");
      AccessServerHandler ace = new AccessServerHandler(es);
      ms.registerReceiver("messagerpc", ace);
      ms.registerRpcReceiver("rpc", ace);
      ms.registerReceiver("message", es);
      System.out.println("Starting event channel server");
      EventChannelServerHandler ecsh = new EventChannelServerHandler(ec_config);
      ms.registerRpcReceiver("rpc_ec", ecsh);
      ms.registerReceiver("message_ec", ecsh);

      ms.listen(current_port,
                Url.Protocol.parse(fuegocore.notify.profile.UserProfile.
                                   DEF_MSG_PROTOCOL));

    }
    catch (Exception ex) {
      ex.printStackTrace(System.err);
    }

    fuegocore.notify.comms.HTTPServlet.PORT = admin_port; // set http port
    fuegocore.notify.comms.HTTPServlet.startHttpServlet();

    try {
      System.out.println("Formulating URL: " + ecserver);
      EventChannelServer.get().setURL(new URL("http://" + ecserver));
    }
    catch (Exception e) {
      e.printStackTrace();
    }

  }

  /**
   * Configure server using a configuration file
   * @param file String
   * @param current int
   */
  public static void initServer(String file, int current) {
    String clientdata = "";

    try {
      ReadFile rf = new ReadFile(file);
      if (rf == null) {
        return;
      }

      // Extract contents
      String tmp = rf.readValue("count");
      int count = 0;

      if (tmp != null) {
        count = (int) (new Integer(tmp.trim())).intValue();
      }
      if (count < 0) {
        count = 0; // limits
      }
      else if (count > 1000) {
        count = 1000;

      }
      for (int i = 0; i < count; i++) {
        if (i == current) {
          server = rf.readValue("server");
          if (server != null) {
            server = server.trim();
          }
          ecserver = rf.readValue("ec-server");
          if (ecserver != null) {
            ecserver = ecserver.trim();
          }
          ec_config = rf.readValue("channel-config");
          if (ec_config != null) {
            ec_config = ec_config.trim();
            getChannels(ec_config, ecserver);
          }
          current_port = fuegocore.notify.util.Auxil.getPort(ecserver);

          tmp = rf.readValue("admin-port");
          if (tmp != null) {
            admin_port = (int) (new Integer(tmp.trim())).intValue();
          }
        }
        else {
          tmp = rf.readValue("server"); // dummy read
          tmp = rf.readValue("ec-server");
          if (tmp != null) {
            tmp = tmp.trim();
          }
          String tmp2 = rf.readValue("channel-config"); // dummy read
          if (tmp2 != null) {
            tmp2 = tmp2.trim();
            getChannels(tmp2, tmp);
          }
          tmp = rf.readValue("admin-port");
        }
      }
      rf.close();
    }
    catch (Exception e) {
      System.out.println("Exception in reading a file:" + e);
      e.printStackTrace();
    }
  }

  /**
   * Builds a master list of event channels in the system
   * @param file String
   * @param host String
   */
  public static void getChannels(String file, String host) {
    if (event_channels == null) {
      event_channels = new Hashtable();
    }
    try {
      ReadFile rf = new ReadFile(file);
      if (rf == null) {
        return;
      }

      // Extract contents
      String tmp = rf.readValue("count");
      int count = 0;
      if (tmp == null) {
        count = 0;
      }
      else {
        count = (int) (new Integer(tmp.trim())).intValue();

      }
      if (count < 0) {
        count = 0; // limits
      }
      else if (count > 1000) {
        count = 1000;

      }
      for (int i = 0; i < count; i++) {
        String type = rf.readValue("type").trim();
        tmp = rf.readValue("maxsubs");
        // We use HTTP as a placeholder
        event_channels.put(type, new URL("http://" + host));
      }
      rf.close();
    }
    catch (Exception e) {
      System.out.println("Exception in reading a file:" + e);
      e.printStackTrace();
    }
  }

}
