/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.comms;

import java.io.*;
import java.net.*;
import java.util.*;

import fuegocore.message.*;
import fuegocore.message.protocol.*;
import fuegocore.notify.core.*;
import fuegocore.notify.core.server.*;
import fuegocore.notify.core.session.*;

/**
 * This class is responsible for buffering and sending notifications
 * using the underlying transport architecture. Sender thread may be
 * paused during disconnections.
 */

public class ServerChannel implements IQueueManager, ConnectionMonitor {
    public EventServer ms = null;
    Vector outbound = new Vector();
    Socket so;
    ServerSenderThread st;
    public String sender = "no def";
    public String origin = null;
    QueueManager qm = new QueueManager(this); // queue manager
    long lastDeadline = 0;
    long RETRANSMISSION_TIMER = 1000; // retransmit timer
    int MAX_RETRY = 10; // maximum number of retries before pausing
    public int MAX_QUEUE = 100; // maximum queue size
    String _connection_id = "";
    boolean orderedDelivery = false; // ordered (acknowledge mode)
    boolean alive = true;
    boolean handover = false;
    boolean pause = false;
    boolean queuemanager = false; // queue manager active?
    int state = 0;
    SessionParams sp = null;

    public ServerChannel() {
        // Activate queue manager
        if (queuemanager) {
            qm.start();
        }
        st = new ServerSenderThread(this);
        pause = false;
        st.start();
    }

    /**
     * Set the maximum queue size
     * @param i the maximum queue size
     */
    public void setQueueSize(int i) {
        MAX_QUEUE = i;
    }

    /** Get the maximum queue size
     *  @return maximum queue size
     */
    public int getMaxQueueSize() {
        return MAX_QUEUE;
    }

    public boolean getPause() {
        return pause;
    }

    public void setPause(boolean b) {
        pause = b;
    }

    /** Set handover bit on.
     */
    public void handOverInProgress() {
        handover = true;
    }

    public ServerChannel(String cn, EventServer s, SessionParams p) {
        this(cn, s);
        sp = p;
    }

    public void setSessionParams(SessionParams p) {
        sp = p;
    }

    public ServerChannel(String cn, EventServer s) {
        _connection_id = cn;

        /*
         Register monitor
         */

        MessageTransportService mts = MessageTransportService.init();

        MessageProtocol mp = null;

        mp = mts.getService().getProtocol(fuegocore.notify.profile.UserProfile.
                                          DEF_MSG_PROTOCOL);

        mp.registerConnectionMonitor(cn, this);

        sender = cn;
        ms = s;
        // Activate queue manager
        if (queuemanager) {
            qm.start();
        }
        st = new ServerSenderThread(this);
        st.start();
    }

    /** Sets the delivery mode
     * @param order true for ordered delivery
     */
    public void setDeliveryMode(boolean order) {
        orderedDelivery = order;
    }

    /** Returns the delivery mode
     * @return true for ordered delivery
     */
    public boolean getDeliveryMode() {
        return orderedDelivery;
    }

    /**
     * Called when a physical connection is broken.
     */
    public void connectionAsleep() {
        fuegocore.notify.util.Log.writeTransport(
                "------------  Connection Asleep");
        pause();
    }

    /**
     * Called when a physical connection is re-established.
     */
    public void connectionAwake() {
        fuegocore.notify.util.Log.writeTransport(
                "------------  Connection Awake");
        kick();
    }

    /**
     * Called when a physical connection fails.
     */
    public void connectionFailure() {
        fuegocore.notify.util.Log.writeTransport(
                "------------  Connection paused due to failures");
        pause();
    }

    /**
     * Returns the state of the MTP.
     * @return int the current state
     */
    public int getState() {
        return state;
    }

    /** Returns the URL of the connection
     * @return String the current URL
     */
    public String getSender() {
        return sender;
    }

    public String getOrigin() {
        return origin;
    }

    /** Set the URL of the connection
     * @param url the data sender
     * @param ori the data origin
     */
    public void setURL(String url, String ori) {
        if (sender != null && sender.equals(url)) {
            return;
        }

        fuegocore.notify.util.Log.writeTransport(
                "------------  Set new connection url");
        String old_url = sender;

        sender = url;

        /*
          Register monitor
         */
        if (!pause && ori != null) {
            MessageTransportService mts = MessageTransportService.init();
            MessageProtocol mp = null;

            mp = mts.getService().getProtocol(fuegocore.notify.profile.
                                              UserProfile.
                                              DEF_MSG_PROTOCOL);

            // unregister old one
            if (origin != null) {
                mp.unregisterConnectionMonitor(origin);

                // register new new
            }
            try {
                mp.registerConnectionMonitor(ori, this);
            } catch (Exception e) {
                System.out.println(
                        "WARNING: Problem with registerConnectionMonitor:" +
                        e);
                System.out.println("Origin: " + ori);
            }
        }
        origin = ori;
    }

    /** Get reference to outbound queue
     * @return Vector the priority queue
     */
    public Vector getOutbound() {
        return outbound;
    }

    /** Iterates over the queue and checks the deadlines.
     * This is called by the QueueManager thread.
     * @return long the last deadline
     */
    public long manageQueue() {
        long min = 1000000000;

        Vector out = null;

        try {
            synchronized (outbound) {
                out = (Vector) outbound.clone();
            }
        } catch (Exception e) {
            out = null;
        }

        lastDeadline = 0;

        fuegocore.notify.util.Log.writeTransport(
                "ManageQueue: processing outbound queue");

        if (out == null) {
            return 0;
        }

        for (int i = 0; i < out.size(); i++) {
            Notification msg = (Notification) out.elementAt(i);

            if (msg == null) {
                continue;
            }

            long deadline = msg.getDeadline();

            if (deadline == 0) {
                continue;
            }

            deadline -= System.currentTimeMillis();

            if (deadline < 0) {
                // Deadline has passed

                out.remove((Notification) msg);
                fuegocore.notify.util.Log.writeTransport("Discarded: [" +
                        msg.toString() +
                        "]");
                /*
                         Currently there is no failure policy.
                 */
            } else {
                if (deadline < min) {
                    min = deadline;
                    lastDeadline = msg.getDeadline();
                }
            }
        }

        return lastDeadline;
    }

    /** Replaces obsolete messages from the session queue.
     *
     * @param id The replace id
     */
    public void replaceObsoleteMessages(String id) {
        Vector removeList = new Vector();

        for (int i = 0; i < outbound.size(); i++) {
            Notification msg = (Notification) outbound.elementAt(i);
            if (msg != null && msg.getReplaceId() != null &&
                msg.getReplaceId().equals(id)) {
                removeList.add(msg);
            }
        }
        synchronized (outbound) {
            for (int i = 0; i < removeList.size(); i++) {
                Notification msg = (Notification) removeList.elementAt(i);
                if (msg != null) {
                    outbound.remove(msg);
                }
            }
        }
    }

    public boolean checkDuplicate(Notification n) {
        boolean flag = false;
        Vector o = null;
        synchronized (outbound) {
            o = (Vector) outbound.clone();
        }
        for (int j = 0; j < o.size(); j++) {
            Notification n2 = (Notification) o.elementAt(j);
            if (n2 != null) {
                if (n2.getACInterface() != null && n.getACInterface() != null &&
                    n2.getACInterface().equals(n.getACInterface()) &&
                    n2.getOrder() == n.getOrder() && n2.getOrder() > 1) {
                    flag = true;
                    break;
                }
            }
        }
        return flag;
    }

    /** Merges new messages and removes duplicates. Duplicate detection
     *  is based on sequence numbers and source addresses.
     *  @param m the messages to merge
     */
    public void mergeNewMessages(Vector m) {
        if (m == null) {
            return;
        }
        for (int i = 0; i < m.size(); i++) {
            Notification n = (Notification) m.elementAt(i);
            boolean flag = false;
            for (int j = 0; j < outbound.size(); j++) {
                Notification n2 = (Notification) outbound.elementAt(j);
                if (n2 != null) {
                    if (n2.getACInterface() != null && n.getACInterface() != null &&
                        n2.getACInterface().equals(n.getACInterface()) &&
                        n2.getOrder() == n.getOrder() && n2.getOrder() > 1) {
                        flag = true;
                        break;
                    }
                }
            }
            if (!flag) {
                outbound.add(n);
            }
        }
    }

    /** One-shot pull functionality (uses template matching)
     * @param templates the templates to match, if null all is pulled up to count
     * @param count the number of notifications to pull, if 0 then all matching are returned
     * @param destroy if true the pulled notifications are destroyed
     * @return Vector the notifications
     */
    public Vector pull(Vector templates, int count, boolean destroy) {
        int cc = 0;
        Vector ret = new Vector();
        Vector removeList = new Vector();

        for (int i = 0; i < outbound.size(); i++) {
            if (count > 0 && cc >= count) {
                break;
            }
            Notification n = (Notification) outbound.elementAt(i);
            if (n != null) {
                // Check if the notification matches the given templates
                if (templates == null) {
                    // pull all
                    ret.add(n);
                    cc++;
                    if (destroy) {
                        removeList.add(n);
                    }
                } else {
                    for (int j = 0; j < templates.size(); j++) {
                        ITemplate t2 = (ITemplate) templates.elementAt(j);
                        if (t2 == null) {
                            continue;
                        }
                        if (((fuegocore.notify.core.filter.IFBase) t2).match(null,
                                n)) {
                            ret.add(n);
                            cc++;
                            if (destroy) {
                                removeList.add(n);
                            }
                            break;
                        }
                    }
                }
            }
        }
        if (destroy) {
            synchronized (outbound) {
                outbound.removeAll(removeList);
            }
        }
        return ret;
    }

    /** One-shot pull functionality (uses subscription identifiers)
     * @param subs susbcription identifiers separated by commas
     * @param count the number of notifications to pull, if 0 then all matching are returned
     * @param destroy if true the pulled notifications are destroyed
     * @return Vector the notifications
     */
    public Vector pullSubs(String subs, int count, boolean destroy) {
        int cc = 0;
        Vector ret = new Vector();
        Vector removeList = new Vector();
        Vector ids = null;

        if (subs != null) {
            ids = new Vector();
            try {
                StringTokenizer tok = new StringTokenizer(subs, ",");
                if (tok != null) {
                    while (tok.hasMoreTokens()) {
                        String s = (String) tok.nextToken().trim();
                        ids.add(s);
                    }
                }
            } catch (Exception e) {
                ids = null;
                e.printStackTrace();
            }
        }

        for (int i = 0; i < outbound.size(); i++) {
            if (count > 0 && cc >= count) {
                break;
            }

            Notification n = (Notification) outbound.elementAt(i);
            if (n != null) {
                // Check if the notification matches the given templates
                if (ids == null || (ids != null && ids.size() < 1) ||
                    (ids != null && ids.equals("*"))) {
                    // pull all
                    ret.add(n);
                    cc++;
                    if (destroy) {
                        removeList.add(n);
                    }
                } else {
                    for (int j = 0; j < ids.size(); j++) {
                        String r = (String) ids.elementAt(j);
                        //System.out.println("Pull: Comparing: " + r + " to " + n.getSubIdString());
                        if (n != null && n.getSubId() != null &&
                            n.getSubId().contains(r)) {
                            ret.add(n);
                            cc++;
                            if (destroy) {
                                removeList.add(n);
                            }
                            break;
                        }
                    }
                }
            }
        }

        if (destroy) {
            synchronized (outbound) {
                outbound.removeAll(removeList);
            }
        }

        return ret;
    }

    /** Client calls this to send a message. The
     *  message is queued in the outgoing queue and
     *  the background thread is woken up if necessary.
     * @param msg the message to send
     * @return true if the send succeeded
     */
    public boolean send(Notification msg) {
        // If handover is in progress all new messages will be discarded
        if (handover) {
            return true;
        }

        // Check if the message is a duplicate
        // We may get some duplicates after a handover so they are removed here.
        if (checkDuplicate(msg)) {
            return true;
        }

        synchronized (outbound) {
            outbound.add(msg);

            // Sort the collection
            Collections.sort(outbound);

            Vector addList = new Vector();

            // Move those notifications that match the priority filter (if any) to top
            if (sp.getPriorityFilter() != null &&
                sp.getPriorityFilter() instanceof Template) {
                for (int i = 0; i < outbound.size(); i++) {
                    Notification n = (Notification) outbound.elementAt(i);
                    if (n == null) {
                        continue;
                    }
                    if (Template.match(sp.getPriorityFilter(),
                                       n.get().getTriplets())) {
                        outbound.remove(n);
                        addList.add(n);
                    }
                }

                if (addList.size() > 0) {
                    addList.addAll(outbound);
                    outbound.clear();
                    outbound.addAll(addList);
                }
            }

            if (outbound.size() > MAX_QUEUE) { // enforce queue size limit
                outbound.remove(outbound.lastElement()); // remove oldest
            }

            // Notify any interested objects
            outbound.notifyAll();
        }

        long t = msg.getDeadline() - System.currentTimeMillis();
        if (t >= 0 && msg.getDeadline() < lastDeadline) {
            lastDeadline = msg.getDeadline();
        }

        if (t < lastDeadline) {
            qm.newDeadline(msg.getDeadline());
        }

        fuegocore.notify.util.Log.writeTransport("Added to queue: " +
                                                 msg.toString());
        return true;
    }

    /** Shutdown comms but not queue management
     * @throws IOException
     */
    public void close() throws IOException {
    }

    /** Pause messaging */
    public void pause() {
        if (st != null) {
            st.shutdown();
            pause = true;
        }
    }

    /** Continue messaging */
    public void kick() {
        if ((st != null && st.alive != true) || st == null) {
            fuegocore.notify.util.Log.writeTransport(
                    "Starting transport thread");
            pause = false;
            st = null;
            st = new ServerSenderThread(this);
            st.start();
        } else {
            fuegocore.notify.util.Log.writeTransport(
                    "Transport thread already active");
        }
    }

    /** Cleanup by waiting for the thread to
     * stop and then closing the record stores.
     */
    public synchronized void destroy() {
        // shut down the queue manager
        qm.shutdown();

        // destroy outbound queue
        outbound.clear();
        st.shutdown();
        st = null;
        outbound = null;

        // Remove reference from transport layer
        MessageTransportService mts = MessageTransportService.init();
        MessageProtocol mp = null;
        mp = mts.getService().getProtocol(fuegocore.notify.profile.UserProfile.
                                          DEF_MSG_PROTOCOL);
        mp.unregisterConnectionMonitor(sender);
        alive = false;
    }
}
