/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core;

import java.text.*;
import java.util.*;

/**
 *  The basic object for event contents.
 */
public class EventTriplet {
    /** Event field / attribute of integer */
    public static final int TYPE_INT = 0;

    /** Event field / attribute of boolean */
    public static final int TYPE_BOOLEAN = 1;

    /** Event field / attribute of double */
    public static final int TYPE_DOUBLE = 2;

    /** Event field / attribute of String */
    public static final int TYPE_STRING = 3;

    /** Event field / attribute type of binary */
    public static final int TYPE_BINARY = 4;

    /** Event field / attribute type of URI */
    public static final int TYPE_URI = 5;

    /** Event field / attribute type of serialized data */
    public static final int TYPE_SERIALIZED = 6;

    /** Event field / attribute type of date */
    public static final int TYPE_DATE = 7;

    String name;
    int type;
    Object o;

    /** Set the triplet name, type and value.
     * @param n the name of the triplet
     * @param t the type of the triplet
     * @param ob the value of the triplet
     */
    public EventTriplet(String n, int t, Object ob) {
        name = n;
        type = t;
        o = ob;
    }

    /** Set the triplet name, value (type implicit).
     * @param n the name of the triplet
     * @param s the value of the triplet
     */
    public EventTriplet(String n, String s) {
        name = n;
        type = TYPE_STRING;
        o = s;
    }

    /** Set the triplet name, value (type implicit).
     * @param n the name of the triplet
     * @param b the value of the triplet
     */
    public EventTriplet(String n, double b) {
        name = n;
        type = TYPE_DOUBLE;
        o = new Double(b);
    }

    /** Set the triplet name, value (type implicit).
     * @param n the name of the triplet
     * @param i the value of the triplet
     */
    public EventTriplet(String n, int i) {
        name = n;
        type = TYPE_INT;
        o = new Integer(i);
    }

    /** Set the triplet name, value (type implicit).
     * @param n the name of the triplet
     * @param b the value of the triplet
     */
    public EventTriplet(String n, boolean b) {
        name = n;
        type = TYPE_BOOLEAN;
        o = new Boolean(b);
    }

    /** Set the triplet name, value (type implicit).
     * @param n the name of the triplet
     * @param d the value of the triplet
     */
    public EventTriplet(String n, Date d) {
        name = n;
        type = TYPE_DATE;
        o = d;
    }

    /**
     * Get the triplet value as String.
     * @return String the triplet as a String
     */
    public String getValueAsString() {
        if (type == TYPE_INT) {
            Integer i = (Integer) o;
            return "" + i.intValue();
        } else if (type == TYPE_BOOLEAN) {
            Boolean b = (Boolean) o;
            return "" + b.booleanValue();
        } else if (type == TYPE_DOUBLE) {
            Double i = (Double) o;
            return "" + i.doubleValue();
        } else if (type == TYPE_STRING) {
            return (String) o;
        } else if (type == TYPE_DATE) {
            Date d = (Date) o;
            return "" + d.toString();
        } else if (type == TYPE_SERIALIZED) {
            return "" + o.toString();
        } else {
            return "unknown type";
        }
    }

    /**
     * Get the type of the triplet as a String.
     * @param type the type as an integer
     * @return the type as a String
     */
    public static String getTypeString(int type) {
        if (type == TYPE_INT) {
            return "integer";
        } else if (type == TYPE_BOOLEAN) {
            return "boolean";
        } else if (type == TYPE_DOUBLE) {
            return "double";
        } else if (type == TYPE_STRING) {
            return "string";
        } else if (type == TYPE_DATE) {
            return "date";
        } else if (type == TYPE_SERIALIZED) {
            return "serialized";
        } else {
            return "unknown type";
        }
    }

    /** Set the triplet name, type and value.  Value is a string
     * and converted to the object format.
     * @param n the name of the triplet
     * @param t the type of the triplet
     * @param s the Value of the triplet as a String
     */
    public EventTriplet(String n, int t, String s) {
        // value conversion
        if (t == TYPE_INT) {
            try {
                o = new Integer(s);
            } catch (Exception e) {
                e.printStackTrace();
            }
        } else if (t == TYPE_BOOLEAN) {
            try {
                o = new Boolean(s);
            } catch (Exception e) {
                e.printStackTrace();
            }
        } else if (t == TYPE_DOUBLE) {
            try {
                o = new Double(s);
            } catch (Exception e) {
                e.printStackTrace();
            }
        } else if (t == TYPE_STRING) {
            o = s;
        } else if (t == TYPE_BINARY) {
            ; // not currently supported
        } else if (t == TYPE_SERIALIZED) {
            ; // not currently supported
        } else if (t == TYPE_DATE) {
            try {
                o = (Date) DateFormat.getInstance().parse(s);
            } catch (Exception e) {
                e.printStackTrace();
            }
        } else {
            fuegocore.notify.util.Log.write("EventTriplet: unknown type " +
                                            type);

            // set other values
        }
        name = n;
        type = t;
    }

    /** Get field/attribute name
     * @return String the name of the triplet
     */
    public String getName() {
        return name;
    }

    /** Set field/attribute name
     * @param n the name of the triplet
     */
    public void setName(String n) {
        name = n;
    }

    /** Get field/attribute type
     * @return int the type of the triplet
     */
    public int getType() {
        return type;
    }

    /** Get field/attribute value as an Object
     * @return Object the value of the triplet
     */
    public Object getValue() {
        return o;
    }

    /** Get field/attribute identifier (name/type)
     * @return String the event triplet identifier (name+type)
     */
    public String getId() {
        return name + "/" + type;
    }

    public String toString() {
        return getId() + " " + getValueAsString();
    }

    /** Auxiliary method for formatting triplet IDs
     * @param n the name of the triplet
     * @param t the type of the triplet
     * @return String the triplet identifier
     */
    public static String formatId(String n, int t) {
        return n + "/" + t;
    }
}
