/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core;

import java.util.*;

/**
 * Interface for templates. Template encapsulates a subscription / filter.
 */
public interface ITemplate {
    /**
     * Set quench value
     * @param b boolean
     */
    public void setQuench(boolean b);

    /**
     * Get quench value
     * @return boolean
     */
    public boolean getQuench();

    /**
     * Flag the template as an advertisement
     * @param b boolean
     */
    public void setAdvertisement(boolean b);

    /**
     * Returns the advertisement flag
     * @return boolean
     */
    public boolean getAdvertisement();

    /** Set the subscription ID
     * @param s the id
     */
    public void setSubId(String s);

    /** Get the subscription ID
     * @return the subscription identifier
     * */
    public String getSubId();

    /** Set the session ID
     * @param s the session ID to set
     */
    public void setSessionId(String s);

    /** Get the session ID
     * @return the session identifier
     */
    public String getSessionId();

    /** Formats a notification according to the required notification fields
     * specified in this template
     * @param n the notification
     * @return the formatted notification
     */
    public Notification formatNotification(Notification n);

    /** Sets the local send property. If this is true notifications from the client are forwarded back to the client
     *  if they match one of the client's active subscriptions.
     * @param b the local send option
     */
    public void setLocalSend(boolean b);

    /** Returns the local send property.  If this is true notifications from the client are forwarded back to the client
     *  if they match one of the client's active subscriptions.
     * @return true if local send is active
     */
    public boolean getLocalSend();

    /** Sets the scope of the subscription (template) to local */
    public void setScopeToLocal();

    /** Sets the scope of the subscription (template) to external */
    public void setScopeToExternal();

    /** Returns the current scope local (true) / external (false)
     * @return true if scope is local */
    public boolean getScope();

    /** Adds the given filter to the template.
     * @param f the attribute filter to add
     */
    public void addFilter(Object f);

    /** Removes the given filter.
     * @param f the filter to remove
     */
    public void removeFilter(Object f);

    /** Returns the set of filters.
     * @return returns a Vector containing the filter objects
     */
    public Vector getFilters();

    /** Set the attribute filters
     *  @param v the Vector
     */
    public void setFilters(Vector v);

    /* *************************************************************************/
    /* *************************************************************************/

    /**
     * Set the subscribing interface
     * @param s String
     */
    public void setInterface(String s);

    /**
     * Get the subscribing interface
     * @return String
     */
    public String getInterface();

    /** Add the type of the subscription to the template (event channel name)
     * @param type the type to add
     */
    public void addType(String type);

    /** Returns the type of the subscription (event channel name)
     * @return String the type of the notification */
    public String getType();

    /** Get the source of the subscription / template
     * @return Notifiable the source of the template
     */
    public Notifiable getSource();

    /** Set the source of the subscription / template
     * @param eb the Notifiable interface
     */
    public void setSource(Notifiable eb);

    /**
     * Set event channel flag (internal use)
     * @param b boolean
     */
    public void setECFlag(boolean b);

    /**
     * Get event channel flag (internal use)
     * @return boolean
     */
    public boolean getECFlag();

    /** Returns the notify field names (separated by commas) as a String.
     *  The notify fields determine which fields are forwarded in the
     *  notification and which fields are omitted. This is an optional feature.
     *  @return the notify field names
     */
    public String getNotifyFields();

    /** Sets the notify field names (separated by commas) as a String.
     *  @param s the notify fields
     */
    public void setNotifyFields(String s);
}
