/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core;

import java.io.*;
import java.util.*;

import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.serialization.*;
import fuegocore.notify.util.*;
import fuegocore.util.xas.*;

/**
 * Notification represents the information contents of an event.
 * Each event consists of n EventTriplet-objects, which have
 * the following structure: (name, type, value). This object
 * currently supports two serialization semantics: XML
 * SaxEventSequence-based serialization, and basic String
 * serialization with ad hoc parsing.
 *
 * The replace identifier functionality supports
 *  the replacement of obsolete notification in session queues. All notifications
 *  with matching replace identifier (with an incoming notification) will be
 *  removed.
 */
public class Notification implements Comparable, IEventSequenceSerializable,
        IStringSerializable {
    private Notifiable source;
    String _val = null;
    int _priority = 0;
    int _order = 0;
    String _type;
    long _deadline;
    String iface = "0";
    String _address; // origin interface / address
    String replaceId = null; // replace ID
    String ACInterface = null; // internal access server identifier
    boolean replaceActive = false; // replace active
    String _subIdentifier; // subscription identifiers for this notification
    NotificationValue nv = new NotificationValue(); // data triplets
    boolean hidesource = false;
    boolean scope = false;
    boolean nolocal = false; /* no local delivery */
    private Vector subhash;

    /**
     * Get source interface (only locally available)
     * @return Notifiable
     */
    public Notifiable getSource() {
        return source;
    }

    /**
     * Set source interface (only locally)
     * @param eb Notifiable
     */
    public void setSource(Notifiable eb) {
        source = eb;
    }

    /**
     * Set AC interface (for internal operations)
     * @param s String
     */
    public void setACInterface(String s) {
        ACInterface = s;
    }

    /**
     * Get AC interface (internal operation)
     * @return String
     */
    public String getACInterface() {
        return ACInterface;
    }

    /** Exclude local delivery.
     * @param b set the local delivery option (true = no local, false = allow local)
     */
    public void setNoLocal(boolean b) {
        nolocal = b;
        scope = !b;
    }

    /** Get replace identifier. The replace identifier functionality supports
     *  the replacement of obsolete notification in session queues. All notifications
     *  with matching replace identifier (with an incoming notification) will be
     *  removed.
     * @return String the replace identifier
     */
    public String getReplaceId() {
        return replaceId;
    }

    /** Set replace identifier. The replace identifier functionality supports
     *  the replacement of obsolete notification in session queues. All notifications
     *  with matching replace identifier (with an incoming notification) will be
     *  removed.
     * @param s the identifier
     */
    public void setReplaceId(String s) {
        replaceId = s;
    }

    /** Set the replace mode (previous notifications matching with the
     * replaceId are replaced by this notification.
     * @param b the replace active value
     */
    public void setReplaceActive(boolean b) {
        replaceActive = b;
    }

    /** Returns true if replace mode is active.
     *  @return true if replace active
     */
    public boolean getReplaceActive() {
        return replaceActive;
    }

    /** Exclude local delivery (true/false). This is different from scope,
     * which determines whether external delivery is used.
     * @return true if local delivery is not active
     */
    public boolean getNoLocal() {
        return nolocal;
    }

    /** Set the subscription IDs (separated by commas). This is set by the
     * the server-side system.
     * @param s the subscription identifier
     */
    public void setSubId(String s) {
        _subIdentifier = s;
        subhash = null;
    }

    /** Get the subscription IDs (separated by commas)
     * @return String the subscription identifiers
     */
    public String getSubIdString() {
        return _subIdentifier;
    }

    /** Get the subscription ID (separated by commas)
     * @return Vector the subscription identifiers as a Vector
     */
    public Vector getSubId() {
        if (_subIdentifier == null) {
            return null;
        }

        if (subhash == null) {
            subhash = new Vector();
            StringTokenizer tok = new StringTokenizer(_subIdentifier, ",");

            while (tok.hasMoreTokens()) {
                String name = tok.nextToken();
                subhash.add(name);
            }
        }
        return subhash;
    }

    /** Set notification scope to local (not forwarded to external subscribers */
    public void setScopeToLocal() {
        scope = true;
    }

    /** Set notification scope to external (forwarded to external subscribers */
    public void setScopeToExternal() {
        scope = false;
    }

    /* Returns notification scope local (true) / external (false)
     * @return true if the scope is local
     */
    public boolean getScope() {
        return scope;
    }

    /** Internal use.
     * @param s the parameter
     */
    public void setInterface(String s) {
        iface = s;
    }

    /** Internal use.
     * @return String the value
     */
    public String getInterface() {
        return iface;
    }

    /** Internal use.
     * @return boolean
     */
    public boolean getHideSource() {
        return hidesource;
    }

    /** Internal use.
     * @param a the parameter
     */
    public void hideSource(boolean a) {
        hidesource = a;
    }

    /** Returns the source address of the message/notification
     * @return String the source address
     */
    public String getAddress() {
        return _address;
    }

    /** Sets the source address of the message/notification
     * @param a the source address to set
     */
    public void setAddress(String a) {
        _address = a;
    }

    /** Returns the string representation of this object.
     * @return String the string representation of the object
     */
    public String toString() {
        String ret = "";
        if (hidesource) {
            ret = "<event-service><type>" + _type + "</type><order>" + _order +
                  "</order>" + "<deadline>" + _deadline + "</deadline>" +
                  "<priority>" + _priority + "</priority><value>" + nv.toString() +
                  "</value>" +
                  "<ra>" + replaceActive + "</ra>" +
                  "<rid>" + replaceId + "</rid>" +
                  "<id>" + _subIdentifier + "</id>" +
                  "<ac>" + ACInterface + "</ac>" +
                  "</event-service>";
        } else {
            ret = "<event-service><type>" + _type + "</type><order>" + _order +
                  "</order>" + "<source>" + _address + "</source><deadline>" +
                  _deadline + "</deadline>" +
                  "<priority>" + _priority + "</priority><value>" + nv.toString() +
                  "</value>" +
                  "<ra>" + replaceActive + "</ra>" +
                  "<rid>" + replaceId + "</rid>" +
                  "<id>" + _subIdentifier + "</id>" +
                  "<ac>" + ACInterface + "</ac>" +
                  "</event-service>";

        }
        return ret;
    }

    /* *************************************************************************/
    /* *************************************************************************/
    /** XAS Serialization */

    /**
     * Insert the notification to an event sequence.
     * @param ser TypedXmlSerializer
     * @throws IOException
     */
    public void toEventSequence(TypedXmlSerializer ser) throws IOException {
        XmlWriter xw0 = new XmlWriter(ser);

        EventSerializer evs = new EventSerializer();

        XmlWriter xw = new XmlWriter(evs);

        nv.toEventSequence(xw);

        xw.typedElement("", "type", XasUtil.XSD_NAMESPACE, "string", _type);

        if (!hidesource) {
            xw.typedElement("", "source", XasUtil.XSD_NAMESPACE, "string",
                            _address);
        } else {
            xw.typedElement("", "source", XasUtil.XSD_NAMESPACE, "string", "");
        }

        xw.typedElement("", "order", XasUtil.XSD_NAMESPACE, "int",
                        new Integer(_order));
        xw.typedElement("", "deadline", XasUtil.XSD_NAMESPACE, "long",
                        new Long(_deadline));
        xw.typedElement("", "priority", XasUtil.XSD_NAMESPACE, "int",
                        new Integer(_priority));

        xw.typedElement("", "ra", XasUtil.XSD_NAMESPACE, "string",
                        String.valueOf(replaceActive));
        xw.typedElement("", "rid", XasUtil.XSD_NAMESPACE, "string", replaceId);
        xw.typedElement("", "id", XasUtil.XSD_NAMESPACE, "string",
                        _subIdentifier);
        xw.typedElement("", "ac", XasUtil.XSD_NAMESPACE, "string", ACInterface);

        xw0.complexElement(Filter.EVENT_NAMESPACE, "event",
                           evs.getCurrentSequence());
    }

    /**
     *  Parse a notification from an event sequence.
     * @param ses EventSequence
     * @throws IOException
     */
    public Notification(EventSequence ses) throws IOException {
        fromEventSequence(ses);
    }

    /**
     * Parse object from the given event sequence
     * @param ses EventSequence
     * @return boolean
     */
    public boolean fromEventSequence(EventSequence ses) {
        XmlReader xr0 = new XmlReader(ses);
        XmlReader xr = null;

        EventSequence esf = xr0.completeContent(Filter.EVENT_NAMESPACE, "event");
        if (esf == null) {
            return false;
        }

        xr = new XmlReader(esf);

        nv.parse(xr);

        _type = (String) xr.typedContent("", "type", XasUtil.XSD_NAMESPACE,
                                         "string");
        _address = (String) xr.typedContent("", "source", XasUtil.XSD_NAMESPACE,
                                            "string");
        Integer i = (Integer) xr.typedContent("", "order",
                                              XasUtil.XSD_NAMESPACE,
                                              "int");
        if (i != null) {
            try {
                _order = i.intValue();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        Long l = (Long) xr.typedContent("", "deadline", XasUtil.XSD_NAMESPACE,
                                        "long");
        if (l != null) {
            try {
                _deadline = l.longValue();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        i = (Integer) xr.typedContent("", "priority", XasUtil.XSD_NAMESPACE,
                                      "int");
        if (i != null) {
            try {
                _priority = i.intValue();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        replaceActive = false;
        try {
            String ra = (String) xr.typedContent("", "ra",
                                                 XasUtil.XSD_NAMESPACE,
                                                 "string");
            if (ra != null && ra.indexOf("true") != -1) {
                replaceActive = true;
            }
        } catch (Exception e) {
        }

        replaceId = (String) xr.typedContent("", "rid", XasUtil.XSD_NAMESPACE,
                                             "string");
        _subIdentifier = (String) xr.typedContent("", "id",
                                                  XasUtil.XSD_NAMESPACE,
                                                  "string");
        ACInterface = (String) xr.typedContent("", "ac", XasUtil.XSD_NAMESPACE,
                                               "string");

        return true;
    }

    /** XAS Serialization */
    /* *************************************************************************/
    /* *************************************************************************/

    public Notification() {}

    public Notification(Notification n) {
        _type = n._type;
        _val = n._val;
        _priority = n._priority;
        _subIdentifier = n._subIdentifier;
        _order = n.getOrder();
        ACInterface = n.ACInterface;
        _deadline = n._deadline;
        iface = n.iface;
        nv = new NotificationValue(n.nv); // data triplets
        hidesource = n.hidesource;
        scope = n.scope;
    }

    public Notification(String str) {
        parse(str);
    }

    /**
     * Parse object from the given string
     * @param str String
     * @return boolean
     */
    public boolean parse(String str) {
        // very simple and quick way of extracting elements
        try {
            _type = Auxil.getTag(str, "type");
        } catch (Exception e) {
            e.printStackTrace(); // mandatory
            System.out.println("String was: " + str);
            return false;
        }
        try {
            _order = (new Integer(Auxil.getTag(str, "order"))).intValue();
        } catch (Exception e) {
            System.out.println("String was: " + str);
        }
        try {
            _deadline = (new Long(Auxil.getTag(str, "deadline"))).longValue();
        } catch (Exception e) {
            System.out.println("String was: " + str);
        }
        try {
            _priority = (new Integer(Auxil.getTag(str, "priority"))).intValue();
        } catch (Exception e) {
            System.out.println("String was: " + str);
        }
        try {
            _val = Auxil.getTag(str, "value");
        } catch (Exception e) {
            System.out.println("String was: " + str);
        }
        try {
            _address = Auxil.getTag(str, "source");
        } catch (Exception e) {
            System.out.println("String was: " + str);
        }
        try {
            _subIdentifier = Auxil.getTag(str, "id");
            if (_subIdentifier != null && _subIdentifier.trim().equals("null")) {
                _subIdentifier = null;
            }
        } catch (Exception e) {
            System.out.println("String was: " + str);
        }

        replaceActive = false;
        try {
            String ra = Auxil.getTag(str, "ra");
            if (ra != null && ra.indexOf("true") != -1) {
                replaceActive = true;
            }
        } catch (Exception e) {
        }

        try {
            replaceId = Auxil.getTag(str, "rid");
            if (replaceId != null && replaceId.trim().equals("null")) {
                replaceId = null;
            }
        } catch (Exception e) {
            e.printStackTrace(); // mandatory
        }

        try {
            ACInterface = Auxil.getTag(str, "ac");
            if (ACInterface != null && ACInterface.trim().equals("null")) {
                ACInterface = null;
            }
        } catch (Exception e) {
            e.printStackTrace(); // mandatory
        }

        // try to parse the notification value
        if (_val != null && _val.length() > 0) {
            nv.parse(_val);
        }
        return true;
    }

    /** Add an event triplet to this notification. The order of triplets is FIFO.
     * @param et the EventTriplet to add
     */
    public void add(EventTriplet et) {
        nv.insert(et);
        if (et.getName() != null && et.getName().equals("type") &&
            et.getType() == et.TYPE_STRING) {
            _type = (String) et.getValue();
        }
    }

    /** Add an event triplet to this notification. The order of triplets is FIFO.
     * @param name the name of the triplet
     * @param obj the data object
     */
    public void add(String name, Object obj) {
        nv.insert(name, obj);
    }

    /* Extract a field value from the notification based on the name
     * @param name the field to retrieve
     * @return Object the field value
     */
    public Object getField(String name) {
        return nv.getTriplet(name);
    }

    /* Extract a field value from the notification based on the name
     * @param name the field to retrieve
     * @param type the type of the field to retrieve
     * @return Object the field value
     */
    public Object getField(String name, int type) {
        return nv.getTriplet(EventTriplet.formatId(name, type));
    }

    /** Get the value of the notification
     * @return NotificationValue the notification contents
     */
    public NotificationValue get() {
        return nv;
    }

    /** Set the value of the notification (does not clone)
     * @param z the NotificationValue object to set
     */
    public void set(NotificationValue z) {
        nv = z;
    }

    /** Get the value of the notification as text
     *  @return String the value of the notification as a String*/
    public String getDataText() {
        _val = nv.toString();
        return _val;
    }

    /** Set the value of the notification using a String
     * @param val the notification to parse
     */
    public void set(String val) {
        _val = val;
        nv = null;
        nv = new NotificationValue();
        nv.parse(val);
    }

    /** Set the priority of the notification
     *  @param p the priority of the notification
     */
    public void setPriority(int p) {
        _priority = p;
    }

    /** Get the priority of the notification
     * @return int the priority of the notification
     */
    public int getPriority() {
        return _priority;
    }

    /** Set the deadline for this notification in milliseconds
     * @param millis the deadline
     */
    public void setDeadline(long millis) {
        _deadline = millis;
    }

    /** Get the deadline for this notification in milliseconds
     *  @return long the deadline value
     */
    public long getDeadline() {
        return _deadline;
    }

    /** Get notification type (event channel name)
     * @return String the type of the notification
     */
    public String getType() {
        return _type;
    }

    /** Set notification type (event channel name). Inserts a new
     *  EventTriplet for the type.
     * @param type the type of the notification to set
     */
    public void setType(String type) {
        _type = type;

        if (nv.getTriplet("type") == null) {
            nv.insert(new EventTriplet("type", EventTriplet.TYPE_STRING, type));
        }
    }


    /** Set notification type (event channel name). Inserts a new
     *  EventTriplet for the type.
     * @param type the type of the notification to set
     */
    public void updateType(String type) {
        _type = type;
    }

    /** Get notification order (updated at event channel)
     * @return int the notification order
     */
    public int getOrder() {
        return _order;
    }

    /** Set notification order
     * @param order the order of the notification
     */
    public void setOrder(int order) {
        _order = order;
    }

    /** Tests if the deadline of the notification has passed
     * @param dl the deadline to test
     * @return true if the deadline has passed
     */
    public boolean deadlinePassed(long dl) {
        if (_deadline > dl) {
            return true;
        }
        return false;
    }

    /** Clones the notification object
     * @return Object the clone*/
    public Object clone() {
        Object obj = null;

        try {
            obj = super.clone();
        } catch (Exception e) {
            fuegocore.notify.util.Log.write("Exception: " + e);
            e.printStackTrace();
        }
        return obj;
    }

    /**
     * Comparison based on order.
     * @param other the other object to compare to
     * @return int the comparison value
     */
    public int compareTo(Object other) {
        if (other instanceof Notification) {
            int a = ((Notification) other).getOrder();
            return a - getOrder();

        } else {
            throw new ClassCastException();
        }
    }

}
