/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core;

import java.io.*;
import java.util.*;

import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.serialization.*;
import fuegocore.notify.util.*;
import fuegocore.util.xas.*;

/**
 * This class represents the notification data contents:
 * name and value pairs.
 */
public class NotificationValue implements IEventSequenceSerializable,
        IStringSerializable {
    TripletTable data;

    public String SEPARATOR = ":";
    public String ROWSEPARATOR = "|";

    public NotificationValue() {
        data = new TripletTable();
    }

    /**
     * Create a copy of a notification value object.
     * Note that event triplets are shared.
     * @param nv NotificationValue
     */
    public NotificationValue(NotificationValue nv) {
        data = new TripletTable();
        Vector v = nv.getVector();
        for (int i = 0; i < v.size(); i++) {
            EventTriplet et = (EventTriplet) v.elementAt(i);
            data.put(et.getId(), et);
        }
    }

    /** Insert data object.
     * @param name the name of the object
     * @param obj the object data
     * @return true if the insert was a success
     */
    public boolean insert(String name, Object obj) {
        if (name == null || obj == null) {
            return false;
        }

        int type = getObjectType(obj);
        if (type == -1) {
            return false;
        }
        EventTriplet e = new EventTriplet(name, type, obj);

        data.put(e.getId(), e);
        return true;
    }

    /**
     * Remove data object.
     * @param name String
     * @param type int
     * @return boolean
     */
    public boolean remove(String name, int type) {
        if (name == null) {
            return false;
        }
        data.remove(name + "/" + type);
        return true;
    }

    /**
     * Remove data object.
     * @param et the EventTriplet to remove
     * @return true if success
     */
    public boolean remove(EventTriplet et) {
        if (et == null) {
            return false;
        }
        data.remove(et.getId(), et);
        return true;
    }

    /** Insert data object (EventTriplet).
     * @param et the EventTriplet to insert
     * @return true if success
     */
    public boolean insert(EventTriplet et) {
        if (et == null) {
            return false;
        }
        data.put(et.getId(), et);
        return true;
    }

    /**
     * Remove EventTriplets based on their name.
     * Note that id needs to be name/type, where type is integer
     * (see EventTriplet formatId method)
     * @param id the identifier to remove
     * @return a Vector containing EventTriplets.
     */
    public Object remove(String id) {
        return data.remove(id);
    }

    /** Returns the triplets of the notification as a TripletTable
     * @return TripletTable the triplets
     */
    public TripletTable getTriplets() {
        return data;
    }

    /** Returns a vector containing EventTriplet objects (in order)
     * @return Vector the data in Vector format
     */
    public Vector getVector() {
        return data.getAllInOrder();
    }

    /** Set the triplets of the notification
     * @param z the new TripletTable to set (not cloned)
     */
    public void setTriplets(TripletTable z) {
        data = z;
    }

    /** Creates the object from the given XML string.
     * Note that in this version the string should not contain
     * commas, ":" or dots.
     * @param str the String to parse
     * @return true if success
     */
    public boolean parse(String str) {
        if (data == null) {
            data = new TripletTable();
        }
        try {
            int i = str.indexOf("<TRIPLET>");
            while (i > -1) {
                // seek end
                int j = str.indexOf("</TRIPLET>", i + 1);
                if (j == -1) {
                    break; // error in XML
                }
                String triplet = str.substring(i + 1, j);
                String name = Auxil.getTag(triplet, "NAME");

                int type = -1;

                String val = Auxil.getTag(triplet, "INT");

                if (val != null) {
                    type = EventTriplet.TYPE_INT;
                } else if ((val = Auxil.getTag(triplet, "BOOLEAN")) != null) {
                    type = EventTriplet.TYPE_BOOLEAN;
                } else if ((val = Auxil.getTag(triplet, "DOUBLE")) != null) {
                    type = EventTriplet.TYPE_DOUBLE;
                } else if ((val = Auxil.getTag(triplet, "STRING")) != null) {
                    type = EventTriplet.TYPE_STRING;
                } else if ((val = Auxil.getTag(triplet, "DATE")) != null) {
                    type = EventTriplet.TYPE_DATE;

                }
                if (name != null && val != null && type > -1) {
                    EventTriplet et = new EventTriplet(name, type, val);
                    data.put(et.getId(), et);
                }

                // next triplet
                i = str.indexOf("<TRIPLET>", i + 1);
            }
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }
        return true;
    }

    /** Returns the value of this object as XML.
     * @return String the XML text representation of the object
     */
    public String toString() {
        String str = "";
        Vector vec = null;

        if (data == null) {
            return "<TRIPLET></TRIPLET>";
        }

        for (Enumeration e = data.elements(); e.hasMoreElements(); ) {
            // we loop over duplicate names
            Object obj = e.nextElement();
            if (obj != null && obj instanceof Vector) {
                vec = (Vector) obj;
                for (int i = 0; i < vec.size(); i++) {
                    EventTriplet em = (EventTriplet) vec.elementAt(i);
                    if (em != null) {
                        str += "<TRIPLET><NAME>" + em.name + "</NAME>\n";
                        if (em.type == EventTriplet.TYPE_INT) {
                            str += "<INT>" + ((Integer) em.o).intValue() +
                                    "</INT>";
                        } else if (em.type == EventTriplet.TYPE_BOOLEAN) {
                            str += "<BOOLEAN>" + ((Boolean) em.o).booleanValue() +
                                    "</BOOLEAN>";
                        } else if (em.type == EventTriplet.TYPE_DOUBLE) {
                            str += "<DOUBLE>" + ((Double) em.o).doubleValue() +
                                    "</DOUBLE>";
                        } else if (em.type == EventTriplet.TYPE_STRING) {
                            str += "<STRING>" + (String) em.o + "</STRING>";
                        } else if (em.type == EventTriplet.TYPE_DATE) {
                            str += "<DATE>" + (String) ((Date) em.o).toString() +
                                    "</DATE>";
                        } else if (em.type == EventTriplet.TYPE_SERIALIZED) {
                            str += "<SERIALIZED></SERIALIZED>";
                        }
                        str += "\n</TRIPLET>\n";
                    }
                }
            } else if (obj != null && obj instanceof EventTriplet) {
                EventTriplet em = (EventTriplet) obj;
                if (em != null) {
                    str += "<TRIPLET><NAME>" + em.name + "</NAME>\n";
                    if (em.type == EventTriplet.TYPE_INT) {
                        str += "<INT>" + ((Integer) em.o).intValue() + "</INT>";
                    } else if (em.type == EventTriplet.TYPE_BOOLEAN) {
                        str += "<BOOLEAN>" + ((Boolean) em.o).booleanValue() +
                                "</BOOLEAN>";
                    } else if (em.type == EventTriplet.TYPE_DOUBLE) {
                        str += "<DOUBLE>" + ((Double) em.o).doubleValue() +
                                "</DOUBLE>";
                    } else if (em.type == EventTriplet.TYPE_STRING) {
                        str += "<STRING>" + (String) em.o + "</STRING>";
                    } else if (em.type == EventTriplet.TYPE_DATE) {
                        str += "<DATE>" + (String) ((Date) em.o).toString() +
                                "</DATE>";
                    } else if (em.type == EventTriplet.TYPE_SERIALIZED) {
                        str += "<SERIALIZED></SERIALIZED>";
                    }
                    str += "\n</TRIPLET>\n";
                }
            }
        }
        return str;
    }

    /* *************************************************************************/
    /* *************************************************************************/
    /** XAS Serialization */

    /**
     * Insert the triples to an event sequence.
     * @param ser TypedXmlSerializer
     * @throws IOException
     */
    public void toEventSequence(TypedXmlSerializer ser) throws IOException {
        XmlWriter xw = new XmlWriter(ser);
        toEventSequence(xw);
    }

    /**
     * Insert the triples to an event sequence.
     * @param xw XmlWriter
     * @throws IOException
     * @return true if success
     */
    public boolean toEventSequence(XmlWriter xw) throws IOException {
        String str = "";
        Vector vec = null;

        for (Enumeration e = data.elements(); e.hasMoreElements(); ) {

            // we loop over duplicate names
            Object obj = e.nextElement();

            if (obj != null && obj instanceof Vector) {
                vec = (Vector) obj;

                for (int i = 0; i < vec.size(); i++) {
                    EventTriplet em = (EventTriplet) vec.elementAt(i);

                    if (em != null) {
                        EventSerializer evs = new EventSerializer();
                        XmlWriter xw2 = new XmlWriter(evs);

                        xw2.typedElement("", "name", XasUtil.XSD_NAMESPACE,
                                         "string",
                                         em.name);
                        if (em.type == EventTriplet.TYPE_INT) {
                            xw2.typedElement("", "int", XasUtil.XSD_NAMESPACE,
                                             "string",
                                             "" + ((Integer) em.o).intValue());
                        } else if (em.type == EventTriplet.TYPE_BOOLEAN) {
                            xw2.typedElement("", "boolean",
                                             XasUtil.XSD_NAMESPACE, "string",
                                             "" + ((Boolean) em.o).booleanValue());
                        } else if (em.type == EventTriplet.TYPE_DOUBLE) {
                            xw2.typedElement("", "double",
                                             XasUtil.XSD_NAMESPACE, "string",
                                             "" + ((Double) em.o).doubleValue());
                        } else if (em.type == EventTriplet.TYPE_STRING) {
                            xw2.typedElement("", "string",
                                             XasUtil.XSD_NAMESPACE, "string",
                                             (String) em.o);
                        } else if (em.type == EventTriplet.TYPE_DATE) {
                            xw2.typedElement("", "date", XasUtil.XSD_NAMESPACE,
                                             "string",
                                             "" +
                                             (String) ((Date) em.o).toString());
                        }
                        xw.complexElement(Filter.EVENT_NAMESPACE, "triplet",
                                          evs.getCurrentSequence());
                    }
                }
            } else if (obj != null && obj instanceof EventTriplet) {
                EventTriplet em = (EventTriplet) obj;
                if (em != null) {
                    EventSerializer evs = new EventSerializer();
                    XmlWriter xw2 = new XmlWriter(evs);
                    xw2.typedElement("", "name", XasUtil.XSD_NAMESPACE,
                                     "string", em.name);
                    if (em.type == EventTriplet.TYPE_INT) {
                        xw2.typedElement("", "int", XasUtil.XSD_NAMESPACE,
                                         "string",
                                         "" + ((Integer) em.o).intValue());
                    } else if (em.type == EventTriplet.TYPE_BOOLEAN) {
                        xw2.typedElement("", "boolean", XasUtil.XSD_NAMESPACE,
                                         "string",
                                         "" + ((Boolean) em.o).booleanValue());
                    } else if (em.type == EventTriplet.TYPE_DOUBLE) {
                        xw2.typedElement("", "double", XasUtil.XSD_NAMESPACE,
                                         "string",
                                         "" + ((Double) em.o).doubleValue());
                    } else if (em.type == EventTriplet.TYPE_STRING) {
                        xw2.typedElement("", "string", XasUtil.XSD_NAMESPACE,
                                         "string",
                                         (String) em.o);
                    } else if (em.type == EventTriplet.TYPE_DATE) {
                        xw2.typedElement("", "date", XasUtil.XSD_NAMESPACE,
                                         "string",
                                         "" + (String) ((Date) em.o).toString());
                    }
                    xw.complexElement(Filter.EVENT_NAMESPACE, "triplet",
                                      evs.getCurrentSequence());
                }
            }
        }
        return true;
    }

    /**
     * Parse object from the given event sequence
     * @param es EventSequence
     * @return boolean
     */
    public boolean fromEventSequence(EventSequence es) {
        XmlReader xr = new XmlReader(es);
        parse(xr);
        return true;
    }

    /**
     * Parse the object from an EventSequence.
     * @param xr the XmlReader
     */
    public void parse(XmlReader xr) {
        if (data == null) {
            data = new TripletTable();
        }

        boolean quit = false;

        // attribute filters follow
        while (!quit && !xr.isFinished()) {
            EventSequence esf = xr.completeContent(Filter.EVENT_NAMESPACE,
                    "triplet");

            if (esf == null) {
                quit = true;
            } else {
                XmlReader xr2 = new XmlReader(esf);
                String name = (String) xr2.typedContent("", "name",
                        XasUtil.XSD_NAMESPACE, "string");

                int type = -1;

                String val = (String) xr2.typedContent("", "int",
                        XasUtil.XSD_NAMESPACE,
                        "string");
                if (val != null) {
                    type = EventTriplet.TYPE_INT;
                } else if ((val = (String) xr2.typedContent("", "boolean",
                        XasUtil.XSD_NAMESPACE, "string")) != null) {
                    type = EventTriplet.TYPE_BOOLEAN;
                } else if ((val = (String) xr2.typedContent("", "int",
                        XasUtil.XSD_NAMESPACE, "double")) != null) {
                    type = EventTriplet.TYPE_DOUBLE;
                } else if ((val = (String) xr2.typedContent("", "string",
                        XasUtil.XSD_NAMESPACE, "string")) != null) {
                    type = EventTriplet.TYPE_STRING;
                } else if ((val = (String) xr2.typedContent("", "date",
                        XasUtil.XSD_NAMESPACE, "string")) != null) {
                    type = EventTriplet.TYPE_DATE;
                }
                if (name != null && val != null && type > -1) {
                    EventTriplet et = new EventTriplet(name, type, val);
                    data.put(et.getId(), et);
                }
            }
        }
    }

    /** XAS Serialization */

    /* *************************************************************************/
    /* *************************************************************************/


    /** Returns the first object corresponding to the given field name.
     * Note that id needs to be name/type, where type is integer
     * (see EventTriplet formatId method). Does not support duplicate names.
     * @param id the object to get (identifier)
     * @return Object the value
     */
    public Object getObject(String id) {
        if (data == null) {
            return null;
        }
        EventTriplet em = (EventTriplet) data.getFirst(id);
        if (em == null) {
            return null;
        }
        return em.o;
    }

    /** Returns the object corresponding to the given field name.
     *  Note that id needs to be name/type, where type is integer
     * (see EventTriplet formatId method). Does not support duplicate names.
     * @param id the identifier
     * @return EventTriplet the triplet
     */
    public EventTriplet getTriplet(String id) {
        if (data == null) {
            return null;
        }
        EventTriplet em = (EventTriplet) data.getFirst(id);
        if (em == null) {
            return null;
        }
        return em;
    }

    /** Returns the object corresponding to the given field name.
     * Note that id needs to be name/type, where type is integer
     * (see EventTriplet formatId method).
     * Supports duplicate objects.
     * @param id the identifier
     * @return Vector the triplets
     */
    public Vector getTripletAll(String id) {
        if (data == null) {
            return null;
        }
        Vector em = (Vector) data.get(id);
        if (em == null) {
            return null;
        }
        return em;
    }

    /** Returns the type of object.
     * @param obj the object
     * @return int the type (EventTriplet type)
     */
    public int getObjectType(Object obj) {
        int type = -1;
        if (obj instanceof Integer) {
            type = EventTriplet.TYPE_INT;
        } else if (obj instanceof Boolean) {
            type = EventTriplet.TYPE_BOOLEAN;
        } else if (obj instanceof Double) {
            type = EventTriplet.TYPE_DOUBLE;
        } else if (obj instanceof String) {
            type = EventTriplet.TYPE_STRING;
        } else if (obj instanceof Date) {
            type = EventTriplet.TYPE_DATE;
        }
        return type;
    }

}
