/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core;

import java.io.*;
import java.util.*;

import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.serialization.*;
import fuegocore.notify.util.*;
import fuegocore.util.xas.*;

/**
 * Template encapsulates a subscription or an advertisement.
 * Default is a subscription, use setAdvertisementFlag method to change.
 * The class defines what events are routed and to which recipients.
 * Each template and subscription has a single type attribute,
 * which uniquely determines the event channel. Templates have
 * associated sessions and subscription identifiers.  Subscription
 * identifiers are used for server-side filtering and efficient
 * client-side dispatching.  The template consists of n attribute
 * filters, which have the format of <name, type, constraint>.
 */
public class Template implements Cloneable, Comparable, ITestComponent,
        ITemplate, IFBase,
        IEventSequenceSerializable, IStringSerializable {
    private Notifiable source;
    private String iface; // subscribing interface
    public static int code = 0;
    private String _type;
    private int depth = 0; // ICovers interface, poset depth
    private String sessionId;
    private boolean sent_ec; // internal send flag (sent to event channel)
    private String fs_id; // internal matching use
    private boolean scope = false; // external
    private boolean _localsend = true; // forward events originated on this host back to this host if they match a filter
    private double overhead = 1; // used in load balancing
    private String coverId = null;
    private boolean advertisement = false; // set to true if template represents an adv
    private boolean quench = false; // quench subscription

    // localsend=true means that events are sent back
    private String _notifyfields = "*";
    private String _subIdentifier = null; // subscription identifier
    private String _expiration = "";
    private int storeIndex = 0; // index for filter store internal use
    private Vector ifaces; // internal matching use
    Hashtable ncache;
    Vector tid; // template reference for matching
    Vector compressedTid;
    double tempOverhead = 1;
    int hitCount = 0;
    Vector filters = new Vector();


    /** Comparator for attribute filters */
    Comparator filterComparator = new Comparator() {
        public int compare(Object o1, Object o2) {
            if (o1 instanceof Filter && o2 instanceof Filter) {
                String a = ((Filter) o1).getCoverId();
                String b = ((Filter) o2).getCoverId();
                return a.compareTo(b);
            } else {
                throw new ClassCastException();
            }
        }
    };

    /**
     * Internal method. Used in matching (does not support concurrent matching).
     * @return int
     */
    public int getHitCount() {
        return 0;
    }

    /**
     * Internal method. Used in matching (does not support concurrent matching).
     * @param hit int
     */
    public void setHitCount(int hit) {
    }

    /**
     * Flag the template as an advertisement
     * @param b boolean
     */
    public void setAdvertisement(boolean b) {
        advertisement = b;
    }

    /**
     * Returns the advertisement flag
     * @return boolean
     */
    public boolean getAdvertisement() {
        return advertisement;
    }

    /**
     * Set temp overhead
     * @param t double
     */
    public void setTempOverhead(double t) {
        tempOverhead = t;
    }

    /**
     * Get temp overhead
     * @return double
     */
    public double getTempOverhead() {
        return tempOverhead;
    }

    /**
     * Used to generate unique id numbers for templates if one is not supplied
     * @return int
     */
    public static int getUniqueCode() {
        return code++;
    }

    /** Returns the cost of the filter (generality).
     *  Higher value denotes more general filters.
     * @return int the cost
     */
    public int getCost() {
        int cost = 0;

        if (filters == null || filters.size() == 0) {
            return 0;
        }
        int c = 1000000;

        for (int i = 0; i < filters.size(); i++) {
            Filter f = (Filter) filters.elementAt(i);
            cost += c + f.getCost();
        }

        cost = cost / filters.size();

        return cost;
    }

    /**
     * Add an interface identifier (used by posets)
     * @param id Object
     */
    public synchronized void addTid(Object id) {
        if (tid == null) {
            tid = new Vector();
            compressedTid = new Vector();
        }
        tid.add(id);
        if (!compressedTid.contains(id)) {
            compressedTid.add(id);
        }
    }

    /**
     * Returns the interface vector (multiset)
     * @return Vector
     */
    public Vector getTid() {
        return tid;
    }

    /**
     * Gets an interface vector without duplicates (a set)
     * @return Vector
     */
    public Vector getCompressedTid() {
        return compressedTid;
    }

    /**
     * Set the depth of this object in the poset tree.
     * @param i the depth to set
     */
    public void setDepth(int i) {
        depth = i;
    }

    public boolean noSelectivity() {
        return false;
    }

    /**
     * Set quench value
     * @param b boolean
     */
    public void setQuench(boolean b) {
        quench = b;
    }

    /**
     * Get quench value
     * @return boolean
     */
    public boolean getQuench() {
        return quench;
    }

    /**
     * Comparison based on depth.
     * @param other the other object to compare to
     * @return int the comparison value
     */
    public int compareTo(Object other) {
        if (other instanceof Template) {
            int a = ((Template) other).getDepth();
            return a - getDepth();

        } else {
            throw new ClassCastException();
        }
    }

    /**
     * Returns the depth of this object in the poset tree.
     * @return the depth as int
     */
    public int getDepth() {
        return depth;
    }

    /**
     * Returns the number of attribute filters
     * @return int
     */
    public int getSize() {
        if (filters == null) {
            return 0;
        }
        return filters.size();
    }

    /** Add a matching interface identifier
     *  @param s interface
     */
    public void addInterface(String s) {
        if (ifaces == null) {
            ifaces = new Vector();
        }
        ifaces.add(s);
    }

    /** Set the subscription ID
     * @param s the id
     */
    public void setSubId(String s) {
        _subIdentifier = s;
    }

    /** Get the subscription ID
     * @return the subscription identifier
     * */
    public String getSubId() {
        return _subIdentifier;
    }

    /** Set the session ID
     * @param s the session ID to set
     */
    public void setSessionId(String s) {
        sessionId = s;
    }

    /** Get the session ID
     * @return the session identifier
     */
    public String getSessionId() {
        return sessionId;
    }

    /** Set the forward ID (internal)
     * @param s the forward id
     */
    public void setFSId(String s) {
        fs_id = s;
    }

    /** Get the forward ID (internal)
     * @return the forward id
     */
    public String getFSId() {
        return fs_id;
    }

    /**
     * Set event channel flag (internal use)
     * @param b boolean
     */
    public void setECFlag(boolean b) {
        sent_ec = b;
    }

    /**
     * Get event channel flag (internal use)
     * @return boolean
     */
    public boolean getECFlag() {
        return sent_ec;
    }

    /**
     * Filter store internal use. This is used to support matching using
     * an array instead of using a Hashtable with Integer objects
     * @param i int
     */
    public void setStoreIndex(int i) {
        storeIndex = i;
    }

    /**
     *  Filter store internal use. This is used to support matching using
     *  an array instead of using a Hashtable with Integer objects
     * @return int
     */
    public int getStoreIndex() {
        return storeIndex;
    }

    /** Return the matching interface identifiers as a vector
     * @return Vector list of interfaces */
    public Vector getInterfaces() {
        return ifaces;
    }

    /** Clear the interface vector */
    public void clearInterfaces() {
        if (ifaces != null) {
            ifaces.clear();
        }
    }

    /** Formats a notification according to the required notification fields
     * specified in this template
     * @param n the notification
     * @return the formatted notification
     */
    public Notification formatNotification(Notification n) {
        if (n == null || _notifyfields == null || _notifyfields == "" ||
            _notifyfields.equals("*")) {
            return n; // no requirements
        }

        Notification n2 = new Notification(n); // clone the notification

        if (ncache == null) {
            ncache = new Hashtable();
            StringTokenizer tok = new StringTokenizer(_notifyfields, ",");

            while (tok.hasMoreTokens()) {
                String name = tok.nextToken();
                ncache.put(name, "true");
            }
        }
        // match fields of the notification and drop those that are not specified
        NotificationValue nv2 = n2.get();
        TripletTable td = (TripletTable) nv2.getTriplets();
        if (td != null) {
            // we remove extra fields
            for (Enumeration e = td.elements(); e.hasMoreElements(); ) {
                EventTriplet et = (EventTriplet) e.nextElement();
                if (et != null && !et.getName().equals("type") &&
                    ncache.get(et.getName()) == null) { // field not found
                    td.remove(et.getId(), et);
                    break;
                }
            }
        }
        return n2;
    }

    /** Returns the notify field names (separated by commas) as a String.
     *  The notify fields determine which fields are forwarded in the
     *  notification and which fields are omitted. This is an optional feature.
     *  @return the notify field names
     */
    public String getNotifyFields() {
        return _notifyfields;
    }

    /** Sets the notify field names (separated by commas) as a String.
     *  @param s the notify fields
     */
    public void setNotifyFields(String s) {
        _notifyfields = s;
    }

    /** Sets the local send property. If this is true notifications from the client are forwarded back to the client
     *  if they match one of the client's active subscriptions.
     * @param b the local send option
     */
    public void setLocalSend(boolean b) {
        _localsend = b;
    }

    /** Returns the local send property.  If this is true notifications from the client are forwarded back to the client
     *  if they match one of the client's active subscriptions.
     * @return true if local send is active
     */
    public boolean getLocalSend() {
        return _localsend;
    }

    /** Sets the scope of the subscription (template) to local */
    public void setScopeToLocal() {
        scope = true;
    }

    /** Sets the scope of the subscription (template) to external */
    public void setScopeToExternal() {
        scope = false;
    }

    /** Returns the current scope local (true) / external (false)
     * @return true if scope is local */
    public boolean getScope() {
        return scope;
    }

    /** Adds the given filter to the template.
     * @param o the filter to add
     */
    public void addFilter(Object o) {
        if (!(o instanceof Filter)) {
            return;
        }
        Filter f = (Filter) o;
        if (f == null) {
            return;
        }
        filters.add(f);
        // set type of the template
        if (f.getName() != null && f.getName().equals("type")) {
            _type = f.getFilterType();
        }
        // sort
        Collections.sort(filters);
    }

    /** Removes the given filter.
     * @param o the filter to remove
     */
    public void removeFilter(Object o) {
        if (!(o instanceof Filter)) {
            return;
        }
        Filter f = (Filter) o;
        filters.remove(f);
    }

    /** Returns the set of filters.
     * @return returns a Vector containing the filter objects
     */
    public Vector getFilters() {
        return filters;
    }

    /** Set the attribute filters
     *  @param v the Vector
     */
    public void setFilters(Vector v) {
        filters = v;
    }

    /**
     * Parse object from string
     * @param s String
     * @return boolean
     */
    public boolean parse(String s) {
        return parseFilters(s);
    }

    /**
     *  Parse filters from string.
     * @param str String
     * @return boolean
     */
    public boolean parseFilters(String str) {
        int i = 0;
        int j = str.indexOf("<dlm/>");
        if (j == -1) {
            return false;
        } while (j != -1) {
            String s = str.substring(i, j);
            Filter f = null;
            try {
                f = new Filter(s);
            } catch (Exception e) {
                System.out.println("Unable to parse filter: " + s);
                f = null;
            }
            if (f != null) {
                filters.add(f);
                // set type of the template
                if (f.getName() != null && f.getName().equals("type")) {
                    _type = f.getFilterType();
                }
            }
            i = j + "<dlm/>".length();
            j = str.indexOf("<dlm/>", i);
        }
        try {
            _notifyfields = Auxil.getTag(str, "nfields");
        } catch (Exception e) {
            e.printStackTrace();
        }
        try {
            _subIdentifier = Auxil.getTag(str, "id");
        } catch (Exception e) {
            _subIdentifier = null;
            e.printStackTrace();
        }
        try {
            iface = Auxil.getTag(str, "interface");
        } catch (Exception e) {
            e.printStackTrace();
        }

        try {
            _localsend = (boolean) (new Boolean(Auxil.getTag(str, "localsend"))).
                         booleanValue();
        } catch (Exception e) {
            e.printStackTrace();
        }

        try {
            advertisement = (boolean) (new Boolean(Auxil.getTag(str, "adv"))).
                            booleanValue();
        } catch (Exception e) {
            e.printStackTrace();
        }

        try {
            quench = (boolean) (new Boolean(Auxil.getTag(str, "quench"))).
                     booleanValue();
        } catch (Exception e) {
            e.printStackTrace();
        }

        try {
            _expiration = Auxil.getTag(str, "expiration");
        } catch (Exception e) {
            e.printStackTrace();
        }
        try {
            sessionId = Auxil.getTag(str, "sessionid");
        } catch (Exception e) {
            e.printStackTrace();
        }
        return true;
    }

    /* *************************************************************************/
    /* *************************************************************************/
    /** XAS Serialization */

    /** Default constructor
     * @param es the EventSequence representation for parsing the notification
     */
    public Template(EventSequence es) {
        parseFilters(es);
    }

    public boolean parseFilters(EventSequence es) {
        return fromEventSequence(es);
    }

    /**
     * Parse object from the given event sequence
     * @param es EventSequence
     * @return boolean
     */
    public boolean fromEventSequence(EventSequence es) {
        XmlReader xr = new XmlReader(es);

        boolean quit = false;

        // check the root tag for a filter
        EventSequence esf = xr.completeContent(Filter.EVENT_NAMESPACE, "filter");

        // Check that the element is correct
        if (esf == null) {
            System.out.println("Template: invalid root element");
            return false;
        }

        xr = new XmlReader(esf);

        while (!quit && !xr.isFinished()) { // attribute filters follow
            EventSequence xes = xr.completeContent(Filter.EVENT_NAMESPACE, "af");
            if (xes == null) {
                quit = true;
            } else {
                Filter f = new Filter(xes);
                if (f != null) {
                    filters.add(f);
                    if (f.getName() != null && f.getName().equals("type")) {
                        _type = f.getFilterType();
                    }
                }
            }
        }

        // now parse the rest of the filter
        _notifyfields = (String) xr.typedContent("", "nfields",
                                                 XasUtil.XSD_NAMESPACE,
                                                 "string");

        String tmp = (String) xr.typedContent("", "localsend",
                                              XasUtil.XSD_NAMESPACE, "string");
        try {
            _localsend = (boolean) (Boolean.valueOf(tmp).booleanValue());
        } catch (Exception e) {
            e.printStackTrace();
        }

        tmp = (String) xr.typedContent("", "adv",
                                       XasUtil.XSD_NAMESPACE, "string");
        try {
            advertisement = (boolean) (Boolean.valueOf(tmp).booleanValue());
        } catch (Exception e) {
            e.printStackTrace();
        }

        tmp = (String) xr.typedContent("", "quench",
                                       XasUtil.XSD_NAMESPACE, "string");
        try {
            quench = (boolean) (Boolean.valueOf(tmp).booleanValue());
        } catch (Exception e) {
            e.printStackTrace();
        }

        _expiration = (String) xr.typedContent("", "expiration",
                                               XasUtil.XSD_NAMESPACE, "string");

        _subIdentifier = (String) xr.typedContent("", "id",
                                                  XasUtil.XSD_NAMESPACE,
                                                  "string");
        iface = (String) xr.typedContent("", "interface", XasUtil.XSD_NAMESPACE,
                                         "string");
        sessionId = (String) xr.typedContent("", "sessionid",
                                             XasUtil.XSD_NAMESPACE,
                                             "string");

        return true;
    }

    /**
     * Insert the filters to a SAX event sequence.
     * @param ser TypedXmlSerializer
     * @throws IOException
     */
    public void toEventSequence(TypedXmlSerializer ser) throws IOException {
        XmlWriter xw = new XmlWriter(ser);

        EventSerializer evs = new EventSerializer();
        XmlWriter xw2 = new XmlWriter(evs);

        // Serialize filters
        for (int i = 0; i < filters.size(); i++) {
            ((Filter) filters.elementAt(i)).xasSerialize(evs);
        }

        xw2.typedElement("", "nfields", XasUtil.XSD_NAMESPACE, "string",
                         _notifyfields);
        xw2.typedElement("", "localsend", XasUtil.XSD_NAMESPACE, "string",
                         String.valueOf(_localsend));
        xw2.typedElement("", "adv", XasUtil.XSD_NAMESPACE, "string",
                         String.valueOf(advertisement));
        xw2.typedElement("", "quench", XasUtil.XSD_NAMESPACE, "string",
                         String.valueOf(quench));

        xw2.typedElement("", "expiration", XasUtil.XSD_NAMESPACE, "string",
                         _expiration);
        xw2.typedElement("", "id", XasUtil.XSD_NAMESPACE, "string",
                         _subIdentifier);
        xw2.typedElement("", "interface", XasUtil.XSD_NAMESPACE, "string",
                         iface);
        xw2.typedElement("", "sessionid", XasUtil.XSD_NAMESPACE, "string",
                         sessionId);

        xw.complexElement(Filter.EVENT_NAMESPACE, "filter",
                          evs.getCurrentSequence());
    }

    /** XAS Serialization */

    /* *************************************************************************/
    /* *************************************************************************/

    /** Create an XML string representation from filters.
     * @return the filters in String format
     */
    public String filtersToString() {
        String str = "";
        for (int i = 0; i < filters.size(); i++) {
            str += ((Filter) filters.elementAt(i)).toString();
            if (i < filters.size() - 1) {
                str += "<dlm/>";
            }
        }
        str += "<dlm/>\n";
        str += "<nfields>" + _notifyfields + "</nfields>";
        str += "<localsend>" + _localsend + "</localsend>";
        str += "<adv>" + advertisement + "</adv>";
        str += "<quench>" + quench + "</quench>";
        str += "<expiration>" + _expiration + "</expiration>";
        str += "<id>" + _subIdentifier + "</id>";
        str += "<interface>" + iface + "</interface>";
        str += "<sessionid>" + sessionId + "</sessionid>";

        return str;
    }

    /** Clones the template
     * @throws CloneNotSupportedException
     * @return Object
     */
    public Object clone() throws CloneNotSupportedException {
        Template t = new Template(this.toString());
        t._localsend = _localsend;
        if (source != null) {
            t.source = source;
        }
        if (_subIdentifier != null) {
            t._subIdentifier = new String(_subIdentifier);
        }
        if (iface != null) {
            t.iface = new String(iface);
        }
        t.sent_ec = sent_ec;
        t.fs_id = fs_id;
        t._type = _type;
        t.advertisement = advertisement;
        t.quench = quench;
        return t;
    }

    public void setInterface(String s) {
        iface = s;
    }

    public String getInterface() {
        return iface;
    }

    /** Return an XML string representation of the subscription/template
     * @return the XML representation as a String
     */
    public String toString() {
        String r2 = filtersToString();

        if (r2 != null && r2.length() > 0) {
            return r2;
        }

        return "";
    }

    public Template() {}

    /** Default constructor
     * @param str the String representation for parsing the notification
     */
    public Template(String str) {
        parseFilters(str);
    }

    /** Add the type of the subscription to the template (event channel name)
     * @param type the type to add
     */
    public void addType(String type) {
        addFilter(new Filter("type", EventTriplet.TYPE_STRING,
                             type));
        _type = type;
    }

    /** Returns the type of the subscription (event channel name)
     * @return String the type of the notification */
    public String getType() {
        return _type;
    }

    /** Returns true if the given template has the same structure
     * as this template and is mergeable using the conjunction algorithm.
     * We count the number of mergeable
     * filters.  All filters in t2 must be mergeable with this
     * object. Selectivity may change for some fields in subsequent
     * (imperfect) merging.<br>
     * Note that template merging has direction (not symmetric)
     * Note that this can no longer be used as a condition for covering.
     * @param t2 the input template
     * @return String the mergeableId of the differing attribute filter or null or all.
     */
    public String mergeableConjunctiveStructure(Template t2) {
        Vector filt = t2.getFilters();
        // Attribute counter
        int count = 0;
        Hashtable mergeable = new Hashtable();
        Hashtable temp = new Hashtable();

        /* First test: same fields and types and selectivities */
        for (int i = 0; i < filt.size(); i++) {
            Filter f2 = (Filter) filt.elementAt(i);
            mergeable.put(f2.getIdforMerge(), f2);
        }

        for (int i = 0; i < getFilters().size(); i++) {
            Filter f2 = (Filter) getFilters().elementAt(i);
            temp.put(f2.getIdforMerge(), f2);
        }

        // Check field count
        if (temp.size() != mergeable.size()) {
            return null;
        }

        // Check total field count
        // We cannot merge if there are differing parameters!
        if (getFilters().size() != filt.size()) {
            return null;
        }

        // Check equality =>
        for (int i = 0; i < getFilters().size(); i++) {
            Filter f2 = (Filter) getFilters().elementAt(i);
            if (mergeable.get(f2.getIdforMerge()) == null) {
                return null;
            }
        }
        // Check equality <=
        for (int i = 0; i < filt.size(); i++) {
            Filter f2 = (Filter) filt.elementAt(i);
            if (temp.get(f2.getIdforMerge()) == null) {
                return null;
            }
        }

        /* We iterate over the set of filters, and for each filter
           find the corresponding filter with the same name,type,selectivity.
         */
        int count2 = 0;
        String diff = "all";
        for (int i = 0; i < getFilters().size(); i++) {
            Filter f2 = (Filter) getFilters().elementAt(i);
            // If the selectivity is any, we have already
            // ascertained that there is a filter, and by definition we know
            // that there can be only one per selectivity type
            // we have to ensure that the type is exact
            Filter f3 = (Filter) mergeable.get(f2.getIdforMerge());
            if (f3 == null) {
                System.out.println("f3 was null");
                return null; // cannot merge field with null field
            } else if (f3.getSelectivity() != f2.getSelectivity()) {
                // Same name and type but different selectivity
                // cannot perform perfect merging!
                System.out.println("Differing selectivity");
                return null;
            }
            // Check that the two filters identical
            if (f2.getFilterString().trim().equals(f3.getFilterString().trim())) {
                count2++;
            } else {
                // test the conjunctive mergeability of the differing filter
                // check conjunctive mergeability of the differing filter
                if (!f2.conjunctivePerfectMerge(f3, true)) { // test mergeability
                    //System.out.println("Not mergeable: " +
                    //                   f2.toString() + " / " +
                    //                   f3.toString());
                    // the filters differ and cannot be merged
                    // the filter sets cannot be merged at all
                    return null;
                } else {
                    diff = f3.getIdforMerge();
                }
            }
        }
        // System.out.println("Count was: "+ count2 + " " + (getFilters().size()-1));
        // Check if perfect merging is used
        if (count2 >= getFilters().size() - 1) {
            return diff;
        }
        return null;
    }

    /** Returns true if the given template has the same structure
     * as this template. We count the number of mergeable
     * filters.  All filters in t2 must be mergeable with this
     * object. Selectivity may change for some fields in subsequent
     * (imperfect) merging.<br>
     * Note that template merging has direction (not symmetric)
     * Note that this can no longer be used as a condition for covering.
     * @param t2 the input template
     * @return int the number of the differing filter or -1 for error
     */
    public boolean mergeableStructure(Template t2) {
        if (mergeableStructureS(t2) != null) {
            return true;
        }
        return false;
    }

    /** Returns true if the given template has the same structure
     * as this template. We count the number of mergeable
     * filters.  All filters in t2 must be mergeable with this
     * object. Selectivity may change for some fields in subsequent
     * (imperfect) merging.<br>
     * Note that template merging has direction (not symmetric)
     * Note that this can no longer be used as a condition for covering.
     * Note that return value of all means that all are identical. This is handled
     * properly by the perfectlyMergeS methods
     * @param t2 the input template
     * @return String the mergeableId of the differing attribute filter or null or all.
     */
    public String mergeableStructureS(Template t2) {
        Vector filt = t2.getFilters();
        // Attribute counter
        int count = 0;
        Hashtable mergeable = new Hashtable();
        Hashtable temp = new Hashtable();

        /* First test: same fields and types and selectivities */
        for (int i = 0; i < filt.size(); i++) {
            Filter f2 = (Filter) filt.elementAt(i);
            mergeable.put(f2.getIdforMerge(), f2);
        }

        for (int i = 0; i < getFilters().size(); i++) {
            Filter f2 = (Filter) getFilters().elementAt(i);
            temp.put(f2.getIdforMerge(), f2);
        }

        // Check field count
        if (temp.size() != mergeable.size()) {
            return null;
        }

        // Check total field count
        // We cannot merge if there are differing parameters!
        if (getFilters().size() != filt.size()) {
            return null;
        }

        // Check equality =>
        for (int i = 0; i < getFilters().size(); i++) {
            Filter f2 = (Filter) getFilters().elementAt(i);
            if (mergeable.get(f2.getIdforMerge()) == null) {
                return null;
            }
        }
        // Check equality <=
        for (int i = 0; i < filt.size(); i++) {
            Filter f2 = (Filter) filt.elementAt(i);
            if (temp.get(f2.getIdforMerge()) == null) {
                return null;
            }
        }

        /* We iterate over the set of filters, and for each filter
           find the corresponding filter with the same name,type,selectivity.
         */
        int count2 = 0;
        String diff = "all";
        for (int i = 0; i < getFilters().size(); i++) {
            Filter f2 = (Filter) getFilters().elementAt(i);
            // If the selectivity is any, we have already
            // ascertained that there is a filter, and by definition we know
            // that there can be only one per selectivity type
            // we have to ensure that the type is exact
            Filter f3 = (Filter) mergeable.get(f2.getIdforMerge());
            if (f3 == null) {
                return null; // cannot merge field with null field
            } else if (f3.getSelectivity() != f2.getSelectivity()) {
                // Same name and type but different selectivity
                // cannot perform perfect merging!
                return null;
            }

            // Check that the two filters identical
            if (f2.getFilterString().trim().equals(f3.getFilterString().trim())) {
                count2++;
            } else {
                diff = f3.getIdforMerge();
            }
        }
        //System.out.println("Count was: "+ count2 + "/" + (getFilters().size()));
        // Check if perfect merging is used
        if (count2 >= getFilters().size() - 1) {
            return diff;
        }
        return null;
    }

    public void nicePrint() {
        for (int i = 0; i < getFilters().size(); i++) {
            Filter f = (Filter) getFilters().elementAt(i);
            System.out.println(f.getFilterString() + "\t");
        }
        System.out.println("*************************");
    }

    /** Returns true if the given template has the same structure
     * as this template. We count the number of mergeable
     * filters.  All filters in t2 must be mergeable with this
     * object. Selectivity may change for some fields in subsequent
     * (imperfect) merging.<br>
     * Note that template merging has direction (not symmetric)
     * Note that this can no longer be used as a condition for covering.
     * Note that return value of all means that all are identical. This is handled
     * properly by the perfectlyMergeS methods
     * @param t2 the input template
     * @return String the mergeableId of the differing attribute filter or null or all.
     */
    public boolean imperfectlyMergeableStructure(Template t2) {
        Vector filt = t2.getFilters();
        // Attribute counter
        int count = 0;
        Hashtable mergeable = new Hashtable();
        Hashtable temp = new Hashtable();

        /* First test: same fields and types and selectivities */
        for (int i = 0; i < filt.size(); i++) {
            Filter f2 = (Filter) filt.elementAt(i);
            mergeable.put(f2.getIdforMerge(), f2);
        }

        for (int i = 0; i < getFilters().size(); i++) {
            Filter f2 = (Filter) getFilters().elementAt(i);
            temp.put(f2.getIdforMerge(), f2);
        }

        // Check field count
        if (temp.size() != mergeable.size()) {
            return false;
        }

        // Check total field count
        // We cannot merge if there are differing parameters!
        if (getFilters().size() != filt.size()) {
            return false;
        }

        // Check equality =>
        for (int i = 0; i < getFilters().size(); i++) {
            Filter f2 = (Filter) getFilters().elementAt(i);
            if (mergeable.get(f2.getIdforMerge()) == null) {
                return false;
            }
        }
        // Check equality <=
        for (int i = 0; i < filt.size(); i++) {
            Filter f2 = (Filter) filt.elementAt(i);
            if (temp.get(f2.getIdforMerge()) == null) {
                return false;
            }
        }

        return true;
    }

    /** Perfectly merge the given template with this template.
     *  Returns a merged template.<br>Note that variable:
     * fuegocore.notify.profile.UserProfile.perfectMerging
     *  determines whether perfect or imperfect merging is used. The imperfect
     *  approach uses the same mechanism, only the mergeability check is different.
     *
     *  This can only be called if the two templates are structurally similar.<br>
     *  Note that template merging has direction (not symmetric!)
     *  This tries to merge in the direction (t1,t2).
     * @param t2 the input template
     * @return Template the perfectly merged template
     */
    public synchronized Template perfectMerge(Template t2) {
        Vector filt = t2.getFilters();
        Template t0 = null;

        try {
            t0 = (Template)this.clone();
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
        //t0 = new Template(toString());

        Hashtable mergeable = new Hashtable();

        /* First test: same fields and types and selectivities */
        for (int i = 0; i < t2.getFilters().size(); i++) {
            Filter f2 = (Filter) t2.getFilters().elementAt(i);
            mergeable.put(f2.getIdforMerge(), f2);
        }

        for (int i = 0; i < t0.getFilters().size(); i++) {
            Filter f = (Filter) t0.getFilters().elementAt(i);
            Filter f2 = (Filter) mergeable.get(f.getIdforMerge());
            // merge may be performed only once per name+type+selectivity
            if (f != null && f2 != null && f.perfectlyMergeable(f2)) {
                f.perfectMerge(f2); // perform the merge operation
                //System.out.println("Merged: " + f.toString());
                //t0.getFilters().set(i,f);
                //System.out.println("Merged:" + f.toString());
            } else {
                System.out.println("Error perfectMerge: no key found:" +
                                   f.getIdforMerge());
            }
        }
        return t0;
    }

    /** Perfectly merge the given template with this template.
     *  Returns a merged template. This method takes advantage of
     *  the knowledge of the differing attribute filter, which
     *  speeds up the merging considerably. This assumes that
     *  the filters are structurally equivalent and the differing
     *  filter has been found, otherwise the merging will not
     *  work properly. <br>
     * @param t2 the input template
     * @param id the mergeId of the differing attribute filter
     * @return Template the perfectly merged template
     */
    public synchronized Template perfectMergeS(Template t2, String id) {
        Vector filt = t2.getFilters();
        Template t0 = null;

        if (id == null) {
            return null; // failure
        }
        if (id.equals("all")) {
            return this; // identical
        }

        try {
            t0 = (Template)this.clone();
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
        //t0 = new Template(toString());

        Hashtable mergeable = new Hashtable();

        Filter m2 = null;
        Filter m1 = null;
        int p = 0;
        /* Find the attribute filters */
        for (int i = 0; i < t2.getFilters().size(); i++) {
            Filter f2 = (Filter) t2.getFilters().elementAt(i);
            if (f2.getIdforMerge().equals(id)) {
                m2 = f2;
                break;
            }
        }
        for (int i = 0; i < t0.getFilters().size(); i++) {
            Filter f2 = (Filter) t0.getFilters().elementAt(i);
            if (f2.getIdforMerge().equals(id)) {
                m1 = f2;
                p = i;
                break;
            }
        }

        if (m1 == null || m2 == null) {
            System.out.println("perfectMergeS m1 or m2 was null! id: " + id);
            return t0;
        }

        m1.perfectMerge(m2); // perform the merge operation
        //System.out.println("result: " + m1.toString());
        //t0.getFilters().set(p,m1);
        return t0;
    }

    /** Perfectly merge the given (conjunctive) template with this template.
     *  Returns a merged template. This method takes advantage of
     *  the knowledge of the differing attribute filter, which
     *  speeds up the merging considerably. This assumes that
     *  the filters are structurally equivalent and the differing
     *  filter has been found, otherwise the merging will not
     *  work properly. <br>
     * @param t2 the input template
     * @param id the mergeId of the differing attribute filter
     * @return Template the perfectly merged template
     */
    public synchronized Template conjunctivePerfectMergeS(Template t2,
            String id) {
        Vector filt = t2.getFilters();
        Template t0 = null;

        if (id == null) {
            return null; // failure
        }
        if (id.equals("all")) {
            return this; // identical
        }

        try {
            t0 = (Template)this.clone();
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }

        //t0 = new Template(toString());
        Hashtable mergeable = new Hashtable();

        Filter m2 = null;
        Filter m1 = null;
        int p = 0;

        /* Find the attribute filters */
        for (int i = 0; i < t2.getFilters().size(); i++) {
            Filter f2 = (Filter) t2.getFilters().elementAt(i);
            if (f2.getIdforMerge().equals(id)) {
                m2 = f2;
                break;
            }
        }
        for (int i = 0; i < t0.getFilters().size(); i++) {
            Filter f2 = (Filter) t0.getFilters().elementAt(i);
            if (f2.getIdforMerge().equals(id)) {
                m1 = f2;
                p = i;
                break;
            }
        }

        if (m1 == null || m2 == null) {
            System.out.println(
                    "conjunctivePerfectMergeS m1 or m2 was null! id: " +
                    id);
            return t0;
        }

        m1.conjunctivePerfectMerge(m2, false); // perform the merge operation
        //System.out.println("result: " + m1.toString());
        return t0;
    }

    /** Imperfectly merge the given template with this template.
     *  Uses perfect merge for single filter merging.
     * Note that template merging has direction (not symmetric)
     * @param t2 the input template
     * @return Template the imperfectly merged template
     */
    public synchronized Template imperfectMerge(Template t2) {
        Vector unique = new Vector();
        Template t0 = null;

        try {
            t0 = (Template)this.clone();
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }

        //t0 = new Template(toString());

        Vector filt = (Vector) getFilters().clone();
        Hashtable mergeable = new Hashtable();
        boolean me[] = new boolean[t2.getFilters().size()];
        Vector remove = new Vector();
        boolean merged;
        int count = 0; // Attribute counter

        if (t2 == null) {
            return null;
        }

        /** we remove selections from this template */

        for (int i = 0; i < filt.size(); i++) {
            Filter f2 = (Filter) filt.elementAt(i);
            Vector v = (Vector) mergeable.get(f2.getId());
            if (v == null) {
                v = new Vector();
                v.add(f2);
                mergeable.put(f2.getId(), v);
            } else {
                v.add(f2);
            }
        }

        // We transform each field with different selectivity into any
        // and merge it, so that we have unique fields with any selectivity if
        // there were many selectivies
        for (Enumeration i = mergeable.elements(); i.hasMoreElements(); ) {
            Vector v = (Vector) i.nextElement();
            Filter f2 = (Filter) v.firstElement();
            if (v != null && v.size() > 1) {
                f2.setSelectivity(Filter.SEL_ANY);
                for (int j = 1; j < v.size(); j++) {
                    Filter f3 = (Filter) v.elementAt(j);
                    f3.setSelectivity(Filter.SEL_ANY);
                    f2.perfectMerge(f3);
                }
            }
            unique.add(f2);
        }

        unique = filt;

        /* end selection removal */

        for (int i = 0; i < unique.size(); i++) {
            Filter f = (Filter) unique.elementAt(i);
            merged = false;
            for (int j = 0; j < t2.getFilters().size(); j++) {
                if (me[j]) {
                    continue;
                }
                Filter f2 = (Filter) t2.getFilters().elementAt(j);

                // Merge may be performed many times, because of selectivity support.
                if (f.imperfectlyMergeable(f2)) {
                    //System.out.println("merging: " + f.toString());
                    //System.out.println("with: " + f2.toString());

                    f.perfectMerge(f2); // perform the merge operation
                    unique.set(i, f);
                    //System.out.println("merged: " + f.toString());
                    merged = true; // mark dirty
                    me[j] = true;
                }
            }
            if (!merged) {
                remove.add(f);
            }
        }

        if (unique.removeAll((Collection) remove)) {
            ;
        }
//     System.out.println("Imperfect prune success, removed " + remove.size() + " fields");
//   else System.out.println("Imperfect prune failed");

        Vector cv = t0.getFilters();
        t0.filters = unique;

        return t0;
    }

    /**
     * Generate a canonical description of the template
     * @return String
     */
    public String getCoverId() {
        Vector ff = (Vector) getFilters();
        Collections.sort(ff, filterComparator); // sort
        if (coverId == null) {
            coverId = "";
            for (int i = 0; i < ff.size(); i++) {
                Filter f = (Filter) ff.elementAt(i);
                coverId += f.getCoverId() + "_";
            }
        }
        return coverId;
    }

    /**
     * Auxiliary method to generate hashcodes
     * @return int
     */
    private int generateHash() {
        int ret = hashCode();
        for (int i = 0; i < getFilters().size(); i++) {
            Filter f = (Filter) getFilters().elementAt(i);
            ret += f.hashCode();
        }
        return ret;
    }

    /** Matches the object with the given object
     * @param et the EventTriplet object to match, may be null
     * @param noti the TripletTable, may not be null
     * @return boolean true if the match succeeded
     */
    public boolean match(Object et, Object noti) {
        if (noti instanceof Notification) {
            return match(this, ((Notification) noti).get().getTriplets());
        } else if (!(noti instanceof TripletTable)) {
            return false;
        }
        return match(this, (TripletTable) noti);
    }

    /** Merges the object with the given object
     * @param obj the object to be merged
     * @return boolean true if the merge succeeded
     */
    public boolean merge(Object obj) {
        if (!(obj instanceof Template)) {
            return false;
        }
        if (!mergeableStructure((Template) obj)) {
            return false;
        }
        Template tt = perfectMerge((Template) obj);
        getFilters().clear();
        getFilters().addAll(tt.getFilters());
        return true;
    }

    /** Merges the current object with the given object
     * @param obj the object to be merged with the current object
     * @return Object the merged result
     */
    public Object mergeObject(Object obj) {
        if (!(obj instanceof Template)) {
            return null;
        }
        if (!mergeableStructure((Template) obj)) {
            return null;
        }
        Template tt = perfectMerge((Template) obj);

        if (((Template) obj).getTid() != null) {
            tt.tid = (Vector) ((Template) obj).getTid().clone();
        }

        return tt;
    }

    /**
     * Match a notification against the given template.
     * @param t Template
     * @param tt TripletTable
     * @return boolean
     */
    public static synchronized boolean match(Template t, TripletTable tt) {
        if (t == null || tt == null) {
            return false;
        }

        int count = 0;

        // This uses the hashtable in TripletTable for fast execution
        for (int i = 0; i < t.getFilters().size(); i++) {
            Filter f = (Filter) t.getFilters().elementAt(i);
            if (f != null) {
                Vector v = tt.get(f.getId());
                // We get just the first triplet, because the filter
                // automatically uses the correct selectivity using TripletTable

                if (v != null && v.size() > 0) {
                    EventTriplet et = (EventTriplet) v.firstElement();
                    if (f.evaluate(et, tt)) {
                        count++;
                    }
                    // These cannot be matched, so they are counted automatically
                    else if (et.getType() == EventTriplet.TYPE_SERIALIZED) {
                        count++;
                    }
                }
            }
        }
        // Return true if all filters were matched
        if (count == t.getFilters().size()) {
            return true;
        }
        return false;
    }

    /**
     * Covering test
     * @param obj the covered object
     * @return true if covered
     */
    public boolean covers(Object obj) {
        if (!(obj instanceof Template)) {
            return false;
        }
        return covers((Template) obj);
    }


    /** Returns true if this template shares common fields with the given template
     * Needed to handle the special case in checking overlapping subsc.
     * @param obj the Template
     * @return true if the current object shares fields
     */
    public boolean commonFields(Object obj) {

        if (!(obj instanceof Template)) {
            return false;
        }

        Template t = (Template) obj;

        Template s, l;

        // Find the smallest set
        if (t.getFilters().size() >= this.getFilters().size()) {
            s = this;
            l = t;
        } else {
            s = t;
            l = this;
        }

        for (int i = 0; i < s.getFilters().size(); i++) {
            Filter f = (Filter) s.getFilters().elementAt(i);
            for (int j = 0; j < l.getFilters().size(); j++) {
                Filter f2 = (Filter) l.getFilters().elementAt(j);
                if (f.getName().equals("type") || f2.getName().equals("type")) {
                    continue; // skip type
                }
                if (f.getId().equals(f2.getId())) {
                    return true;
                }
            }
        }
        return false;
    }


    /** Returns true if this template overlaps with the given object
     * @param obj the Template
     * @return true if the current object overlaps with the input object
     */
    public boolean overlaps(Object obj) {

        if (!(obj instanceof Template)) {
            return false;
        }

        int count = 0;

        Template t = (Template) obj;

        Template s, l;

        // Find the smallest set
        if (t.getFilters().size() >= this.getFilters().size()) {
            s = this;
            l = t;
        } else {
            s = t;
            l = this;
        }

        for (int i = 0; i < s.getFilters().size(); i++) {
            Filter f = (Filter) s.getFilters().elementAt(i);
            for (int j = 0; j < l.getFilters().size(); j++) {
                Filter f2 = (Filter) l.getFilters().elementAt(j);
                // if they share a name+type and the tuple does not overlap
                // they cannot overlap
                if (f.getId().equals(f2.getId()) && !f.overlaps(f2)) {
                    return false;
                }
            }
        }
        return true;
    }

    public double getOverhead() {
        return overhead;
    }

    public void setOverhead(double o) {
        overhead = o;
    }

    /** returns the distance of two overlapping filters
     * @param obj the Template
     * @return the distance metric (1=no overlap, 0<x<1 overlap,0=overlap on all AF).
     */
    public double getOverlapCost(Object obj) {

        if (!(obj instanceof Template)) {
            return 1;
        }

        double dcount = 0;
        int count = 0;
        Template t = (Template) obj;

        Template s, l;

        // Find the smallest set
        if (t.getFilters().size() > this.getFilters().size()) {
            s = this;
            l = t;
        } else {
            s = t;
            l = this;
        }

        for (int i = 0; i < s.getFilters().size(); i++) {
            Filter f = (Filter) s.getFilters().elementAt(i);
            for (int j = 0; j < l.getFilters().size(); j++) {
                Filter f2 = (Filter) l.getFilters().elementAt(j);
                if (f.covers(f2) || f2.covers(f)) {
                    dcount++;
                    count++;
                    break;
                } else if (f.overlaps(f2)) {
                    // modified
                    dcount += f.getOverlapCost(f2);
                    count++;
                    break;
                }
            }
        }
        // System.out.println("Overlap count: " + count);
        // If all attribute filters overlap return true
        if (count == s.getFilters().size()) {
//      System.out.println("huu count:"+dcount);
//      System.out.println("size:"+ l.getFilters().size());
            double d = 1 - ((double) dcount / (double) l.getFilters().size());
            //     System.out.println("d:"+d);
            return d;
        }
        return 1;
    }

    /**
     * Covering test: is the given object covered
     * @param obj the object that covers
     * @return true if the current object was covered
     */
    public boolean isCovered(Object obj) {
        if (!(obj instanceof Template)) {
            return false;
        }
        Template tobj = (Template) obj;
        if (tobj == null) {
            return false;
        }
        return tobj.covers(this);
    }

    /** Returns true if the template covers the given template. This is
     *  based on filter counting. Template A covers B if all filters in A are
     *  covered in B.
     *  Duplicate handling: Since there may be several tuples with the same name,
     *  but different selectivity: we must loop over all the duplicates before
     *  any filter may cover one in A (to avoid false positives)
     *  @param t2 the input template
     *  @return true if this template covers the input template
     */
    public boolean covers(Template t2) {

        // If this has fewer attribute filters, it cannot cover
        if (getFilters().size() > t2.getFilters().size()) {
            return false;
        }

        Hashtable duplicates = new Hashtable();
        Vector filt = getFilters();
        int count = 0;

        // Create a hashtable for the duplicates
        for (int i = 0; i < t2.getFilters().size(); i++) {
            Filter f2 = (Filter) t2.getFilters().elementAt(i);
            Hashtable hsh = (Hashtable) duplicates.get(f2.getId());
            // we count only duplicate elements
            if (hsh == null) {
                hsh = new Hashtable();
                duplicates.put(f2.getId(), hsh);
            }
            hsh.put("" + count, f2);
            count++;
        }
        count = 0; // f1 counter

        synchronized (this) {
            for (int i = 0; i < filt.size(); i++) {
                Filter f = (Filter) filt.elementAt(i);

                // Get hashtable for this name+type
                Hashtable h2 = (Hashtable) duplicates.get(f.getId());

                if (h2 == null) {
                    break; // no reason to continue
                }

                int cc = 0;

                for (Enumeration j = h2.elements(); j.hasMoreElements(); ) {
                    Filter f2 = (Filter) j.nextElement();
                    // if not any, we can skip different selectivities
                    // for any we can check different selectivities
                    if (f.getSelectivity() != Filter.SEL_ANY) {
                        if (f.getSelectivity() == f2.getSelectivity() &&
                            f.covers(f2)) {
                            cc++;
                            break; // we can break since it is enought that one covers
                        }
                    } else if (f.covers(f2)) { // any
                        cc++;
                        break; // we can break since it is enought that one covers
                    }
                }

                if (cc > 0) { // increment count
                    count++;
                }
            }
        }
        if (count == filt.size()) {
            return true;
        }
        return false;
    }

    public Template(Notifiable s) {
        source = s;
    }

    /** Get the source of the subscription / template
     * @return Notifiable the source of the template
     */
    public Notifiable getSource() {
        return source;
    }

    /** Set the source of the subscription / template
     * @param eb the Notifiable interface
     */
    public void setSource(Notifiable eb) {
        source = eb;
    }

    // test driver
    public static void main(String[] args) {
        Template t = new Template();
        t.runTest();
    }

    public boolean runTest() {

        // First test mergability
        if (!testMergeableStructure()) {
            return false;
        }

        // Then run more specific tests

        Template t = new Template();
        t.addFilter(new Filter("type", EventTriplet.TYPE_STRING,
                               "abc"));
        t.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                               "select any&gr 10"));
        t.addFilter(new Filter("field2", EventTriplet.TYPE_INT,
                               "select 1&gr 10"));

        // ----
        Template t2 = new Template();
        t2.addFilter(new Filter("type", EventTriplet.TYPE_STRING,
                                "abc"));
        t2.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                "select any&gr 5"));
        t2.addFilter(new Filter("field2", EventTriplet.TYPE_INT,
                                "select 1&gr 5"));
        t2.addFilter(new Filter("field2", EventTriplet.TYPE_INT,
                                "select 2&gr 5"));

        // ----
        Template t3 = new Template();
        t3.addFilter(new Filter("type", EventTriplet.TYPE_STRING,
                                "abc"));
        t3.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                "select any&gr 5"));
        t3.addFilter(new Filter("field2", EventTriplet.TYPE_INT,
                                "select all&gr 10"));

        Notification n = new Notification();
        n.setType("abc");
        n.add(new EventTriplet("field1", 20));
        n.add(new EventTriplet("field1", 2));
        n.add(new EventTriplet("field2", 10));
        n.add(new EventTriplet("field2", 30));

        if (t.match(null, n)) {
            System.out.println("List test #1 ok");
        } else {
            System.out.println("List test #1 failed");
            return false;
        }

        if (!t2.match(null, n)) {
            System.out.println("List test #2 ok");
        } else {
            System.out.println("List test #2 failed");
            return false;
        }

        n.add(new EventTriplet("field2", 30));

        if (t2.match(null, n)) {
            System.out.println("List test #2 ok");
        } else {
            System.out.println("List test #2 failed");
            return false;
        }

        if (!t3.match(null, n)) {
            System.out.println("List test #3 ok");
        } else {
            System.out.println("List test #3 failed");
            return false;
        }

        Notification n2 = new Notification();
        n2.setType("abc");
        n2.add(new EventTriplet("field1", 20));
        n2.add(new EventTriplet("field1", 20));
        n2.add(new EventTriplet("field2", 11));
        n2.add(new EventTriplet("field2", 30));

        if (t3.match(null, n2)) {
            System.out.println("List test #4 ok");
        } else {
            System.out.println("List test #4 failed");
            return false;
        }
        System.out.println("Performing cover test");

        Template t4 = new Template();
        t4.addFilter(new Filter("type", EventTriplet.TYPE_STRING,
                                "abc"));
        t4.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                "select any&gr 5"));
        t4.addFilter(new Filter("field2", EventTriplet.TYPE_INT,
                                "select any&gr 10"));

        Template t5 = new Template();
        t5.addFilter(new Filter("type", EventTriplet.TYPE_STRING,
                                "abc"));
        t5.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                "select 2&gr 15"));
        t5.addFilter(new Filter("field2", EventTriplet.TYPE_INT,
                                "select any&gr 19"));

        if (t4.covers(t5)) {
            System.out.println("Cover test #1 ok (same # fields,covered)");
        } else {
            System.out.println("Cover test #1 failed");
            return false;
        }
        if (!t5.covers(t4)) {
            System.out.println("Cover test #2 ok (reverse #1)");
        } else {
            System.out.println("Cover test #2 failed");
            return false;
        }

        Template t6 = new Template();
        t6.addFilter(new Filter("type", EventTriplet.TYPE_STRING,
                                "abc"));
        t6.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                "select any&gr 5"));
        t6.addFilter(new Filter("field2", EventTriplet.TYPE_INT,
                                "select any&gr 10"));

        Template t7 = new Template();
        t7.addFilter(new Filter("type", EventTriplet.TYPE_STRING,
                                "abc"));
        t7.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                "select 2&gr 15"));
        t7.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                "select 1&gr 10"));
        t7.addFilter(new Filter("field2", EventTriplet.TYPE_INT,
                                "select any&gr 19"));
        t7.addFilter(new Filter("field2", EventTriplet.TYPE_INT,
                                "select all&gr 12"));
        t7.addFilter(new Filter("field2", EventTriplet.TYPE_INT,
                                "select 2&gr 100"));

        if (t6.covers(t7)) {
            System.out.println("Cover test #3 ok (differing # fields, cover)");
        } else {
            System.out.println("Cover test #3 failed");
            return false;
        }
        if (!t7.covers(t6)) {
            System.out.println("Cover test #4 ok (reverse 3)");
        } else {
            System.out.println("Cover test #4 failed");
            return false;
        }
        // -----------------------------------
        Template t8 = new Template();
        t8.addFilter(new Filter("type", EventTriplet.TYPE_STRING,
                                "abc"));
        t8.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                "select any&gr 5"));
        t8.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                "select 1&gr 10"));
        t8.addFilter(new Filter("field2", EventTriplet.TYPE_INT,
                                "select any&gr 15"));
        t8.addFilter(new Filter("field2", EventTriplet.TYPE_INT,
                                "select all&gr 10"));

        Template t9 = new Template();
        t9.addFilter(new Filter("type", EventTriplet.TYPE_STRING,
                                "abc"));
        t9.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                "select 2&gr 15"));
        t9.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                "select 1&gr 10"));
        t9.addFilter(new Filter("field2", EventTriplet.TYPE_INT,
                                "select any&gr 19"));
        t9.addFilter(new Filter("field2", EventTriplet.TYPE_INT,
                                "select all&gr 12"));
        t9.addFilter(new Filter("field2", EventTriplet.TYPE_INT,
                                "select 2&gr 100"));

        if (t8.covers(t9)) {
            System.out.println("Cover test #5 ok (differing # fields, cover)");
        } else {
            System.out.println("Cover test #5 failed");
            return false;
        }
        if (!t9.covers(t8)) {
            System.out.println("Cover test #6 ok (reverse 3)");
        } else {
            System.out.println("Cover test #6 failed");
            return false;
        }

        // -----------------------------------
        Template t10 = new Template();
        t10.addFilter(new Filter("type", EventTriplet.TYPE_STRING,
                                 "abc"));
        t10.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                 "select any&gr 5"));

        Template t11 = new Template();
        t11.addFilter(new Filter("type", EventTriplet.TYPE_STRING,
                                 "abc"));
        t11.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                 "select 2&gr 15"));
        t11.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                 "select 1&gr 10"));

        if (t10.covers(t11)) {
            System.out.println("Cover test #7 ok (differing # fields, cover)");
        } else {
            System.out.println("Cover test #7 failed");
            return false;
        }
        if (!t11.covers(t10)) {
            System.out.println("Cover test #8 ok (reverse 3)");
        } else {
            System.out.println("Cover test #8 failed");
            return false;
        }

        // Perfect merging tests
        Template x1 = new Template();
        x1.addFilter(new Filter("type", EventTriplet.TYPE_STRING,
                                "abc"));
        x1.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                "select any&gr 5&neq 5"));
        x1.addFilter(new Filter("field2", EventTriplet.TYPE_INT,
                                "select 1&lt 10"));

        Template x2 = new Template();
        x2.addFilter(new Filter("type", EventTriplet.TYPE_STRING,
                                "abc"));
        x2.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                "select any&lt 10&neq 5"));
        x2.addFilter(new Filter("field2", EventTriplet.TYPE_INT,
                                "select 1&lt 10"));
        Template z = null;

        if (x1.mergeableStructure(x2)) {
            z = x1.perfectMerge(x2);
            System.out.println("Merge test #1 ok: " + z.toString());
        } else {
            System.out.println("Merge test #1 failed");
            return false;
        }
        x2 = null;
        x2 = new Template();
        x2.addFilter(new Filter("type", EventTriplet.TYPE_STRING,
                                "abc"));
        x2.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                "select any&lt 10&neq 5"));
        x2.addFilter(new Filter("field2", EventTriplet.TYPE_INT,
                                "select 1&lt 10"));

        String str = null;
        str = x1.mergeableStructureS(x2);
        System.out.println("Differing id:" + str);
        if (str != null) {
            z = x1.perfectMergeS(x2, str);
            System.out.println("Merge test #2 ok: " + z.toString());
        } else {
            System.out.println("Merge test #2 failed");
            return false;
        }
        x2 = null;
        x2 = new Template();
        x2.addFilter(new Filter("type", EventTriplet.TYPE_STRING,
                                "abc"));
        x2.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                "select any&lt 10&neq 5"));
        x2.addFilter(new Filter("field2", EventTriplet.TYPE_INT,
                                "select 1&lt 10"));

        str = null;
        str = x1.mergeableConjunctiveStructure(x2);
        System.out.println("Differing id:" + str);
        if (str != null) {
            z = x1.conjunctivePerfectMergeS(x2, str);
            System.out.println("Merge test #3 ok: " + z.toString());
        } else {
            System.out.println("Merge test #3 failed:" + str);
            return false;
        }

        // Perfect merging tests
        Template z1 = new Template();
        z1.addFilter(new Filter("type", EventTriplet.TYPE_STRING,
                                "abc"));
        z1.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                "rng 10 15"));

        Template z2 = new Template();
        z2.addFilter(new Filter("type", EventTriplet.TYPE_STRING,
                                "abc"));
        z2.addFilter(new Filter("field1", EventTriplet.TYPE_INT,
                                "rng 9 14"));

        z = null;

        if (z1.mergeableStructure(z2)) {
            z = z1.perfectMerge(z2);
            System.out.println("Merge test #4 ok: " + z.toString());
        } else {
            System.out.println("Merge test #4 failed");
            return false;
        }

        str = z1.mergeableStructureS(z2);
        System.out.println("Differing id:" + str);
        if (str != null) {
            z = z1.perfectMergeS(z2, str);
            System.out.println("Merge test #5 ok: " + z.toString());
        } else {
            System.out.println("Merge test #5 failed");
            return false;
        }
        str = z1.mergeableConjunctiveStructure(z2);
        System.out.println("Differing id:" + str);
        if (str != null) {
            z = z1.conjunctivePerfectMergeS(z2, str);
            System.out.println("Merge test #6 ok: " + z.toString());
        } else {
            System.out.println("Merge test #6 failed:" + str);
            return false;
        }
        return true;
    }

    /**
     * Tests mergeable structure of templates
     * @return boolean
     */
    public boolean testMergeableStructure() {
        Vector temps = new Vector();

        /**********************************************************/

        System.out.println("Performing mergeability test #1");

        for (int i = 0; i < 1000; i++) {
            Template tm = new Template();
            tm.addFilter(new Filter("val" + i, EventTriplet.TYPE_INT,
                                    "equ " + i));
            tm.addFilter(new Filter("val2" + i, EventTriplet.TYPE_INT,
                                    "equ 1"));
            tm.addFilter(new Filter("val3", EventTriplet.TYPE_INT,
                                    "equ 50"));
            tm.addFilter(new Filter("val4", EventTriplet.TYPE_STRING,
                                    "equ abc"));
            temps.add(tm);
        }

        int conj = 0, merg = 0;

        for (int i = 0; i < temps.size(); i++) {
            Template t = (Template) temps.elementAt(i);
            for (int j = 0; j < temps.size(); j++) {
                if (i != j) {
                    Template t2 = (Template) temps.elementAt(j);
                    if (t.mergeableConjunctiveStructure(t2) != null) {
                        conj++;
                    }
                    if (t.mergeableStructureS(t2) != null) {
                        merg++;
                    }
                }
            }
        }
        System.out.println("Conjuncts: " + conj + " normal: " + merg);
        if (conj != 0 || merg != 0) {
            System.out.println("Template mergeability test #1 failed");
            return false;
        } else {
            System.out.println("Template mergeability test #1 ok");

            /**********************************************************/
        }
        temps.clear();

        System.out.println("Performing mergeability test #2");

        for (int i = 0; i < 1000; i++) {
            Template tm = new Template();
            tm.addFilter(new Filter("val0", EventTriplet.TYPE_INT,
                                    "equ " + i % 100));
            tm.addFilter(new Filter("val1", EventTriplet.TYPE_INT,
                                    "equ 1"));
            tm.addFilter(new Filter("val2", EventTriplet.TYPE_INT,
                                    "equ 50"));
            tm.addFilter(new Filter("val3", EventTriplet.TYPE_STRING,
                                    "equ abc"));
            temps.add(tm);
        }

        conj = 0;
        merg = 0;

        for (int i = 0; i < temps.size(); i++) {
            Template t = (Template) temps.elementAt(i);
            for (int j = 0; j < temps.size(); j++) {
                if (i != j) {
                    Template t2 = (Template) temps.elementAt(j);
                    if (t.mergeableConjunctiveStructure(t2) != null) {
                        conj++;
                    }
                    if (t.mergeableStructureS(t2) != null) {
                        merg++;
                    }
                }
            }
        }
        System.out.println("Conjuncts: " + conj + " normal: " + merg);
        // All except the template itself will be matched
        // there are 1000/10 conjunctively mergeable (identical) fields
        // perfect merging using disjunctions is able to merge them all
        if (conj != (1000 / 100) * temps.size() - temps.size() ||
            merg != temps.size() * temps.size() - temps.size()) {
            System.out.println("Template mergeability test #2 failed");
            return false;
        } else {
            System.out.println("Template mergeability test #2 ok");

            /**********************************************************/

        }
        System.out.println("Performing mergeability test #3");

        conj = 0;
        merg = 0;

        for (int i = 0; i < temps.size(); i++) {
            Template t = (Template) temps.elementAt(i);
            String s = t.mergeableConjunctiveStructure(t);
            if (s != null && s.equals("all")) {
                conj++;
            }
            s = t.mergeableStructureS(t);
            if (s != null && s.equals("all")) {
                merg++;
            }
        }
        System.out.println("Conjuncts: " + conj + " normal: " + merg);
        // All except the template itself will be matched
        if (conj != temps.size() || merg != temps.size()) {
            System.out.println("Template mergeability test #3 failed");
            return false;
        } else {
            System.out.println("Template mergeability test #3 ok");

        }

        return true;
    }

}
