/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.channel;

import java.io.*;
import java.util.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.filter.merge.*;
import fuegocore.notify.core.serialization.*;
import fuegocore.util.xas.*;

/**
 * This class represents a set of filters that match a specific interface.
 * This is used to store event channel modification and reroute information.
 */
public class ECRRerouteInfo {
    /** Vector of Templates that are added */
    private Vector active_filters;

    /** Templates that are removed */
    private Vector remove_filters;

    /** Name of the channel */
    private String channel_name;

    /** The interface */
    private String iface;

    /** Merging algorithm */
    private String merger = "perfect";

    public void setRemoved(Vector v) {
        remove_filters = v;
    }

    public void setMergingAlgorithm(String s) {
        merger = s;
    }

    public static void main(String[] args) {
        Template t = new Template();
        Template t2 = new Template();
        Vector a = new Vector();
        a.add(t);
        Vector b = new Vector();
        b.add(t);
        b.add(t2);

        ECRRerouteInfo ecr = new ECRRerouteInfo(a, b, "abc");

        ECRRerouteInfo ecr3 = new ECRRerouteInfo(a, a, "abc2");

        String s = ecr.toString();

        System.out.println("ECR:" + s);

        System.out.println("------------------");

        ECRRerouteInfo ecr2 = new ECRRerouteInfo();

        ecr2.parse(s);

        System.out.println("Active: " + ecr2.getActive().size() +
                           " should be 1");
        System.out.println("Remove: " + ecr2.getRemoved().size() +
                           " should be 2");
        System.out.println("------------------");
        System.out.println("Parsed:" + ecr2.toString());
        System.out.println("------------------");
        Hashtable n = new Hashtable();
        n.put("1", ecr);
        n.put("2", ecr2);
        n.put("3", ecr3);
        String s2 = multiplex(n);
        Hashtable h = demultiplex(s2);
        System.out.println("After demultiplexing:" + s2);

        ECRRerouteInfo ecr4 = (ECRRerouteInfo) h.get("1");

        System.out.println("Active: " + ecr4.getActive().size() +
                           " should be 1");
        System.out.println("Remove: " + ecr4.getRemoved().size() +
                           " should be 2");

        ECRRerouteInfo ecr5 = (ECRRerouteInfo) h.get("3");
        System.out.println("Active: " + ecr5.getActive().size() +
                           " should be 1");
        System.out.println("Remove: " + ecr5.getRemoved().size() +
                           " should be 1");
    }

    public ECRRerouteInfo() {
        active_filters = new Vector();
        remove_filters = new Vector();
    }

    /**
     * Constructor
     * @param channel channel name
     */
    public ECRRerouteInfo(String channel) {
        channel_name = channel;
        active_filters = new Vector();
        remove_filters = new Vector();
    }

    /**
     * Constructor
     * @param a Vector the active set
     * @param b Vector the set of filters to be removed
     * @param n String the channel name
     */
    public ECRRerouteInfo(Vector a, Vector b, String n) {
        active_filters = a;
        remove_filters = b;
        channel_name = n;
    }

    /**
     * Returns the interface
     * @return String
     */
    public String getInterface() {
        return iface;
    }

    /**
     * Sets the interface
     * @param s String
     */
    public void setInterface(String s) {
        iface = s;
    }

    /**
     * Returns active filters
     * @return Vector
     */
    public Vector getActive() {
        return active_filters;
    }

    /**
     * Returns removed filters
     * @return Vector
     */
    public Vector getRemoved() {
        return remove_filters;
    }

    /**
     * Returns channel name
     * @return String
     */
    public String getChannel() {
        return channel_name;
    }

    /**
     * Sets the channel name
     * @param ch String
     */
    public void setChannel(String ch) {
        channel_name = ch;
    }

    /**
     * Returns a string representation of this object. Merges the new templates
     * for the event channel
     * @return String
     */
    public String getMergedString() {
        String ret = "";
        ret += "<channel>" + getChannel() + "</channel>";
        ret += "<filter_type_delimiter/>";
        if (active_filters != null) {
            Vector active_b = MergingService.get(merger).mergeSet(
                    active_filters);
            for (int i = 0; i < active_b.size(); i++) {
                Template t = (Template) active_b.elementAt(i);
                if (t != null) {
                    ret += t.toString();
                    ret += "<template_dlm/>";
                }
            }
        }

        ret += "<filter_type_delimiter/>";

        if (remove_filters != null) {
            for (int i = 0; i < remove_filters.size(); i++) {
                Template t = (Template) remove_filters.elementAt(i);
                if (t != null) {
                    ret += t.toString();
                    ret += "<template_dlm/>";
                }
            }
        }
        return ret;
    }

    /** XAS Serialization */

    /* *************************************************************************/
    /* *************************************************************************/

    /**
     * Insert the templates etc. to an event sequence.
     * @param ser TypedXmlSerializer
     * @throws IOException
     */
    public void getMergedSaxSequence(TypedXmlSerializer ser) throws IOException {
        XmlWriter xw = new XmlWriter(ser);

        xw.typedElement("", "channel", XasUtil.XSD_NAMESPACE, "string",
                        getChannel());

        Vector active_b = MergingService.get(merger).mergeSet(active_filters);

        xw.complexElement(Filter.EVENT_NAMESPACE, "active",
                          EventSerializationService.toEventSequence(active_b));

        xw.complexElement(Filter.EVENT_NAMESPACE, "remove",
                          EventSerializationService.toEventSequence(
                remove_filters));
    }

    /**
     * Insert the templates etc. to an event sequence.
     * @param ser TypedXmlSerializer
     * @throws IOException
     */
    public void getEventSequence(TypedXmlSerializer ser) throws IOException {
        XmlWriter xw = new XmlWriter(ser);

        xw.typedElement("", "channel", XasUtil.XSD_NAMESPACE, "string",
                        getChannel());

        xw.complexElement(Filter.EVENT_NAMESPACE, "active",
                          EventSerializationService.toEventSequence(
                active_filters));

        xw.complexElement(Filter.EVENT_NAMESPACE, "remove",
                          EventSerializationService.toEventSequence(
                remove_filters));
    }

    /**
     * Parse filters from an event sequence
     * @param ses EventSequence
     * @throws IOException
     */
    public void parse(EventSequence ses) throws IOException {
        XmlReader xr = new XmlReader(ses);
        channel_name = (String) xr.typedContent("", "channel",
                                                XasUtil.XSD_NAMESPACE, "string");

        EventSequence active = xr.completeContent(Filter.EVENT_NAMESPACE,
                                                  "active");
        active_filters = EventSerializationService.fromEventSequence(active);
        EventSequence remove = xr.completeContent(Filter.EVENT_NAMESPACE,
                                                  "remove");
        remove_filters = EventSerializationService.fromEventSequence(remove);
    }

    /**
     * Demultiplexes a number of objects from an event sequence
     * @param ses EventSequence
     * @throws IOException
     * @return Hashtable
     */
    public static Hashtable demultiplexSax(EventSequence ses) throws
            IOException {
        Hashtable ret = new Hashtable();
        boolean quit = false;

        XmlReader xr = new XmlReader(ses);

        if (xr == null || xr.isFinished()) {
            return ret;
        }

        // attribute filters follow
        while (!quit && !xr.isFinished()) {
            EventSequence xes = xr.completeContent(Filter.EVENT_NAMESPACE,
                    "ecrreroute");
            if (xes == null) {
                quit = true;
            } else {
                ECRRerouteInfo ecr = new ECRRerouteInfo();
                ecr.parse(xes);
                if (ecr != null) {
                    ret.put(ecr.getChannel(), ecr);
                }
            }
        }
        return ret;
    }

    /**
     * Multiplexes a number of objects into an event stream.
     * @param ser TypedXmlSerializer
     * @param hsh Hashtable
     * @throws IOException
     */
    public static void multiplexSax(TypedXmlSerializer ser, Hashtable hsh) throws
            IOException {
        if (hsh == null) {
            return;
        }

        XmlWriter xw = new XmlWriter(ser);

        for (Enumeration j = hsh.elements(); j.hasMoreElements(); ) {
            EventSerializer evs = new EventSerializer();
            XmlWriter xw2 = new XmlWriter(evs);
            ECRRerouteInfo ecr = (ECRRerouteInfo) j.nextElement();
            if (ecr == null) {
                return;
            }
            ecr.getEventSequence(evs);
            xw.complexElement(Filter.EVENT_NAMESPACE, "ecrreroute",
                              evs.getCurrentSequence());
        }
    }

    /**
     * Multiplexes a number of objects into an event stream.
     * @param ser TypedXmlSerializer
     * @param hsh Hashtable
     * @throws IOException
     */
    public static void multiplexSaxMerge(TypedXmlSerializer ser, Hashtable hsh) throws
            IOException {
        if (hsh == null) {
            return;
        }
        XmlWriter xw = new XmlWriter(ser);

        for (Enumeration j = hsh.elements(); j.hasMoreElements(); ) {
            EventSerializer evs = new EventSerializer();
            XmlWriter xw2 = new XmlWriter(evs);
            ECRRerouteInfo ecr = (ECRRerouteInfo) j.nextElement();
            if (ecr == null) {
                return;
            }
            ecr.getMergedSaxSequence(evs);
            xw.complexElement(Filter.EVENT_NAMESPACE, "ecrreroute",
                              evs.getCurrentSequence());
        }
    }

    /** XAS Serialization */

    /* *************************************************************************/
    /* *************************************************************************/

    /**
     * Returns a string representation of this object
     * @return String
     */
    public String toString() {
        String ret = "";
        ret += "<channel>" + getChannel() + "</channel>";
        ret += "<filter_type_delimiter/>";
        if (active_filters != null) {
            for (int i = 0; i < active_filters.size(); i++) {
                Template t = (Template) active_filters.elementAt(i);
                if (t != null) {
                    ret += t.toString();
                    ret += "<template_dlm/>";
                }
            }
        }

        ret += "<filter_type_delimiter/>";

        if (remove_filters != null) {
            for (int i = 0; i < remove_filters.size(); i++) {
                Template t = (Template) remove_filters.elementAt(i);
                if (t != null) {
                    ret += t.toString();
                    ret += "<template_dlm/>";
                }
            }
        }
        return ret;
    }

    /**
     * Parse this object from a string representation
     * @param str String
     * @return boolean
     */
    public boolean parse(String str) {
        try {
            int i = 0;
            int j = str.indexOf("<filter_type_delimiter/>");

            String ch = str.substring(i, j);
            setChannel(fuegocore.notify.util.Auxil.getTag(ch, "channel"));

            i = j + "<filter_type_delimiter/>".length();
            j = str.indexOf("<filter_type_delimiter/>", i);

            String active = str.substring(i, j);
            String remove = str.substring(j + "<filter_type_delimiter/>".length(),
                                          str.length());

            active_filters = parseVector(active);

            remove_filters = parseVector(remove);
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }
        return true;
    }

    /**
     * Parse this object from the given string.
     * @param str String
     * @return Vector
     */
    private Vector parseVector(String str) {
        Vector ret = new Vector();

        int i = 0;
        int j = str.indexOf("<template_dlm/>");
        if (j == -1) {
            return null;
        } while (j != -1) {
            String s = str.substring(i, j);
            Template t = new Template(s);
            if (t != null) {
                ret.add(t);
            }
            i = j + "<template_dlm/>".length();
            j = str.indexOf("<template_dlm/>", i);
        }
        return ret;
    }

    /**
     * Multiplexes a number of objects into a string
     * @param hsh Hashtable
     * @return String
     */
    public static String multiplex(Hashtable hsh) {
        String ret = "";

        if (hsh == null) {
            return null;
        }
        Enumeration k = hsh.keys();
        for (Enumeration j = hsh.elements(); j.hasMoreElements(); ) {
            ECRRerouteInfo ecr = (ECRRerouteInfo) j.nextElement();
            if (ecr == null) {
                return null;
            }
            String ch = (String) k.nextElement();
            ecr.setChannel(ch);
            ret += ecr.toString();
            ret += "<ecrreroute/>";
        }

        return ret;
    }

    /**
     * Multiplexes a number of objects into a string. Uses merging
     * for new subscriptions for the event channel.
     * @param hsh Hashtable
     * @return String
     */
    public static String multiplexMerge(Hashtable hsh) {
        String ret = "";

        if (hsh == null) {
            return null;
        }
        for (Enumeration j = hsh.elements(); j.hasMoreElements(); ) {
            ECRRerouteInfo ecr = (ECRRerouteInfo) j.nextElement();
            if (ecr == null) {
                return null;
            }
            ret += ecr.getMergedString();
            ret += "<ecrreroute/>";
        }

        return ret;
    }

    /**
     * Demultiplexes a number of objects from a string
     * @param str String
     * @return Hashtable
     */
    public static Hashtable demultiplex(String str) {
        Hashtable ret = new Hashtable();

        int i = 0;
        int j = str.indexOf("<ecrreroute/>");
        if (j == -1) {
            return null;
        } while (j != -1) {
            String s = str.substring(i, j);
            ECRRerouteInfo ecr = new ECRRerouteInfo();
            ecr.parse(s);

            ret.put(ecr.getChannel(), ecr);
            i = j + "<ecrreroute/>".length();
            j = str.indexOf("<ecrreroute/>", i);
        }

        return ret;
    }

}
