/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.channel;

import java.net.*;
import java.util.*;

import fuegocore.message.*;
import fuegocore.message.protocol.*;
import fuegocore.message.service.*;
import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.filter.poset.*;
import fuegocore.notify.core.serialization.*;
import fuegocore.notify.core.server.*;
import fuegocore.notify.util.*;
import fuegocore.util.xas.*;

/**
 * This class represents the subscription information set of a single
 * event channel. Channel allows subscriptions, unsubscriptions,
 * handovers (subscription updates), and supports event routing.
 *
 */
public class EventChannel {
    private String type; // event channel type

    /** Maximum sequence number for published events */
    public static int MAX_COUNT = (int) Math.pow(2, 32) - 1;

    private int counter = 0; // event counter
    private int maxSubs = 100; // maximum number of subscriptions
    private int expiredNotification = 0;
    private int strayNotification = 0;
    private int historySize = 1000; // default history/log size
    private int subId = 0;

    /** true if temporary channel */
    private boolean temporary = false;

    /** true if a proxy channel */
    private boolean proxy = false;

    /** History vector */
    Vector history = new Vector();

    /** Filter poset for routing and matching */
    IFilterPoset pos = PosetService.get(); // get default implementation

    /** Filter poset for advertisements */
    IFilterPoset posA = PosetService.get(); // get default implementation

    /** Vector of subscriptions, this is for debug purposes. Set to null in order
     *  to save memory */
    Vector subscriptions = new Vector();

    /** Vector of quench subscriptions */
    Vector quench = new Vector();

    /**
     * Returns the subscription poset (mainly for debug)
     * @return IFilterPoset
     */
    public IFilterPoset getSubPoset() {
        return pos;
    }

    /**
     * Returns the advertisement poset (mainly for debug)
     * @return IFilterPoset
     */
    public IFilterPoset getAdvPoset() {
        return posA;
    }

    /**
     * Returns the set of active quench subscriptions
     * @return Vector
     */
    public synchronized Vector getQuench() {
        return quench;
    }

    /**
     * Returns the set of active quench subscriptions
     * @param iface String
     * @return Vector
     */
    public synchronized Vector getQuench(String iface) {
        Vector temps = new Vector();
        synchronized (quench) {
            for (int i = 0; i < quench.size(); i++) {
                Object o = quench.elementAt(i);
                if (o instanceof Template) {
                    Template t = (Template) o;
                    if (t.getInterface().equals(iface)) {
                        temps.add((Template) o);
                    }
                }
            }
        }
        return temps;
    }

    /**
     * Debug method to retrieve the current subscriptions at the event channel
     * @return Vector
     */
    public synchronized Vector getSubscriptions() {
        Vector all = pos.getAllNodes();
        Vector temps = new Vector();
        for (int i = 0; i < all.size(); i++) {
            Object o = all.elementAt(i);
            if (o instanceof Template) {
                temps.add((Template) o);
            }
        }
        return temps;
    }

    /**
     * Returns subscriptions for the given interface
     * @param iface the interface
     * @return Vector
     */
    public synchronized Vector getSubscriptions(String iface) {
        Vector all = pos.getAllNodes();
        Vector temps = new Vector();
        for (int i = 0; i < all.size(); i++) {
            Object o = all.elementAt(i);
            if (o instanceof Template) {
                Template t = (Template) o;
                if (t.getInterface().equals(iface)) {
                    temps.add((Template) o);
                }
            }
        }
        return temps;
    }

    /**
     * Returns subscriptions with the given type
     * @param type String
     * @return Vector
     */
    public synchronized Vector getSubscriptionsByType(String type) {
        Vector all = pos.getAllNodes();
        Vector temps = new Vector();
        for (int i = 0; i < all.size(); i++) {
            Object o = all.elementAt(i);
            if (o instanceof Template) {
                Template t = (Template) o;
                if (t.getType().equals(type)) {
                    temps.add((Template) o);
                }
            }
        }
        return temps;
    }

    /**
     * Debug method to retrieve the current advertisements at the event channel
     * @return Vector
     */
    public synchronized Vector getAdvertisements() {
        Vector all = posA.getAllNodes();
        Vector temps = new Vector();
        for (int i = 0; i < all.size(); i++) {
            Object o = all.elementAt(i);
            if (o instanceof Template) {
                temps.add((Template) o);
            }
        }
        return temps;
    }

    /**
     * Returns advertisements for the given interface
     * @param iface the interface
     * @return Vector
     */
    public synchronized Vector getAdvertisements(String iface) {
        Vector all = posA.getAllNodes();
        Vector temps = new Vector();
        for (int i = 0; i < all.size(); i++) {
            Object o = all.elementAt(i);
            if (o instanceof Template) {
                Template t = (Template) o;
                if (t.getInterface().equals(iface)) {
                    temps.add((Template) o);
                }
            }
        }
        return temps;
    }

    /**
     * Returns advertisements with the given type
     * @param type String
     * @return Vector
     */
    public synchronized Vector getAdvertisementsByType(String type) {
        Vector all = posA.getAllNodes();
        Vector temps = new Vector();
        for (int i = 0; i < all.size(); i++) {
            Object o = all.elementAt(i);
            if (o instanceof Template) {
                Template t = (Template) o;
                if (t.getType().equals(type)) {
                    temps.add((Template) o);
                }
            }
        }
        return temps;
    }

    /**
     * Returns unique interfaces
     * @return Vector
     */
    public synchronized Vector getInterfaces() {
        Vector all = pos.getAllNodes();
        Vector temps = new Vector();
        for (int i = 0; i < all.size(); i++) {
            Object o = all.elementAt(i);
            if (o instanceof Template) {
                Template t = (Template) o;
                if (!temps.contains(t.getInterface())) {
                    temps.add(t.getInterface());
                }
            }
        }
        return temps;
    }

    /**
     * Returns reference to event channel history
     * @return Vector
     */
    public Vector getHistory() {
        return history;
    }

    /**
     * Returns the name of the channel
     * @return String
     */
    public String getName() {
        return type;
    }

    /**
     * Add notification to history
     * @param n Notification
     */
    public void addHistory(Notification n) {
        // No history for proxies
        if (temporary && !proxy) {
            return;
        }

        if (history.size() < historySize) {
            history.add(n);
        } else {
            history.remove(0); // we remove the oldest
            history.add(n); // add a new one to the tail
        }
    }

    /**
     * One-shot pull functionality that pulls notification from history (uses template matching)
     * @param subscriptions  subscriptions the subscriptions to match, must not be null
     * @param count the number of notifications to pull, if 0 then all matching are returned
     * @return Vector
     */
    public Vector pull(Vector subscriptions, int count) {
        if (temporary && !proxy) { // proxies are not pullable
            return null;
        }

        int cc = 0;
        Vector ret = new Vector();
        if (subscriptions == null) {
            return null;
        }

        for (int i = 0; i < history.size(); i++) {
            if (count > 0 && cc >= count) {
                break;
            }
            Notification n = (Notification) history.elementAt(i);
            if (n == null) {
                continue;
            }
            for (int j = 0; j < subscriptions.size(); j++) {
                Template t2 = (Template) subscriptions.elementAt(j);
                if (t2 == null) {
                    continue;
                }
                if (t2.match(null, n)) {
                    ret.add(n);
                    cc++;
                    break;
                }
            }
        }
        return ret;
    }

    /**
     * One-shot pull functionality that pulls notification from history (uses order matching)
     * @param identifiers comma separated list of identifiers (order) to match, must not be null
     * @param count the number of notifications to pull, if 0 then all matching are returned
     * @return Vector
     */
    public Vector pull(String identifiers, int count) {
        if (temporary && !proxy) { // proxies are not pullable
            return null;
        }

        int cc = 0;
        Vector ret = new Vector();
        Vector ids = null;

        if (identifiers != null) {
            ids = new Vector();
            try {
                StringTokenizer tok = new StringTokenizer(identifiers, ",");
                if (tok != null) {
                    while (tok.hasMoreTokens()) {
                        Integer ij = new Integer((String) tok.nextToken().trim());
                        ids.add(ij);
                    }
                }
            } catch (Exception e) {
                ids = null;
            }
        } else {
            return null;
        }

        for (int i = 0; i < history.size(); i++) {
            if (count > 0 && cc >= count) {
                break;
            }
            Notification n = (Notification) history.elementAt(i);
            if (n == null) {
                continue;
            }
            for (int j = 0; j < ids.size(); j++) {
                Integer order = (Integer) ids.elementAt(j);
                if (n.getOrder() == order.intValue()) {
                    ret.add(n);
                    cc++;
                    break;
                }
            }
        }
        return ret;
    }

    /** Returns event channel information: event channel properties and subscriptions
     * @return the event channel description as an XML String
     */
    public String toString() {
        String ret = "";
        ret += "<historysize>" + historySize + "</historysize>\n";
        ret += "<maxsubs>" + maxSubs + "</maxsubs>\n";
        ret += "<counter>" + counter + "</counter>\n";
        ret += "<subscriptions>\n";
        Vector v = getSubscriptions();
        for (int i = 0; i < v.size(); i++) {
            Template t = (Template) v.elementAt(i);
            ret += t.toString();
        }
        ret += "</subscriptions>\n";
        return ret;
    }

    /**
     * Sets the history size
     * @param a int
     */
    public void setHistorySize(int a) {
        historySize = a;
    }

    /**
     * Gets the history size
     * @return int
     */
    public int getHistorySize() {
        return historySize;
    }

    /**
     *  Set temporary channel flag
     * @param b boolean
     */
    public void setTemporary(boolean b) {
        temporary = b;
    }

    /**
     * Get temporary channel flag
     * @return boolean
     */
    public boolean getTemporary() {
        return temporary;
    }

    /**
     * Set proxy mode
     * @param b boolean
     */
    public void setProxy(boolean b) {
        proxy = b;
    }

    /**
     * Get proxy mode status
     * @return boolean
     */
    public boolean getProxy() {
        return proxy;
    }

    /**
     * Create a new channel
     * @param type String
     * @param maxs int
     */
    public EventChannel(String type, int maxs) {
        this.type = type;
        maxSubs = maxs;
    }

    /**
     * Get a unique sequence number for published events
     * @return int
     */
    private synchronized int getSequence() {
        counter++;
        counter = counter % MAX_COUNT;
        return counter;
    }

    /**
     * Publish a notification in this channel. Updates the
     * sequence number. Proxy channels do not update sequence numbers
     * @param n Notification
     * @return boolean
     */
    public boolean publish(Notification n) {
        // Assign a new sequence number for the notification
        if (!proxy) { // proxies do not update sequence numbers
            n.setOrder(getSequence());
            // Add to history
        }
        addHistory(n);
        return route(n);
    }

    /**
     * Modify a subscription (used by merging)
     * @param t Template
     * @param old Template
     * @return boolean
     */
    public boolean modifySubscription(Template t, Template old) {
        Log.writeEC("Got template in modifySubscription: " + t.toString());

        // register to filter service
        if (t.getInterface() == null || old.getInterface() == null) {
            Log.writeEC("Interface was null");
            return false;
        }

        if (!t.getInterface().equals(old.getInterface())) {
            System.out.println(
                    "New interface and old interface are not the same!");
            return false;
        }

        Log.writeEC("Using interface: " + t.getInterface());
        int k = 0;
        try {
            // First remove old
            k = pos.remove(old, old.getInterface());
        } catch (PosetException e) {
            k = FilterForest.POSET_ERROR;
        }

        if (k == FilterForest.POSET_ERROR) {
            System.out.println("EC: remove failure in modifySubscription");
            return false;
        }

        if (pos.removable(t, t.getInterface())) {
            Log.writeEC("Template already subscribed for the given interface");
            Log.writeEC("Template: " + t.toString());
            return false;
        }
        try {
            // Then insert new
            pos.add(t, t.getInterface());
        } catch (PosetException e) {
            // put it back
            try {
                pos.add(old, old.getInterface());
            } catch (PosetException e2) {
                e2.printStackTrace();
            }
            return false;
        }

        // UPDATE ACCESS SERVERS FOR TEMPORARY CHANNELS
        if (temporary && !proxy) {
            sendMergedSetsToAccessServers(t.getInterface());
        }
        return true;
    }

    /**
     * Unsubscribe with an uncovered set of subscriptions
     * @param t Template
     * @param csub Vector
     * @return boolean
     */
    public boolean unsubscribe(Template t, Vector csub) {
        if (t == null || csub == null) {
            return false;
        }
        if (!unsubscribe(t)) {
            return false;
        }
        boolean ret = true;
        for (int i = 0; i < csub.size(); i++) {
            Template _t = (Template) csub.elementAt(i);
            if (!subscribe(_t)) {
                ret = false;
            }
        }
        return ret;
    }

    /**
     * Unadvertise with an uncovered set of advertisements
     * @param t Template
     * @param csub Vector
     * @return boolean
     */
    public boolean unadvertise(Template t, Vector csub) {
        if (t == null || csub == null) {
            return false;
        }
        if (!unadvertise(t)) {
            return false;
        }
        boolean ret = true;
        for (int i = 0; i < csub.size(); i++) {
            Template _t = (Template) csub.elementAt(i);
            if (!advertise(_t)) {
                ret = false;
            }
        }
        return ret;
    }

    /**
     * Unquench with an uncovered set of subscriptions
     * @param t Template
     * @param csub Vector
     * @return boolean
     */
    public boolean unquench(Template t, Vector csub) {
        if (t == null || csub == null) {
            return false;
        }
        if (!unquench(t)) {
            return false;
        }
        boolean ret = true;
        for (int i = 0; i < csub.size(); i++) {
            Template _t = (Template) csub.elementAt(i);
            if (!quench(_t)) {
                ret = false;
            }
        }
        return ret;
    }

    /**
     * Quench
     * @param t Template
     * @return boolean
     */
    public boolean quench(Template t) {
        if (t == null) {
            return false;
        }
        quench.add(t);
        return true;
    }

    /**
     * Unquench
     * @param t Template
     * @return boolean
     */
    public boolean unquench(Template t) {
        if (t == null) {
            return false;
        }
        if (!quench.remove(t)) {
            return false;
        }
        return true;
    }

    /**
     * Advertise
     * @param t Template
     * @return boolean
     */
    public boolean advertise(Template t) {
        if (t == null) {
            return false;
        }

        // register to filter service
        if (t.getInterface() == null) {
            Log.writeEC("Interface was null");
            return false;
        }
        Log.writeEC("Using interface: " + t.getInterface());

        if (posA.removable(t, t.getInterface())) {
            Log.writeEC("Template already subscribed for the given interface");
            Log.writeEC("Template: " + t.toString());
            return false;
        }
        try {
            posA.add(t, t.getInterface());
        } catch (PosetException e) {
            e.printStackTrace();
        }

        return true;
    }

    /**
     * Unadvertise
     * @param t Template
     * @return boolean
     */
    public boolean unadvertise(Template t) {
        if (t == null) {
            return false;
        }

        if (t.getInterface() == null) {
            Log.writeEC("Interface was null");
            return false;
        }

        Log.writeEC("Using interface: " + t.getInterface());

        if (!posA.removable(t, t.getInterface())) {
            Log.writeEC("Could not remove: " + t.toString());
            return false;
        }
        try {
            posA.remove(t, t.getInterface());
        } catch (PosetException e) {
            e.printStackTrace();
        }

        return true;
    }

    /**
     * Activate subscription
     * @param t Template
     * @return boolean
     */
    public boolean subscribe(Template t) {
        Log.writeEC("Got template in subscribe: " + t.toString());

        // register to filter service
        if (t.getInterface() == null) {
            Log.writeEC("Interface was null");
            return false;
        }
        Log.writeEC("Using interface: " + t.getInterface());

        if (pos.removable(t, t.getInterface())) {
            Log.writeEC("Template already subscribed for the given interface");
            Log.writeEC("Template: " + t.toString());
            return false;
        }
        try {
            pos.add(t, t.getInterface());
        } catch (PosetException e) {
            e.printStackTrace();
        }

        // UPDATE ACCESS SERVERS FOR TEMPORARY CHANNELS
        if (temporary && !proxy) {
            sendMergedSetsToAccessServers(t.getInterface());
        }
        return true;
    }

    /**
     * Terminate subscription
     * @param t Template
     * @return boolean
     */
    public boolean unsubscribe(Template t) {
        Log.writeEC("Got template in unsubscribe: " + t.toString());

        if (t.getInterface() == null) {
            Log.writeEC("Interface was null");
            return false;
        }

        Log.writeEC("Using interface: " + t.getInterface());

        if (!pos.removable(t, t.getInterface())) {
            Log.writeEC("Could not remove: " + t.toString());
            return false;
        }
        try {
            pos.remove(t, t.getInterface());
        } catch (PosetException e) {
            e.printStackTrace();
        }

        // UPDATE ACCESS SERVERS FOR TEMPORARY CHANNELS
        if (temporary && !proxy) {
            sendMergedSetsToAccessServers(t.getInterface());
        }
        return true;
    }

    /** Sends merged filter sets to corresponding proxy event channels
     *  at the access servers. The filter sets are sent using an
     *  asynchronous message.
     */
    public void sendMergedSetsToAccessServers() {
        sendMergedSetsToAccessServers(null, null);
    }

    /**
     * Sends merged filter sets to corresponding proxy event channels
     * @param exclude server to be excluded
     */
    public void sendMergedSetsToAccessServers(String exclude) {
        sendMergedSetsToAccessServers(exclude, null);
    }

    /** Sends merged filter sets to corresponding proxy event channels
     *  at the access servers. The filter sets are sent using an
     *  asynchronous message.
     * @param exclude the address to exclude (the source)
     * @param exclude2 the address to exclude (the second source, used for handover)
     */
    public void sendMergedSetsToAccessServers(String exclude, String exclude2) {
        Vector v = getSubscriptions(type);

        // Retrieve list of access servers
        EventServer eb = EventServer.get();

        Hashtable vh = new Hashtable();

        // We are assuming that access servers and resolutions servers share the port

        for (Enumeration k = eb.getRoutingEngine().getRouterManager().
                             getChannels().
                             elements();
                             k.hasMoreElements(); ) {
            URL host = (URL) k.nextElement();

            if (vh.get(host) != null || host.toString().indexOf(
                    fuegocore.notify.profile.CommProfile.event_server_id) != -1 ||
                (exclude != null && host.toString().indexOf(exclude) != -1) ||
                (exclude2 != null && host.toString().indexOf(exclude2) != -1)) {
                ; // ensure uniqueness and do not send to this server
            } else {
//                System.out.println("*** Sending to: " + host.getHost() + ":" +
//                                   host.getPort());
                sendsubscriptions(host.getHost() + ":" + host.getPort(), v);
                vh.put(host, "true");
            }
        }
    }

    public void removeAll() {
        pos.reset();
        // UPDATE ACCESS SERVERS FOR TEMPORARY CHANNELS
        if (temporary && !proxy) {
            sendMergedSetsToAccessServers();
        }
    }

    /**
     * Remove an interface
     * @param iface String
     */
    private void pruneInterface(String iface) {
        Vector all = this.getSubscriptions();
        for (int i = 0; i < all.size(); i++) {
            Template t = (Template) all.elementAt(i);
            if (t.getInterface().equals(iface)) {
                try {
                    pos.remove(t, t.getInterface());
                } catch (PosetException e) {
                    e.printStackTrace();
                }
            }
        }
    }

    /**
     * Reset subscriptions for the given interface
     * @param iface String
     * @return boolean
     */
    public boolean resetSubscriptions(String iface) {
        pruneInterface(iface);

        // UPDATE ACCESS SERVERS FOR TEMPORARY CHANNELS
        if (temporary && !proxy) {
            sendMergedSetsToAccessServers(iface);
        }
        return true;
    }

    /* Handover */
    public boolean handover(Vector acc, Vector rem, String iface, String old,
                            String channel) {
        Log.writeEC("Starting handover");

        if (iface == null || old == null) {
            return false;
        }

        fuegocore.notify.util.Log.writeEC("EC: removing:" + rem.size() +
                                          " for: " + old);
        for (int i = 0; i < rem.size(); i++) {
            Template t = (Template) rem.elementAt(i);
            fuegocore.notify.util.Log.writeEC(t.toString());
            try {
                pos.remove(t, old);
            } catch (PosetException e) {
                e.printStackTrace();
            }
        }

        fuegocore.notify.util.Log.writeEC(
                "--------------------------------------");

        // register to local filter service
        if (acc != null && iface != null) {
            fuegocore.notify.util.Log.writeEC("EC: adding:" + acc.size() +
                                              " for " +
                                              iface);
            for (int i = 0; i < acc.size(); i++) {
                Template t = (Template) acc.elementAt(i);
                fuegocore.notify.util.Log.writeEC(t.toString());
                if (!pos.removable(t, iface)) {
                    fuegocore.notify.util.Log.writeEC("Adding template " + i +
                            " for " +
                            iface);
                    fuegocore.notify.util.Log.writeEC("HA: " + t.toString());
                    try {
                        pos.add(t, iface);
                    } catch (PosetException e) {
                        e.printStackTrace();
                    }
                }
            }
        } else {
//            System.out.println("Subsc. for new server was null");
        }

        // UPDATE ACCESS SERVERS FOR TEMPORARY CHANNELS
        if (temporary && !proxy) {
            sendMergedSetsToAccessServers(old, iface);

        }
        return true;
    }

    /* Set subscriptions */
    public boolean setSubscriptions(String ch, String ac, ECRRerouteInfo ecr) {
        Log.writeEC("Setting event channel subscriptions for: " + ac);
        try {
            // First we reset the subscriptions for the given access server
            pruneInterface(ac);

            Vector acc = ecr.getActive();

            // The set may be null if there are no subscriptions
            if (acc != null) {
                fuegocore.notify.util.Log.writeEC("EC adding: " + acc.size() +
                                                  " subscriptions");

                for (int i = 0; i < acc.size(); i++) {
                    Template t = (Template) acc.elementAt(i);
                    if (!pos.removable(t, ac)) {
                        pos.add(t, ac);
                    }
                }
            } else {
 //               System.out.println("New subs were null for " + ac);
            }
        } catch (Exception e) {
            System.out.println("Exception in setSubscriptions");
            e.printStackTrace();
            return false;
        }

        // UPDATE ACCESS SERVERS FOR TEMPORARY CHANNELS
        if (temporary && !proxy) {
            sendMergedSetsToAccessServers(ac);
        }
        return true;
    }

    /**
     * Route an event coming from a transport to listeners based on
     * event type and template matching.
     * @param n Notification
     * @return boolean
     */
    public boolean route(Notification n) {
        boolean f = false;

        // check if the event is expired
        if (n == null || n.getDeadline() > 0 &&
            (n.getDeadline() - System.currentTimeMillis()) < 0) {
            Log.writeEC("Expired notification, dropping");
            expiredNotification++;
            return true; // not an error
        }

        int noti = 0;
        Vector v = (Vector) pos.match(n);
        Log.writeEC("EC: Matching notification: " + v.size());
        if (v != null && v.size() > 0) {
            for (int i = 0; i < v.size(); i++) {
                try {
                    String no = (String) v.elementAt(i);
                    Log.writeEC("Routing interface: " + i + " " + no);

                    if (no != null) {
                        noti++;
                        n.setInterface(no);
                        // send event to proper interface
                        send(n);
                    }
                } catch (Exception e) {
                    e.printStackTrace();
                    noti = -1;
                }
            } // for
        } // if
        if (noti == 0) {
            strayNotification++;
        }
        if (noti == -1) {
            return false;
        }
        return true;
    }

    /**
     * Internal message send method
     * @param n Notification
     */
    private void send(Notification n) {
        // open one-shot connection
        try {
            Log.writeEC("Adding message to outbound queue");

            // Fast local delivery
            if (n.getInterface().equals(
                    fuegocore.notify.profile.CommProfile.access_server_id)) {
                EventServer.get().route(n);
            } else {
                MessageTransportService mts = MessageTransportService.init();
                CodecIndustry.registerContentFactory(new EventCodecFactory());
                MessageFactory mf = mts.getFactory();
                fuegocore.message.service.Message m2 = mf.createMessage();

                m2.putData("saxdata", n);

                // get host and port
                String host = fuegocore.notify.util.Auxil.getHost(n.
                        getInterface());
                int port = fuegocore.notify.util.Auxil.getPort(n.getInterface());

                m2.setDestination(new Url(Url.Protocol.parse(fuegocore.notify.
                        profile.
                        UserProfile.DEF_MSG_PROTOCOL),
                                          "message",
                                          Url.toDestination(host, port)));

                MessageService ms = mts.getService();
                ms.sendMessage(m2);
            }
        } catch (Exception e) {
            Log.writeEC("Exception:" + e);
        }

    }

    private void sendsubscriptions(String dst, Vector v) {
        // open one-shot connection
        try {
            Log.writeEC("Adding proxy message to outbound queue");

            MessageTransportService mts = MessageTransportService.init();
            CodecIndustry.registerContentFactory(new EventCodecFactory());
            MessageFactory mf = mts.getFactory();
            fuegocore.message.service.Message m2 = mf.createMessage();
            m2.putData("saxdata", v);
            m2.putData("type", "proxy-channel-update");

            // get host and port
            String host = fuegocore.notify.util.Auxil.getHost(dst);
            int port = fuegocore.notify.util.Auxil.getPort(dst);

            m2.setDestination(new Url(Url.Protocol.parse(fuegocore.notify.
                    profile.
                    UserProfile.DEF_MSG_PROTOCOL),
                                      "message",
                                      Url.toDestination(host, port)));
            MessageService ms = mts.getService();
            ms.sendMessage(m2);
        } catch (Exception e) {
            Log.writeEC("Exception:" + e);
        }

    }

    public static void main(String[] args) {
        EventChannel ec = new EventChannel("test", 100);

        Template t1 = new Template();
        Filter f1 = new Filter("type", EventTriplet.TYPE_STRING,
                               "equ presence");
        Filter f2 = new Filter("number", EventTriplet.TYPE_INT,
                               "lte 10");
        t1.setInterface("koe");
        t1.addFilter(f1);
        t1.addFilter(f2);
        ec.subscribe(t1);

        Template t2 = new Template();
        Filter f3 = new Filter("type", EventTriplet.TYPE_STRING,
                               "equ presence");
        Filter f4 = new Filter("number", EventTriplet.TYPE_INT,
                               "gt 10");
        t2.addFilter(f3);
        t2.addFilter(f4);
        t2.setInterface("koe");
        ec.modifySubscription(t2, t1);

    }
}
