/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.channel;

import java.net.*;
import java.util.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.serialization.*;
import fuegocore.notify.util.*;

/**
 * Event Channel Server.
 * This class is responsible for managing a number of event channels.
 * Each event channel handles a particular type of event. Event channel
 * server is an active object listening to a specific port. All
 * update operations are done using SOAP RPC.
 */
public class EventChannelServer {
    URL server;
    int strayNotification = 0;
    int expiredNotification = 0;
    public static EventChannelServer es = null;

    /** hashtable of event channels on this server */
    public Hashtable event_channels;

    /** hashtable of event channels on this server */
    public Hashtable proxy_event_channels;

    /**
     * Get a reference to the Event Channel server (singleton)
     * @return EventChannelServer
     */
    public synchronized static EventChannelServer get() {
        if (es == null) {
            es = new EventChannelServer(80, "");
        }
        return es;
    }

    /**
     * Get a reference to the Event Channel server (singleton)
     * @param port int
     * @param config String
     * @return EventChannelServer
     */
    public synchronized static EventChannelServer get(int port, String config) {
        if (es == null) {
            es = new EventChannelServer(port, config);
        }
        return es;
    }

    /**
     *  Get a reference to the Event Channel server (singleton)
     * @param config String
     * @return EventChannelServer
     */
    public static EventChannelServer get(String config) {
        if (es == null) {
            es = new EventChannelServer(config);
        }
        return es;
    }

    /**
     * Constructor
     * @param port int
     * @param config String
     */
    public EventChannelServer(int port, String config) {
        event_channels = new Hashtable();
        initServer(config);
    }

    /**
     * Constructor
     * @param config String
     */
    public EventChannelServer(String config) {
        event_channels = new Hashtable();
        initServer(config);
    }

    /**
     * Sets the URL
     * @param url URL
     */
    public void setURL(URL url) {
        server = url;
    }

    /**
     * Returns the URL
     * @return URL
     */
    public URL getURL() {
        return server;
    }

    /**
     * Lookup event channel given its type
     * @param type String
     * @return EventChannel
     */
    public EventChannel lookup(String type) {
        if (event_channels == null || type == null) {
            return null;
        }
        return (EventChannel) event_channels.get(type);
    }

    /**
     * Lookup event channel given its type
     * @param type String
     * @return EventChannel
     */
    public EventChannel lookupProxy(String type) {
        if (proxy_event_channels == null || type == null) {
            return null;
        }
        return (EventChannel) proxy_event_channels.get(type);
    }

    /**
     * Add a new event channel
     * @param name String
     * @param ecs EventChannel
     */
    public synchronized void addEventChannel(String name, EventChannel ecs) {
        if (event_channels == null) {
            event_channels = new Hashtable();

        }
        String r = (String) event_channels.get(name);
        if (r == null) { // new channel
            event_channels.put(name, ecs);
        }
    }

    /**
     * Add a new event channel
     * @param name String
     * @param ecs EventChannel
     */
    public synchronized void addProxyEventChannel(String name, EventChannel ecs) {
        if (proxy_event_channels == null) {
            proxy_event_channels = new Hashtable();
        }
        String r = (String) proxy_event_channels.get(name);
        if (r == null) { // new channel
            proxy_event_channels.put(name, ecs);
        }
    }

    /**
     *  Configure server using a configuration file
     * @param file String
     */
    public void initServer(String file) {
        if (event_channels == null) {
            event_channels = new Hashtable();
        }
        try {
            System.out.println("Reading:" + file);
            ReadFile rf = new ReadFile(file);
            if (rf == null) {
                return;
            }
            // Extract contents
            String tmp = rf.readValue("count");
            int count = 0;
            if (tmp == null) {
                count = 0;
            } else {
                count = (int) (new Integer(tmp.trim())).intValue();

            }
            if (count < 0) {
                count = 0; // limits
            } else if (count > 1000) {
                count = 1000;

            }
            for (int i = 0; i < count; i++) {
                System.out.println("Reading Event Channel: " + i);
                String type = rf.readValue("type").trim();
                tmp = rf.readValue("maxsubs");
                int m = (int) (new Integer(tmp.trim())).intValue();
                EventChannel ec = new EventChannel(type, m);
                addEventChannel(type, ec);
            }

            rf.close();
        } catch (Exception e) {
            System.out.println("Exception in reading a file [" + file + "]:" +
                               e);
            e.printStackTrace();
        }
    }

    /**
     * Reset proxies
     */
    public void resetAllProxies() {
        for (Enumeration j = proxy_event_channels.elements(); j.hasMoreElements(); ) {
            EventChannel ec2 = (EventChannel) j.nextElement();
            if (ec2 == null) {
                continue;
            }
            ec2.removeAll();
        }
    }

    /**
     * Return a hashtable containing all the event channels on this server
     * @return Hashtable
     */
    public Hashtable getEventChannels() {
        return event_channels;
    }

    /**
     * Return a hashtable containing all the event channels on this server
     * @return Hashtable
     */
    public Hashtable getEventChannelProxies() {
        return proxy_event_channels;
    }

    /**
     * Creates an event channel for this resolution server
     * @param type String
     * @param properties String
     * @return boolean
     */
    public synchronized boolean createChannel(String type, String properties) {
        EventChannelParams ecp = new EventChannelParams();
        ecp.parse(properties);

        EventChannel ec = new EventChannel(type, ecp.getMaxSubscribers());
        ec.setTemporary(true);
        ec.setHistorySize(ecp.getEventHistorySize());
        addEventChannel(type, ec);
        return true;
    }

    /**
     * Creates a proxy event channel for this resolution server
     * @param type String
     * @param properties String
     * @return boolean
     */
    public synchronized boolean createProxyChannel(String type,
            String properties) {
        EventChannelParams ecp = new EventChannelParams();
        ecp.parse(properties);

        EventChannel ec = new EventChannel(type, ecp.getMaxSubscribers());
        ec.setProxy(true);
        ec.setTemporary(true);
        ec.setHistorySize(ecp.getEventHistorySize());
        addProxyEventChannel(type, ec);
        return true;
    }

    /**
     * Removes an event channel from this resolution server
     * @param type String
     * @return boolean
     */
    public synchronized boolean removeChannel(String type) {
        EventChannel ec = (EventChannel) event_channels.get(type);
        if (ec == null) {
            return false;
        }
        // Only temporary event channels may be removed
        if (!ec.getTemporary()) {
            return false;
        }
        event_channels.remove(type);
        ec.removeAll();
        ec = null;
        return true;
    }

    /**
     * Removes an event channel from this resolution server
     * @param type String
     * @return boolean
     */
    public synchronized boolean removeProxyChannel(String type) {
        EventChannel ec = (EventChannel) proxy_event_channels.get(type);
        if (ec == null) {
            return false;
        }
        // Only temporary event channels may be removed
        if (!ec.getTemporary()) {
            return false;
        }
        proxy_event_channels.remove(type);
        return true;
    }

    /** Pulls notifications from the given event channel. Pull does not work
     *  for proxy channels.
     *  @param type the event channel name
     *  @param temps the templates to match in string format
     *  @param count the number of notifications to pull
     *  @return Vector the matching notifications
     */
    public Vector pull(String type, String temps, int count) {
        EventChannel ec = (EventChannel) event_channels.remove(type);

        Vector vo = EventSerializationService.get("Template").demultiplex(temps);
        Vector v = ec.pull(vo, count);
        return v;
    }

    /** Pulls notifications from the given event channel. Pull does not
     *  work for proxy channels.
     *  @param type the event channel name
     *  @param temps the notification identifiers (order) to retrieve
     *  @param count the number of notifications to pull
     *  @return Vector the matching notifications
     */
    public Vector pullOrder(String type, String temps, int count) {
        EventChannel ec = (EventChannel) event_channels.remove(type);

        Vector v = ec.pull(temps, count);
        return v;
    }

    /**
     * Returns debug and log information pertaining to all event channels
     * @param reserved String
     * @return String
     */
    public String channelLog(String reserved) {
        String ret = "";

        for (Enumeration k = event_channels.elements(); k.hasMoreElements(); ) {
            EventChannel ec = (EventChannel) k.nextElement();
            ret += ec.toString();
        }
        return ret;
    }

    /**
     *  RPC publish operation
     * @param str String
     * @return boolean
     */
    public boolean publish(String str) {
        Notification n = new Notification(str);
        if (n == null) {
            Log.writeEC("Invalid notification");
            return false;
        }
        return publish(n);
    }

    /**
     * Publish
     * @param n Notification
     * @return boolean
     */
    public boolean publish(Notification n) {
        String type = n.getType();
        EventChannel ec = lookup(type);
        if (ec == null) {
            Log.writeEC("Event channel type lookup failed:" + type);
            return false;
        }
        return ec.publish(n); // handle publication
    }

    /**
     * Handover RPC handler
     * @param reserved String
     * @param iface String
     * @param reroute String
     * @param old String
     * @return boolean
     */
    public boolean handover(String reserved, String iface, String reroute,
                            String old) {
        Log.writeEC("EventChannelServer: Starting handover");

        // Demultiplex the incoming templates
        Hashtable hsh = ECRRerouteInfo.demultiplex(reroute);
        if (hsh == null) {
            return false;
        }
        for (Enumeration k = hsh.elements(); k.hasMoreElements(); ) {
            ECRRerouteInfo ecr = (ECRRerouteInfo) k.nextElement();
            if (ecr == null) {
                Log.writeEC("Invalid object");
                return false;
            }
            EventChannel ec = lookup(ecr.getChannel());
            if (ec == null) {
                Log.writeEC("Event channel type lookup failed:" +
                            ecr.getChannel());
                return false;
            }
            if (!ec.handover(ecr.getActive(), ecr.getRemoved(), iface, old,
                             ecr.getChannel())) {
                return false;
            }
        }
        return true;
    }

    /**
     * Set subscriptions RPC handler
     * @param channel String
     * @param ac String
     * @param ecrs String
     * @return boolean
     */
    public boolean setSubscriptions(String channel, String ac, String ecrs) {
        ECRRerouteInfo ecr = new ECRRerouteInfo();
        ecr.parse(ecrs);

        if (ecr == null) {
            System.out.println("Invalid object");
            Log.writeEC("Invalid object");
            return false;
        }
        EventChannel ec = lookup(channel);
        if (ec == null) {
            System.out.println("Local lookup failed");
            Log.writeEC("Event channel type lookup failed:" + channel);
            return false;
        }
        return ec.setSubscriptions(channel, ac, ecr);
    }

    /**
     * RPC subscribe operation
     * @param template String
     * @return boolean
     */
    public boolean subscribe(String template) {
        Template t = new Template(template);
        if (t == null) {
            Log.writeEC("Invalid template");
            return false;
        }
        return subscribe(t);
    }

    /**
     * Subscribe
     * @param t Template
     * @return boolean
     */
    public boolean subscribe(Template t) {
        EventChannel ec = lookup(t.getType());
        if (ec == null) {
            Log.writeEC("Event channel type lookup failed:" + t.getType());
            return false;
        }
        return ec.subscribe(t);
    }

    /**
     * Advertise
     * @param t Template
     * @return boolean
     */
    public boolean advertise(Template t) {
        EventChannel ec = lookup(t.getType());
        if (ec == null) {
            Log.writeEC("Event channel type lookup failed:" + t.getType());
            return false;
        }
        return ec.advertise(t);
    }

    /**
     * Quench
     * @param t Template
     * @return boolean
     */
    public boolean quench(Template t) {
        EventChannel ec = lookup(t.getType());
        if (ec == null) {
            Log.writeEC("Event channel type lookup failed:" + t.getType());
            return false;
        }
        return ec.quench(t);
    }

    /**
     * RPC modifySubscription operation
     * @param template String
     * @param old String
     * @return boolean
     */
    public boolean modifySubscription(String template, String old) {
        Template t = new Template(template);
        Template o = new Template(old);
        if (t == null || o == null) {
            Log.writeEC("Invalid template");
            return false;
        }
        EventChannel ec = lookup(t.getType());
        if (ec == null) {
            Log.writeEC("Event channel type lookup failed:" + t.getType());
            return false;
        }
        return ec.modifySubscription(t, o);
    }

    /**
     *  RPC subscribe operation
     * @param template String
     * @return boolean
     */
    public boolean unsubscribe(String template) {
        Template t = new Template(template);
        if (t == null) {
            Log.writeEC("Invalid template");
            return false;
        }
        return unsubscribe(t);
    }

    /**
     * Unsubscribe
     * @param t Template
     * @return boolean
     */
    public boolean unsubscribe(Template t) {
        EventChannel ec = lookup(t.getType());
        if (ec == null) {
            Log.writeEC("Event channel type lookup failed:" + t.getType());
            return false;
        }
        return ec.unsubscribe(t);

    }

    /**
     * Unadvertise
     * @param t Template
     * @return boolean
     */
    public boolean unadvertise(Template t) {
        EventChannel ec = lookup(t.getType());
        if (ec == null) {
            Log.writeEC("Event channel type lookup failed:" + t.getType());
            return false;
        }
        return ec.unadvertise(t);

    }

    /**
     * Unquench
     * @param t Template
     * @return boolean
     */
    public boolean unquench(Template t) {
        EventChannel ec = lookup(t.getType());
        if (ec == null) {
            Log.writeEC("Event channel type lookup failed:" + t.getType());
            return false;
        }
        return ec.unquench(t);

    }

    /**
     * Unadvertise
     * @param t Template
     * @return boolean
     */
    public boolean unadvertise(Template t, Vector csub) {
        EventChannel ec = lookup(t.getType());
        if (ec == null) {
            Log.writeEC("Event channel type lookup failed:" + t.getType());
            return false;
        }
        return ec.unadvertise(t, csub);
    }

    /**
     * Unquench
     * @param t Template
     * @return boolean
     */
    public boolean unquench(Template t, Vector csub) {
        EventChannel ec = lookup(t.getType());
        if (ec == null) {
            Log.writeEC("Event channel type lookup failed:" + t.getType());
            return false;
        }
        return ec.unquench(t, csub);
    }

    /**
     * Unsubscribe
     * @param t Template
     * @return boolean
     */
    public boolean unsubscribe(Template t, Vector csub) {
        EventChannel ec = lookup(t.getType());
        if (ec == null) {
            Log.writeEC("Event channel type lookup failed:" + t.getType());
            return false;
        }
        return ec.unsubscribe(t, csub);
    }
}
