/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.channel;

import java.util.*;

import fuegocore.message.service.*;
import fuegocore.notify.core.*;
import fuegocore.notify.core.serialization.*;
import fuegocore.util.xas.*;

/**
 *  Event Channel Server Handler.
 *  Used for handling Axis RPC (Application scope).
 */
public class EventChannelServerHandler implements RpcReceiver,
        fuegocore.message.service.MessageReceiver {
    EventChannelServer ecs;

    public EventChannelServerHandler() {
        ecs = EventChannelServer.get(); // get reference
    }

    public EventChannelServerHandler(String config) {
        ecs = EventChannelServer.get(config); // get reference
    }

    /** Incoming async messages
     * @param msg the Message to process
     * @return Message */
    public fuegocore.message.service.Message invoke(fuegocore.message.service.
            Message msg) {

        try {
            EventSequence ses = (EventSequence) msg.getData("saxdata");
            String type_ = (String) msg.getData("type");
            String op = (String) msg.getData("operation");

            // Process proxy update messages
            if (type_ != null && type_.equalsIgnoreCase("access-server-msg")) {
                if (op == null) {
                    System.out.println("Op was null for msg: " + msg);
                } else if (op.equals("publish")) {
                    Notification n = new Notification(ses);
                    ecs.publish(n);
                } else if (op.equals("processSubAdvQuench")) {
                    Template t = new Template(ses);
                    if (t.getAdvertisement()) {
                        ecs.advertise(t);
                    } else if (t.getQuench()) {
                        ecs.quench(t);
                    } else {
                        ecs.subscribe(t);
                    }
                } else if (op.equals("removeSubAdvQuenchOneshot")) {
                    Template t = new Template(ses);
                    if (t.getAdvertisement()) {
                        ecs.unadvertise(t);
                    } else if (t.getQuench()) {
                        ecs.unquench(t);
                    } else {
                        ecs.unsubscribe(t);
                    }
                } else if (op.equals("removeSubAdvQuench")) {
                    EventSequence ses2 = (EventSequence) msg.getData("saxdata2");
                    Vector v = EventSerializationService.fromEventSequence((
                            EventSequence) ses2);
                    Template t = new Template(ses);
                    if (t.getAdvertisement()) {
                        ecs.unadvertise(t, v);
                    } else if (t.getQuench()) {
                        ecs.unquench(t, v);
                    } else {
                        ecs.unsubscribe(t, v);
                    }
                } else if (op.equals("handover")) {
                }
            } else {
                System.out.println("Unknown message: " + msg);
            }
        } catch (Exception e) {
            fuegocore.notify.util.Log.write("Error during notification route: " +
                                            e);
            e.printStackTrace();
        }
        return null;
    }

    /**
     * SOAP RPC handler for event channels
     * @param method String
     * @param arguments NameValuePair[]
     * @return Object
     */
    public Object invoke(String method, NameValuePair[] arguments) {
        //System.out.println("Method " + method + " called");

        String arg1 = null, arg2 = null, arg3 = null, arg4 = null;
        String address = null;
        String origin = null;
        int count = 0;
        boolean flag = false;

        for (int i = 0; i < arguments.length; i++) {
            if (arguments[i].getName().equals("arg1") ||
                arguments[i].getName().equals("text")) {
                arg1 = (String) arguments[i].getValue();
            } else if (arguments[i].getName().equals("arg2")) {
                arg2 = (String) arguments[i].getValue();
            } else if (arguments[i].getName().equals("arg3")) {
                arg3 = (String) arguments[i].getValue();
            } else if (arguments[i].getName().equals("arg4")) {
                arg4 = (String) arguments[i].getValue();
            } else if (arguments[i].getName().equals("count")) {
                try {
                    Integer ii = (Integer) arguments[i].getValue();
                    count = ii.intValue();
                } catch (Exception e) {
                    e.printStackTrace();
                    return "failure";
                }
            } else if (arguments[i].getName().equals("flag")) {
                try {
                    Boolean b = (Boolean) arguments[i].getValue();
                    flag = b.booleanValue();
                } catch (Exception e) {
                    e.printStackTrace();
                    return "failure";
                }
            } else if (arguments[i].getName().equals(fuegocore.message.service.
                    PropertyNames.MESG_DATA_SENDER)) {
                address = (String) arguments[i].getValue();
            } else if (arguments[i].getName().equals(fuegocore.message.service.
                    PropertyNames.MESG_DATA_ORIGIN)) {
                origin = (String) arguments[i].getValue();
            }
        }

        if (arguments == null) {
            return new Boolean(false);
        }
        if (method.equals("publish")) {
            return new Boolean(publish(arg1));
        } else if (method.equals("subscribe")) {
            return new Boolean(subscribe(arg1));
        } else if (method.equals("unsubscribe")) {
            return new Boolean(unsubscribe(arg1));
        } else if (method.equals("handover")) {
            return new Boolean(handover(arg1, arg2, arg3, arg4));
        } else if (method.equals("setsubscriptions")) {
            return new Boolean(ecs.setSubscriptions(arg1, arg2, arg3));
        } else if (method.equals("modifysubscription")) {
            return new Boolean(ecs.modifySubscription(arg1, arg2));
        } else if (method.equals("createChannel")) {
            return new Boolean(ecs.createChannel(arg1, arg2));
        } else if (method.equals("removeChannel")) {
            return new Boolean(ecs.removeChannel(arg1));
        } else if (method.equals("channelLog")) {
            return new Boolean(ecs.channelLog(arg1));
        } else if (method.equals("pullHistory")) {
            EventSequence evs = null;
            try {
                evs = EventSerializationService.toEventSequence(
                        ecs.pull(arg1, arg2, count));
            } catch (Exception e) {
                e.printStackTrace();
                evs = null;
            }
            return evs;
        } else if (method.equals("pullHistoryOrder")) {
            EventSequence evs = null;
            try {
                evs = EventSerializationService.toEventSequence(
                        ecs.pullOrder(arg1, arg2, count));
            } catch (Exception e) {
                e.printStackTrace();
                evs = null;
            }
            return evs;
        }
        return "";
    }

    /**
     * RPC publish operation
     * @param str String
     * @return boolean
     */
    public boolean publish(String str) {
        return ecs.publish(str);
    }

    /**
     * RPC subscribe operation
     * @param template String
     * @return boolean
     */
    public boolean subscribe(String template) {
        return ecs.subscribe(template);
    }

    /**
     * RPC subscribe operation
     * @param template String
     * @return boolean
     */
    public boolean unsubscribe(String template) {
        return ecs.unsubscribe(template);
    }

    /**
     * RPC handover operation
     * @param channel String
     * @param iface String
     * @param erc String
     * @param old String
     * @return boolean
     */
    public boolean handover(String channel, String iface, String erc,
                            String old) {
        return ecs.handover(channel, iface, erc, old);
    }
}
