package fuegocore.notify.core.client;

import java.io.*;
import java.util.*;

import fuegocore.message.service.Message;
import fuegocore.message.service.MessageFactory;
import fuegocore.message.service.LightMessageService;
import fuegocore.message.service.MessageReceiver;
import fuegocore.message.protocol.MessageProtocol;
import fuegocore.message.protocol.Url;
import fuegocore.message.service.NameValuePair;
import fuegocore.message.service.RpcCall;

import fuegocore.util.Util;
import fuegocore.util.xas.Event;
import fuegocore.util.xas.EventSequence;
import fuegocore.util.xas.EventSerializer;
import fuegocore.util.xas.EventList;
import fuegocore.util.xas.XasUtil;
import fuegocore.util.xas.TypedXmlSerializer;
import fuegocore.util.xas.CodecFactory;

import fuegocore.notify.core.*;
import fuegocore.notify.core.session.*;
import fuegocore.notify.core.channel.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.serialization.*;
import fuegocore.message.MessageTransportService;
import fuegocore.message.encoding.XebuCodecFactory;
import fuegocore.notify.util.TripletTable;

/**
 * <p>Title: Fuego Core Rendezvous-Notify</p>
 * <p>Description:
 * EventConnection represents the client-side event bus.
 * All events are passed as local method invocations.
 * We assume that the client and server have synchronized
 * clocks.<br>
 *
 * To do:
 *  - support for multiple sessions. For example: close session
 * does not remove subscriptions locally.
 * - moveSession is only for special cases, typically the server detects
 *   client relocation and moves the reopened session from the old location
 * - Currently multiple connections are not supported, but it is
 *  easy to add support for them, number message interfaces and update
 * server senderThread to send to correct interface.
 *  <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Helsinki Institute for Information Technology</p>
 * @author Sasu Tarkoma
 * @version 1.0
 */
public class EventConnection {
  int strayNotification = 0;
  int expiredNotification = 0;
  public static long CLIENT_TIMEOUT = 40000; // client time-out
  public static String REOPEN = "_reopen";
  public static String OP_FAILURE = "failure";

  int subNumber = 0; // identifier for subscriptions

  // local interfaces, with multiple sessions this is session specific
  public Hashtable localSubs = new Hashtable();

  // reference
  static EventConnection eb = null;

  private LightMessageService lms;
  private MessageProtocol mp;
  private MessageFactory mf;

  public String conHost = "";
  public int conPort = 80;

  /** Client user id, used in identifying sessions */
  public static String name = "mec_test_name";

  /** Default Fuego protocol */
  public static String DEF_MSG_PROTOCOL = "mop";

  /** Last server that was used (used in handover management) */
  public static String last_server = "-";
  SimpleMessageProtocol simple = null;

  public int NOTIFY_SIMPLEMESSAGE_PROTOCOL = 0;
  public int NOTIFY_FUEGO_PROTOCOL = 1;

  public int NOT_ACTIVE = 0;
  public int ACTIVE = 1;
  public int FAILURE = 2;
  public int SUCCESS = 3;
  public int SIMPLEMESSAGE_PROTOCOL_STATE = NOT_ACTIVE;

  /** simple message protocol is a very lightweight without any
   *  bells and whistles. Just one blocking thread (for one session) and
   *  operations using a blocking request thread
   */
  // public int protocol = NOTIFY_FUEGO_PROTOCOL;
  public int protocol = NOTIFY_SIMPLEMESSAGE_PROTOCOL;

  /**
   * Returns the name of the last server.
   * @return String
   */
  public static String getLastServer() {
    return last_server;
  }

  /**
   * Sets the name of the last server.
   * @param s String
   */
  public static void setLastServer(String s) {
    last_server = s;
  }

  /**
   *  Returns the name of the user.
   * @return String
   */
  public static String getName() {
    return name;
  }

  /**
   * Sets the name of the user.
   * @param s String
   */
  public static void setName(String s) {
    name = s;
  }

  /* ----------------------------------------------------------------*/

  /**
   * Set the host and port
   * @param h String
   * @param p int
   */
  public void setHostPort(String h, int p) {
    conHost = h;
    conPort = p;
  }

  /**
   * Get host
   * @return String
   */
  public String getHost() {
    return conHost;
  }

  /**
   * Get port
   * @return int
   */
  public int getPort() {
    return conPort;
  }

  /* ----------------------------------------------------------------*/
  /**
   * Start an emulated push connection implemented using pull (midp http)
   * @param sessionId String
   */
  public void startPullConnection(String sessionId) {
    if (protocol != NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
      return;
    }
    if (simple != null) {
      System.out.println("Pull connection already active for:" +
                         simple.getSessionId());
      return;
    }
    simple = new SimpleMessageProtocol(getHost(), getPort(), sessionId);
    simple.start();
  }

  /**
   * Stop pull connection
   * @param sessionId String
   */
  public void stopPullConnection(String sessionId) {
    if (protocol != NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
      return;
    }
    if (simple == null) {
      System.out.println("No active pull connection");
      return;
    }
    simple.close();
  }

  /* ----------------------------------------------------------------*/

  /** Constructor. Registers RPC and async message listener */
  public EventConnection() {

    if (protocol != NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
      try {
          /*
        XebuCodecFactory.init();
        MessageTransportService mts = MessageTransportService.init();
        lms = mts.getService();
        mp = mts.getProtocol();
        mf = mts.getFactory();
        lms.registerReceiver("message", this);
*/
      }
      catch (Exception ex) {
        ex.printStackTrace();
      }
    }
  }

  /**
   * Logging method
   * @param s String
   */
  public void logPrint(String s) {
    //   System.out.println(s);
  }

  /** Get reference to the event bus.
   *  Implements the singleton pattern
   *  @return EventConnection the handle to the event bus
   */
  public static EventConnection get() {
    if (eb == null) {
      eb = new EventConnection();
    }
    return eb;
  }

  /**
   *  RPC call
   * @param op String
   * @param o1 Object
   * @param o2 Object
   * @throws Exception
   * @return Object
   */
  public Object invoke(String op, Object o1, Object o2) throws Exception {
    return invoke(op, o1, o2, null, null, null);
  }

  /**
   * RPC call
   * @param op String
   * @param o1 Object
   * @param o2 Object
   * @param o3 Object
   * @throws Exception
   * @return Object
   */
  public Object invoke(String op, Object o1, Object o2, Object o3) throws
      Exception {
    return invoke(op, o1, o2, o3, null, null);
  }

  /**
   * RPC call
   * @param op String
   * @param o1 Object
   * @param o2 Object
   * @param o3 Object
   * @param o4 Object
   * @throws Exception
   * @return Object
   */
  public Object invoke(String op, Object o1, Object o2, Object o3, Object o4) throws
      Exception {
    return invoke(op, o1, o2, o3, o4, null);
  }

  /**
   * RPC call
   * @param op String
   * @param o1 Object
   * @param o2 Object
   * @param o3 Object
   * @param o4 Object
   * @param o5 Object
   * @throws Exception
   * @return Object
   */
  public Object invoke(String op, Object o1, Object o2, Object o3, Object o4,
                       Object o5) throws Exception {

    Object ret = null;

    Message message = mf.createMessage();

    logPrint("Creating message");

    message.putData("op", op);
    if (o1 != null) {
      message.putData("attr1", o1);
    }
    if (o2 != null) {
      message.putData("attr2", o2);
    }
    if (o3 != null) {
      message.putData("attr3", o3);
    }
    if (o4 != null) {
      message.putData("attr4", o4);
    }
    if (o5 != null) {
      message.putData("attr5", o5);

    }
    Url url = new Url(Url.Protocol.parse(DEF_MSG_PROTOCOL), "messagerpc",
                      Url.toDestination(getHost(), getPort()));

    message.setDestination(url);

    try {
      Message rmsg = lms.sendSynch(message, CLIENT_TIMEOUT);
      ret = (Object) rmsg.getData("data");
    }
    catch (fuegocore.message.service.TimeoutException e) {
      ret = null;
    }

    return ret;
  }

  /* ----------------------------------------------------------------*/
  /* PUBLISH/SUBSCRIBE RPC OPERATIONS 							   */
  /* ----------------------------------------------------------------*/

  /**
   * Subscribes the given ITemplate locally+externally (if applicable)
   * Using the given subscription identifier (this used to recover state
   * after the failure of the client side system).
   *
   * Set/get sessions using ITemplate setSessionId/GetSessionId.
   * Set/get subscription id using ITemplate setSubId/getSubId
   *
   * @param t the ITemplate to subscribe
   * @param id the subscription identifier
   * @return String subscription identifier
   */
  public String subscribe(ITemplate t, String id) {
    boolean ret = false;

    if (!t.getScope()) {
      t.setSubId(id);

      // set client device id
      if (t.getSessionId() != null) {
        t.setInterface(t.getSessionId());
      }
      else {
        t.setInterface(getName());
      }

      Boolean reply = null;
      try {
        try {
          if (protocol == NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
            String sret = SimpleMessageProtocol.call("subscribe",
                t.getInterface(), t.toString(), false, 0);
            if (sret != null && sret.indexOf("failure") == -1) {
              reply = new Boolean(true);
            }
            else {
              reply = new Boolean(false);
            }
          }
          else {
            reply = (Boolean) invoke("subscribe",
                                     EventSerializationService.toEventSequence(
                t),
                                     null, null);
          }
        }
        catch (Exception e) {
          reply = new Boolean(false);
          e.printStackTrace();
        }

        ret = reply.booleanValue();

        logPrint("*************** Reply was: " + ret);

        if (ret) {
          // store interface reference
          localSubs.put(id, t);
        }
        return id;
      }
      catch (Exception e) {
        e.printStackTrace();
      }
      return null;
    }
    else {
      // store interface reference
      id = t.getSubId();
      if (id == null) {
        id = "" + t.getSource().hashCode();
        t.setSubId(id);
      }
      localSubs.put(id, t);
      return id;
    }
  }

  public void resetSubscriptions() {
    localSubs.clear();
  }

  /**
   * Subscribes the given ITemplate locally + externally (if applicable)
   * Set/get sessions using ITemplate setSessionId/GetSessionId.
   * Set/get subscription id using ITemplate setSubId/getSubId
   * @param t the ITemplate to subscribe
   * @return String subscription identifier
   */
  public String subscribe(ITemplate t) {
    boolean ret = false;

    String id = t.getSubId();

    logPrint("Subscribing ITemplate: " + t.toString() +
             " scope:" + t.getScope());

    if (!t.getScope()) {
      subNumber++;

      // set client device id
      if (t.getSessionId() != null) {
        t.setInterface(t.getSessionId());
      }
      else {
        t.setInterface(getName());
      }

      if (t.getSubId() == null || t.getSubId().length() < 1) {
        // Create domain wide unique identifier
        id = t.getInterface() + (t.getSource().hashCode() + ":" + subNumber);
        t.setSubId(id);
      }

      try {
        Boolean reply = null;
        try {
          if (protocol == NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
            String sret = SimpleMessageProtocol.call("subscribe",
                t.getInterface(), t.toString(), false, 0);
            if (sret != null && sret.indexOf("failure") == -1) {
              reply = new Boolean(true);
            }
            else {
              reply = new Boolean(false);
            }
          }
          else {
            reply = (Boolean) invoke("subscribe",
                                     EventSerializationService.toEventSequence(
                t),
                                     null, null);
          }
        }
        catch (Exception e) {
          reply = new Boolean(false);
        }

        if (reply != null) {
          ret = reply.booleanValue();

        }
        logPrint("*************** Reply was: " + ret);

        if (ret) {
          // store interface reference
          localSubs.put(id, t);
        }
      }
      catch (Exception e) {
        e.printStackTrace();
      }
      if (ret) {
        System.out.println("Returning id: " + id);
        return id;
      }
      else {
        return null;
      }
    }
    else {
      // store interface reference
      if (id == null) {
        id = "" + t.getSource().hashCode();
        t.setSubId(id);
      }
      localSubs.put(id, t);
      return id;
    }
  }

  /** Unsubscribe an event.
   * @param subID the subscription identifier to unsubscribe
   * @return boolean true if success
   */
  public boolean unsubscribe(String subID) {
    logPrint("Looking up ITemplate");

    if (subID == null) {
      return false;
    }

    ITemplate t = (ITemplate) localSubs.get(subID);
    if (t == null) {
      return false;
    }
    return unsubscribe(t);
  }

  /* Unsubscribe an event.
   * @param ITemplate the ITemplate to unsubscribe
   * @return boolean true if success
   */
  public boolean unsubscribe(ITemplate t) {
    boolean ret = false;
    // we should check that it is subscribed...
    String id = t.getSubId();

    if (!t.getScope()) {
      // set client device id
      if (t.getSessionId() != null) {
        t.setInterface(t.getSessionId());
      }
      else {
        t.setInterface(getName());

      }
      try {
        Boolean reply = null;
        try {
          if (protocol == NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
            String sret = SimpleMessageProtocol.call("unsubscribe",
                t.getInterface(), t.toString(), false, 0);
            if (sret != null && sret.indexOf("failure") == -1) {
              reply = new Boolean(true);
            }
            else {
              reply = new Boolean(false);
            }
          }
          else {
            reply = (Boolean) invoke("unsubscribe",
                                     EventSerializationService.toEventSequence(
                t),
                                     null, null);
          }
        }
        catch (Exception e) {
          reply = new Boolean(false);
        }

        ret = reply.booleanValue();

        logPrint("Reply was: " + ret);

        if (ret) {
          if (t.getFilters() != null) {
            localSubs.remove(id);
          }
        }
      }
      catch (Exception e) {
        e.printStackTrace();
      }
    }
    else {
      if (t.getFilters() != null) {
        localSubs.remove(id);
      }
      ret = true;
    }
    return ret;
  }

  /** Publish an event
   *  @param n the notification to publish
   *  @return true if the operation succeeded
   */
  public boolean publish(Notification n) {
    boolean ret = false;

    if (!n.getHideSource() && n.getAddress() == null) {
      n.setAddress(getName());
    }

    if (!n.getScope()) { // external event
      try {
        Boolean reply = null;
        try {
          if (protocol == NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
            String sret = SimpleMessageProtocol.call("publish", "", n.toString(), false,
                0);
            if (sret != null && sret.indexOf("failure") == -1) {
              reply = new Boolean(true);
            }
            else {
              reply = new Boolean(false);
            }
          }
          else {
            reply = (Boolean) invoke("publish",
                                     EventSerializationService.toEventSequence(
                n),
                                     null, null);
          }
        }
        catch (Exception e) {
          reply = new Boolean(false);
        }

        ret = reply.booleanValue();

        logPrint("Reply was: " + ret);
      }
      catch (Exception e) {
        e.printStackTrace();
      }
      if (!ret) {
        return false; // break here
      }
      else {
        return true;
      }
    }
    // Check local subscriptions
    else if (n.getScope()) {
      logPrint("Sending locally");
      route(n);
    }
    return true;
  }

  /*** Aggregate functionality */

  /** Publish a list of events (external)
   *  @param v a vector of notifications to publish
   *  @return String ("failure " + offset in the vector if publish failed)
   */
  public String publishAll(Vector v) {
    String reply = null;

    if (v == null) {
      return null;
    }
    for (int i = 0; i < v.size(); i++) {
      Notification n = (Notification) v.elementAt(i);
      if (!n.getHideSource() && n.getAddress() == null) {
        n.setAddress(getName());
      }
    }

    try {
      try {
        if (protocol == NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
          reply = SimpleMessageProtocol.call("publishAll", "",
                                             EventSerializationService.get(
              "Notification").multiplex(v), false, 0);
        }
        else {
          reply = (String) invoke("publishAll",
                                  EventSerializationService.toEventSequence(v),
                                  null, null);
            }
      }
      catch (Exception e) {
        reply = null;
      }
    }
    catch (Exception e) {
      e.printStackTrace();
    }
    return reply;
  }

  /**
   * Subscribes the given ITemplates externally
   * @param v the ITemplates to subscribe
   * @return String (failure + vector offset if subscribe failed)
   */
  public String subscribeAll(Vector v) {
    String reply = null;

    if (v == null) {
      return null;
    }
    for (int i = 0; i < v.size(); i++) {
      ITemplate t = (ITemplate) v.elementAt(i);
      subNumber++;
      // set client device id
      if (t.getSessionId() != null) {
        t.setInterface(t.getSessionId());
      }
      else {
        t.setInterface(getName());
      }

      if (t.getSubId() == null || t.getSubId().length() < 1) {
        // Create domain wide unique identifier
        String id = t.getInterface() +
            (t.getSource().hashCode() + ":" + subNumber);
        t.setSubId(id);
      }
    }

    try {

      try {
        if (protocol == NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
          reply = SimpleMessageProtocol.call("subscribeAll", "",
                                             EventSerializationService.get(
              "Template").multiplex(v), false, 0);
        }
        else {
          reply = (String) invoke("subscribeAll",
                                  EventSerializationService.toEventSequence(v),
                                 null, null);
        }
      }
      catch (Exception e) {
        reply = null;
      }

      if (reply != null && reply.indexOf(OP_FAILURE) == -1) {
        // store interface reference
        for (int i = 0; i < v.size(); i++) {
          ITemplate t = (ITemplate) v.elementAt(i);
          // register to filter service
          if (t.getFilters() != null) {
            localSubs.put(t.getSubId(), t);
          }
        }
      }
    }
    catch (Exception e) {
      e.printStackTrace();
    }
    return reply;
  }

  /** Unsubscribe a vector of ITemplates
   *  @param v the ITemplates to unsubscribe
   *  @return string (failure + offset if it failed)
   */
  public String unsubscribeAll(Vector v) {
    String reply = null;

    if (v == null) {
      return null;
    }

    try {
      try {
        if (protocol == NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
          reply = SimpleMessageProtocol.call("unsubscribeAll", "",
                                             EventSerializationService.get(
              "Template").multiplex(v), false, 0);
        }
        else {
          reply = (String) invoke("unsubscribeAll",
                                  EventSerializationService.toEventSequence(v),
                                  null, null);

        }
      }
      catch (Exception e) {
        reply = null;
      }

      // We remove the ITemplates locally even if the unsubscribe failed
      // given that the unsubscribe was processed
      if (reply != null) {
        for (int i = 0; i < v.size(); i++) {
          ITemplate t = (ITemplate) v.elementAt(i);
          if (t.getFilters() != null) {
            localSubs.remove(t.getSubId());
          }
        }
      }
    }
    catch (Exception e) {
      e.printStackTrace();
    }
    return reply;
  }

  /* ----------------------------------------------------------------*/
  /* ----------------------------------------------------------------*/
  /* EVENT CHANNEL MANAGEMENT OPERATIONS */

  /** Create a new event channel
   *  @param channel the channel name
   *  @param params the event channel parameters
   *  @return true if the operation succeeded
   */
  public boolean createEventChannel(String channel, String params) {
    Boolean reply = null;
    try {
      try {
        if (protocol == NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
          String sret = SimpleMessageProtocol.call("createEventChannel",
              channel, params, false, 0);
          if (sret != null && sret.indexOf("failure") == -1) {
            reply = new Boolean(true);
          }
          else {
            reply = new Boolean(false);
          }
        }
        else {
          reply = (Boolean) invoke("createEventChannel", channel, params);
        }
      }
      catch (Exception e) {
        reply = null;
      }
      logPrint("Reply was: " + reply);
    }
    catch (Exception e) {
      e.printStackTrace();
    }

    return reply.booleanValue();
  }

  /** Remove an event channel
   *  @param channel the channel name to be removed
   *  @return true if the operation succeeded
   */
  public boolean removeEventChannel(String channel) {
    Boolean reply = null;
    try {
      try {
        if (protocol == NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
          String ret = SimpleMessageProtocol.call("removeEventChannel", channel,
                                                  "", false, 0);
          if (ret != null && ret.indexOf("failure") == -1) {
            reply = new Boolean(true);
          }
          else {
            reply = new Boolean(false);
          }
        }
        else {
          reply = (Boolean) invoke("removeEventChannel", channel, null);
        }
      }
      catch (Exception e) {
        reply = null;
      }
      logPrint("Reply was: " + reply);
    }
    catch (Exception e) {
      e.printStackTrace();
    }

    return reply.booleanValue();
  }

  /* ----------------------------------------------------------------*/
  /* ----------------------------------------------------------------*/


  /* ----------------------------------------------------------------*/
  /* ----------------------------------------------------------------*/

  /** Open session.
   * @param id session identifier
   * @param sp the SessionParams object
   * @return the session identifier
   */
  public String openSession(String id, SessionParams sp) {
    String reply = null;
    logPrint("Open session to: " + getHost() +
             " port: " + getPort());
    try {
      if (protocol == NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
        reply = SimpleMessageProtocol.call("openSession", id, sp.toString(), false,
                                           0);
      }
      else {
        reply = (String) invoke("openSession", id, getName(), getLastServer(),
                                sp.toString());
      }
      logPrint("Reply was: " + reply);
    }
    catch (Exception e) {
      e.printStackTrace();
      reply = null;
    }

    if (reply != null && reply.indexOf(OP_FAILURE) != -1) {
      return null;
    }
    else {
      setLastServer(getHost() + ":" + getPort());
    }

    return reply;
  }

  /* This helper method returns true if the session was reopened
   * @param s the session identifier
   * @return true if the operation succeeded
   */
  public boolean reopenedSession(String s) {
    if (s == null) {
      return false;
    }
    if (s.indexOf(REOPEN) != -1) {
      return true;
    }
    return false;
  }

  /** Pull notifications from the given session using the ITemplates
   * and deliver them locally to the correct subscribers. This supports
   * server-side filtering so client-side filtering is not necessary
   * @param id the session identifier
   * @param ITemplates Vector containing the ITemplates
   * @param count the number of notifications to pull
   * @param destroy destructive receipt? true destroys notifications from the session
   * @return the number of pulled and delivered events
   */
  public int pullAndDeliver(String id, Vector ITemplates, int count,
                            boolean destroy) {
    int r = 0;

    Vector v = pull(id, ITemplates, count, destroy);
    if (v == null) {
      return 0;
    }
    for (int i = 0; i < v.size(); i++, r++) {
      Notification n = (Notification) v.elementAt(i);
      this.route(n);
    }
    return r;
  }

  /** Pull notifications from the given session using the given subscription identifiers
   * and deliver them locally to the correct subscribers. This supports
   * server-side filtering so client-side filtering is not necessary
   * @param id the session identifier
   * @param subs the notification identifiers (order)
   * @param count the number of notifications to pull
   * @param destroy destructive receipt? true destroys notifications from the session
   *  @return true if the operation succeeded
   */
  public int pullAndDeliver(String id, String subs, int count,
                            boolean destroy) {
    int r = 0;

    Vector v = pull(id, subs, count, destroy);
    if (v == null) {
      return 0;
    }
    for (int i = 0; i < v.size(); i++, r++) {
      Notification n = (Notification) v.elementAt(i);
      this.route(n);
    }
    return r;
  }

  /** Pulls notifications from the given session and returns them in a
   * Vector.
   * @param id the session identifier
   * @param templates Vector containing the ITemplates
   * @param count the number of notifications to pull
   * @param destroy destructive receipt? true destroys notifications from the session
   *  @return Vector containing the pulled notifications
   */
  public Vector pull(String id, Vector templates, int count, boolean destroy) {
    String reply = null;
    Object obj = null;
    try {

      if (id == null || templates == null ||
          templates.size() == 0) {
        return null;
      }

      try {
        if (protocol == NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
          obj = (String) SimpleMessageProtocol.call("pullSession",
              id,
              EventSerializationService.get("Template").multiplex(templates),
              destroy, count);
        }
        else {
          obj = (Object) invoke("pullSession", id, getName(),
                                EventSerializationService.toEventSequence(
              templates),
                                new Integer(count),
                                new Boolean(destroy));

        }
      }
      catch (Exception e) {
        obj = OP_FAILURE;
      }
    }
    catch (Exception e) {
      e.printStackTrace();
    }
    // RETURN VALUE PROCESSING
    // We may receive a String (with error message) or an Sax sequence
    // or a plain string representation
    if (obj instanceof String) {
      reply = (String) obj;
      if (reply == null || reply != null && reply.indexOf(OP_FAILURE) != -1) {
        return null;
      }
      // Possibly string format response
      return EventSerializationService.get("Notification").demultiplex(reply);
    }
    else if (obj instanceof EventSequence) {
      Vector ret = null;
      try {
        ret = EventSerializationService.fromEventSequence( (EventSequence) obj);
      }
      catch (Exception z) {
        z.printStackTrace();
        ret = null;
      }
      return ret;
    }
    else {
      System.out.println("Invalid type:" + obj);
    }
    return null;
  }

  /** Pulls notifications from the given session and returns them in a
   * Vector. Uses subscription identifiers.
   * @param id the session identifier
   * @param subs the notification identifiers to pull
   * @param count the number of notifications to pull
   * @param destroy destructive receipt? true destroys notifications from the session
   *  @return Vector containing the pulled notifications
   */
  public Vector pull(String id, String subs, int count, boolean destroy) {
    String reply = null;
    Object obj = null;

    logPrint("Pull: " + getHost() +
             " port: " + getPort());

    try {
      if (protocol == NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
        obj = (String) SimpleMessageProtocol.call("pullSessionSubs", id, subs,
                                                  destroy, count);
      }
      else {
        obj = (Object) invoke("pullSessionSubs", id, getName(), subs,
                              new Integer(count), new Boolean(destroy));
      }
    }
    catch (Exception e) {
      obj = OP_FAILURE;
    }
    // RETURN VALUE PROCESSING
    // We may receive a String (with error message) or an Sax sequence
    // or a plain string representation
    if (obj instanceof String) {
      reply = (String) obj;
      if (reply == null || reply != null && reply.indexOf(OP_FAILURE) != -1) {
        return null;
      }
      // Possibly string format response
      return EventSerializationService.get("Notification").demultiplex(reply);
    }
    else if (obj instanceof EventSequence) {
      Vector ret = null;
      try {
        ret = EventSerializationService.fromEventSequence( (EventSequence) obj);
      }
      catch (Exception z) {
        z.printStackTrace();
        ret = null;
      }
      return ret;
    }
    else {
      System.out.println("Invalid type:" + obj);
    }
    return null;
  }

  /** Pulls notifications from the given event channel and returns them
   *  as a vector.
   *  @param channel the event channel
   *  @param ITemplates the ITemplates to use in matching
   *  @param count the number of notifications to pull
   *  @return Vector containing the pulled notifications
   */
  public Vector pullEventChannelHistory(String channel, Vector templates,
                                        int count) {
    String reply = null;
    Object obj = null;
    try {
      if (templates == null || templates.size() == 0) {
        return null;
      }

      try {
        if (protocol == NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
          obj = (String) SimpleMessageProtocol.call("pullEventChannelHistory",
              channel,
              EventSerializationService.get("Template").multiplex(templates), false,
              count);
        }
        else {
          obj = (Object) invoke("pullEventChannelHistory", channel,
                                EventSerializationService.toEventSequence(
              templates),
                                new Integer(count));
        }
      }
      catch (Exception e) {
        obj = OP_FAILURE;
      }

    }
    catch (Exception e) {
      e.printStackTrace();
    }

    // RETURN VALUE PROCESSING
    // We may receive a String (with error message) or an Sax sequence
    // or a plain string representation
    if (obj instanceof String) {
      reply = (String) obj;
      if (reply == null || reply != null && reply.indexOf(OP_FAILURE) != -1) {
        return null;
      }
      // Possibly string format response
      return EventSerializationService.get("Notification").demultiplex(reply);
    }
    else if (obj instanceof EventSequence) {
      Vector ret = null;
      try {
        ret = EventSerializationService.fromEventSequence( (EventSequence) obj);
      }
      catch (Exception z) {
        z.printStackTrace();
        ret = null;
      }
      return ret;
    }
    else {
      System.out.println("Invalid type:" + obj);
    }
    return null;
  }

  /** Pulls notifications from the given event channel and returns them
   *  as a vector.
   *  @param channel the event channel
   *  @param ids comma separated list of notification identifiers (order)
   *  @param count the number of notifications to pull
   *  @return Vector containing the pulled notifications
   */
  public Vector pullEventChannelHistoryOrder(String channel, String ids,
                                             int count) {
    String reply = null;
    Object obj = null;

    try {
      if (ids == null) {
        return null;
      }

      try {
        if (protocol == NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
          obj = (String) SimpleMessageProtocol.call(
              "pullEventChannelHistoryOrder", channel, ids, false, count);
        }
        else {
          obj = (Object) invoke("pullEventChannelHistoryOrder", channel, ids,
                                new Integer(count));
        }
      }
      catch (Exception e) {
        obj = OP_FAILURE;
      }

    }
    catch (Exception e) {
      e.printStackTrace();
    }

    // RETURN VALUE PROCESSING
    // We may receive a String (with error message) or an Sax sequence
    // or a plain string representation
    if (obj instanceof String) {
      reply = (String) obj;
      if (reply == null || reply != null && reply.indexOf(OP_FAILURE) != -1) {
        return null;
      }
      // Possibly string format response
      return EventSerializationService.get("Notification").demultiplex(reply);
    }
    else if (obj instanceof EventSequence) {
      Vector ret = null;
      try {
        ret = EventSerializationService.fromEventSequence( (EventSequence) obj);
      }
      catch (Exception z) {
        z.printStackTrace();
        ret = null;
      }
      return ret;
    }
    else {
      System.out.println("Invalid type:" + obj);
    }
    return null;
  }

  /** Reset the subscription status of the given session.
   * @param id session identifier
   * @return the session identifier
   */
  public String resetSession(String id) {
    String reply = null;
    try {
      logPrint("ResetSession: " + getHost() +
               " port: " + getPort());

      try {
        if (protocol == NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
          reply = (String) SimpleMessageProtocol.call("resetSession", id, "", false,
              0);
        }
        else {
          reply = (String) invoke("resetSession", id, getName(), getLastServer(), null);
        }
      }
      catch (Exception e) {
        reply = null;
      }
      logPrint("Reply was: " + reply);
    }
    catch (Exception e) {
      e.printStackTrace();
    }

    if (reply != null && reply.indexOf(OP_FAILURE) != -1) {
      return null;
    }

    return reply;
  }

  /** Move session to a new server by connecting to thew new server.
   *  Note that this is only for special cases. Normally when the
   *  client changes the access server a session is moved when it is
   *  reopened.
   * @param id session identifier
   * @param server the current server
   * @param sport the port
   * @return the session identifier
   */
  public String moveSession(String id, String server, int sport) {
    String reply = null;

    logPrint("------ move: " + server);

    try {
      String host = getHost();
      int port = getPort();

      // we connect to the new server
      setHostPort(server, sport);

      logPrint("Sending roam request to port: " + sport +
               " server: " +
               server);
      try {
        if (protocol == NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
          reply = (String) SimpleMessageProtocol.call("moveSession", id,
              host + ":" + port, false, 0);
        }
        else {
          reply = (String) invoke("moveSession", id, host + ":" + port, getName(), null);
        }
      }
      catch (Exception e) {
        reply = null;
      }

      logPrint("*** Reply was: " + reply);

      if (reply != null) {
        setLastServer(host + ":" + port);
      }
      else { // reset value
        setHostPort(host, port);
      }
    }
    catch (Exception e) {
      e.printStackTrace();
    }

    if (reply != null && reply.equals(OP_FAILURE)) {
      return null;
    }

    return reply;
  }

  /** Close session. Currently one session per client.
   * @param id session identifier
   * @return the session identifier
   */
  public String closeSession(String id) {
    String reply = null;
    try {
      if (protocol == NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
        reply = (String) SimpleMessageProtocol.call("closeSession", id, "", false,
            0);
      }
      else {
        reply = (String) invoke("closeSession", id, getName(), getLastServer(), null);
      }
    }
    catch (Exception e) {
      reply = null;
    }
    logPrint("Reply was: " + reply);

    if (reply != null && reply.indexOf(OP_FAILURE) != -1) {
      return null;
    }

    return reply;
  }

  /** Pause session. Currently one session per client.
   * @param id session identifier
   * @return the session identifier
   */
  public String pauseSession(String id) {
    String reply = null;
    try {
      if (protocol == NOTIFY_SIMPLEMESSAGE_PROTOCOL) {
        reply = (String) SimpleMessageProtocol.call("pauseSession", id, "", false,
            0);
      }
      else {
        reply = (String) invoke("pauseSession", id, getName(), getLastServer(), null);
      }
    }
    catch (Exception e) {
      reply = null;
    }
    logPrint("Reply was: " + reply);

    if (reply != null && reply.equals(OP_FAILURE)) {
      return null;
    }

    return reply;
  }

  /**
   * Route an event coming from transport to listeners based on
   * event type and template matching.
   * Note that if the client state is reset there are two basic
   * recovery approaches: Just renew local subs and do local filtering
   * (this takes computing power, and no notification field dropping must be done
   * The second approach is to reset the subscription status, and do a resubscription
   * In this approach also, there must not be any field dropping, because the
   * system cannot associate the old subscription IDs with the new subscription
   * IDs. If the client is able to store state, it may update the subscription
   * IDs using the update subscription id method.
   *
   * So if the client is not able to store state (subscription ids)
   * notification field dropping is not recommended, because it may
   * lead to duplicate notifications because the subscription identifiers
   * have changed (either for all or for buffered notifications).
   * @param n the notification to route
   */
  public void route(Notification n) {
    boolean f = false;
    int noti = 0;

    // check if the event is expired
    if (n.getDeadline() > 0 &&
        (n.getDeadline() - System.currentTimeMillis()) < 0) {
      expiredNotification++;
      // to be done, currently expired notifications are silently discarded
      return;
    }

    /* Server filtered */
    Vector id = n.getSubId();

    if (id != null && id.size() > 0) {
      fuegocore.notify.util.Log.write("server filtered: " + id.size());

      int procNum = 0;
      for (int i = 0; i < id.size(); i++) {
        // try to find matching interface
        fuegocore.notify.util.Log.write("Proc id: " + (String) id.elementAt(i));
        ITemplate te = (ITemplate) localSubs.get( (String) id.elementAt(i));
        Notifiable nf = null;
        if (te != null) {
          nf = (Notifiable) te.getSource();
        }
        else {
          fuegocore.notify.util.Log.write("No match for: " + n.getSubIdString() +
                                          " iface: " + n.getInterface());
        }
        if (nf != null) {
          fuegocore.notify.util.Log.write(
              "Server-filtered: Calling notifyIncoming");
          noti++;
          procNum++;
          nf.notifyIncoming(n);
        }
      }
      // Here we make the assumption that if one of the subscription ids
      // was properly delivered, they all were and we can exit.
      // Otherwise the fallback mechanism may lead to multiple notifications
      // locally
      if (procNum == id.size()) {
        return;
      }
    }

    // the server filtered falls back to this if the subscription state
    // was reset.

    /* Not server filtered */
    fuegocore.notify.util.Log.write("not server filtered: ");

    NotificationValue nv = n.get();
    TripletTable h = nv.getTriplets();

    // We use a hashtable to prevent multiple delivery
    // A notification will be delivered only once per subscribing interface
    Hashtable excludeList = new Hashtable();

    if (h != null) {
      // we simply interate all events and match
      for (Enumeration k = localSubs.elements(); k.hasMoreElements(); ) {
        ITemplate t = (ITemplate) k.nextElement();
        if (t != null &&
            ( (fuegocore.notify.core.filter.IFBase) t).match(null, h)) {
          Notifiable no = (Notifiable) t.getSource();
          if (no != null) {
            // We do notification formatting here
            // This should not be used if multiple events with overlapping
            // filters are subscribed because we do not know how to
            // format them uniquely
            Notification n2 = t.formatNotification(n);
            noti++;
            no.notifyIncoming(n2);
            continue;
          }
        }
      }
    }

    if (noti == 0) {
      strayNotification++;
    }
  }

  /* Incoming messages (SOAP). We do not currently support other than
   * async. messages so null is returned.
   * @param msg the incoming message
   * @return Message the outgoing message (reply)
   */
  public fuegocore.message.service.Message invoke(fuegocore.
                                                  message.service.Message msg) {
    try {
      EventSequence ses = (EventSequence) msg.getData("saxdata");
      if (ses != null) {
        Notification n = new Notification(ses);
        route(n);
      }
      else {
        String data = (String) msg.getData("data");
        Notification n = new Notification(data);
        route(n);
      }
    }
    catch (Exception e) {
      logPrint("Error during notification route: " + e);
      return null;
    }
    return null;
  }
}
