/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.client;

import java.io.*;
import java.util.Vector;
import java.util.Enumeration;

import org.xmlpull.v1.XmlPullParserException;

import fuegocore.util.xas.CodecFactory;
import fuegocore.util.xas.TypedXmlSerializer;
import fuegocore.util.xas.XasUtil;
import fuegocore.util.xas.ContentCodecFactory;
import fuegocore.util.xas.Qname;
import fuegocore.util.xas.XmlWriter;
import fuegocore.util.xas.EventStream;
import fuegocore.util.xas.TypedEventStream;
import fuegocore.util.xas.XmlReader;
import fuegocore.util.xas.TypedXmlParser;
import fuegocore.util.xas.ContentDecoder;
import fuegocore.util.xas.EventSequence;
import fuegocore.util.xas.Event;
import fuegocore.util.xas.CodecIndustry;
import fuegocore.util.Util;
import fuegocore.notify.core.filter.Filter;

/**
 * Represents a lightweight service request
 */
public class MicroRequestXAS {
    public String op = "";
    public String id = "";
    public String lastserver = "-";
    public String deviceId = "";
    public int count = 0;
    public int flag = 0;
    public Object payload = null;

    private EventStream es = null;
    private TypedEventStream tes = null;
    private XmlReader xr = null;

    public String toString() {
        String ret = "";
        ret += "Op:" + op;
        ret += "Id:" + id;
        ret += "Lastserver:" + lastserver;
        ret += "Count:" + count;
        ret += "DeviceId:" + deviceId;
        ret += "Flag:" + flag;
        ret += "Payload:" + payload;
        return ret;
    }


    public byte[] persist(Object key) throws IOException {
        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        persist(bout, key);
        return bout.toByteArray();
    }

    public void persist(OutputStream os, Object key) throws IOException {
        CodecFactory fac = CodecIndustry.getFactory("application/x-ebu+data");
        TypedXmlSerializer ser = fac.getNewEncoder(key);
        ser.setOutput(os, "ISO-8859-1");
        ser.startDocument(null, null);
        ser.setPrefix("xsd", XasUtil.XSD_NAMESPACE);
        ser.setPrefix("xsi", XasUtil.XSI_NAMESPACE);
        ser.setPrefix("xas", XasUtil.XAS_NAMESPACE);
        Vector prefixes = ContentCodecFactory.getNamespaces();
        for (Enumeration e = prefixes.elements(); e.hasMoreElements(); ) {
            Qname qn = (Qname) e.nextElement();
            ser.setPrefix(qn.getName(), qn.getNamespace());
        }
        ser.startTag(Filter.EVENT_NAMESPACE, "request");
        XmlWriter xw = new XmlWriter(ser);
        xw.typedElement("", "op", XasUtil.XSD_NAMESPACE, "string", op);
        xw.typedElement("", "id", XasUtil.XSD_NAMESPACE, "string", id);
        xw.typedElement("", "ls", XasUtil.XSD_NAMESPACE, "string", lastserver);
        xw.typedElement("", "did", XasUtil.XSD_NAMESPACE, "string", deviceId);
        xw.typedElement("", "count", XasUtil.XSD_NAMESPACE, "int",
                        new Integer(count));
        xw.typedElement("", "flag", XasUtil.XSD_NAMESPACE, "int",
                        new Integer(flag));
        if (payload != null) {
            Qname name = ContentCodecFactory.getXmlName(payload.getClass());
            if (name != null) {
                xw.typedElement("", "payload", name.getNamespace(),
                                name.getName(),
                                payload);
            } else {
                throw new IOException("Cannot serialize payload of type "
                                      + payload.getClass());
            }
        }
        ser.endTag(Filter.EVENT_NAMESPACE, "request");
        ser.endDocument();
        ser.flush();
    }


    public void resurrect(byte[] data, Object key) throws IOException {
        ByteArrayInputStream bin = new ByteArrayInputStream(data);
        resurrect(bin, key);
    }

    public void resurrect(InputStream in, Object key) throws IOException {
        CodecFactory fac = CodecIndustry.getFactory("application/x-ebu+data");
        TypedXmlParser par = fac.getNewDecoder(key);
        try {
            par.setInput(in, "ISO-8859-1");
            ContentDecoder dec = (ContentDecoder)
                                 par.getProperty(XasUtil.PROPERTY_CONTENT_CODEC);
            dec.setRoot(dec);
            if (es == null) {
                es = new EventStream(par);
            } else {
                es.reset(par);
            }
            if (tes == null) {
                tes = new TypedEventStream(es, dec);
            } else {
                tes.reset(es, dec);
            }
            if (xr == null) {
                xr = new XmlReader(tes);
            } else {
                xr.reset(tes);
            }
        } catch (XmlPullParserException ex) {
            ex.printStackTrace();
            throw new IOException(ex.getMessage());
        }
        EventSequence req = xr.nextInsideElement(Filter.EVENT_NAMESPACE,
                                                 "request");
        xr.reset(req);
        xr.advance();
        op = (String) xr.typedContent("", "op", XasUtil.XSD_NAMESPACE, "string");
        id = (String) xr.typedContent("", "id", XasUtil.XSD_NAMESPACE, "string");
        lastserver = (String) xr.typedContent("", "ls", XasUtil.XSD_NAMESPACE,
                                              "string");
        deviceId = (String) xr.typedContent("", "did", XasUtil.XSD_NAMESPACE,
                                            "string");
        count = ((Integer) xr.typedContent("", "count", XasUtil.XSD_NAMESPACE,
                                           "int")).intValue();
        flag = ((Integer) xr.typedContent("", "flag", XasUtil.XSD_NAMESPACE,
                                          "int")).intValue();
        Event ev = xr.advance();
        if (ev.getType() == Event.START_ELEMENT
            && Util.equals(ev.getNamespace(), "")
            && Util.equals(ev.getName(), "payload")) {
            do {
                ev = xr.advance();
            } while (ev != null && ev.getType() != Event.TYPED_CONTENT);
            payload = ev.getValue();
        }
    }
}
