/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.filter;

import java.io.*;
import java.util.*;
import java.util.zip.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.disjuncts.*;
import fuegocore.notify.core.filter.predicates.*;
import fuegocore.notify.core.serialization.*;
import fuegocore.notify.util.*;
import fuegocore.util.xas.*;

/**
 * This class defines the attribute filter. An attribute
 * filter is defined on a single field of a notification and
 * consists of a set of subfilters.  The subfilter may include
 * a selectivity constraint (all fields with the same name and
 * type, any field, and a specific field.
 *
 * Added support for selectivity. Note that field operator
 * uses only the first field (no duplicates). Selectivity supports
 * all, any, and specific fields. You can specify several
 * filters for the same field, so it is relatively flexible.
 *
 * Note that this class does not know anything how to process predicates
 * and therefore atomic predicates and commands may be added in the
 * corresponding classes. In addition, custom type handling may also
 * be implemented by adding support in EventTriplet and to this class,
 * predicates and conjuncts.
 */

public class Filter implements Cloneable, Comparable, IFBase,
        IEventSequenceSerializable,
        IStringSerializable {

    String name; // name/value pair
    int type; // type of field
    int depth = 0; // depth in the poset tree, used to implement Comparable
    String filter; // filter constraint
    String filterstring = null; // filter constraint as an id
    Vector filter_components_conditional; // all filter components, including conditional
    int selectivity = SEL_NORMAL; // selectivity level
    int[] oldCoverCount; // old cover count
    Vector tid; // template reference for matching
    Vector compressedTid;

    public static final String EVENT_NAMESPACE =
            "http://www.hiit.fi/fuego/fc/notify";

    // positive selectivity indicates a specific duplicate field of the notification
    // normal selectivity indicates that the first tuple matches
    // the method noSelectivity() returns true in this case
    public static final int SEL_NORMAL = 0;
    public static final int SEL_ANY = -1;
    public static final int SEL_ALL = -2;

    // Fixed commands and formats
    public static final String OPERAND_DIVISOR = " ";
    public static final String FILTER_DIVISOR = "&";
    public static final String FORMATDELIMITER = ";";
    public static final String OR_DIVISOR = "|";
    public static final String NOT_EQUAL = "neq";
    public static final String EQUAL = "equ";
    public static final String IN_STRING = "in";
    public static final String NOT_IN_STRING = "nin";
    public static final String STARTS_WITH = "stw";
    public static final String ENDS_WITH = "end";
    public static final String EXISTENCE = "exists";
    public static final String SELECT = "select";
    public static final String SELECT_ANY = "any";
    public static final String SELECT_ALL = "all";

    // number and date operators/predicates
    public static final String OP_EQU = "equ";
    public static final String OP_GR = "gr";
    public static final String OP_LT = "lt";
    public static final String OP_NEQ = "neq";
    public static final String OP_LTE = "lte";
    public static final String OP_GTE = "gte";
    public static final String OP_RNG = "rng";

    public static final String CURRENT_TIME = "$now";

    public static void main(String[] args) {
        System.out.println("Testing Event Serialization");
        try {
            Filter sf1 = new Filter("field1", EventTriplet.TYPE_STRING,
                                    "in abc&ends th");
            EventSerializer evs = new EventSerializer();
            sf1.xasSerialize(evs);
            Filter sf2 = new Filter(evs.getCurrentSequence());

            System.out.println("Filter 1: " + sf1.toString());
            System.out.println("Filter 2: " + sf2.toString());

            System.out.println("XML rep:" + evs.toString());
        } catch (Exception e) {
            e.printStackTrace();
        }
        System.exit(0);

        System.out.println("Some examples of merging");

        Filter f1 = new Filter("field1", EventTriplet.TYPE_STRING,
                               "in abc&ends th");
        Filter f2 = new Filter("field1", EventTriplet.TYPE_STRING,
                               "nin abc&ends th");

        f1.perfectMerge(f2);

        System.out.println("Output a: " + f1.toString());

        Filter f1b = new Filter("field1", EventTriplet.TYPE_STRING,
                                "in abc&ends th");
        Filter f2b = new Filter("field1", EventTriplet.TYPE_STRING,
                                "nin abc&ends th");

        f1b.perfectMerge(f2b);

        System.out.println("Output b: " + f1b.toString());

        Filter f3 = new Filter("field1", EventTriplet.TYPE_INT,
                               "rng 7 14");
        Filter f4 = new Filter("field1", EventTriplet.TYPE_INT,
                               "rng 0 10");
        Filter f5 = new Filter("field1", EventTriplet.TYPE_INT,
                               "rng 8 15");

        f3.perfectMerge(f4);
        f3.perfectMerge(f5);

        System.out.println("Output: " + f3.toString());
        // test sorting
        Filter f6 = new Filter("field1", EventTriplet.TYPE_INT,
                               "rng 8 15|equ 10|gr 100|lt 10|neq 4");
        System.out.println("Sorted: " + f6.toString());

        Filter f7 = new Filter("field1", EventTriplet.TYPE_INT,
                               "gr 10&neq 10|lt 2|rng 10 100|equ 100");
        Filter f8 = new Filter("field1", EventTriplet.TYPE_INT,
                               "gr 10&equ 10|lt 3|rng 0 110|equ 200");

        f7.perfectMerge(f8);
        System.out.println("Output a: " + f7.toString());

        Filter f7b = new Filter("field1", EventTriplet.TYPE_INT,
                                "gr 10&neq 10|lt 2|rng 10 100|equ 100");
        Filter f8b = new Filter("field1", EventTriplet.TYPE_INT,
                                "gr 10&equ 10|lt 3|rng 0 110|equ 200");

        f7b.perfectMerge(f8b);

        System.out.println("Output b: " + f7b.toString());

        Filter f9 = new Filter("field1", EventTriplet.TYPE_INT,
                               "gr 10&neq 10|lt 2|rng 10 100|equ 100");
        Filter f10 = new Filter("field1", EventTriplet.TYPE_INT,
                                "gr 5|lt 10|rng 40 50|neq 100");

        f9.perfectMerge(f10);
        System.out.println("Output: " + f9.toString());

        Filter f11 = new Filter("field1", EventTriplet.TYPE_INT,
                                "gr 10&neq 10|lt 2|rng 10 100|equ 100");
        Filter f12 = new Filter("field1", EventTriplet.TYPE_INT,
                                "gr 5|lt 10|rng 40 50|neq 100");

        f11.perfectMerge(f12);
        System.out.println("Output: " + f11.toString());

        System.out.println("Existence cover test");

        Filter f1c = new Filter("field1", EventTriplet.TYPE_INT,
                                "gr 10");
        Filter f2c = new Filter("field1", EventTriplet.TYPE_INT,
                                "exists");
        Filter f3c = new Filter("field1", EventTriplet.TYPE_INT,
                                "lt 20");

        f1c.perfectMerge(f3c);
        if (f1c.covers(f2c) && f2c.covers(f1c)) {
            System.out.println("Test ok");
        } else {
            System.out.println("Existence cover test failed");

        }
    }

    public Filter() {
    }

    /**
     * Returns the size of the object
     * If filter the # disjuncts, if disjunct the # predicates
     * @return the size
     */
    public int getSize() {
        if (filter_components_conditional == null) {
            return 0;
        }
        return filter_components_conditional.size();
    }

    public boolean noSelectivity() {
        if (selectivity == SEL_NORMAL || selectivity == 1) {
            return true;
        }
        return false;
    }

    /**
     * Returns the depth of this object in the poset tree.
     * @return the depth as int
     */
    public int getDepth() {
        return depth;
    }

    /**
     * Set the depth of this object in the poset tree.
     * @param i the depth to set
     */

    public void setDepth(int i) {
        depth = i;
    }

    /** Returns the cost of the filter (generality).
     *  Higher value denotes more general filters.
     * @return int the cost
     */
    public int getCost() {
        int cost = 0;

        // selectivity of ALL is most restrictive
        // ANY is most general
        if (selectivity == SEL_ALL) {
            cost = filter_components_conditional.size() * 1000;
        } else if (selectivity == SEL_ANY) {
            cost = filter_components_conditional.size() * 10000;
        } else {
            cost = filter_components_conditional.size() * 5000;

        }
        for (int i = 0; i < filter_components_conditional.size(); i++) {
            Disjunct dd = (Disjunct) filter_components_conditional.elementAt(i);
            cost += dd.getCost();
        }
        return cost;
    }

    /**
     * Comparison based on depth.
     * @param other the other object to compare to
     * @return int the comparison value
     */
    public int compareTo(Object other) {
        if (other instanceof Filter) {
            int a = ((Filter) other).getDepth();
            return a - getDepth();

        } else {
            throw new ClassCastException();
        }
    }

    /****************************************************************************/
    /****************************************************************************/
    /** XAS Serialization */

    /**
     * Insert the filters to a SAX event sequence.
     * @param ser TypedXmlSerializer
     * @throws IOException
     */
    public void toEventSequence(TypedXmlSerializer ser) throws IOException {
        xasSerialize(ser);
    }

    /**
     * Parse object from the given event sequence
     * @param ses EventSequence
     * @return boolean
     */
    public boolean fromEventSequence(EventSequence ses) {
        String f = convertToInternal(ses);
        if (f == null) {
            fuegocore.notify.util.Log.writeFilter("Filter was null: " + f);
            return false;
        }

        filter_components_conditional = new Vector();

        String[] fs = parseConditionalClauses();

        if (fs == null || fs.length == 0) {
            return false;
        }

        for (int i = 0; i < fs.length; i++) {
            f = fs[i];
            Disjunct conj = null;

            // Decompose filter string
            if (type == EventTriplet.TYPE_INT ||
                type == EventTriplet.TYPE_DOUBLE) {
                conj = new NumberDisjunct(this);
                conj.decompose(f);
            } else if (type == EventTriplet.TYPE_STRING) {
                if (name.equals("type")) {
                    // type is a special filter field
                    // it does not have a command, so we add one
                    // it is always equals due to mapping to event channels
                    if (f != null && f.indexOf("equ") == -1) {
                        f = "equ " + f;
                    }
                }
                conj = new StringDisjunct(this);
                conj.decompose(f);
            } else if (type == EventTriplet.TYPE_BOOLEAN) {
                conj = new BooleanDisjunct(this);
                conj.decompose(f);
            } else if (type == EventTriplet.TYPE_DATE) {
                conj = new DateDisjunct(this);
                conj.decompose(f);
            }
            // add reference to the conditional vector
            filter_components_conditional.add(conj);
        }
        Collections.sort(filter_components_conditional);
        return true;
    }

    /**
     * Serialize using xas
     * @param ser TypedXmlSerializer
     * @throws IOException
     */
    public void xasSerialize(TypedXmlSerializer ser) throws IOException {
        XmlWriter xw = new XmlWriter(ser);

        EventSerializer evs = new EventSerializer();
        XmlWriter xw2 = new XmlWriter(evs);

        xw2.typedElement("", "name", XasUtil.XSD_NAMESPACE, "string", name);
        xw2.typedElement("", "type", XasUtil.XSD_NAMESPACE, "int",
                         new Integer(type));
        xw2.typedElement("", "filter", XasUtil.XSD_NAMESPACE, "string",
                         convertFilterToInternalForm());

        xw.complexElement(EVENT_NAMESPACE, "af",
                          evs.getCurrentSequence());
    }

    /**
     * Convert given EventSequence to internal form
     * @param es EventSequence
     * @return String
     */
    public String convertToInternal(EventSequence es) {
        XmlReader xr0 = new XmlReader(es);
        XmlReader xr = null;
        EventSequence esf = xr0.completeContent(EVENT_NAMESPACE, "af");
        if (esf == null) {
            xr = xr0;
        } else {
            xr = new XmlReader(esf);

        }
        name = (String) xr.typedContent("", "name", XasUtil.XSD_NAMESPACE,
                                        "string");

        Integer typ = (Integer) xr.typedContent("", "type",
                                                XasUtil.XSD_NAMESPACE,
                                                "int");
        if (typ != null) {
            try {
                type = typ.intValue();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        filter = (String) xr.typedContent("", "filter", XasUtil.XSD_NAMESPACE,
                                          "string");
        return filter;
    }

    /** This constructor takes an event sequence as input
     *  @param ses the SaxEventSequence
     */
    public Filter(EventSequence ses) {
        fromEventSequence(ses);
    }

    /** XAS Serialization */
    /****************************************************************************/
    /****************************************************************************/


    /** Returns the length of the filter XML representation in GZIP format
     * @return int the size of the GZIPPED filter
     */
    public int filterGZIPLength() {
        try {
            ByteArrayOutputStream bay = new ByteArrayOutputStream();
            GZIPOutputStream gzip = new GZIPOutputStream(bay);
            String content = this.toString();

            gzip.write(content.getBytes(), 0, content.getBytes().length);
            gzip.finish();
            gzip.close();
            byte[] bb = bay.toByteArray();
            return bb.length;
        } catch (Exception e) {
            e.printStackTrace();
        }
        return -1;
    }

    public Object clone() {
        Filter f = new Filter(toString());
        return f;
    }

    public String getName() {
        return name;
    }

    public int getType() {
        return type;
    }

    public String getSelectivityString() {
        if (selectivity == SEL_NORMAL) {
            return null;
        } else if (selectivity == SEL_ALL) {
            return this.SELECT + OPERAND_DIVISOR + this.SELECT_ALL;
        } else if (selectivity == this.SEL_ANY) {
            return this.SELECT + this.OPERAND_DIVISOR + this.SELECT_ANY;
        } else {
            return this.SELECT + this.OPERAND_DIVISOR + selectivity;
        }
    }

    /** Returns the filter string
     * @return String the filter
     */
    public String getFilter() {
        return filter;
    }

    /** Returns plain type
     * @return String the type
     */
    public String getFilterType() {
        int i = filter.indexOf("equ");
        if (i > -1) {
            return filter.substring(i + 3, filter.length()).trim();
        } else {
            return filter;
        }
    }

    public String toStringOld() {
        String s = name + FORMATDELIMITER + type + FORMATDELIMITER + filter;
        return s;
    }

    /** Get field/attribute identifier (name/type)
     * @return String the identifier
     */
    public String getId() {
        return name + "/" + type;
    }

    /** Get field/attribute identifier (name/type/selectivity)
     * @return String the identifier for merge
     */
    public String getIdforMerge() {
        return name + "/" + type + "/" + getSelectivity();
    }

    /** Returns the XML String representation of the object
     * @return String the XML String representation
     */
    public String toString() {
        return toXML();
    }

    /** XML representation of a single filter. Currently subfilters
     * are in internal notation.
     * @return String the XML String
     */
    public String toXML() {
        String s = "<name>" + name + "</name>" +
                   "<type>" + type + "</type>" + "<filter>" +
                   convertFilterToInternalForm() +
                   "</filter>";
        return s;
    }

    public String convertToInternal(String s) {
        name = Auxil.getTag(s, "name");
        String typ = Auxil.getTag(s, "type");
        filter = Auxil.getTag(s, "filter");
        type = -1;
        if (typ != null) {
            try {
                type = (new Integer(typ.trim())).intValue();
            } catch (Exception e) {
                e.printStackTrace();
                fuegocore.notify.util.Log.writeFilter("Exception: " + e);
                fuegocore.notify.util.Log.writeFilter("String was: " + typ);
            }
        }
        //filter = convertFilterToInternalForm();

        return filter;
    }

    /** Convert filter string to internal representation.
     *  Called only when the given filter is in XML.
     * @return String the String representation of the constraint
     */
    public String convertFilterToInternalForm() {
        // Loop over all conditional clauses
        String ret = "";

        // General objects can only be tested for existence
        if (type == EventTriplet.TYPE_SERIALIZED) {
            return "exists";
        }

        // add the selectivity constraint as a prefix (not sorted)
        String sel = getSelectivityString();
        if (sel != null) {
            ret += sel + "&"; // it does not matter where it is...

        }
        if (filter_components_conditional == null) {
            return ret;
        }

        for (int i = 0; i < filter_components_conditional.size(); i++) {
            // Get one conjunctive formula
            Disjunct dd = (Disjunct) filter_components_conditional.elementAt(i);
            ret += dd.getFilterString();
            if (i < filter_components_conditional.size() - 1) {
                ret += OR_DIVISOR;
            }
        }

        return ret; /* currently always in internal mode */
    }

    /** Convert filter to a unique identifier
     * @return String the filter identifier
     */
    public String createFilterId() {
        // Loop over all conditional clauses
        String ret = "";
        if (filter_components_conditional == null) {
            return null;
        }

        // General objects can only be tested for existence
        if (type == EventTriplet.TYPE_SERIALIZED) {
            return "exists";
        }

        // add the selectivity constraint as a prefix (not sorted)
        String sel = getSelectivityString();
        if (sel != null) {
            ret += sel + "&"; // it does not matter where it is...

        }
        String[] cons = new String[filter_components_conditional.size()];

        for (int i = 0; i < filter_components_conditional.size(); i++) {
            Disjunct dd = (Disjunct) filter_components_conditional.elementAt(i);
            ret += dd.getFilterString();
            cons[i] = ret;
            ret = "";
        }

        /* Sort */
        for (int i = 0; i < filter_components_conditional.size(); i++) {
            for (int j = i; j < filter_components_conditional.size(); j++) {
                if (cons[i].compareTo(cons[j]) > 0) {
                    String temp = cons[i];
                    cons[i] = cons[j];
                    cons[j] = temp;
                }
            }
        }

        // Formulate the string
        ret = "";
        for (int i = 0; i < filter_components_conditional.size(); i++) {
            ret += cons[i];
            if (i < filter_components_conditional.size() - 1) {
                ret += OR_DIVISOR;
            }
        }

        // System.out.println("Unique id: " + ret);

        return ret.trim(); /* currently always in internal mode */
    }

    /** Returns statistics about the filter.
     * @return String the statistics
     */
    public String getComplexity() {
        // Loop over all conditional clauses
        String ret = "";
        int Disjunct = 0;
        int size = 0;

        for (int i = 0; i < filter_components_conditional.size(); i++) {
            // Get one conjunctive formula
            Hashtable b = (Hashtable) filter_components_conditional.elementAt(i);
            for (Enumeration j = b.elements(); j.hasMoreElements(); ) {
                IAtomicPredicate e_b = (IAtomicPredicate) j.nextElement();
                if (e_b != null) {
                    Disjunct++;
                    size += e_b.toString().length();
                }
                if (j.hasMoreElements()) {
                    ret += FILTER_DIVISOR;
                }
            }
            if (i < filter_components_conditional.size() - 1) {
                size++;
            }
        }

        ret = "Disjunctions:" + filter_components_conditional.size() +
              " Conjuncts: " + Disjunct + " Size: " + size;

        return ret;
    }

    /** Returns the filter string
     * @return String the constraint as a String
     */
    public String getFilterString() {
        if (filterstring == null) {
            filterstring = createFilterId();
        }
        return filterstring;
    }

    public String getCoverId() {
        return getId() + "/" + getFilterString();
    }

    /** Returns the selectivity of this filter (against duplicate names)
     * @return int the selectivity
     */
    public int getSelectivity() {
        return selectivity;
    }

    /** Sets the selectivity of this filter
     * @param i the selectivity to set
     */
    public void setSelectivity(int i) {
        selectivity = i;
    }

    /** Extracts conditional clauses
     * @return String[] the conditional clauses
     */
    public String[] parseConditionalClauses() {
        String[] ret = null;
        // equ y:ends z:or:
        try {
            StringTokenizer st = new StringTokenizer(filter, OR_DIVISOR);

            fuegocore.notify.util.Log.writeFilter("Found " + st.countTokens());

            int size = st.countTokens();
            if (size == 0) { // no conditional clause
                ret = new String[1];
                ret[0] = filter;
                return ret;
            }

            ret = new String[size];
            int i = 0;
            while (st.hasMoreTokens()) {
                String subfilter = st.nextToken();
                fuegocore.notify.util.Log.writeFilter("subfilter: " + subfilter);
                ret[i++] = subfilter;
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return ret;
    }

    /* ***************************************************************** */
    /* COVERING 														 */
    /* ***************************************************************** */

    /** This method returns true if the given filter is perfectly mergeable
     * with this one.
     * @param fi the filter
     * @return true if the filters are perfectly mergeable
     */
    public boolean perfectlyMergeable(Filter fi) {
        // We require that the name and type and selectivity are identical
        if (fi.getType() == getType() && fi.getName().equals(getName()) &&
            (fi.getSelectivity() == getSelectivity())) {
//         fi.getSelectivity() == SEL_ANY ||
//         getSelectivity() == SEL_ANY)) {
            return true;
        }
        return false;
    }

    /** This method returns true if the given filter is imperfectly mergeable
     * with this one.
     * @param fi the filter
     * @return true if the filters are perfectly mergeable
     */

    public boolean imperfectlyMergeable(Filter fi) {
        // We require that the name and type are identical
        if (fi.getType() == getType() && fi.getName().equals(getName())) {
            return true;
        }
        return false;
    }


    /** Merges the current object with the given object
     * @param obj the object to be merged with the current object
     * @return Object the merged result
     */
    public Object mergeObject(Object obj) {
        Filter fx = (Filter)this.clone();
        if (fx.merge(obj)) {
            return fx;
        } else {
            return null;
        }
    }

    /** Perfectly merges two filters and returns true if it succeeded.
     * Filter merging is done using disjunctive normal form.
     * Conditional filters are conjunctions.
     * @param fi the filter to merge
     * @return true if the perfect merge succeeded
     */
    public synchronized boolean perfectMerge(Filter fi) {
        if (!perfectlyMergeable(fi)) {
            return false;
        }

        Vector v = fi.getAllComponents();
        if (v == null) {
            return false;
        }
        // Merge disjuncts with the maximality assumption
        mergeDisjunctsMinimalityAssumption(v);
        Collections.sort(filter_components_conditional);

        return true;
    }

    /**
     * Perfectly merges two conjunctive attribute filters and returns true if it succeeded.
     * Conditional filters are conjunctions.
     * @param fi Filter
     * @param test boolean
     * @return boolean
     */
    public synchronized boolean conjunctivePerfectMerge(Filter fi, boolean test) {
        if (!perfectlyMergeable(fi)) {
            return false;
        }

        // Set the selectivity
        if (fi.getSelectivity() != getSelectivity()) {
            if (fi.getSelectivity() == SEL_ANY ||
                getSelectivity() == SEL_ANY) {
                setSelectivity(SEL_ANY);
            }
        }

        // Get all components of the given filter
        Vector v = fi.getAllComponents();
        if (v == null || v.size() < 1) {
            return false;
        }

        Disjunct me = (Disjunct) filter_components_conditional.
                      elementAt(0);

        // Perform the conjunctive merge if possible
        boolean f = me.conjunctiveMerge(test, (Disjunct) v.elementAt(0));

        return f;
    }

    /**
     * Merge disjuncts perfectly and remove duplicates.
     * We use pair-wise comparisons of the each of the conjunctive clauses to determine
     * the covering relations between them.
     * @param af2 Vector
     */
    public void mergeDisjunctsMinimalityAssumption(Vector af2) {
        Hashtable coveredPreds = new Hashtable();
        // Loop over all new disjuncts
        Vector removeSet = new Vector();
        boolean repeat = false;
        Disjunct cond = null;

        int i = 0;

        while (i < af2.size() || repeat) {
            // Get one conjunctive formula
            //System.out.println("Processing:"  + i + " repeat: " + repeat);

            if (!repeat) {
                if (i >= af2.size()) {
                    break;
                }
                cond = (Disjunct) af2.elementAt(i++);
            } else {
                repeat = false;
            }

            removeSet.clear();

            boolean skip = false;

            // compare to each existing disjunct
            for (int k = 0; k < filter_components_conditional.size(); k++) {
                Disjunct cond2 = (Disjunct) filter_components_conditional.
                                 elementAt(k);
                // We check if this is covered
                if (cond2 != null) {
                    // if cond2 covers cond
                    int re = cond.coverAndMerge(cond2);
                    //System.out.println("cond2 covers cond:" + re);
                    if (re == Disjunct.COVERED) {
                        // cond was covered so it is not added to filter_components
                        skip = true;
                        break;
                    } else if (re == Disjunct.MODIFIED) {
                        // cond2 was modified so we remove cond and re-iterate with cond2
                        // We remove cond2 and add it to the add set
                        // then iterate again and add it as a new if not covered
                        repeat = true;
                        skip = true;
                        cond = cond2; // change reference
                        removeSet.add(cond2);
                        break;
                    }

                    // existence is captured below
                    //System.out.println("cond covers cond2:" + re);
                    re = cond2.coverAndMerge(cond);
                    if (re == Disjunct.COVERED) {
                        // cond2 was covered by cond so cond2 is removed and cond is added
                        removeSet.add(cond2);
                        // we have to iterate the others as well
                    } else if (re == Disjunct.MODIFIED) {
                        // cond was modified so we remove cond2 and re-iterate with cond
                        removeSet.add(cond2);
                        repeat = true;
                        skip = true; // not yet orthogonal, repeat to make sure
                        break;
                    }

                    // Check for existence
                    if (cond2.size() == 1) {
                        Object obj = cond2.elements().nextElement();

                        try {
                            IAtomicPredicate pred = (IAtomicPredicate) obj;
                            if (pred != null &&
                                pred.getOperand().equals(EXISTENCE)) {
                                // We found an existence test, which covers all other predicates
                                // We replace all predicates with this
                                filter_components_conditional.clear();
                                filter_components_conditional.add(cond2);
                                // Update internal data structures
                                filter = convertFilterToInternalForm();
                                return;
                            }
                        } catch (ClassCastException ce) {
                            ce.printStackTrace();
                            System.out.println(
                                    "Fatal error in mergeDisjunctsMinimalityAssumption");
                            System.out.println("Class:" +
                                               obj.getClass().getName());
                            System.out.println(obj);
                            System.exit(0);
                        }
                    }
                }
            }

            if (removeSet.size() > 0) {
                filter_components_conditional.removeAll(removeSet);
            }

            if (!skip) {
                //System.out.println("Adding cond");
                // Cond was orthogonal so we add it to the set
                filter_components_conditional.add(cond);
            }
        }
        // Update internal data structures
        filter = convertFilterToInternalForm();
    }

    /** Merge predicates perfectly and remove duplicates.
     * We use pair-wise comparisons of the each of the conjunctive clauses to determine
     * the covering relations between them.  Cost metric (distance) is not needed
     * because we merge the first possible alternative.
     **/
    public void mergePredicatesPerfectly() {
        Hashtable coveredPreds = new Hashtable();
        boolean covered = false;

        // purge list
        boolean removeList[] = new boolean[filter_components_conditional.size()];
        int coverCount[] = new int[filter_components_conditional.size()];
        int purge = 0;
        int removeCount = 0;
        int neqCount = 0;
        String lastNeq = null;

        // Loop over all conditional clauses
        for (int i = 0; i < filter_components_conditional.size(); i++) {
            // Get one conjunctive formula
            Disjunct cond = (Disjunct) filter_components_conditional.elementAt(
                    i);
            covered = false;

            // add old cover count
            if (oldCoverCount != null && i < oldCoverCount.length) {
                coverCount[i] += oldCoverCount[i];

                // Compare each component with all existing alternatives
            }
            for (int k = 0; k < filter_components_conditional.size(); k++) {
                if (k == i || removeList[k]) {
                    continue; // component covers itself trivially
                }

                Disjunct cond2 = (Disjunct) filter_components_conditional.
                                 elementAt(k);

                // We check if this Disjunct is covered
                if (cond2 != null) {
                    // if cond2 covers cond
                    int re = cond.coverAndMerge(cond2);
                    if (re != cond.FAILED) {
                        covered = true;
                        coverCount[k]++;
                        //System.out.println("cover count: " + coverCount[k] + " " + k);

                        // Check for existence
                        if (cond2.size() == 1 &&
                            ((IAtomicPredicate) (cond2.elements().nextElement())).
                            getOperand().equals(EXISTENCE)) {
                            // We found an existence test, which covers all other predicates
                            // We replace all predicates with this
                            filter_components_conditional.clear();
                            filter_components_conditional.add(cond2);
                            // Update internal data structures
                            filter = convertFilterToInternalForm();
                            return;
                        }
                        break; // break loop
                    }
                }
            }
            // Check covering
            if (covered) {
                removeCount++;
                removeList[i] = true;
            }
        }

        Vector v = (Vector) filter_components_conditional.clone();
        filter_components_conditional.clear();

        // sort set according to the cover metric (counter)
        // this is required to improve matching time
        for (int i = 0; i < v.size(); i++) {
            for (int j = i; j < v.size(); j++) {
                if (coverCount[i] < coverCount[j]) {
                    int t = coverCount[j];
                    coverCount[j] = coverCount[i];
                    coverCount[i] = t;

                    boolean tm = removeList[j];
                    removeList[j] = removeList[i];
                    removeList[i] = tm;

                    Disjunct tmp = (Disjunct) v.elementAt(j);
                    v.set(j, (Disjunct) v.elementAt(i));
                    v.set(i, tmp);
                }
            }
        }

        int newCoverCount[] = new int[v.size() - removeCount];
        int j = 0;
        // remove covered
        for (int i = 0; i < v.size(); i++) {
            if (!removeList[i]) {
                newCoverCount[j++] = coverCount[i];
                filter_components_conditional.add((Disjunct) v.elementAt(i));
            }
        }
        v = null;
        oldCoverCount = newCoverCount; // set reference

        // Update internal data structures
        filter = convertFilterToInternalForm();
    }

    /**
     * Covering test
     * @param obj the covered object
     * @return true if covered
     */
    public boolean covers(Object obj) {
        if (!(obj instanceof Filter)) {
            return false;
        }

        return covers((Filter) obj);
    }

    /**
     * Merge
     * @param obj the object to be merged
     * @return boolean true if the merge succeeded
     */
    public boolean merge(Object obj) {
        if (!(obj instanceof Filter)) {
            return false;
        }
        return perfectMerge((Filter) obj);
    }

    /**
     * Match
     * @param field Object
     * @param noti Object
     * @return boolean
     */
    public boolean match(Object field, Object noti) {
        if (!(field instanceof EventTriplet)) {
            return false;
        }
        if (!(noti instanceof TripletTable)) {
            return false;
        }
        return evaluate((EventTriplet) field, (TripletTable) noti);
    }

    /**
     * Covering test: is the given object covered
     * @param obj the object that covers
     * @return true if the current object was covered
     */
    public boolean isCovered(Object obj) {
        if (!(obj instanceof Filter)) {
            return false;
        }
        Filter tobj = (Filter) obj;
        if (tobj == null) {
            return false;
        }
        return tobj.covers(this);
    }

    /** Returns true if the filter covers the given filter.
     *  The covering test for a filter is similar to the covering test for
     * conjuncts; however, with disjunctions the requirement is as follows: a
     *  filter B is covered by the filter A if and only if for all conjuncts in
     * B exists a conjunct in A that covers B.  Instead of counting the number
     * of covering predicates, we count the number of covered conjuncts.
     * The covering test for a set of filters is achieved by counting the number
     * of covered filters. If each filter in template T1 covers a filter in
     * template T2, T1 is said to cover T2.
     * @param f2 the filter that is covered
     * @return true if the current filter covers the input filter
     */
    public boolean covers(Filter f2) {
        Hashtable coveredPreds = new Hashtable();
        int coverCount = 0;

        // First requirement for covering
        if (!perfectlyMergeable(f2)) {
            // It may be that all or any cover something other
            // we need to check this special case (the template class handles
            // the duplicates for the higher level by checking them)
            // note that for all other selection they must be 1:1

            if (getSelectivity() == SEL_ANY) {
                ; // continue to check the attribute
            } else {
                return false;
            }
        }

        // Check filter equivalence this is faster than doing the covering check.
        if (this.getFilterString().equals(f2.getFilterString())) {
            return true;
        }

        // Loop over all conditional clauses
        for (int k = 0; k < f2.getAllComponents().size(); k++) {
            // Get one conjunctive formula
            Disjunct cond2 = (Disjunct) f2.getAllComponents().elementAt(k);

            for (int i = 0; i < filter_components_conditional.size(); i++) {
                Disjunct cond = (Disjunct) filter_components_conditional.
                                elementAt(
                                        i);
                // Compare each component with all existing alternatives
                // We check if this conjugate is covered
                if (cond2 != null) {
                    // if cond covers cond2
                    if (cond.covers(cond2)) {
                        coverCount++;
                        break; // break loop
                    }
                }
            }
        }

        //System.out.println("f1: " + getFilterString());
        //System.out.println("f2: " + f2.getFilterString());
        //System.out.println("coverCount: " + coverCount +
        //                  " == " + f2.getAllComponents().size());
        // All conjuncts in f2 were covered by this filter
        if (coverCount == f2.getAllComponents().size()) {
            return true;
        }
        return false;
    }

    /* ***************************************************************** */

    /**
     * Returns true if the filter overlaps with the given filter.
     * @param obj Object
     * @return boolean
     */
    public boolean overlaps(Object obj) {
        Hashtable coveredPreds = new Hashtable();
        int coverCount = 0;

        if (!(obj instanceof Filter)) {
            return false;
        }

        Filter f2 = (Filter) obj;

        // First requirement for covering
        if (!perfectlyMergeable(f2)) {
            // It may be that all or any cover something other
            // we need to check this special case (the template class handles
            // the duplicates for the higher level by checking them)
            // note that for all other selection they must be 1:1

            if (getSelectivity() == SEL_ANY || getSelectivity() == SEL_ALL) {
                ; // continue to check the attribute
            } else {
                return false;
            }
        }

        // Check filter equivalence this is faster than doing the covering check.
        if (this.getFilterString().equals(f2.getFilterString())) {
            return true;
        }

        // Loop over all conditional clauses
        for (int k = 0; k < f2.getAllComponents().size(); k++) {
            // Get one conjunctive formula
            Disjunct cond2 = (Disjunct) f2.getAllComponents().elementAt(k);

            for (int i = 0; i < filter_components_conditional.size(); i++) {
                Disjunct cond = (Disjunct) filter_components_conditional.
                                elementAt(
                                        i);
                // Compare each component with all existing alternatives
                if (cond.overlaps(cond2)) {
                    // In a disjunctive formula it is enough that any
                    // two disjuncts overlap.
                    return true;
                }
            }
        }

        return false;
    }

    /**
     * Returns the estimated overlap cost to the given filter. The
     * cost is symmetric. Returns the overlap of the most selective pair
     * of disjuncts.
     * @param f2 Filter
     * @return double
     */
    public double getOverlapCost(Filter f2) {
        // Check filter equivalence this is faster than doing the covering check.
        if (this.getFilterString().equals(f2.getFilterString())) {
            return 1; // maximal overlap
        }

        double maxOverlap = 0;
        int c = 0;
        // Loop over all conditional clauses
        for (int k = 0; k < f2.getAllComponents().size(); k++) {
            // Get one conjunctive formula
            Disjunct cond2 = (Disjunct) f2.getAllComponents().elementAt(k);

            for (int i = 0; i < filter_components_conditional.size(); i++) {
                Disjunct cond = (Disjunct) filter_components_conditional.
                                elementAt(
                                        i);

                double d = 0;

                // if either covers the other
                if (cond2.covers(cond) || cond.covers(cond2)) {
                    d = 1;
                } else {
                    // Compare each component with all existing alternatives
                    d = cond.getOverlapCost(cond2);
                }

                maxOverlap += d;
                c++;
            }
        }
        double ret = maxOverlap / (double) c;
        return ret;
    }

    /* ***************************************************************** */

    /** Returns a vector consisting of hashtables (the conjunctive clauses)
     * @return Vector the components
     */
    public Vector getAllComponents() {
        return filter_components_conditional;
    }

    /** Constructor
     * @param n the name of the filter
     * @param t the type of the filter
     * @param f the filter clause
     */
    public Filter(String n, int t, String f) {
        name = n;
        type = t;
        filter = f;

        filter_components_conditional = new Vector();

        String[] fs = parseConditionalClauses();

        if (fs == null || fs.length == 0) {
            return;
        }

        for (int i = 0; i < fs.length; i++) {
            f = fs[i];

            Disjunct conj = null;

            // Decompose filter string
            if (type == EventTriplet.TYPE_INT ||
                type == EventTriplet.TYPE_DOUBLE) {
                conj = new NumberDisjunct(this);
                conj.decompose(f);
            } else if (type == EventTriplet.TYPE_STRING) {
                if (name.equals("type")) {
                    // type is a special filter field
                    // it does not have a command, so we add one
                    // it is always equals due to mapping to event channels
                    if (f != null && f.indexOf("equ") == -1) {
                        f = "equ " + f;
                    }
                }
                conj = new StringDisjunct(this);
                conj.decompose(f);
            } else if (type == EventTriplet.TYPE_BOOLEAN) {
                conj = new BooleanDisjunct(this);
                conj.decompose(f);
            } else if (type == EventTriplet.TYPE_DATE) {
                conj = new DateDisjunct(this);
                conj.decompose(f);
            }
            // Custom processing may be added here

            // add reference to the conditional vector
            filter_components_conditional.add(conj);
        }
        Collections.sort(filter_components_conditional);
    }

    public synchronized void addTid(Object id) {
        if (tid == null) {
            tid = new Vector();
            compressedTid = new Vector();
        }
        tid.add(id);
        if (!compressedTid.contains(id)) {
            compressedTid.add(id);
        }
    }

    public Vector getTid() {
        return tid;
    }

    public Vector getCompressedTid() {
        return compressedTid;
    }

    /**
     * Parse the object from the given String
     * @param str String
     * @return boolean
     */
    public boolean parse(String str) {
        String f = convertToInternal(str);
        if (f == null) {
            fuegocore.notify.util.Log.writeFilter("Filter was null: " + f);
            return false;
        }

        filter_components_conditional = new Vector();

        String[] fs = parseConditionalClauses();

        if (fs == null || fs.length == 0) {
            return false;
        }

        for (int i = 0; i < fs.length; i++) {
            f = fs[i];
            Disjunct conj = null;

            // Decompose filter string
            if (type == EventTriplet.TYPE_INT ||
                type == EventTriplet.TYPE_DOUBLE) {
                conj = new NumberDisjunct(this);
                conj.decompose(f);
            } else if (type == EventTriplet.TYPE_STRING) {
                if (name.equals("type")) {
                    // type is a special filter field
                    // it does not have a command, so we add one
                    // it is always equals due to mapping to event channels
                    if (f != null && f.indexOf("equ") == -1) {
                        f = "equ " + f;
                    }
                }
                conj = new StringDisjunct(this);
                conj.decompose(f);
            } else if (type == EventTriplet.TYPE_BOOLEAN) {
                conj = new BooleanDisjunct(this);
                conj.decompose(f);
            } else if (type == EventTriplet.TYPE_DATE) {
                conj = new DateDisjunct(this);
                conj.decompose(f);
            }
            // add reference to the conditional vector
            filter_components_conditional.add(conj);
        }
        Collections.sort(filter_components_conditional);
        return true;
    }

    /** This constructor takes XML string as input
     * @param str the XML string
     */
    public Filter(String str) {
        parse(str);
    }

    /** Evaluate this filter with an event triplet. Takes also a TripletTable
     * to the all event triplets of the notification as a parameter in order
     * for field comparison constraints (extra computation required is not great).
     * Supports duplicate field names in notification.
     * NOTE: if selectivity is used this OVERRIDES the given event triplet!
     * The correct triplet is found using the TripletTable
     * @param ev the triplet to match against the filter
     * @param h TripletTable for all the triplets in the notification to be matched
     * @return true if the condition holds
     */
    public synchronized boolean evaluate(EventTriplet ev, TripletTable h) {
        fuegocore.notify.util.Log.writeFilter("Filter: evaluating type: " +
                                              ev.getType() + " name: " +
                                              name);

        if (selectivity == SEL_NORMAL) { // normal selectivity
            return evaluateSingle(ev, h);
        } else if (selectivity > 0) { // specific triplet
            EventTriplet ev2 = (EventTriplet) h.get(ev.getId(), selectivity);
            if (ev2 != null) {
                return evaluateSingle(ev2, h);
            }
        } else if (selectivity == SEL_ANY) {
            Vector dups = (Vector) h.get(ev.getId());
            for (int i = 0; i < dups.size(); i++) {
                fuegocore.notify.util.Log.writeFilter(
                        "Selecting triplet (any): " + i);
                EventTriplet ev2 = (EventTriplet) dups.elementAt(i);
                if (ev2 != null) {
                    fuegocore.notify.util.Log.writeFilter(
                            "Triplet found (any): " + i);
                    if (evaluateSingle(ev2, h)) {
                        return true;
                    }
                }
            }
            return false; // no match found
        } else if (selectivity == SEL_ALL) {
            Vector dups = (Vector) h.get(ev.getId());
            for (int i = 0; i < dups.size(); i++) {
                fuegocore.notify.util.Log.writeFilter(
                        "Selecting triplet (all): " + i);
                EventTriplet ev2 = (EventTriplet) dups.elementAt(i);
                if (ev2 != null) {
                    fuegocore.notify.util.Log.writeFilter(
                            "Triplet found (all): " + i);
                    if (!evaluateSingle(ev2, h)) {
                        return false;
                    }
                }
            }
            return true; // all were true
        }
        return false;
    }

    /** Evaluate this filter with an event triplet. Takes also a TripletTable
     * to the all event triplets of the notification as a parameter in order
     * for field comparison constraints (extra computation required is not great).
     * @param ev the triplet to match against the filter
     * @param h TripletTable for all the triplets in the notification to be matched
     * @return true if the condition holds
     */
    public boolean evaluateSingle(EventTriplet ev, TripletTable h) {
        boolean ret = false;

        fuegocore.notify.util.Log.writeFilter("Filter: evaluating type: " +
                                              ev.getType() + " name: " +
                                              name);

        // We iterate over the set of conditional filters
        // it is enough that one of them matches

        if (filter_components_conditional == null ||
            filter_components_conditional.size() == 0) {
            return false;
        }

        if (!(name.trim().equals(ev.getName().trim()) &&
              type == ev.getType())) { // check that the type is correct
            return false;
        }

        // General objects can only be tested for existence, so if we
        // found the correct type, it matches
        if (type == EventTriplet.TYPE_SERIALIZED) {
            return true;
        }

        for (int j = 0; j < filter_components_conditional.size(); j++) {
            // extract new expression vector for matching
            Disjunct conj = (Disjunct) filter_components_conditional.
                            elementAt(j);

            if (conj.evaluateExpression(ev, h)) {
                return true; // if true return and break matching
            }
        }
        return false; // no match
    }
}
