/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.filter.disjuncts;

import java.text.*;
import java.util.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.filter.predicates.*;
import fuegocore.notify.util.*;

/**
 * Implementation of a date disjunct
 */

public class DateDisjunct extends Disjunct {

    public DateDisjunct(Filter filter) {
        super(filter);
    }

    /** Decomposes the filter into its components. Dates.
     * @param filter2 the filter string to decompose
     */
    public void decompose(String filter2) {
        int selectivity = Filter.SEL_NORMAL;

        if (filter2 == null) {
            fuegocore.notify.util.Log.writeFilter("Filter was null: " + filter2);
            return;
        }

        StringTokenizer st = new StringTokenizer(filter2, Filter.FILTER_DIVISOR);
        while (st.hasMoreTokens()) {
            String subfilter = st.nextToken();
            try {
                StringTokenizer st2 = new StringTokenizer(subfilter.trim(),
                        Filter.OPERAND_DIVISOR);
                String operand = st2.nextToken();

                String opval = "";
                Date d = null;
                DatePredicate de = null;
                double dv = 0;

                opval = st2.nextToken();
                while (st2.hasMoreTokens()) {
                    opval += " " + st2.nextToken();

                }
                if (opval != null && !opval.startsWith("$") &&
                    operand.indexOf(Filter.SELECT) == -1) { // clearly not a date in this case
                    try {
                        d = DateFormat.getInstance().parse(opval);
                    } catch (Exception e) {
                        d = Auxil.parseDate(opval);
                        if (d == null) {
                            e.printStackTrace();
                            operand = Filter.EXISTENCE;
                        }
                    }
                }

                // Add this subcomponent to this filter's components
                if (operand.indexOf(Filter.SELECT) != -1) {
                    // select constraint
                    if (opval.indexOf(Filter.SELECT_ANY) != -1) {
                        selectivity = Filter.SEL_ANY;
                    } else if (opval.indexOf(Filter.SELECT_ALL) != -1) {
                        selectivity = Filter.SEL_ALL;
                    } else {
                        // check if the argument is a valid integer
                        Integer ik = null;
                        try {
                            ik = new Integer(opval);
                        } catch (Exception e) {
                            ik = null;
                        }
                        if (ik == null) {
                            // no valid argument: no selection
                        } else { // set selectivity
                            if (ik.intValue() > 0) {
                                selectivity = ik.intValue();
                            }
                        }
                    }
                    if (filter != null) {
                        filter.setSelectivity(selectivity);
                    }
                } else {
                    if (!opval.startsWith("$")) { // clearly not a number in this case
                        de = new DatePredicate(operand.trim(), d);
                    } else {
                        de = new DatePredicate(operand.trim(), opval);
                    }
                    fuegocore.notify.util.Log.writeFilter("Processing: " +
                            operand +
                            " " +
                            opval);

                    this.put(de.getId(), de);
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        getCoverId(); // update cover id cache
    }

    /** Evaluates the conjunct given the EventTriplet
     * @param et the EventTriplet
     * @param h TripletTable containing all triplets
     * @return true if the condition holds
     */
    public boolean evaluateExpression(EventTriplet et, TripletTable h) {
        Date at = (Date) et.getValue();

        // We evaluate all subfilters

        Double b = null;
        long a = at.getTime();

        // We evaluate all subfilters
        // We break as soon as one is false (AND)
        for (Enumeration i = this.elements(); i.hasMoreElements(); ) {
            DatePredicate de = (DatePredicate) i.nextElement();
            if (de != null) {
                fuegocore.notify.util.Log.writeFilter("Evaluating:" + a + " " +
                        de.toString());
                long se = 0;
                // Check if dollar sign field is used (loose checking)
                if (de.getValue() != null && de.getValue().startsWith("$")) {
                    fuegocore.notify.util.Log.writeFilter("Current time");
                    if (de.getValue().equals(Filter.CURRENT_TIME)) {
                        se = System.currentTimeMillis();
                    } else {
                        fuegocore.notify.util.Log.writeFilter(
                                "External date constraint");
                        Object obj = (Object) transformValue(de.getValue().trim(),
                                type, h);

                        if (obj == null) {
                            return false;
                        }

                        if (obj instanceof Date) {
                            se = (long) ((Date) obj).getTime();
                        } else {
                            return false; // incorrect type
                        }
                    }
                } else {
                    se = de.getDateValue().getTime();
                    // match using the predicate
                }
                if (!de.match((double) se, (double) a)) {
                    Log.writeFilter("Match returned false: compared" + se +
                                    " to: " +
                                    a);
                    return false;
                }
            }
        }
        return true;
    }

}
