/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.filter.disjuncts;

import java.util.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.filter.predicates.*;
import fuegocore.notify.util.*;

/**
 * Abstract class for disjuncts
 */

public abstract class Disjunct extends Hashtable implements IDisjunct, IFBase,
        Comparable {
    int type; // type of Disjunct (TYPE_STRING,.. in EventTriplet)
    String name; // Name of the field
    public static final int FAILED = 0;
    public static final int COVERED = 1;
    public static final int MODIFIED = 2;
    public static final int EXISTENCE = 3;
    Filter filter;
    String coveridcache = null;

    /** Constructor
     * @param type the type of this Disjunct
     */
    public Disjunct(int type) {
        this.type = type;
    }

    public Disjunct(Filter filter) {
        this.filter = filter;
        this.name = filter.getName();
        this.type = filter.getType();
    }

    public Vector getTid() {
        return null;
    }

    public void addTid(Object obj) {}

    public String getName() {
        return name;
    }

    public Filter getFilter() {
        return filter;
    }

    /**
     * Returns the size of the object
     * If filter the # disjuncts, if disjunct the # predicates
     * @return the size
     */
    public int getSize() {
        return this.size();
    }

    public boolean noSelectivity() {
        if (filter.getSelectivity() == Filter.SEL_NORMAL ||
            filter.getSelectivity() == 1) {
            return true;
        }
        return false;
    }

    /** Constructor
     * @param name the name of the element
     * @param type the type of this Disjunct
     */
    public Disjunct(String name, int type) {
        this.type = type;
        this.name = name;
    }

    public Disjunct() {}

    /**
     *  Returns the type of this Disjunct.
     * @return type
     */
    public int getType() {
        return type;
    }

    /** Get field/attribute identifier (name/type)
     * @return String the identifier
     */
    public String getId() {
        return name + "/" + type;
    }

    /**
     * Matches the object with the given object
     * @param et2 the EventTriplet object to match, may be null
     * @param noti  the TripletTable, may not be null
     * @return boolean true if the match succeeded
     */
    public boolean match(Object et2, Object noti) {
        if (!(et2 instanceof EventTriplet)) {
            System.out.println("Disjunct type error 1");
            return false;
        }
        if (!(noti instanceof TripletTable)) {
            System.out.println("Disjunct type error 2");
            return false;
        }
        EventTriplet et = (EventTriplet) et2;
        TripletTable h = (TripletTable) noti;

        if (filter.getSelectivity() == Filter.SEL_NORMAL) { // normal selectivity
            return evaluateExpression(et, h);
        } else if (filter.getSelectivity() > 0) { // specific triplet
            EventTriplet ev2 = (EventTriplet) h.get(et.getId(),
                    filter.getSelectivity());
            if (ev2 != null) {
                return evaluateExpression(ev2, h);
            }
        } else if (filter.getSelectivity() == Filter.SEL_ANY) {
            Vector dups = (Vector) h.get(et.getId());
            for (int i = 0; i < dups.size(); i++) {
                fuegocore.notify.util.Log.writeFilter(
                        "Selecting triplet (any): " + i);
                EventTriplet ev2 = (EventTriplet) dups.elementAt(i);
                if (ev2 != null) {
                    fuegocore.notify.util.Log.writeFilter(
                            "Triplet found (any): " + i);
                    if (evaluateExpression(ev2, h)) {
                        return true;
                    }
                }
            }
            return false; // no match found
        } else if (filter.getSelectivity() == Filter.SEL_ALL) {
            Vector dups = (Vector) h.get(et.getId());
            for (int i = 0; i < dups.size(); i++) {
                fuegocore.notify.util.Log.writeFilter(
                        "Selecting triplet (all): " + i);
                EventTriplet ev2 = (EventTriplet) dups.elementAt(i);
                if (ev2 != null) {
                    fuegocore.notify.util.Log.writeFilter(
                            "Triplet found (all): " + i);
                    if (!evaluateExpression(ev2, h)) {
                        return false;
                    }
                }
            }
            return true; // all were true
        }
        return false;
    }

    /** Merges the current object with the given object
     * @param obj the object to be merged with the current object
     * @return Object the merged result
     */
    public Object mergeObject(Object obj) {
        Disjunct d = (Disjunct)this.clone();
        if (d.merge(obj)) {
            return d;
        } else {
            return null;
        }
    }

    /** Merges the object with the given object
     * @param obj the object to be merged
     * @return boolean true if the merge succeeded
     */
    public boolean merge(Object obj) {
        if (!(obj instanceof Filter)) {
            return false;
        }
        if (coverAndMerge(this, (Disjunct) obj) == FAILED) {
            return false;
        }
        return true;
    }

    /**
     *  Performs a conjunctive merge (no disjunctions).
     *  Used to benchmark the more complex disjunctive merging algorithm.
     *  @param test if true no modifications are made
     *  @param a hashtable with predicates
     *  @return true if the merge was possible
     **/
    public boolean conjunctiveMerge(boolean test, Hashtable a) {
        return conjunctiveMerge(test, a, this);
    }

    /**
     *  Performs a conjunctive merge (no disjunctions).
     *  Used to benchmark the more complex disjunctive merging algorithm.
     *  @param test if true no modifications are made
     *  @param a hashtable with predicates
     *  @param b hashtable with predicates
     *  @return true if the merge was possible
     **/
    public boolean conjunctiveMerge(boolean test, Hashtable a, Hashtable b) {
        int count = 0, count2 = 0;
        IAtomicPredicate chosen = null;
        IAtomicPredicate chosen_b = null;

        if (a == null || b == null) {
            return false;
        }

        coveridcache = null;
        boolean identical[] = new boolean[a.size()];
        boolean covers[] = new boolean[b.size()];
        int k = 0;

        for (Enumeration i = b.elements(); i.hasMoreElements(); ) {
            IAtomicPredicate e_b = (IAtomicPredicate) i.nextElement();
            boolean idc = false;
            if (e_b != null) {
                int c = 0;
                for (Enumeration j = a.elements(); j.hasMoreElements(); ) {
                    IAtomicPredicate e_a = (IAtomicPredicate) j.nextElement();
                    if (e_a != null && e_b.identical(e_a)) {
                        if (!identical[c]) {
                            count++;
                        }
                        idc = true;
                        identical[c] = true;
                    }
                    c++;
                }
            }
            if (!idc) {
                chosen_b = e_b;
            }
            k++;
        }

        // find the second part of the distinctive atom
        if (chosen_b != null) {
            int c = 0;
            for (Enumeration j = a.elements(); j.hasMoreElements(); ) {
                IAtomicPredicate e_a = (IAtomicPredicate) j.nextElement();
                if (!identical[c] && !chosen_b.identical(e_a) &&
                    chosen_b.perfectMergeRuleForExists(e_a) ||
                    chosen_b.perfectMergeRuleForModified(e_a, false)) {
                    chosen = e_a;
                    break;
                }
                c++;
            }
        }

        // if all but one of the covered predicates are identical
        // with the covering predicates, we use perfect filtering
        // rules to find if a simple existence tests is enough,
        // or if we can remove some of the predicates
        // the sets need to be of the same length in order for the merge
        // to work

        if (count == a.size() - 1 && a.size() == b.size() && chosen != null &&
            chosen_b != null) {

            // ---

            // Now, attempt to merge the not-identical predicate with the covering predicate set (b)
            if (chosen_b.perfectMergeRuleForExists(chosen)) {
                // e_a can be merged with e_b
                if (b.size() == 1) {
                    // We replace the e_b predicate using existence test
                    if (!test) {
                        b.clear(); // clear hashtable
                        IAtomicPredicate exists = CreateExistsPredicate();
                        b.put(exists.getId(), exists);
                        Log.writeFilter("Adding existence test: " +
                                        chosen_b.getId());
                        //System.out.println("Adding existence test: " + chosen_b.getId());
                    }
                    // We return true
                    return true;
                } else {
                    // The predicate merge resulted in existence test,
                    // however, there are other predicates as well so
                    // we remove the predicate altogether. Since we
                    // have already determined that the other predicates
                    // are duplicates, we return true.
                    if (!test) {
                        b.remove(chosen_b.getId());
                        Log.writeFilter("Removing: " + chosen_b.getId());
                        //System.out.println("Removing: " + chosen_b.getId());
                    }
                    return true;
                }
            } else if (chosen_b.perfectMergeRuleForModified(chosen, !test)) {
                // the predicates were merged and resulted in a modified predicate
                return true;
            }
        } else if (count == a.size()) { // end if count
            return true; // the clauses are identical
        }
        /** END: PERFECT MERGING RULES FOR EXISTENCE TEST REPLACEMENT */

        return false; // no merge was possible
    }

    /** Returns true if this Disjunctive predicate set is covered by another (b).
     * @param b covering hashtable
     * @return true if this Disjunction was covered
     */
    public int coverAndMerge(Hashtable b) {
        return coverAndMerge(this, b);
    }

    /**
     * Returns true if this predicate set overlaps with the given set
     * @param obj Object
     * @return boolean
     */
    public boolean overlaps(Object obj) {
        int count = 0;

        if (!(obj instanceof Hashtable)) {
            return false;
        }

        Hashtable a = (Hashtable) obj;

        Hashtable s, l;

        // Find the smallest set
        if (a.size() > this.size()) {
            s = this;
            l = a;
        } else {
            s = a;
            l = this;
        }
        int max = 0;
        for (Enumeration i = s.elements(); i.hasMoreElements(); ) {
            IAtomicPredicate e_b = (IAtomicPredicate) i.nextElement();
            if (e_b != null) {
                for (Enumeration j = l.elements(); j.hasMoreElements(); ) {
                    IAtomicPredicate e_a = (IAtomicPredicate) j.nextElement();
                    if (e_a != null && e_b.overlaps(e_a)) {
                        count++;
                    }
                    max++;
                }
            }
        }
        // If all predicates overlap return true
        // System.out.println("Disj: " + count + " max " + max);
        if (count == max) {
            return true;
        }
        return false;
    }

    /**
     *  Returns the cost of overlap
     * @param obj  a hashtable representing the input set
     * @return double the cost in [0,1], 0 is orthogonal, 1 is identical
     */
    public double getOverlapCost(Object obj) {
        int count = 0;

        if (!(obj instanceof Hashtable)) {
            return 0;
        }

        Hashtable a = (Hashtable) obj;

        // we use the cost to estimate overlap
        if (a.size() == 1 && this.size() == 1) {
            Enumeration i = a.elements();
            IAtomicPredicate e_a = (IAtomicPredicate) i.nextElement();
            i = this.elements();
            IAtomicPredicate e_b = (IAtomicPredicate) i.nextElement();
            if (e_a.overlaps(e_a)) {
                double d = e_a.getCost();
                double d2 = e_b.getCost();
                double r = (d * d2) / (100 * 100);
                return r;
            }
            return 0;
        }

        Hashtable s, l;

        // Find the smallest set
        if (a.size() > this.size()) {
            s = this;
            l = a;
        } else {
            s = a;
            l = this;
        }
        int max = 0;
        for (Enumeration i = s.elements(); i.hasMoreElements(); ) {
            IAtomicPredicate e_b = (IAtomicPredicate) i.nextElement();
            if (e_b != null) {
                for (Enumeration j = l.elements(); j.hasMoreElements(); ) {
                    IAtomicPredicate e_a = (IAtomicPredicate) j.nextElement();
                    if (e_a != null && e_b.overlaps(e_a)) {
                        count++;
                    }
                    max++;
                }
            }
        }
        // If all predicates overlap return true
        if (count == max) {
            return 1;
        }
        return 0;
    }

    /** Returns true if a disjunctive predicate set (a) is covered by another (b).
     * This means that each element in a must be covered by b. Note that this
     * method also performs perfect merging of predicates in the following way:
     * if all but one predicates are identical, a perfect merge operation is attempted
     * which may result in 1) removal of the distinct predicate 2) replace whole Disjunct
     * with an existence test 3) nothing.  If no operation is performed or there
     * are no identical predicates, the covering and merging is determined by
     * counting the number of covering predicates which must be equal to the
     * number of predicates in b for the method to return true.
     * @param a hashtable representing the covered set
     * @param b hashtable representing the covering set
     * @return FAILED if failure
     */
    public int coverAndMerge(Hashtable a, Hashtable b) {
        int count = 0, count2 = 0;
        IAtomicPredicate chosen = null;
        IAtomicPredicate chosen_b = null;

        if (a == null || b == null) {
            return FAILED;
        }

        coveridcache = null;

        boolean identical[] = new boolean[a.size()];
        boolean covers[] = new boolean[b.size()];
        boolean idc = false;
        int k = 0;

        for (Enumeration i = b.elements(); i.hasMoreElements(); ) {
            IAtomicPredicate e_b = (IAtomicPredicate) i.nextElement();
            idc = false;

            if (e_b != null) {
                int c = 0;

                for (Enumeration j = a.elements(); j.hasMoreElements(); ) {

                    IAtomicPredicate e_a = (IAtomicPredicate) j.nextElement();

                    if (e_a != null && e_b.identical(e_a)) {
                        if (!identical[c]) {
                            count++;
                        }
                        if (!covers[k]) {
                            count2++;
                        }

                        covers[k] = true;
                        idc = true;
                        identical[c] = true;

                        if (chosen != null) { // we have found the distinctive predicate if one exists
                            break; // we assume that Disjuncts do not have duplicates
                        }
                    }

                    if (e_a != null && e_b.covers(e_a)) {
                        if (!covers[k]) {
                            count2++;
                        }
                        covers[k] = true;
                        if (chosen != null) { // we have found the distinctive predicate if one exists
                            break;
                        }
                    }
                    c++;
                }
            }
            if (!idc) {
                chosen_b = e_b;
            }
            k++;
        }

        // find the second part of the distinctive atom
        if (chosen_b != null) {
            int c = 0;
            for (Enumeration j = a.elements(); j.hasMoreElements(); ) {
                IAtomicPredicate e_a = (IAtomicPredicate) j.nextElement();
                if (!identical[c] && !chosen_b.identical(e_a) &&
                    chosen_b.perfectMergeRuleForExists(e_a) ||
                    chosen_b.perfectMergeRuleForModified(e_a, false)) {
                    chosen = e_a;
                    break;
                }
                c++;
            }
        }
        /** START: PERFECT MERGING RULES FOR EXISTENCE TEST REPLACEMENT */

        if (count == a.size() - 1 && a.size() == b.size() && chosen != null &&
            chosen_b != null) {
            // ---
            // Now, attempt to merge the not-identical predicate with the covering predicate set (b)
            // Actually we should test all for the possibility of existence test.
            // This merges with the first possibility

            // ---
            // Now, attempt to merge the not-identical predicate with the covering predicate set (b)
            if (chosen_b.perfectMergeRuleForExists(chosen)) {
                // e_a can be merged with e_b
                if (b.size() == 1) {
                    // We replace the e_b predicate using existence test
                    b.clear(); // clear hashtable
                    IAtomicPredicate exists = CreateExistsPredicate();
                    b.put(exists.getId(), exists);
                    Log.writeFilter("Adding existence test: " + chosen_b.getId());

                    //System.out.println("Adding existence test: " + chosen_b.getId());
                    // We return true
                    return EXISTENCE;
                } else {
                    // The predicate merge resulted in existence test,
                    // however, there are other predicates as well so
                    // we remove the predicate altogether. Since we
                    // have already determined that the other predicates
                    // are duplicates, we return true.
                    b.remove(chosen_b.getId());
                    Log.writeFilter("Removing: " + chosen_b.getId());
                    //System.out.println("Removing: " + chosen_b.getId());
                    return MODIFIED;
                }
            } else if (chosen_b.perfectMergeRuleForModified(chosen, true)) {
                // the predicates were merged and resulted in a modified predicate
                return MODIFIED;
            }

        } // end if count

        /** END: PERFECT MERGING RULES FOR EXISTENCE TEST REPLACEMENT */

        // If all elements were matched
        if (count2 == b.size()) {
            return COVERED;
        }
        return FAILED;
    }

    /** Auxiliary method to sort a hashtable of IAtomicPredicates
     * @return Vector the sorted hashtable
     */
    public Vector sortHashtable() {
        Collection c = this.values();
        Vector v = new Vector(c);

        /* Sort */
        for (int i = 0; i < v.size(); i++) {
            IAtomicPredicate a1 = (IAtomicPredicate) v.elementAt(i);
            String val1 = a1.toString();
            for (int j = i; j < v.size(); j++) {
                IAtomicPredicate a2 = (IAtomicPredicate) v.elementAt(j);
                String val2 = a2.toString();

                if (val1.compareTo(val2) > 0) {
                    v.set(i, a2);
                    v.set(j, a1);
                }
            }
        }
        return v;
    }

    /**
     * Returns the filter string
     * @return String
     */
    public String getFilterString() {
        // We sort the hashtable
        Vector v = sortHashtable();
        String ret = "";

        // add the selectivity constraint as a prefix (not sorted)
        String sel = filter.getSelectivityString();
        if (sel != null) {
            ret += sel + "&"; // it does not matter where it is...
        }

        for (int j = 0; j < v.size(); j++) {
            IAtomicPredicate e_b = (IAtomicPredicate) v.elementAt(j);
            if (e_b != null) {
                ret += e_b.toString();
            }
            if (j < v.size() - 1) {
                ret += Filter.FILTER_DIVISOR;
            }
        }
        return ret;
    }

    /**
     * Covering test
     * @param obj the covered object
     * @return true if covered
     */
    public boolean covers(Object obj) {
        if (!(obj instanceof Disjunct)) {
            System.out.println("Class error in Disjunct");
            return false;
        }
        Disjunct d = (Disjunct) obj;

        if (d.getType() == getType() && d.getName().equals(getName()) &&
            (d.getFilter().getSelectivity() == filter.getSelectivity() ||
             filter.getSelectivity() == Filter.SEL_ANY)) {
            return covers(d, this);
        }
        return false;
    }

    /**
     * Covering test: is the given object covered
     * @param obj the object that covers
     * @return true if the current object was covered
     */
    public boolean isCovered(Object obj) {
        if (!(obj instanceof Disjunct)) {
            System.out.println("Class error in Disjunct");
            return false;
        }
        Disjunct d = (Disjunct) obj;
        if (d == null) {
            return false;
        }

        if (d.getType() == getType() && d.getName().equals(getName()) &&
            (d.getFilter().getSelectivity() == filter.getSelectivity() ||
             filter.getSelectivity() == Filter.SEL_ANY)) {
            return d.covers(this);
        }
        return false;
    }

    /**
     * Get id for hash lookups.
     * @return String the identifier
     */
    public String getCoverId() {
        if (coveridcache != null) {
            return coveridcache;
        }
        coveridcache = getId() + "/" + getFilterString();
        return coveridcache;
    }

    /**
     * Set the depth of this object in the poset tree.
     * @param i the depth to set
     */
    public void setDepth(int i) {}

    /**
     * Returns the depth of this object in the poset tree.
     * @return the depth as int
     */
    public int getDepth() {
        return 0;
    }

    /** returns true if this predicate set covers the given set
     * @param a hashtable representing the covered set
     * @return true if b covers a
     */
    public boolean covers(Hashtable a) {
        return covers(a, this);
    }

    /** returns true if a Disjunctive predicate set (a) is covered by another (b).
     * This means that each element in a must be covered by b.
     * @param a hashtable representing the covered set
     * @param b hashtable representing the covering set
     * @return true if b covers a
     */
    public boolean covers(Hashtable a, Hashtable b) {
        int count = 0, count2 = 0;
        if (a == null || b == null) {
            return false;
        }

        boolean covers[] = new boolean[b.size()];
        int k = 0;

        // now we iterate over the conditional clause and mark those that cover this
        for (Enumeration i = b.elements(); i.hasMoreElements(); ) {
            IAtomicPredicate e_b = (IAtomicPredicate) i.nextElement();
            if (e_b != null) {
                int c = 0;
                for (Enumeration j = a.elements(); j.hasMoreElements(); ) {
                    IAtomicPredicate e_a = (IAtomicPredicate) j.nextElement();
                    if (e_a != null && e_b.covers(e_a)) {
                        //System.out.println("covered: " + e_a.toString());
                        if (!covers[k]) {
                            count2++;
                        }
                        covers[k] = true;
                        break;
                    }
                    c++;
                }
            }
            k++;
        }
        // If all elements were matched
        if (count2 == b.size()) {
            return true;
        }

        return false;
    }

    /** Creates an exists predicate of the same type as the filter field/attribute
     * @return IAtomicPredicate the exists predicate
     */
    public IAtomicPredicate CreateExistsPredicate() {
        if (getType() == EventTriplet.TYPE_STRING) {
            return new StringPredicate(Filter.EXISTENCE, "");
        } else if (getType() == EventTriplet.TYPE_INT) {
            return new NumberPredicate(Filter.EXISTENCE, 0);
        } else if (getType() == EventTriplet.TYPE_DOUBLE) {
            return new NumberPredicate(Filter.EXISTENCE, 0.0);
        } else if (getType() == EventTriplet.TYPE_DATE) {
            return new DatePredicate(Filter.EXISTENCE, "");
        } else if (getType() == EventTriplet.TYPE_BOOLEAN) {
            return new BooleanPredicate(Filter.EXISTENCE, 0);
        } else {
            System.out.println("Fatal error, unknown type:" + getType());
        }
        return null;
    }

    /** Looks up an external field from the same notification
     * @param s the original value (placeholder)
     * @param type the type
     * @param h the TripletTable
     * @return Object the replaced value (using the placeholder)*/
    public Object transformValue(String s, int type, TripletTable h) {
        /* evaluate FIELD element */
        if (s.startsWith("$")) {
            String fs = s.substring(1, s.length());
            EventTriplet et = (EventTriplet) h.getFirst(EventTriplet.formatId(
                    fs,
                    type));
            fuegocore.notify.util.Log.writeFilter("Got: " + et + " type: " +
                                                  type);
            if (et != null && et.getType() == type) {
                Object o = null;
                try {
                    o = et.getValue();
                } catch (Exception e) {
                    fuegocore.notify.util.Log.writeFilter(
                            "Exception in Filter: " + e);
                    e.printStackTrace();
                }
                return o;
            }
        }
        return null;
    }

    /** Returns the equality predicate of this disjunct if one exists.
     * Clearly any proper disjunct contains only one equality predicate,
     * since two different equality predicates are orthogonal.
     * @return String the equality predicate or null
     */
    public String getEqualityPredicate() {
        for (Enumeration i = this.elements(); i.hasMoreElements(); ) {
            IAtomicPredicate ip = (IAtomicPredicate) i.nextElement();
            if (ip.getEqualityValue(getType()) != null) {
                return ip.getEqualityValue(getType());
            }
        }
        return null;
    }

    /** Returns the cost of the disjunct (generality). Does not take
     *  hierarchical selectivity into account (this is in filter).
     *  Higher value denotes more specific filters.
     * @return int the cost
     */
    public int getCost() {
        if (size() == 0) {
            return 0;
        }
        int cost = 0;
        for (Enumeration i = this.elements(); i.hasMoreElements(); ) {
            IAtomicPredicate ip = (IAtomicPredicate) i.nextElement();
            cost += ip.getCost();
        }
        // Divide by the number of conjuncts since they are more restricting
        cost = cost / this.size();
        return cost;
    }

    /**
     * Comparison based on generality in order to optimize matching.
     * @param other the other object to compare to
     * @return int the comparison value
     */
    public int compareTo(Object other) {
        if (other instanceof Disjunct) {
            int a = ((Disjunct) other).getCost();
            return a - getCost();
        } else {
            throw new ClassCastException();
        }
    }
}
