/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.filter.disjuncts;

import java.util.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.predicates.*;
import fuegocore.notify.util.*;

/**
 * Interface for disjuncts
 */

public interface IDisjunct {
    /**
     *  Returns the type of this conjunct.
     * @return type
     */
    public int getType();

    /**
     *  Performs a conjunctive merge (no disjunctions).
     *  Used to benchmark the more complex disjunctive merging algorithm.
     *  @param test if true no modifications are made
     *  @param a hashtable with predicates
     *  @return true if the merge was possible
     **/
    public boolean conjunctiveMerge(boolean test, Hashtable a);

    /** returns true if this conjunctive predicate set is covered by another (b).
     * @param b covering hashtable
     * @return int COVERED if b covers a
     */
    public int coverAndMerge(Hashtable b);

    /** Returns true if a conjunctive predicate set (a) is covered by another (b).
     * This means that each element in a must be covered by b. Note that this
     * method also performs perfect merging of predicates in the following way:
     * if all but one predicates are identical, a perfect merge operation is attempted
     * which may result in 1) removal of the distinct predicate 2) replace whole conjunct
     * with an existence test 3) nothing.  If no operation is performed or there
     * are no identical predicates, the covering and merging is determined by
     * counting the number of covering predicates which must be equal to the
     * number of predicates in b for the method to return true.
     * @param a hashtable representing the covered set
     * @param b hashtable representing the covering set
     * @return int COVERED if b covers a
     */
    public int coverAndMerge(Hashtable a, Hashtable b);

    /** returns true if this predicate set covers the given set
     * @param a hashtable representing the covered set
     * @return true if the current object covers a
     */
    public boolean covers(Hashtable a);

    /** returns true if this predicate set overlaps with the given set
     * @param a hashtable representing the input set
     * @return true if the current object overlaps with a
     */
    public boolean overlaps(Object a);

    /** returns true if a conjunctive predicate set (a) is covered by another (b).
     * This means that each element in a must be covered by b.
     * @param a hashtable representing the covered set
     * @param b hashtable representing the covering set
     * @return true if b covers a
     */
    public boolean covers(Hashtable a, Hashtable b);

    /** Creates an exists predicate of the same type as the filter field/attribute
     * @return IAtomicPredicate the exists predicate
     */
    public IAtomicPredicate CreateExistsPredicate();

    /** Looks up an external field from the same notification
     * @param s the original value (placeholder)
     * @param type the type
     * @param h the TripletTable
     * @return Object the replaced value (using the placeholder)*/
    public Object transformValue(String s, int type, TripletTable h);

    /** Decomposes the filter into its components.
     * @param filter2 the filter string to decompose
     */
    public void decompose(String filter2);

    /** Evaluates the conjunct given the EventTriplet
     * @param et the EventTriplet
     * @param h TripletTable containing all triplets
     * @return true if the condition holds
     */
    public boolean evaluateExpression(EventTriplet et, TripletTable h);

    /** Returns the equality predicate of this disjunct if one exists.
     * Clearly any proper disjunct contains only one equality predicate,
     * since two different equality predicates are orthogonal.
     * @return String the equality predicate or null
     */
    public String getEqualityPredicate();

}
