/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.filter.disjuncts;

import java.util.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.filter.predicates.*;
import fuegocore.notify.util.*;

/**
 * Implements the number disjunct
 */
public class NumberDisjunct extends Disjunct {

    public NumberDisjunct(Filter filter) {
        super(filter);
    }

    /** Decomposes the filter into its components.
     * @param filter2 the filter string to decompose
     */
    public void decompose(String filter2) {
        int selectivity = Filter.SEL_NORMAL;

        if (filter2 == null) {
            fuegocore.notify.util.Log.writeFilter("Filter was null: " + filter2);
            return;
        }

        StringTokenizer st = new StringTokenizer(filter2, Filter.FILTER_DIVISOR);
        fuegocore.notify.util.Log.writeFilter("filter2:(" + filter2 + ")");

        while (st.hasMoreTokens()) {
            String subfilter = st.nextToken();
            fuegocore.notify.util.Log.writeFilter("Working on subfilter: " +
                                                  subfilter);
            try {
                StringTokenizer st2 = new StringTokenizer(subfilter.trim(),
                        Filter.OPERAND_DIVISOR);
                String operand = st2.nextToken();

                String opval = "";
                String opval2 = "";
                Double d = null;
                Double d2 = null;
                NumberPredicate de = null;

                try {
                    opval = st2.nextToken();

                    if (st2.hasMoreTokens()) {
                        opval2 = st2.nextToken();

                    }
                    if (opval != null && operand.indexOf(Filter.SELECT) == -1 &&
                        !opval.startsWith("$")) { // clearly not a number in this case
                        d = new Double(opval);
                    }

                    if (opval2 != null && operand.indexOf(Filter.OP_RNG) != -1 &&
                        !opval2.startsWith("$")) {
                        d2 = new Double(opval2);
                    }
                } catch (Exception e) {
                    operand = Filter.EXISTENCE;
                    d = null;
                    d2 = null;
                }

                fuegocore.notify.util.Log.writeFilter("Adding op:" + operand +
                        " val: " +
                        opval);

                if (operand.indexOf(Filter.SELECT) != -1) {
                    // select constraint
                    if (opval.indexOf(Filter.SELECT_ANY) != -1) {
                        selectivity = Filter.SEL_ANY;
                    } else if (opval.indexOf(Filter.SELECT_ALL) != -1) {
                        selectivity = Filter.SEL_ALL;
                    } else {
                        // check if the argument is a valid integer
                        Integer ik = null;
                        try {
                            ik = new Integer(opval);
                        } catch (Exception e) {
                            ik = null;
                        }
                        if (ik == null) {
                            // no valid argument: no selection
                        } else { // set selectivity
                            if (ik.intValue() > 0) {
                                selectivity = ik.intValue();
                            }
                        }
                    }
                    if (filter != null) {
                        filter.setSelectivity(selectivity);
                    }
                } else {
                    if (!opval.startsWith("$") && d != null) { // clearly not a number in this case
                        if (d2 != null) { // two parameter command
                            de = new NumberPredicate(operand.trim(),
                                    d.doubleValue(),
                                    d2.doubleValue());
                        } else { // one parameter command
                            de = new NumberPredicate(operand.trim(),
                                    d.doubleValue());
                        }
                    } else {
                        de = new NumberPredicate(operand.trim(), opval); // store $value
                    }

                    fuegocore.notify.util.Log.writeFilter("Processing: " +
                            operand +
                            " " +
                            opval);

                    // Add this subcomponent to this conjunct
                    this.put(de.getId(), de);
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        getCoverId(); // update cover id cache
    }

    /** Evaluates the conjunct given the EventTriplet
     * @param et the EventTriplet
     * @param h TripletTable containing all triplets
     * @return true if the condition holds
     */
    public boolean evaluateExpression(EventTriplet et, TripletTable h) {
        double a = 0;

        if (filter != null && filter.getType() == EventTriplet.TYPE_INT) {
            try {
                Integer i = (Integer) et.getValue();
                a = (double) i.intValue();
            } catch (Exception e) {
                e.printStackTrace();
            }
        } else {
            a = ((Double) et.getValue()).doubleValue();
        }
        // We evaluate all subfilters

        Double b = null;

        // We evaluate all subfilters
        // We break as soon as one is false (AND)
        for (Enumeration i = this.elements(); i.hasMoreElements(); ) {
            NumberPredicate de = (NumberPredicate) i.nextElement();
            if (de != null) {
                fuegocore.notify.util.Log.writeFilter("Evaluating:" + a + " " +
                        de.toString());
                double se = 0;
                // Check if dollar sign field is used (loose checking)
                if (de.getValue() != null && de.getValue().startsWith("$")) {
                    Log.writeFilter("dollar sign detect");
                    Object obj = (Object) transformValue(de.getValue().trim(),
                            type, h);

                    if (obj == null) {
                        return false;
                    }

                    if (obj instanceof Double) {
                        se = ((Double) obj).doubleValue();
                    } else if (obj instanceof Integer) {
                        se = (double) (((Integer) obj).intValue());
                    } else {
                        return false; // incorrect type
                    }
                } else {
                    se = de.getDValue();

                    // match using the predicate
                }
                if (!de.match(se, a)) {
                    return false;
                }
            }
        }
        return true;
    }

}
