/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.filter.disjuncts;

import java.util.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.filter.predicates.*;
import fuegocore.notify.util.*;

/**
 * Implements the String disjunct
 */

public class StringDisjunct extends Disjunct {

    public StringDisjunct(Filter filter) {
        super(filter);
    }

    /**
     * Decomposes the filter into its components.
     *
     * @param filter2 the filter string to decompose
     */
    public void decompose(String filter2) {
        int selectivity = Filter.SEL_NORMAL;

        if (filter2 == null) {
            fuegocore.notify.util.Log.writeFilter("Filter was null: " + filter2);
            return;
        }

        StringTokenizer st = new StringTokenizer(filter2, Filter.FILTER_DIVISOR);
        while (st.hasMoreTokens()) {
            String subfilter = st.nextToken();
            try {
                StringTokenizer st2 = new StringTokenizer(subfilter,
                        Filter.OPERAND_DIVISOR);
                String operand = st2.nextToken();
                String opval = "";
                try {
                    opval = st2.nextToken();
                    while (st2.hasMoreTokens()) {
                        opval += " " + st2.nextToken();
                    }
                } catch (Exception e) {
                    operand = Filter.EXISTENCE;
                }

                if (opval.trim() == "") {
                    operand = Filter.EXISTENCE;

                }
                fuegocore.notify.util.Log.writeFilter("Processing: " + operand +
                        " " +
                        opval);

                if (operand.indexOf(Filter.SELECT) != -1) {
                    // select constraint
                    if (opval.indexOf(Filter.SELECT_ANY) != -1) {
                        selectivity = Filter.SEL_ANY;
                    } else if (opval.indexOf(Filter.SELECT_ALL) != -1) {
                        selectivity = Filter.SEL_ALL;
                    } else {
                        // check if the argument is a valid integer
                        Integer ik = null;
                        try {
                            ik = new Integer(opval);
                        } catch (Exception e) {
                            ik = null;
                        }
                        if (ik == null) {
                            // no valid argument: no selection
                        } else { // set selectivity
                            if (ik.intValue() > 0) {
                                selectivity = ik.intValue();
                            }
                        }
                    }
                    if (filter != null) {
                        filter.setSelectivity(selectivity);
                    }
                } else {
                    fuegocore.notify.util.Log.writeFilter("Adding op:" +
                            operand +
                            " val: " + opval);
                    StringPredicate de = new StringPredicate(operand.trim(),
                            opval.trim());
                    // Add this subcomponent to this filter's components
                    this.put(de.getId(), de);
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        getCoverId(); // update cover id cache
    }

    /** Evaluates the conjunct given the EventTriplet
     * @param et the EventTriplet
     * @param h TripletTable containing all triplets
     * @return true if the condition holds
     */
    public boolean evaluateExpression(EventTriplet et, TripletTable h) {
        String v = (String) et.getValue();
        // We evaluate all subfilters
        // We break as soon as one is false (AND)

        for (Enumeration i = this.elements(); i.hasMoreElements(); ) {
            StringPredicate de = (StringPredicate) i.nextElement();
            String se = null;
            if (de != null) {
                // Check if dollar sign field is used
                if (de.getValue() != null && de.getValue().startsWith("$")) {
                    fuegocore.notify.util.Log.writeFilter("Dollar detected:" +
                            de.getValue());
                    se = (String) transformValue(de.getValue().trim(),
                                                 EventTriplet.TYPE_STRING, h);
                    fuegocore.notify.util.Log.writeFilter("Dollar got:" + se);
                    if (se == null) { // failed or was not used
                        return false;
                    }
                } else {
                    se = de.getValue().trim();
                    // match using the predicate
                }
                if (!de.match(se, v)) {
                    return false;
                }
            }
        }
        fuegocore.notify.util.Log.writeFilter("String filter returned true");
        // filter passed
        return true;
    }

}
