/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.filter.merge;

import java.util.*;

import fuegocore.notify.core.*;

/**
 * Interface for the merging service
 */

public interface IMergingService {
    /** Add a new merging algorithm to the framework.
     *
     * @param name Name of the algorithm
     * @param im Instance of the algorithm
     */
    public void register(String name, IMergingAlgorithm im);

    /** Remove a merging algorithm from the framework
     *
     * @param name the name of the algorithm
     */
    public void deregister(String name);

    /** Get covered count for forward-type of behaviour
     * @param t the input vector
     * @return int the covered count
     */
    public int getCoveredCountForward(Vector t);

    /** Get covered count for minimum cover set
     * @param t the input vector
     * @return int[] the covered count
     */
    public int[] getCoveredCount(Vector t);

    /** Get non-covered vector
     * @param t the input vector
     * @return Vector the output
     */
    public Vector getNonCovered(Vector t);


    /**
     * Returns the set of templates that are not covered.
     * @param v1 Vector
     * @param v2 Vector
     * @return Vector
     */
    public Vector getNotCovered(Vector v1, Vector v2);


    /** Returns true if the set of templates v1 covers the second
     *  set of templates v2.  This is based on counting covered templates.
     *  All templates in v2 need to be covered by at least one template
     *  in v1. If all templates in v2 are covered by some template in v1
     *  the set v1 covers v2.
     * @param v1 the covering set
     * @param v2 the covered set
     * @return true if v1 covers v2
     */
    public boolean covers(Vector v1, Vector v2);

    /** Returns true if the set the given template is covered by the
     *  given set of templates.
     *  This is based on counting covered templates.
     * @param v1 the covering set
     * @param v2 the covered template
     * @return true if v1 covers v2
     */
    public boolean covers(Vector v1, Template v2);

    /**
     * Returns true if the set the given template overlaps with the
     * given set of templates.
     * @param v1 Vector
     * @param v2 Template
     * @return boolean
     */
    public boolean overlaps(Vector v1, Template v2);

}
