/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.filter.merge;

import java.util.*;

import fuegocore.notify.core.*;
import fuegocore.notify.util.*;

/**
 * This class implements the imperfect merging algorithm
 */
public class ImperfectMerger implements IMergingAlgorithm {
    public static boolean COVERCHECK = true;


    /**
     * If set to false, filter covering is not checked after merging
     * @param check boolean
     */
    public void setCoverCheck(boolean check) {
        COVERCHECK = check;
    }

    /** Returns the set of perfectly merged templates for a single template
     *  We count the number of identical fields.  For perfect filtering,
     *  the mergeable filters must have the same structure.
     *
     *  Updates the given set of templates.
     *  @param t the set of templates
     *  @param tp the template to be incrementally merged
     *  @param merge boolean variable whether to actually do the merge or not
     *  @return index that was changed in the given vector
     */
    public int mergeSet(Vector t, Template tp, boolean merge) {
        for (int i = 0; i < t.size(); i++) {
            Template t2 = (Template) t.elementAt(i);
            if (t2 == null) {
                continue;
            }
            if (t2.imperfectlyMergeableStructure(tp)) {
                if (merge) {
                    t2 = t2.imperfectMerge(tp);
                }
                return i;
            }
        }
        return -1;
    }

    /** Returns the set of imperfectly merged templates for a vector
     *  of templates. We count the number of identical fields.  For perfect filtering,
     *  the mergeable filters must have the same structure.
     * @param t Vector of templates to be perfectly merged
     * @return Vector containing the merge result
     */
    public Vector mergeSet(Vector t) {
        Vector merged = new Vector();

        if (t == null || t.size() < 1) {
            return merged;
        }

        int kk = 0;
        int kk2 = 0;
        int mergeCount = 0;

        try {
            for (int i = 0; i < t.size(); i++) {
                Template t2 = (Template) t.elementAt(i);
                merged.add(t2.clone());
            }
        } catch (Exception e) {
            e.printStackTrace();
            return t;
        }

        if (t.size() == 1) {
            return merged;
        }

        Vector removeSet = new Vector();

        boolean skiplist[] = new boolean[merged.size()];
        int cc = 0;

        if (t == null) {
            return null;
        }

        /** merging order of templates is not important, both directions
         * will be examined. */
        for (int i = 0; i < merged.size(); i++) {
            Template t2 = (Template) merged.elementAt(i);
            if (t2 != null && !skiplist[i]) {
                //System.out.println("Looking at: " + t2.toString());
                for (int j = 0; j < merged.size(); j++) {
                    if (i != j && !skiplist[j]) {
                        Template t3 = (Template) merged.elementAt(j);
                        if (t3 != null &&
                            t2.getFilters().size() == t3.getFilters().size() &&
                            t2.getType().equals(t3.getType())) {
                            // We test if the two templates have the same structure
                            // We use the knowledge of the event type to optimize this procedure
                            if (t2.imperfectlyMergeableStructure(t3)) {
                                //System.out.println("with: " + t3.toString());
                                Template t4 = t2.imperfectMerge(t3);
                                merged.set(i, t4);
                                t2 = t4;

                                mergeCount++;
                                removeSet.add(t3); // remove t3

                                skiplist[j] = true;
                                cc++;
                            }
                        }
                    }
                }
            }
        }

        // remove merged templates
        if (!merged.removeAll((Collection) removeSet)) {
            Log.writeFilterService("Problem with remove collection");
        }

        // We need to check the cover
        if (merged.size() > 0 && COVERCHECK) {
            merged = MergingService.get().getNonCovered(merged);
        }

        return merged;
    }
}
