/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.filter.merge;

import java.util.*;

import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.*;
import fuegocore.notify.util.*;

/**
 * An implementation of the merging service
 */

public class MergingService implements IMergingService {
    private static MergingService ms;
    private Hashtable algorithms;

    public MergingService() {
        algorithms = new Hashtable();
        register("perfect", new PerfectMerger());
        register("imperfect", new ImperfectMerger());
        register("conjunctive", new ConjunctiveMerger());
    }

    /** Add a new merging algorithm to the framework.
     *
     * @param name Name of the algorithm
     * @param im Instance of the algorithm
     */
    public void register(String name, IMergingAlgorithm im) {
        algorithms.put(name, im);
    }

    /** Remove a merging algorithm from the framework
     *
     * @param name the name of the algorithm
     */
    public void deregister(String name) {
        algorithms.remove(name);
    }

    /**
     * Factory method to get a merging algorithm
     * @param name String
     * @return IMergingAlgorithm
     */
    public static synchronized IMergingAlgorithm get(String name) {
        if (ms == null) {
            ms = new MergingService();
        }
        if (ms.algorithms != null) {
            return (IMergingAlgorithm) ms.algorithms.get(name);
        }
        return null;
    }

    /**
     * Factory method to get a merging algorithm
     * @return IMergingService
     */
    public static IMergingService get() {
        if (ms == null) {
            ms = new MergingService();
        }
        return ms;
    }

    /** Get covered count for forward-type of behaviour
     * @param t the input vector
     * @return int the covered count
     */
    public int getCoveredCountForward(Vector t) {
        int kk = 0;
        Vector merged = new Vector();
        for (int i = 0; i < t.size(); i++) {
            Template t2 = (Template) t.elementAt(i);
            if (!covers(merged, t2)) {
                merged.add(new Template(t2.toString()));
            } else {
                kk++;
            }
        }
        return kk;
    }

    /** Get covered count for minimum cover set
     * @param t the input vector
     * @return int[] the covered count
     */
    public int[] getCoveredCount(Vector t) {
        int kk = 0;
        Hashtable types = new Hashtable();
        int ret[] = new int[2];
        long timer1 = System.currentTimeMillis();
        Vector merged = new Vector();

        for (int i = 0; i < t.size(); i++) {
            Template t2 = (Template) t.elementAt(i);

            // Optimization, if the type is not included the template cannot
            // be covered. This is
            if (types.get(t2.getType()) != null && !covers(merged, t2)) {
                // We check whether this new addition covers
                // any of the existing covering templates
                Vector merged2 = (Vector) merged.clone();

                for (int z = 0; z < merged2.size(); z++) {
                    // We use the type attribute filter for optimization
                    Template z1 = (Template) merged2.elementAt(z);
                    if (t2.getType().equals(z1.getType()) &&
                        t2.covers(z1)) {

                        // remove this template because new addition covers it

                        merged.removeElement(z1);
                    }
                }
                types.put(t2.getType(), "true");
                merged.add(new Template(t2.toString()));
            } else if (types.get(t2.getType()) == null) {
                // new type, we add it to the cover list
                types.put(t2.getType(), "true");
                merged.add(new Template(t2.toString()));
            } else {
                kk++;
            }
        }

        kk = t.size() - merged.size();
        long timer2 = System.currentTimeMillis();
        ret[0] = kk;
        ret[1] = (int) (timer2 - timer1); // time
        return ret;
    }

    /** Get non-covered vector
     * @param t the input vector
     * @return Vector the output
     */
    public Vector getNonCovered(Vector t) {
        Vector noncovered = new Vector();
        boolean[] skip = new boolean[t.size()];

        for (int i = 0; i < t.size(); i++) {
            boolean coverFlag = false;
            IFBase f0 = (IFBase) t.elementAt(i);
            for (int j = 0; j < t.size(); j++) {
                if (i == j || skip[j]) {
                    continue;
                }
                IFBase f = (IFBase) t.elementAt(j);
                if (f.covers(f0)) {
                    skip[i] = true;
                    coverFlag = true;
                    break;
                }
            }
            if (!coverFlag) {
                noncovered.add(f0);
            }
        }

        return noncovered;
    }

    /** Returns true if the set of templates v1 covers the second
     *  set of templates v2.  This is based on counting covered templates.
     *  All templates in v2 need to be covered by at least one template
     *  in v1. If all templates in v2 are covered by some template in v1
     *  the set v1 covers v2.
     * @param v1 the covering set
     * @param v2 the covered set
     * @return true if v1 covers v2
     */
    public boolean covers(Vector v1, Vector v2) {
        int cc = 0; // cover counter

        if (v1 == null || v2 == null || v1.size() < 1 || v2.size() < 1) {
            Log.writeFilterService("Either of the cover sets was null");
            Log.writeFilterService("v1 " + v1 + " v2 " + v2);
            return false;
        }

        for (int j = 0; j < v2.size(); j++) {
            Template t3 = (Template) v2.elementAt(j);
            if (t3 != null) {
                for (int i = 0; i < v1.size(); i++) {
                    Template t2 = (Template) v1.elementAt(i);
                    if (t2 != null) {
                        if (t2.covers(t3)) {
                            // t2 covers t3 so we add one to the t3 counter
                            cc++;
                            break; // end loop because covering already established
                        }
                    }
                }
            }
        }
        //System.out.println("Cover test: " + cc + " / " + v2.size());
        if (cc == v2.size()) {
            return true;
        }
        return false;
    }

    /** Returns true if the set the given template is covered by the
     *  given set of templates.
     *  This is based on counting covered templates.
     * @param v1 the covering set
     * @param v2 the covered template
     * @return true if v1 covers v2
     */
    public boolean covers(Vector v1, Template v2) {
        int cc = 0; // cover counter
        if (v1 == null || v2 == null) {
            Log.writeFilterService("Either of the cover sets was null");
            Log.writeFilterService("v1 " + v1 + " v2 " + v2);
            return false;
        }
        if (v1.size() == 0) {
            return false;
        }

        for (int i = 0; i < v1.size(); i++) {
            Template t2 = (Template) v1.elementAt(i);
            if (t2.covers(v2)) {
                Log.writeFilterService(t2.toString() + " covers " + v2.toString());
                cc++;
                return true;
            } else {
                Log.writeFilterService(t2.toString() + " does not cover " +
                                       v2.toString());
            }
        }
        return false;
    }

    /**
     * Returns true if the set the given template overlaps with the
     * given set of templates.
     * @param v1 Vector
     * @param v2 Template
     * @return boolean
     */
    public boolean overlaps(Vector v1, Template v2) {
        int cc = 0; // cover counter

        if (v1 == null || v2 == null) {
            Log.writeFilterService("Either of the overlap sets was null");
            Log.writeFilterService("v1 " + v1 + " v2 " + v2);
            return false;
        }
        if (v1.size() == 0) {
            return false;
        }

        for (int i = 0; i < v1.size(); i++) {
            Template t2 = (Template) v1.elementAt(i);
            if (t2.overlaps(v2)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Returns the set of templates that are not covered.
     * @param v1 Vector
     * @param v2 Vector
     * @return Vector
     */
    public Vector getNotCovered(Vector v1, Vector v2) {
        int cc = 0; // cover counter
        Vector covered = new Vector();

        if (v1 == null || v2 == null) {
            Log.writeFilterService("Either of the cover sets was null");
            Log.writeFilterService("v1 " + v1 + " v2 " + v2);
            return null;
        }

        for (int j = 0; j < v2.size(); j++) {
            Template t3 = (Template) v2.elementAt(j);
            if (t3 != null) {
                for (int i = 0; i < v1.size(); i++) {
                    Template t2 = (Template) v1.elementAt(i);
                    if (t2 != null) {
                        if (t2.covers(t3)) {
                            // t2 covers t3 so we add one to the t3 counter
                            covered.add(t3);
                            cc++;
                            break; // end loop because covering already established
                        }
                    }
                }
            }
        }

        if (cc == v2.size()) {
            return null;
        }

        Vector result = (Vector) v2.clone();
        result.removeAll(covered);

        return result;
    }
}
