package fuegocore.notify.core.filter.poset;

import java.util.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.util.*;

/**
 * Implements the balanced colored filter forest data structure. Uses
 * the ColorIndex helper class for implementing the color index.
 */
public class BalancedFilterForest
    extends FilterForest {

  /** Color Index structure for maintaining information about colors */
  protected ColorIndex cin = new ColorIndex();

  public BalancedFilterForest() {
    // Enable testing
    test = false;
  }

  /**
   * Returns the color index for the given filter
   * @param f IFBase
   * @return Vector
   */
  public Vector getColorIndex(IFBase f) {
    Node n = (Node) templates.get(f.getCoverId());
    if (n == null) {
      return null;
    }
    Vector cols = cin.getColorVector(n);
    return cols;
  }

  /* ********************************************************************* */

  /** Comparator for color indices */
  protected Comparator colorComparator = new Comparator() {
    public int compare(Object o1, Object o2) {
      if (o1 instanceof Node && o2 instanceof Node) {
        int a = ( (Node) o1).getCount();
        int b = ( (Node) o2).getCount();
        return b - a;
      }
      else {
        throw new ClassCastException();
      }
    }
  };

  /* ********************************************************************* */
  /**
   * Returns true if a color is not present in the subtree
   * @param n Node
   * @param tids Vector
   * @return boolean
   */
  protected boolean testColorsInSubtree(Node n, Vector tids) {
    boolean f = true;
    for (int k = 0; k < tids.size(); k++) {
      Object color = tids.elementAt(k);
      if (cin.getColorCount(n, color) > 0) {
        f = false;
        break;
      }
    }
    return f;
  }

  void eliminateC(Node x, Vector tids, Node start) throws PosetException {

    if (NO_ELIMINATE) {
      return;
    }

    if (debug) {
      System.out.println("BalancedFilterForest: eliminateC");
    }

    if (x == null) {
      return;
    }

    if (tids == null || tids.size() == 0) {
      return;
    }

    IFBase cur = (IFBase) x.getData();

    for (int k = 0; k < tids.size(); k++) {
      Object color = tids.elementAt(k);
      if (cur.getTid().contains(color)) {
        cur.getTid().remove(color);
      }
      // remove color indice in any case
      cin.delColorIndex(x, color); // update color index
    }

    // move the children under parent
    if (cur.getTid() == null ||
        (cur.getTid() != null && cur.getTid().size() == 0)) {

      if (test) {
        System.out.println("* removing node");
      }
      templates.remove(cur.getCoverId());

      cin.resetColorIndex(x); // update color index

      // Notify listeners about removal
      fireEventToListeners(DELETE_EVENT_ELIMINATE, cur);

      if (x.getParent() != null) {
        x.getParent().getChildren().remove(x);

        if (SIBLING_PURITY_ON_ELIMINATE) {
          moveChildrenDown(x, x.getParent());
        }
        else {
          // We clone the vector so flag is not required
          Vector vec = (Vector) x.getChildren().clone();

          moveSiblings(x.getChildren(), x.getParent().getChildren(), null,
                       x.getParent());
          for (int i = 0; i < vec.size(); i++) {
            Node n = (Node) vec.elementAt(i);
            // color index optimization: we skip subtrees that do not have the color
            if (testColorsInSubtree(n, tids)) {
              continue;
            }
            eliminateC(n, (Vector) tids.clone(), start);
//          eliminateC(n, tids,start);
          }
        }
      }
      else {
        System.out.println("eliminate: parent was null");
        throw new PosetException();
      }

      x.getChildren().clear();
    }
    else {
      // We clone the vector so flag is not required
      tids.addAll(cur.getTid()); // add remaining colors
      Vector vec = (Vector) x.getChildren().clone();
      for (int i = 0; i < vec.size(); i++) {
        Node n = (Node) vec.elementAt(i);
        // color index optimization: we skip subtrees that do not have the color
        if (testColorsInSubtree(n, tids)) {
          continue;
        }
        eliminateC(n, (Vector) tids.clone(), start);
      }
    }

    return;
  }

  /**
   * Eliminate a vector of colors
   * @param x Node
   * @param v Vector
   * @throws PosetException
   */
  public void eliminateColor(Node x, Vector v) throws PosetException {
    if (NO_ELIMINATE) {
      return;
    }

    Vector c = (Vector) x.getChildren().clone();

    if (v == null) {
      v = new Vector();
    }

    // ensure that the colors of the start level are included
    for (int p = 0; p < x.getData().getTid().size(); p++) {
      Object col = x.getData().getTid().elementAt(p);
      if (!v.contains(col)) {
        v.add(col);
      }
    }

    for (int i = 0; i < c.size(); i++) {
      Node y = (Node) c.elementAt(i);
      // color index optimization: we skip subtrees that do not have the color
      if (testColorsInSubtree(y, v)) {
        continue;
      }
      eliminateC(y, v, x);
    }
  }

  void eliminateC(Node x, Object color, Node start) throws PosetException {
    if (debug) {
      System.out.println("BalancedFilterForest: eliminateC single");
    }

    if (x == null || color == null) {
      return;
    }
    IFBase cur = (IFBase) x.getData();

    if (test) {
      System.out.println("Eliminate-color: " + color + "Looking at colors:" +
                         cur.getTid());
    }

    // remove color from index in any case
    cin.delColorIndex(x, color); // update color index

    if (cur.getTid().contains(color)) {
      if (debug || test) {
        System.out.println("Eliminate_color: color found");
      }

      cur.getTid().remove(color);

      // Notify listeners about removal
      fireEventToListeners(DELETE_EVENT_ELIMINATE, cur);

      // move the children under parent
      if (cur.getTid().size() == 0) {
        if (debug || test) {
          System.out.println("Eliminate_color: Removing node and terminating: " +
                             cur.toString() + " color: " + color);
        }

        templates.remove(cur.getCoverId());
        cin.resetColorIndex(x); // update color index

        if (x.getParent() != null) {
          x.getParent().getChildren().remove(x);

          if (SIBLING_PURITY_ON_ELIMINATE) {
            moveChildrenDown(x, x.getParent());
          }
          else {
            // We clone the vector so flag is not required
            Vector vec = (Vector) x.getChildren().clone();

            moveSiblings(x.getChildren(), x.getParent().getChildren(), null,
                         x.getParent());

            // We have to ensure that there are no color indices for the
            // deleted color. Tids are safe by definition.
            for (int i = 0; i < vec.size(); i++) {
              Node n = (Node) vec.elementAt(i);
              // color index optimization: we skip subtrees that do not have the color
              if (cin.getColorCount(n, color) == 0) {
                if (debug || test) {
                  System.out.println(
                      "Skipping update in eliminate after delete");
                }
                continue;
              }
              eliminateC(n, color, start);
            }

          }

        }
        else {
          System.out.println("eliminate: parent was null");
          throw new PosetException();
        }
        x.getChildren().clear();
      }
    }
    else {
      if (debug || test) {
        System.out.println("Eliminate_color: iterating tree");
        // We clone the vector so flag is not required
      }
      Vector vec = (Vector) x.getChildren().clone();
      for (int i = 0; i < vec.size(); i++) {
        Node n = (Node) vec.elementAt(i);
        // color index optimization: we skip subtrees that do not have the color
        if (cin.getColorCount(n, color) == 0) {
          if (debug || test) {
            System.out.println("Skipping update in eliminate");
          }
          continue;
        }
        eliminateC(n, color, start);
      }
    }
    return;
  }

  public boolean testEliminateAll(IFBase xc, Object col) throws PosetException {
    if (NO_ELIMINATE) {
      return true;
    }
    boolean disable = DISABLE_LISTENERS;
    Enumeration el = templates.elements();
    while (el.hasMoreElements()) {
      Node x = (Node) el.nextElement();
      if (x.getData().getTid().contains(col)) {
        if (xc.covers(x.getData())) {
          DISABLE_LISTENERS = true; // prevent notification
          this.remove(x, col);
          DISABLE_LISTENERS = disable;
        }
        else if (x.getData().covers(xc)) {
          return false;
        }
      }
    }
    return true;
  }

  /**
   * Eliminate all nodes that are covered by x and have the same color.
   * Note: add must be called after this for the node.
   * @param xc Node
   * @param col Object
   * @return boolean false if x is covered by a filter with the same color, otherwise true
   * @throws PosetException
   */
  public boolean eliminateAll(IFBase xc, Object col) throws PosetException {
    Vector stack = new Vector();
    // Create an imaginary node
    Node imaginary = new Node();
    stack.add(imaginary);

    Vector all = getLevelOnlyColored(null, col);
    imaginary.children.addAll(all);

    // Traverse the tree
    while (stack.size() > 0) {
      Node n = (Node) stack.remove(0);
      if (n != null) {
        Vector ci = getLevelOnlyColored(n, col);
        if (ci != null) {
          Iterator k = ci.iterator();
          while (k.hasNext()) {
            Node n2 = (Node) k.next();

            IFBase t = (IFBase) n2.getData();

            if (!t.getTid().contains(col)) {
// --
              /*
                             comparisons++;
                            if (xc.covers(t)) {
                eliminateColor(n2, col);
                            }
                            else
               */
              //--
              if (n2.getChildren().size() > 0) {
                stack.add(n2);
              }
            }
            else {
              if (NO_COVERED_UPDATES && n2.getParent() == null) {
                continue;
              }

              boolean doCoverCheck = true;

              if (NO_COVERED_UPDATES) {
                doCoverCheck = false;
              }

              if (doCoverCheck) {
                comparisons++;

                if (t.covers(xc)) { // if the node is covered, then we can stop
                  // this is needed to lose redundancy
                  return false;
                }
              }

              comparisons++;

              if (xc.covers(t)) {
                // we eliminate only on non-roots, add must be called after
                // add will move covered root nodes and eliminate color
                if (n2.getParent() != null) {
                  // We simply remove the node
                  // No reason to check the descendants
                  eliminateC(n2, col, n); // only local effect
                  removeNodeEvent(n2, col);

                }
              }
            }
          }
        }
      }
    }
    /*
         // REMOVEME
         if (!testEliminateAll(xc, col)) {
      System.out.println("EliminateAll failure");
      System.exit(0);
         }
         // REMOVEME
     */
    return true;
  }

  protected boolean testColorsToRoot(Node n, Vector cols) {
    //System.out.println("Level: " + n.hashCode());
    for (int i = 0; i < cols.size(); i++) {
      Object c = (Object) cols.elementAt(i);
      if (cin.getColorCount(n, c) == 0) {
        System.out.println("**************************************");
        System.out.println("Color: " + c + " missing");
        System.out.println("Active cols: " + (Hashtable) cin.cIndex.get(n));
        System.out.println("Parent: " + n.getParent());
        System.out.println("Cols: " + cols);
        System.out.println("Tids: " + n.getData().getTid());
        return false;
      }
    }

    //System.out.println("Cols: " + n.getData().getTid());

    cols.addAll(n.getData().getTid());
    if (n.parent != null) {
      return testColorsToRoot(n.parent, cols);
    }
    else {
      //System.out.println("Root");
      return true;
    }
  }

  /**
   * Tests that there are no duplicate color indices. Ensures
   * that no node below the given node has the color in the index.
   * @param n Node
   * @param v Vector The colors advertised by the node (tid)
   * @param skip boolean
   * @return boolean false if failure
   */
  protected boolean testColorIndexDuplicates(Node n, Vector v, boolean skip) {
    if (!skip) { // skip the first node because it owns v
      Vector cols = cin.getColorVector(n);
      if (cols == null) {
        System.out.println("Color vector was null!");
        System.out.println("Input vector: " + v);
        System.out.println("Cols: " + cols);
        System.out.println("Parent: " + n.parent);
        System.out.println("Node:" + n);
        System.out.println("Tid: " + n.getData().getTid());
        System.out.println("Data: " + n.getData());

        return true;
      }

      for (int i = 0; i < v.size(); i++) {
        Object col = v.elementAt(i);
        if (cols.contains(col)) {
          System.out.println("A duplicate color was found: " + col);
          System.out.println("Input vector: " + v);
          System.out.println("Cols: " + cols);
          System.out.println("Parent: " + n.parent);
          System.out.println("Node:" + n);
          System.out.println("Tid: " + n.getData().getTid());
          System.out.println("Data: " + n.getData());
          System.out.println("Explanation: a node below the input node " +
                             " advertises the color in the color index");

          printColorsToRoot(n);
          System.out.println("Color in tid away from the root");
          printColors(n, col);
          return true;
        }
      }
    }
    if (n != null && n.getChildren() != null) {
      for (int i = 0; i < n.getChildren().size(); i++) {
        if (testColorIndexDuplicates( (Node) n.getChildren().elementAt(i), v, false)) {
          return true;
        }
      }
    }
    return false;
  }

  /**
   * Tests that all color indices are being used in lower levels
   * @param n Node
   * @param v Vector
   */
  private void testColorUsage(Node n, Vector v) {
    Vector cols = n.getData().getTid();

    v.removeAll(cols);
    if (v.size() == 0) {
      return;
    }

    if (n.getChildren() != null) {
      for (int i = 0; i < n.getChildren().size(); i++) {
        testColorUsage( (Node) n.getChildren().elementAt(i), v);
      }
    }
  }

  public void printColors(Node n) {
    Vector cols = n.getData().getTid();
    System.out.println("Node: " + n.hashCode() + ":" + cols);
    if (n.getChildren() != null) {
      for (int i = 0; i < n.getChildren().size(); i++) {
        printColors( (Node) n.getChildren().elementAt(i));
      }
    }
  }

  public void printColors(Node n, Object col) {
    Vector cols = n.getData().getTid();
    if (cols.contains(col)) {
      System.out.println("Node: " + n + " cols:" + cols);
    }
    if (n.getChildren() != null) {
      for (int i = 0; i < n.getChildren().size(); i++) {
        printColors( (Node) n.getChildren().elementAt(i), col);
      }
    }
  }

  public void printColorsToRoot(Node n) {
    Node p = n;
    do {
      System.out.println("Node: " + p + " colors: " + p.getData().getTid());
      System.out.println("Color index: " + cin.getColorVector(p));
      p = p.parent;
    }
    while (p != null);
  }

  public String runTests() {
    String str = super.runTests();
    boolean ret = true;
    String myName = this.getClass().getName();

    str += "TEST: Color index colors to root, color duplicates, node cover\n";
    Vector v = getAll();
    for (int i = 0; i < v.size(); i++) {
      Node n = (Node) v.elementAt(i);
      if (n.parent != null && !testColorsToRoot(n, new Vector())) {
        str += "Fatal:  colors not active to root\n";
        printColorsToRoot(n);
        ret = false;
        break;
      }
      if (testColorIndexDuplicates(n, n.getData().getTid(), true)) {
        ret = false;
      }
      if (!testNodeCover(n)) {
        ret = false;
      }

    }

    if (ret) {
      str += "RESULT: Color index ok\n";
    }
    else {
      str += "RESULT: Color index failed\n";
      ret = true;
    }

    str += "TEST: all nodes, minimum color, color usage, further color-index tests, redundancy if applicable\n";

    // testing minimum color property
    for (int i = 0; i < nodes.size(); i++) {
      Node k = (Node) nodes.elementAt(i);
      if (k == null) {
        str += "testColorIndex: node was null\n";
      }
      if (k.getData() == null) {
        str += "testColorIndex: node.getData() was null\n";
      }
      if (k.getData().getTid() == null) {
        str += "testColorIndex: node.getData().getTid() was null\n";
      }

      if (k.getData().getTid().size() == 0) {
        str +=
            "" + myName + ": A root node was found with no color!\n";
        System.out.println(k.getData());
        ret = false;
      }

      if (!testMinimumColor(k, new Vector())) {
        str += "Fatal: " + myName + ": minimum color property " +
            " violation\n";
        ret = false;
        break;
      }
      Vector test = new Vector();
      Hashtable h = (Hashtable) cin.cIndex.get(k);
      if (h == null) {
        str += "Fatal: " + myName +
            ": color index missing for a node: " +
            k.hashCode() + "\n";
        ret = false;
        break;
      }

      test.addAll(h.keySet());
      testColorUsage(k, test);
      if (test.size() != 0) {
        str += "Fatal: " + myName + ": color usage error. Size: " +
            test.size() + "\n";
        str += "Missing color: " + test + "\n";
        ret = false;
        break;
      }

      if (REMOVE_ALL_REDUNDANCY) {
        try {
          // Test if a covering color was found
          for (int p = 0; p < k.getData().getTid().size(); p++) {
            Object col = k.getData().getTid().elementAt(p);
            if (findCoveringTest(k.getData(), col)) {
              str += "Found a covering color: " + col + "\n";
              str += "Node: " + k.getData() + "\n";
              ret = false;
              break;
            }
          }
        }
        catch (Exception e2) {
          e2.printStackTrace();
          ret = false;
        }
      }
    }

    if (ret) {
      str += "RESULT: tests ok\n";
    }
    else {
      str += "RESULT: tests failed\n";
    }

    return str;
  }

  /** Tests the color index
   *  @return boolean false if tests fail
   */
  public boolean testColorIndex() {
    boolean ret = true;

    String myName = this.getClass().getName();

    System.out.println("[" + myName + "] testColorIndex. Root size: " +
                       nodes.size());

    Vector v = getAll();
    for (int i = 0; i < v.size(); i++) {
      Node n = (Node) v.elementAt(i);
      if (n.parent != null && !testColorsToRoot(n, new Vector())) {
        System.out.println("Fatal: " + myName + ": colors not active to root");
        printColorsToRoot(n);
        ret = false;
        break;
      }
      if (testColorIndexDuplicates(n, n.getData().getTid(), true)) {
        ret = false;
      }
      if (!testNodeCover(n)) {
        ret = false;
      }

    }

    // testing minimum color property
    for (int i = 0; i < nodes.size(); i++) {
      Node k = (Node) nodes.elementAt(i);
      if (k == null) {
        System.out.println("testColorIndex: node was null");
      }
      if (k.getData() == null) {
        System.out.println("testColorIndex: node.getData() was null");
      }
      if (k.getData().getTid() == null) {
        System.out.println("testColorIndex: node.getData().getTid() was null");
      }

      if (k.getData().getTid().size() == 0) {
        System.out.println(
            "" + myName + ": A root node was found with no color!");
        System.out.println(k.getData());
        ret = false;
      }

      if (!testMinimumColor(k, new Vector())) {
        System.out.println("Fatal: " + myName + ": minimum color property " +
                           " violation");
        ret = false;
        break;
      }
      Vector test = new Vector();
      Hashtable h = (Hashtable) cin.cIndex.get(k);
      if (h == null) {
        System.out.println("Fatal: " + myName +
                           ": color index missing for a node: " +
                           k.hashCode());
        ret = false;
        break;
      }

      test.addAll(h.keySet());
      //System.out.println("Input set: " + test);
      //System.out.println("Cur cols: " +k.getData().getTid());
      testColorUsage(k, test);

      if (test.size() != 0) {
        System.out.println("Fatal: " + myName + ": color usage error. Size: " +
                           test.size());
        System.out.println("Missing color: " + test);
        ret = false;
        printColors(k);
        break;
      }

      if (REMOVE_ALL_REDUNDANCY) {
        try {
          // Test if a covering color was found
          for (int p = 0; p < k.getData().getTid().size(); p++) {
            Object col = k.getData().getTid().elementAt(p);
            if (findCoveringTest(k.getData(), col)) {
              System.out.println("Found a covering color: " + col);
              System.out.println("Node: " + k.getData());
              ret = false;
              break;
            }
          }
        }
        catch (Exception e2) {
          e2.printStackTrace();
          ret = false;
        }
      }
    }
    if (!ret) {
      System.out.println("[" + myName +
                         "] testColorIndex. There were errors.");
      System.exit(0);
    }
    return ret;
  }

  public void reset() {
    super.reset();
    cin.reset();
  }

  /**
   * Get the children sorted based on the color property.
   * Used for balancing. Clones the vector.
   * @param x Node
   * @param color Object
   * @return Vector
   */
  synchronized Vector getLevel(Node x, Object color) {
    Vector v = null;

    if (color == null) {
      return super.getLevel(x, color);
    }

    if (x == null) {
      v = nodes;
    }
    else {
      v = x.getChildren();
    }
    if (v == null) {
      return null;
    }

    Vector ret = new Vector();
    synchronized (v) {
      for (int i = 0; i < v.size(); i++) {
        Node k = (Node) v.elementAt(i);
        // set count for sorting
        k.setCount(cin.getColorCount(k, color));
        // Simple way to sort the binary values
        if (k.getCount() > 0) {
          ret.add(0, k);
        }
        else {
          ret.add(ret.size(), k);
        }
      }
    }
    return ret;
  }

  /**
   * Returns only nodes with the given color. Clones the vector.
   * @param x Node
   * @param color Object
   * @return Vector
   */
  synchronized Vector getLevelOnlyColored(Node x, Object color) {
    Vector v = null;

    if (x == null) {
      v = nodes;
    }
    else {
      v = x.getChildren();
    }
    if (v == null) {
      return null;
    }

    Vector ret = new Vector();
    synchronized (v) {
      for (int i = 0; i < v.size(); i++) {
        Node k = (Node) v.elementAt(i);
        // set count for sorting
        k.setCount(cin.getColorCount(k, color));
        if (k.getCount() > 0) {
          ret.add(k);
        }
      }
    }
//    System.out.println("Color reduced: " + ret.size() + " / " + v.size());
    return ret;
  }

  /**
   * Get the children sorted based on the color property.
   * Used for balancing. Clones the vector.
   * @param x Node
   * @param color Object
   * @return Vector
   */
  synchronized Vector getLevelSortedWithComparator(Node x, Object color) {
    Vector v = null;

    if (color == null) {
      return super.getLevel(x, color);
    }

    if (x == null) {
      v = (Vector) nodes;
    }
    else {
      v = (Vector) x.getChildren().clone();
    }
    if (v == null) {
      return null;
    }

    for (int i = 0; i < v.size(); i++) {
      Node k = (Node) v.elementAt(i);
      // set count for sorting
      k.setCount(cin.getColorCount(k, color));
    }
    // we use an inner class for sorting based on the color
    Collections.sort(v, colorComparator);
    return v;
  }

  /*** EXTENSION MECHANISMS **********************************************/

  /**
   * Updates the color index
   * @param n Node
   * @param color Object
   */
  public void addNodeEvent(Node n, Object color) {
    if (test) {
      System.out.println("AddEvent color: " + color);

    }
    if (n == null) {
      System.out.println("BalancedFilterForest: fatal input node was null");
    }
    else {
      cin.updateColorBasedOnChildren(n, color);
    }
  }

  /**
   * Updates the color index
   * @param n Node
   * @param colors Vector
   * @throws PosetException
   */
  public void addNodeEvent(Node n, Vector colors) throws PosetException {

    if (test) {
      System.out.println("AddEvent colors: " + n.getData().getTid());
    }
    if (n == null) {
      System.out.println("BalancedFilterForest: fatal input node was null");
    }
    else {
      cin.updateColorBasedOnChildren(n);
    }
  }

  /**
   * Updates the color index
   * @param n Node
   * @param colors Vector
   */
  public void removeNodeEvent(Node n, Vector colors) {
    if (n == null) {
      System.out.println("BalancedFilterForest: fatal input node was null");
      return;
    }

    if (test) {
      System.out.println("RemoveEvent colors: " + n.getData().getTid());
      cin.testColorStatus(n);
      System.out.println("Number of children: " + n.getChildren().size());
      cin.printAllCols(n);
    }
    Vector cols = cin.getColorVector(n);
    if (test) {
      System.out.println("Removing colors for: " + cols);
    }

    if (cols == null) {
      return;
    }
    for (int i = 0; i < cols.size(); i++) {
      Object col = cols.elementAt(i);
      cin.removeColorIndex(n.getParent(), n, col);
    }
    cols.clear();
  }

  /**
   * Updates the color index
   * @param n Node
   * @param color Object
   */
  public void removeNodeEvent(Node n, Object color) {
    if (n == null) {
      System.out.println("BalancedFilterForest: fatal input node was null");
      return;
    }

    if (test) {
      System.out.println("RemoveEvent color: " + color);

    }
    cin.delColorIndex(n, color);
    cin.removeColorIndex(n.getParent(), n, color);
  }

  /*** EXTENSION MECHANISMS **********************************************/


  /**
   * Match a notification using the poset or forest. Assumes that
   * the objects are ITemplates and groups them by their interface name
   *  (getInterface).
   *  @param e the notification to match
   *  @return TripletTable contains the matched objects
   */
  public TripletTable matchT(Object e) {
    if (! (e instanceof Notification)) {
      return null;
    }
    comparisons = 0;
    TripletTable ret = new TripletTable();
    Vector stack = new Vector();

    // Create an imaginary node
    Node imaginary = new Node();
    imaginary.children.addAll(nodes);
    stack.add(imaginary);
    // Insert type-specific optimizations here
    // Insert a pre-selection process here

    boolean quit = false;
    HashSet result = new HashSet();

    // Traverse the tree
    while (stack.size() > 0 && !quit) {
      Node n = (Node) stack.remove(0);
      if (n != null) {
        Vector ci = new Vector();

        synchronized (n.getChildren()) {
          // Color-based optimization
          for (int i = 0; i < n.getChildren().size(); i++) {
            Node c = (Node) n.getChildren().elementAt(i);
            if (!cin.colorVectorContained(c, result)) {
              ci.add(c);
            }
          }
        }

        if (ci != null) {
          Iterator k = ci.iterator();
          while (k.hasNext()) {
            Node n2 = (Node) k.next();
            IFBase t = (IFBase) n2.getData();

            // if the current node has been matched already
            // we may skip the matching
            if (result.containsAll(t.getTid())) {
              if (n2.getChildren().size() > 0) {
                stack.add(n2);
              }
            }
            else {
              comparisons++;
              if (t.match(null, e)) {
                addToResults(result, t);
                if (n2.getChildren().size() > 0) {
                  stack.add(n2);
                }
                ITemplate _t = (ITemplate) t;

                ret.insert(_t.getInterface(), _t);
              }
            }
          }
        }
      }
    }
    return ret;
  }

  /**
   * Match an object against the data structure with color-specific
   * optimizations.
   * @param e Object
   * @return Vector
   */
  public Vector match(Object e) {
    if (! (e instanceof Notification)) {
      return null;
    }

    comparisons = 0;

    Vector stack = new Vector();

    // Create an imaginary node
    Node imaginary = new Node();
    imaginary.children.addAll(nodes);
    stack.add(imaginary);
    // Insert type-specific optimizations here
    // Insert a pre-selection process here

    boolean quit = false;
    HashSet result = new HashSet();

    // Traverse the tree
    while (stack.size() > 0 && !quit) {
      Node n = (Node) stack.remove(0);

      if (n != null) {
        Vector ci = new Vector();

        synchronized (n.getChildren()) {
          // Color-based optimization
          for (int i = 0; i < n.getChildren().size(); i++) {
            Node c = (Node) n.getChildren().elementAt(i);
            if (!cin.colorVectorContained(c, result)) {
              ci.add(c);
            }
          }
        }

        if (ci != null && ci.size() > 0) {
          Iterator k = ci.iterator();
          while (k.hasNext()) {
            Node n2 = (Node) k.next();
            IFBase t = (IFBase) n2.getData();

            // if the current node has been matched already
            // we may skip the matching
            if (result.containsAll(t.getTid())) {
              if (n2.getChildren().size() > 0) {
                stack.add(n2);
              }
            }
            else {
              comparisons++;
              if (t.match(null, e)) {
                addToResults(result, t);
                if (n2.getChildren().size() > 0) {
                  stack.add(n2);
                }
              }
            }
          }
        }
      }
    }
    Vector res = new Vector(result);
    return res;
  }
}
