package fuegocore.notify.core.filter.poset;

import java.util.*;

/** This is a helper class for the balanced filter forest and
 * maintains a color index for the nodes. This version supports only
 * true/false value for colors, which are maintained for each color
 * present under a node.
 *
 */
public class ColorIndex {
  Hashtable cIndex = new Hashtable(); // color index for nodes
  boolean debug = false;

  /** Constructor
   */
  public ColorIndex() {
  }

  /**
   * Update the color index locally
   * @param n Node
   */
  public void updateColorBasedOnChildrenLocally(Node n) {
    resetColorIndex(n);
    // the update the color status for this node
    for (int i = 0; i < n.getChildren().size(); i++) {
      Node k = (Node) n.getChildren().elementAt(i);
      combineColorIndices(k, n);
    }
  }

  /**
   * Update the color index globally
   * @param n Node
   * @throws PosetException
   */
  public void updateColorBasedOnChildren(Node n) throws PosetException {
    updateColorIndex(n);

    // the update the color status for this node
    // these are already counted so they are not taken into account
    if (debug) {
      System.out.println("Multiple color indice update: Node: " + n.hashCode());
      System.out.println("Combining color indices: " + n.getChildren().size());
      System.out.println("Cur cols: " + this.getColorVector(n));
    }

    for (int i = 0; i < n.getChildren().size(); i++) {
      Node k = (Node) n.getChildren().elementAt(i);
      combineColorIndices(k, n);
    }

    // update color count to root
    if (n.getParent() != null) {
      if (debug) {
        Hashtable b = new Hashtable();
        for (int i = 0; i < n.getChildren().size(); i++) {
          Node k = (Node) n.getChildren().elementAt(i);
          b.putAll( (Hashtable) cIndex.get(k));
        }

        System.out.println("Cur cols: " + this.getColorVector(n));
        System.out.println("Recons cols: " + b);
        System.out.println("Own cols: " + n.getData().getTid());
        printAllCols(n);
        System.out.println("--");
      }
      addColorIndex(n.getParent(), (Hashtable) cIndex.get(n));
    }
  }

  /**
   * Prints all colors of the input node
   * @param n Node
   */
  public void printAllCols(Node n) {
    for (int i = 0; i < n.children.size(); i++) {
      Node k = (Node) n.children.elementAt(i);
      System.out.println("CCOL: " + k.getData().getTid());
      printAllCols(k);
    }
  }

  /**
   * Debug colors status
   * @param n Node
   */
  public void testColorStatus(Node n) {
    Hashtable b = new Hashtable();
    for (int i = 0; i < n.getChildren().size(); i++) {
      Node k = (Node) n.getChildren().elementAt(i);
      b.putAll( (Hashtable) cIndex.get(k));
    }
    System.out.println("Reconstructed colors of children: " + b);
    System.out.println("Own cols: " + n.getData().getTid());
  }

  /**
   * Update the color index globally for the given color
   * @param n Node
   * @param color Object
   */
  public void updateColorBasedOnChildren(Node n, Object color) {
    updateColorIndex(n);

    if (debug) {
      System.out.println("Sigle color index update: Node: " + n.hashCode() +
                         " " + n.getData().getTid());
      System.out.println("Combining color indices: " + n.getChildren().size());
    }

    // the update the color status for this node
    for (int i = 0; i < n.getChildren().size(); i++) {
      Node k = (Node) n.getChildren().elementAt(i);
      combineColorIndices(k, n);
    }

    // update color from this object to the root
    if (n.getParent() != null) {
      addColorIndex(n.getParent(), color);
    }
  }

  /** Resets color index for given node
   * @param x the node
   */
  public void removeColorIndex(Node x) {
    Hashtable h = (Hashtable) cIndex.remove(x);
    if (h == null) {
      return;
    }
    h.clear();
    h = null;
  }

  /**
   * Return the color vector for the given node
   * @param n Node
   * @return Vector
   */
  public Vector getColorVector(Node n) {
    Hashtable h = (Hashtable) cIndex.get(n);
    if (h == null) {
      return null;
    }
    Set s = h.keySet();
    Vector r = new Vector();
    if (s != null) {
      r.addAll(s);
    }
    return r;
  }

  /**
   * Return the color vector for the given node
   * @param n Node
   * @return Vector
   */
  public HashSet getColorVectorHash(Node n) {
    Hashtable h = (Hashtable) cIndex.get(n);
    if (h == null) {
      return null;
    }
    Set s = h.keySet();
    HashSet r = new HashSet();
    if (s != null) {
      r.addAll(s);
    }
    return r;
  }

  /**
   * Tests whether the given node's color vector is contained with the
   * given color vector.
   * @param n Node
   * @param comp HashSet
   * @return boolean
   */
  public boolean colorVectorContained(Node n, HashSet comp) {
    Hashtable h = (Hashtable) cIndex.get(n);
    if (h == null) {
      return false;
    }
    Set s = h.keySet();
    Iterator i = s.iterator();
    while (i.hasNext()) {
      Object col = i.next();
      if (!comp.contains(col)) {
        return false;
      }
    }
    return true;
  }

  /**
   * Returns true if the node has colors.
   * @param x Node
   * @return boolean true if the node has colors
   */
  public boolean hasColors(Node x) {
    Hashtable h = (Hashtable) cIndex.get(x);
    if (h == null) {
      return false;
    }
    if (h.size() > 0) {
      return true;
    }
    return false;
  }

  /** Resets color index for given node
   * @param x the node
   */
  public void resetColorIndex(Node x) {
    Hashtable h = (Hashtable) cIndex.get(x);
    if (h == null) {
      h = new Hashtable();
      cIndex.put(x, h);
    }
    h.clear();
    Vector v = (Vector) x.getData().getTid();
    for (int i = 0; i < v.size(); i++) {
      h.put(v.elementAt(i), new Integer(1));
    }
  }

  /** Updates color index for given node
   * @param x the node
   */
  public void updateColorIndex(Node x) {
    Hashtable h = (Hashtable) cIndex.get(x);
    if (h == null) {
      h = new Hashtable();
      cIndex.put(x, h);
    }
    else {
      h.clear();
    }

    Vector v = (Vector) x.getData().getTid();
    if (v == null) {
      System.out.println("updateColorIndex: Found an empty tid");
      return;
    }
    for (int i = 0; i < v.size(); i++) {
      if (h.get(v.elementAt(i)) == null) {
        h.put(v.elementAt(i), new Integer(1));
      }
    }
  }

  /**
   * Adds a new color to the forest starting from this node and
   * proceeding towards the root. Descandants of this node do not
   * need to be updated.
   * @param x Node
   * @param color Object
   */
  public void addColorIndex(Node x, Object color) {
    if (getColorCount(x, color) > 0) {
      if (debug) {
        System.out.println("Color " + color + " already in " +
                           x.getData().getTid());
      }
      return;
    }
    insColorIndex(x, color);
    if (x.parent != null) {
      addColorIndex(x.parent, color);
    }
  }

  /**
   * Adds a new color to the forest starting from this node and
   * proceeding towards the root. Descandants of this node do not
   * need to be updated.
   * @param x Node
   * @param colors Hashtable of colors
   * @throws PosetException
   */
  public void addColorIndex(Node x, Hashtable colors) throws PosetException {
    Enumeration k = colors.keys();
    int c = 0;
    for (Enumeration e = colors.elements(); e.hasMoreElements(); ) {
      Object color = k.nextElement();
      Integer count = (Integer) e.nextElement();
      if (getColorCount(x, color) > 0) {
        c++;
      }
      else {
        insColorIndex(x, color);
      }
    }
    if (c == colors.size()) {
      if (debug) {
        System.out.println("All colors covered:" + colors.size());
      }
      return; // no need to continue
    }
    else if (c > colors.size()) {
      System.out.println("Failure: size assert in addColorIndex");
      throw new PosetException();
    }

    if (x.parent != null) {
      addColorIndex(x.parent, colors);
    }
  }

  /**
   * Removes a new color to the forest starting from this node and
   * proceeding towards the root.Descandants of this node do not
   * need to be updated since they cannot have the color.
   * @param x Node
   * @param source the source Node
   * @param color Object
   */
  public void removeColorIndex(Node x, Node source, Object color) {
    boolean flag = false;

    if (x == null) {
      return;
    }

    // we remove the color and continue only if the current node does
    // not have other children with the same color
    Vector vec = (Vector) x.getChildren().clone();
    for (int i = 0; i < vec.size(); i++) {
      Node k = (Node) vec.elementAt(i);
      if (k != source && getColorCount(k, color) > 0) {
        flag = true;
        break;
      }
    }

    if (flag) {
      if (debug) {
        System.out.println("Color " + color +
                           " was covered by children in removeColor");
      }
      return;
    }

    delColorIndex(x, color);
    if (x.parent != null) {
      removeColorIndex(x.parent, x, color);
    }
    if (!hasColors(x)) {
      removeColorIndex(x);
    }
  }

  /**
   * Returns true if the given node has the given color
   * @param x Node the node
   * @param color Object the color
   * @return boolean true if node has the given color
   */
  public boolean hasColorIndex(Node x, Object color) {
    Hashtable h = (Hashtable) cIndex.get(x);
    if (h == null) {
      return false;
    }
    Integer obj = (Integer) h.get(color);
    if (obj == null) {
      return false;
    }
    return true;
  }

  /**
   * Get the total number of different colors
   * @param x Node the node
   * @return int the number of colors
   */
  public int getTotalColorCount(Node x) {
    Hashtable h = (Hashtable) cIndex.get(x);
    if (h == null) {
      return -1;
    }
    int k = 0;
    for (Enumeration i = h.elements(); i.hasMoreElements(); ) {
      Integer it = (Integer) i.nextElement();
      if (it.intValue() > 0) {
        k++;
      }
    }
    return k;
  }

  /** Get color count
   * @param x the Node
   * @param color the color
   * @return int the color count
   */
  public int getColorCount(Node x, Object color) {
    if (x == null) {
      return -1;
    }

    Hashtable h = (Hashtable) cIndex.get(x);
    if (h == null) {
      return -1;
    }
    Integer obj = (Integer) h.get(color);
    if (obj == null) {
      return 0;
    }
    return 1;
  }

  /**
   * Combines the color indices of the two input nodes by adding the colors of x
   * to y. This is called when
   * nodes are moved under a new node during add
   * @param x Node
   * @param y Node
   */
  void combineColorIndices(Node x, Node y) {
    Hashtable h1 = (Hashtable) cIndex.get(x);
    Hashtable h2 = (Hashtable) cIndex.get(y);

    if (h1 == null) {
      System.out.println("combineColorIndices h1 was null");
      return;
    }

    if (h2 == null) {
      h2 = new Hashtable();
      cIndex.put(y, h2);
    }

    if (debug) {
      System.out.println("Combining colors for: " + h1.size());
    }
    Enumeration j = h1.keys();
    for (Enumeration i = h1.elements(); i.hasMoreElements(); ) {
      Integer it = (Integer) i.nextElement();
      Object col = j.nextElement();
      Integer it2 = (Integer) h2.get(col);
      if (it2 == null) {
        h2.put(col, new Integer(1));
      }
      // otherwise the object exists already
    }
  }

  /**
   * Insert color index
   * @param x Node
   * @param color Object
   * @return int
   */
  public int insColorIndex(Node x, Object color) {
    // note that this double hashtable implementation takes a lot of memory
    // Java is not very good for bit operations
    Hashtable h = (Hashtable) cIndex.get(x);
    if (h == null) {
      h = new Hashtable();
      cIndex.put(x, h);
    }
    Integer obj = (Integer) h.get(color);
    if (obj == null) {
      obj = new Integer(1);
      h.put(color, obj);
    }
    return obj.intValue();
  }

  /**
   * Delete color index
   * @param x Node
   * @param color Object
   * @return int
   */
  public int delColorIndex(Node x, Object color) {
    int ret = -1;

    // note that this double hashtable implementation takes a lot of memory
    // Java is not very good for bit operations
    Hashtable h = (Hashtable) cIndex.get(x);
    if (h == null) {
      return ret;
    }
    Integer obj = (Integer) h.get(color);
    if (obj != null) {
      if (h.remove(color) != null) {
        ret = 1;
      }
    }
    return ret;
  }

  /** Reset the data structure */
  public void reset() {
    cIndex.clear();
  }

  /** Destroy the data structure */
  public void destroy() {
    cIndex = null;
  }
}
