package fuegocore.notify.core.filter.poset;

import java.io.*;
import java.util.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.filter.poset.forwards.*;
import fuegocore.notify.core.server.ui.*;
import fuegocore.notify.util.*;

/**
 * The poset derived forest data structure
 * Sasu Tarkoma, 2004-2005
 */
public class FilterForest
    implements IFilterPoset {
  int counter = 0;
  int comparisons = 0; // for benchmarking
  Hashtable templates = new Hashtable();
  Vector nodes = new Vector();
  public boolean debug = false; // for debug output
  public boolean test = false;

  /** Used if all excess filters are removed (eliminateAll during add) */
  public boolean REMOVE_ALL_REDUNDANCY = false;

  public boolean SKIP_REDUNDANCY_CHECK = false;

  /** Do not perform eliminate color */
  public boolean NO_ELIMINATE = false;

  /** Eliminate only with del */
  public boolean ELIMINATE_ALL_DEL = false;

  /** Prevent all changes to interfaces (tids) */
  public boolean NO_INTERFACE_CHANGES = false;

  /** If covered filters are allowed from the same iface */
  public boolean NO_COVERED_UPDATES = false;

  public boolean DISABLE_LISTENERS = false;

  /** Enforces sibling purity */
  public static boolean SIBLING_PURITY_ON_ELIMINATE = false;

  /** Skip sibling purity checking */
  public static boolean SKIP_SIBLING_PURITY = !SIBLING_PURITY_ON_ELIMINATE;

  /* ********************************************************************* */


  /** Return code for duplicate entry */
  public static int POSET_DUPLICATE = -1;

  /** Error code for general error */
  public static int POSET_ERROR = 0;

  /** Return code for entry into the root set */
  public static int POSET_ROOT = 1;

  /** Return code for entry into non-root set */
  public static int POSET_NOTROOT = 2;

  /** Return code for non-duplicate entry */
  public static int POSET_NOTDUPLICATE = 3;

  /** Return code for contained color ad filter */
  public static int POSET_CONTAINED = -1;

  public static int POSET_NOT_CONTAINED = -1;

  // Vector of listeners
  private Vector listeners = new Vector();

  public static String DELETE_EVENT_ELIMINATE = "delEventEliminate";

  public static String DELETE_EVENT = "delEvent";

  public static String DELETE_EVENT_BEFORE = "delEventBefore";

  public static String DELETE_EVENT_AFTER = "delEventAfter";

  public static String ADD_EVENT = "addEvent";

  public static String ADD_EVENT_BEFORE = "addEventBefore";

  public static String ADD_EVENT_AFTER = "addEventAfter";

  public static String RESET = "reset";

  /** For temporary operation */
  public Vector nodeCache = new Vector();

  /** Stores nodes in add if this is true (for SRF) */
  public static boolean useNodeCache = false;

  // Pertaining to DoubleForest match
  public final int QUERYTOPROFILE = 0;
  public final int PROFILETOQUERY = 1;

  /* ********************************************************************* */
  /* ********************************************************************* */
  /* ********************************************************************* */
  /**
   * Sets the settings for the peer to peer routing mode
   */
  public void setDistributed() {
    // eliminate all
    REMOVE_ALL_REDUNDANCY = true;
    // No basic color-eliminatin, done by external routine
//    NO_ELIMINATE = true;
    // no covering filters will be added
    NO_COVERED_UPDATES = true;
  }

  /**
   * Sets the settings for the hierarchical. This is not a good model
   * for peer-to-peer, because the second level may contain additional
   * entries.
   */
  public void setDistributedWithEliminationAtDel() {
    // eliminate all
    REMOVE_ALL_REDUNDANCY = true;
    // force update at del only (may contain redundant entries until del)
    ELIMINATE_ALL_DEL = true;
    // No basic color-eliminatin, done by external routine
    //  NO_ELIMINATE = true;
    // no covering filters will be added
    NO_COVERED_UPDATES = true;
  }

  /** Set local mode */
  public void setLocal() {
    // no changes to default settings
  }

  /* ********************************************************************* */
  /* ********************************************************************* */

  public static void main(String[] args) {
    // Simple test follows
    Filter f1 = new Filter("field1", EventTriplet.TYPE_INT,
                           "rng 200 300|rng 0 100");

    Filter f2 = new Filter("field1", EventTriplet.TYPE_INT,
                           "rng 9 14");

    Filter f3 = new Filter("field1", EventTriplet.TYPE_INT,
                           "rng 201 209");

    Filter f4 = new Filter("field1", EventTriplet.TYPE_INT,
                           "rng 202 206");

    Filter f5 = new Filter("field1", EventTriplet.TYPE_INT,
                           "rng 301 400");

    Filter f6 = new Filter("field1", EventTriplet.TYPE_INT,
                           "rng 199 280");

    FilterForest p = new FilterForest();
    int ops = 0;
    try {
      p.add(f3, "koe3");
      ops += p.comparisons;
      p.add(f5, "koe3");
      ops += p.comparisons;
      p.add(f2, "koe3");
      ops += p.comparisons;
      p.add(f6, "koe3");
      ops += p.comparisons;
      p.add(f4, "koe3");
      ops += p.comparisons;
      p.add(f1, "koe1");
      ops += p.comparisons;

      System.out.println("Ops total: " + ops);

      p.remove(f1, "koe1");
    }
    catch (PosetException e) {
      e.printStackTrace();
    }
    System.out.println("Ops: " + p.comparisons);

    try {
      PrintWriter fout = new PrintWriter(new FileWriter(
          "testposet.dat"), false);
      p.printDotFull(fout);
      fout.close();
    }
    catch (Exception e2) {
      e2.printStackTrace();
    }
    System.out.println(p.toString());
  }

  /**
   * Adds a change listener to the data structure
   * @param cl PosetChangeListener
   */
  public void addListener(PosetChangeListener cl) {
    if (!listeners.contains(cl)) {
      listeners.add(cl);
    }
  }

  /**
   * Adds a change listener to the data structure
   * @param cl PosetChangeListener
   * @return boolean
   */
  public boolean removeListener(PosetChangeListener cl) {
    return listeners.remove(cl);
  }

  /**
   * Fires an internal event to the listeners
   * @param event String
   * @param x IFBase
   */
  public void fireEventToListeners(String event, IFBase x) {
    if (DISABLE_LISTENERS) {
      return;
    }
    for (int i = 0; i < listeners.size(); i++) {
      PosetChangeListener pl = (PosetChangeListener) listeners.elementAt(i);
      pl.eventOccurred(event, x);
    }
  }

  /** Return cover test count for insert/remove
   * @return int the cover count number
   */
  public int getOps() {
    return comparisons;
  }

  public void resetOps() {
    comparisons = 0;
  }

  /**
   * Add both profiles and queries in random order. Used to test structure.
   * @param pro Vector
   * @param que Vector
   * @param biglist Vector
   * @return int
   * @throws PosetException
   */
  public synchronized int addRandomly(Vector pro, Vector que, Vector biglist) throws
      PosetException {
    int cost = 0;

    for (int i = 0; i < biglist.size(); i++) {
      IFBase f = (IFBase) biglist.elementAt(i);
      comparisons = 0;
      if (pro.contains(f)) {
        if (add(f, ( (ITemplate) f).getInterface()) ==
            FilterForest.POSET_ERROR) {
          // break;
        }
        cost += comparisons;
      }
      else {
        if (add(f, ( (ITemplate) f).getInterface()) ==
            FilterForest.POSET_ERROR) {
          // break;
        }
        cost += comparisons;
      }
    }
    return cost;
  }

  /**
   * Lookup profiles and queries in random order. Used to test structure.
   * @param pro Vector
   * @param que Vector
   * @param biglist Vector
   * @return int
   * @throws PosetException
   */
  public synchronized int lookupRandomly(Vector pro, Vector que, Vector biglist) throws
      PosetException {
    int cost = 0;

    for (int i = 0; i < biglist.size(); i++) {
      IFBase f = (IFBase) biglist.elementAt(i);
      comparisons = 0;
      if (pro.contains(f)) {
        Vector v2 = match(f, 1, false, null, null);
      }
      else {
        Vector v2 = match(f, 0, false, null, null);
      }
      cost += comparisons;
    }
    return cost;
  }

  /**
   * Delete profiles and queries in random order. Used to test structure.
   * @param pro Vector
   * @param que Vector
   * @param biglist Vector
   * @return int
   * @throws PosetException
   */
  public synchronized int deleteRandomly(Vector pro, Vector que, Vector biglist) throws
      PosetException {
    int cost = 0;
    for (int i = 0; i < biglist.size(); i++) {
      comparisons = 0;
      IFBase f = (IFBase) biglist.elementAt(i);
      if (pro.contains(f)) {
        if (this.remove(f, ( (ITemplate) f).getInterface()) ==
            FilterForest.POSET_ERROR) {
          // break;
        }
        cost += comparisons;
      }
      else {
        if (this.remove(f, ( (ITemplate) f).getInterface()) ==
            FilterForest.POSET_ERROR) {
          // break;
        }
        cost += comparisons;
      }
    }
    return cost;
  }

  /** Update overheads using getOverhead and by setting setTempOverhead
   *
   */
  public void updateOverhead() {
    // Update root set weights
    for (int i = 0; i < nodes.size(); i++) {
      Node n = (Node) nodes.elementAt(i);
      Vector fs = FilterForest.getCoveredNodes(n); // includes the node
      double cost = 0;
      for (int j = 0; j < fs.size(); j++) {
        Template t = (Template) fs.elementAt(j);
        cost += t.getOverhead();
      }
      ( (Template) n.getData()).setTempOverhead(cost);
    }
  }

  /**
   * Returns all nodes as Nodes
   * @return Vector
   */
  public Vector getAll() {
    Vector ret = new Vector();
    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      Node n = (Node) e.nextElement();
      ret.add(n);
    }
    return ret;
  }

  /**
   * Returns all nodes as IFBase
   * @return Vector
   */
  public Vector getAllNodes() {
    Vector ret = new Vector();
    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      Node n = (Node) e.nextElement();
      ret.add(n.getData());
    }
    return ret;
  }

  /**
   * Returns all objects in the forest as templates. Note that the
   * interfaces do not reflect the current situation because they
   * are collapsed!
   * @return Vector
   */
  public Vector getAllAsTemplates() {
    Vector ret = new Vector();
    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      Node n = (Node) e.nextElement();
      if (n.getData() instanceof Template) {
        ITemplate t = (ITemplate) n.getData();
        if (!connectedToRoot(n)) {
          System.out.println("Found a node not connected to root!!");
        }
        ret.add(t);
      }
    }
    return ret;
  }

  /**
   * Print ifaces
   * @param x Object
   */
  public void printAllIface(Object x) {

    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      Node n = (Node) e.nextElement();
      IFBase t = (IFBase) n.getData();
      if (t.getTid().contains(x)) {
        System.out.println("************************");
        System.out.println("Template:" + t.toString());
        System.out.println("Tid: " + t.getTid());
        System.out.println("## Printing to root");
        printColorsFromNode(n, null);
      }
      else if ( ( (Object) ( (ITemplate) t).getInterface()).equals(x)) {
        System.out.println("Missing tid!");
        System.out.println("Template:" + t.toString());
        System.out.println("Tid: " + t.getTid());
      }
    }
    System.out.println("Tested : " + templates.size());
  }

  /** Returns true if the node is connected to root
   * @param n the Node
   * @return true if connected
   */
  public boolean connectedToRoot(Node n) {
    Node t = n;
    while (t.parent != null) {
      if (!t.parent.getData().covers(t.getData())) {
        System.out.println("Fatal: parent didn't cover child");
        System.out.println("Parent: " + t.parent.getData());
        System.out.println("Child: " + t.getData());
        return false;
      }
      t = t.parent;
    }
    if (!nodes.contains(t)) {
      return false;
    }
    return true;
  }

  /**
   * Debug method
   * @param en Notification
   * @param iface Object
   * @return boolean
   */
  public boolean testMatch(Notification en, Object iface) {
    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      Node n = (Node) e.nextElement();
      IFBase t = (IFBase) n.getData();
      if (t.getTid().contains(iface) && t.match(null, en)) {
        return true;
      }
    }
    return false;
  }

  /**
   * Match a notification using the poset or forest. Assumes that
   * the objects are ITemplates and groups them by their interface name
   *  (getInterface).
   *  @param e the notification to match
   *  @return TripletTable contains the matched objects
   */
  public TripletTable matchT(Object e) {
    if (! (e instanceof Notification)) {
      return null;
    }
    comparisons = 0;
    TripletTable ret = new TripletTable();
    Vector stack = new Vector();

    // Create an imaginary node
    Node imaginary = new Node();
    imaginary.children.addAll(nodes);
    stack.add(imaginary);
    // Insert type-specific optimizations here
    // Insert a pre-selection process here

    boolean quit = false;
    Vector result = new Vector();

    // Traverse the tree
    while (stack.size() > 0 && !quit) {
      Node n = (Node) stack.remove(0);
      if (n != null) {
        Vector ci = null;
        synchronized (n.getChildren()) {
          ci = (Vector) n.getChildren().clone();
        }
        if (ci != null) {
          Iterator k = ci.iterator();
          while (k.hasNext()) {
            Node n2 = (Node) k.next();
            IFBase t = (IFBase) n2.getData();

            // if the current node has been matched already
            // we may skip the matching
            if (result.containsAll(t.getTid())) {
              if (n2.getChildren().size() > 0) {
                stack.add(n2);
              }
            }
            else {
              comparisons++;

              if (t.match(null, e)) {
                if (n2.getChildren().size() > 0) {
                  stack.add(n2);
                }
                ITemplate _t = (ITemplate) t;
//                ret.insert(_t.getInterface(), _t);

                       // This expansion is needed for proper multicast
                       for (int i = 0; i < t.getTid().size(); i++)  {
                           Object o = t.getTid().elementAt(i);
                           ret.insert((String)o, _t);
                       }

              }
            }
          }
        }
      }
    }
    return ret;
  }

  /**
   * Match an object against the data structure
   * @param e Object
   * @return Vector
   */
  public Vector match(Object e) {
    if (! (e instanceof Notification)) {
      return null;
    }

    comparisons = 0;

    // note no optimizations
    Vector stack = new Vector();

    // Create an imaginary node
    Node imaginary = new Node();
    imaginary.children.addAll(nodes);
    stack.add(imaginary);
    boolean quit = false;
    HashSet result = new HashSet();

    // Traverse the tree
    while (stack.size() > 0 && !quit) {
      Node n = (Node) stack.remove(0);
      if (n != null) {
        Vector ci = null;
        synchronized (n.getChildren()) {
          ci = (Vector) n.getChildren().clone();
        }

        if (ci != null) {
          Iterator k = ci.iterator();
          while (k.hasNext()) {
            Node n2 = (Node) k.next();
            IFBase t = (IFBase) n2.getData();

            // if the current node has been matched already
            // we may skip the matching

            if (result.containsAll(t.getTid())) {
              if (n2.getChildren().size() > 0) {
                stack.add(n2);
              }
            }
            else {
              comparisons++;

              if (t.match(null, e)) {
                addToResults(result, t);
                if (n2.getChildren().size() > 0) {
                  stack.add(n2);
                }
              }
            }
          }
        }
      }
    }
    Vector res = new Vector(result);
    return res;
  }

  /**
   * Adds nodes to the given vector
   * @param rs Vector
   * @param nx Node
   * @return boolean
   */
  public boolean addNodesToRoot(Vector rs, Node nx) {
    Node t = nx;
    if (t.parent == null) {
      if (!rs.contains(t.getData())) {
        rs.add(t.getData());
      }
    }
    else {
      while (t.parent != null) {
        if (!rs.contains(t.getData())) {
          rs.add(t.getData());
        }
        addChildren(rs, t);
        t = t.parent;
      }
    }
    return true;
  }

  public void addChildren(Vector rs, Node nx) {
    if (nx.getChildren() != null) {
      for (int i = 0; i < nx.getChildren().size(); i++) {
        Node k = (Node) nx.getChildren().elementAt(i);
        if (!rs.contains(k)) {
          rs.add(k);
        }
      }
    }
  }

  /**
   * Checks if the given node is in the bound
   * @param bound HashSet
   * @param nx Node
   * @return boolean
   */
  public boolean parentInBound(HashSet bound, Node nx) {
    Node t = nx;

    while (t != null) {
      if (bound.contains(t)) {
        return true;
      }
      t = t.parent;
    }
    return false;
  }

  /**
   *
   * @param nx Node
   * @return Node
   */
  public Node findFirstCover(Node nx, IFBase x) {
    Node t = nx;

    if (t.parent == null) {
      comparisons++;
      if (nx.getData().covers(x)) {
        return nx;
      }
    }
    else {
      while (t.parent != null) {
        comparisons++;
        if (t.getData().covers(x)) {
          return t;
        }
        t = t.parent;
      }
    }
    return null;
  }

  /**
   * @param rs Vector
   * @param nx Node
   * @param x IFBase
   */
  public void addCoveredNodes(Vector rs, Node nx, IFBase x) {
    Node t = nx;

    if (t.parent == null) {
      comparisons++;
      if (x.covers(t.getData())) {
        if (!rs.contains(t.getData())) {
          rs.add(t.getData());
        }
      }
      return;
    }
    else {
      while (t.parent != null) {
        comparisons++;
        if (x.covers(t.getData())) {
          if (!rs.contains(t.getData())) {
            rs.add(t.getData());
          }
        }
        else {
          return;
        }
        // root
        t = t.parent;
      }
    }
    return;
  }

  /**
   *
   * @param rs Vector
   * @param nx Node
   * @param x IFBase
   */
  public void removeOverlappingNodesToRoot(Vector rs, Node nx, IFBase x) {
    Node t = nx;

    if (t.parent == null) {
      comparisons++;
      if (!x.overlaps(nx.getData())) {
        rs.remove(nx.getData());
      }
      return;
    }
    else {
      while (t.parent != null) {
        comparisons++;
        if (!x.overlaps(nx.getData())) {
          rs.remove(nx.getData());
        }
        else {
          return;
        }
        t = t.parent;
      }
    }
    return;
  }

  /** For testing DoubleForest */
  public int DF_counter = 0;

  public Vector matchQueryToProfilesOld(IFBase e, HashSet startSet,
                                        HashSet endSet) {
    HashSet ret = new HashSet();
    Vector stack = new Vector();
    stack.addAll(endSet);
    HashSet visited = new HashSet();
    visited.addAll(endSet);

    while (!stack.isEmpty()) {
      Node x = (Node) stack.remove(0);
      ret.add(x.getData());
      Node parent = x.getParent();
      if (parent != null) {
        if (!visited.contains(parent) && startSet.contains(parent)) {
          if (e.covers(parent.getData())) {
            stack.add(parent);
            visited.add(parent);
          }
        }
      }
    }
    return new Vector(ret);
  }

  public Vector matchProfileToQueriesOld(IFBase e, HashSet startSet,
                                         HashSet endSet) {
    HashSet ret = new HashSet();
    Vector stack = new Vector();
    stack.addAll(endSet);
    HashSet visited = new HashSet();
    visited.addAll(endSet);

    while (!stack.isEmpty()) {
      Node x = (Node) stack.remove(0);
      ret.add(x.getData());
      Vector children = x.getChildren();
      if (children != null && children.size() > 0) {
        for (int i = 0; i < children.size(); i++) {
          Node y = (Node) children.elementAt(i);
          if (!visited.contains(y) && startSet.contains(y)) {
            if (y.getData().covers(e)) {
              stack.add(y);
              visited.add(y);
            }
          }
        }
      }
    }
    return new Vector(ret);
  }

  /**
   * Match an object against the data structure using the given matching
   * type.
   * @param e IFBase
   * @param matchType int 0 = e covers x, 1 = x covers e, 2 = overlap P->Q,
   * 3 Q->P
   * @param rootOnly boolean rootOnly if true processes only the first two levels
   * @param startSet HashSet the upper bound set (optimization) as nodes
   * @param endSet HashSet the lower bound set (optimization). The children of
   * this set are not processed. As nodes.
   * @return Vector
   */
  public Vector match(IFBase e, int matchType, boolean rootOnly,
                      HashSet startSet, HashSet endSet) {
    comparisons = 0;

    Vector stack = new Vector();
    HashSet res = new HashSet();

    // Create an imaginary node
    Node imaginary = new Node();
    imaginary.children.addAll(nodes);

    stack.add(imaginary);
    // Insert type-specific optimizations here
    // Insert a pre-selection process here

    boolean quit = false;

// random shuffle for experiments

    // Traverse the tree
    while (stack.size() > 0 && !quit) {
      Node n = (Node) stack.remove(0);

      if (n != null) {
        Vector ci = n.getChildren();
        Template parent = (Template) n.getData();

        if (ci != null && ci.size() > 0) {
          Iterator k = ci.iterator();

          while (k.hasNext()) {
            Node n2 = (Node) k.next();
            IFBase t = (IFBase) n2.getData();

            boolean mthc = false;
            boolean iterate = false;

            if (endSet != null && endSet.contains(n2)) {
              mthc = true;
            }
            // Upper and lower bound optimizations
            else if (startSet != null && !startSet.contains(n2)) {
              // if there is an upper bound, the node must be contained in it
              // to be interesting

              // if parent is in the set, this is a transition node so
              // we can stop iterating
              if (matchType == QUERYTOPROFILE) {
                iterate = true; // skip this node
              }
            }
            else if (matchType == QUERYTOPROFILE) { // added query
              comparisons++;
              DF_counter++;
              if (e.covers(t)) {
                mthc = true;
              }
              else {
                iterate = true; // have to check children
              }
            }
            else if (matchType == PROFILETOQUERY) { // added profile
              comparisons++;
              DF_counter++;
              if (t.covers(e)) {
                mthc = true;
              }
            }

            if (mthc) {
              if (matchType == QUERYTOPROFILE) {
                // all all descendants, simply add this and stop iter
                mthc = false;
                iterate = false;
                addDescendantsAndSelf(res, n2);
              }
              else if (!res.contains(n2)) {
                res.add(n2);
              }
            }

            if (iterate || mthc) {
              if (n2.getChildren().size() > 0) {
                // if rootOnly flag is set, we process only the first two levels
                if (rootOnly) {
                  if ( (n2.getParent() == null || n2.getParent().getParent() == null)) {
                    stack.add(n2);
                  }
                }
                else {
                  stack.add(n2);
                }
              }
            }
          }
        }
      }
    }
    Vector ret = new Vector(res);
    return ret;
  }

  /**
   * Match subspace
   * @param e IFBase
   * @param rootOnly boolean
   * @param startSet HashSet
   * @param endSet HashSet
   * @param exclude HashSet
   * @return Vector
   */
  public Vector matchSubspace(IFBase e, boolean rootOnly, HashSet startSet,
                              HashSet endSet, HashSet exclude) {
    comparisons = 0;

    Vector stack = new Vector();
    HashSet res = new HashSet();

    // Create an imaginary node
    Node imaginary = new Node();
    imaginary.children.addAll(nodes);

    stack.add(imaginary);
    // Insert type-specific optimizations here
    // Insert a pre-selection process here

    boolean quit = false;

    // Traverse the tree
    while (stack.size() > 0 && !quit) {
      Node n = (Node) stack.remove(0);

      if (n != null) {
        Vector ci = n.getChildren();
        Template parent = (Template) n.getData();

        if (ci != null && ci.size() > 0) {
          Iterator k = ci.iterator();

          while (k.hasNext()) {
            Node n2 = (Node) k.next();
            IFBase t = (IFBase) n2.getData();

            boolean mthc = false;
            boolean iterate = false;
            boolean lower = false;

            if (endSet != null && endSet.contains(n2)) {
              lower = true;
            }
            // Upper and lower bound optimizations
            if (startSet != null && !startSet.contains(n2)) {
              ;
            }
            else if (exclude != null && exclude.contains(n2)) {
              ;
            }
            else {

              if (lower) {
                mthc = true;
              }
              else {
                comparisons++;
                DF_counter++;
                if (t.covers(e)) {
                  mthc = true;
                }
              }

              if (mthc) {
                if (!res.contains(n2)) {
                  res.add(n2.getData());
                }
              }
            }
            if (iterate || mthc) {
              if (n2.getChildren().size() > 0) {
                // if rootOnly flag is set, we process only the first two levels
                if (rootOnly) {
                  if ( (n2.getParent() == null || n2.getParent().getParent() == null)) {
                    stack.add(n2);
                  }
                }
                else {
                  stack.add(n2);
                }
              }
            }
          }
        }
      }
    }
    Vector ret = new Vector(res);
    return ret;
  }

  /**
   * Match an object against the data structure using the given matching
   * type.
   * @param e IFBase
   * @param rootOnly boolean rootOnly if true processes only the first two levels
   * @param startSet HashSet the upper bound set (optimization) as nodes
   * @param endSet HashSet the lower bound set (optimization). The children of
   * this set are not processed. As nodes.
   * @return Vector
   */
  public Vector matchOverlap(IFBase e, boolean rootOnly,
                             HashSet startSet, HashSet endSet) {

    comparisons = 0;

    Vector stack = new Vector();
    HashSet res = new HashSet();

    // Create an imaginary node
    Node imaginary = new Node();
    imaginary.children.addAll(nodes);

    stack.add(imaginary);
    // Insert type-specific optimizations here
    // Insert a pre-selection process here

    boolean quit = false;

    // Traverse the tree
    while (stack.size() > 0 && !quit) {
      Node n = (Node) stack.remove(0);

      if (n != null) {
        Vector ci = n.getChildren();
        Template parent = (Template) n.getData();

        if (ci != null && ci.size() > 0) {
          Iterator k = ci.iterator();

          while (k.hasNext()) {
            Node n2 = (Node) k.next();
            IFBase t = (IFBase) n2.getData();

            boolean mthc = false;
            boolean iterate = false;
            boolean lower = false;

            if (startSet != null && !startSet.contains(n2)) {
              ; // skip
            }
            else {
              if (endSet != null && endSet.contains(n2)) {
                mthc = true;
              }
              else {
                comparisons++;
                DF_counter++;
                if (e.overlaps(t)) {
                  mthc = true;
                }
              }

              if (mthc) {
                if (!res.contains(n2)) {
                  res.add(n2);
                }
              }
            }
            if (iterate || mthc) {
              if (n2.getChildren().size() > 0) {
                // if rootOnly flag is set, we process only the first two levels
                if (rootOnly) {
                  if ( (n2.getParent() == null || n2.getParent().getParent() == null)) {
                    stack.add(n2);
                  }
                }
                else {
                  stack.add(n2);
                }
              }
            }
          }
        }
      }
    }
    return new Vector(res);
  }

  /**
   * Add all descendants of the given node to the input vector
   * @param v Vector
   * @param b IFBase
   */
  public void addDescendantsAndSelf(Vector v, Node b) {
    Vector a = b.getChildren();
    if (!v.contains(b.getData())) {
      v.add(b.getData());
    }
    for (int i = 0; i < a.size(); i++) {
      Node z = (Node) a.elementAt(i);
      if (!v.contains(z.getData())) {
        v.add(z.getData());
      }
      Vector c = z.getChildren();
      if (c != null && c.size() > 0) {
        addDescendantsAndSelf(v, z);
      }
    }
  }

  /**
   * Add all descendants of the given node to the input vector
   * @param v HashSet
   * @param b IFBase
   */
  public void addDescendantsAndSelf(HashSet v, Node b) {
    Vector a = b.getChildren();
    if (!v.contains(b)) {
      v.add(b);
    }
    for (int i = 0; i < a.size(); i++) {
      Node z = (Node) a.elementAt(i);
      if (!v.contains(z)) {
        v.add(z);
      }
      Vector c = z.getChildren();
      if (c != null && c.size() > 0) {
        addDescendantsAndSelf(v, z);
      }
    }
  }

  /**
   * Updates the result set
   * @param result HashSet
   * @param f IFBase
   */
  public void addToResults(HashSet result, IFBase f) {
    for (int j = 0; j < f.getTid().size(); j++) {
      Object o = f.getTid().elementAt(j);
      if (!result.contains(o)) {
        result.add(o);
      }
    }

  }

  /**
   * Updates the result set
   * @param result Vector
   * @param f IFBase
   */
  public void addToResults(Vector result, IFBase f) {
    synchronized (f.getTid()) {
      for (int j = 0; j < f.getTid().size(); j++) {
        Object o = f.getTid().elementAt(j);
        if (!result.contains(o)) {
          result.add(o);
        }
      }
    }
  }

  /**
   * Eliminate all nodes that are covered by x and have the same color.
   * Note: add must be called after this for the node.
   * @param xc IFBase
   * @param col Object
   * @return boolean false if x is covered by a filter with the same color, otherwise true
   * @throws PosetException
   */
  public boolean eliminateAll(IFBase xc, Object col) throws PosetException {

    if (NO_ELIMINATE) {
      return true;
    }

    Vector stack = new Vector();

    if (debug) {
      System.out.println("eliminateAll called: " + xc);
    }

    // Create an imaginary node
    Node imaginary = new Node();
    imaginary.children.addAll(nodes);
    stack.add(imaginary);
    boolean quit = false;

    // Traverse the tree
    while (stack.size() > 0 && !quit) {
      Node n = (Node) stack.remove(0);

      if (n != null) {
        Vector ci = null;

        synchronized (n.getChildren()) {
          ci = (Vector) n.getChildren().clone();
        }

        if (ci != null) {
          Iterator k = ci.iterator();

          while (k.hasNext()) {
            Node n2 = (Node) k.next();
            IFBase t = (IFBase) n2.getData();

            if (t.getTid().contains(col)) {
              if (debug) {
                System.out.println("eliminateAll: Covered");
              }

              // If the filters are identical we skip, this will be handled
              // by the normal mechanisms.

              if (t.getCoverId().equals(xc.getCoverId())) {
                if (debug) {
                  System.out.println("eliminateAll: Identical found");
                }
                continue;
              }
              comparisons++;
              // check if covered by the same color
              if (t.covers(xc)) {
                if (debug) {
                  System.out.println("eliminateAll: same color found:" + col);
                }
                return false;
              }

              comparisons++;
              if (xc.covers(t)) {
                //System.out.println("eliminateAll: xc covers:"+t);
                if (n2.getParent() != null) {
                  if (debug) {
                    System.out.println("Eliminating");
                  }
                  eliminateC(n2, col, n);
                }
              }
            }
            if (n2.getChildren().size() > 0) {
              if (debug) {
                System.out.println("eliminateAll: adding to stack");
              }
              stack.add(n2);
            }

          }
        }
      }
    }
    return true;
  }

  /**
   * Move siblings from one node to another. Pushes them down the tree.
   * @param x Node
   * @param parent Node
   * @throws PosetException
   */
  void moveChildrenDown(Node x, Node parent) throws PosetException {
    if (debug) {
      System.out.println("MoveChildrenDown: parent " + parent);
      System.out.println("MoveChildrenDown: x " + x);
    }
    Vector v1 = x.getChildren();

    if (v1 == null) {
      return;
    }

    boolean d = this.NO_ELIMINATE;
    NO_ELIMINATE = true;

    for (int i = 0; i < v1.size(); i++) {
      Node n = (Node) v1.elementAt(i);

      if (debug) {
        System.out.println("MoveChildrenDown: Old parent for node: " +
                           n.getParent());
      }

      if (parent == null) {
        if (debug) {
          System.out.println(
              "MoveChildrenDown: Moving node after eliminate [parent=null]: " +
              n);
        }
        add(n);
      }
      else {
        if (debug) {
          System.out.println(
              "MoveChildrenDown: Moving node after eliminate [parent!=null]: " +
              n);
        }
        // add must be capable of performing relocations further down the
        // tree to guarantee sibling-purity
        add(parent, n);
      }
      if (debug) {
        System.out.println("MoveChildrenDown: New parent for node: " +
                           n.getParent());
      }
    }
    v1.clear();
    NO_ELIMINATE = d;
  }

  /**
   * Test method for finding a covering filter (other than the given filter)
   * with the same color
   * @param xc IFBase
   * @param col Object
   * @return boolean true if found
   * @throws PosetException
   */
  public boolean findCoveringTest(IFBase xc, Object col) throws PosetException {
    Vector stack = new Vector();

    // Create an imaginary node
    Node imaginary = new Node();
    imaginary.children.addAll(nodes);
    stack.add(imaginary);
    boolean quit = false;

    // Traverse the tree
    while (stack.size() > 0 && !quit) {
      Node n = (Node) stack.remove(0);

      if (n != null) {
        Vector ci = null;

        synchronized (n.getChildren()) {
          ci = (Vector) n.getChildren().clone();
        }

        if (ci != null) {
          Iterator k = ci.iterator();

          while (k.hasNext()) {
            Node n2 = (Node) k.next();
            IFBase t = (IFBase) n2.getData();
            if (t.covers(xc)) {
              // If the filters are identical we skip, this will be handled
              // by the normal mechanisms.
              if (t.getCoverId().equals(xc.getCoverId())) {
                continue;
              }
              // check if covered by the same color
              else if (t.getTid().contains(col)) {
                System.out.println("Covering node:" + t.toString());
                System.out.println("Tids: " + t.getTid());
                return true;
              }
              else if (n2.getChildren().size() > 0) {
                stack.add(n2);
              }
            }
          }
        }
      }
    }
    return false;
  }

  /**
   * Returns the total overhead of the root nodes
   * @return double
   */
  public double getOverhead() {
    double cost = 0;

    // Update root set weights
    for (int i = 0; i < nodes.size(); i++) {
      Node n = (Node) nodes.elementAt(i);
      Template t = (Template) n.getData();
      cost += t.getOverhead();
    }
    return cost;
  }

  /** Sort trees
   * @param sort true if sort is to be used
   */
  public void setSort(boolean sort) {
    // not used
  }

  public FilterForest() {
  }

  /** Sorts the trees by size: smallest size subtree first in order
   * to optimize insertions
   */
  public void sortTreesBySize() {
    int depth[] = new int[nodes.size()];

    // Get the subtreesizes for all nodes (the slow way)
    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      Node n = (Node) e.nextElement();
      n.setSubTreeSize(getNodeCount(n));
    }

    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      Node n = (Node) e.nextElement();
      Vector v = n.getChildren();
      Collections.sort(v);
    }

    // Finally sort root node
    Collections.sort(nodes);
  }

  /**
   * Returns the top-level objects in the poset (the non-covered set)
   * @return Vector the non-covered objects
   */
  public Vector get() {
    Vector ret = new Vector();
    for (int i = 0; i < nodes.size(); i++) {
      Node n = (Node) nodes.elementAt(i);
      if (n != null) {
        ret.add(n.data);
      }
    }
    return ret;
  }

  /**
   * Returns the top-level objects in the poset (the non-covered set)
   * @return Vector the non-covered objects
   */
  public Vector getAsNodes() {
    Vector ret = new Vector();
    for (int i = 0; i < nodes.size(); i++) {
      Node n = (Node) nodes.elementAt(i);
      if (n != null) {
        ret.add(n);
      }
    }
    return ret;
  }

  /**
   * Eliminate colors
   * @param x Node
   * @param tids Vector
   * @param start Node
   * @throws PosetException
   */
  void eliminateC(Node x, Vector tids, Node start) throws PosetException {

    if (NO_ELIMINATE) {
      return;
    }

    if (debug) {
      System.out.println("FilterForest: eliminateC");
    }

    if (x == null) {
      return;
    }

    if (tids == null || tids.size() == 0) {
      return;
    }

    IFBase cur = (IFBase) x.getData();

    for (int k = 0; k < tids.size(); k++) {
      Object color = tids.elementAt(k);
      if (cur.getTid() != null && cur.getTid().contains(color)) {
        cur.getTid().remove(color);
      }
    }

    // execute callback for extensions
    eliminateColorCallback(x, tids, start);

    // move the children under parent
    if (cur.getTid() == null ||
        (cur.getTid() != null && cur.getTid().size() == 0)) {
      if (test) {
        System.out.println("* removing node");
      }
      templates.remove(cur.getCoverId());

      // Notify listeners about removal
      fireEventToListeners(DELETE_EVENT_ELIMINATE, cur);

      if (x.getParent() != null) {
        x.getParent().getChildren().remove(x);

        if (SIBLING_PURITY_ON_ELIMINATE) {
          moveChildrenDown(x, x.getParent());
          // the add method performs color-elimination for necessary parts
        }
        else {
          // We clone the vector so flag is not required
          Vector vec = (Vector) x.getChildren().clone();

          moveSiblings(x.getChildren(), x.getParent().getChildren(), null,
                       x.getParent());
          for (int i = 0; i < vec.size(); i++) {
            Node n = (Node) vec.elementAt(i);
            eliminateC(n, (Vector) tids.clone(), start);
          }
        }
      }
      else {
        System.out.println("eliminate: parent was null");
        throw new PosetException();
      }

      x.getChildren().clear();
      x.setParent(null);
    }
    else {
      // We clone the vector so flag is not required
      tids.addAll(cur.getTid()); // add remaining colors
      Vector vec = (Vector) x.getChildren().clone();
      for (int i = 0; i < vec.size(); i++) {
        Node n = (Node) vec.elementAt(i);
        eliminateC(n, (Vector) tids.clone(), start);
      }
    }

    return;
  }

  /**
   * Eliminate a vector of colors
   * @param x Node
   * @param v Vector
   * @throws PosetException
   */
  public void eliminateColor(Node x, Vector v) throws PosetException {
    if (NO_ELIMINATE) {
      return;
    }

    Vector c = (Vector) x.getChildren().clone();

    if (v == null) {
      v = new Vector();
    }

    // ensure that the colors of the start level are included
    for (int p = 0; p < x.getData().getTid().size(); p++) {
      Object col = x.getData().getTid().elementAt(p);
      if (!v.contains(col)) {
        v.add(col);
      }
    }

    for (int i = 0; i < c.size(); i++) {
      Node y = (Node) c.elementAt(i);
      eliminateC(y, v, x);
    }
  }

  /**
   * Eliminate a single color
   * @param x Node
   * @param color Object
   * @throws PosetException
   */
  public void eliminateColor(Node x, Object color) throws PosetException {
    if (NO_ELIMINATE) {
      return;
    }
    if (debug) {
      System.out.println("size: " + x.getChildren());
    }
    Vector c = (Vector) x.getChildren().clone();
    for (int i = 0; i < c.size(); i++) {
      Node y = (Node) c.elementAt(i);
      eliminateC(y, color, x);
    }
  }

  /**
   * Eliminate a color
   * @param x Node
   * @param color Object
   * @param start Node
   * @throws PosetException
   */
  void eliminateC(Node x, Object color, Node start) throws PosetException {
    if (NO_ELIMINATE) {
      return;
    }

    if (debug) {
      System.out.println("FilterForest: eliminateC single");
    }

    if (x == null || color == null) {
      return;
    }
    IFBase cur = (IFBase) x.getData();

    if (test) {
      System.out.println("Eliminate-color: " + color + "Looking at colors:" +
                         cur.getTid());
    }

    if (cur.getTid() != null && cur.getTid().contains(color)) {
      if (debug || test) {
        System.out.println("Eliminate_color: color found");
      }

      cur.getTid().remove(color);

      // execute callback for extensions
      eliminateColorCallback(x, color, start);

      // move the children under parent
      if (cur.getTid().size() == 0) {
        if (debug || test) {
          System.out.println("Eliminate_color: Removing node and terminating: " +
                             cur.toString() + " color: " + color);
        }

        templates.remove(cur.getCoverId());

        // Notify listeners about removal
        fireEventToListeners(DELETE_EVENT_ELIMINATE, cur);

        if (x.getParent() != null) {
          x.getParent().getChildren().remove(x);

          if (SIBLING_PURITY_ON_ELIMINATE) {
            moveChildrenDown(x, x.getParent());
          }
          else {
            moveSiblings(x.getChildren(), x.getParent().getChildren(), null,
                         x.getParent());
          }
        }
        else {
          System.out.println("eliminate: parent was null");
          throw new PosetException();
        }
        x.getChildren().clear();
        x.setParent(null);
      }
    }
    else {
      if (debug || test) {
        System.out.println("Eliminate_color: iterating tree");
        // We clone the vector so flag is not required
      }
      Vector vec = (Vector) x.getChildren().clone();
      for (int i = 0; i < vec.size(); i++) {
        Node n = (Node) vec.elementAt(i);
        // color index optimization: we skip subtrees that do not have the color
        eliminateC(n, color, start);
      }
    }
    return;
  }

  void printColorsFromNode(Node n, Node last) {
    System.out.println("--");
    System.out.println(n.getData().getTid());
    System.out.println(n.getData() + " hash: " + n.hashCode());
    if (last != null) {
      if (!n.getChildren().contains(last)) {
        System.out.println("Fatal: orphan detected!");
      }
    }
    if (n.parent != null) {
      printColorsFromNode(n.parent, n);
    }
    else {
      if (!nodes.contains(n)) {
        System.out.println("Fatal: root node not in root set");
      }
    }
  }

  boolean testMinimumColor(Node n, Vector v) {
    Vector cols = n.getData().getTid();
    if (cols.size() == 0) {
      System.out.println("Fatal: detected a node without colors");
      printColorsFromNode(n, null);
      System.exit(0);
    }
    for (int i = 0; i < v.size(); i++) {
      Object o = (Object) v.elementAt(i);
      if (cols.contains(o)) {
        System.out.println("Minimum color error: " + o);
        System.out.println("Input vector" + v);
        System.out.println("Current node colors:" + cols);

        System.out.println("Explanation: a node was detected with the " +
                           "same color as a covering node");

        printColorsFromNode(n, null);
        System.out.println("root");
        return false; // System.exit(0);
      }
    }

    v.addAll(cols);
    if (n.getChildren() != null) {
      for (int i = 0; i < n.getChildren().size(); i++) {
        if (!testMinimumColor( (Node) n.getChildren().elementAt(i),
                              (Vector) v.clone())) {
          return false;
        }
      }
    }
    return true;
  }

  /**
   * Tests that the current nodes covers its children
   * @param x Node
   * @return boolean
   */
  public boolean testNodeCover(Node x) {
    Vector v = x.getChildren();
    if (v == null) {
      return true; // nothing to test
    }

    for (int i = 0; i < v.size(); i++) {
      Node n = (Node) v.elementAt(i);
      if (n.getData().covers(x.getData())) {
        // In some cases (merging forest) you may have identical nodes
        // that have different colors..
        if (!x.getData().covers(n.getData())) {
          System.out.println("Fatal: node was covered by its child!");
          System.out.println("Node:" + x.getData());
          System.out.println("Child:" + n.getData());
          return false;
        }
      }
    }
    return true;
  }

  public boolean testSiblingPurity(Node n) {
    if (SKIP_SIBLING_PURITY) {
      return true;
    }
    Node parent = n.getParent();
    if (parent != null) {
      Vector v = parent.getChildren();
      if (v != null) {
        for (int i = 0; i < v.size(); i++) {
          Node c = (Node) v.elementAt(i);
          if (c == n) {
            continue;
          }
          for (int k = 0; k < parent.getData().getTid().size(); k++) {
            Object col = parent.getData().getTid().elementAt(k);
            if (c.getData().getTid().contains(col)) {
              System.out.println("Color-elim error in testSiblingPurity");
              return false;
            }
          }
          if (n.getData().covers(c.getData())) {
            for (int j = 0; j < n.getChildren().size(); j++) {
              Node o = (Node) n.getChildren().elementAt(j);
              System.out.println("Child:" + j + " " + o.hashCode());
            }
            System.out.println("Node: " + n.getData() + " hash:" + n.hashCode());
            System.out.println("Node tid: " + n.getData().getTid());
            System.out.println("Covers: " + c.getData() + " hash:" + c.hashCode());
            System.out.println("Covered parent hash:" + c.getParent().hashCode());
            System.out.println("Covered tid: " + c.getData().getTid());
            System.out.println("But latter is not a child of the former");
            if (n.getChildren() != null) {
              System.out.println("Containment: " + n.getChildren().contains(c));
            }

            System.out.println("Parent: " + n.getParent() + " hash: " +
                               n.getParent().hashCode());
            System.out.println("Parent's parent: " + n.getParent().getParent());

            int d = 0;
            Node x = n;
            while (x.getParent() != null) {
              x = x.getParent();
              d++;
            }
            System.out.println("Depth: " + d);
            return false;
          }
        }
      }
    }
    return true;
  }

  public String runTests() {
    String myName = this.getClass().getName();
    String str = "[" + myName + "] test structure\n\n";
    boolean ret = true;

    str += "TEST: Minimum color property\n";

    // testing minimum color property
    for (int i = 0; i < nodes.size(); i++) {
      Node k = (Node) nodes.elementAt(i);
      if (k.getData().getTid().size() == 0) {
        str += "A root node was found with no color!\n";
        str += k.getData() + "\n";
        ret = false;
      }
      if (!testMinimumColor(k, new Vector())) {
        str += "Fatal: testColorIndex: minimum color property " +
            " violation\n";
        ret = false;
        break;
      }
    }

    if (ret) {
      str += "RESULT: Minimum color property ok\n";
    }
    else {
      str += "RESULT: Minimum color property failed\n";
    }

    str += "TEST: Checking integrity (orphans, covering, connection to root)\n";
    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      Node n = (Node) e.nextElement();

      if (!connectedToRoot(n)) {
        System.out.println("Found a node in templates not connected to root");
        System.out.println(n);
        System.exit(0);
      }
      if (!testNodeCover(n)) {
        System.exit(0);
      }

      if (!testSiblingPurity(n)) {
        System.out.println("Sibling purity failed");
        System.exit(0);
      }

      if (n.getData().getTid().size() == 0) {
        System.out.println("Fatal: node with empty color set");
        System.out.println("Probably an orphan");
        System.out.println(n);
        System.exit(0);
      }
      if (REMOVE_ALL_REDUNDANCY) {
        try {
          // Test if a covering color was found
          for (int p = 0; p < n.getData().getTid().size(); p++) {
            Object col = n.getData().getTid().elementAt(p);
            if (findCoveringTest(n.getData(), col)) {
              System.out.println("Found a covering color: " + col);
              System.out.println("Node: " + n.getData());
              System.out.println("Parent: " + n.getParent());
              System.exit(0);
            }
          }
        }
        catch (Exception e2) {
          e2.printStackTrace();
          ret = false;
        }
      }
    }
    if (ret) {
      str += "RESULT: integrity ok\n";
    }
    else {
      str += "RESULT: integrity failed\n";
    }
    return str;
  }

  /** Tests the color index
   *  @return boolean false if tests fail
   */
  public boolean testColorIndex() {
    boolean ret = true;
    String myName = this.getClass().getName();

    System.out.println("[" + myName + "] testColorIndex");

    // testing minimum color property
    for (int i = 0; i < nodes.size(); i++) {
      Node k = (Node) nodes.elementAt(i);
      if (k.getData().getTid().size() == 0) {
        System.out.println("A root node was found with no color!");
        System.out.println(k.getData());
        System.exit(0);
      }
      if (!testMinimumColor(k, new Vector())) {
        System.out.println("Fatal: testColorIndex: minimum color property " +
                           " violation");
        ret = false;
        break;
      }
    }
    System.out.println("Checking orphans ************************");
    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      Node n = (Node) e.nextElement();

      if (!testSiblingPurity(n)) {
        System.out.println("Sibling purity failed");
        System.exit(0);
      }

      if (!connectedToRoot(n)) {
        System.out.println("Found a node in templates not connected to root");
        System.out.println(n);
        System.exit(0);
      }
      if (!testNodeCover(n)) {
        System.exit(0);
      }
      if (n.getData().getTid().size() == 0) {
        System.out.println("Fatal: node with empty color set");
        System.out.println("Probably an orphan");
        System.out.println(n);
        System.exit(0);
      }
      if (REMOVE_ALL_REDUNDANCY) {
        try {
          // Test if a covering color was found
          for (int p = 0; p < n.getData().getTid().size(); p++) {
            Object col = n.getData().getTid().elementAt(p);
            if (findCoveringTest(n.getData(), col)) {
              System.out.println("Found a covering color: " + col);
              System.out.println("Node: " + n.getData());
              System.out.println("Parent: " + n.getParent());
              System.exit(0);
            }
          }
        }
        catch (Exception e2) {
          e2.printStackTrace();
          ret = false;
        }
      }
    }
    return ret;
  }

  /**
   * Returns the top-level objects in the poset (the non-covered set)
   * @return Vector the non-covered objects (Nodes)
   */
  public Vector getNodes() {
    return nodes;
  }

  /** Reset the data structure */
  public void reset() {
    resetNodes();
    nodes.clear();
    templates.clear();
    // Notify listeners about reset
    fireEventToListeners(RESET, null);
  }

  /**
   * Returns true if the object is removable
   * @param t IFBase
   * @param color Object
   * @return boolean
   */
  public boolean removable(IFBase t, Object color) {
    Node n = (Node) templates.get(t.getCoverId());

    if (n == null) {
      //System.out.println("FilterForest: node not found");
      return false;
    }

    IFBase f = (IFBase) n.getData();
    if (f == null) {
      System.out.println("FilterForest: object was null");
      return false;
    }
    if (f.getTid().contains(color)) {
      return true;
    }
    return false;
  }

  /** Removes the given object from the poset.
   * @param t the object to add, must be a cloned copy of the original object
   * @param color the color
   * @throws PosetException
   * @return 0 if remove failed, 1 if remove was the root node, greater than 1 otherwise,
   * returns -1 if remove was succesful but there were multiple copies
   */
  public synchronized int remove(IFBase t, Object color) throws PosetException {
    comparisons = 0;
    int comparisons2 = 0; // reset counter for remove

    if (test) {
      System.out.println("REMOVE CALLED for: " + color);
    }

    // first find the object
    Node n = (Node) templates.get(t.getCoverId());

    if (n == null) {
      if (debug) {
        System.out.println("Fatal: Object not found:" + t.getCoverId());
      }
      return POSET_ERROR;
    }

    int ret = remove(n, color);

    return ret;
  }

  /**
   * Remove a node
   * @param n Node
   * @param color Object
   * @throws PosetException
   * @return int
   */
  protected synchronized int remove(Node n, Object color) throws PosetException {
    IFBase f = (IFBase) n.getData();

    if (f.getTid() == null) {
      System.out.println("FilterForest: remove: getid() == null");
      return POSET_ERROR;
    }

    boolean rem = false;

    if (NO_INTERFACE_CHANGES) {
      rem = true;
    }
    else {
      rem = f.getTid().remove(color);
    }

    if (!rem && f.getTid().size() > 0) { // should be only one
      System.out.println("Note: Color not found: " + color + " tid: " +
                         f.getTid());
      return POSET_ERROR;
    }
    else if (!rem && f.getTid().size() == 0) {
      System.out.println("Note: getTid() was 0 in remove but remove failed ");
      System.out.println("Probably duplicate insertions/deletions. " +
                         "Garbage collecting");
    }

    if (f.getTid().size() > 0) {
      removeNodeEvent(n, color); // callback for extensions

      // Notify listeners about removal
      fireEventToListeners(DELETE_EVENT, n.getData());

      if (test) {
        System.out.println("REMOVE ENDED (DUPLICATE)");
      }

      return POSET_DUPLICATE; // remove was succesful but a copy still exists
    }

    removeNodeEvent(n, n.getData().getTid()); // callback for extensions

    f.getTid().add(color); // ensure that the listener gets this
    // Notify listeners about removal
    fireEventToListeners(DELETE_EVENT_BEFORE, n.getData());
    f.getTid().remove(color);

    // permanent deletion
    templates.remove(f.getCoverId());

    if (test) {
      System.out.println("Remove: deleted node for " + color);

    }
    if (nodes.contains(n)) { // root node

      if (n.parent != null) {
        System.out.println("Fatal: in root set but not root");
      }

      nodes.remove(n); // the list remains sorted

      if (test) {
        System.out.println("top level node");
      }

      Vector ch = (Vector) n.children.clone();

      for (int i = 0; i < ch.size(); i++) {
        Node n2 = (Node) ch.elementAt(i);
        add(n2); // add to the poset
      }

      n.children.clear();
      ch = null;
      n = null;

      if (test) {
        System.out.println("After remove");
        this.testColorIndex();
      }
      // Notify listeners about removal
      fireEventToListeners(DELETE_EVENT_AFTER, f);

      return POSET_ROOT;
    }
    else { // some other node
      Node parent = n.parent;

      if (parent == null) {
        System.out.println("Parent was null but not a top-level node");
      }
      else {
        if (!parent.children.remove(n)) { // remove this branch
          //        if (debug) {
          System.out.println(
              "Warning: failed to remove from parent list. Contains:"
              + parent.children.contains(n));
          //        }
        }
      }
      if (test) {
        System.out.println("not top level node");

        // add the descendants to the poset again
      }
      Vector ch = (Vector) n.children.clone();
      for (int i = 0; i < ch.size(); i++) {
        Node n2 = (Node) ch.elementAt(i);
        if (test) {
          System.out.println("Calling add: " + i + "/" + ch.size());
        }
        add(parent, n2); // add to the poset
      }

      n.children.clear();
      ch = null;
      n = null;

      if (test) {
        System.out.println("After remove 2");
      }
      // Notify listeners about removal
      fireEventToListeners(DELETE_EVENT_AFTER, f);

      return POSET_NOTROOT;
    }
  }

  /**
   * Get filter objects for the given Node.
   * @param pos Node the node
   * @return Vector filters
   */
  public static Vector getCoveredNodes(Node pos) {
    Vector v = new Vector();

    v.add(pos.getData());

    if (pos == null) {
      System.out.println("FilterPoset: getCoveringNode, node not found");
      return v;
    }

    if (pos.children == null || pos.children.size() == 0) {
      return v;
    }

    Vector stack = new Vector();
    stack.add(pos);

    while (stack.size() > 0) {
      Node n = (Node) stack.remove(0);
      if (n != null) {
        if (n.children != null) {
          Iterator i = n.children.iterator();
          while (i.hasNext()) {
            Node n2 = (Node) i.next();
            v.add(n2.data);
            stack.add(n2);
          }
        }
      }
    }
    return v;
  }

  /** Returns the set of covered nodes (below this node in the tree)
   *  Used to enhance matching. Does not include the given object.
   *  Note that this does not return ALL covered nodes!
   * @param t the node
   * @param v Vector that will contain the covered nodes
   */
  public void getCoveredNodes(IFBase t, Vector v) {
    Node pos = (Node) templates.get(t.getCoverId());
    if (pos == null) {
      if (debug) {
        System.out.println("FilterPoset: getCoveringNode, node not found");
      }
      return;
    }

    if (pos.children == null || pos.children.size() == 0) {
      return;
    }

    Vector stack = new Vector();
    stack.add(pos);

    while (stack.size() > 0) {
      Node n = (Node) stack.remove(0);
      if (n != null) {
        if (n.children != null) {
          Iterator i = n.children.iterator();
          while (i.hasNext()) {
            Node n2 = (Node) i.next();
            v.add(n2.data);
            stack.add(n2);
          }
        }
      }
    }
  }

  /**
   * Returns true if covered nodes exist
   * @param t IFBase
   * @return boolean
   */
  public boolean coveredNodesExist(IFBase t) {
    Node pos = (Node) templates.get(t.getCoverId());
    if (pos == null) {
      if (debug) {
        System.out.println("FilterPoset: getCoveringNode, node not found");
      }
      return false;
    }

    if (pos.children.size() > 0) {
      return true;
    }

    return false;
  }

  /** Returns the set of covering nodes (above this node in the tree)
   *  Used to enhance matching. Does not include the given object.
   * @param t the node
   * @param v of covering ICovering objects (cast to the proper class)
   */
  public synchronized void getCoveringNodes(IFBase t, Vector v) {
    Node n = (Node) templates.get(t.getCoverId());
    if (n == null) {
      if (debug) {
        System.out.println("FilterPoset: getCoveringNode, node not found");
      }
      return;
    }

    n = n.parent;
    while (n != null) {
      v.add(n.data);
      n = n.parent;
    }

    return;
  }

  /**
   * Lookup a node
   * @param t IFBase
   * @return IFBase
   */
  public IFBase lookup(IFBase t) {
    Node n0 = (Node) templates.get(t.getCoverId());

    // Check duplicate entries
    if (n0 != null) {
      return n0.data;
    }
    return null;
  }

  /**
   * Lookup a Node object
   * @param t IFBase
   * @return Node
   */
  public Node lookupNode(IFBase t) {
    Node n0 = (Node) templates.get(t.getCoverId());
    return n0;
  }

  /**
   * Get the children sorted based on the color property.
   * Used for balancing. Clones the vector.
   * @param x Node
   * @param color Object
   * @return Vector
   */
  synchronized Vector getLevel(Node x, Object color) {
    Vector v = null;

    if (x == null) {
      v = (Vector) nodes.clone();
    }
    else {
      v = (Vector) x.getChildren().clone();
    }
    // random shuffle here for experiments

    return v;
  }

  /* EXTENSION MECHANISMS **********************************************/

  /**
   * Called when a color is eliminated. Reserved for extension mechanisms.
   * @param x Node
   * @param tids Vector
   * @param start Node
   */
  public void eliminateColorCallback(Node x, Vector tids, Node start) {
    // NOP
  }

  /**
   * Called when a color is eliminated. Reserved for extension mechanisms.
   * @param x Node
   * @param color Object
   * @param start Node
   */
  public void eliminateColorCallback(Node x, Object color, Node start) {
    // NOP
  }

  /**
   * Called when a node is added. Reserved for extension mechanisms.
   * @param n Node
   * @param color Object
   */
  public void addNodeEvent(Node n, Object color) {
    // NOP
  }

  /**
   * Called when a node is added. Reserved for extension mechanisms.
   * @param n Node
   * @param colors Vector
   * @throws PosetException
   */
  public void addNodeEvent(Node n, Vector colors) throws PosetException {
    // NOP
  }

  /**
   * Called when a node is removed. Reserved for extension mechanisms.
   * @param n Node
   * @param colors Vector
   */
  public void removeNodeEvent(Node n, Vector colors) {
    // NOP
  }

  /**
   * Called when a node is removed. Reserved for extension mechanisms.
   * @param n Node
   * @param color Object
   */
  public void removeNodeEvent(Node n, Object color) {
    // NOP
  }

  /* EXTENSION MECHANISMS **********************************************/

  /** Checks if the filter is already contained in the structure
   * @param t the object to add
   * @param color the color/interface to add
   * @throws PosetException
   * @return code
   */
  public int contains(IFBase t, Object color) throws PosetException {
    Node n0 = (Node) templates.get(t.getCoverId());

    // Check duplicate entries
    if (n0 != null) {
      IFBase f = (IFBase) n0.getData();

      if (f.getTid() == null) {
        return POSET_ERROR;
      }

      if (!f.getTid().contains(color)) { // if iface/color is new we perform
        // Check that there does not exist a parent with the same color
        Node p = n0.getParent();
        while (p != null) {
          if (p.getData().getTid().contains(color)) {
            return POSET_CONTAINED; // a covering node with the same color was found
          }
          p = p.getParent();
        }
        return POSET_NOT_CONTAINED;
      }
      else {
        return POSET_CONTAINED;
      }
    }
    return FilterForest.POSET_NOTDUPLICATE;
  }

  /** Adds the given object to the poset. Resets covering ops counter.
   * @param t the object to add. Note object is not cloned. Only one TID
   * @param iface the interface/color
   * @throws PosetException
   * @return int status code
   */
  public synchronized int add(IFBase t, Object iface) throws PosetException {
    comparisons = 0; // reset counter

    int ret = addNC(t, iface);

    return ret;
  }

  /** Adds the given object to the poset after del. Resets covering ops counter.
   * @param t the object to add. Note object is not cloned. Only one TID
   * @param iface the interface/color
   * @throws PosetException
   * @return int status code
   */
  public synchronized int addAfterDel(IFBase t, Object iface) throws
      PosetException {
    comparisons = 0; // reset counter
    // no reason to eliminate colors in this case

    SKIP_REDUNDANCY_CHECK = true;
    int ret = addNC(t, iface);
    SKIP_REDUNDANCY_CHECK = false;
    return ret;
  }

  /** Adds the given object to the poset. Does not reset cov. ops counter.
   * @param t the object to add. Note object is not cloned. Only one TID
   * @param iface the interface/color
   * @throws PosetException
   * @return int status code
   */
  protected synchronized int addNC(IFBase t, Object iface) throws
      PosetException {
    boolean flag = false;
    int ret = POSET_ERROR;

    if (test) {
      System.out.println("ADD CALLED");
    }

    if (iface == null) {
      System.out.println("Forest: iface was: " + iface);
      throw new PosetException();
    }

    if (useNodeCache && nodeCache != null) { // SRF, P set comp.
      nodeCache.clear();
    }

    // Eliminate all colors

    if (!NO_ELIMINATE) {
      if (REMOVE_ALL_REDUNDANCY && !SKIP_REDUNDANCY_CHECK) {
        if (!eliminateAll(t, iface)) {
          return POSET_NOTROOT; // eliminated
        }
      }
    }

    Node n0 = (Node) templates.get(t.getCoverId());

    // Check duplicate entries
    if (n0 != null) {
      IFBase f = (IFBase) n0.getData();

      if (f.getTid() == null) {
        System.out.println("Fatal: tid was null for existing object");
        return POSET_ERROR;
      }

      if (!f.getTid().contains(iface)) { // if iface/color is new we perform
        if (!NO_ELIMINATE) {
          // Check that there does not exist a parent with the same color
          Node p = n0.getParent();
          while (p != null) {
            if (p.getData().getTid().contains(iface)) {
              return POSET_NOTROOT; // a covering node with the same color was found
            }
            p = p.getParent();
          }
          // eliminate color procedure
          eliminateColor(n0, iface);
        }

        // we can add the iface
        if (!NO_INTERFACE_CHANGES) {
          f.addTid(iface);
        }

        addNodeEvent(n0, iface); // callback for extensions

        // Notify listeners about addition
        fireEventToListeners(ADD_EVENT, n0.getData());
      }
      else {
        System.out.println("Warning: tried to insert identical filter with " +
                           "the same color");
        System.out.println("New Object: " + t.hashCode() + " Old: " +
                           f.hashCode());
        System.out.println("Color: " + iface);
      }
      if (test) {
        System.out.println("AFTER ADD 1");
        this.testColorIndex();
      }
      return POSET_DUPLICATE;
    }

    // Notify listeners about addition
    fireEventToListeners(ADD_EVENT_BEFORE, t);

    // clearly this is the first time the object is added
    // so we clear the tid vector
    if (!NO_INTERFACE_CHANGES) {
      if (t.getTid() != null) {
        t.getTid().clear();
      }
      // and add the first interface/color
      t.addTid(iface);
    }

    Vector l2 = getLevel(null, iface);

    boolean cover1 = false;
    boolean cover2 = false;

    for (int i = 0; i < l2.size(); i++) {
      Node n = (Node) l2.elementAt(i);

      if (n.data == null) {
        continue;
      }

      comparisons++;
      cover1 = n.data.covers(t);
      // add the node to descend to in any case
      if (useNodeCache && nodeCache != null) { // SRF, P set comp.
        nodeCache.add(n);
      }

      if (cover1) {
        ret = POSET_NOTROOT;
        if (!NO_ELIMINATE) {
          if (containsColor(n.data, iface)) {
            flag = true;
            if (debug) {
              System.out.println("Color covered " + t.toString() + " Under " +
                                 n.getData().toString());
            }
            break;
          }
        }
        Vector ifaces = new Vector();
        Node n2 = findSubTree(n, null, t, ifaces, iface);
        if (ifaces.size() > 0) { // a covering color was already there
          flag = true;
          break;
        }

        if (debug) {
          System.out.println("Inserting " + t.toString() + " Under " +
                             n2.getData().toString());
          // add n3 under n2 and move any covered nodes under n3
        }
        Node n3 = new Node();
        n3.parent = n2;
        n3.count = counter++;
        n3.data = t;

        moveSiblings(n2.children, n3.children, n3.data, n3);

        n2.children.add(n3);
        templates.put(t.getCoverId(), n3);
        eliminateColor(n3, iface);
        addNodeEvent(n3, iface); // callback for extensions
        flag = true;
        break;
      }
    }

    // The new subscription covers a top level subscription
    if (!cover1) { // add new node as a root
      ret = POSET_ROOT;
      Node n3 = new Node();
      n3.count = counter++;
      n3.parent = null;
      n3.data = t;
      // moves all covered top level nodes under the new node

      moveSiblings(nodes, n3.children, n3.data, n3);
      nodes.add(n3);
      templates.put(t.getCoverId(), n3);

      if (debug) {
        System.out.println("new root: " + t.toString());
      }
      eliminateColor(n3, iface);
      addNodeEvent(n3, iface); // callback for extensions
      flag = true;
    }

    // Notify listeners about addition
    fireEventToListeners(ADD_EVENT_AFTER, t);

    return ret;
  }

  /** Adds the given object to the poset. Does not reset ops counter.
   * Called when a root node was removed. A special case for colors!
   * The color index is updated since the node came from outside and
   * it is not a local change.
   * @param node the Node object to add
   * @throws PosetException
   */
  protected synchronized void add(Node node) throws PosetException {
    boolean flag = false;

    if (test) {
      System.out.println("ADD ROOT CALLED for " + node.getData().getTid() +
                         " hash: " + node.hashCode());
    }

    for (int i = 0; i < nodes.size(); i++) {
      Node n = (Node) nodes.elementAt(i);

      if (n.data == null) {
        continue;
      }

      boolean cover = n.data.covers(node.data);
      comparisons++;

      if (cover) {
        Vector ifaces = new Vector();
        ifaces.addAll(n.getData().getTid());

        // ifaces contains now all colors from n to n2, including them both
        Node n2 = findSubTree(n, null, node.data, ifaces, null);

        if (n2 == null) { // a covering color was already there
          flag = true;
          System.out.println("Filter poset: fatal error in add(node)");
          throw new PosetException();
        }

        // add node under n2 and move any covered nodes under node
        node.parent = n2;
        if (SIBLING_PURITY_ON_ELIMINATE) {
          moveSiblingsPure(n2.children, node.children, node.data, node);
        }
        else {
          moveSiblings(n2.children, node.children, node.data, node);
        }

        n2.children.add(node);

        if (test) {
          System.out.println("+Adding under: " + n2.hashCode());
          System.out.println("+Eliminating color for: " + ifaces);
          System.out.println("++Adding under: " + n2.getData().getTid());
        }
        subTreeAdd(node, n2, ifaces);
        flag = true;
        break;
      }
    }

    // descendants did not cover, add to root level
    if (!flag) {
      if (test) {
        System.out.println("Adding as new node");
        // add new poset
      }
      node.parent = null;
      nodes.add(node);

      // the uncovered node cannot cover any root node

      addNodeEvent(node, node.getData().getTid()); // callback for extensions
    }

    if (test) {
      System.out.println("End add root");
    }
    test = false;
  }

  /**
   * Adds the given Vector to the poset.
   * @param v the objects to add
   * @throws PosetException
   * @return int the number of elements inserted starting from the beginning
   */
  public synchronized int add(Vector v) throws PosetException {
    if (v == null || v.size() == 0 || ! (v.firstElement() instanceof ITemplate)) {
      return 0;
    }
    for (int i = 0; i < v.size(); i++) {
      IFBase ic = (IFBase) v.elementAt(i);
      if (add(ic, ( (ITemplate) ic).getInterface()) == POSET_ERROR) {
        // break;
      }
    }
    return v.size();
  }

  /** Adds the given object to the poset. Does not reset ops counter.
   * The node is relocated within the tree and the color index has
   * already been updated!
   * @param parent the parent Node
   * @param node the Node object to add
   * @throws PosetException
   */
  protected synchronized void add(Node parent, Node node) throws PosetException {
    boolean flag = false;

    if (test) {
      System.out.println("ADD CALLED for " + node.getData().getTid());
    }

    if (parent.children == null) {
      parent.children = new Vector();
    }

    for (int i = 0; i < parent.children.size(); i++) {
      Node n = (Node) parent.children.elementAt(i);

      if (n.data == null) {
        continue;
      }

      comparisons++;
      boolean cover1 = n.data.covers(node.data);

      if (cover1) {
        Vector ifaces = new Vector();
        ifaces.addAll(n.getData().getTid());
        // ifaces contains now all colors from n to n2, including them both
        Node n2 = findSubTree(n, null, node.data, ifaces, null);
        if (n2 == null) {
          flag = true;
          System.out.println("Filter poset: fatal error in add(parent,node)");
          throw new PosetException();
        }

        // add node under n2 and move any covered nodes under node
        node.parent = n2;
        n2.children.add(node);

        if (SIBLING_PURITY_ON_ELIMINATE) {
          moveSiblingsPure(n2.children, node.children, node.data, node);
        }
        else {
          moveSiblings(n2.children, node.children, node.data, node);
        }

        if (test) {
          System.out.println("Eliminating color for: " + ifaces);
        }

        subTreeAdd(node, n2, ifaces);
        flag = true;
        break;
      }
    }

    if (!flag) {
      if (test) {
        System.out.println("Adding as new node");
      }

      node.parent = parent;
      parent.children.add(node);

      // this is needed to maintain sibling purity with dels
      if (REMOVE_ALL_REDUNDANCY) {
        if (SIBLING_PURITY_ON_ELIMINATE) {
          moveSiblingsPure(parent.children, node.children, node.data, node);
        }
      }
      /*
            else {
              moveSiblings(parent.children, node.children, node.data, node);
            }
            if (!REMOVE_ALL_REDUNDANCY) {
              eliminateColor(node, null);
            }
       */
      addNodeEvent(node, node.getData().getTid()); // callback for extensions
    }

    if (test) {
      System.out.println("ADD ENDED CALLED");
    }
  }

  /**
   * Adds a subtree to the given location and eliminates the given set
   * of interfaces.
   * @param node Node
   * @param parent Node
   * @param ifaces Vector
   * @throws PosetException
   */
  void subTreeAdd(Node node, Node parent, Vector ifaces) throws PosetException {
    // If the root of the added subtree is eliminated
    // --> the parent's set of children may have new elements
    // We start from node, and ifaces contains all the necessary colors
    // including the colors of the parent
    eliminateC(node, ifaces, parent);

    // reconstruct the colors of the parent
    addNodeEvent(parent, ifaces); // callback for extensions

    // reconstruct the given node if it exists
    if (node.getData().getTid().size() > 0) {
      addNodeEvent(node, ifaces); // callback for extensions
    }
  }

  /**
   * Move siblings from one node to another.
   * @param v1 The node whose children are moved
   * @param v2 The target node's set of children
   * @param t2 The object that must cover all moved, if null all are moved
   * @param parent The new parent node
   */
  void moveSiblings(Vector v1, Vector v2, IFBase t2, Node parent) {
    Vector removeList = new Vector();
    for (int i = 0; i < v1.size(); i++) {
      Node n = (Node) v1.elementAt(i);
      if (n != null && n != parent) {
        boolean cover1 = false;
        if (t2 == null) {
          cover1 = true;
        }
        else {
          comparisons++; // increment comparison counter
          cover1 = t2.covers(n.data);
        }
        if (cover1) {
          v2.add(n);
          n.parent = parent;
          removeList.add(n);
        }
      }
    }
    v1.removeAll(removeList);
  }

  /**
   * Move siblings from one node to another.
   * @param v0 The node whose children are moved
   * @param v2 The target node's set of children
   * @param t2 The object that must cover all moved, if null all are moved
   * @param parent The new parent node
   */
  void moveSiblingsPure(Vector v0, Vector v2, IFBase t2, Node parent) throws
      PosetException {
    Vector removeList = new Vector();

    // We turn eliminate off from the add methods
    boolean d = this.NO_ELIMINATE;
    // If we know that the forest is non-redundant, it is not necessary to eliminate anymore
    if (REMOVE_ALL_REDUNDANCY) {
      NO_ELIMINATE = true;
    }
    Vector v1 = (Vector) v0.clone();

    for (int i = 0; i < v1.size(); i++) {
      Node n = (Node) v1.elementAt(i);
      if (n != null && n != parent) {
        boolean cover1 = false;
        if (t2 == null) {
          cover1 = true;
        }
        else {
          comparisons++; // increment comparison counter
          cover1 = t2.covers(n.data);
        }
        if (cover1) {
          v0.remove(n);

          if (debug) {
            System.out.println("moveSiblingsPure: removing " + n.hashCode());
          }

          add(parent, n);

          if (debug) {
            System.out.println("Popped:" + n.hashCode());
          }
        }
      }
    }
    NO_ELIMINATE = d;
  }

  /**
   * Returns the node count
   * @return int
   */
  public int getNodeCount() {
    int removeCount = 0;
    Vector stack = new Vector();
    for (int i = 0; i < nodes.size(); i++) {
      Node n = (Node) nodes.elementAt(i);
      stack.add(n);
    }
    while (stack.size() > 0) {
      Node n = (Node) stack.elementAt(0);
      if (n != null) {
        if (n.children != null) {
          for (int i = 0; i < n.children.size(); i++) {
            Node n2 = (Node) n.children.elementAt(i);
            stack.add(n2);
          }
        }
      }
      stack.remove(0);
      removeCount++;
    }
    return removeCount;
  }

  private int getNodeCount(Node pos) {
    int removeCount = 0;
    Vector stack = new Vector();
    stack.add(pos);
    while (stack.size() > 0) {
      Node n = (Node) stack.elementAt(0);
      if (n != null) {
        if (n.children != null) {
          for (int i = 0; i < n.children.size(); i++) {
            Node n2 = (Node) n.children.elementAt(i);
            stack.add(n2);
          }
        }
      }
      stack.remove(0);
      removeCount++;
    }
    return removeCount;
  }

  private int resetNodes() {
    int removeCount = 0;
    Vector stack = new Vector();
    for (int i = 0; i < nodes.size(); i++) {
      Node n = (Node) nodes.elementAt(i);
      stack.add(n);
    }
    while (stack.size() > 0) {
      Node n = (Node) stack.elementAt(0);
      if (n != null) {
        if (n.children != null) {
          for (int i = 0; i < n.children.size(); i++) {
            Node n2 = (Node) n.children.elementAt(i);
            stack.add(n2);
          }
        }
      }
      n = (Node) stack.remove(0);
      n.children.clear();
      n.children = null;
      n = null;
      removeCount++;
    }
    return removeCount;
  }

  protected boolean containsColor(IFBase x, Object y) {
    if (x == null || y == null) {
      return false;
    }
    if (x.getTid() != null && x.getTid().contains(y)) {
      return true;
    }
    return false;
  }

  /**
   * Finds the last covering subtree in a tree
   * @param n Node
   * @param parent Node
   * @param t IFBase
   * @param ifaces Vector
   * @param color Object
   * @return Node
   */
  protected Node findSubTree(Node n, Node parent, IFBase t, Vector ifaces,
                             Object color) {
    if (n.children == null || n.children.size() == 0) {
      if (test) {
        System.out.println("No children");
      }
      return n;
    }
    try {
      Vector nv = n.getChildren();
      for (int i = 0; i < nv.size(); i++) {
        Node n2 = (Node) nv.elementAt(i);
        if (n2 != null) {
          comparisons++; // increment counter
          boolean cover1 = n2.data.covers(t);

          if (test) {
            System.out.println(n2.hashCode() + " covers: " + cover1);
          }

          if (debug) {
            System.out.println("\n" + n2.data.toString() + " covers " +
                               t.toString() + " :" + cover1);
            System.out.println("Cur color: " +
                               n2.data.getTid().firstElement().toString());
            System.out.println("Node color: " +
                               t.getTid().firstElement().toString());
          }

          // add the node in any case
          if (useNodeCache && nodeCache != null) { // SRF, P set comp.
            nodeCache.add(n2);
          }

          if (cover1) {
            if (debug) {
              System.out.println("** Subtree for color " + color + " cur: " +
                                 n2.getData().getTid());

              // if we are inserting a single element and the same color is already there
              // we exit.
            }
            if (!NO_ELIMINATE) {
              if (color != null && containsColor(n2.data, color)) {
                ifaces.add(n2);
              }
            }
            if (ifaces != null && color == null) {
              IFBase f = (IFBase) n2.getData();
              Vector tid = f.getTid();
              if (tid != null) {
                for (int k = 0; k < tid.size(); k++) {
                  if (!ifaces.contains( (Object) tid.elementAt(k))) {
                    ifaces.add(tid.elementAt(k));
                  }
                }
              }
            }

            return findSubTree(n2, n, t, ifaces, color);
          }
        }
      }
    }
    catch (Exception e) {
      e.printStackTrace();
      return null;
    }
    return n;
  }

  /* ********************************************************************* */
  /* ********************************************************************* */
  /* ********************************************************************* */
  /* ********************************************************************* */
  public void incOps() {
    comparisons++;
  }

  public void incOps(int inc) {
    comparisons += inc;
  }

  /* ********************************************************************* */
  /* ********************************************************************* */
  /* ********************************************************************* */
  /* ********************************************************************* */


  /**
   * Get the data structure as HTML
   * @return String
   */
  public String getHTMLOutput() {
    StringBuffer buf = new StringBuffer();
    buf.append("<html><body>");
    for (int i = 0; i < nodes.size(); i++) {
      Node n = (Node) nodes.elementAt(i);
      if (n != null) {
        printHtml(n, buf);
      }
    }
    buf.append("</body></html>");
    return buf.toString();
  }

  /**
   * Print roots
   * @param fn PrintWriter
   */
  public void printRoots(PrintWriter fn) {
    for (int i = 0; i < this.nodes.size(); i++) {
      Node t = (Node) nodes.elementAt(i);
      fn.println( ( (IFBase) t.getData()).toString());
    }
  }

  /**
   *  Print dot format output
   * @param fn PrintWriter
   */
  public void printDot(PrintWriter fn) {
    fn.println("digraph poset {");
    fn.println("size=\"6,6\";");
    fn.println("node [color=lightblue2, style=filled]");

    for (int i = 0; i < nodes.size(); i++) {
      Node n = (Node) nodes.elementAt(i);
      if (n != null) {
        printDot(n, fn);
      }
    }
    fn.println("}");
  }

  private void printDot(Node n, PrintWriter fn) {
    if (n.children == null || n.children.size() == 0) { // orphan node
      fn.println("\"" + n.count + "\";");
      return;
    }
    for (int i = 0; i < n.children.size(); i++) {
      Node n2 = (Node) n.children.elementAt(i);
      fn.println("\"" + n.count + "\"->\"" + n2.count +
                 "\";");
      printDot(n2, fn);
    }
  }

  public void printDotFull(PrintWriter fn) {
    fn.println("digraph poset {");
    fn.println("size=\"6,6\";");
    fn.println("node [color=lightblue2, style=filled]");

    for (int i = 0; i < nodes.size(); i++) {
      Node n = (Node) nodes.elementAt(i);
      printDotFull(n, fn);
    }
    fn.println("}");
  }

  void printDotFull(Node n, PrintWriter fn) {
    Filter f = null;
    String s1 = "";
    if (n.data instanceof ITemplate) {
      ITemplate t0 = (ITemplate) n.data;
      f = (Filter) t0.getFilters().firstElement();
      s1 = t0.toString();
    }
    else if (n.data instanceof Filter) {
      f = (Filter) n.data;
      s1 = f.toString();
    }
    if (n.children == null || n.children.size() == 0) { // orphan node
      fn.println("\"" + f.getFilter() + "\";");
      return;
    }
    try {
      for (int i = 0; i < n.children.size(); i++) {
        Node n2 = (Node) n.children.elementAt(i);
        Filter f2 = null;
        String s2 = "";
        if (n.data instanceof ITemplate) {
          ITemplate t = (Template) n2.data;
          f2 = (Filter) t.getFilters().firstElement();
          s2 = t.toString();
        }
        else if (n.data instanceof Filter) {
          f2 = (Filter) n2.data;
          s2 = f2.getFilter();
        }

        fn.println("\"" + s1 + "\"->\"" + s2 +
                   "\";");
        printDotFull(n2, fn);
      }
    }
    catch (Exception e) {
      e.printStackTrace();
    }
  }

  void printHtml(Node n, StringBuffer buf) {
    Filter f = null;
    String s1 = "";
    if (! (n.data instanceof ITemplate)) {
      return;
    }
    ITemplate t = (ITemplate) n.data;
    s1 = ManagementUI.printFilterTable(t);

    if (n.children == null || n.children.size() == 0) {

      buf.append("<table bgcolor=\"#AAAAAF\"><tr><td>");
      buf.append(s1 + "<br>");
      buf.append("</td></tr></table>");
      return;
    }
    try {
      for (int i = 0; i < n.children.size(); i++) {
        Node n2 = (Node) n.children.elementAt(i);
        if (n.data instanceof ITemplate) {
          ITemplate t2 = (Template) n2.data;

          String s2 = ManagementUI.printFilterTable(t2);
          buf.append("<table><tr><td>");
          buf.append(s1 + "</td><td>");
          buf.append("</td><td>--></td><td>");
          buf.append(s2);
          buf.append("</td></tr></table>");
          printHtml(n2, buf);
        }
      }
    }
    catch (Exception e) {
      e.printStackTrace();
    }
  }
}
