package fuegocore.notify.core.filter.poset;

import java.util.*;

import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.filter.poset.forwards.*;
import fuegocore.notify.util.*;

/**
 * The basic interface for filter posets and forests. The data structure
 * is used to store objects by their covering relation. The two concrete
 * implementations of this interface are: the Filter Forest and the Filters
 * Poset. The former stores objects only part of the relations whereas the
 * latter stores all relations between objects.
 */
public interface IFilterPoset {
  /**
   * Adds a change listener to the data structure
   * @param cl PosetChangeListener
   */
  public void addListener(PosetChangeListener cl);

  /**
   * Adds a change listener to the data structure
   * @param cl PosetChangeListener
   * @return boolean
   */
  public boolean removeListener(PosetChangeListener cl);

  /**
   * Update overhead counters. Used for load balancing.
   */
  public void updateOverhead();

  /** Sorts the trees by size.
   */
  public void sortTreesBySize();

  /** Return cover test count for insert/remove. Used for benchmarking.
   * @return int the cover count number
   */
  public int getOps();

  /**
   * Returns the top-level objects in the poset (the non-covered set)
   * Note that elements may have multiple interfaces (getTid() in IFBase)
   * @return Vector the non-covered objects
   */
  public Vector get();

  /**
   * Match a notification using the poset or forest.
   *  @param e the notification to match
   *  @return Vector contains the matched objects
   */

  public Vector match(Object e);

  /**
   * Match a notification using the poset or forest. Assumes that
   * the objects are Templates and groups them by their interface name
   *  (getInterface).
   *  @param e the notification to match
   *  @return TripletTable contains the matched objects
   */
  public TripletTable matchT(Object e);

  /**
   * Returns all nodes as IFBase
   * @return Vector
   */
  public Vector getAllNodes();

  /**
   * Returns the top-level objects in the poset (the non-covered set)
   * @return Vector the non-covered objects (Nodes)
   */
  public Vector getNodes();

  /** Reset the data structure */
  public void reset();

  /** Removes the given object from the poset.
   * @param t the object to remove
   * @param color the color/interface to remove
   * @throws PosetException
   * @return 0 if remove failed, 1 if remove was the root node, greater than 1 otherwise,
   * returns -1 if remove was succesful but there were multiple copies
   */
  public int remove(IFBase t, Object color) throws PosetException;

  /**
   * Returns true if the object is removable
   * @param t IFBase
   * @param color Object
   * @return boolean
   */
  public boolean removable(IFBase t, Object color);

  /** Returns the set of covered nodes (below this node in the tree)
   *  Used to enhance matching. Does not include the given object.
   * For forests it is not guaranteed that all covered nodes are returned.
   * @param t the node
   * @param v Vector that will contain the covered nodes
   */
  public void getCoveredNodes(IFBase t, Vector v);

  /**
   * Returns true if covered nodes exist for the given object.
   * @param t IFBase the input object
   * @return boolean true if covered nodes exist
   */
  public boolean coveredNodesExist(IFBase t);

  /** Returns the set of covering nodes (above this node in the tree)
   *  Used to enhance matching. Does not include the given object.
   * For forests it is not guaranteed that all covered nodes are returned.
   * @param t the node
   * @param v Vector of covering ICovering objects (cast to the proper class)
   */
  public void getCoveringNodes(IFBase t, Vector v);

  /**
   * Looks up the given object.
   * @param t IFBase the input object
   * @return IFBase the object or null if it was not found in the structure
   */
  public IFBase lookup(IFBase t);

  /** Adds the given object to the poset.
   * @param t the object to add
   * @param color the color/interface to add
   * @throws PosetException
   * @return code
   */
  public int add(IFBase t, Object color) throws PosetException;

  /** Checks if the filter is already contained in the structure
   * @param t the object to add
   * @param color the color/interface to add
   * @throws PosetException
   * @return code
   */
  public int contains(IFBase t, Object color) throws PosetException;

  /**
   * Adds the given Vector to the poset.
   * @param v the objects to add
   * @throws PosetException
   * @return int the number of elements inserted starting from the beginning
   */
  public int add(Vector v) throws PosetException;

  /**
   * Returns the number of nodes in the structure
   * @return int the number of nodes
   */
  public int getNodeCount();

  /**
   * Get the data structure as HTML
   * @return String
   */
  public String getHTMLOutput();
}
