package fuegocore.notify.core.filter.poset;

import java.io.*;
import java.util.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.filter.poset.forwards.*;
import fuegocore.notify.core.server.ui.*;
import fuegocore.notify.util.*;

/**
 * Old forest implementation for filter store that does not support
 * color elimination and allows duplicate interfaces.
 */
public class PlainFilterForest
    implements IFilterPoset {
  int counter = 0;
  int comparisons = 0; // for benchmarking
  Hashtable templates = new Hashtable();
  Vector nodes = new Vector();
  boolean debug = false; // for debug output
  boolean test = false;

  /* ********************************************************************* */

  /** Return code for duplicate entry */
  public static int POSET_DUPLICATE = -1;

  /** Error code for general error */
  public static int POSET_ERROR = 0;

  /** Return code for entry into the root set */
  public static int POSET_ROOT = 1;

  /** Return code for entry into non-root set */
  public static int POSET_NOTROOT = 2;

  /** Return code for non-duplicate entry */
  public static int POSET_NOTDUPLICATE = 3;

  /** Return code for contained color ad filter */
  public static int POSET_CONTAINED = -1;

  public static int POSET_NOT_CONTAINED = -1;

  /* ********************************************************************* */
  /* ********************************************************************* */
  /* ********************************************************************* */

  public static void main(String[] args) {
    // Simple test follows
    Filter f1 = new Filter("field1", EventTriplet.TYPE_INT,
                           "rng 200 300|rng 0 100");

    Filter f2 = new Filter("field1", EventTriplet.TYPE_INT,
                           "rng 9 14");

    Filter f3 = new Filter("field1", EventTriplet.TYPE_INT,
                           "rng 201 209");

    Filter f4 = new Filter("field1", EventTriplet.TYPE_INT,
                           "rng 202 206");

    Filter f5 = new Filter("field1", EventTriplet.TYPE_INT,
                           "rng 301 400");

    Filter f6 = new Filter("field1", EventTriplet.TYPE_INT,
                           "rng 199 280");

    PlainFilterForest p = new PlainFilterForest();
    int ops = 0;
    try {
      p.add(f3, "koe3");
      ops += p.comparisons;
      p.add(f5, "koe3");
      ops += p.comparisons;
      p.add(f2, "koe3");
      ops += p.comparisons;
      p.add(f6, "koe3");
      ops += p.comparisons;
      p.add(f4, "koe3");
      ops += p.comparisons;
      p.add(f1, "koe1");
      ops += p.comparisons;

      System.out.println("Ops total: " + ops);

      p.remove(f1, "koe1");
    }
    catch (PosetException e) {
      e.printStackTrace();
    }
    System.out.println("Ops: " + p.comparisons);

    try {
      PrintWriter fout = new PrintWriter(new FileWriter(
          "testposet.dat"), false);
      p.printDotFull(fout);
      fout.close();
    }
    catch (Exception e2) {
      e2.printStackTrace();
    }
    System.out.println(p.toString());
  }

  /**
   * Adds a change listener to the data structure
   * @param cl PosetChangeListener
   */
  public void addListener(PosetChangeListener cl) {
  }

  /**
   * Adds a change listener to the data structure
   * @param cl PosetChangeListener
   * @return boolean
   */
  public boolean removeListener(PosetChangeListener cl) {
    return true;
  }

  /** Return cover test count for insert/remove
   * @return int the cover count number
   */
  public int getOps() {
    return comparisons;
  }

  /** Update overheads using getOverhead and by setting setTempOverhead
   *
   */
  public void updateOverhead() {
    // Update root set weights
    for (int i = 0; i < nodes.size(); i++) {
      Node n = (Node) nodes.elementAt(i);
      Vector fs = PlainFilterForest.getCoveredNodes(n); // includes the node
      double cost = 0;
      for (int j = 0; j < fs.size(); j++) {
        Template t = (Template) fs.elementAt(j);
        cost += t.getOverhead();
      }
      ( (Template) n.getData()).setTempOverhead(cost);
    }
  }

  /**
   * Returns all nodes as IFBase
   * @return Vector
   */
  public Vector getAllNodes() {
    Vector ret = new Vector();
    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      Node n = (Node) e.nextElement();
      ret.add(n.getData());
    }
    return ret;
  }

  /**
   * Returns all objects in the forest as templates. Note that the
   * interfaces do not reflect the current situation because they
   * are collapsed!
   * @return Vector
   */
  public Vector getAllAsTemplates() {
    Vector ret = new Vector();
    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      Node n = (Node) e.nextElement();
      if (n.getData() instanceof Template) {
        ITemplate t = (ITemplate) n.getData();
        if (!connectedToRoot(n)) {
          System.out.println("Found a node not connected to root!!");
        }
        ret.add(t);
      }
    }
    return ret;
  }

  /** Returns true if the node is connected to root
   * @param n the Node
   * @return true if connected
   */
  public boolean connectedToRoot(Node n) {
    Node t = n;
    while (t.parent != null) {
      t = t.parent;
    }
    if (!nodes.contains(t)) {
      return false;
    }
    return true;
  }

  /**
   * Debug method
   * @param en Notification
   * @param iface Object
   * @return boolean
   */
  public boolean testMatch(Notification en, Object iface) {
    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      Node n = (Node) e.nextElement();
      IFBase t = (IFBase) n.getData();
      if (t.getTid().contains(iface) && t.match(null, en)) {
        return true;
      }
    }
    return false;
  }

  /**
   * Match a notification using the poset or forest. Assumes that
   * the objects are ITemplates and groups them by their interface name
   *  (getInterface).
   *  @param e the notification to match
   *  @return TripletTable contains the matched objects
   */
  public TripletTable matchT(Object e) {
    if (! (e instanceof Notification)) {
      return null;
    }
    TripletTable ret = new TripletTable();

    // note no optimizations
    Vector stack = new Vector();

    // Create an imaginary node
    Node imaginary = new Node();
    imaginary.children.addAll(nodes);
    stack.add(imaginary);
    boolean quit = false;
    Vector result = new Vector();

    // Traverse the tree
    while (stack.size() > 0 && !quit) {
      Node n = (Node) stack.remove(0);
      if (n != null) {
        Vector ci = null;
        synchronized (n.getChildren()) {
          ci = (Vector) n.getChildren().clone();
        }
        if (ci != null) {
          Iterator k = ci.iterator();
          while (k.hasNext()) {
            Node n2 = (Node) k.next();
            IFBase t = (IFBase) n2.getData();

            if (t.match(null, e)) {
              if (n2.getChildren().size() > 0) {
                stack.add(n2);
              }
              ITemplate _t = (ITemplate) t;

              ret.insert(_t.getInterface(), _t);
            }
          }
        }
      }
    }
    return ret;
  }

  /**
   * Calculates the forwards set given the base set of interfaces
   * @param ifaces Vector
   * @param t IFBase
   * @return Vector
   */
  public Vector calculateForwards(Vector ifaces, IFBase t) {
    Node n0 = (Node) templates.get(t.getCoverId());

    if (n0 == null) {
      return null;
    }

    IFBase f = (IFBase) n0.getData();

    // remove the interfaces from the set
    ifaces.removeAll(f.getTid());

    Node p = n0.getParent();
    while (p != null) {
      f = (IFBase) p.getData();
      // remove the interfaces from the set
      ifaces.removeAll(f.getTid());
      p = p.getParent();
    }
    return ifaces;
  }

  /**
   * Match an object against the data structure
   * @param e Object
   * @return Vector
   */
  public Vector match(Object e) {
    if (! (e instanceof Notification)) {
      return null;
    }

    // note no optimizations
    Vector stack = new Vector();

    // Create an imaginary node
    Node imaginary = new Node();
    imaginary.children.addAll(nodes);
    stack.add(imaginary);
    boolean quit = false;
    Vector result = new Vector();

    // Traverse the tree
    while (stack.size() > 0 && !quit) {
      Node n = (Node) stack.remove(0);
      if (n != null) {
        Vector ci = null;
        synchronized (n.getChildren()) {
          ci = (Vector) n.getChildren().clone();
        }
        if (ci != null) {
          Iterator k = ci.iterator();
          while (k.hasNext()) {
            Node n2 = (Node) k.next();
            IFBase t = (IFBase) n2.getData();

            if (t.match(null, e)) {
              if (n2.getChildren().size() > 0) {
                stack.add(n2);
              }
              result.addAll(t.getTid());
            }
          }
        }
      }
    }

    // compress result
    Vector ret = new Vector();
    for (int k = 0; k < result.size(); k++) {
      Object o = (Object) result.elementAt(k);
      if (!ret.contains(o)) {
        ret.add(o);
      }
    }
    return ret;
  }

  /**
   * Returns the total overhead of the root nodes
   * @return double
   */
  public double getOverhead() {
    double cost = 0;

    // Update root set weights
    for (int i = 0; i < nodes.size(); i++) {
      Node n = (Node) nodes.elementAt(i);
      Template t = (Template) n.getData();
      cost += t.getOverhead();
    }
    return cost;
  }

  /** Sort trees
   * @param sort true if sort is to be used
   */
  public void setSort(boolean sort) {
    // not used
  }

  public PlainFilterForest() {
  }

  /** Sorts the trees by size: smallest size subtree first in order
   * to optimize insertions
   */
  public void sortTreesBySize() {
    int depth[] = new int[nodes.size()];

    // Get the subtreesizes for all nodes (the slow way)
    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      Node n = (Node) e.nextElement();
      n.setSubTreeSize(getNodeCount(n));
    }

    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      Node n = (Node) e.nextElement();
      Vector v = n.getChildren();
      Collections.sort(v);
    }

    // Finally sort root node
    Collections.sort(nodes);
  }

  /**
   * Returns the top-level objects in the poset (the non-covered set)
   * @return Vector the non-covered objects
   */
  public Vector get() {
    Vector ret = new Vector();
    for (int i = 0; i < nodes.size(); i++) {
      Node n = (Node) nodes.elementAt(i);
      if (n != null) {
        ret.add(n.data);
      }
    }
    return ret;
  }

  /**
   * Returns the top-level objects in the poset (the non-covered set)
   * @return Vector the non-covered objects (Nodes)
   */
  public Vector getNodes() {
    return nodes;
  }

  /** Reset the data structure */
  public void reset() {
    resetNodes();
    nodes.clear();
    templates.clear();
  }

  /**
   * Returns true if the object is removable
   * @param t IFBase
   * @param color Object
   * @return boolean
   */
  public boolean removable(IFBase t, Object color) {
    Node n = (Node) templates.get(t.getCoverId());

    if (n == null) {
      System.out.println("PlainFilterForest: node not found");
      return false;
    }

    IFBase f = (IFBase) n.getData();
    if (f == null) {
      System.out.println("PlainFilterForest: object was null");
      return false;
    }
    if (f.getTid().contains(color)) {
      return true;
    }
    return false;
  }

  /** Removes the given object from the poset.
   * @param t the object to add, must be a cloned copy of the original object
   * @param color the color
   * @throws PosetException
   * @return 0 if remove failed, 1 if remove was the root node, greater than 1 otherwise,
   * returns -1 if remove was succesful but there were multiple copies
   */
  public synchronized int remove(IFBase t, Object color) throws PosetException {
    comparisons = 0;

    if (test) {
      System.out.println("REMOVE CALLED for: " + color);
    }

    // first find the object
    Node n = (Node) templates.get(t.getCoverId());

    if (n == null) {
      if (debug) {
        System.out.println("Fatal: Object not found:" + t.getCoverId());
      }
      return POSET_ERROR;
    }
    return remove(n, color);
  }

  /**
   * Remove a node
   * @param n Node
   * @param color Object
   * @throws PosetException
   * @return int
   */
  protected synchronized int remove(Node n, Object color) throws PosetException {
    IFBase f = (IFBase) n.getData();

    if (f.getTid() == null) {
      System.out.println("PlainFilterForest: remove: getid() == null");
      return POSET_ERROR;
    }

    boolean rem = f.getTid().remove(color);

    if (!rem && f.getTid().size() > 0) { // should be only one
      System.out.println("Note: Color not found: " + color + " tid: " +
                         f.getTid());
      return POSET_ERROR;
    }
    else if (!rem && f.getTid().size() == 0) {
      System.out.println("Note: getTid() was 0 in remove but remove failed ");
      System.out.println("Probably duplicate insertions/deletions. " +
                         "Garbage collecting");
    }

    if (f.getTid().size() > 0) {
      removeNodeEvent(n, color); // callback for extensions

      if (test) {
        System.out.println("REMOVE ENDED (DUPLICATE)");
      }

      return POSET_DUPLICATE; // remove was succesful but a copy still exists
    }

    removeNodeEvent(n, n.getData().getTid()); // callback for extensions

    // permanent deletion
    templates.remove(f.getCoverId());

    if (test) {
      System.out.println("Remove: deleted node for " + color);
    }

    if (nodes.contains(n)) { // root node

      if (n.parent != null) {
        System.out.println("Fatal: in root set but not root");

      }
      nodes.remove(n); // the list remains sorted

      if (test) {
        System.out.println("top level node");
      }

      Vector ch = (Vector) n.children.clone();
      for (int i = 0; i < ch.size(); i++) {
        Node n2 = (Node) ch.elementAt(i);
        add(n2); // add to the poset, resets counter
      }
      n.children.clear();
      ch = null;
      n = null;

      return POSET_ROOT;
    }
    else { // some other node
      Node parent = n.parent;

      if (parent == null) {
        System.out.println("Parent was null but not a top-level node");
      }
      else {
        if (!parent.children.remove(n)) { // remove this branch
          System.out.println(
              "Warning: failed to remove from parent list. Contains:"
              + parent.children.contains(n));
        }
      }
      if (test) {
        System.out.println("not top level node");

        // add the siblings to the poset again
      }
      Vector ch = (Vector) n.children.clone();
      for (int i = 0; i < ch.size(); i++) {
        Node n2 = (Node) ch.elementAt(i);
        if (test) {
          System.out.println("Calling add: " + i + "/" + ch.size());
        }
        add(parent, n2); // add to the poset
      }

      n.children.clear();
      ch = null;
      n = null;

      if (test) {
        System.out.println("After remove 2");
      }
      return POSET_NOTROOT;
    }
  }

  /**
   * Get filter objects for the given Node.
   * @param pos Node the node
   * @return Vector filters
   */
  public static Vector getCoveredNodes(Node pos) {
    Vector v = new Vector();

    v.add(pos.getData());

    if (pos == null) {
      System.out.println("FilterPoset: getCoveringNode, node not found");
      return v;
    }

    if (pos.children == null || pos.children.size() == 0) {
      return v;
    }

    Vector stack = new Vector();
    stack.add(pos);

    while (stack.size() > 0) {
      Node n = (Node) stack.remove(0);
      if (n != null) {
        if (n.children != null) {
          Iterator i = n.children.iterator();
          while (i.hasNext()) {
            Node n2 = (Node) i.next();
            v.add(n2.data);
            stack.add(n2);
          }
        }
      }
    }
    return v;
  }

  /** Returns the set of covered nodes (below this node in the tree)
   *  Used to enhance matching. Does not include the given object.
   *  Note that this does not return ALL covered nodes!
   * @param t the node
   * @param v Vector that will contain the covered nodes
   */
  public void getCoveredNodes(IFBase t, Vector v) {
    Node pos = (Node) templates.get(t.getCoverId());
    if (pos == null) {
      if (debug) {
        System.out.println("FilterPoset: getCoveringNode, node not found");
      }
      return;
    }

    if (pos.children == null || pos.children.size() == 0) {
      return;
    }

    Vector stack = new Vector();
    stack.add(pos);

    while (stack.size() > 0) {
      Node n = (Node) stack.remove(0);
      if (n != null) {
        if (n.children != null) {
          Iterator i = n.children.iterator();
          while (i.hasNext()) {
            Node n2 = (Node) i.next();
            v.add(n2.data);
            stack.add(n2);
          }
        }
      }
    }
  }

  /**
   * Returns true if covered nodes exist
   * @param t IFBase
   * @return boolean
   */
  public boolean coveredNodesExist(IFBase t) {
    Node pos = (Node) templates.get(t.getCoverId());
    if (pos == null) {
      if (debug) {
        System.out.println("FilterPoset: getCoveringNode, node not found");
      }
      return false;
    }

    if (pos.children.size() > 0) {
      return true;
    }

    return false;
  }

  /** Returns the set of covering nodes (above this node in the tree)
   *  Used to enhance matching. Does not include the given object.
   * @param t the node
   * @param v of covering ICovering objects (cast to the proper class)
   */
  public synchronized void getCoveringNodes(IFBase t, Vector v) {
    Node n = (Node) templates.get(t.getCoverId());
    if (n == null) {
      if (debug) {
        System.out.println("FilterPoset: getCoveringNode, node not found");
      }
      return;
    }

    n = n.parent;
    while (n != null) {
      v.add(n.data);
      n = n.parent;
    }

    return;
  }

  /**
   * Lookup a node
   * @param t IFBase
   * @return IFBase
   */
  public IFBase lookup(IFBase t) {
    Node n0 = (Node) templates.get(t.getCoverId());

    // Check duplicate entries
    if (n0 != null) {
      return n0.data;
    }
    return null;
  }

  /**
   * Lookup a Node object
   * @param t IFBase
   * @return Node
   */
  public Node lookupNode(IFBase t) {
    Node n0 = (Node) templates.get(t.getCoverId());
    return n0;
  }

  /**
   * Get the children sorted based on the color property.
   * Used for balancing. Clones the vector.
   * @param x Node
   * @param color Object
   * @return Vector
   */
  synchronized Vector getLevel(Node x, Object color) {
    Vector v = null;

    if (x == null) {
      v = (Vector) nodes.clone();
    }
    else {
      v = (Vector) x.getChildren().clone();
    }
    return v;
  }

  /* EXTENSION MECHANISMS **********************************************/

  /**
   * Called when a color is eliminated. Reserved for extension mechanisms.
   * @param x Node
   * @param tids Vector
   * @param start Node
   */
  public void eliminateColorCallback(Node x, Vector tids, Node start) {
    // NOP
  }

  /**
   * Called when a color is eliminated. Reserved for extension mechanisms.
   * @param x Node
   * @param color Object
   * @param start Node
   */
  public void eliminateColorCallback(Node x, Object color, Node start) {
    // NOP
  }

  /**
   * Called when a node is added. Reserved for extension mechanisms.
   * @param n Node
   * @param color Object
   */
  public void addNodeEvent(Node n, Object color) {
    // NOP
  }

  /**
   * Called when a node is added. Reserved for extension mechanisms.
   * @param n Node
   * @param colors Vector
   * @throws PosetException
   */
  public void addNodeEvent(Node n, Vector colors) throws PosetException {
    // NOP
  }

  /**
   * Called when a node is removed. Reserved for extension mechanisms.
   * @param n Node
   * @param colors Vector
   */
  public void removeNodeEvent(Node n, Vector colors) {
    // NOP
  }

  /**
   * Called when a node is removed. Reserved for extension mechanisms.
   * @param n Node
   * @param color Object
   */
  public void removeNodeEvent(Node n, Object color) {
    // NOP
  }

  /* EXTENSION MECHANISMS **********************************************/

  /** Checks if the filter is already contained in the structure
   * @param t the object to add
   * @param color the color/interface to add
   * @throws PosetException
   * @return code
   */
  public int contains(IFBase t, Object color) throws PosetException {
    Node n0 = (Node) templates.get(t.getCoverId());

    // Check duplicate entries
    if (n0 != null) {
      IFBase f = (IFBase) n0.getData();

      if (f.getTid() == null) {
        return POSET_ERROR;
      }

      if (!f.getTid().contains(color)) { // if iface/color is new we perform
        // Check that there does not exist a parent with the same color
        Node p = n0.getParent();
        while (p != null) {
          if (p.getData().getTid().contains(color)) {
            return POSET_CONTAINED; // a covering node with the same color was found
          }
          p = p.getParent();
        }
        return POSET_NOT_CONTAINED;
      }
      else {
        return POSET_CONTAINED;
      }
    }
    return PlainFilterForest.POSET_NOTDUPLICATE;
  }

  /** Adds the given object to the poset.
   * @param t the object to add. Note object is not cloned. Only one TID
   * @param iface the interface/color
   * @throws PosetException
   * @return int status code
   */
  public synchronized int add(IFBase t, Object iface) throws PosetException {
    boolean flag = false;
    int ret = POSET_ERROR;

    if (test) {
      System.out.println("ADD CALLED");
    }

    if (iface == null) {
      System.out.println("Forest: iface was: " + iface);
      throw new PosetException();
    }

    comparisons = 0; // reset counter

    Node n0 = (Node) templates.get(t.getCoverId());

    // Check duplicate entries
    if (n0 != null) {
      IFBase f = (IFBase) n0.getData();

      if (f.getTid() == null) {
        System.out.println("Fatal: tid was null for existing object");
        return POSET_ERROR;
      }

      //  if (!f.getTid().contains(iface)) { // if iface/color is new we perform
      // we can add the iface
      f.addTid(iface);

      addNodeEvent(n0, iface); // callback for extensions
      //   }
      /*
         else {
       System.out.println("Warning: tried to insert identical filter with " +
                                 "the same color");
              System.out.println("New Object: " + t.hashCode() + " Old: " +
                                 f.hashCode());
              System.out.println("Color: " + iface);
            }
       }*/
      return POSET_DUPLICATE;
    }

    // clearly this is the first time the object is added
    // so we clear the tid vector
    if (t.getTid() != null) {
      t.getTid().clear();
    }
    // and add the first interface/color
    t.addTid(iface);
    Vector l2 = getLevel(null, iface);

    for (int i = 0; i < l2.size(); i++) {
      Node n = (Node) l2.elementAt(i);

      if (n.data == null) {
        continue;
      }

      // We count the number of comparisons, they are conditional
      boolean cover1 = t.covers(n.data);

      boolean cover2 = false;

      comparisons++;

      if (!cover1) {
        comparisons++;
        cover2 = n.data.covers(t);
      }

      // The new subscription covers a top level subscription
      if (cover1) { // add new node as a root
        ret = POSET_ROOT;
        nodes.remove(n); // remove from the actual vector and not clone
        Node n3 = new Node();
        n3.count = counter++;
        n3.parent = null;
        n.parent = n3;
        n3.data = t;
        // moves all covered top level nodes under the new node
        moveSiblings(nodes, n3.children, n3.data, n3);
        n3.children.add(n);
        nodes.add(n3);
        templates.put(t.getCoverId(), n3);

        if (debug) {
          System.out.println("new root: " + t.toString());
        }

        addNodeEvent(n3, iface); // callback for extensions

        flag = true;
        break;
      }
      else if (cover2) {
        ret = POSET_NOTROOT;

        Vector ifaces = new Vector();
        Node n2 = findSubTree(n, null, t, ifaces, iface);

        if (debug) {
          System.out.println("Inserting " + t.toString() + " Under " +
                             n2.getData().toString());
          // add n3 under n2 and move any covered nodes under n3
        }
        Node n3 = new Node();
        n3.parent = n2;
        n3.count = counter++;
        n3.data = t;
        moveSiblings(n2.children, n3.children, n3.data, n3);

        n2.children.add(n3);
        templates.put(t.getCoverId(), n3);

        addNodeEvent(n3, iface); // callback for extensions

        flag = true;
        break;
      }
    }

    if (!flag) {
      if (debug) {
        System.out.println("Inserting " + t.toString());

      }
      ret = POSET_ROOT;
      // add new root
      Node n = new Node();
      n.count = counter++;
      n.parent = null;
      n.data = t;
      nodes.add(n);
      templates.put(t.getCoverId(), n);
      addNodeEvent(n, iface); // callback for extensions
    }
    return ret;
  }

  /** Adds the given object to the poset. Does not reset ops counter.
   * Called when a root node was removed. A special case for colors!
   * The color index is updated since the node came from outside and
   * it is not a local change.
   * @param node the Node object to add
   * @throws PosetException
   */
  protected synchronized void add(Node node) throws PosetException {
    boolean flag = false;

    if (test) {
      System.out.println("ADD ROOT CALLED for " + node.getData().getTid());
    }

    for (int i = 0; i < nodes.size(); i++) {
      Node n = (Node) nodes.elementAt(i);

      if (n.data == null) {
        continue;
      }

      boolean cover = n.data.covers(node.data);
      comparisons++;

      if (cover) {
        Node n2 = findSubTree(n, null, node.data, null, null);

        // add node under n2 and move any covered nodes under node
        node.parent = n2;
        moveSiblings(n2.children, node.children, node.data, node);
        n2.children.add(node);

        flag = true;
        break;
      }
    }

    // siblings did not cover, add to root level
    if (!flag) {
      if (test) {
        System.out.println("Adding as new node");
        // add new poset
      }
      node.parent = null;
      nodes.add(node);
      addNodeEvent(node, node.getData().getTid()); // callback for extensions
    }

    if (test) {
      System.out.println("End add root");
    }
  }

  /**
   * Adds the given Vector to the poset.
   * @param v the objects to add
   * @throws PosetException
   * @return int the number of elements inserted starting from the beginning
   */
  public synchronized int add(Vector v) throws PosetException {
    if (v == null || v.size() == 0 || ! (v.firstElement() instanceof ITemplate)) {
      return 0;
    }
    int lastIndex = 0;
    for (int i = 0; i < v.size(); i++) {
      IFBase ic = (IFBase) v.elementAt(i);
      if (add(ic, ( (ITemplate) ic).getInterface()) == POSET_ERROR) {
        break;
      }
      lastIndex = i + 1;
    }
    return lastIndex;
  }

  /** Adds the given object to the poset. Does not reset ops counter.
   * The node is relocated within the tree and the color index has
   * already been updated!
   * @param parent the parent Node
   * @param node the Node object to add
   * @throws PosetException
   */
  protected synchronized void add(Node parent, Node node) throws PosetException {
    boolean flag = false;

    if (test) {
      System.out.println("ADD CALLED for " + node.getData().getTid());
    }

    for (int i = 0; i < parent.children.size(); i++) {
      Node n = (Node) parent.children.elementAt(i);

      if (n.data == null) {
        continue;
      }

      comparisons++;
      boolean cover1 = n.data.covers(node.data);

      if (cover1) {
        Node n2 = findSubTree(n, null, node.data, null, null);

        // add node under n2 and move any covered nodes under node
        node.parent = n2;
        moveSiblings(n2.children, node.children, node.data, node);
        n2.children.add(node);

        flag = true;
        break;
      }
    }

    if (!flag) {
      if (test) {
        System.out.println("Adding as new node");
      }
      node.parent = parent;
      parent.children.add(node);
      addNodeEvent(node, node.getData().getTid()); // callback for extensions
    }

    if (test) {
      System.out.println("ADD ENDED CALLED");
    }
  }

  /**
   * Move siblings from one node to another.
   * @param v1 The node whose children are moved
   * @param v2 The target node
   * @param t2 The object that must cover all moved, if null all are moved
   * @param parent The parent node
   */
  void moveSiblings(Vector v1, Vector v2, IFBase t2, Node parent) {
    Vector removeList = new Vector();
    for (int i = 0; i < v1.size(); i++) {
      Node n = (Node) v1.elementAt(i);
      if (n != null && n != parent) {
        boolean cover1 = false;
        if (t2 == null) {
          cover1 = true;
        }
        else {
          comparisons++; // increment comparison counter
          cover1 = t2.covers(n.data);
        }
        if (cover1) {
          v2.add(n);
          n.parent = parent;
          removeList.add(n);
        }
      }
    }
    v1.removeAll(removeList);
  }

  /**
   * Returns the node count
   * @return int
   */
  public int getNodeCount() {
    int removeCount = 0;
    Vector stack = new Vector();
    for (int i = 0; i < nodes.size(); i++) {
      Node n = (Node) nodes.elementAt(i);
      stack.add(n);
    }
    while (stack.size() > 0) {
      Node n = (Node) stack.elementAt(0);
      if (n != null) {
        if (n.children != null) {
          for (int i = 0; i < n.children.size(); i++) {
            Node n2 = (Node) n.children.elementAt(i);
            stack.add(n2);
          }
        }
      }
      stack.remove(0);
      removeCount++;
    }
    return removeCount;
  }

  private int getNodeCount(Node pos) {
    int removeCount = 0;
    Vector stack = new Vector();
    stack.add(pos);
    while (stack.size() > 0) {
      Node n = (Node) stack.elementAt(0);
      if (n != null) {
        if (n.children != null) {
          for (int i = 0; i < n.children.size(); i++) {
            Node n2 = (Node) n.children.elementAt(i);
            stack.add(n2);
          }
        }
      }
      stack.remove(0);
      removeCount++;
    }
    return removeCount;
  }

  private int resetNodes() {
    int removeCount = 0;
    Vector stack = new Vector();
    for (int i = 0; i < nodes.size(); i++) {
      Node n = (Node) nodes.elementAt(i);
      stack.add(n);
    }
    while (stack.size() > 0) {
      Node n = (Node) stack.elementAt(0);
      if (n != null) {
        if (n.children != null) {
          for (int i = 0; i < n.children.size(); i++) {
            Node n2 = (Node) n.children.elementAt(i);
            stack.add(n2);
          }
        }
      }
      n = (Node) stack.remove(0);
      n.children.clear();
      n.children = null;
      n = null;
      removeCount++;
    }
    return removeCount;
  }

  protected boolean containsColor(IFBase x, Object y) {
    if (x == null || y == null) {
      return false;
    }
    if (x.getTid() != null && x.getTid().contains(y)) {
      return true;
    }
    return false;
  }

  /**
   * Finds the last covering subtree in a tree
   * @param n Node
   * @param parent Node
   * @param t IFBase
   * @param ifaces Vector
   * @param color Object
   * @return Node
   */
  protected Node findSubTree(Node n, Node parent, IFBase t, Vector ifaces,
                             Object color) {
    if (n.children == null || n.children.size() == 0) {
      return n;
    }
    try {
      Vector nv = n.getChildren();
      for (int i = 0; i < nv.size(); i++) {
        Node n2 = (Node) nv.elementAt(i);
        if (n2 != null) {
          comparisons++; // increment counter
          boolean cover1 = n2.data.covers(t);
          if (cover1) {
            return findSubTree(n2, n, t, ifaces, color);
          }
        }
      }
    }
    catch (Exception e) {
      e.printStackTrace();
      return null;
    }
    return n;
  }

  /**
   * Get the data structure as HTML
   * @return String
   */
  public String getHTMLOutput() {
    StringBuffer buf = new StringBuffer();
    buf.append("<html><body>");
    for (int i = 0; i < nodes.size(); i++) {
      Node n = (Node) nodes.elementAt(i);
      if (n != null) {
        printHtml(n, buf);
      }
    }
    buf.append("</body></html>");
    return buf.toString();
  }

  /**
   * Print roots
   * @param fn PrintWriter
   */
  public void printRoots(PrintWriter fn) {
    for (int i = 0; i < this.nodes.size(); i++) {
      Node t = (Node) nodes.elementAt(i);
      fn.println( ( (IFBase) t.getData()).toString());
    }
  }

  /**
   *  Print dot format output
   * @param fn PrintWriter
   */
  public void printDot(PrintWriter fn) {
    fn.println("digraph poset {");
    fn.println("size=\"6,6\";");
    fn.println("node [color=lightblue2, style=filled]");

    for (int i = 0; i < nodes.size(); i++) {
      Node n = (Node) nodes.elementAt(i);
      if (n != null) {
        printDot(n, fn);
      }
    }
    fn.println("}");
  }

  private void printDot(Node n, PrintWriter fn) {
    if (n.children == null || n.children.size() == 0) { // orphan node
      fn.println("\"" + n.count + "\";");
      return;
    }
    for (int i = 0; i < n.children.size(); i++) {
      Node n2 = (Node) n.children.elementAt(i);
      fn.println("\"" + n.count + "\"->\"" + n2.count +
                 "\";");
      printDot(n2, fn);
    }
  }

  public void printDotFull(PrintWriter fn) {
    fn.println("digraph poset {");
    fn.println("size=\"6,6\";");
    fn.println("node [color=lightblue2, style=filled]");

    for (int i = 0; i < nodes.size(); i++) {
      Node n = (Node) nodes.elementAt(i);
      printDotFull(n, fn);
    }
    fn.println("}");
  }

  void printDotFull(Node n, PrintWriter fn) {
    Filter f = null;
    String s1 = "";
    if (n.data instanceof ITemplate) {
      ITemplate t0 = (ITemplate) n.data;
      f = (Filter) t0.getFilters().firstElement();
      s1 = t0.toString();
    }
    else if (n.data instanceof Filter) {
      f = (Filter) n.data;
      s1 = f.toString();
    }
    if (n.children == null || n.children.size() == 0) { // orphan node
      fn.println("\"" + f.getFilter() + "\";");
      return;
    }
    try {
      for (int i = 0; i < n.children.size(); i++) {
        Node n2 = (Node) n.children.elementAt(i);
        Filter f2 = null;
        String s2 = "";
        if (n.data instanceof ITemplate) {
          ITemplate t = (Template) n2.data;
          f2 = (Filter) t.getFilters().firstElement();
          s2 = t.toString();
        }
        else if (n.data instanceof Filter) {
          f2 = (Filter) n2.data;
          s2 = f2.getFilter();
        }

        fn.println("\"" + s1 + "\"->\"" + s2 +
                   "\";");
        printDotFull(n2, fn);
      }
    }
    catch (Exception e) {
      e.printStackTrace();
    }
  }

  void printHtml(Node n, StringBuffer buf) {
    Filter f = null;
    String s1 = "";
    if (! (n.data instanceof ITemplate)) {
      return;
    }
    ITemplate t = (ITemplate) n.data;
    s1 = ManagementUI.printFilterTable(t);

    if (n.children == null || n.children.size() == 0) { // orphan node

      buf.append("<table bgcolor=\"#AAAAAF\"><tr><td>");
      buf.append(s1 + "<br>");
      buf.append("</td></tr></table>");
      return;
    }
    try {
      for (int i = 0; i < n.children.size(); i++) {
        Node n2 = (Node) n.children.elementAt(i);
        if (n.data instanceof ITemplate) {
          ITemplate t2 = (Template) n2.data;

          String s2 = ManagementUI.printFilterTable(t2);
          buf.append("<table><tr><td>");
          buf.append(s1 + "</td><td>");
          buf.append("</td><td>--></td><td>");
          buf.append(s2);
          buf.append("</td></tr></table>");
          printHtml(n2, buf);
        }
      }
    }
    catch (Exception e) {
      e.printStackTrace();
    }
  }
}
