package fuegocore.notify.core.filter.poset;

import java.io.*;
import java.util.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.filter.poset.forwards.*;
import fuegocore.notify.util.*;

/**
 * Implements a directed acyclic graph filters poset
 * Based on the Siena Java code
 */
public class Poset
    implements IFilterPoset {
  private Set roots;

  /** Predecessor list */
  private Hashtable pre = new Hashtable();

  /** Successor list */
  private Hashtable anc = new Hashtable();

  boolean debug = false;
  boolean test = false;

  int ops = 0;
  public Hashtable templates = new Hashtable(); // fast access

  /** Error code for duplicate entry */
  public static int POSET_DUPLICATE = -1;

  /** Error code for general error */
  public static int POSET_ERROR = 0;

  /** Return code for entry into the root set */
  public static int POSET_ROOT = 1;

  /** Return code for entry into non-root set */
  public static int POSET_NOTROOT = 2;

  /** Return code for contained color ad filter */
  public static int POSET_CONTAINED = -1;

  public static int POSET_NOT_CONTAINED = -1;

  // Vector of listeners
  private Vector listeners = new Vector();
  public static String DELETE_EVENT_ELIMINATE = "delEventEliminate";

  public static String DELETE_EVENT = "delEvent";

  public static String DELETE_EVENT_BEFORE = "delEventBefore";

  public static String DELETE_EVENT_AFTER = "delEventAfter";

  public static String ADD_EVENT = "addEvent";

  public static String ADD_EVENT_BEFORE = "addEventBefore";

  public static String ADD_EVENT_AFTER = "addEventAfter";

  public static String RESET = "reset";

  /** Do not perform eliminate color */
  public boolean NO_ELIMINATE = false;

  /* ********************************************************************* */
  /**
   * Adds a change listener to the data structure
   * @param cl PosetChangeListener
   */
  public void addListener(PosetChangeListener cl) {
    if (!listeners.contains(cl)) {
      listeners.add(cl);
    }
  }

  /**
   * Adds a change listener to the data structure
   * @param cl PosetChangeListener
   * @return boolean
   */
  public boolean removeListener(PosetChangeListener cl) {
    return listeners.remove(cl);
  }

  /**
   * Increment ops counter
   */
  public void incOps() {
    ops++;
  }

  public Hashtable getTemplates() {
    return templates;
  }

  /**
   * Reset ops
   */
  public void resetOps() {
    ops = 0;
  }

  /**
   *  Add both profiles and queries in random order. Used to test structure.
   * @param pro Vector
   * @param que Vector
   * @param biglist Vector
   * @return int
   * @throws PosetException
   */
  public synchronized int addRandomly(Vector pro, Vector que, Vector biglist) throws
      PosetException {
    int cost = 0;

    for (int i = 0; i < biglist.size(); i++) {
      IFBase f = (IFBase) biglist.elementAt(i);
      if (pro.contains(f)) {
        if (add(f, ( (ITemplate) f).getInterface()) ==
            FilterForest.POSET_ERROR) {
          // break;
        }
        cost += ops;
      }
      else {
        if (add(f, ( (ITemplate) f).getInterface()) ==
            FilterForest.POSET_ERROR) {
          // break;
        }
        cost += ops;
      }
    }
    return cost;
  }

  /**
   * Lookup profiles and queries in random order. Used to test structure.
   * @param pro Vector
   * @param que Vector
   * @param biglist Vector
   * @return int the cost
   * @throws PosetException
   */
  public synchronized int lookupRandomly(Vector pro, Vector que, Vector biglist) throws
      PosetException {
    for (int i = 0; i < biglist.size(); i++) {
      IFBase f = (IFBase) biglist.elementAt(i);
      if (pro.contains(f)) {
        HashSet v2 = new HashSet();
        addPredecessors(v2, f);
      }
      else {
        HashSet v2 = new HashSet();
        addSuccessors(v2, f);
      }
    }
    return 0;
  }

  /**
   * Delete profiles and queries in random order. Used to test structure.
   * @param pro Vector
   * @param que Vector
   * @param biglist Vector
   * @return int
   * @throws PosetException
   */
  public synchronized int deleteRandomly(Vector pro, Vector que, Vector biglist) throws
      PosetException {
    int cost = 0;
    for (int i = 0; i < biglist.size(); i++) {
      this.ops = 0;
      IFBase f = (IFBase) biglist.elementAt(i);
      if (pro.contains(f)) {
        if (this.remove(f, ( (ITemplate) f).getInterface()) ==
            FilterForest.POSET_ERROR) {
          // break;
        }
      }
      else {
        if (this.remove(f, ( (ITemplate) f).getInterface()) ==
            FilterForest.POSET_ERROR) {
          // break;
        }
      }
      cost += ops;
    }
    return cost;
  }

  /**
   * Fires an internal event to the listeners
   * @param event String
   * @param x IFBase
   */
  public void fireEventToListeners(String event, IFBase x) {
    for (int i = 0; i < listeners.size(); i++) {
      PosetChangeListener pl = (PosetChangeListener) listeners.elementAt(i);
      pl.eventOccurred(event, x);
    }
  }

  /** Sorts the trees by size
   */
  public void sortTreesBySize() {
    return;
  }

  /**
   * Returns all objects in the forest as templates
   * @return Vector
   */
  public Vector getAllAsTemplates() {
    Vector ret = new Vector();
    ret.addAll(templates.values());
    return ret;
  }

  public void updateOverhead() {}

  /** Return cover test count for insert/remove
   * @return int the cover count number
   */
  public int getOps() {
    return ops;
  }

  public Vector getPre(String s) {
    IFBase f0 = (IFBase) templates.get(s);
    Vector v = (Vector) pre.get(f0);
    return v;
  }

  public Vector getSuccessor(String s) {
    IFBase f0 = (IFBase) templates.get(s);
    Vector v = (Vector) anc.get(f0);
    return v;
  }

  public Vector getPre(IFBase s) {
    Vector v = (Vector) pre.get(s);
    if (v == null) {
      v = new Vector();
      pre.put(s, v);
    }
    return v;
  }

  /**
   * Returns the mapping size
   * @return int
   */
  public int getPredecessorSize() {
    int size = 0;

    Vector v = getAllAsTemplates();
    for (int i = 0; i < v.size(); i++) {
      int level = 0;
      IFBase t = (IFBase) v.elementAt(i);
      Vector pred = this.getPre(t);
      if (pred != null) {
        level += pred.size();
      }
      size += level;
    }
    return size;
  }

  /**
   * Returns the mapping size
   * @return int
   */
  public int getSize() {
    int size = 0;

    Vector v = getAllAsTemplates();
    for (int i = 0; i < v.size(); i++) {
      int level = 0;
      IFBase t = (IFBase) v.elementAt(i);
      Vector pred = this.getPre(t);
      Vector succ = this.getSuccessor(t);
      if (pred != null) {
        level += pred.size();
      }
      if (succ != null) {
        level += succ.size();
      }
      size += level;
    }
    return size;
  }

  public Vector getSuccessor(IFBase s) {
    Vector v = (Vector) anc.get(s);
    if (v == null) {
      v = new Vector();
      anc.put(s, v);
    }
    return v;
  }

  /**
   * Returns the top-level objects in the poset (the non-covered set)
   * @return Vector the non-covered objects
   */
  public Vector get() {
    Vector root = new Vector(roots);
    return root;
  }

  /**
   * Returns the top-level objects in the poset (the non-covered set)
   * @return Vector the non-covered objects (Nodes)
   */
  public Vector getNodes() {
    return null;
  }

  /** Reset the data structure */
  public void reset() {
    clear();
  }

  /** Returns the set of covered nodes (below this node in the tree)
   *  Used to enhance matching. Does not include the given object.
   * @param t the node
   * @param v Vector that will contain the covered nodes
   */
  public void getCoveredNodes(IFBase t, Vector v) {
    Vector vz = getSuccessor(t);
    if (vz == null) {
      return;
    }
    v.addAll(vz);
  }

  public boolean coveredNodesExist(IFBase t) {
    Vector vz = getSuccessor(t);
    if (vz == null) {
      return false;
    }
    return true;
  }

  /** Returns the set of covering nodes (above this node in the tree)
   *  Used to enhance matching. Does not include the given object.
   * @param t the node
   * @param v Vector of covering ICovering objects (cast to the proper class)
   */
  public void getCoveringNodes(IFBase t, Vector v) {
    Vector vz = getPre(t);
    if (vz == null) {
      return;
    }
    v.addAll(vz);
  }

  /**
   * Returns node count
   * @return int
   */
  public int getNodeCount() {
    return count();
  }

  public Poset() {
    roots = new HashSet();
  }

  public void clear() {
    roots.clear();
    templates.clear();
    pre.clear();
    anc.clear();
    // Notify listeners about reset
    fireEventToListeners(RESET, null);
  }

  /**
   * Checks if the given object is a root
   * @param s IFBase
   * @return boolean
   */
  public boolean isRootNode(IFBase s) {
    return getPre(s).isEmpty();
  }

  private boolean empty() {
    return roots.isEmpty();
  }

  public Iterator getRootsIterator() {
    return roots.iterator();
  }

  /**
   * Insert an object between direct ancestors and predecessors
   * @param nfilt IFBase
   * @param pre Collection
   * @param post Collection
   */
  public void insert(IFBase nfilt,
                     Collection pre, Collection post) {
    IFBase x, y;
    Iterator i, j;

    // test pre and post collections
    if (test) {
      Vector c1 = new Vector(pre);
      c1.removeAll(post);
      Vector c3 = new Vector(post);
      c3.removeAll(pre);

      if (c1.size() != pre.size()) {
        Vector c2 = new Vector(pre);
        c2.removeAll(c1);
        Vector c4 = new Vector(post);
        c4.removeAll(c3);

        IFBase pro = (IFBase) c4.firstElement();
        for (int k = 0; k < c2.size(); k++) {
          IFBase fb = (IFBase) c2.elementAt(k);
          if (fb.covers(nfilt)) {
            System.out.println("Problem element covers f");
            System.out.println("Post element" + pro);
            System.out.println("Pre element" + fb);
          }
          else {
            System.out.println("Element did not cover f");
          }
          if (nfilt.covers(fb)) {
            System.out.println("Problem element is covered by f");
          }
          else {
            System.out.println("Element was not covered by f");
          }
          if (pro == fb) {
            System.out.println("Problem element is equal to f");
          }
          else {
            System.out.println("Element was not identical to f");
          }
        }

        System.out.println("1 Intersection of pre and post not zero");
        System.out.println("Original: " + pre.size() + " new " + c1.size() +
                           " intersection:" + c2.size());
        System.exit(0);
      }
      System.out.println("Insert Integrity check 1");
      testAncPreSet();
    }

    if (pre.isEmpty()) {
      roots.add(nfilt);
      roots.removeAll(post);
      if (debug) {
        System.out.println("New root");
      }
    }
    else {
      i = pre.iterator();
      while (i.hasNext()) {
        x = (IFBase) i.next();
        j = post.iterator();
        while (j.hasNext()) {
          disconnectNodes(x, (IFBase) j.next());
        }
        if (!connectNodes(x, nfilt)) {
          if (test) {
            System.out.println("insert: pre: " + pre);
            System.exit(0);
          }
        }
      }
    }

    // ----------------------------------------------------------
    // iterate over the post set
    j = post.iterator();

    // connect each node in the post set with the filter
    while (j.hasNext()) {
      y = (IFBase) j.next();
      if (!connectNodes(nfilt, y)) {
        if (test) {
          System.out.println("insert post: " + post);
          System.exit(0);
        }
      }
    }

    // ----------------------------------------------------------
    if (test) {
      System.out.println("Insert Integrity check 2");
      testAncPreSet();
    }
  }

  private boolean disconnectNodes(IFBase x, IFBase y) {

    getSuccessor(x).remove(y);
    getPre(y).remove(x);

    return true;
  }

  private boolean testIntegrity(IFBase x) {
    Vector v = getPre(x);
    Vector v2 = getSuccessor(x);

    for (int i = 0; i < v.size(); i++) {
      IFBase x2 = (IFBase) v.elementAt(i);
      if (v2.contains(x2)) {
        System.out.println("Integrity failure");
        System.exit(0);
      }
    }
    return true;
  }

  private boolean connectNodes(IFBase x, IFBase y) {

    if (test) {
      testIntegrity(y);
      testIntegrity(x);
    }

    if (x == y) {
      System.out.println("ConnectNodes error: x == y");
      return false;
    }

    if (getSuccessor(x).add(y)) {
      getPre(y).add(x);
    }

    // remove y if it is no longer a root
    if (roots.contains(y)) {
      roots.remove(y);
    }
    return true;
  }

  /**
   * Tests that the successors set is correct and only contains the direct
   * successors of a node
   * @return boolean
   */
  public boolean testDirectSuccessors() {
    // loop over all nodes
    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      IFBase n = (IFBase) e.nextElement();

      // Test that the set of direct successors is correct
      Vector directSuccessors = getSuccessor(n);
      HashSet seq = new HashSet();

      if (directSuccessors == null ||
          directSuccessors.size() == 0) {
        continue;
      }

      // Phase 1: prepare set
      for (int i = 0; i < directSuccessors.size(); i++) {
        IFBase f = (IFBase) directSuccessors.elementAt(i);
        Vector s = getSuccessor(f);
        if (s != null) {
          seq.addAll(s);
        }
      }
      // Phase 2: check that any successor of an element in the
      // successor set is not contain in the successor set
      while (!seq.isEmpty()) {
        IFBase f = (IFBase) seq.iterator().next();
        seq.remove(f);
        if (directSuccessors.contains(f)) {
          System.out.println("testDirectSuccessors: fatal error");
          System.out.println("DirectSuccessors contains:" + f.toString());
          System.out.println("Problem node:" + n.toString());

          return false;
        }
        Vector s = getSuccessor(f);
        if (s != null) {
          seq.addAll(s);
        }
      }
    }
    return true;
  }

  /**
   * Tests that the union of all predecessor sets to the root are
   * equal to the set of covering filters. This must be the case for
   * the forwards set computation to work correctly.
   * @return boolean
   */
  public boolean testCoverSets() {

    Set temps = templates.entrySet();

    // loop over all nodes
    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      IFBase n = (IFBase) e.nextElement();

      HashSet coverSet = new HashSet();
      HashSet computedCoverSet = new HashSet();

      Vector p = getPre(n);
      if (p != null) {
        Vector v = new Vector();
        v.addAll(p);
        while (!v.isEmpty()) {
          IFBase x = (IFBase) v.remove(0);
          computedCoverSet.add(x);
          p = getPre(x);
          v.addAll(p);
        }

        // Compute true cover set
        for (Enumeration e2 = templates.elements(); e2.hasMoreElements(); ) {
          IFBase n2 = (IFBase) e2.nextElement();
          if (n2 != n && n2.covers(n)) {
            coverSet.add(n2);
          }
        }

        // Test set equivalence
        boolean b1 = coverSet.containsAll(computedCoverSet);
        boolean b2 = computedCoverSet.containsAll(coverSet);
        if (! (b1 && b2)) {
          System.out.println("Problem node: " + n.toString());
          System.out.println("testCoverSets: cover computation error");
          System.out.println("Actual contains all computed: " + b1);
          System.out.println("Computed contains all actual: " + b2);
          System.out.println("-------------------------------");
          System.out.println("Actual: " + coverSet);
          System.out.println("Computed: " + computedCoverSet);
          System.out.println("-------------------------------");
          return false;
        }
      }
    }
    return true;
  }

  /**
   * Tests that the predecessors set is correct and only contains the direct
   * predecessors of a node
   * @return boolean
   */
  public boolean testDirectPredecessors() {

    // loop over all nodes
    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      IFBase n = (IFBase) e.nextElement();
      // Test that the set of direct successors is correct
      Vector directPredecessors = getPre(n);

      if (directPredecessors == null ||
          directPredecessors.size() == 0) {
        continue;
      }

      HashSet seq = new HashSet();

      // Phase 1: prepare set
      // We add only the direct predecessors of the direct predecessor to
      // the current node
      for (int i = 0; i < directPredecessors.size(); i++) {
        IFBase f = (IFBase) directPredecessors.elementAt(i);
        Vector s = getPre(f);
        if (s != null) {
          seq.addAll(s);
        }
      }
      // Phase 2: check that any successor of an element in the
      // successor set is not contain in the successor set
      while (!seq.isEmpty()) {
        IFBase f = (IFBase) seq.iterator().next();
        seq.remove(f);

        if (directPredecessors.contains(f)) {
          System.out.println("testDirectPredecessors: fatal error");
          System.out.println("DirectPredecessors contains:" + f.toString());
          System.out.println("Problem node:" + n.toString());

          return false;
        }

        Vector s = getPre(f);
        if (s != null) {
          seq.addAll(s);
        }
      }
    }
    return true;
  }

  /**
   * Tests the integrity of the graph
   * @return boolean
   */
  public boolean testAncPreSet() {
    Hashtable za = new Hashtable();

    // we loop over all nodes
    for (Enumeration e = anc.keys(); e.hasMoreElements(); ) {
      IFBase key = (IFBase) e.nextElement();

      if (za.get(key) != null) {
        System.out.println("testAndPre: Duplicate entry in hash");
        System.exit(0);
      }
      za.put(key, "true");

      Vector v = getPre(key);
      Vector va = getPre(key);

      if (v != null) {
        if (v.contains(key)) {
          System.out.println("testAncPre: loop detected: pre(f) contains f");
          return false;
        }
        Vector v2 = getSuccessor(key);
        Iterator i = v2.iterator();
        while (i.hasNext()) {
          IFBase f = (IFBase) i.next();
          if (v.contains(f)) {
            System.out.println("testAncPre: loop detected: pre(f) contains " +
                               " element of anc(f)");
            return false;
          }
          if (va.contains(f)) {
            System.out.println(
                "testAncPre: indirect cover detected: anc(f) contains " +
                " element of anc(anc(f))");
            return false;
          }
        }
      }
    }
    return true;
  }

  /**
   * Tests that the root set does not have covering nodes
   * @return boolean
   */
  public boolean testRoots() {
    Iterator i = roots.iterator();
    while (i.hasNext()) {
      IFBase b = (IFBase) i.next();
      if (!getPre(b).isEmpty()) {
        System.out.println("Root node has predecessors!");
        System.out.println("Node: " + b);
        System.out.println("Predecessors: " + getPre(b));
        return false;
      }

      // Test root is really root
      Iterator j = roots.iterator();
      while (j.hasNext()) {
        IFBase c = (IFBase) j.next();
        if (c == b) {
          continue;
        }
        if (c.covers(b)) {
          System.out.println("Root was covered!");
          System.out.println("Node: " + b);
          System.out.println("Covering node: " + c);
          return false;
        }
      }
    }
    return true;
  }

  public boolean printRoots() {
    Iterator i = roots.iterator();
    while (i.hasNext()) {
      IFBase b = (IFBase) i.next();
      System.out.println("Node: " + b);
      System.out.println("Tid: " + b.getTid());
      System.out.println("Predecessors: " + getPre(b));
      System.out.println("Ancestors: " + getSuccessor(b));
      System.out.println("******************************");
    }
    return true;
  }

  /**
   * Lookup a node
   * @param t IFBase
   * @return IFBase
   */
  public IFBase lookup(IFBase t) {
    Iterator yi = getPre(t).iterator();
    while (yi.hasNext()) {
      IFBase y = (IFBase) yi.next();
      if (y == t) {
        return y;
      }
    }
    return t;
  }

  /**
   * Returns true if the object is removable
   * @param t IFBase
   * @param color Object
   * @return boolean
   */
  public boolean removable(IFBase t, Object color) {
    IFBase f = (IFBase) templates.get(t.getCoverId());

    if (f == null) {
      return false;
    }
    if (f.getTid().contains(color)) {
      return true;
    }
    return false;
  }

  /** Removes the given object from the poset.
   * @param s the object to remove
   * @param color the color of the object
   * @return 0 if remove failed, 1 if remove was the root node, greater than 1 otherwise,
   * returns -1 if remove was succesful but there were multiple copies
   */
  public synchronized int remove(IFBase s, Object color) {
    int ret = POSET_ERROR;
    ops = 0;

    if (test) {
      System.out.println("Integrity check 1");
      testAncPreSet();
    }

    IFBase f0 = (IFBase) templates.get(s.getCoverId());

    if (f0 != null) {
      if (f0.getTid() == null) {
        System.out.println("Poset: color or tid was null");
        System.out.println("Color: " + color);
        System.out.println("Poset: " + f0);
        return POSET_ERROR;
      }

      if (color != null) {
        if (f0.getTid().contains(color)) {
          // Notify listeners about deletion
          f0.getTid().remove(color);

          if (f0.getTid().size() > 0) {
            if (color != null) {
              fireEventToListeners(DELETE_EVENT, f0);
            }
            System.out.println("Poset remove: was duplicate");
            return POSET_DUPLICATE;
          }
        }
        else {
          System.out.println("Poset Fatal: color " + color + " not found");
          return POSET_ERROR;
        }
      }
      s = f0; // operate on original object
    }
    else {
      if (debug) {
        System.out.println("Poset remove: filter not found");
      }
      return POSET_ERROR;
    }

    // Notify listeners about addition
    if (color != null) {
      f0.getTid().add(color);
      fireEventToListeners(DELETE_EVENT_BEFORE, s);
      f0.getTid().remove(color);
    }
    templates.remove(s.getCoverId());

    Set result = new HashSet();
    IFBase x, y;
    Iterator xi, yi;

    // Disconnect nodes
    yi = ( (Vector) getSuccessor(s).clone()).iterator();

    while (yi.hasNext()) {
      y = (IFBase) yi.next();
      getPre(y).remove(s);
    }

    if (test) {
      System.out.println("Integrity check 2");
      testAncPreSet();
    }

    if (getPre(s).isEmpty()) {
      // root node
      ret = POSET_ROOT;

      yi = ( (Vector) getSuccessor(s).clone()).iterator();

      while (yi.hasNext()) {

        y = (IFBase) yi.next();

        if (getPre(y).isEmpty()) {
          roots.add(y);
          result.add(y);
        }
      }
      roots.remove(s);
    }
    else {
      // connect nodes
      ret = POSET_NOTROOT;

      xi = ( (Vector) getPre(s).clone()).iterator();

      while (xi.hasNext()) {

        x = (IFBase) xi.next(); // x in pre(s)

        getSuccessor(x).remove(s);

        if (test) {
          if (getSuccessor(s).contains(x)) {
            System.out.println("Fatal: anc contained x");
          }
          if (getPre(x).contains(s)) {
            System.out.println("Fatal 2: anc contained x");
          }
        }

        yi = getSuccessor(s).iterator();

        while (yi.hasNext()) {

          y = (IFBase) yi.next();

          if (!indirectSuccessorTest(x, y)) {

            if (!connectNodes(x, y)) {
              System.out.println("connectNodes error in remove");
              if (test) {
                System.exit(0);
              }
            }
          }
        }
      }
    }
    // ensure deletion
    getPre(s).clear();
    getSuccessor(s).clear();

    if (test) {
      System.out.println("Integrity check 3");
      testAncPreSet();
    }

    // Notify listeners about addition
    if (color != null) {
      fireEventToListeners(DELETE_EVENT_AFTER, s);
    }

    return ret;
  }

  /**
   * Test if x indirectly covers y
   * @param x IFBase
   * @param y IFBase
   * @return boolean
   */
  public boolean indirectSuccessorTest(IFBase x, IFBase y) {
    Iterator i = getSuccessor(x).iterator();
    while (i.hasNext()) {
      IFBase z = (IFBase) i.next();
      ops++;
      if (z.covers(y)) {
        return true;
      }
    }
    return false;
  }

  /**
   * Computes the set immediate predecessors for f that do not contain s
   * @param f IFBase
   * @param s Object
   * @return Set
   */
  public Set getImmediatePredecessors(IFBase f, Object s) {
    IFBase sub, y;
    LinkedList visitlist = new LinkedList();
    Set visited = new HashSet();
    boolean f_lower;
    Iterator i = roots.iterator();

    if (debug) {
      System.out.println("Pred. iterating root set");
    }

    while (i.hasNext()) {
      sub = (IFBase) i.next();
      ops++;

      if (sub.covers(f)) {
        if (debug) {
          System.out.println("Pred: Sub: " + sub + " covers and " + sub.getTid() +
                             " color: " + s);
        }
        if (sub.getTid().contains(s)) {
          return null;
        }
        else {
          visitlist.addLast(sub);
        }
      }
    }

    Set result = new HashSet();
    HashSet cache = new HashSet();

    if (debug) {
      System.out.println("Pred. starting process");
    }
    ListIterator li;

    while ( (li = visitlist.listIterator()).hasNext()) {
      sub = (IFBase) li.next();
      li.remove();
      i = getSuccessor(sub).iterator();
      f_lower = false;

      if (debug) {
        System.out.println("****** next level: parent: " + sub);
        // try to locate lower covering sub
      }

      while (i.hasNext()) { // iterate successors
        y = (IFBase) i.next();

        if (visited.add(y)) { // process new nodes only
          ops++;
          if (y.covers(f)) {
            cache.add(y);
            if (debug) {
              System.out.println("Pred 1: Sub: " + y + " covers and " +
                                 y.getTid() +
                                 " color: " + s);
            }
            if (y.getTid().contains(s)) { // if color contained
              return null;
            }
            else {
              f_lower = true;
              visitlist.addLast(y);
            }
          }
        }

        // We can avoid this comparison if we remove the predecessor set
        else if (!f_lower) {
//          ops++;
          if (cache.contains(y)) {
//          if (y.covers(f)) {
            if (debug) {
              System.out.println("Pred 2: Sub: " + y + " covers and " +
                                 y.getTid() +
                                 " color: " + s);
            }
            f_lower = true;
          }
        }
      }
      if (debug) {
        System.out.println("f_lower:" + f_lower);
      }

      if (!f_lower) { // if a lower covering sub was not found
        result.add(sub);
      }
    }

    return result;
  }

  /**
   * Computes the set immediate predecessors for f that do not contain s
   * @param f IFBase
   * @param s Object
   * @return Set
   */
  private Set getImmediatePredecessorsOriginal(IFBase f, Object s) {
    IFBase sub, y;
    LinkedList visitlist = new LinkedList();
    Set visited = new HashSet();
    boolean f_lower;
    Iterator i = roots.iterator();

    if (debug) {
      System.out.println("Pred. iterating root set");
    }

    while (i.hasNext()) {
      sub = (IFBase) i.next();
      ops++;

      if (sub.covers(f)) {
        if (debug) {
          System.out.println("Pred: Sub: " + sub + " covers and " + sub.getTid() +
                             " color: " + s);
        }
        if (sub.getTid().contains(s)) {
          return null;
        }
        else {
          visitlist.addLast(sub);
        }
      }
    }
    Set result = new HashSet();

    if (debug) {
      System.out.println("Pred. starting process");
    }
    ListIterator li;

    while ( (li = visitlist.listIterator()).hasNext()) {
      sub = (IFBase) li.next();
      li.remove();
      i = getSuccessor(sub).iterator();
      f_lower = false;

      if (debug) {
        System.out.println("****** next level: parent: " + sub);
        // try to locate lower covering sub
      }

      while (i.hasNext()) { // iterate successors
        y = (IFBase) i.next();

        if (visited.add(y)) { // process new nodes only
          ops++;
          if (y.covers(f)) {
            if (debug) {
              System.out.println("Pred 1: Sub: " + y + " covers and " +
                                 y.getTid() +
                                 " color: " + s);
            }
            if (y.getTid().contains(s)) { // if color contained
              return null;
            }
            else {
              f_lower = true;
              visitlist.addLast(y);
            }
          }
        }
        else if (!f_lower) {
          ops++;
          if (y.covers(f)) {
            if (debug) {
              System.out.println("Pred 2: Sub: " + y + " covers and " +
                                 y.getTid() +
                                 " color: " + s);
            }
            f_lower = true;
          }
        }
      }
      if (debug) {
        System.out.println("f_lower:" + f_lower);
      }

      if (!f_lower) { // if a lower covering sub was not found
        result.add(sub);
      }
    }

    return result;
  }

  /**
   * Computes the intersection of the successor sets of the given collection
   * @param pred Collection
   * @return HashSet
   */
  public HashSet computeIntersection(Collection pred) {
    HashSet I = new HashSet();
    Iterator i = pred.iterator();

    boolean first = true;
    while (i.hasNext()) {
      Vector s = getSuccessor( (IFBase) i.next());
      if (s == null) {
        continue;
      }
      if (first) {
        I.clear();
        I.addAll(s);
        first = false;
      }
      else {
        I.retainAll(s);
      }
    }
    return I;
  }

  /**
   * Compute the set of immediate successors of f given the set of
   * immediate predecessors
   * @param f IFBase
   * @param pred Collection
   * @return Set
   */
  public Set getImmediateSuccessors(IFBase f, Collection pred) {
    IFBase sub, y;
    LinkedList visitlist = new LinkedList();
    HashSet visited = new HashSet();
    Iterator i;

    if (pred == null || pred.isEmpty()) {
      if (debug) {
        System.out.println("getImmediateSuccessors: was root");
      }
      visitlist.addAll(roots);
    }
    else {
      i = pred.iterator();
      // sufficient to add any
      // we would want to consider only the intersection of
      // the successors but that would require traversing the
      // successor sets
//        while (i.hasNext()) {
      visitlist.addLast( (IFBase) i.next());
//        }
    }

    Set result = new HashSet();
    ListIterator li;

    // find the direct covered descendants
    while ( (li = visitlist.listIterator()).hasNext()) {
      sub = (IFBase) li.next();
      li.remove();
      // ensure that the node should be visited

      if (visited.contains(sub)) {
        continue;
      }

      // It is not necessary to test the pred set if duplicates
      // are handled properly. The element cannot cover its predecessors.
      boolean coverFlag = false;

      if (pred != null && pred.contains(sub)) {
        ;
      }
      else {
        ops++;
        coverFlag = f.covers(sub);
      }

      if (coverFlag) {
        if (test) {
          if (pred.contains(sub)) {
            System.out.println("Successors: tried to insert pred");
            System.out.println("[" + f.getCoverId() + "]");
            System.out.println("[" + sub.getCoverId() + "]");
            if (sub.getCoverId().equals(f.getCoverId())) {
              System.out.println("identical cover id");
            }
            System.exit(0);
          }
        }
        visited.add(sub);

        // This is needed to remove redundancy from the successor set
        HashSet sibs = new HashSet();
        addSuccessors(sibs, sub);
        // We have to do two things:
        // 1. add all siblings to visited list
        // 2. remove any siblings from the result set (otherwise there may be
        // redundant elements)
        if (debug) {
          System.out.println("Sibs set size: " + sibs.size());
        }
        visited.addAll(sibs);
        result.removeAll(sibs);
        visitlist.removeAll(sibs);
        result.add(sub);
      }
      else {
        visited.add(sub);
        i = getSuccessor(sub).iterator();
        while (i.hasNext()) {
          y = (IFBase) i.next();
          //   if (visited.add(y)) {
          visitlist.addLast(y);
          //  }
        }
      }
    }
    return result;
  }

  /**
   * Returns all nodes as IFBase
   * @return Vector
   */
  public Vector getAllNodes() {
    Vector ret = new Vector();
    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      IFBase n = (IFBase) e.nextElement();
      ret.add(n);
    }
    return ret;
  }

  /** Checks if the filter is already contained in the structure
   * @param f the object to add
   * @param s the color/interface to add
   * @throws PosetException
   * @return code
   */
  public int contains(IFBase f, Object s) throws PosetException {
    IFBase f0 = (IFBase) templates.get(f.getCoverId());

    if (f0 != null) {
      // duplicate
      if (!f0.getTid().contains(s)) { // new color
        // We should check that predecessors do not contain
        // the color
        if (containsColor(f0, s)) {
          return POSET_CONTAINED;
        }
        return POSET_NOT_CONTAINED;
      }
      return POSET_CONTAINED;
    }
    return FilterForest.POSET_NOTDUPLICATE;
  }

  /**
   * Iterates over predecessors to find the given color
   * @param f IFBase
   * @param s Object
   * @return boolean
   */
  public boolean containsColor(IFBase f, Object s) {
    Vector v = (Vector) pre.get(f);
    if (v == null) {
      return false;
    }
    Vector stack = new Vector();
    stack.addAll(v);

    Vector visited = new Vector();

    while (stack.size() > 0) {
      IFBase f2 = (IFBase) stack.remove(0);
      if (f2.getTid().contains(s)) {
        if (debug) {
          System.out.println("!!!" + f2 + " contained color: " + s);
        }
        return true;
      }
      if (!visited.contains(f2)) {
        v = (Vector) pre.get(f2);
        if (v != null) {
          stack.addAll(v);
        }
        visited.add(f2);
      }

    }
    return false;
  }

  /** Adds the given object to the poset.
   * @param f the object to add
   * @param s the color/interface
   * @return null if failure, original if duplicate or the given object
   */
  public int add(IFBase f, Object s) {
    int ret = POSET_ERROR;
    ops = 0;

    IFBase f0 = (IFBase) templates.get(f.getCoverId());

    if (f0 != null) {
      if (f == f0) {
        System.out.println("Fatal: tried to insert the same filter");
        return POSET_DUPLICATE;
      }

      // duplicate
      if (!f0.getTid().contains(s)) { // new color
        // We should check that predecessors do not contain
        // the color. This is now done before.

        if (containsColor(f0, s)) {
          return POSET_DUPLICATE;
        }

        eliminate_color(f0, s);
        f0.addTid(s);

        // Notify listeners about addition
        fireEventToListeners(ADD_EVENT, f0);
      }
      return POSET_DUPLICATE;
    }

    // else we reset the interface vector

    if (f.getTid() != null) {
      f.getTid().clear();
    }

    getSuccessor(f).clear();
    getPre(f).clear();

    // We first check predecessors for duplicate detection
    // This walks the poset from the start...
//    long t1 = System.currentTimeMillis();
    Set pred = getImmediatePredecessors(f, s);
//    System.out.println("ins1: " + (System.currentTimeMillis()-t1));

    if (pred == null) {
      return ret;
    }

    f.addTid(s);
    templates.put(f.getCoverId(), f);

    //  t1 = System.currentTimeMillis();
    Collection c = getImmediateSuccessors(f, pred);
    /*
      System.out.println("ins2: " + (System.currentTimeMillis()-t1));
      if ((System.currentTimeMillis()-t1) >  50) {
        System.out.println("Collection size: " + c.size());
        debug = true;
        getImmediateSuccessors(f, pred);
        debug = false;
      }
     */
    if (debug) {
      System.out.println("Predecessors: " + pred);
    }
    if (debug) {
      System.out.println("Successors: " + c);
    }

    //  t1 = System.currentTimeMillis();

    insert(f, pred, c);

    // new color so we perform the eliminate procedure
    eliminate_color(f, s);
    //  System.out.println("ins3: " + (System.currentTimeMillis()-t1));

    if (debug) {
      System.out.println("Add test color index");
      this.testColorIndex();
    }
    if (test) {
      System.out.println("Add test color index");
      this.testColorIndex();

      if (!this.testRoots()) {
        System.out.println("Insertion failure");
        System.out.println("Node:" + f);
        System.out.println("Successors: " + c);
        System.out.println("Predecessors: " + pred);
        remove(f, s);

        debug = true;
        pred = getImmediatePredecessors(f, s);
        System.out.println("Predecessors: " + pred);

        System.exit(0);
      }
    }

    // Notify listeners about addition
    fireEventToListeners(ADD_EVENT, f);

    if (isRootNode(f)) {
      return POSET_ROOT;
    }

    return POSET_NOTROOT;
  }

  /**
   * Eliminates colors excluding the input node
   * @param start IFBase
   * @param s Object
   */
  public void eliminate_color(IFBase start, Object s) {
    if (NO_ELIMINATE) {
      return;
    }

    LinkedList visitlist = new LinkedList();
    Set visited = new HashSet();

    visitlist.addAll(getSuccessor(start));
    Vector removeList = new Vector();

    if (test) {
      System.out.println("Eliminate-color Integrity check 1");
      testAncPreSet();
    }

    ListIterator li;

    while ( (li = visitlist.listIterator()).hasNext()) {
      IFBase sub = (IFBase) li.next();
      li.remove();
      if (visited.add(sub)) {
        // remove all colors
        if (sub.getTid().contains(s)) {
          while (sub.getTid().remove(s)) {
            ;
          }
          if (sub.getTid().isEmpty()) {
            if (debug) {
              System.out.println("Removing sub: " + s);
              System.out.println("Hash code: " + sub.hashCode());
            }
            // Notify listeners about addition
            fireEventToListeners(DELETE_EVENT_ELIMINATE, sub);

            remove(sub, null);
          }
        }
        else { // if contained color, siblings cannot contain
          visitlist.addAll(getSuccessor(sub));
        }
      }
    }
  }

  /**
   * Match a notification
   * @param e Object
   * @return TripletTable
   */
  public TripletTable matchT(Object e) {
    if (! (e instanceof Notification)) {
      return null;
    }

    ops = 0;

    Vector stack = new Vector();
    Set visited = new HashSet();
    boolean quit = false;
    TripletTable ret = new TripletTable();
    Iterator i = roots.iterator();

    while (i.hasNext()) {
      IFBase sub = (IFBase) i.next();
      ops++;
      visited.add(sub);
      if (sub.match(null, e)) {
        ret.insert( ( (Template) sub).getInterface(), sub);
        stack.add(sub);
      }
      else {
        Vector a = getSuccessor(sub);
        addSuccessors(visited, sub);
      }
    }

    // Traverse the tree
    while (stack.size() > 0 && !quit) {
      IFBase n = (IFBase) stack.remove(0);
      if (n != null) {
        Vector ci = getSuccessor(n);
        if (ci != null) {
          Iterator k = ci.iterator();
          while (k.hasNext()) {
            IFBase n2 = (IFBase) k.next();
            if (visited.add(n2)) {
              ops++;
              Vector a = getSuccessor(n2);
              if (n2.match(null, e)) {

                // This expansion is needed for proper multicast
                for (int j = 0; j < n2.getTid().size(); j++) {
                  Object o = n2.getTid().elementAt(j);
                  //       ret.insert(_t.getInterface(), _t);
                  ret.insert( (String) o, n2);
                }

//           ret.insert( ( (Template) n2).getInterface(), n2);
                if (a != null && a.size() > 0) {
                  stack.add(n2);
                }
              }
              else {
                // We add all the children to the visited list
                // because their parent did not match
                // This prevents unnecessary checking from other
                // branches...
                addSuccessors(visited, n2);
              }
            }
          }
        }
      }
    }
    return ret;
  }

  /**
   * Match an object against the data structure using the given matching
   * type.
   * @param e IFBase
   * @param matchType int 0 = e covers x, 1 = x covers e, 2 = overlap
   * @param rootOnly boolean rootOnly if true processes only the first two levels
   * @return Vector
   */
  public Vector match(IFBase e, int matchType, boolean rootOnly) {
    ops = 0;

    Vector stack = new Vector();
    Set visited = new HashSet();
    boolean quit = false;
    HashSet result = new HashSet();

    Iterator i = roots.iterator();
    while (i.hasNext()) {
      IFBase sub = (IFBase) i.next();
      boolean matchFlag = false;
      if (result.containsAll(sub.getTid())) {
        matchFlag = true;
        stack.add(sub);
      }
      else {
        ops++;
        boolean mthc = false;
        if (matchType == 0) {
          if (e.covers(sub)) {
            addToResults(result, sub);
            stack.add(sub);
            mthc = true;
          }
        }
        else if (matchType == 1) {
          if (sub.covers(e)) {
            // no need to go into this branch
            addSuccessors(result, sub);
            mthc = true;
          }
          else {
            // move into
            stack.add(sub);
            mthc = true;
          }
        }
        else if (matchType == 2) {
          if (sub.overlaps(e)) {
            // no need to go into this branch
            addSuccessors(result, sub);
            mthc = true;
          }
          else {

          }
        }
        if (mthc) {
          matchFlag = true;
        }
      }
      if (!matchFlag) {
        addSuccessors(visited, sub);
      }
    }

    // Traverse the tree
    while (stack.size() > 0 && !quit) {
      IFBase n = (IFBase) stack.remove(0);
      if (n != null) {
        Vector ci = getSuccessor(n);
        if (ci != null) {
          Iterator k = ci.iterator();
          while (k.hasNext()) {
            IFBase n2 = (IFBase) k.next();
            if (!visited.contains(n2)) { // if not yet visited n2
              visited.add(n2);
              Vector a = getSuccessor(n2);

              boolean matchFlag = false;

              // if the current node has been matched already
              // we may skip the matching

              if (result.containsAll(n2.getTid())) {
                matchFlag = true;
                // we have to go deeper
                if (a != null && a.size() > 0) {
                  // if rootOnly flag is set, we process only the first two levels
                  if (rootOnly && !findRootForForwards(n2, null)) {
                    stack.add(n2);
                  }
                  else {
                    stack.add(n2);
                  }
                }
              }
              else {
                ops++;
                boolean mthc = false;
                if (matchType == 0) {
                  if (e.covers(n2)) {
                    mthc = true;
                  }
                }
                else if (matchType == 1) {
                  if (n2.covers(e)) {
                    mthc = true;
                  }
                }
                else if (matchType == 2) {
                  if (n2.overlaps(e)) {
                    mthc = true;
                  }
                }
                if (mthc) {
                  matchFlag = true;
                  addToResults(result, n2);
                  if (a != null && a.size() > 0) {
                    // if rootOnly flag is set, we process only the first two levels
                    if (rootOnly && !findRootForForwards(n2, null)) {
                      stack.add(n2);
                    }
                    else {
                      stack.add(n2);
                    }
                  }
                }
              }

              if (!matchFlag) {
                // We add all the children to the visited list
                // because their parent did not match
                // This prevents unnecessary checking from other
                // branches...
                addSuccessors(visited, n2);
              }
            }
          }
        }
      }
    }

    Vector res = new Vector(result);
    return res;
  }

  /** For testing DoubleForest */
  public int DF_counter = 0;

  /**
   * Match subspace
   * @param e IFBase
   * @param rootOnly boolean
   * @param startSet HashSet
   * @param endSet HashSet
   * @return Vector
   */
  public Vector matchSubspace(IFBase e, boolean rootOnly,
                              HashSet startSet, HashSet endSet) {

    ops = 0;

    Vector stack = new Vector();
    Set visited = new HashSet();
    HashSet res = new HashSet();

    Iterator i = roots.iterator();
    while (i.hasNext()) {
      IFBase sub = (IFBase) i.next();
      ops++;
      DF_counter++;
      visited.add(sub);
      if (sub.covers(e)) {
        res.add(sub);
        stack.add(sub);
      }
    }

    boolean quit = false;

    // Traverse the tree
    while (stack.size() > 0 && !quit) {
      IFBase n = (IFBase) stack.remove(0);

      if (n != null) {
        Vector ci = this.getSuccessor(n);

        if (ci != null && ci.size() > 0) {
          Iterator k = ci.iterator();

          while (k.hasNext()) {
            IFBase t = (IFBase) k.next();

            if (visited.contains(t)) {
              continue;
            }

            visited.add(t);

            boolean mthc = false;
            boolean iterate = false;
            boolean lower = false;

            if (endSet != null && endSet.contains(t)) {
              lower = true;
            }
            // Upper and lower bound optimizations
            if (startSet != null && !startSet.contains(t)) {
              ;
            }
            else {
              if (lower) {
                mthc = true;
              }
              else {
                ops++;
                DF_counter++;
                if (t.covers(e)) {
                  mthc = true;
                }
              }
            }

            if (mthc) {
              if (!res.contains(t)) {
                res.add(t);
              }
            }

            if (iterate || mthc) {
              if (rootOnly) {
                // TBD
              }
              stack.add(t);
            }
          }
        }
      }
    }
    Vector ret = new Vector(res);
    return ret;
  }

  /**
   * Tests the path to root (also in PosetForwardsSet)
   * @param f IFBase
   * @param root Vector
   * @return boolean if the node is covered by at least 2 parent nodes. In this
   * case the forwards set is zero sized.
   */
  public boolean findRootForForwards(IFBase f, Vector root) {
    Vector a = getPre(f.getCoverId());
    for (int i = 0; i < a.size(); i++) {
      IFBase x = (IFBase) a.elementAt(i);
      Vector p = getPre(x.getCoverId());
      if (p == null || p.size() == 0) {
        if (root != null) {
          root.add(x);
        }
      }
      else {
        return true;
      }
    }
    return false;
  }

  /**
   * Match a notification
   * @param e Object
   * @return Vector
   */
  public Vector match(Object e) {
    if (! (e instanceof Notification)) {
      return null;
    }

    ops = 0;

    Vector stack = new Vector();
    Set visited = new HashSet();
    boolean quit = false;
    HashSet result = new HashSet();

    Iterator i = roots.iterator();
    while (i.hasNext()) {
      IFBase sub = (IFBase) i.next();
      boolean matchFlag = false;
      if (result.containsAll(sub.getTid())) {
        matchFlag = true;
        stack.add(sub);
      }
      else {
        ops++;
        if (sub.match(null, e)) {
          matchFlag = true;
          addToResults(result, sub);
          stack.add(sub);
        }
      }
      if (!matchFlag) {
        addSuccessors(visited, sub);
      }
    }

    // Traverse the tree
    while (stack.size() > 0 && !quit) {
      IFBase n = (IFBase) stack.remove(0);
      if (n != null) {
        Vector ci = getSuccessor(n);
        if (ci != null) {
          Iterator k = ci.iterator();
          while (k.hasNext()) {
            IFBase n2 = (IFBase) k.next();
            if (!visited.contains(n2)) { // if not yet visited n2
              boolean matchFlag = false;

              visited.add(n2);
              Vector a = getSuccessor(n2);

              // if the current node has been matched already
              // we may skip the matching

              if (result.containsAll(n2.getTid())) {
                matchFlag = true;
                // we have to go deeper
                if (a != null && a.size() > 0) {
                  stack.add(n2);
                }
              }
              else {
                ops++;
                if (n2.match(null, e)) {
                  matchFlag = true;
                  addToResults(result, n2);
                  if (a != null && a.size() > 0) {
                    stack.add(n2);
                  }
                }
              }

              if (!matchFlag) {
                // We add all the children to the visited list
                // because their parent did not match
                // This prevents unnecessary checking from other
                // branches...
                //addSuccessors(visited, n2);
                visited.addAll(a);
              }
            }
          }
        }
      }
    }
    Vector res = new Vector(result);
    return res;
  }

  /**
   * Updates the result set
   * @param result HashSet
   * @param f IFBase
   */
  public void addToResults(HashSet result, IFBase f) {
    for (int j = 0; j < f.getTid().size(); j++) {
      Object o = f.getTid().elementAt(j);
      if (!result.contains(o)) {
        result.add(o);
      }
    }

  }

  /**
   * Add all successors of the given node to the input vector
   * @param v Vector
   * @param b IFBase
   */
  public void addSuccessors(Set v, IFBase b) {
    Vector a = getSuccessor(b);
    for (int i = 0; i < a.size(); i++) {
      IFBase z = (IFBase) a.elementAt(i);
      if (!v.contains(z)) {
        v.add(z);
        Vector c = getSuccessor(z);
        if (c != null && c.size() > 0) {
          addSuccessors(v, z);
        }
      }
    }
  }

  /**
   * Add all predecessors of the given node to the input vector
   * @param v Vector
   * @param b IFBase
   */
  public void addPredecessors(Set v, IFBase b) {
    Vector a = getPre(b);
    for (int i = 0; i < a.size(); i++) {
      IFBase z = (IFBase) a.elementAt(i);
      if (!v.contains(z)) {
        v.add(z);
        Vector c = getPre(z);
        if (c != null && c.size() > 0) {
          addPredecessors(v, z);
        }
      }
    }
  }

  /**
   * Add all siblings of the given node to the input vector
   * @param v Vector
   * @param b IFBase
   */
  private void addSuccessorsOld(Set v, IFBase b) {
    Vector a = getSuccessor(b);
    for (int i = 0; i < a.size(); i++) {
      IFBase z = (IFBase) a.elementAt(i);
      v.add(z);
      Vector c = getSuccessor(z);
      if (c != null && c.size() > 0) {
        addSuccessors(v, z);
      }
    }
  }

  /** Returns true if the node is connected to root
   * @param n the IFBase
   * @return true if connected
   */
  public boolean connectedToRoot(IFBase n) {
    IFBase t = n;
    for (; ; ) {
      Vector p = getPre(t);
      if (p == null || p.size() < 1) {
        break;
      }
      t = (IFBase) p.firstElement();
      if (t == null) {
        return false;
      }
    }
    if (roots.contains(t)) {
      return true;
    }
    return false;
  }

  boolean testMinimumColor(IFBase n, Vector v, Vector route) {

    Vector stack = new Stack();
    Vector attrs = new Vector();
    attrs.add(n);
    attrs.add(v);
    attrs.add(route);
    stack.add(attrs);
    int c = 0;
    while (!stack.isEmpty()) {
      Vector atr = (Vector) stack.remove(0);
      n = (IFBase) atr.elementAt(0);
      v = (Vector) atr.elementAt(1);
      route = (Vector) atr.elementAt(2);

      Vector cols = n.getTid();
      route.add(n);
      if (cols.size() == 0) {
        System.out.println("Fatal: detected a node without colors");
        System.exit(0);
      }

      for (int i = 0; i < v.size(); i++) {
        Object o = (Object) v.elementAt(i);
        if (cols.contains(o)) {
          System.out.println("Minimum color error: " + o);
          System.out.println("Input vector" + v);
          System.out.println("Current node colors:" + cols);
          System.out.println("Hash code: " + n.hashCode());
          System.out.println("Route colors: ");
          for (int k = 0; k < route.size(); k++) {
            IFBase ifb = (IFBase) route.elementAt(k);
            if (roots.contains(ifb)) {
              System.out.println(ifb.getTid() + " hash:" + ifb.hashCode() +
                                 " [root]");
            }
            else {
              System.out.println(ifb.getTid() + " hash:" + ifb.hashCode());
            }
          }
          System.exit(0);
        }
      }

      if (getPre(n).contains(n)) {
        System.out.println("Fatal pre contained n");
        System.exit(0);
      }
      v.addAll(cols);

      Vector ch2 = getPre(n);
      for (int i = 0; i < ch2.size(); i++) {
        IFBase x = (IFBase) ch2.elementAt(i);
        if (getSuccessor(n).contains(x)) {
          System.out.println("Test: Fatal anc contained parent");
          System.exit(0);
        }
      }

      Vector ch = getSuccessor(n);
      if (ch != null) {
        for (int i = 0; i < ch.size(); i++) {
          if (ch.contains(n)) {
            System.out.println("Fatal anc contained parent");
            System.exit(0);
          }
          if (getPre(n).contains(ch.elementAt(i))) {
            System.out.println("Fatal pre contained ch");
            System.exit(0);
          }
          attrs = new Vector();
          attrs.add(ch.elementAt(i));
          attrs.add(v.clone());
          attrs.add(route.clone());
          stack.add(attrs);
        }
      }
    }
    return true;
  }

  public String runTests() {
    String myName = this.getClass().getName();
    String str = "[" + myName + "] test structure\n\n";
    boolean ret = true;

    str += "TEST: predecessor and successor set tests\n";
    ret = testAncPreSet(); // test integrity of the graph

    if (ret) {
      str += "RESULT: predecessor and successor set tests ok\n";
    }
    else {
      str += "RESULT: predecessor and successor set tests failed\n";
      ret = true;
    }

    str += "TEST: direct predecessor set test\n";
    if (testDirectPredecessors()) {
      str += "RESULT:  direct predecessor set test ok\n";
    }
    else {
      str += "RESULT:  direct predecessor set test failed\n";
    }

    str += "TEST: direct successor set test\n";
    if (testDirectSuccessors()) {
      str += "RESULT:  direct predecessor set test ok\n";
    }
    else {
      str += "RESULT:  direct successor set test failed\n";
    }

    str +=
        "TEST: coverage of union of predecessor sets to root (all nodes)\n";
    if (testCoverSets()) {
      str += "RESULT:  coverage ok\n";
    }
    else {
      str += "RESULT:  coverage failed\n";
    }

    str += "TEST: root tests\n";
    ret = testRoots();

    if (ret) {
      str += "RESULT: root tests ok\n";
    }
    else {
      str += "RESULT: root tests failed\n";
      ret = true;
    }

    str += "TEST: Minimum color property\n";

    // testing minimum color property
    Iterator i = roots.iterator();
    while (i.hasNext()) {
      IFBase k = (IFBase) i.next();
      if (k.getTid().size() == 0) {
        str += "Poset: A root node was found with no color!\n";
        ret = false;
      }
      Vector route = new Vector();
      if (!testMinimumColor(k, new Vector(), route)) {
        str += "Fatal: testColorIndex: minimum color property " +
            " violation\n";
        ret = false;
        break;
      }
    }

    if (ret) {
      str += "RESULT: Minimum color property ok\n";
    }
    else {
      str += "RESULT: Minimum color property failed\n";
      ret = true;
    }

    str += "TEST: Checking orphans and connections to root\n";

    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      IFBase n = (IFBase) e.nextElement();
      if (!connectedToRoot(n)) {
        str += "Found a node in templates not connected to root\n";
        ret = false;
      }
      if (n.getTid().size() == 0) {
        str += "Fatal: node with empty color set\n";
        str += "Probably an orphan\n";
        ret = false;
      }
    }

    if (ret) {
      str += "RESULT: Checking orphans and connections to root ok\n";
    }
    else {
      str += "RESULT: Checking orphans and connections to root failed\n";
      ret = true;
    }

    return str;
  }

  /** Tests the color index
   *  @return boolean false if tests fail
   */
  public boolean testColorIndex() {
    boolean ret = true;

    System.out.println("[Poset] testColorIndex");

    if (!testAncPreSet()) { // test integrity of the graph
      System.out.println("Anc and Pre set test failure");
      System.exit(0);
    }

    if (!testDirectPredecessors()) {
      System.out.println("Direct predecessor test failure");
      System.exit(0);
    }

    if (!testDirectSuccessors()) {
      System.out.println("Direct predecessor test failure");
      System.exit(0);
    }

    if (!testCoverSets()) {
      System.out.println("Node coverage test failed");
      System.exit(0);
    }

    if (!testRoots()) {
      System.out.println("Root test failed");
      System.exit(0);
    }

    // testing minimum color property
    Iterator i = roots.iterator();
    while (i.hasNext()) {
      IFBase k = (IFBase) i.next();
      if (k.getTid().size() == 0) {
        System.out.println("Poset: A root node was found with no color!");
        System.out.println(k);
        System.exit(0);
      }
      Vector route = new Vector();
      if (!testMinimumColor(k, new Vector(), route)) {
        System.out.println("Fatal: testColorIndex: minimum color property " +
                           " violation");
        ret = false;
        break;
      }
    }
    System.out.println("Checking orphans ************************");
    for (Enumeration e = templates.elements(); e.hasMoreElements(); ) {
      IFBase n = (IFBase) e.nextElement();
      if (!connectedToRoot(n)) {
        System.out.println("Found a node in templates not connected to root");
        System.out.println(n);
        System.exit(0);
      }
      if (n.getTid().size() == 0) {
        System.out.println("Fatal: node with empty color set");
        System.out.println("Probably an orphan");
        System.out.println(n);
        System.exit(0);
      }
    }

    return ret;
  }

  /* ********************************************************************* */
  /* ********************************************************************* */
  /* ********************************************************************* */
  /* ********************************************************************* */


  /**
   * Debug match routine
   * @param e Object
   * @param x Object
   * @return Vector
   */
  public Vector match(Object e, Object x) {
    if (! (e instanceof Notification)) {
      return null;
    }

    Vector result = new Vector();
    Iterator i = roots.iterator();
    LinkedList visitlist = new LinkedList();
    Set visited = new HashSet();
    IFBase sub;

    System.out.println(e.toString());

    while (i.hasNext()) {
      sub = (IFBase) i.next();
      // the poset doesn't work properly so we prevent false positives
      if (sub.match(null, e)) {
        if (x != null) {
          if (sub.getTid().contains(x)) {
            System.out.println(sub);
            if (templates.get(sub.getCoverId()) != null) {
              System.out.println("in templates");
            }
            else {
              System.out.println("Not in templates");
              //        System.out.println("Anc: " + getSuccessor(sub));
            }
          }
        }
        visited.add(sub);
        visitlist.addLast(sub);
        result.addAll(sub.getTid());
      }
    }

    ListIterator li;
    while ( (li = visitlist.listIterator()).hasNext()) {
      sub = (IFBase) li.next();
      li.remove();
      i = getSuccessor(sub).iterator();

      while (i.hasNext()) {
        IFBase y = (IFBase) i.next();
        if (visited.add(y) && y.match(null, e)) {
          if (x != null) {
            if (sub.getTid().contains(x)) {
              System.out.println(sub);
              if (templates.get(sub.getCoverId()) != null) {
                System.out.println("in templates");
              }
              else {
                System.out.println("Not in templates");
              }
            }
          }

          visitlist.addLast(y);
          result.addAll(y.getTid());
        }
      }
    }

    // compress result
    Vector ret = new Vector();
    for (int k = 0; k < result.size(); k++) {
      Object o = (Object) result.elementAt(k);
      if (!ret.contains(o)) {
        ret.add(o);
      }
    }

    return ret;
  }

  /**
   * Adds the given Vector to the poset. Works only for Templates.
   * @param v the objects to add
   * @throws PosetException
   * @return int the number of elements inserted starting from the beginning
   */
  public synchronized int add(Vector v) {
    if (v == null || v.size() == 0 ||
        ! (v.firstElement() instanceof Template)) {
      return 0;
    }
    for (int i = 0; i < v.size(); i++) {
      Template ic = (Template) v.elementAt(i);
      if (add(ic, ic.getInterface()) == POSET_ERROR) {
        //break;
      }
    }
    return v.size();
  }

  void printDotFull(IFBase f, PrintWriter fn) {
    Vector pre = getPre(f);
    Vector anc = getSuccessor(f);

    try {

      String dis = "";
      if (f instanceof Filter) {
        dis = ( (Filter) f).getFilter();
      }
      else {
        dis = f.toString();

      }
      if (pre != null && pre.size() > 0) {
      }
      else {
        fn.println("\"" + dis + "\";"); // root

      }
      if (anc != null && anc.size() > 0) {
        for (int i = 0; i < anc.size(); i++) {
          IFBase f2 = (IFBase) anc.elementAt(i);
          String str = "";
          if (f2 instanceof Filter) {
            str = ( (Filter) f2).getFilter();
          }
          else {
            str = f2.toString();

          }
          fn.println("\"" + dis + "\"->\"" + str + "\";");
        }
      }

    }
    catch (Exception e) {
      e.printStackTrace();
    }
  }

  public static void main(String[] args) {

    // Simple test follows
    Filter f1 = new Filter("field1", EventTriplet.TYPE_INT,
                           "rng 200 300|rng 0 100");

    Filter f2 = new Filter("field1", EventTriplet.TYPE_INT,
                           "rng 9 14");

    Filter f3 = new Filter("field1", EventTriplet.TYPE_INT,
                           "rng 201 209");

    Filter f4 = new Filter("field1", EventTriplet.TYPE_INT,
                           "rng 202 206");

    Filter f5 = new Filter("field1", EventTriplet.TYPE_INT,
                           "rng 301 400");

    Filter f6 = new Filter("field1", EventTriplet.TYPE_INT,
                           "rng 199 280");

    Poset p = new Poset();
    int ops = 0;
    p.add(f1, "koe1");
    ops += p.ops;
    p.add(f2, "koe2");
    ops += p.ops;
    p.add(f3, "koe3");
    ops += p.ops;
    p.add(f4, "koe4");
    ops += p.ops;
    p.add(f5, "koe5");
    ops += p.ops;
    p.add(f6, "koe6");
    ops += p.ops;

    System.out.println("Ops Total: " + ops);
    p.remove(f5, "koe5");
    ops += p.ops;
    System.out.println("Ops Total: " + ops);

    try {
      PrintWriter fout = new PrintWriter(new FileWriter(
          "testposet.dat"), false);
      p.print(fout);
      fout.close();
    }
    catch (Exception e2) {
      e2.printStackTrace();
    }
    System.out.println(p.toString());
  }

  public void test() {
    String s = "<name>type</name><type>3</type><filter>equ koe</filter><dlm/><name>val1</name><type>0</type><filter>equ 53.0</filter><dlm/><name>val2</name><type>0</type><filter>lt 60.0</filter><dlm/>  <nfields>*</nfields><localsend>true</localsend><expiration></expiration><id>null</id><interface>iface0</interface><sessionid>null</sessionid>";
    Template t = new Template(s);
    String s2 = "<name>type</name><type>3</type><filter>equ koe</filter><dlm/><name>val1</name><type>0</type><filter>equ 20.0</filter><dlm/><name>val0</name><type>0</type><filter>lt 50.0</filter><dlm/>  <nfields>*</nfields><localsend>true</localsend><expiration></expiration><id>null</id><interface>iface1</interface>  <sessionid>null</sessionid>";
    Template t2 = new Template(s2);
    String s3 = "<name>type</name><type>3</type><filter>equ koe</filter><dlm/><name>val0</name><type>0</type><filter>lt 67.0</filter><dlm/>  <nfields>*</nfields><localsend>true</localsend><expiration></expiration><id>null</id><interface>iface3</interface><sessionid>null</sessionid>";
    Template t3 = new Template(s3);
    String s4 = "<name>type</name><type>3</type><filter>equ koe</filter><dlm/><name>val2</name><type>0</type><filter>neq 74.0</filter><dlm/>  <nfields>*</nfields><localsend>true</localsend><expiration></expiration><id>null</id><interface>iface2</interface><sessionid>null</sessionid>";
    Template t4 = new Template(s4);

    this.add(t, "koe");
    this.add(t2, "koe");
    this.add(t3, "koe");
    this.add(t4, "koe");

    try {
      PrintWriter fout = new PrintWriter(new FileWriter(
          "poset_bug.dat"), false);
      printRoots(fout);
      fout.close();
    }
    catch (Exception e2) {
      e2.printStackTrace();
    }

  }

  public void printRoots(PrintWriter fn) {
    for (int i = 0; i < this.get().size(); i++) {
      IFBase t = (IFBase) get().elementAt(i);
      fn.println(t.toString());
    }
  }

  public int count() {
    int count = 0;
    LinkedList visitlist = new LinkedList();
    Set visited = new HashSet();
    visitlist.addAll(roots);
    ListIterator li;
    while ( (li = visitlist.listIterator()).hasNext()) {
      IFBase sub = (IFBase) li.next();
      li.remove();
      if (visited.add(sub)) {
        if (sub.getTid().isEmpty()) {
          System.out.println("Found a colorless node in poset count");
          System.out.println("Pre:" + getPre(sub));
          System.out.println("------------------");
          System.out.println("Anc:" + getPre(sub));
          System.out.println("------------------");
          if (templates.get(sub.getCoverId()) != null) {
            System.out.println("In templates");
          }
          else {
            System.out.println("NOT in templates");
          }
        }
        else {
          count++;
        }
        visitlist.addAll(getSuccessor(sub));
      }
    }
    return count;
  }

  /**
   * Get the data structure as HTML
   * @return String
   */
  public String getHTMLOutput() {
    StringBuffer buf = new StringBuffer();
    buf.append("<html><body>");

    //
    // prints the poset
    //
    LinkedList visitlist = new LinkedList();
    Set visited = new HashSet();

    visitlist.addAll(roots);

    ListIterator li;
    while ( (li = visitlist.listIterator()).hasNext()) {
      IFBase sub = (IFBase) li.next();
      li.remove();
      if (visited.add(sub)) {
        printHtml(sub, buf);
        visitlist.addAll(getSuccessor(sub));
      }
    }

    buf.append("</body></html>");
    return buf.toString();
  }

  void printHtml(IFBase n, StringBuffer buf) {
    Filter f = null;
    String s1 = "";

    ITemplate t = (ITemplate) n;
    s1 = fuegocore.notify.core.server.ui.ManagementUI.printFilterTable(t);

    Vector a = getPre(n);
    Vector b = getSuccessor(n);

    if (b == null || b.size() == 0) {

      buf.append("<table bgcolor=\"#AAAAAF\"><tr><td>");
      buf.append(s1 + "<br>");
      buf.append("</td></tr></table>");
      return;
    }

    try {
      for (int i = 0; i < b.size(); i++) {
        IFBase n2 = (IFBase) b.elementAt(i);
        if (n2 instanceof ITemplate) {
          ITemplate t2 = (Template) n2;

          String s2 = fuegocore.notify.core.server.ui.ManagementUI.
              printFilterTable(t2);
          buf.append("<table><tr><td>");
          buf.append(s1 + "</td><td>");
          buf.append("</td><td>--></td><td>");
          buf.append(s2);
          buf.append("</td></tr></table>");
        }
      }
    }
    catch (Exception e) {
      e.printStackTrace();
    }
  }

  public void print(PrintWriter out) throws IOException {
    //
    // prints the poset
    //
    LinkedList visitlist = new LinkedList();
    Set visited = new HashSet();

    visitlist.addAll(roots);

    ListIterator li;
    while ( (li = visitlist.listIterator()).hasNext()) {
      IFBase sub = (IFBase) li.next();
      li.remove();
      if (visited.add(sub)) {
        printDotFull(sub, out);
        visitlist.addAll(getSuccessor(sub));
      }
    }
  }
}
