package fuegocore.notify.core.filter.poset.forwards;

import java.util.*;

import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.filter.poset.*;

/**
 * Implements an event-based forwards set. The set is updated when
 * the associated data structure changes.
 */
public class ForestForwardsSet
    implements PosetChangeListener {
  /**
   * The underlying data structure
   */
  FilterForest poset;
  /**
   * The forwards set
   */
  Hashtable forwards;
  /**
   * The set of neighbours
   */
  Vector neighbours;

  Hashtable tempDelStorage;

  boolean debug = false;

  public Hashtable changeSet;

  /** Total number of added filters to update sets */
  int total_filters = 0;

  /** Use approximative forwards sets */
  public static boolean NONCOMPLETE = false;

  /** Move nodes not on level two to level three */
  public static boolean REBALANCE_LEVEL2 = false;

  public int getTotal() {
    return total_filters;
  }

  public void setApproximate(boolean val) {
    NONCOMPLETE = val;
  }

  public void resetTotal() {
    total_filters = 0;
  }

  public FilterForest getPoset() {
    return poset;
  }

  /**
   * Create a poset forwards set
   * @param pos IFilterPoset
   * @param neighb Vector
   */
  public ForestForwardsSet(FilterForest pos, Vector neighb) {
    poset = pos;
    poset.addListener(this);
    forwards = new Hashtable();
    neighbours = neighb;
    tempDelStorage = new Hashtable();
    changeSet = new Hashtable();
  }

  /**
   * Returns the changed set
   * @return Hashtable
   */
  public Hashtable getChangeSet() {
    return changeSet;
  }

  /**
   * Clears the changed set
   */
  public void resetChanges() {
    changeSet.clear();
  }

  /**
   * Returns the forwards table
   * @return Hashtable
   */
  public Hashtable get() {
    return forwards;
  }

  public boolean checkChange(IFBase x, Vector newForwards) {

    Vector oldForwards = (Vector) forwards.get(x.getCoverId());
    /*
      if ( (oldForwards == null || oldForwards.size() == 0) &&
          (newForwards.size() == 0)) {
        return false;
      }
     */
    if (oldForwards == null) {
      return true;
    }

    if (oldForwards.equals(newForwards)) {
      return false;
    }
    return true;
  }

  /**
   * Update forwards set for a level 1 (root) node and level 2
   * @param ifb IFBase
   */
  public void computeForwardsAdd(IFBase ifb) {
    if (debug) {
      System.out.println("Forwards size: " + forwards.size());
    }

    Node n = poset.lookupNode(ifb);
    if (n == null) { // not found
      return;
    }

    if (n.getParent() == null) {
      if (debug) {
        System.out.println("computeForwardsAdd: " + ifb + ": parent was null");
      }
    }
    else if (n.getParent().getParent() == null) {
      if (debug) {
        System.out.println("computeForwardsAdd: " + ifb +
                           ": parent.parent was null");
      }
    }
    else {
      if (debug) {
        System.out.println("computeForwardsAdd: not interesting");
      }

      // else not interesting
      forwards.remove(ifb.getCoverId());
      return;
    }

    /*
          1L add to root
          Insertion of F may alter the forwards sets
          of covered nodes.

          For each covered node that has a forwards
          entry contained in forwards(F), it is removed.
     */
    Vector roots = poset.getAsNodes();

    // This may affect the forwards sets of covered nodes
    Vector newForwards = new Vector();

    // If covered by some 1-level or 2-level node, return immediately
    // the forwards set will be zero sized
    computeForwardsAdd(n, neighbours,
                       newForwards, NONCOMPLETE);

    boolean b = checkChange(n.getData(), newForwards);

    if (!b) {
      return;
    }

    changeSet.put(n.getData(), newForwards);
    forwards.put(n.getData().getCoverId(), newForwards);
    total_filters += (newForwards.size());
    if (debug) {
      System.out.println("ComputeForwardsAdd: " + total_filters);
    }

    if (n.getParent() == null && newForwards.size() < neighbours.size() - 1) {
      System.out.println("computeForwardsAdd: newForwards size error: " +
                         newForwards.size());
      System.exit(0);
    }

    if (newForwards.size() == 0) {
      if (debug) {
        System.out.println("Empty forwards set.");
      }
      return;
    }
    /*
        Vector level2 = new Vector();

        // We should ensure that excess forwards sets are removed

        if (n.getParent() == null) {
          // reduce the forwards sets
          for (int k = 0; k < n.getChildren().size(); k++) {
            Node n3 = (Node) n.getChildren().elementAt(k);
            Vector f3 = (Vector) forwards.get(n3.getData().getCoverId());
            if (f3 != null) {
              f3.removeAll(newForwards);
            }
          }
        }
     */
    // We break here
    // This causes some filters to have redundant elements in their
    // forwards sets
    return;
    /*
        // First level
        for (int i = 0; i < roots.size(); i++) {
          Node n2 = (Node) roots.elementAt(i);
          if (n2 == n) {
            continue;
          }

          // Compute second level forwards
          Vector tempset = (Vector) forwards.get(n2.getData().getCoverId());

          // approximate second level forwards set
          // we build a candidate set that could have been blocked by the
          // deleted node

          if (tempset != null && tempset.size() > 0) {
            Vector approx = (Vector) neighbours.clone();
            approx.removeAll(tempset);
            if (approx.size() > 0) {
              if (newForwards.contains(approx.firstElement())) {
                if (n.getParent() == null) {
                  level2.addAll(n2.getChildren());
                }
                else {
                  poset.incOps();
                  if (n2.getData().covers(n.getData())) {
                    level2.addAll(n2.getChildren());
                  }
                }
              }
            }
          }
        }

        // Second level
        for (int i = 0; i < level2.size(); i++) {
          Node y = (Node) level2.elementAt(i);
          if (y.getData() == ifb) {
            continue;
          }

          Vector f2 = (Vector) forwards.get(y.getData().getCoverId());
          if (f2 == null) {
            continue;
          }

          if (newForwards.containsAll(f2)) {
            poset.incOps();
            if (n.getData().covers(y.getData())) {
              f2.clear();
              forwards.remove(y.getData().getCoverId());
            }
          }
        }
     */
  }

  /**
   * Computes the forwards set for del
   * @param ifb IFBase
   * @param removed IFBase
   * @return boolean
   */
  public boolean computeForwardsForDel(IFBase ifb, IFBase removed) {
    if (debug) {
      System.out.println("computeForwardsForDel");
    }

    Node n = poset.lookupNode(ifb);

    if (n == null) {
      if (debug) {
        System.out.println("Node not found");
      }
      return false; // eliminated
    }

    if (n.getParent() == null) {
      /*
          if (debug) {
       System.out.println("computeForwardsDel: " + ifb + ": parent was null");
          }
       */
    }
    else if (n.getParent().getParent() == null) {
      /*
            if (debug) {
              System.out.println("computeForwardsDel: " + ifb +
                                 ": parent.parent was null");
            }
       */
    }
    else {
      if (debug) {
        System.out.println("computeForwardsDel: not interesting");
      }

      // else not interesting
      forwards.remove(ifb.getCoverId());
      return false;
    }

    /*
          1L add to root
          Insertion of F may alter the forwards sets
          of covered nodes.

          For each covered node that has a forwards
          entry contained in forwards(F), it is removed.
     */
    // This may affect the forwards sets of covered nodes
    Vector newForwards = new Vector();

    // If covered by some 1-level or 2-level node, return immediately
    // the forwards set will be zero sized
    computeForwardsAdd(n, neighbours,
                       newForwards, NONCOMPLETE);

//   boolean b = checkChange(n.getData(), newForwards);
//   if (!b) return;

    changeSet.put(n.getData(), newForwards);
    forwards.put(n.getData().getCoverId(), newForwards);
    if (removed != null) {
      // Only forward where the unsub was sent
      Vector temp = (Vector) newForwards.clone();
      Vector fw = new Vector();
      Vector temp2 = (Vector) changeSet.get(removed);

      // take the intersection of the unsub set and this fw set
      for (int k = 0; k < newForwards.size(); k++) {
        Object col = newForwards.elementAt(k);
        if (temp2.contains(col)) {
          fw.add(col);
        }
      }
      if (debug) {
        System.out.println("ComputeForwardsForDel: " + total_filters);
        System.out.println("Forest unsub [delete]:" + fw);
      }
      total_filters += (fw.size());
    }
    else {
      if (debug) {
        System.out.println("ComputeForwardsForDel: " + total_filters);
        System.out.println("Forest unsub [no-delete]:" + newForwards);
      }
      total_filters += (newForwards.size());
    }

    if (n.getParent() == null && newForwards.size() < neighbours.size() - 1) {
      System.out.println("computeForwardsAdd: newForwards size error: " +
                         newForwards.size());
    }

    if (newForwards.size() == 0 || newForwards.size() == 1) {
      return false;
    }

    // Have to update the next level as well
    if (!NONCOMPLETE) {
      if (n.getParent() == null && n.getChildren() != null) {
        for (int i = 0; i < n.getChildren().size(); i++) {
          Node z = (Node) n.getChildren().elementAt(i);
          if (!changeSet.contains(z.getData())) {
            computeForwardsForDel(z.getData(), removed);
          }
        }
      }
    }
    return true;
  }

  /**
   * Resets the forwards set for all nodes that are not on the 1/2 levels
   */
  public void resetForwards() {
    Hashtable copy = new Hashtable();
    Vector roots = (Vector) poset.getAsNodes().clone();
    while (!roots.isEmpty()) {
      Node n2 = (Node) roots.remove(0);
      if (n2.getParent() == null || n2.getParent().getParent() == null) {
        // Remove any forwards set entry
        //forwards.remove(n2.getData().getCoverId());
        Vector v = (Vector) forwards.get(n2.getData().getCoverId());
        if (v != null) {
          copy.put(n2.getData().getCoverId(), v);
        }
        if (n2.getParent() == null && n2.getChildren() != null) {
          roots.addAll(n2.getChildren());
        }
      }
    }
    forwards.clear();
    forwards.putAll(copy);
  }

  /**
   * Update forwards set for a level 1 (root) node and level 2.
   * @param ifb IFBase
   */
  public boolean computeForwardsDel(IFBase ifb, Vector candidateSet,
                                    boolean realUpdate, boolean skipCheck) {

    Node n = poset.lookupNode(ifb);

    if (n == null) {
      return false;
    }

    if (n != null && n.getParent() == null) {
      ;
    }
    else if (n.getParent().getParent() == null) {
      ;
    }
    else {
      // else not interesting
      forwards.remove(ifb.getCoverId());
      return false;
    }

    /*
       For del we check all nodes on the second level that are
       covered by F and have an empty forwards set due to covering.
     */
    Vector roots = poset.getAsNodes();

    Vector newForwards = (Vector) forwards.remove(n.getData().getCoverId());

    // may contain an additional entry
    if (newForwards == null || newForwards.size() == 0) {
      return false;
    }

    Vector oldForwards = (Vector) newForwards.clone();

    // we recompute this
    newForwards.clear();

    // If covered by some 1-level or 2-level node, return immediately
    // the forwards set will be zero sized
    computeForwardsAdd(n, neighbours,
                       newForwards, skipCheck);

    if (newForwards == null || newForwards.size() == 0) {
      return false;
    }

    if (realUpdate) { // put it back
      forwards.put(n.getData().getCoverId(), newForwards.clone());
      Vector temp = (Vector) newForwards.clone();
      newForwards.clear();
      newForwards.addAll(oldForwards);
      newForwards.removeAll(temp); // remove where it has already
    }

    changeSet.put(n.getData(), newForwards);
    total_filters += (newForwards.size());
    if (debug) {
      System.out.println("ComputeForwardsForDel2: " + total_filters);
      System.out.println("Set: " + newForwards + " tid: " +
                         n.getData().getTid());
    }

    Vector vc = getCandidates(ifb, newForwards, neighbours);
    candidateSet.addAll(vc);

    /*
      Vector level2 = new Vector();

      // First level
      for (int i = 0; i < roots.size(); i++) {
        Node n2 = (Node) roots.elementAt(i);
        if (n2.getData() == ifb) {
          continue;
        }
        // Compute second level forwards
        Vector tempset = (Vector) forwards.get(n2.getData().getCoverId());

        if (tempset == null) {
          if (debug) {
            System.out.println("Fatal: could not locate forwards set for: " +
                               n2.getData());
          }
          System.exit(0);
        }

        // approximate second level forwards set
        // we build a candidate set that could have been blocked by the
        // deleted node
        Vector approx = (Vector) neighbours.clone();
        approx.removeAll(tempset);

        // note the optimizations
        // If a root contains the forwards set
        // then there is no need to consider the second level since they
        // are already established at the remote node and won't be affected
        // by the unsubscription
        if (tempset.size() != neighbours.size() &&
            !tempset.containsAll(newForwards)) {
          if (approx.size() > 0) {
            if (newForwards.contains(approx.firstElement())) {
              if (n.getParent() == null) {
                level2.addAll(n2.getChildren());
              }
              else {
                poset.incOps();
                if (n2.getData().covers(n.getData())) {
                  level2.addAll(n2.getChildren());
                }
              }
            }
          }
        }
      }

      // Second level
      for (int i = 0; i < level2.size(); i++) {
        Node y = (Node) level2.elementAt(i);
        if (y.getData() == ifb) {
          continue;
        }

        Vector f2 = (Vector) forwards.get(y.getData().getCoverId());
        // should be contained in the forwards
        if (f2 == null || f2.size() == 0 || intersects(newForwards, f2)) {
          poset.incOps();
          if (n.getData().covers(y.getData())) {
            candidateSet.add(y);
          }
        }
      }
     */
    if (debug) {
      System.out.println("Candidate set size:" + candidateSet.size());
    }

    return true;
  }

  public boolean intersects(Vector a, Vector b) {
    for (int i = 0; i < a.size(); i++) {
      Object o = a.elementAt(i);
      if (b.contains(o)) {
        return true;
      }
    }
    return false;
  }

  /**
   * Computes the forwards set for the added filter. Called after add.
   * @param n Node
   * @param neighbours Vector
   * @param forw Vector
   * @param skip boolean
   * @return int
   */
  public int computeForwardsAdd(Node n, Vector neighbours,
                                Vector forw, boolean skip) {

    if (n == null || forw == null || neighbours == null) {
      return 0;
    }

    // We need to find the covering set
    Node a = n.getParent();

    if (a == null) {
      forw.clear();
      ForwardsSet.computeForwards(n.getData(), forw,
                                  neighbours);
    }
    else {
      // Not a ROOT-node
      Vector parentForwards = new Vector();
      parentForwards = (Vector) forwards.get(a.getData().getCoverId());

      HashSet cand = new HashSet();
      cand.addAll(neighbours);
      if (parentForwards != null) {
        cand.removeAll(parentForwards);
      }
      else {
        System.out.println(
            "computeForwardsAdd: fatal error parentForwards was null");
      }
      forw.clear();

      if (cand.size() == 0) {
        return 1;
      }

      Object fo = cand.iterator().next();
      if (fo == null) {
        System.out.println("computeForwardsAdd: missing forward object");
        return 0;
      }

      // Now we check that is this forward object correct
      // There should not exist any filter that covers this filter
      // and has already been sent to this interface
      if (!skip) {
        boolean r = checkForwardsCover(n, fo, neighbours);
        if (debug) {
          System.out.println("checkForwardsCover returned:" + r);
        }
        if (!r) {
          forw.add(fo);
          return 1;
        }
      }
      else {
        forw.add(fo);
      }
      return 2;
    }
    return 1;
  }

  /**
   * Checks whether or not the given filter is covered by another filter
   * that has been sent to the given forward interface.
   * @param n Node
   * @param fo Object
   * @param neighbours Vector
   * @return boolean
   */
  public boolean checkForwardsCover(Node n, Object fo, Vector neighbours) {
    Vector seq = new Vector();
    seq.addAll(poset.getAsNodes());
    IFBase f = n.getData();

    // we know that n is a second-level node
    while (seq.size() > 0) {
      Node x = (Node) seq.remove(0);

      // this requires that the second level is sibling pure
      if (x.getData() != f && x.getData() != n.getParent().getData()) {
        if (x.getParent() == null) {
          poset.incOps();
          if (x.getData().covers(f)) {
            if (!x.getData().getTid().contains(fo)) {
              return true;
            }
            else {
              seq.addAll(x.getChildren());
            }
          }
        }
        else {
          if (!x.getData().getTid().contains(fo)) {
            poset.incOps();
            if (x.getData().covers(f)) {
              if (this.REBALANCE_LEVEL2) {
                // We move this node under the covering node, and thus rebalance second level
                if (debug) {
                  System.out.println("Balancing node: " + n);
                }
                System.out.println("Balancing node: " + n + " hash: " +
                                   n.hashCode());
                try {
                  poset.addNodeEvent(n, neighbours); // latter parameter is a dummy
                  n.getParent().getChildren().remove(n);
                  x.getChildren().add(n);
                  n.setParent(x);
                  poset.addNodeEvent(n, neighbours); // latter parameter is a dummy
                }
                catch (Exception e) {
                  e.printStackTrace();
                  System.exit(0);
                }
              }

              return true; // second level node covers --> cannot have forwards set
            }
          }
        }
      }
    }
    return false;
  }

  public Vector getCandidates(IFBase y, Vector newForwards, Vector neighbours) {
    Vector ret = new Vector();

    Vector seq = new Vector();
    seq.addAll(poset.getAsNodes());

    // note that this is in a different direction than before

    // we know that y is a second-level node
    while (seq.size() > 0) {
      Node x = (Node) seq.remove(0);

      if (x.getData() != y) {
        if (x.getParent() == null && x.getData().getTid().size() == 1) {
          // if tid size is greater, the forwards sets of descendants are empty already
          seq.addAll(x.getChildren());
        }
        else {
          Vector f2 = (Vector) forwards.get(x.getData().getCoverId());
          // should be contained in the forwards
          boolean forward = false;

          if (f2 == null || f2.size() == 0 || intersects(newForwards, f2)) {
            forward = true;
          }

          if (forward) {
            poset.incOps();
            if (y.covers(x.getData())) {
              ret.add(x);
            }
          }
        }
      }
    }
    return ret;
  }

  /**
   * Called when the data structure changes
   * @param event String
   * @param x IFBase
   */
  public void eventOccurred(String event, IFBase x) {
    if (debug) {
      System.out.println("###################################");
      System.out.println("Event:" + event + " occurred");
    }

    if (event.equals(FilterForest.DELETE_EVENT_ELIMINATE)) {
      // Eliminate color was used to remove the node
      if (x.getTid() == null || x.getTid().size() == 0) {
        Vector ft = (Vector) forwards.remove(x.getCoverId());
        if (ft != null) {
          ft.clear();
        }
      }
    }
    else if (event.equals(FilterForest.DELETE_EVENT)) {
      Vector uncovered = new Vector();
      Node z = (Node) poset.lookupNode(x);
      // updates the uncovered set!
      if (computeForwardsDel(x, uncovered, true, false)) {
        uncovered.addAll(z.getChildren());
        for (int i = 0; i < uncovered.size(); i++) {
          Node n = (Node) uncovered.elementAt(i);
          if (!changeSet.contains(n.getData())) {
            computeForwardsForDel(n.getData(), x);
          }
        }
      }
    }
    else if (event.equals(FilterForest.DELETE_EVENT_BEFORE)) {
      if (x.getTid() == null || x.getTid().size() == 0) {
        System.out.println(
            "Fatal in ForestForwardsSet: filter did not have tid");
        System.exit(0);
      }
      Vector uncovered = new Vector();
      Node z = (Node) poset.lookupNode(x);
      // updates the uncovered set!
      // potential second level nodes
      if (computeForwardsDel(x, uncovered, false, false)) {
        uncovered.addAll(z.getChildren()); // potential root nodes
        tempDelStorage.put(x, uncovered);
      }
    }
    else if (event.equals(FilterForest.DELETE_EVENT_AFTER)) {
      Vector uncovered = (Vector) tempDelStorage.remove(x);
      if (uncovered == null) {
        if (debug) {
          System.out.println(
              "ForestForwardsSet: failure uncovered set missing in del");
        }
      }
      else {
        // Update successor set
        for (int i = 0; i < uncovered.size(); i++) {
          Node n = (Node) uncovered.elementAt(i);
          if (debug) {
            System.out.println("Forest processing: " + n.getData().hashCode());
          }
          if (!changeSet.contains(n.getData())) {
            if (debug) {
              System.out.println("Update " + i + " was not in change set");
            }
            computeForwardsForDel(n.getData(), x);
          }
          else
          if (debug) {
            System.out.println("Update " + i + " was in change set");
          }

        }
      }
    }
    else if (event.equals(FilterForest.RESET)) {
      forwards.clear();
      changeSet.clear();
      tempDelStorage.clear();
      resetTotal();
    }
    else if (event.equals(FilterForest.ADD_EVENT)) {
      // possibly a new entry in a forwards set (only for a root node)
      computeForwardsAdd(x);
    }
    else if (event.equals(FilterForest.ADD_EVENT_AFTER)) {
      computeForwardsAdd(x);
    }
  }
}
