package fuegocore.notify.core.filter.poset.forwards;

import java.util.*;

import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.filter.poset.*;

/**
 * Implements an event-based forwards set. The set is updated when
 * the associated data structure changes.
 */
public class ForestForwardsSetPlain
    implements PosetChangeListener {
  /**
   * The underlying data structure
   */
  FilterForest poset;
  /**
   * The forwards set
   */
  Hashtable forwards;
  /**
   * The set of neighbours
   */
  Vector neighbours;

  Hashtable tempDelStorage;

  boolean debug = false;

  Hashtable changeSet;

  /** Total number of added filters to update sets */
  int total_filters = 0;

  public int getTotal() {
    return total_filters;
  }

  public void resetTotal() {
    total_filters = 0;
  }

  /**
   * Returns the changed set
   * @return Hashtable
   */
  public Hashtable getChangeSet() {
    return changeSet;
  }

  /**
   * Clears the changed set
   */
  public void resetChanges() {
    changeSet.clear();
  }

  /**
   * Create a poset forwards set
   * @param pos IFilterPoset
   * @param neighb Vector
   */
  public ForestForwardsSetPlain(FilterForest pos, Vector neighb) {
    poset = pos;
    poset.addListener(this);
    forwards = new Hashtable();
    neighbours = neighb;
    tempDelStorage = new Hashtable();
    changeSet = new Hashtable();
  }

  /**
   * Returns the forwards table
   * @return Hashtable
   */
  public Hashtable get() {
    return forwards;
  }

  public boolean checkChange(IFBase x, Vector newForwards) {

    Vector oldForwards = (Vector) forwards.get(x.getCoverId());
    /*
      if ( (oldForwards == null || oldForwards.size() == 0) &&
          (newForwards.size() == 0)) {
        return false;
      }
     */
    if (oldForwards == null) {
      return true;
    }

    if (oldForwards.equals(newForwards)) {
      return false;
    }
    return true;
  }

  /**
   * Update forwards set for a level 1 (root) node and level 2
   * @param ifb IFBase
   */
  public void computeForwardsAdd(IFBase ifb) {
    if (debug) {
      System.out.println("Forwards size: " + forwards.size());
    }

    Node n = poset.lookupNode(ifb);

    if (n.getParent() == null) {
      if (debug) {
        System.out.println("computeForwardsAdd: " + ifb + ": parent was null");
      }
    }
    else if (n.getParent().getParent() == null) {
      if (debug) {
        System.out.println("computeForwardsAdd: " + ifb +
                           ": parent.parent was null");
      }
    }
    else {
      if (debug) {
        System.out.println("computeForwardsAdd: not interesting");
      }

      // else not interesting
      forwards.remove(ifb.getCoverId());
      return;
    }

    /*
          1L add to root
          Insertion of F may alter the forwards sets
          of covered nodes.

          For each covered node that has a forwards
          entry contained in forwards(F), it is removed.
     */
    Vector roots = poset.getAsNodes();

    // This may affect the forwards sets of covered nodes
    Vector newForwards = new Vector();

    // If covered by some 1-level or 2-level node, return immediately
    // the forwards set will be zero sized
    computeForwardsAdd(n, neighbours,
                       newForwards);

    forwards.put(n.getData().getCoverId(), newForwards);

//    boolean b = checkChange(n.getData(), newForwards);

    /*
         if (debug) {
           System.out.println("Computed new forwards: " + newForwards.size() +
                              " changed: " + b);
         }
        if (!b) { // this is relevant only for duplicates
          if (debug) {
            System.out.println("No change in forwards set.");
          }
          return;
        }
     */
    if (n.getParent() == null && newForwards.size() < neighbours.size() - 1) {
      System.out.println("computeForwardsAdd: newForwards size error: " +
                         newForwards.size());
    }

    if (newForwards.size() == 0) {
      if (debug) {
        System.out.println("Empty forwards set.");
      }
      return;
    }

    Vector level2 = new Vector();

    // We should ensure that excess forwards sets are removed

    if (n.getParent() == null) {
      // reduce the forwards sets
      for (int k = 0; k < n.getChildren().size(); k++) {
        Node n3 = (Node) n.getChildren().elementAt(k);
        Vector f3 = (Vector) forwards.get(n3.getData().getCoverId());
        if (f3 != null) {
          f3.removeAll(newForwards);
        }
      }
    }

    // First level
    for (int i = 0; i < roots.size(); i++) {
      Node n2 = (Node) roots.elementAt(i);
      if (n2 == n) {
        continue;
      }

      // Compute second level forwards
      Vector tempset = (Vector) forwards.get(n2.getData().getCoverId());

      // approximate second level forwards set
      // we build a candidate set that could have been blocked by the
      // deleted node

      if (tempset != null && tempset.size() > 0) {
        Vector approx = (Vector) neighbours.clone();
        approx.removeAll(tempset);
        if (approx.size() > 0) {
          if (newForwards.contains(approx.firstElement())) {
            if (n.getParent() == null) {
              level2.addAll(n2.getChildren());
            }
            else {
              poset.incOps();
              if (n2.getData().covers(n.getData())) {
                level2.addAll(n2.getChildren());
              }
            }
          }
        }
      }
    }

    // Second level
    for (int i = 0; i < level2.size(); i++) {
      Node y = (Node) level2.elementAt(i);
      if (y.getData() == ifb) {
        continue;
      }

      Vector f2 = (Vector) forwards.get(y.getData().getCoverId());
      if (f2 == null) {
        continue;
      }

      if (newForwards.containsAll(f2)) {
        poset.incOps();
        if (n.getData().covers(y.getData())) {
          f2.clear();
          forwards.remove(y.getData().getCoverId());
        }
      }
    }
  }

  /**
   * Resets the forwards set for all nodes that are not on the 1/2 levels
   */
  public void resetForwards() {
    Hashtable copy = new Hashtable();
    Vector roots = (Vector) poset.getAsNodes().clone();
    while (!roots.isEmpty()) {
      Node n2 = (Node) roots.remove(0);
      if (n2.getParent() == null || n2.getParent().getParent() == null) {
        // Remove any forwards set entry
        //forwards.remove(n2.getData().getCoverId());
        Vector v = (Vector) forwards.get(n2.getData().getCoverId());
        if (v != null) {
          copy.put(n2.getData().getCoverId(), v);
        }
        if (n2.getParent() == null && n2.getChildren() != null) {
          roots.addAll(n2.getChildren());
        }
      }
    }
    forwards.clear();
    forwards.putAll(copy);
  }

  /**
   * Update forwards set for a level 1 (root) node and level 2.
   * @param ifb IFBase
   */
  public void computeForwardsDel(IFBase ifb, Vector candidateSet) {

    Node n = poset.lookupNode(ifb);

    if (n == null || n.getParent() == null) {
      ;
    }
    else if (n.getParent().getParent() == null) {
      ;
    }
    else {
      // else not interesting
      forwards.remove(ifb.getCoverId());
      return;
    }

    /*
       For del we check all nodes on the second level that are
       covered by F and have an empty forwards set due to covering.
     */
    Vector roots = poset.getAsNodes();

    Vector newForwards = (Vector) forwards.remove(n.getData().getCoverId());

    if (newForwards == null || newForwards.size() == 0) {
      return;
    }

    Vector level2 = new Vector();

    // First level
    for (int i = 0; i < roots.size(); i++) {
      Node n2 = (Node) roots.elementAt(i);
      if (n2 == n) {
        continue;
      }
      // Compute second level forwards
      Vector tempset = (Vector) forwards.get(n2.getData().getCoverId());
      if (tempset == null) {
        if (debug) {
          System.out.println("Fatal: could not locate forwards set for: " +
                             n2.getData());
        }
        continue;
      }
      // approximate second level forwards set
      // we build a candidate set that could have been blocked by the
      // deleted node
      Vector approx = (Vector) neighbours.clone();
      approx.removeAll(tempset);

      if (tempset.size() != neighbours.size()) {
        if (approx.size() > 0) {
          if (newForwards.contains(approx.firstElement())) {
            if (n.getParent() == null) {
              level2.addAll(n2.getChildren());
            }
            else {
              poset.incOps();
              if (n2.getData().covers(n.getData())) {
                level2.addAll(n2.getChildren());
              }
            }
          }
        }
      }
    }

    // Second level
    for (int i = 0; i < level2.size(); i++) {
      Node y = (Node) level2.elementAt(i);
      if (y.getData() == ifb) {
        continue;
      }

      Vector f2 = (Vector) forwards.get(y.getData().getCoverId());
      if (f2 == null || f2.size() == 0) {
        poset.incOps();
        if (n.getData().covers(y.getData())) {
          candidateSet.add(y);
        }
      }
    }
  }

  /**
   * Computes the forwards set for the added filter. Called after add.
   * @param n Node
   * @param neighbours Vector
   * @param forw Vector
   * @return int
   */
  public int computeForwardsAdd(Node n, Vector neighbours,
                                Vector forw) {

    if (n == null || forw == null || neighbours == null) {
      return 0;
    }

    // We need to find the covering set
    Node a = n.getParent();

    if (a == null) {
      forw.clear();
      ForwardsSet.computeForwards(n.getData(), forw,
                                  neighbours);
    }
    else {
      // Not a ROOT-node
      Vector parentForwards = new Vector();
      parentForwards = (Vector) forwards.get(a.getData().getCoverId());

      HashSet cand = new HashSet();
      cand.addAll(neighbours);
      cand.removeAll(parentForwards);

      forw.clear();

      if (cand.size() == 0) {
        return 1;
      }

      Object fo = cand.iterator().next();
      if (fo == null) {
        System.out.println("computeForwardsAdd: missing forward object");
        return 0;
      }

      // Now we check that is this forward object correct
      // There should not exist any filter that covers this filter
      // and has already been sent to this interface

      boolean r = checkForwardsCover(n, fo, neighbours);
      if (!r) {
        forw.add(fo);
        return 1;
      }
      return 2;
    }
    return 1;
  }

  /**
   * Checks whether or not the given filter is covered by another filter
   * that has been sent to the given forward interface.
   * @param n Node
   * @param fo Object
   * @param neighbours Vector
   * @return boolean
   */
  public boolean checkForwardsCover(Node n, Object fo, Vector neighbours) {
    Vector seq = new Vector();
    seq.addAll(poset.getAsNodes());
    IFBase f = n.getData();
    // we know that n is a second-level node
    while (seq.size() > 0) {
      Node x = (Node) seq.remove(0);
      if (x.getData() != f && x.getData() != n.getParent().getData()) {
        Vector forw = (Vector) forwards.get(x.getData().getCoverId());
        poset.incOps();
        if (x.getData().covers(f)) {
          if (x.getParent() == null) {
            if (forw != null && forw.contains(fo)) {
              return true;
            }
            else {
              seq.addAll(x.getChildren());
            }
          }
          else {
            return true; // second level node covers --> cannot have forwards set
          }

        }
      }
    }
    return false;
  }

  /**
   * Called when the data structure changes
   * @param event String
   * @param x IFBase
   */
  public void eventOccurred(String event, IFBase x) {
    if (event.equals(FilterForest.DELETE_EVENT_ELIMINATE)) {
      // Eliminate color was used to remove the node
      Vector ft = (Vector) forwards.remove(x.getCoverId());
      if (ft != null) {
        ft.clear();
      }
    }
    else if (event.equals(FilterForest.DELETE_EVENT)) {
      // In order for the forwards set to change, the node must
      // be a root node
      computeForwardsAdd(x);
    }
    else if (event.equals(FilterForest.DELETE_EVENT_BEFORE)) {
      Vector newForwards = new Vector();
      Vector uncovered = new Vector();
      Node z = (Node) poset.lookupNode(x);
      uncovered.addAll(z.getChildren());
      computeForwardsDel(x, uncovered);

      // store the set for later use
      tempDelStorage.put(x, uncovered);
    }
    else if (event.equals(FilterForest.DELETE_EVENT_AFTER)) {
      Vector uncovered = (Vector) tempDelStorage.remove(x);
      if (uncovered == null) {
        if (debug) {
          System.out.println(
              "ForestForwardsSet: failure uncovered set missing in del");
        }
      }
      else {
        // Update successor set

        for (int i = 0; i < uncovered.size(); i++) {
          Node n = (Node) uncovered.elementAt(i);
          computeForwardsAdd(n.getData());
        }
        resetForwards();
      }
    }
    else if (event.equals(FilterForest.RESET)) {
      forwards.clear();
      tempDelStorage.clear();
    }
    else if (event.equals(FilterForest.ADD_EVENT)) {
      // possibly a new entry in a forwards set (only for a root node)
      computeForwardsAdd(x);
    }
    else if (event.equals(FilterForest.ADD_EVENT_AFTER)) {
      computeForwardsAdd(x);
      resetForwards();
    }
  }
}
