package fuegocore.notify.core.filter.poset.forwards;

import java.util.*;

import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.filter.merge.*;

/**
 * This class computes the forwards set for peer-to-peer routers
 * using a naive method. This used used to test the correctness
 * of the filters poset and the forest data structures.
 */
public class ForwardsSet {
  /** Total number of added filters to update sets */
  int total_filters = 0;
  static boolean debug = false;

  public int getTotal() {
    return total_filters;
  }

  public void resetTotal() {
    total_filters = 0;
  }

  public static int computeForwardsSetRemove(Vector originalSet,
                                             Vector neighbours, IFBase t,
                                             Hashtable h) {

    Vector fw = (Vector) h.get(t); // old forwards set

    // We simply remove the filter, and compute a new set
    // then compare any changes in the forwards sets
    Hashtable h2 = computeForwardsSet(originalSet, neighbours);

    Vector fw2 = (Vector) h2.get(t); // new forwards set

    int cost = 0;

    // 1. determine where t was sent
    if (fw == null || fw.size() == 0) {
      if (debug) {
        System.out.println("ForwardsSet: ** no forwards set");
      }
      return 0; // no forwards set
    }

    if (fw.size() == neighbours.size()) {
      if (debug) {
        System.out.println("ForwardsSet:  ** Forward all neighbours:" +
                           neighbours.size());
        System.out.println("Neighbours: " + fw + " tid " + t.getTid());
      }
    }

    if (fw2 != null) { // there is still the old set in existence
      fw.removeAll(fw2);
    }

    cost += fw.size();

    if (debug) {
      System.out.println("ForwardsSet:  ** Forward unsub:" + fw + " " +
                         t.getTid());
      if (fw.size() > 1 && fw.size() < neighbours.size() - 1) {
        System.out.println("ForwardsSet: Fatal size error for fw set: " +
                           fw.size());
        System.out.println("Neighbours: " + neighbours.size());
        System.exit(0);
      }
    }

    // then check uncovered, any new entries in the forwards set

    Enumeration k = h2.keys();
    Enumeration k2 = h2.elements();
    while (k.hasMoreElements()) {
      IFBase key = (IFBase) k.nextElement();
      Vector fws = (Vector) k2.nextElement();
      Vector old = (Vector) h.get(key);

      if (fws != null && fws.size() > 0) {
        // ok. so there is a set, is it larger?
        if (old != null) {
          fws.removeAll(old); // must be an old set
        }
        if (fws.size() > 0) {
          if (debug) {
            System.out.println("Testing candidate:" + fws.size() + " old " +
                               old +
                               " cur " + fws);
            System.out.println("ForwardsSet:  ** level 2 Forward sub:" + fws +
                               " " + key.hashCode());
          }
          cost += fws.size();
        }
      }
    }
    return cost;
  }

  public static Hashtable computeForwardsSet(Vector originalSet,
                                             Vector neighbours) {
    return computeForwardsSet(originalSet,
                              neighbours, null);

  }

  /**
   * Computes the forwards set for all filters
   * @param originalSet Vector
   * @param neighbours Vector
   * @return Hashtable
   */
  public static Hashtable computeForwardsSet(Vector originalSet,
                                             Vector neighbours, IFBase debug) {
    Hashtable ret = new Hashtable();

    // First compute the root set
    Vector root = MergingService.get().getNonCovered(originalSet);

    System.out.println("Computing forwards for root: " + root.size());
    Vector second = (Vector) originalSet.clone();
    second.removeAll(root);

    // then the second level
    HashSet L2 = new HashSet(MergingService.get().getNonCovered(second));

    Vector third = (Vector) originalSet.clone();
    third.removeAll(root);
    third.removeAll(second);

    HashSet L3 = new HashSet(MergingService.get().getNonCovered(third));

    // Iterate over the root set, and compute the forwards sets
    for (int i = 0; i < root.size(); i++) {
      IFBase x = (IFBase) root.elementAt(i);
      Vector fx = new Vector();
      computeForwards(x, fx, neighbours);
      ret.put(x, fx);
    }

    if (debug != null) {
      System.out.println("L1 contains given filter: " + root.contains(debug));
      System.out.println("L2 contains given filter: " + L2.contains(debug));
    }

    // Iterate over the successor set and finalize the computation
    Iterator ik = L2.iterator();
    while (ik.hasNext()) {
      IFBase x = (IFBase) ik.next();

      if (x == debug) {
        System.out.println("Computing given filter");
      }

      computeUnion(x, root, neighbours, ret);
    }

    // Iterate over the root set, and compute the forwards sets
    for (int i = 0; i < third.size(); i++) {
      IFBase x = (IFBase) third.elementAt(i);
      Vector fx = new Vector();
      ret.put(x, fx);
    }

    return ret;
  }

  public static void getRootCover(IFBase x, Vector rs, Vector root) {
    for (int i = 0; i < root.size(); i++) {
      IFBase y = (IFBase) root.elementAt(i);
      if (!y.getCoverId().equals(x.getCoverId()) && y.covers(x)) {
        rs.add(y);
      }
    }
  }

  /**
   * Computes the union of the forwards sets of the given predecessor set
   * @param y IFBase
   * @param root Vector
   * @param neighbours Vector
   * @param ret Hashtable
   * @return Vector
   */
  public static Vector computeUnion(IFBase y, Vector root, Vector neighbours,
                                    Hashtable ret) {
    HashSet forw = new HashSet();

    Vector rs = new Vector();
    getRootCover(y, rs, root);

    // Level 1 forwards sets
    for (int i = 0; i < rs.size(); i++) {
      IFBase x = (IFBase) rs.elementAt(i);
      if (x != y) {
        Vector fx = (Vector) ret.get(x);
        forw.addAll(fx);
      }
    }

    Vector neigh = (Vector) neighbours.clone();
    neigh.removeAll(forw);
    ret.put(y, neigh);

    if (y.getTid().size() == 1 && neigh.contains(y.getTid().firstElement())) {
      System.out.println("Fatal forwards set error. Sending back to source");
      System.out.println(y);
      System.out.println("Tid:" + y.getTid());
      System.out.println(neigh);
      System.out.println("Predecessor fw:" + forw);

      for (int i = 0; i < rs.size(); i++) {
        IFBase x = (IFBase) rs.elementAt(i);
        System.out.println(x);
      }

      System.exit(0);
    }

    return neigh;
  }

  /**
   * Compute the forwards set for the given node
   * @param x IFBase
   * @param forwards Vector
   * @param neighbours Vector
   */
  public static void computeForwards(IFBase x, Vector forwards,
                                     Vector neighbours) {
    // ROOT-node
    forwards.clear();
    forwards.addAll(neighbours);

    if (x.getTid() == null) {
      System.out.println("computeForwards: Fatal error with tid");
    }

    // If there are more than one color, we may send the message to all
    if (x.getTid().size() > 1 || allNotContained(neighbours, x.getTid())) {
      return;
    }

    if (x.getTid().size() > 0) {
      forwards.remove(x.getTid().firstElement());
    }
  }

  public static boolean isLocalClient(IFBase x, Vector neighbours) {
    // If there are more than one color, we may send the message to all
    if (allNotContained(neighbours, x.getTid())) {
      return true;
    }
    return false;
  }

  public static boolean allNotContained(Vector a, Vector b) {
    if (b == null) {
      return false;
    }
    for (int j = 0; j < b.size(); j++) {
      if (!a.contains(b.elementAt(j))) {
        return true;
      }
    }
    return false;
  }
}
