package fuegocore.notify.core.filter.poset.forwards;

import java.util.*;

import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.filter.poset.*;

/**
 * Implements an event-based forwards set. The set is updated when
 * the associated data structure changes.
 */
public class PosetForwardsSet
    implements PosetChangeListener {
  /**
   * The underlying data structure
   */
  Poset poset;
  /**
   * The forwards set
   */
  Hashtable forwards;
  /**
   * The set of neighbours
   */
  Vector neighbours;

  Hashtable tempDelStorage;

  boolean debug = false;

  Hashtable changeSet;

  /** Total number of added filters to update sets */
  int total_filters = 0;

  public int getTotal() {
    return total_filters;
  }

  public Poset getPoset() {
    return poset;
  }

  public void resetTotal() {
    total_filters = 0;
  }

  /**
   * Create a poset forwards set
   * @param pos IFilterPoset
   * @param neighb Vector
   */
  public PosetForwardsSet(Poset pos, Vector neighb) {
    poset = pos;
    poset.addListener(this);
    forwards = new Hashtable();
    neighbours = neighb;
    tempDelStorage = new Hashtable();
    changeSet = new Hashtable();
  }

  /**
   * Returns the changed set
   * @return Hashtable
   */
  public Hashtable getChangeSet() {
    return changeSet;
  }

  /**
   * Clears the changed set
   */
  public void resetChanges() {
    changeSet.clear();
  }

  /**
   * Returns the forwards table
   * @return Hashtable
   */
  public Hashtable get() {
    return forwards;
  }

  public Vector getNewForwards(IFBase x) {
    Vector newForwards = new Vector();
    computeForwardsAdd(x, neighbours, newForwards);
    return newForwards;
  }

  public boolean checkChange(IFBase x, Vector newForwards) {
    Vector oldForwards = (Vector) forwards.get(x.getCoverId());
    /*
        if ( (oldForwards == null || oldForwards.size() == 0) &&
            (newForwards.size() == 0)) {
          return false;
        }
     */
    if (oldForwards == null) {
      return true;
    }

    // if (oldForwards.equals(newForwards)) {
    if (oldForwards.containsAll(newForwards) &&
        newForwards.containsAll(oldForwards)) {
      return false;
    }
    return true;
  }

  public void computeForwards(IFBase x) {
    // The node was a duplicate and its color was removed
    // This may affect the forwards sets of covered nodes
    Vector newForwards = new Vector();
    computeForwardsAdd(x, neighbours,
                       newForwards);
    boolean b = checkChange(x, newForwards);

    if (b) {
      Vector oldForwards = (Vector) forwards.get(x.getCoverId());

      forwards.put(x.getCoverId(), newForwards.clone());

      changeSet.put(x, newForwards);

      if (oldForwards != null) {
        newForwards.removeAll(oldForwards); // have to remove those that have already received
      }

      total_filters += (newForwards.size());
      if (debug) {
        System.out.println("computeForwards: " + newForwards.size());
      }

      // We have to check the direct successors
      Vector post = poset.getSuccessor(x);

      if (poset.getPre(x.getCoverId()) == null ||
          poset.getPre(x.getCoverId()).size() == 0) {
        updateSuccessors(post, true, true, null);
      }
      else {
        refreshSuccessors(post);
      }
    }
  }

  HashSet visitedList = new HashSet();

  public void updateSuccessors(Vector post, boolean isDefinitelyRoot,
                               boolean wasAdd, IFBase deleted) {

    Vector iter = (Vector) post.clone();
    while (!iter.isEmpty()) {
      IFBase y = (IFBase) iter.remove(0);

      if (visitedList.contains(y)) {
        continue;
      }

      visitedList.add(y);

      Vector newForwards2 = new Vector();

      computeForwardsAdd(y, neighbours,
                         newForwards2);

      boolean b = checkChange(y, newForwards2);

      if (debug) {
        System.out.println("b:" + b + " fw:" + newForwards2);
      }

      if (b && newForwards2.size() >= newForwards2.size() &&
          newForwards2.size() > 0) {

        if (!wasAdd) { // compute intersection
          // forwards set has changed
          changeSet.put(y, newForwards2);

          if (deleted != null) {
            // Only forward where the unsub was sent
            Vector temp = (Vector) newForwards2.clone();
            Vector fw = new Vector();
            Vector temp2 = (Vector) changeSet.get(deleted);

            // take the intersection of the unsub set and this fw set
            for (int k = 0; k < newForwards2.size(); k++) {
              Object col = newForwards2.elementAt(k);
              if (temp2.contains(col)) {
                fw.add(col);
              }
            }
            total_filters += (fw.size());
            if (debug) {
              System.out.println("Poset forward to [deleted]:" + fw + " id " +
                                 y.hashCode());
            }
          }
          else {
            total_filters += (newForwards2.size());
            if (debug) {
              System.out.println("Poset forward to [no-delete]:" + newForwards2);
            }
          }
        }

        forwards.put(y.getCoverId(), newForwards2);

        // for del the successors may have become root nodes..
        if (!isDefinitelyRoot && poset.isRootNode(y)) { // called after del so we can test rootness
          // The second level must be checked also!
          iter.addAll( (Vector) poset.getSuccessor(y.getCoverId()));
          /*
            if (debug) System.out.println("Recursive call to successors");

            updateSuccessors(poset.getSuccessor(y.getCoverId()), true, wasAdd,
                             deleted);
           */
        }
        else {
          if (newForwards2.size() > 1) {
            System.out.println(
                "Fatal error: updateSuccessors, forwards set > 1 and was not a root.");
            System.out.println("Root: " + poset.isRootNode(y));
            System.out.println("Forwards set:" + newForwards2);
            System.out.println("Tid: " + y.getTid());
            System.out.println("** running computation using debug mode");
            debug = true;
            computeForwardsAdd(y, neighbours,
                               newForwards2);

            System.exit(0);
          }
          // Update successors --> remove forwards set
          Vector succ = poset.getSuccessor(y);
          refreshSuccessors(succ);
        }
      }
      /*
             else {
        // Update successors --> remove forwards set
        Vector succ = poset.getSuccessor(y);
        refreshSuccessors(succ);
             }
       */
    }
  }

  public void refreshSuccessors(Vector v) {
    if (v != null) {
      for (int k = 0; k < v.size(); k++) {
        IFBase z = (IFBase) v.elementAt(k);
        forwards.remove(z.getCoverId());
      }
    }
  }

  /**
   * Computes the forwards set for the added filter. Called after add.
   * @param f IFBase
   * @param neighbours Vector
   * @param forwards Vector
   * @return boolean
   */
  public boolean computeForwardsAdd(IFBase f, Vector neighbours,
                                    Vector forw) {

    if (f == null || forw == null || neighbours == null) {
      return false;
    }

    // Get predecessors
    Vector a = poset.getPre(f.getCoverId());

    if (a == null || a.size() == 0) {
      Vector b = new Vector();
      ForwardsSet.computeForwards(f, b, neighbours);
      forw.addAll(b);
    }
    else {
      // Not a ROOT-node
      Vector root = new Vector();
      if (findRootForForwards(f, root)) {
        // at least 2 cover nodes, no forwards set in this case
        forw.clear();
        return true;
      }

      if (root == null) {
        System.out.println("computeForwardsAdd: fatal error. Root was null");
        return false;
      }
      HashSet unionF = new HashSet();
      recursiveAdder(f, unionF);

      forw.clear();
      forw.addAll(neighbours);
      forw.removeAll(unionF);
    }
    // Then we iterate down the tree and calculate the forwards sets
    return true;
  }

  public void recursiveAdder(IFBase x, HashSet set) {
    Vector pre = poset.getPre(x.getCoverId());
    if (pre != null && pre.size() > 0) {
      for (int i = 0; i < pre.size(); i++) {
        IFBase y = (IFBase) pre.elementAt(i);
        Vector fset = (Vector) forwards.get(y.getCoverId());

        // must have a forwards set, if does not have it must be recomputed
        // must be a root node, so it is easy to recompute
        if (fset == null) {
          if (!poset.isRootNode(y)) {
            System.out.println("RecursiveAdder: fatal node not root");
            System.exit(0);
          }
          fset = new Vector();
          ForwardsSet.computeForwards(y, fset, neighbours);
        }
        if (fset != null) {
          set.addAll(fset);
        }
      }
    }
    if (set.size() == 0) {
      System.out.println("Missing forwards set in recursiveAdder!");
      System.out.println("Pre set size:" + pre.size());
      IFBase y = (IFBase) pre.elementAt(0);
      Vector fset = (Vector) forwards.get(y.getCoverId());
      System.out.println("forwards: " + fset);
      ForwardsSet.computeForwards(y, fset, neighbours);
      System.out.println("forwards computed: " + fset);
      System.exit(0);
    }
  }

  /**
   * Removes non 1L and 2L nodes from the forwards set.
   */
  public void resetForwards() {
    Enumeration e = forwards.keys();
    while (e.hasMoreElements()) {
      String f = (String) e.nextElement();
      IFBase k = (IFBase) poset.templates.get(f);
      if (k == null || findRootForForwards(k, null)) {
        forwards.remove(f);
      }
    }
  }

  /**
   * Tests the path to root
   * @param count int
   * @param root Vector
   * @return true if the node is covered by at least 2 parent nodes. In this
   * case the forwards set is zero sized.
   */
  public boolean findRootForForwards(IFBase f, Vector root) {
    Vector a = poset.getPre(f.getCoverId());
    for (int i = 0; i < a.size(); i++) {
      IFBase x = (IFBase) a.elementAt(i);
      Vector p = poset.getPre(x.getCoverId());
      if (p == null || p.size() == 0) {
        if (root != null) {
          root.add(x);
        }
      }
      else {
        return true;
      }
    }
    return false;
  }

  /**
   * Called when the data structure changes
   * @param event String
   * @param x IFBase
   */
  public void eventOccurred(String event, IFBase x) {
    visitedList.clear();
    if (debug) {
      System.out.println("Poset:" + event);
    }

    if (event.equals(FilterForest.DELETE_EVENT_ELIMINATE)) {
      // Eliminate color was used to remove the node
      Vector ft = (Vector) forwards.remove(x.getCoverId());
      if (ft != null) {
        ft.clear();
      }
    }
    // duplicate filter removed
    else if (event.equals(FilterForest.DELETE_EVENT)) {
      // In order for the forwards set to change, the node must
      // be a root node or a second level node
      Vector uncovered = new Vector();
      computeForwardsDel(x, neighbours, uncovered, true);
      if (uncovered.size() > 0) {
        sortSuccessorsByRoot(uncovered);
        updateSuccessors(uncovered, false, false, x);
      }
    }
    else if (event.equals(FilterForest.DELETE_EVENT_BEFORE)) {
      if (debug) {
        System.out.println("Before  del -----------------------");
      }
      if (x.getTid() == null || x.getTid().size() == 0) {
        System.out.println(
            "Fatal in PosetForwardsSet: filter did not have tid");
        System.exit(0);
      }

      Vector uncovered = new Vector();
      computeForwardsDel(x, neighbours, uncovered, false);
      // store the set for later use
      tempDelStorage.put(x, uncovered);
    }
    else if (event.equals(FilterForest.DELETE_EVENT_AFTER)) {
      if (debug) {
        System.out.println("After del -----------------------");
      }
      Vector uncovered = (Vector) tempDelStorage.remove(x);
      if (uncovered == null) {
        if (debug) {
          System.out.println(
              "PosetForwardsSet: failure uncovered set missing in del");
        }
      }
      else if (uncovered.size() > 0) {
        // Update successor set
        // First we process new root nodes
        sortSuccessorsByRoot(uncovered);
        updateSuccessors(uncovered, false, false, x);
      }
      forwards.remove(x.getCoverId());
    }
    else if (event.equals(FilterForest.ADD_EVENT)) {
      // possibly a new entry in a forwards set (only for a root node)
      computeForwards(x);
    }
    else if (event.equals(FilterForest.RESET)) {
      forwards.clear();
      changeSet.clear();
      tempDelStorage.clear();
      resetTotal();
    }

  }

  public void sortSuccessorsByRoot(Vector a) {
    if (a == null || a.size() == 0) {
      return;
    }
    Vector pre = new Vector();
    Vector post = new Vector();
    for (int i = 0; i < a.size(); i++) {
      IFBase x = (IFBase) a.elementAt(i);
      Vector z = poset.getPre(x);
      if (z == null || z.size() == 0) {
        pre.add(x);
      }
      else {
        post.add(x);
      }
    }
    a.clear();
    a.addAll(pre);
    a.addAll(post);
  }

  /**
   * Computes the forwards set for the given filter and the set of uncovered
   * filters. Used when a filter is removed. Called BEFORE removing the filter.
   * The uncovered set contains filters that need to be re-established.
   * @param f IFBase
   * @param neighbours Vector
   * @param forwards Vector
   * @param uncovered Vector
   * @return boolean
   */
  public boolean computeForwardsDel(IFBase f, Vector
                                    neighbours,
                                    Vector uncovered, boolean realUpdate) {

    if (f == null || forwards == null || neighbours == null) {
      return false;
    }

    Vector newForwards = new Vector();
    Vector oldForwards = (Vector) forwards.get(f.getCoverId());

    computeForwardsAdd(f, neighbours,
                       newForwards);

    if (newForwards == null || newForwards.size() == 0) {
      return true;
    }

    if (realUpdate) {
      forwards.put(f.getCoverId(), (newForwards.clone()));
      Vector temp = (Vector) newForwards.clone();
      newForwards.clear();
      newForwards.addAll(oldForwards);
      newForwards.removeAll(temp);
    }

    changeSet.put(f, newForwards);
    total_filters += (newForwards.size());
    if (debug) {
      System.out.println("ComputeForwardsDel:" + newForwards.size());
    }

    // Now we compute the set of uncovered filters
    uncovered.clear();
    uncovered.addAll(poset.getSuccessor(f.getCoverId()));

    return true;
  }

}
