/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.filter.predicates;

import java.util.*;

import fuegocore.notify.core.filter.*;

/**
 * This class implements the boolean predicate.
 */
public class BooleanPredicate implements IAtomicPredicate {
    private String op;
    private Date dt;
    private double dval;
    private String val;

    public BooleanPredicate(String o, double v) {
        op = o;
        dval = v;
    }

    public BooleanPredicate(String o, String v) {
        op = o;
        val = v;
    }

    public BooleanPredicate(String o, Date v) {
        op = o;
        dt = v;
    }

    /** Returns id for this object
     * @return String the identifier
     */
    public String getId() {
        if (val != null) {
            return op + "_" + val;
        } else if (dt != null) {
            return op + "_" + dt.getTime();
        } else {
            return op + "_" + dval;
        }
    }

    /** Overlapping relation. Returns true if this object overlaps with the given object.
     * @param de the input object
     * @return true if the input object overlaps with the current object
     */
    public boolean overlaps(IAtomicPredicate de) {
        if (getOperand().equals(Filter.SELECT) &&
            de.getOperand().equals(Filter.SELECT) &&
            getValue().trim().equals(de.getValue().trim())) {
            return true;
        }

        // If either value has a dollar sign (placeholder)
        if (getValue() != null && (getValue().startsWith("$") ||
                                   de.getValue().startsWith("$"))) {
            if (getValue().equals(de.getValue()) &&
                getOperand().equals(de.getOperand())) {
                return true;
            }
        } else if (getOperand().equals(Filter.EQUAL)) {
            if (de.getOperand().equals(Filter.EQUAL)) {
                if (getValue().trim().equals(de.getValue().trim())) {
                    return true;
                }
            }
        } else if (getOperand().equals(Filter.NOT_EQUAL)) {
            if (de.getOperand().equals(Filter.NOT_EQUAL)) {
                if (getValue().trim().equals(de.getValue().trim())) {
                    return true;
                }
            }
        } else if (getOperand() == Filter.EXISTENCE) {
            // existence covers all
            return true;
        }

        return false;
    }

    /**
     *  Returns the value of an equality predicate (or null)
     * @param type int
     * @return String  the value of the equality predicate
     */
    public String getEqualityValue(int type) {
        if (getOperand().equals(Filter.EQUAL)) {
            return "" + getValue().trim();
        }
        return null;
    }

    /** Covering relation. Returns true if this object covers the given object.
     * @param de the input object
     * @return true if the input object is covered
     */
    public boolean covers(IAtomicPredicate de) {
        if (getOperand().equals(Filter.SELECT) &&
            de.getOperand().equals(Filter.SELECT) &&
            getValue().trim().equals(de.getValue().trim())) {
            return true;
        } else if (getOperand().equals(Filter.EQUAL)) {
            if (de.getOperand().equals(Filter.EQUAL)) {
                if (getValue().trim().equals(de.getValue().trim())) {
                    return true;
                }
            }
        } else if (getOperand().equals(Filter.NOT_EQUAL)) {
            if (de.getOperand().equals(Filter.NOT_EQUAL)) {
                if (getValue().trim().equals(de.getValue().trim())) {
                    return true;
                }
            }
        } else if (getOperand() == Filter.EXISTENCE) {
            // existence covers all
            return true;
        }

        return false;
    }

    /** Tests whether the given predicate a merges with b and results
     *  in an existence test. Existence test is the most general filter
     *  and if one is detected, we can break the merging process for
     *  this field. Existence test covers all other predicates.
     * @param b the input object
     * @return true if the merge results in an existence test
     */
    public boolean perfectMergeRuleForExists(IAtomicPredicate b) {

        return false;
    }

    /** Tests whether the given predicate a merges with b and results
     *  in a modified predicate.
     * @param b the input object
     * @param change true if changes should be made
     * @return true if the merge results in an existence test
     */
    public boolean perfectMergeRuleForModified(IAtomicPredicate b,
                                               boolean change) {
        return false;
    }

    /** Matching functionality
     * @param se the first input value
     * @param a the second input value
     * @return true if the values match the condition
     */
    public boolean match(double se, double a) {
        return false;
    }

    /** Matching functionality
     * @param a the first input value
     * @param b the second input value
     * @return true if the values match the condition
     */
    public boolean match(String a, String b) {
        if (getOperand().equals(Filter.EQUAL)) {
            if (a.equals(b)) {
                return true;
            }
        } else if (getOperand().equals(Filter.NOT_EQUAL)) {
            if (!a.equals(b)) {
                return true;
            }
        } else if (getOperand().equals(Filter.EXISTENCE)) {
            // existence covers all
            return true;
        }
        return false;
    }

    /** Matching functionality
     * @param a the first input value
     * @param b the second input value
     * @return true if the values match the condition
     */
    public boolean match(Date a, Date b) {
        return false;
    }

    /** Tests for equivalence of atomic predicates
     * @param ip the input object
     * @return true if the two objects are identical
     */
    public boolean identical(IAtomicPredicate ip) {
        if (ip == null) {
            return false;
        }

        if (getOperand().equals(ip.getOperand())) {
            if (getValue() == null || ip.getValue() == null) {
                return false;
            }
            if (getValue() != null &&
                (getValue().startsWith("$") || ip.getValue().startsWith("$"))) {
                if (getValue().equals(ip.getValue())) {
                    return true;
                }
            }
            if (getValue().equals(ip.getValue())) {
                return true;
            }
        }
        return false;
    }

    /** Tests for the equivalence of the value part of two predicates
     * @param ip the input object
     * @return true if the two objects are identical
     */
    public boolean identicalValue(IAtomicPredicate ip) {
        if (ip == null) {
            return false;
        }
        if (getValue() != null &&
            (getValue().startsWith("$") || ip.getValue().startsWith("$"))) {
            if (getValue().equals(ip.getValue())) {
                return true;
            }
        }
        if (getValue().equals(ip.getValue())) {
            return true;
        }
        return false;
    }

    /** Get operand
     * @return String the operand
     */
    public String getOperand() {
        return op;
    }

    /** Get value
     * @return String the value
     */
    public String getValue() {
        return val;
    }

    /** Get double value
     * @return double the value
     */
    public double getDValue() {
        return dval;
    }

    /** Get second double value
     * @return double the value
     */
    public double getDValue2() {
        return 0;
    }

    /** Get date value
     * @return Date the date value
     */
    public Date getDateValue() {
        return dt;
    }

    /** Get time value
     * @return long the time value
     */
    public long getTimeValue() {
        return dt.getTime();
    }

    /** Returns the cost/generality of the predicate
     *  @return int the cost
     */
    public int getCost() {
        return 0;
    }

    /** Return the value in string format
     * @return String the String representation of the object
     */
    public String toString() {
        return op + Filter.OPERAND_DIVISOR + val;
    }
}
