/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.filter.predicates;

import java.util.*;

import fuegocore.notify.core.filter.*;

/**
 * This class implements the date predicate.
 */
public class DatePredicate implements IAtomicPredicate {
    private String op;
    private Date dt;
    private double dval;
    private String val;

    public DatePredicate(String o, double v) {
        op = o.trim();
        dval = v;
    }

    public DatePredicate(String o, String v) {
        op = o.trim();
        val = v.trim();
    }

    public DatePredicate(String o, Date v) {
        op = o.trim();
        dt = v;
    }

    /** Returns id for this object
     * @return String the identifier
     */
    public String getId() {
        if (val != null) {
            return op + "_" + val;
        } else if (dt != null) {
            return op + "_" + dt.getTime();
        } else {
            return op + "_" + dval;
        }
    }

    /**
     *  Returns the value of an equality predicate (or null)
     * @param type int
     * @return String  the value of the equality predicate
     */
    public String getEqualityValue(int type) {
        if (getOperand().equals(Filter.OP_EQU)) {
            return ("" + getTimeValue()).trim();
        }
        return null;
    }

    /** Covering relation. Returns true if this object covers the given object.
     * @param de the input object
     * @return true if the input object is covered
     */
    public boolean covers(IAtomicPredicate de) {

        if (getDateValue() == null || de.getDateValue() == null) {
            return false;
        }

        if (getOperand().equals(Filter.SELECT) &&
            de.getOperand().equals(Filter.SELECT) &&
            getValue().equals(de.getValue())) {
            return true;
        } else if (getOperand().equals(Filter.OP_EQU)) {
            if (de.getOperand().equals(Filter.OP_EQU)) {
                if (getTimeValue() == de.getTimeValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_NEQ)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.OP_GR)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.OP_LT)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.OP_LTE)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.OP_GTE)) {
                // does not cover
            }
        } else if (getOperand().equals(Filter.OP_NEQ)) {
            if (de.getOperand().equals(Filter.OP_EQU)) {
                if (getTimeValue() != de.getTimeValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_NEQ)) {
                if (getTimeValue() == de.getTimeValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_GR)) {
                if (getTimeValue() <= de.getTimeValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LT)) {
                if (getTimeValue() >= de.getTimeValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LTE)) {
                if (getTimeValue() > de.getTimeValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_GTE)) {
                if (getTimeValue() < de.getTimeValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                return true;
            }
        } else if (getOperand().equals(Filter.OP_GR)) {
            if (de.getOperand().equals(Filter.OP_EQU)) {
                // If either value has a dollar sign (placeholder)
                // they must match exactly
                if ((getValue().startsWith("$") ||
                     de.getValue().startsWith("$"))) {
                    if (getValue().equals(de.getValue())) {
                        return true;
                    } else {
                        return false;
                    }
                }

                if (de.getTimeValue() > getTimeValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_NEQ)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.OP_GR)) {
                // If either value has a dollar sign (placeholder)
                // they must match exactly
                if ((getValue().startsWith("$") ||
                     de.getValue().startsWith("$"))) {
                    if (getValue().equals(de.getValue())) {
                        return true;
                    } else {
                        return false;
                    }
                }
                if (de.getTimeValue() >= getTimeValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LT)) {
            } else if (de.getOperand().equals(Filter.OP_LTE)) {
            } else if (de.getOperand().equals(Filter.OP_GTE)) {
                // If either value has a dollar sign (placeholder)
                // they must match exactly
                if ((getValue().startsWith("$") ||
                     de.getValue().startsWith("$"))) {
                    if (getValue().equals(de.getValue())) {
                        return true;
                    } else {
                        return false;
                    }
                }
                if (de.getTimeValue() > getTimeValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                return true;
            }

        } else if (getOperand().equals(Filter.OP_LT)) {
            if (de.getOperand().equals(Filter.OP_EQU)) {
                // If either value has a dollar sign (placeholder)
                // they must match exactly
                if ((getValue().startsWith("$") ||
                     de.getValue().startsWith("$"))) {
                    if (getValue().equals(de.getValue())) {
                        return true;
                    } else {
                        return false;
                    }
                }

                if (de.getTimeValue() < getTimeValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_NEQ)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.OP_GR)) {
            } else if (de.getOperand().equals(Filter.OP_LT)) {
                // If either value has a dollar sign (placeholder)
                // they must match exactly
                if ((getValue().startsWith("$") ||
                     de.getValue().startsWith("$"))) {
                    if (getValue().equals(de.getValue())) {
                        return true;
                    } else {
                        return false;
                    }
                }
                if (de.getTimeValue() <= getTimeValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LTE)) {
                // If either value has a dollar sign (placeholder)
                // they must match exactly
                if ((getValue().startsWith("$") ||
                     de.getValue().startsWith("$"))) {
                    if (getValue().equals(de.getValue())) {
                        return true;
                    } else {
                        return false;
                    }
                }
                if (de.getTimeValue() < getTimeValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_GTE)) {
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                return true;
            }
        } else if (getOperand().equals(Filter.OP_LTE)) {
            if (de.getOperand().equals(Filter.OP_EQU)) {
                // If either value has a dollar sign (placeholder)
                // they must match exactly
                if ((getValue().startsWith("$") ||
                     de.getValue().startsWith("$"))) {
                    if (getValue().equals(de.getValue())) {
                        return true;
                    } else {
                        return false;
                    }
                }

                if (de.getTimeValue() <= getTimeValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_NEQ)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.OP_GR)) {
            } else if (de.getOperand().equals(Filter.OP_LT)) {
                // If either value has a dollar sign (placeholder)
                // they must match exactly
                if ((getValue().startsWith("$") ||
                     de.getValue().startsWith("$"))) {
                    if (getValue().equals(de.getValue())) {
                        return true;
                    } else {
                        return false;
                    }
                }
                if (de.getTimeValue() <= getTimeValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LTE)) {
                // If either value has a dollar sign (placeholder)
                // they must match exactly
                if ((getValue().startsWith("$") ||
                     de.getValue().startsWith("$"))) {
                    if (getValue().equals(de.getValue())) {
                        return true;
                    } else {
                        return false;
                    }
                }
                if (de.getTimeValue() <= getTimeValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_GTE)) {
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                return true;
            }
        } else if (getOperand().equals(Filter.OP_GTE)) {
            if (de.getOperand().equals(Filter.OP_EQU)) {
                // If either value has a dollar sign (placeholder)
                // they must match exactly
                if ((getValue().startsWith("$") ||
                     de.getValue().startsWith("$"))) {
                    if (getValue().equals(de.getValue())) {
                        return true;
                    } else {
                        return false;
                    }
                }

                if (de.getTimeValue() >= getTimeValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_NEQ)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.OP_GR)) {
                // If either value has a dollar sign (placeholder)
                // they must match exactly
                if ((getValue().startsWith("$") ||
                     de.getValue().startsWith("$"))) {
                    if (getValue().equals(de.getValue())) {
                        return true;
                    } else {
                        return false;
                    }
                }
                if (de.getTimeValue() >= getTimeValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LT)) {
            } else if (de.getOperand().equals(Filter.OP_LTE)) {
            } else if (de.getOperand().equals(Filter.OP_GTE)) {
                // If either value has a dollar sign (placeholder)
                // they must match exactly
                if ((getValue().startsWith("$") ||
                     de.getValue().startsWith("$"))) {
                    if (getValue().equals(de.getValue())) {
                        return true;
                    } else {
                        return false;
                    }
                }
                if (de.getTimeValue() >= getTimeValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                return true;
            }
        } else if (getOperand().equals(Filter.EXISTENCE)) {
            // existence covers all
            return true;
        }
        return false;
    }

    /** Overlapping relation. Returns true if this object overlaps with the given object.
     * @param de the input object
     * @return true if the input object overlaps with the current object
     */
    public boolean overlaps(IAtomicPredicate de) {
        return false; // tbd
    }

    /** Tests whether the given predicate a merges with b and results
     *  in a modified predicate.
     * @param b the input object
     * @param change true if changes should be made
     * @return true if the merge results in an existence test
     */
    public boolean perfectMergeRuleForModified(IAtomicPredicate b,
                                               boolean change) {
        return false;
    }

    /** Tests whether the given predicate a merges with b and resuts
     *  in an existence test. Existence test is the most general filter
     *  and if one is detected, we can break the merging process for
     *  this field. Existence test covers all other predicates.
     * @param b the input object
     * @return true if the merge results in an existence test
     */
    public boolean perfectMergeRuleForExists(IAtomicPredicate b) {
        if (getValue() != null &&
            (getValue().startsWith("$") || b.getValue().startsWith("$"))
            &&
            !(getOperand().equals(Filter.OP_EQU) ||
              getOperand().equals(Filter.OP_NEQ))) {
            // the placeholder cannot be merged unless equ or neq is used, because
            // we do not know the value yet.
            return false;
        }

        if (getOperand().equals(Filter.OP_EQU) &&
            b.getOperand().equals(Filter.OP_NEQ) &&
            identicalValue(b)) {
            return true;
        } else if (b.getOperand().equals(Filter.OP_EQU) &&
                   getOperand().equals(Filter.OP_NEQ) &&
                   identicalValue(b)) {
            return true;
        } else if (getOperand().equals(Filter.OP_LT) &&
                   b.getOperand().equals(Filter.OP_GR) &&
                   getTimeValue() > b.getTimeValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_LT) &&
                   b.getOperand().equals(Filter.OP_GTE) &&
                   getTimeValue() >= b.getTimeValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_LTE) &&
                   b.getOperand().equals(Filter.OP_GR) &&
                   getTimeValue() >= b.getTimeValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_LTE) &&
                   b.getOperand().equals(Filter.OP_GTE) &&
                   getTimeValue() >= b.getTimeValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_GR) &&
                   b.getOperand().equals(Filter.OP_LT) &&
                   getTimeValue() < b.getTimeValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_GR) &&
                   b.getOperand().equals(Filter.OP_LTE) &&
                   getTimeValue() <= b.getTimeValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_GTE) &&
                   b.getOperand().equals(Filter.OP_LT) &&
                   getTimeValue() <= b.getTimeValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_GTE) &&
                   b.getOperand().equals(Filter.OP_LTE) &&
                   getTimeValue() <= b.getTimeValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_NEQ) &&
                   b.getOperand().equals(Filter.OP_NEQ) &&
                   getTimeValue() != b.getTimeValue()) {
            return true;
        }

        return false;
    }

    /** Matching functionality
     * @param se the first input value
     * @param a the second input value
     * @return true if the values match the condition
     */
    public boolean match(double se, double a) {

        if (getOperand().equals(Filter.OP_EQU)) {
            if (se != a) {
                return false;
            }
        } else if (getOperand().equals(Filter.OP_GR)) {
            if (a <= se) {
                return false;
            }
        } else if (getOperand().equals(Filter.OP_LT)) {
            if (a >= se) {
                return false;
            }
        } else if (getOperand().equals(Filter.OP_NEQ)) {
            if (a == se) {
                return false;
            }
        } else if (getOperand().equals(Filter.OP_LTE)) {
            if (a > se) {
                return false;
            }
        } else if (getOperand().equals(Filter.OP_GTE)) {
            if (a < se) {
                return false;
            }
        } else if (getOperand().equals(Filter.EXISTENCE)) {
            // true
        }
        return true;
    }

    /** Matching functionality
     * @param a the first input value
     * @param b the second input value
     * @return true if the values match the condition
     */
    public boolean match(String a, String b) {
        return false;
    }

    /** Matching functionality
     * @param dse the first input value
     * @param da the second input value
     * @return true if the match succeeded
     */
    public boolean match(Date dse, Date da) {
        long se, a;

        se = dse.getTime();
        a = da.getTime();

        if (getOperand().equals(Filter.OP_EQU)) {
            if (se != a) {
                return false;
            }
        } else if (getOperand().equals(Filter.OP_GR)) {
            if (a <= se) {
                return false;
            }
        } else if (getOperand().equals(Filter.OP_LT)) {
            if (a >= se) {
                return false;
            }
        } else if (getOperand().equals(Filter.OP_NEQ)) {
            if (a == se) {
                return false;
            }
        } else if (getOperand().equals(Filter.OP_LTE)) {
            if (a > se) {
                return false;
            }
        } else if (getOperand().equals(Filter.OP_GTE)) {
            if (a < se) {
                return false;
            }
        } else if (getOperand().equals(Filter.EXISTENCE)) {
            // true
        }
        return true;
    }

    /** Tests for equivalence of atomic predicates
     * @param ip the input object
     * @return true if the two objects are identical
     */
    public boolean identical(IAtomicPredicate ip) {
        if (ip == null) {
            return false;
        }

        if (getOperand().equals(ip.getOperand())) {
            if (getValue() != null &&
                (getValue().startsWith("$") || ip.getValue().startsWith("$"))) {
                if (getValue().equals(ip.getValue())) {
                    return true;
                }
            }
            if (getDateValue().equals(ip.getDateValue())) {
                return true;
            }
        }
        return false;
    }

    /** Tests for the equivalence of the value part of two predicates
     * @param ip the input object
     * @return true if the two objects are identical
     */
    public boolean identicalValue(IAtomicPredicate ip) {
        if (ip == null) {
            return false;
        }
        if (getValue() != null &&
            (getValue().startsWith("$") || ip.getValue().startsWith("$"))) {
            if (getValue().equals(ip.getValue())) {
                return true;
            }
        }
        if (getDateValue().equals(ip.getDateValue())) {
            return true;
        }
        return false;
    }

    /** Get operand
     * @return String the operand
     */
    public String getOperand() {
        return op;
    }

    /** Get value
     * @return String the value
     */
    public String getValue() {
        return val;
    }

    /** Get double value
     * @return double the double value
     */
    public double getDValue() {
        return dval;
    }

    /** Get second double value
     * @return double the value
     */
    public double getDValue2() {
        return 0;
    }

    /** Get date value
     * @return Date the date value
     */
    public Date getDateValue() {
        return dt;
    }

    /** Get time value
     * @return long the time value
     */
    public long getTimeValue() {
        return dt.getTime();
    }

    /** Returns the cost/generality of the predicate
     *  @return int the cost
     */
    public int getCost() {
        int cost = 0;
        if (getOperand().equals(Filter.OP_EQU)) {
            cost = 0; // restricting
        } else if (getOperand().equals(Filter.OP_NEQ)) {
            cost = 100; // less restricting
        } else if (getOperand().equals(Filter.OP_GR)) {
            cost = 50;
        } else if (getOperand().equals(Filter.OP_LT)) {
            cost = 50;
        } else if (getOperand().equals(Filter.OP_LTE)) {
            cost = 50;
        } else if (getOperand().equals(Filter.OP_GTE)) {
            cost = 50;
        }
//     else if (getOperand().equals(Filter.OP_RNG)) {
//       cost = 25;
//     }
        else if (getOperand().equals(Filter.EXISTENCE)) {
            cost = 100;
        }
        return cost;
    }

    /** Return the value in string format
     * @return String the String representation
     */
    public String toString() {
        if (dt == null) {
            return null;
        }
        return op + Filter.OPERAND_DIVISOR + dt.getTime();
    }

}
