/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.filter.predicates;

import java.util.*;

/**
 * Atomic Predicate Interface. This supports the definition of
 * different types in filtering.
 */
public interface IAtomicPredicate {

    /** Returns id for this object
     * @return String the identifier
     */
    public String getId();

    /** Returns the cost/generality of the predicate
     *  @return int the cost
     */
    public int getCost();

    /** Covering relation. Returns true if this object covers the given object.
     * @param de the input object
     * @return true if the input object is covered
     */
    public boolean covers(IAtomicPredicate de);

    /** Overlapping relation. Returns true if this object overlaps with the given object.
     * @param de the input object
     * @return true if the input object overlaps with the current object
     */
    public boolean overlaps(IAtomicPredicate de);

    /** Tests whether the given predicate a merges with b and results
     *  in an existence test. Existence test is the most general filter
     *  and if one is detected, we can break the merging process for
     *  this field. Existence test covers all other predicates.
     * @param b the input object
     * @return true if the merge results in an existence test
     */
    public boolean perfectMergeRuleForExists(IAtomicPredicate b);

    /** Tests whether the given predicate a merges with b and results
     *  in a modified predicate.
     * @param b the input object
     * @param change true if changes should be made
     * @return true if the merge results in an existence test
     */
    public boolean perfectMergeRuleForModified(IAtomicPredicate b,
                                               boolean change);

    /** Returns true if the two values match given the predicate condition,
     * both values are given as parameters because the original value may
     * be a placeholder.
     * @param v the first input value
     * @param v2 the second input value
     * @return true if the values match the condition
     */
    public boolean match(String v, String v2);

    /** Returns true if the two values match given the predicate condition,
     * both values are given as parameters because the original value may
     * be a placeholder
     * @param v the first input value
     * @param v2 the second input value
     * @return true if the values match the condition
     */
    public boolean match(double v, double v2);

    /** Returns true if the two values match given the predicate condition,
     * both values are given as parameters because the original value may
     * be a placeholder
     * @param v the first input value
     * @param v2 the second input value
     * @return true if the values match the condition
     */
    public boolean match(Date v, Date v2);

    /** Tests for equivalence of atomic predicates
     * @param ip the input object
     * @return true if the two objects are identical
     */
    public boolean identical(IAtomicPredicate ip);

    /** Tests for the equivalence of the value part of two predicates
     * @param ip the input object
     * @return true if the two objects are identical
     */
    public boolean identicalValue(IAtomicPredicate ip);

    /** Get predicate operand
     * @return String the operand
     */
    public String getOperand();

    /** Get predicate value (String)
     * @return String the value
     */
    public String getValue();

    /** Get predicate double value
     * @return double the double value
     */
    public double getDValue();

    /** Get second predicate double value
     * @return double the double value
     */
    public double getDValue2();

    /** Get predicate date value
     * @return Date the date value
     */
    public Date getDateValue();

    /** Get predicate time value (long)
     * @return long the time value
     */
    public long getTimeValue();

    /* Returns string representation of the predicate
     * @return String the String representation
     */
    public String toString();

    /**
     *  Returns the value of an equality predicate (or null)
     * @param type int
     * @return String  the value of the equality predicate
     */
    public String getEqualityValue(int type);
}
