/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */


package fuegocore.notify.core.filter.predicates;

import java.util.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;

/**
 * This class implements the number predicate.
 */
public class NumberPredicate implements IAtomicPredicate {
    private String op;
    private Date dt;
    private double dval;
    private double dval2;
    private String val;

    public NumberPredicate(String o, double v) {
        op = o.trim();
        dval = v;
    }

    public NumberPredicate(String o, double v, double v2) {
        op = o.trim();
        dval = v;
        dval2 = v2;
    }

    public NumberPredicate(String o, String v) {
        op = o.trim();
        val = v.trim();
    }

    public NumberPredicate(String o, Date v) {
        op = o.trim();
        dt = v;
    }

    /** Returns id for this object
     * @return String the identifier
     */
    public String getId() {
        if (val != null) {
            return op + "_" + val;
        } else if (dt != null) {
            return op + "_" + dt.getTime();
        } else {
            if (op.indexOf(Filter.OP_RNG) != -1) {
                return op + "_" + dval + "_" + dval2;
            } else {
                return op + "_" + dval;
            }
        }
    }

    /**
     *  Returns the value of an equality predicate (or null)
     * @param type int
     * @return String  the value of the equality predicate
     */
    public String getEqualityValue(int type) {
        if (getOperand().equals(Filter.OP_EQU)) {
            if (type == EventTriplet.TYPE_DOUBLE) {
                return ("" + getDValue()).trim();
            } else {
                return ("" + (int) (getDValue())).trim();
            }
        }
        return null;
    }

    /** Overlapping relation. Returns true if this object overlaps with the given object.
     * @param de the input object
     * @return true if the input object overlaps with the current object
     */
    public boolean overlaps(IAtomicPredicate de) {

        if (getOperand().equals(Filter.SELECT) &&
            de.getOperand().equals(Filter.SELECT) &&
            getValue().equals(de.getValue())) {
            return true;
        }

        // If either value has a dollar sign (placeholder)
        if (getValue() != null && (getValue().startsWith("$") ||
                                   de.getValue().startsWith("$"))) {
            if (getValue().equals(de.getValue()) &&
                getOperand().equals(de.getOperand())) {
                // Identical values (must be for placeholders)
                if (de.getOperand().equals(getOperand())) {
                    return true;
                }
                /*
                        Overlap (symmetric)
                         op,op
                         >=,>
                         <=,<
                         >=,<=
                 */
                if ((de.getOperand().equals(Filter.OP_GTE) &&
                     getOperand().equals(Filter.OP_GR)) ||
                    (de.getOperand().equals(Filter.OP_GR) &&
                     getOperand().equals(Filter.OP_GTE)) ||
                    (de.getOperand().equals(Filter.OP_LTE) &&
                     getOperand().equals(Filter.OP_LT)) ||
                    (de.getOperand().equals(Filter.OP_LT) &&
                     getOperand().equals(Filter.OP_LTE)) ||
                    (de.getOperand().equals(Filter.OP_LTE) &&
                     getOperand().equals(Filter.OP_GTE)) ||
                    (de.getOperand().equals(Filter.OP_GTE) &&
                     getOperand().equals(Filter.OP_LTE))) {
                    return true;
                }
            }
            // otherwise we cannot compare
            return false;
        }

        /* ***************************************************** */
        // Overlapping is a symmetric relation
        if (getOperand().equals(Filter.OP_EQU)) {
            if (de.getOperand().equals(Filter.OP_EQU)) {
                if (getDValue() == de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_NEQ)) {
                if (getDValue() != de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_GR)) {
                if (getDValue() > de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LT)) {
                if (getDValue() < de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LTE)) {
                if (getDValue() <= de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_GTE)) {
                if (getDValue() >= de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_RNG)) {
                if (getDValue() >= de.getDValue() &&
                    getDValue() <= de.getDValue2()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                return true;
            }
        }
        /* ***************************************************** */
        else if (getOperand().equals(Filter.OP_NEQ)) {
            if (de.getOperand().equals(Filter.OP_EQU)) {
                if (getDValue() != de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_NEQ)) {
                // always overlaps
                return true;
            } else if (de.getOperand().equals(Filter.OP_GR)) {
                return true;
            } else if (de.getOperand().equals(Filter.OP_LT)) {
                return true;
            } else if (de.getOperand().equals(Filter.OP_LTE)) {
                return true;
            } else if (de.getOperand().equals(Filter.OP_GTE)) {
                return true;
            } else if (de.getOperand().equals(Filter.OP_RNG)) {
                // if range is not the point then there is overlap
                if (getDValue() == de.getDValue() &&
                    de.getDValue2() == de.getDValue()) {
                    return false; //special case
                } else {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                return true;
            }
        }
        /* ***************************************************** */
        else if (getOperand().equals(Filter.OP_GR)) {
            if (de.getOperand().equals(Filter.OP_EQU)) {
                if (de.getDValue() > getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_NEQ)) {
                return true;
            } else if (de.getOperand().equals(Filter.OP_GR)) {
                // > x, >y overlaps always
                return true;
            } else if (de.getOperand().equals(Filter.OP_LT)) {
                // > x, < y, overlaps iff y > x
                if (de.getDValue() > getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LTE)) {
                // > x, < y, overlaps iff y > x
                if (de.getDValue() > getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_GTE)) {
                return true;
            } else if (de.getOperand().equals(Filter.OP_RNG)) {
                // > x, [y,z] overlaps when z > x
                if (de.getDValue2() > getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                // does not cover
                return true;
            }

        }
        /* ***************************************************** */
        else if (getOperand().equals(Filter.OP_LT)) {
            if (de.getOperand().equals(Filter.OP_EQU)) {
                if (de.getDValue() < getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_NEQ)) {
                return true;
            } else if (de.getOperand().equals(Filter.OP_GR)) {
                // < x, > y overlaps if y < x
                if (de.getDValue() < getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LT)) {
                return true;
            } else if (de.getOperand().equals(Filter.OP_LTE)) {
                return true;
            } else if (de.getOperand().equals(Filter.OP_GTE)) {
                // < x, > y overlaps if y < x
                if (de.getDValue() < getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_RNG)) {
                // < x, [y,z] overlaps if y < x
                if (de.getDValue() < getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                return true;
            }
        }
        /* ***************************************************** */
        else if (getOperand().equals(Filter.OP_LTE)) {
            if (de.getOperand().equals(Filter.OP_EQU)) {
                if (de.getDValue() <= getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_NEQ)) {
                return true;
            } else if (de.getOperand().equals(Filter.OP_GR)) {
                // <= x, > y overlaps if y < x
                if (de.getDValue() < getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LT)) {
                return true;
            } else if (de.getOperand().equals(Filter.OP_LTE)) {
                return true;
            } else if (de.getOperand().equals(Filter.OP_GTE)) {
                // <= x, >= y overlaps if y <= x
                if (de.getDValue() <= getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_RNG)) {
                // <= x, [y,z] overlaps if y <= x
                if (de.getDValue() <= getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                return true;
            }
        }
        /* ***************************************************** */
        else if (getOperand().equals(Filter.OP_GTE)) {
            if (de.getOperand().equals(Filter.OP_EQU)) {
                // equ 10 gr 9 10>=9
                if (de.getDValue() >= getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_NEQ)) {
                return true;
            } else if (de.getOperand().equals(Filter.OP_GR)) {
                // > x, >y overlaps always
                return true;
            } else if (de.getOperand().equals(Filter.OP_LT)) {
                // >= x, < y, overlaps iff y > x
                if (de.getDValue() > getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LTE)) {
                // >= x, <= y, overlaps iff y >= x
                if (de.getDValue() >= getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_GTE)) {
                return true;
            } else if (de.getOperand().equals(Filter.OP_RNG)) {
                // >= x, [y,z] overlaps when z >= x
                if (de.getDValue2() >= de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                return true;
            }

        }
        /* ***************************************************** */
        else if (getOperand().equals(Filter.OP_RNG)) {
            if (de.getOperand().equals(Filter.OP_EQU)) {
                if (de.getDValue() >= getDValue() &&
                    de.getDValue() <= getDValue2()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_NEQ)) {
                if (getDValue() == de.getDValue() && getDValue2() == getDValue()) {
                    return false; //special case
                }
                return true;
            } else if (de.getOperand().equals(Filter.OP_GR)) {
                // [x,y] >z overlaps if y > z
                if (getDValue2() > de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LT)) {
                // [x,y] <z overlaps if x < z
                if (getDValue() < de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LTE)) {
                // [x,y] <=z overlaps if x <= z
                if (getDValue() <= de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_GTE)) {
                // [x,y] >=z overlaps if y >= z
                if (getDValue2() >= de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_RNG)) {
                if ((getDValue2() >= de.getDValue() &&
                     de.getDValue2() >= getDValue()) ||
                    (de.getDValue2() >= getDValue() &&
                     getDValue2() >= de.getDValue())) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                return true;
            }
        } else if (getOperand().equals(Filter.EXISTENCE)) {
            return true;
        }

        return false;
    }

    /** Covering relation. Returns true if this object covers the given object.
     *  @param de the input object
     *  @return true if the input object is covered
     */
    public boolean covers(IAtomicPredicate de) {

        if (getOperand().equals(Filter.SELECT) &&
            de.getOperand().equals(Filter.SELECT) &&
            getValue().equals(de.getValue())) {
            return true;
        }

        boolean placeholder = false;

        // If either value has a dollar sign (placeholder)
        if (getValue() != null && (getValue().startsWith("$") ||
                                   de.getValue().startsWith("$"))) {
            if (getValue().equals(de.getValue())) {
                // Identical values (must be for placeholders)
                if (de.getOperand().equals(getOperand())) {
                    return true;
                }
                /*
                         Cover
                         op,op
                         >=,>
                         <=,<
                 */
                if ((getOperand().equals(Filter.OP_GTE) &&
                     de.getOperand().equals(Filter.OP_GR)) ||
                    (getOperand().equals(Filter.OP_LTE) &&
                     de.getOperand().equals(Filter.OP_LT))) {
                    return true;
                }
            }
            // otherwise we cannot compare
            return false;
        }

        /* ***************************************************** */
        if (getOperand().equals(Filter.OP_EQU)) {
            if (de.getOperand().equals(Filter.OP_EQU)) {
                if (getDValue() == de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_NEQ)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.OP_GR)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.OP_LT)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.OP_LTE)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.OP_GTE)) {
                // does not cover
            }
        }
        /* ***************************************************** */
        else if (getOperand().equals(Filter.OP_NEQ)) {
            if (de.getOperand().equals(Filter.OP_EQU)) {
                if (getDValue() != de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_NEQ)) {
                if (getDValue() == de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_GR)) {
                if (getDValue() <= de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LT)) {
                if (getDValue() >= de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LTE)) {
                if (getDValue() > de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_GTE)) {
                if (getDValue() < de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_RNG)) {
                if (!(getDValue() >= de.getDValue() &&
                      getDValue() <= de.getDValue2())) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                //       return true;
            }
        }
        /* ***************************************************** */
        else if (getOperand().equals(Filter.OP_GR)) {
            if (de.getOperand().equals(Filter.OP_EQU)) {
                if (de.getDValue() > getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_NEQ)) {
            } else if (de.getOperand().equals(Filter.OP_GR)) {
                if (de.getDValue() >= getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LT)) {
            } else if (de.getOperand().equals(Filter.OP_LTE)) {
            } else if (de.getOperand().equals(Filter.OP_GTE)) {
                if (de.getDValue() > getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_RNG)) {
                if (getDValue() < de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                //      return true;
            }

        }
        /* ***************************************************** */
        else if (getOperand().equals(Filter.OP_LT)) {
            if (de.getOperand().equals(Filter.OP_EQU)) {
                if (de.getDValue() < getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_NEQ)) {
            } else if (de.getOperand().equals(Filter.OP_GR)) {
            } else if (de.getOperand().equals(Filter.OP_LT)) {
                if (de.getDValue() <= getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LTE)) {
                if (de.getDValue() < getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_GTE)) {
            } else if (de.getOperand().equals(Filter.OP_RNG)) {
                if (getDValue() > de.getDValue2()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
            }
        }
        /* ***************************************************** */
        else if (getOperand().equals(Filter.OP_LTE)) {
            if (de.getOperand().equals(Filter.OP_EQU)) {
                if (de.getDValue() <= getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_NEQ)) {
            } else if (de.getOperand().equals(Filter.OP_GR)) {
            } else if (de.getOperand().equals(Filter.OP_LT)) {
                if (de.getDValue() <= getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LTE)) {
                if (de.getDValue() <= getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_RNG)) {
                if (getDValue() >= de.getDValue2()) {
                    return true;
                }

            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                //       return true;
            }
        }
        /* ***************************************************** */
        else if (getOperand().equals(Filter.OP_GTE)) {
            if (de.getOperand().equals(Filter.OP_EQU)) {
                if (de.getDValue() >= getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_NEQ)) {
            } else if (de.getOperand().equals(Filter.OP_GR)) {
                if (de.getDValue() >= getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_LT)) {
            } else if (de.getOperand().equals(Filter.OP_LTE)) {
            } else if (de.getOperand().equals(Filter.OP_GTE)) {
                if (de.getDValue() >= getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_RNG)) {
                if (getDValue() <= de.getDValue()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                //       return true;
            }
        }
        /* ***************************************************** */
        else if (getOperand().equals(Filter.OP_RNG)) {
            if (de.getOperand().equals(Filter.OP_RNG)) {
                if (de.getDValue() >= getDValue() &&
                    de.getDValue2() <= getDValue2()) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.OP_EQU)) {
                if (de.getDValue() >= getDValue() &&
                    de.getDValue() <= getDValue2()) {
                    return true;
                }
            }
        } else if (getOperand().equals(Filter.EXISTENCE)) {
            // existence covers all
            return true;
        }

        return false;
    }

    /** Tests whether the given predicate a merges with b and results
     *  in an existence test. Existence test is the most general filter
     *  and if one is detected, we can break the merging process for
     *  this field. Existence test covers all other predicates.
     * @param b the input object
     * @return true if the merge results in an existence test
     */
    public boolean perfectMergeRuleForExists(IAtomicPredicate b) {
        if (getValue() != null &&
            (getValue().startsWith("$") || b.getValue().startsWith("$"))
            &&
            !(getOperand().equals(Filter.OP_EQU) ||
              getOperand().equals(Filter.OP_NEQ))) {
            // the placeholder cannot be merged unless equ or neq is used, because
            // we do not know the value yet.
            return false;
        }

        if (getOperand().equals(Filter.OP_EQU) &&
            b.getOperand().equals(Filter.OP_NEQ) &&
            identicalValue(b)) {
            return true;
        } else if (b.getOperand().equals(Filter.OP_EQU) &&
                   getOperand().equals(Filter.OP_NEQ) &&
                   identicalValue(b)) {
            return true;
        } else if (getOperand().equals(Filter.OP_LT) &&
                   b.getOperand().equals(Filter.OP_GR) &&
                   getDValue() > b.getDValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_LT) &&
                   b.getOperand().equals(Filter.OP_GTE) &&
                   getDValue() >= b.getDValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_LTE) &&
                   b.getOperand().equals(Filter.OP_GR) &&
                   getDValue() >= b.getDValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_LTE) &&
                   b.getOperand().equals(Filter.OP_GTE) &&
                   getDValue() >= b.getDValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_GR) &&
                   b.getOperand().equals(Filter.OP_LT) &&
                   getDValue() < b.getDValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_GR) &&
                   b.getOperand().equals(Filter.OP_LTE) &&
                   getDValue() <= b.getDValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_GTE) &&
                   b.getOperand().equals(Filter.OP_LT) &&
                   getDValue() <= b.getDValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_GTE) &&
                   b.getOperand().equals(Filter.OP_LTE) &&
                   getDValue() <= b.getDValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_NEQ) &&
                   b.getOperand().equals(Filter.OP_NEQ) &&
                   getDValue() != b.getDValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_NEQ) &&
                   b.getOperand().equals(Filter.OP_GR) &&
                   b.getDValue() < getDValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_NEQ) &&
                   b.getOperand().equals(Filter.OP_GTE) &&
                   b.getDValue() <= getDValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_NEQ) &&
                   b.getOperand().equals(Filter.OP_LT) &&
                   b.getDValue() > getDValue()) {
            return true;
        } else if (getOperand().equals(Filter.OP_NEQ) &&
                   b.getOperand().equals(Filter.OP_LTE) &&
                   b.getDValue() >= getDValue()) {
            return true;
        } else if (b.getOperand().equals(Filter.OP_NEQ) &&
                   getOperand().equals(Filter.OP_GR) &&
                   getDValue() < b.getDValue()) {
            return true;
        } else if (b.getOperand().equals(Filter.OP_NEQ) &&
                   getOperand().equals(Filter.OP_GTE) &&
                   getDValue() <= b.getDValue()) {
            return true;
        } else if (b.getOperand().equals(Filter.OP_NEQ) &&
                   getOperand().equals(Filter.OP_LT) &&
                   getDValue() > b.getDValue()) {
            return true;
        } else if (b.getOperand().equals(Filter.OP_NEQ) &&
                   getOperand().equals(Filter.OP_LTE) &&
                   getDValue() >= b.getDValue()) {
            return true;
        } else if (b.getOperand().equals(Filter.OP_NEQ) &&
                   getOperand().equals(Filter.OP_RNG) &&
                   b.getDValue() >= getDValue() &&
                   b.getDValue() <= getDValue2()) {
            return true;
        } else if (b.getOperand().equals(Filter.OP_RNG) &&
                   getOperand().equals(Filter.OP_NEQ) &&
                   getDValue() >= b.getDValue() &&
                   getDValue() <= b.getDValue2()) {
            return true;
        }

        return false;
    }

    /** Tests whether the given predicate a merges with b and results
     *  in a modified predicate.
     * @param b the input object
     * @param change true if changes should be made
     * @return true if the merge results in an existence test
     */
    public boolean perfectMergeRuleForModified(IAtomicPredicate b,
                                               boolean change) {
        if (getOperand().equals(Filter.OP_RNG) &&
            b.getOperand().equals(Filter.OP_RNG)) {
            // check if the two regions overlap, if they do they can be
            // merged without false positives
            if (b.getDValue() <= getDValue2() &&
                getDValue() <= b.getDValue2()) { // Note that THIS OBJECT WILL CHANGE
                if (change) {
                    dval = Math.min(b.getDValue(), getDValue());
                    dval2 = Math.max(b.getDValue2(), getDValue2());
                    //System.out.println("Changing object: " + dval + "," + dval2);
                }
                return true;
            }
        }
        return false;
    }

    /** Matching functionality
     * @param se the first input value
     * @param a the second input value
     * @return true if the values match the condition
     */
    public boolean match(double se, double a) {
        if (op.equals(Filter.OP_EQU)) {
            if (se != a) {
                return false;
            }
        } else if (op.equals(Filter.OP_GR)) {
            if (a <= se) {
                return false;
            }
        } else if (op.equals(Filter.OP_LT)) {
            if (a >= se) {
                return false;
            }
        } else if (op.equals(Filter.OP_NEQ)) {
            if (a == se) {
                return false;
            }
        } else if (op.equals(Filter.OP_LTE)) {
            if (a > se) {
                return false;
            }
        } else if (op.equals(Filter.OP_GTE)) {
            if (a < se) {
                return false;
            }
        } else if (op.equals(Filter.OP_RNG)) {
            // does not currently support placeholders
            if (a < getDValue() || a > getDValue2()) {
                return false;
            }
        } else if (op.equals(Filter.EXISTENCE)) {
            // true
        }
        return true;
    }

    /** Matching functionality
     * @param a the first input value
     * @param b the second input value
     * @return true if the values match the condition
     */
    public boolean match(String a, String b) {
        return false;
    }

    /** Matching functionality
     * @param a the first input value
     * @param b the second input value
     * @return true if the values match the condition
     */
    public boolean match(Date a, Date b) {
        return false;
    }

    /** Tests for equivalence of atomic predicats
     * @param ip the input object
     * @return true if the two objects are identical
     */
    public boolean identical(IAtomicPredicate ip) {
        if (ip == null) {
            return false;
        }

        if (getOperand().equals(ip.getOperand())) {
            // for ranges we do not currently support placeholders
            if (getOperand().equals(Filter.OP_RNG)) {
                if (getDValue() == ip.getDValue() &&
                    getDValue2() == ip.getDValue2()) {
                    return true;
                } else {
                    return false;
                }
            }

            if (getValue() != null &&
                (getValue().startsWith("$") || ip.getValue().startsWith("$"))) {
                if (getValue().equals(ip.getValue())) {
                    return true;
                }
            }
            if (getDValue() == ip.getDValue()) {
                return true;
            }
        }
        return false;
    }

    /** Tests for the equivalence of the value part of two predicates
     * @param ip the input object
     * @return true if the two objects are identical
     */
    public boolean identicalValue(IAtomicPredicate ip) {
        if (ip == null) {
            return false;
        }

        if (getValue() != null &&
            (getValue().startsWith("$") || ip.getValue().startsWith("$"))) {
            if (getValue().equals(ip.getValue())) {
                return true;
            }
        }
        if (getDValue() == ip.getDValue()) {
            return true;
        }
        return false;
    }

    /** Returns the operand of the predicate
     * @return String the operand
     */
    public String getOperand() {
        return op;
    }

    /** Returns the String value of the predicate
     * @return String the value
     */
    public String getValue() {
        return val;
    }

    /** Returns the double value of the predicate
     * @return double the double value
     */
    public double getDValue() {
        return dval;
    }

    /** Returns the double value of the predicate
     * @return double the double value
     */
    public double getDValue2() {
        return dval2;
    }

    /** Returns the date value of the predicate
     * @return Date the date value
     */
    public Date getDateValue() {
        return dt;
    }

    /** Returns the time (long) value of the predicate
     * @return long the time value
     */
    public long getTimeValue() {
        return dt.getTime();
    }

    /** Returns the cost/generality of the predicate
     *  @return int the cost
     */
    public int getCost() {
        int cost = 0;
        if (getOperand().equals(Filter.OP_EQU)) {
            cost = 0; // restricting
        } else if (getOperand().equals(Filter.OP_NEQ)) {
            cost = 100; // less restricting
        } else if (getOperand().equals(Filter.OP_GR)) {
            cost = 50;
        } else if (getOperand().equals(Filter.OP_LT)) {
            cost = 50;
        } else if (getOperand().equals(Filter.OP_LTE)) {
            cost = 50;
        } else if (getOperand().equals(Filter.OP_GTE)) {
            cost = 50;
        } else if (getOperand().equals(Filter.OP_RNG)) {
            cost = 25;
        } else if (getOperand().equals(Filter.EXISTENCE)) {
            cost = 100;
        }
        return cost;
    }

    /** Return the value in string format
     * @return String the String representation
     */
    public String toString() {
        if (op.indexOf(Filter.OP_RNG) != -1) {
            return op + Filter.OPERAND_DIVISOR + dval + Filter.OPERAND_DIVISOR +
                    dval2;
        } else {
            return op + Filter.OPERAND_DIVISOR + dval;
        }
    }
}
