/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.filter.predicates;

import java.util.*;

import fuegocore.notify.core.filter.*;

/**
 * This class implements the string predicate.
 */
public class StringPredicate implements IAtomicPredicate {
    private String op;
    private Date dt;
    private double dval;
    private String val;

    public StringPredicate(String o, double v) {
        op = o.trim();
        dval = v;
    }

    public StringPredicate(String o, String v) {
        op = o.trim();
        val = v.trim();
    }

    public StringPredicate(String o, Date v) {
        op = o.trim();
        dt = v;
    }

    /** Returns id for this object
     * @return String the identifier
     */
    public String getId() {
        if (val != null) {
            return op + "_" + val;
        } else if (dt != null) {
            return op + "_" + dt.getTime();
        } else {
            return op + "_" + dval;
        }
    }

    /**
     *  Returns the value of an equality predicate (or null)
     * @param type int
     * @return String  the value of the equality predicate
     */
    public String getEqualityValue(int type) {
        if (getOperand().equals(Filter.EQUAL)) {
            return getValue().trim();
        }
        return null;
    }

    /** Covering relation. Returns true if this object covers the given object.
     * @param de the input object
     * @return true if the input object is covered
     */
    public boolean covers(IAtomicPredicate de) {
        //System.out.println("Covers:" + getOperand() + " val: " + getValue());
        if (getOperand().equals(Filter.SELECT) &&
            de.getOperand().equals(Filter.SELECT) &&
            getValue().equals(de.getValue())) {
            return true;
        }

        if (de.getValue() == null) {
            System.out.println("StringPredicate: de was null");
            return false;
        }

        // If either value has a dollar sign (placeholder)
        if (getValue() != null && (getValue().startsWith("$") ||
                                   de.getValue().startsWith("$"))) {
            if (getValue().equals(de.getValue())) {
                // Identical values (must be for placeholders)
                if (de.getOperand().equals(getOperand())) {
                    return true;
                }
                /*
                         op,op
                         in, equ
                         in,stw
                         in, end
                         neq,nin
                         stw, equ
                         end,equ

                 */
                if ((getOperand().equals(Filter.IN_STRING) &&
                     de.getOperand().equals(Filter.EQUAL)) ||
                    (getOperand().equals(Filter.IN_STRING) &&
                     de.getOperand().equals(Filter.STARTS_WITH)) ||
                    (getOperand().equals(Filter.IN_STRING) &&
                     de.getOperand().equals(Filter.ENDS_WITH)) ||
                    (getOperand().equals(Filter.NOT_EQUAL) &&
                     de.getOperand().equals(Filter.NOT_IN_STRING)) ||
                    (getOperand().equals(Filter.STARTS_WITH) &&
                     de.getOperand().equals(Filter.EQUAL)) ||
                    (getOperand().equals(Filter.ENDS_WITH) &&
                     de.getOperand().equals(Filter.EQUAL))) {
                    return true;
                }
            }
            // otherwise we cannot compare
            return false;
        }

        /* ********************************************** */

        if (getOperand().equals(Filter.EQUAL)) {
            if (de.getOperand().equals(Filter.EQUAL)) {
                if (getValue().equals(de.getValue())) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_EQUAL)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.IN_STRING)) {

                // does not cover
            } else if (de.getOperand().equals(Filter.NOT_IN_STRING)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.STARTS_WITH)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.ENDS_WITH)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                return true;
            }
        }
        /* ********************************************** */
        else if (getOperand().equals(Filter.NOT_EQUAL)) {
            if (de.getOperand().equals(Filter.EQUAL)) {
                if (!getValue().equals(de.getValue())) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_EQUAL)) {
                if (getValue().equals(de.getValue())) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.IN_STRING)) {
                if (getValue().indexOf(de.getValue()) == -1) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_IN_STRING)) {
                if (getValue().indexOf(de.getValue()) != -1) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.STARTS_WITH)) {
                if (!getValue().startsWith(de.getValue())) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.ENDS_WITH)) {
                if (!getValue().endsWith(de.getValue())) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                return true;
            }
        }
        /* ********************************************** */
        else if (getOperand().equals(Filter.IN_STRING)) {
            if (de.getOperand().equals(Filter.EQUAL)) {
                if (de.getValue().indexOf(getValue()) != -1) {
                    // covered
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_EQUAL)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.IN_STRING)) {
                if (de.getValue().indexOf(getValue()) != -1) {
                    // covered
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_IN_STRING)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.STARTS_WITH)) {
                if (de.getValue().indexOf(getValue()) != -1) {
                    // covered
                    return true;
                }
            } else if (de.getOperand().equals(Filter.ENDS_WITH)) {
                if (de.getValue().indexOf(getValue()) != -1) {
                    // covered
                    return true;
                }
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                return true;
            }
        }
        /* ********************************************** */
        else if (getOperand().equals(Filter.NOT_IN_STRING)) {
            if (de.getOperand().equals(Filter.EQUAL)) {
                if (de.getValue().indexOf(getValue()) == -1) {
                    // covered
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_EQUAL)) {
                if (getValue().indexOf(de.getValue()) != -1) {
                    // covered
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_IN_STRING)) {
                if (getValue().indexOf(de.getValue()) != -1) {
                    // covered
                    return true;
                }
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                return true;
            }
        }
        /* ********************************************** */
        else if (getOperand().equals(Filter.STARTS_WITH)) {
            if (de.getOperand().equals(Filter.EQUAL)) {
                if (de.getValue().startsWith(getValue())) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_EQUAL)) {
            } else if (de.getOperand().equals(Filter.IN_STRING)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.NOT_IN_STRING)) {
            } else if (de.getOperand().equals(Filter.STARTS_WITH)) {
                if (de.getValue().startsWith(getValue())) {
                    // covered
                    return true;
                }
            } else if (de.getOperand().equals(Filter.ENDS_WITH)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                return true;
            }
        }
        /* ********************************************** */
        else if (getOperand().equals(Filter.ENDS_WITH)) {
            if (de.getOperand().equals(Filter.EQUAL)) {
                if (de.getValue().endsWith(getValue())) {
                    // covered
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_EQUAL)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.IN_STRING)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.NOT_IN_STRING)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.STARTS_WITH)) {
                // does not cover
            } else if (de.getOperand().equals(Filter.ENDS_WITH)) {
                if (de.getValue().endsWith(getValue())) {
                    // covered
                    return true;
                }
            } else if (de.getOperand().equals(Filter.EXISTENCE)) {
                return true;
            }
        }
        /* ********************************************** */
        else if (getOperand().equals(Filter.EXISTENCE)) {
            // existence covers all
            return true;
        }

        return false;
    }

    /** Overlapping relation. Returns true if this object overlaps with the given object.
     * @param de the input object
     * @return true if the input object overlaps with the current object
     */
    public boolean overlaps(IAtomicPredicate de) {
        //System.out.println("Covers:" + getOperand() + " val: " + getValue());
        if (getOperand().equals(Filter.SELECT) &&
            de.getOperand().equals(Filter.SELECT) &&
            getValue().equals(de.getValue())) {
            return true;
        }

        if (de.getValue() == null) {
            System.out.println("StringPredicate: de was null");
            return false;
        }

        // If either value has a dollar sign (placeholder)
        if (getValue() != null && (getValue().startsWith("$") ||
                                   de.getValue().startsWith("$"))) {
            if (getValue().equals(de.getValue()) &&
                getOperand().equals(de.getOperand())) {
                // Identical values (must be for placeholders)
                if (de.getOperand().equals(getOperand())) {
                    return true;
                }
                /*
                          Overlap (symmetric)
                          op,op
                          in, equ
                          in, stw
                          in end

                          end equ
                          stw,equ
                          neq nin
                 */
                if ((de.getOperand().equals(Filter.IN_STRING) &&
                     getOperand().equals(Filter.EQUAL)) ||
                    (de.getOperand().equals(Filter.IN_STRING) &&
                     getOperand().equals(Filter.STARTS_WITH)) ||
                    (de.getOperand().equals(Filter.IN_STRING) &&
                     getOperand().equals(Filter.ENDS_WITH)) ||
                    (de.getOperand().equals(Filter.STARTS_WITH) &&
                     getOperand().equals(Filter.EQUAL)) ||
                    (de.getOperand().equals(Filter.ENDS_WITH) &&
                     getOperand().equals(Filter.EQUAL)) ||
                    (de.getOperand().equals(Filter.NOT_EQUAL) &&
                     getOperand().equals(Filter.NOT_IN_STRING)) ||

                    (getOperand().equals(Filter.IN_STRING) &&
                     de.getOperand().equals(Filter.EQUAL)) ||
                    (getOperand().equals(Filter.IN_STRING) &&
                     de.getOperand().equals(Filter.STARTS_WITH)) ||
                    (getOperand().equals(Filter.IN_STRING) &&
                     de.getOperand().equals(Filter.ENDS_WITH)) ||
                    (getOperand().equals(Filter.STARTS_WITH) &&
                     de.getOperand().equals(Filter.EQUAL)) ||
                    (getOperand().equals(Filter.ENDS_WITH) &&
                     de.getOperand().equals(Filter.EQUAL)) ||
                    (getOperand().equals(Filter.NOT_EQUAL) &&
                     de.getOperand().equals(Filter.NOT_IN_STRING))) {
                    return true;
                }
            }
            // otherwise we cannot compare
            return false;
        }

        if (getOperand().equals(Filter.EXISTENCE) ||
            de.getOperand().equals(Filter.EXISTENCE)) {
            return true;
        }

        /* ********************************************** */

        if (getOperand().equals(Filter.EQUAL)) {
            if (de.getOperand().equals(Filter.EQUAL)) {
                if (getValue().equals(de.getValue())) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_EQUAL)) {
                if (!getValue().equals(de.getValue())) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.IN_STRING)) {
                if (getValue().indexOf(de.getValue()) != -1) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_IN_STRING)) {
                if (getValue().indexOf(de.getValue()) == -1) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.STARTS_WITH)) {
                if (getValue().startsWith(de.getValue())) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.ENDS_WITH)) {
                if (getValue().endsWith(de.getValue())) {
                    return true;
                }
            }
        }
        /* ********************************************** */
        else if (getOperand().equals(Filter.NOT_EQUAL)) {
            if (de.getOperand().equals(Filter.EQUAL)) {
                if (!getValue().equals(de.getValue())) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_EQUAL)) {
                // always overlap
                return true;
            } else if (de.getOperand().equals(Filter.IN_STRING)) {
                return true;
            } else if (de.getOperand().equals(Filter.NOT_IN_STRING)) {
                return true;
            } else if (de.getOperand().equals(Filter.STARTS_WITH)) {
                return true;
            } else if (de.getOperand().equals(Filter.ENDS_WITH)) {
                return true;
            }
        }
        /* ********************************************** */
        else if (getOperand().equals(Filter.IN_STRING)) {
            if (de.getOperand().equals(Filter.EQUAL)) {
                if (de.getValue().indexOf(getValue()) != -1) {
                    // covered
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_EQUAL)) {
                return true;
            } else if (de.getOperand().equals(Filter.IN_STRING)) {
                if (de.getValue().indexOf(getValue()) != -1 ||
                    getValue().indexOf(de.getValue()) != -1) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_IN_STRING)) {
                // if in contains nin they do not overlap
                if (getValue().indexOf(de.getValue()) != -1) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.STARTS_WITH)) {
                if (de.getValue().indexOf(getValue()) != -1) {
                    // covered
                    return true;
                }
            } else if (de.getOperand().equals(Filter.ENDS_WITH)) {
                if (de.getValue().indexOf(getValue()) != -1) {
                    // covered
                    return true;
                }
            }
        }
        /* ********************************************** */
        else if (getOperand().equals(Filter.NOT_IN_STRING)) {
            if (de.getOperand().equals(Filter.EQUAL)) {
                if (de.getValue().indexOf(getValue()) == -1) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_EQUAL)) {
                return true;
            } else if (de.getOperand().equals(Filter.IN_STRING)) {
                // if in contains nin they do not overlap
                if (de.getValue().indexOf(getValue()) != -1) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_IN_STRING)) {
                // if in contains nin they do not overlap
                if (getValue().indexOf(de.getValue()) != -1 ||
                    de.getValue().indexOf(getValue()) != -1) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.STARTS_WITH)) {
                if (de.getValue().indexOf(getValue()) == -1) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.ENDS_WITH)) {
                if (de.getValue().indexOf(getValue()) == -1) {
                    return true;
                }
            }
        }
        /* ********************************************** */
        else if (getOperand().equals(Filter.STARTS_WITH)) {
            if (de.getOperand().equals(Filter.EQUAL)) {
                if (de.getValue().startsWith(getValue())) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_EQUAL)) {
                return true;
            } else if (de.getOperand().equals(Filter.IN_STRING)) {
                if (getValue().indexOf(de.getValue()) != -1) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_IN_STRING)) {
                if (getValue().indexOf(de.getValue()) == -1) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.STARTS_WITH)) {
                if (de.getValue().startsWith(getValue()) ||
                    getValue().startsWith(de.getValue())) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.ENDS_WITH)) {
                // they may potentially overlap, we do not know it
                // but we cannot leave it out either.
                return true;
            }
        }
        /* ********************************************** */
        else if (getOperand().equals(Filter.ENDS_WITH)) {
            if (de.getOperand().equals(Filter.EQUAL)) {
                if (de.getValue().endsWith(getValue())) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_EQUAL)) {
                return true;
            } else if (de.getOperand().equals(Filter.IN_STRING)) {
                if (getValue().indexOf(de.getValue()) != -1) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.NOT_IN_STRING)) {
                if (getValue().indexOf(de.getValue()) == -1) {
                    return true;
                }
            } else if (de.getOperand().equals(Filter.STARTS_WITH)) {
                // they may potentially overlap, we do not know it
                // but we cannot leave it out either.
                return true;
            }
        } else if (de.getOperand().equals(Filter.ENDS_WITH)) {
            if (de.getValue().endsWith(getValue()) ||
                getValue().endsWith(de.getValue())) {
                return true;
            }
        }
        return false;
    }

    /** Tests whether the given predicate a merges with b and resuts
     *  in an existence test. Existence test is the most general filter
     *  and if one is detected, we can break the merging process for
     *  this field. Existence test covers all other predicates.
     * @param b the input object
     * @return true if the merge results in an existence test
     */
    public boolean perfectMergeRuleForExists(IAtomicPredicate b) {
        if (getOperand().equals(Filter.EQUAL) &&
            b.getOperand().equals(Filter.NOT_EQUAL) &&
            identicalValue(b)) {
            return true;
        } else if (b.getOperand().equals(Filter.EQUAL) &&
                   getOperand().equals(Filter.NOT_EQUAL) &&
                   identicalValue(b)) {
            return true;
        } else if (b.getOperand().equals(Filter.NOT_EQUAL) &&
                   getOperand().equals(Filter.NOT_EQUAL) &&
                   !identicalValue(b)) {
            return true;
        } else if (getOperand().equals(Filter.IN_STRING) &&
                   b.getOperand().equals(Filter.NOT_IN_STRING) &&
                   b.getValue().indexOf(getValue()) != -1) {
            return true;
        } else if (b.getOperand().equals(Filter.IN_STRING) &&
                   getOperand().equals(Filter.NOT_IN_STRING) &&
                   getValue().indexOf(b.getValue()) != -1) {
            return true;
        } else if (b.getOperand().equals(Filter.NOT_IN_STRING) &&
                   getOperand().equals(Filter.NOT_EQUAL) &&
                   getValue().indexOf(b.getValue()) == -1) {
            return true;
        } else if (b.getOperand().equals(Filter.NOT_EQUAL) &&
                   getOperand().equals(Filter.NOT_IN_STRING) &&
                   b.getValue().indexOf(getValue()) == -1) {
            return true;
        } else if (b.getOperand().equals(Filter.NOT_IN_STRING) &&
                   getOperand().equals(Filter.NOT_IN_STRING) &&
                   getValue().indexOf(b.getValue()) == -1 &&
                   b.getValue().indexOf(getValue()) == -1) {
            return true;
        }
        // ----------------
        else if (b.getOperand().equals(Filter.NOT_EQUAL) &&
                 getOperand().equals(Filter.STARTS_WITH) &&
                 b.getValue().startsWith(getValue())) {
            return true;
        } else if (getOperand().equals(Filter.NOT_EQUAL) &&
                   b.getOperand().equals(Filter.STARTS_WITH) &&
                   getValue().startsWith(b.getValue())) {
            return true;
        } else if (b.getOperand().equals(Filter.NOT_EQUAL) &&
                   getOperand().equals(Filter.ENDS_WITH) &&
                   b.getValue().endsWith(getValue())) {
            return true;
        } else if (getOperand().equals(Filter.NOT_EQUAL) &&
                   b.getOperand().equals(Filter.ENDS_WITH) &&
                   getValue().endsWith(b.getValue())) {
            return true;
        }
        // ---------
        else if (b.getOperand().equals(Filter.NOT_IN_STRING) &&
                 getOperand().equals(Filter.STARTS_WITH) &&
                 b.getValue().startsWith(getValue())) {
            return true;
        } else if (getOperand().equals(Filter.NOT_IN_STRING) &&
                   b.getOperand().equals(Filter.STARTS_WITH) &&
                   getValue().startsWith(b.getValue())) {
            return true;
        } else if (b.getOperand().equals(Filter.NOT_IN_STRING) &&
                   getOperand().equals(Filter.ENDS_WITH) &&
                   b.getValue().endsWith(getValue())) {
            return true;
        } else if (getOperand().equals(Filter.NOT_IN_STRING) &&
                   b.getOperand().equals(Filter.ENDS_WITH) &&
                   getValue().endsWith(b.getValue())) {
            return true;
        }

        return false;
    }

    /** Matching functionality
     * @param se the first input value
     * @param a the second input value
     * @return true if the values match the condition
     */
    public boolean match(double se, double a) {
        return false;
    }

    /** Matching functionality
     * @param se the first input value
     * @param v the second input value
     * @return true if the values match the condition
     */
    public boolean match(String se, String v) {

        if (getOperand().equals(Filter.NOT_EQUAL)) {
            if (se.equals(v.trim())) {
                return false;
            }
        } else if (getOperand().equals(Filter.EQUAL)) {
            if (!se.equals(v.trim())) {
                return false;
            }
        } else if (getOperand().equals(Filter.IN_STRING)) {
            if (v.indexOf(se) == -1) {
                return false;
            }
        } else if (getOperand().equals(Filter.NOT_IN_STRING)) {
            if (v.indexOf(se) != -1) {
                return false;
            }
        } else if (getOperand().equals(Filter.STARTS_WITH)) {
            if (!v.startsWith(se)) {
                return false;
            }
        } else if (getOperand().equals(Filter.ENDS_WITH)) {
            if (!v.endsWith(se)) {
                return false;
            }
        } else if (getOperand().equals(Filter.EXISTENCE)) {
            // true
        }
        return true;
    }

    /** Tests whether the given predicate a merges with b and results
     *  in a modified predicate.
     * @param b the input object
     * @param change true if changes should be made
     * @return true if the merge results in an existence test
     */
    public boolean perfectMergeRuleForModified(IAtomicPredicate b,
                                               boolean change) {
        return false;
    }

    /** Matching functionality
     * @param a the first input value
     * @param b the second input value
     * @return true if the values match the condition
     */
    public boolean match(Date a, Date b) {
        return false;
    }

    /** Tests for equivalence of atomic predicats
     * @param ip the input object
     * @return true if the two objects are identical
     */
    public boolean identical(IAtomicPredicate ip) {
        if (ip == null) {
            return false;
        }
        if (getOperand().equals(ip.getOperand()) &&
            getValue().equals(ip.getValue())) {
            return true;
        }
        return false;
    }

    /** Tests for the equivalence of the value part of two predicates
     * @param ip the input object
     * @return true if the two objects are identical
     */
    public boolean identicalValue(IAtomicPredicate ip) {
        if (ip == null) {
            return false;
        }
        if (getValue() != null &&
            (getValue().startsWith("$") || ip.getValue().startsWith("$"))) {
            if (getValue().equals(ip.getValue())) {
                return true;
            }
        }
        if (getValue().equals(ip.getValue())) {
            return true;
        }
        return false;
    }

    /** Returns the operand of the predicate
     * @return String the operand
     */
    public String getOperand() {
        return op;
    }

    /** Returns the String value of the predicate
     * @return String the value
     */
    public String getValue() {
        return val;
    }

    /** Returns the double value of the predicate
     * @return double the double value
     */
    public double getDValue() {
        return dval;
    }

    /** Get second double value
     * @return double the value
     */
    public double getDValue2() {
        return 0;
    }

    /** Returns the date value of the predicate
     * @return Date the date value
     */
    public Date getDateValue() {
        return dt;
    }

    /** Returns the time (long) value of the predicate
     * @return long the time value
     */
    public long getTimeValue() {
        return dt.getTime();
    }

    /** Return the value in string format
     * @return String the String representation
     */
    public String toString() {
        return op + Filter.OPERAND_DIVISOR + val;
    }

    /** Returns the cost/generality of the predicate
     *  Smaller value means more restricting
     *  @return int the cost
     */
    public int getCost() {
        int cost = 0;

        if (getOperand().equals(Filter.EQUAL)) {
            cost = 0; // restricting
        } else if (getOperand().equals(Filter.NOT_EQUAL)) {
            cost = 100; // less restricting
        } else if (getOperand().equals(Filter.IN_STRING)) {
            cost = 50;
        } else if (getOperand().equals(Filter.NOT_IN_STRING)) {
            cost = 50;
        } else if (getOperand().equals(Filter.STARTS_WITH)) {
            cost = 50;
        } else if (getOperand().equals(Filter.ENDS_WITH)) {
            cost = 50;
        } else if (getOperand().equals(Filter.EXISTENCE)) {
            cost = 100;
        }
        return cost;
    }

}
