/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.filter.test;

import java.io.*;
import java.util.*;
import java.util.zip.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.filter.merge.*;
import fuegocore.notify.core.filter.poset.*;
import fuegocore.notify.util.*;

/**
 * Auxiliary methods for testing.
 */
public class FilterServiceTestDriver  {
    public static int REPLICATIONS = 5; // number of replications for the benchmark
    public static int TEMPLATES = 100; // number of templates (subscriptions)
    public static int TEMPLATEFIELDS = 2; // fields per subscription+(type not included)
    public static int FIELDNAMES = 3; // possible field names
    public static int TYPENAMES = 1; // possible event types (channels)
    public static int FILTERNAMES = 100; // possible filter names
    public static int RANGE = 100; // range for number tests
    public static int TESTNUMBER = 1000; // number of notifications to test
    public static int OUTSIDENUMBER = 0; // this many notifications with out-of-range values
    public static int TYPE_MODE = 4;
    public static boolean RANGE_ONLY = false;

    /* force only inequalities in numbers */
    boolean VARIABLE = false;
    int test_counter = 0; // for dumping out the test data

    /* TYPE_MODE: 0 = int, 1 = string,
     2 = int and string, 3= int,string,boolean, 4=int,string,boolean,double,
     */

    /** special case, equality only in string and number */
    public static boolean EQU_ONLY = false;

    /** Omit neq feature in the tests */
    public static boolean OMIT_NEQ = false;

    public boolean allowVariableLength = false; // allow variable length notifications
    boolean enforceStructure = false; // enforce structure

    String randomFieldNames[] = new String[11001];
    String randomFilters[] = new String[11001];
    String randomFilters2[] = new String[11001];
    String typeNames[] = new String[11001];
    Notification notifs[] = new Notification[11001];

    // For detecting false positives and false negatives
    boolean correct[] = new boolean[11001];

    public FilterServiceTestDriver(boolean b) {}

    public FilterServiceTestDriver() {
   }

    /** Dumps out test data (Vector of templates)
     */
    public void dumpData(Vector t, String suffix) {
        PrintWriter fout = null;

        if (t == null) {
            return;
        }

        try {
            fout = new PrintWriter(new FileWriter(
                    "merge_" + this.test_counter + "_" + suffix + ".dat"), true); // append
        } catch (Exception e2) {
            e2.printStackTrace();
        }

        for (int i = 0; i < t.size(); i++) {
            Template tt = (Template) t.elementAt(i);
            fout.println(tt.toString());
        }
        fout.close();
    }


    public String getParams() {
        String ret = "";
        ret += "typenames: " + TYPENAMES + " templatefields:" + TEMPLATEFIELDS;
        ret += "filternames: " + FILTERNAMES + " range: " + RANGE;

        return ret;
    }

    public Vector getSchemas() {
        Vector ret = new Vector();

        for (int i = 0; i < TYPENAMES; i++) {
            Template t = new Template();
            String etype = typeNames[i];

            // mandatory type
            Filter ft = new Filter("type", EventTriplet.TYPE_STRING,
                                   etype);
            t.addFilter(ft);
            Hashtable usedName = new Hashtable();

            for (int j = 0; j < TEMPLATEFIELDS; j++) {
                int type = getRandomType();
                // ensure that the filter names are unique, there should not
                // be duplicates.
                String name = getRandomFieldName();
                // crude way to realize uniqueness
                while (name.equals("type") || usedName.get(name + "/" + type) != null) {
                    name = getRandomFieldName();
                }
                usedName.put(name + "/" + type, "true");
                String constraint = getConstraint(type);
                Filter f = new Filter(name, type, "");
                t.addFilter(f);
            }
            ret.add(t);
        }
        return ret;
    }

    public Vector createTemplatesUsingSchemas(int n, Vector schemas) {
        Vector ret = new Vector();
        // Create set of templates to be merged
        for (int i = 0; i < n; i++) {
            int temps = (int) (Math.random() * schemas.size());
            Template t = new Template();
            Template tt = null;
            try {
                Template t2 = (Template) schemas.elementAt(temps);
                tt = (Template) t2.clone();
            } catch (Exception e) {
                e.printStackTrace();
            }
            Vector tv = tt.getFilters();
            for (int j = 0; j < tv.size(); j++) {
                // get filter template
                Filter fx = (Filter) tv.elementAt(j);
                int type = fx.getType();
                String name = fx.getName();
                if (!name.equals("type")) {
                    // Get new constraint
                    String constraint = getConstraint(type);
                    // create a new filter according to the specification
                    Filter fx2 = new Filter(name, type, constraint);
                    // set the new vector
                    tv.set(j, fx2);
                }
            }
            ret.add(tt);
        }

        return ret;
    }

    public Vector createVariableTemplatesUsingSchemas(int n, Vector schemas) {
        return createVariableTemplatesUsingSchemas(n, schemas, 1);
    }

    public Vector createVariableTemplatesUsingSchemas(int n, Vector schemas,
            int offset) {
        Vector ret = new Vector();

        // Create set of templates to be merged

        for (int i = 0; i < n; i++) {
            int temps = (int) (Math.random() * schemas.size());
            Template t = new Template();
            Template tt = null;

            try {
                Template t2 = (Template) schemas.elementAt(temps);
                tt = (Template) t2.clone();
            } catch (Exception e) {
                e.printStackTrace();
            }

            Vector tv = tt.getFilters();
            Vector ntv = new Vector();
            // random template fields
            // we get the size and then shuffle and take n first
            int len = (int) (Math.random() *
                             (double) (TEMPLATEFIELDS - offset + 1));
            int kj = offset + len;
            if (kj > TEMPLATEFIELDS) {
                kj = TEMPLATEFIELDS;
            }
            ntv.add(new Filter("type", EventTriplet.TYPE_STRING, tt.getType()));
            Collections.shuffle(tv);
            int kk = 0;
            for (int j = 0; j < tv.size(); j++) {
                if (kk >= kj) {
                    break;
                }
                // get filter template
                Filter fx = (Filter) tv.elementAt(j);
                int type = fx.getType();
                String name = fx.getName();
                if (!name.equals("type")) {
                    // Get new constraint
                    String constraint = getConstraint(type);
                    // create a new filter according to the specification
                    Filter fx2 = new Filter(name, type, constraint);
                    // set the new vector
                    tv.set(j, fx2);
                    ntv.add(fx2);
                    kk++;
                }
            }
            tt.setFilters(ntv);
            ret.add(tt);
        }
        return ret;
    }

    public Vector createVariableTemplatesWithDisjunctionsUsingSchemas(int n,
            int nd, Vector schemas) {
        Vector ret = new Vector();

        // Create set of templates to be merged

        for (int i = 0; i < n; i++) {
            int temps = (int) (Math.random() * schemas.size());
            Template t = new Template();
            Template tt = null;

            try {
                Template t2 = (Template) schemas.elementAt(temps);
                tt = (Template) t2.clone();
            } catch (Exception e) {
                e.printStackTrace();
            }

            Vector tv = tt.getFilters();
            Vector ntv = new Vector();
            // random template fields
            // we get the size and then shuffle and take n first

            int kj = 1 + (int) (Math.random() * (double) TEMPLATEFIELDS);
            ntv.add(new Filter("type", EventTriplet.TYPE_STRING, tt.getType()));
            Collections.shuffle(tv);
            int kk = 0;
            for (int j = 0; j < tv.size(); j++) {
                // get filter template
                Filter fx = (Filter) tv.elementAt(j);
                int type = fx.getType();
                String name = fx.getName();
                if (!name.equals("type")) {
                    String constraint = "";
                    for (int z = 0; z < nd; z++) {
                        // Get new constraint
                        constraint += getConstraint(type);
                        if (z < nd - 1) {
                            constraint += "|";
                        }
                    }
                    // create a new filter according to the specification
                    Filter fx2 = new Filter(name, type, constraint);
                    // set the new vector
                    tv.set(j, fx2);
                    ntv.add(fx2);
                    kk++;
                }
                if (kk >= kj) {
                    break;
                }
            }
            tt.setFilters(ntv);
            ret.add(tt);
        }
        return ret;
    }

    public Notification getNotificationUsingSchema(Template tt) {
        int tc = 0;
        Notification n = new Notification();
        EventTriplet et2 = new EventTriplet("type", tt.getType());
        n.add(et2);

        for (int j = 0; j < tt.getFilters().size(); j++) {
            Filter f = (Filter) tt.getFilters().elementAt(j);
            if (!f.getName().equals("type")) {
                EventTriplet et = createTriplet(f.getType());
                et.setName(f.getName());
                n.add(et);
            }
        }
        return n;
    }

    /** This method prints a poset (partially ordered set) from the given
     *  set of templates.
     *
     * @param templates the templates to be inserted into the poset
     */
    public void printPoset(FilterForest fp, int reps) {

        System.out.println("Poset node count: " + fp.toString());
        // Output GraphViz datafile
        try {
            PrintWriter fout = new PrintWriter(new FileWriter(
                    "poset_" + this.test_counter + "_" + reps + ".dat"), false); // append
            fp.printDot(fout);
            fout.close();

        } catch (Exception e2) {
            e2.printStackTrace();
        }
    }


    /** Returns the length of the template XML representation in GZIP format */
    public int filterGZIPLength(String temps) {
        try {
            ByteArrayOutputStream bay = new ByteArrayOutputStream();
            GZIPOutputStream gzip = new GZIPOutputStream(bay);

            gzip.write(temps.getBytes(), 0, temps.getBytes().length);
            gzip.finish();
            gzip.close();
            byte[] bb = bay.toByteArray();
            return bb.length;
        } catch (Exception e) {
            e.printStackTrace();
        }
        return -1;
    }

    /** Generates random and unique field names and filter constraints */
    public void populateFields() {

        Hashtable usedName = new Hashtable();

        // Generate a random set of field names
        for (int i = 0; i < FIELDNAMES; i++) {
            do {
                randomFieldNames[i] = "";

                for (int j = 0; j < 3 + 8 * Math.random(); j++) {
                    int val = (int) (Math.random() * (double) ('z' - 'a'));
                    randomFieldNames[i] += (char) ('a' + val);
                }
            } while (usedName.get(randomFieldNames[i]) != null);
            usedName.put(randomFieldNames[i], "true");
        }

        usedName.clear();

        // Generate a random set of filter constraints
        for (int i = 0; i < FILTERNAMES; i++) {
            randomFilters[i] = "";
            for (int j = 0; j < 3 + 8 * Math.random(); j++) {
                int val = (int) (Math.random() * (double) ('z' - 'a'));
                randomFilters[i] += (char) ('a' + val);
            }
            usedName.put(randomFilters[i], "true");
        }

        // usedName must not be cleared here to create an orthogonal set

        // Generate a random set of filter constraints: different set
        for (int i = 0; i < FILTERNAMES; i++) {
            do {
                randomFilters2[i] = "";
                for (int j = 0; j < 3 + 8 * Math.random(); j++) {
                    int val = (int) (Math.random() * (double) ('z' - 'a'));
                    randomFilters2[i] += (char) ('a' + val);
                }
            } while (usedName.get(randomFilters2[i]) != null);
        }

        // Generate a random set of unique type names
        usedName.clear();

        for (int i = 0; i < TYPENAMES; i++) {
            do {
                typeNames[i] = "";
                for (int j = 0; j < 3 + 8 * Math.random(); j++) {
                    int val = (int) (Math.random() * (double) ('z' - 'a'));
                    typeNames[i] += (char) ('a' + val);
                }
            } while (usedName.get(typeNames[i]) != null);
            usedName.put(typeNames[i], "true");
        }
    }

    /** Returns a random field name */
    public String getRandomFieldName() {
        int i = (int) (Math.random() * FIELDNAMES);
        return randomFieldNames[i];
    }

    /** Returns a random filter */
    public String getRandomFilter() {
        int i = (int) (Math.random() * FILTERNAMES);
        return randomFilters[i];
    }

    /** Returns a random filter */
    public String getRandomFilterTest(int count) {
        String ret = "";
        int l = (int) (Math.random() * count);
        ret += (char) ('a' + l);
        return ret;
    }

    /** Returns a random field name */
    public String getRandomFilterOutside() {
        int i = (int) (Math.random() * FILTERNAMES);
        return randomFilters2[i];
    }

    /** Returns a random field name */
    public String getRandomTypeName() {
        int i = (int) (Math.random() * TYPENAMES);
        return typeNames[i];
    }

    /** Returns a random field type */
    public int getRandomType() {
        int rand = 4;

        /* TYPE_MODE: 0 = int, 1 = string,
         2 = int and string, 3= int,string,boolean, 4=int,string,boolean,double
         */
        if (TYPE_MODE == 0) {
            return EventTriplet.TYPE_INT;
        } else if (TYPE_MODE == 1) {
            return EventTriplet.TYPE_STRING;
        } else if (TYPE_MODE == 2) {
            rand = 2;
        } else if (TYPE_MODE == 3) {
            rand = 3;
        } else if (TYPE_MODE == 4) {
            rand = 4;

        }
        int i = (int) (Math.random() * rand);
        if (i == 0) {
            return EventTriplet.TYPE_INT;
        } else if (i == 1) {
            return EventTriplet.TYPE_STRING;
        } else if (i == 2) {
            return EventTriplet.TYPE_BOOLEAN;
        } else if (i == 3) {
            return EventTriplet.TYPE_DOUBLE;
        }
        // no dates
        return -1;
    }

    public EventTriplet createTriplet(int type) {
        EventTriplet et = null;

        String fname = getRandomFieldName();

        if (type == EventTriplet.TYPE_INT) {
            double j = (double) (Math.random() * RANGE);
            et = new EventTriplet(fname, (int) j);
        }
        if (type == EventTriplet.TYPE_DOUBLE) {
            float j = (float) (Math.random() * RANGE);
            et = new EventTriplet(fname, (double) j);
        } else if (type == EventTriplet.TYPE_BOOLEAN) {
            int k = (int) (Math.random() * 2);
            if (k == 0) {
                et = new EventTriplet(fname, true);
            } else if (k == 1) {
                et = new EventTriplet(fname, false);
            } else {
                System.out.println("Error in generating type boolean");
                System.exit(0);
            }
        } else if (type == EventTriplet.TYPE_STRING) {
            String co = getRandomFilter();
            et = new EventTriplet(fname, co);
        }
        return et;
    }

    public EventTriplet createTripletOutside(int type) {
        EventTriplet et = null;

        String fname = getRandomFieldName();

        if (type == EventTriplet.TYPE_INT) {
            double j = (double) (Math.random() * RANGE) + RANGE;
            et = new EventTriplet(fname, (int) j);
        }
        if (type == EventTriplet.TYPE_DOUBLE) {
            float j = (float) (Math.random() * RANGE) + RANGE;
            et = new EventTriplet(fname, (double) j);
        } else if (type == EventTriplet.TYPE_BOOLEAN) {
            int k = (int) (Math.random() * 2);
            if (k == 0) {
                et = new EventTriplet(fname, true);
            } else if (k == 1) {
                et = new EventTriplet(fname, false);
            } else {
                System.out.println("Error in generating type boolean");
                System.exit(0);
            }
        } else if (type == EventTriplet.TYPE_STRING) {
            String co = getRandomFilterOutside();
            et = new EventTriplet(fname, co);
        }
        return et;
    }

    /** Returns filter constaint */
    public String getConstraint(int type) {
        String str = "";
        int rand = 5;
        if (type == EventTriplet.TYPE_INT || type == EventTriplet.TYPE_DOUBLE) {
            if (OMIT_NEQ) {
                rand = 4;
            }
            int k = (int) (Math.random() * rand);
            String a = null;
            String b = null;
            float j = 0, j2 = 0;

            // convert to integer if integer is used.
            if (type == EventTriplet.TYPE_INT) {
                j = (int) (Math.random() * RANGE);

                do { // find unique j and j2
                    j2 = (int) (Math.random() * RANGE);
                } while (j == j2);

                int tmp = (int) j;
                a = "" + tmp;
                j = (float) tmp;
                tmp = (int) j2;
                b = "" + tmp;
                j2 = (float) tmp;
            } else {
                j = (float) (Math.random() * RANGE);
                do { // find unique j and j2
                    j2 = (float) (Math.random() * RANGE);
                } while (j == j2);

                a = "" + j;
                b = "" + j2;
            }

            if (EQU_ONLY) {
                str += "equ " + a;
                return str;
            }

            if (RANGE_ONLY) {
                k = 1;
            }

            if (k == 0) {
                str += "equ " + a;
            } else if (k == 1) {
                if (j == j2) {
                    System.out.println("Uniqueness error in getConstraint()");
                    System.exit(0);
                } else if (j > j2) {
                    str += "rng " + b + " " + a;
                    //str += "gr " + b + "&lt " + a;
                } else {
                    str += "rng " + a + " " + b;
                    //str += "gr " + a + "&lt " + b;
                }
            } else if (k == 2) {
                str += "gr " + a;
            } else if (k == 3) {
                str += "lt " + a;
            } else if (k == 4) {
                str += "neq " + a;
            } else {
                System.out.println("Filter generation error");
                System.exit(0);
            }

        } else if (type == EventTriplet.TYPE_BOOLEAN) {
            int k = (int) (Math.random() * 2);
            if (k == 0) {
                str += "equ true";
            } else if (k == 1) {
                str += "equ false";
            } else {
                System.out.println("Filter generation error");
                System.exit(0);
            }
        } else if (type == EventTriplet.TYPE_STRING) {
            String co = getRandomFilter();

            if (EQU_ONLY) {
                str += "equ " + co;
                return str;
            }

            if (OMIT_NEQ) {
                rand = 5; // skip neq
            } else {
                rand = 6;
            }

            int k = (int) (Math.random() * rand);

            if (k == 0) {
                str += "equ " + co;
            } else if (k == 1) {
                int i = (int) (Math.random() * Math.max(co.length() - 2, 1));
                i += 3;
                if (i >= co.length()) {
                    i = co.length();
                }
                co = co.substring(0, i);
                str += "in " + co;
            } else if (k == 2) {
                int i = (int) (Math.random() * Math.max(co.length() - 2, 1));
                i += 3;
                if (i >= co.length()) {
                    i = co.length();
                }
                co = co.substring(0, i);
                str += "nin " + co;
            } else if (k == 3) {
                int i = (int) (Math.random() * co.length());
                i += 3;
                if (i >= co.length()) {
                    i = co.length() - 1;
                }
                co = co.substring(0, i);
                str += "stw " + co;
            } else if (k == 4) {
                int i = (int) (Math.random() * co.length());
                i += 3;
                if (i >= co.length()) {
                    i = co.length() - 1;
                }
                co = co.substring(Math.min(co.length() - 3, co.length() - i),
                                  co.length());
                str += "end " + co;
            } else if (k == 5) {
                str += "neq " + co;
            } else {
                System.out.println("Filter generation error");
                System.exit(0);
            }
        }
        return str;

    }

    public static void main(String[] args) {
        FilterServiceTestDriver filterServiceTestDriver1 = new
                FilterServiceTestDriver();
    }
}
