/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.serialization;

import java.util.Calendar;

import fuegocore.notify.core.EventTriplet;
import fuegocore.notify.core.NotificationValue;
import fuegocore.notify.core.Notification;
import fuegocore.notify.core.Template;
import fuegocore.notify.core.channel.EventChannelParams;
import fuegocore.notify.core.channel.ECRRerouteInfo;
import fuegocore.notify.core.filter.Filter;
import fuegocore.notify.core.session.SessionParams;
import fuegocore.util.xas.ContentDecoder;
import fuegocore.util.xas.ChainedContentDecoder;
import fuegocore.util.xas.XmlReader;
import fuegocore.util.xas.EventList;
import fuegocore.util.xas.XasUtil;
import fuegocore.util.Util;

/**
 * A typed content decoder for the types used in the event service.
 * This class extends the {@link ContentDecoder} abstraction for
 * decoding event service objects from XML.
 */
public class EventDecoder extends ChainedContentDecoder {

    public EventDecoder(ContentDecoder chain) {
        super(null);
        if (chain == null) {
            throw new IllegalArgumentException("Chained decoder must be"
                                               + " non-null");
        }
        this.chain = chain;
    }

    public Object decode(String namespace, String name, XmlReader reader,
                         EventList attributes) {
        Object result = null;
        if (Util.equals(namespace, Filter.EVENT_NAMESPACE)) {
            int pos = reader.getCurrentPosition();
            if (Util.equals(name, "triplet")) {
                String etname = (String) expect("", "name", reader);
                if (etname != null) {
                    Object val = expect("", "int", reader);
                    if (val != null) {
                        result =
                                new EventTriplet(etname,
                                                 ((Integer) val).intValue());
                    } else if ((val = expect("", "boolean", reader)) != null) {
                        result =
                                new EventTriplet(etname,
                                                 ((Boolean) val).booleanValue());
                    } else if ((val = expect("", "double", reader)) != null) {
                        result =
                                new EventTriplet(etname,
                                                 ((Double) val).doubleValue());
                    } else if ((val = expect("", "string", reader)) != null) {
                        result = new EventTriplet(etname, (String) val);
                    } else if ((val = expect("", "date", reader)) != null) {
                        result = new EventTriplet(etname,
                                                  ((Calendar) val).getTime());
                    } else if ((val = expect("", "serialized", reader))
                                      != null) {
                        result = new EventTriplet
                                 (etname, EventTriplet.TYPE_SERIALIZED, val);
                    }
                }
            } else if (Util.equals(name, "value")) {
                NotificationValue nv = new NotificationValue();
                EventTriplet et;
                while ((et = (EventTriplet) expect(Filter.EVENT_NAMESPACE,
                        "triplet", reader))
                             != null) {
                    nv.insert(et);
                }
                result = nv;
            } else if (Util.equals(name, "notification")) {
                Notification n = new Notification();
                n.set((NotificationValue) expect("", "value", reader));
                n.updateType((String) expect("", "type", reader));
                n.setAddress((String) expect("", "source", reader));
                Integer i = (Integer) expect("", "order", reader);
                if (i != null) {
                    n.setOrder(i.intValue());
                }
                Long l = (Long) expect("", "deadline", reader);
                if (l != null) {
                    n.setDeadline(l.longValue());
                }
                i = (Integer) expect("", "priority", reader);
                if (i != null) {
                    n.setPriority(i.intValue());
                }
                Boolean b = (Boolean) expect("", "ra", reader);
                if (b != null) {
                    n.setReplaceActive(b.booleanValue());
                }
                n.setReplaceId((String) expect("", "rid", reader));
                n.setSubId((String) expect("", "id", reader));
                n.setACInterface((String) expect("", "ac", reader));
                result = n;
            } else if (Util.equals(name, "template")) {
                Template t = new Template();
                Object o;
                while ((o = expect(Filter.EVENT_NAMESPACE, "af", reader))
                            != null) {
                    t.addFilter(o);
                }
                t.setNotifyFields((String) expect("", "nfields", reader));
                Boolean b = (Boolean) expect("", "localsend", reader);
                if (b != null) {
                    t.setLocalSend(b.booleanValue());
                }
                b = (Boolean) expect("", "adv", reader);
                if (b != null) {
                    t.setAdvertisement(b.booleanValue());
                }
                b = (Boolean) expect("", "quench", reader);
                if (b != null) {
                    t.setQuench(b.booleanValue());
                }
                t.setSubId((String) expect("", "id", reader));
                t.setInterface((String) expect("", "interface", reader));
                t.setSessionId((String) expect("", "sessionid", reader));
                result = t;
            } else if (Util.equals(name, "filter")) {
                String s = (String) expect("", "name", reader);
                Integer i = (Integer) expect("", "type", reader);
                String f = (String) expect("", "filter", reader);
                result = new Filter(s, i.intValue(), f);
            } else if (Util.equals(name, "session")) {
                SessionParams sp = new SessionParams();
                Boolean b = (Boolean) expect("", "od", reader);
                if (b != null) {
                    sp.setOrderedDelivery(b.booleanValue());
                }
                Integer i = (Integer) expect("", "mq", reader);
                if (i != null) {
                    sp.setMaxQueueSize(i.intValue());
                }
                b = (Boolean) expect("", "pe", reader);
                if (b != null) {
                    sp.setPersistency(b.booleanValue());
                }
                b = (Boolean) expect("", "pa", reader);
                if (b != null) {
                    sp.setPause(b.booleanValue());
                }
                b = (Boolean) expect("", "bp", reader);
                if (b != null) {
                    sp.setBlockedPull(b.booleanValue());
                }
                sp.filterLanguage = (String) expect("", "la", reader);
                sp.securityToken = (String) expect("", "sec", reader);
                i = (Integer) expect("", "exp", reader);
                if (i != null) {
                    sp.setExpiration(i.intValue());
                }
                Template t = (Template) expect("", "pf", reader);
                if (t != null) {
                    sp.setPriorityFilter(t);
                }
                result = sp;
            } else if (Util.equals(name, "channel")) {
                EventChannelParams ecp = new EventChannelParams();
                Integer i = (Integer) expect("", "maxsubs", reader);
                if (i != null) {
                    ecp.setMaxSubscribers(i.intValue());
                }
                i = (Integer) expect("", "historysize", reader);
                if (i != null) {
                    ecp.setEventHistorySize(i.intValue());
                }
                Boolean b = (Boolean) expect("", "orderandlog", reader);
                if (b != null) {
                    ecp.setOrderAndLogging(b.booleanValue());
                }
                result = ecp;
                //} else if (Util.equals(name, "reroute")) {
            }
            if (result == null) {
                reader.setCurrentPosition(pos);
            }
        } else if (chain != null) {
            result = chain.decode(namespace, name, reader, attributes);
        }
        return result;
    }

}
