/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.serialization;

import java.util.Calendar;
import java.util.Date;
import java.util.Vector;
import java.util.Enumeration;
import java.io.IOException;

import fuegocore.notify.core.EventTriplet;
import fuegocore.notify.core.NotificationValue;
import fuegocore.notify.core.Notification;
import fuegocore.notify.core.Template;
import fuegocore.notify.core.ITemplate;
import fuegocore.notify.core.channel.EventChannelParams;
import fuegocore.notify.core.channel.ECRRerouteInfo;
import fuegocore.notify.core.filter.Filter;
import fuegocore.notify.core.session.SessionParams;
import fuegocore.util.xas.ContentEncoder;
import fuegocore.util.xas.ChainedContentEncoder;
import fuegocore.util.xas.TypedXmlSerializer;
import fuegocore.util.xas.ContentCodecFactory;
import fuegocore.util.xas.XmlWriter;
import fuegocore.util.xas.Qname;
import fuegocore.util.xas.XasUtil;
import fuegocore.util.Util;

/**
 * A typed content encoder for the types used in the event service.
 * This class implements the {@link ContentEncoder} interface for
 * encoding event service objects into XML.
 */
public class EventEncoder extends ChainedContentEncoder {

    public EventEncoder(ContentEncoder chain) {
        this.chain = chain;
    }

    public boolean encode(Object o, String namespace, String name,
                          TypedXmlSerializer ser) throws IOException {
        boolean result = false;
        if (Util.equals(namespace, Filter.EVENT_NAMESPACE)) {
            if (Util.equals(name, "triplet")) {
                if (o instanceof EventTriplet) {
                    putTypeAttribute(namespace, name, ser);
                    EventTriplet et = (EventTriplet) o;
                    XmlWriter xw = new XmlWriter(ser);
                    xw.typedElement("", "name", XasUtil.XSD_NAMESPACE,
                                    "string", et.getName());
                    String localName = "";
                    String typeNamespace = XasUtil.XSD_NAMESPACE;
                    String typeName = "";
                    Object value = et.getValue();
                    switch (et.getType()) {
                    case EventTriplet.TYPE_INT:
                        localName = "int";
                        typeName = "int";
                        break;
                    case EventTriplet.TYPE_BOOLEAN:
                        localName = "boolean";
                        typeName = "boolean";
                        break;
                    case EventTriplet.TYPE_DOUBLE:
                        localName = "double";
                        typeName = "double";
                        break;
                    case EventTriplet.TYPE_STRING:
                        localName = "string";
                        typeName = "string";
                        break;
                    case EventTriplet.TYPE_DATE:
                        localName = "date";
                        typeName = "dateTime";
                        Calendar c = Calendar.getInstance();
                        c.setTime((Date) value);
                        value = c;
                        break;
                    case EventTriplet.TYPE_SERIALIZED:
                        Qname qname =
                                ContentCodecFactory.getXmlName(value.getClass());
                        if (qname != null) {
                            typeNamespace = qname.getNamespace();
                            typeName = qname.getName();
                        } else {
                            throw new IOException("Cannot serialize type "
                                                  + value.getClass()
                                                  + " with value " + value);
                        }
                        localName = "serialized";
                        break;
                    }
                    xw.typedElement("", localName, typeNamespace, typeName,
                                    value);
                    result = true;
                }
            } else if (Util.equals(name, "value")) {
                if (o instanceof NotificationValue) {
                    putTypeAttribute(namespace, name, ser);
                    NotificationValue nv = (NotificationValue) o;
                    XmlWriter xw = new XmlWriter(ser);
                    Vector data = nv.getVector();
                    for (Enumeration e = data.elements();
                                         e.hasMoreElements(); ) {
                        xw.typedElement(Filter.EVENT_NAMESPACE, "triplet",
                                        Filter.EVENT_NAMESPACE, "triplet",
                                        e.nextElement());
                    }
                    result = true;
                }
            } else if (Util.equals(name, "notification")) {
                if (o instanceof Notification) {
                    putTypeAttribute(namespace, name, ser);
                    Notification n = (Notification) o;
                    XmlWriter xw = new XmlWriter(ser);
                    xw.typedElement("", "value", Filter.EVENT_NAMESPACE,
                                    "value", n.get());
                    xw.typedElement("", "type", XasUtil.XSD_NAMESPACE,
                                    "string", n.getType());
                    if (!n.getHideSource()) {
                        String s = n.getAddress();
                        if (s != null) {
                            xw.typedElement("", "source",
                                            XasUtil.XSD_NAMESPACE, "string",
                                            s);
                        }
                    } else {
                        xw.typedElement("", "source", XasUtil.XSD_NAMESPACE,
                                        "string", "");
                    }
                    xw.typedElement("", "order", XasUtil.XSD_NAMESPACE, "int",
                                    new Integer(n.getOrder()));
                    xw.typedElement("", "deadline", XasUtil.XSD_NAMESPACE,
                                    "long", new Long(n.getDeadline()));
                    xw.typedElement("", "priority", XasUtil.XSD_NAMESPACE,
                                    "int", new Integer(n.getPriority()));
                    xw.typedElement("", "ra", XasUtil.XSD_NAMESPACE, "boolean",
                                    Boolean.valueOf(n.getReplaceActive()));
                    String s = n.getReplaceId();
                    if (s != null) {
                        xw.typedElement("", "rid", XasUtil.XSD_NAMESPACE,
                                        "string", s);
                    }
                    s = n.getSubIdString();
                    if (s != null) {
                        xw.typedElement("", "id", XasUtil.XSD_NAMESPACE,
                                        "string", s);
                    }
                    s = n.getACInterface();
                    if (s != null) {
                        xw.typedElement("", "ac", XasUtil.XSD_NAMESPACE,
                                        "string", s);
                    }
                    result = true;
                }
            } else if (Util.equals(name, "template")) {
                if (o instanceof ITemplate) {
                    putTypeAttribute(namespace, name, ser);
                    ITemplate t = (ITemplate) o;
                    XmlWriter xw = new XmlWriter(ser);
                    Vector filters = t.getFilters();
                    for (Enumeration e = filters.elements();
                                         e.hasMoreElements(); ) {
                        xw.typedElement(Filter.EVENT_NAMESPACE, "af",
                                        Filter.EVENT_NAMESPACE, "filter",
                                        e.nextElement());
                    }
                    String s = t.getNotifyFields();
                    if (s != null) {
                        xw.typedElement("", "nfields", XasUtil.XSD_NAMESPACE,
                                        "string", s);
                    }
                    xw.typedElement("", "localsend", XasUtil.XSD_NAMESPACE,
                                    "boolean",
                                    Boolean.valueOf(t.getLocalSend()));
                    xw.typedElement("", "adv", XasUtil.XSD_NAMESPACE,
                                    "boolean",
                                    Boolean.valueOf(t.getAdvertisement()));
                    xw.typedElement("", "quench", XasUtil.XSD_NAMESPACE,
                                    "boolean", Boolean.valueOf(t.getQuench()));
                    //xw.typedElement("", "expiration", XasUtil.XSD_NAMESPACE,
                    //"string", t.getExpiration());
                    s = t.getSubId();
                    if (s != null) {
                        xw.typedElement("", "id", XasUtil.XSD_NAMESPACE,
                                        "string", s);
                    }
                    s = t.getInterface();
                    if (s != null) {
                        xw.typedElement("", "interface", XasUtil.XSD_NAMESPACE,
                                        "string", s);
                    }
                    s = t.getSessionId();
                    if (s != null) {
                        xw.typedElement("", "sessionid", XasUtil.XSD_NAMESPACE,
                                        "string", s);
                    }
                    result = true;
                }
            } else if (Util.equals(name, "filter")) {
                if (o instanceof Filter) {
                    putTypeAttribute(namespace, name, ser);
                    Filter f = (Filter) o;
                    XmlWriter xw = new XmlWriter(ser);
                    String s = f.getName();
                    if (s != null) {
                        xw.typedElement("", "name", XasUtil.XSD_NAMESPACE,
                                        "string", s);
                    }
                    xw.typedElement("", "type", XasUtil.XSD_NAMESPACE, "int",
                                    new Integer(f.getType()));
                    s = f.convertFilterToInternalForm();
                    if (s != null) {
                        xw.typedElement("", "filter", XasUtil.XSD_NAMESPACE,
                                        "string", s);
                    }
                    result = true;
                }
            } else if (Util.equals(name, "session")) {
                if (o instanceof SessionParams) {
                    putTypeAttribute(namespace, name, ser);
                    SessionParams sp = (SessionParams) o;
                    XmlWriter xw = new XmlWriter(ser);
                    xw.typedElement("", "od", XasUtil.XSD_NAMESPACE, "boolean",
                                    Boolean.valueOf(sp.getOrderedDelivery()));
                    xw.typedElement("", "mq", XasUtil.XSD_NAMESPACE, "int",
                                    new Integer(sp.getMaxQueueSize()));
                    xw.typedElement("", "pe", XasUtil.XSD_NAMESPACE, "boolean",
                                    Boolean.valueOf(sp.getPersistency()));
                    xw.typedElement("", "pa", XasUtil.XSD_NAMESPACE, "boolean",
                                    Boolean.valueOf(sp.getPause()));
                    xw.typedElement("", "bp", XasUtil.XSD_NAMESPACE, "boolean",
                                    Boolean.valueOf(sp.getBlockedPull()));
                    String s = sp.filterLanguage;
                    if (s != null) {
                        xw.typedElement("", "la", XasUtil.XSD_NAMESPACE,
                                        "string", s);
                    }
                    s = sp.securityToken;
                    if (s != null) {
                        xw.typedElement("", "sec", XasUtil.XSD_NAMESPACE,
                                        "string", s);
                    }
                    xw.typedElement("", "exp", XasUtil.XSD_NAMESPACE, "int",
                                    new Integer(sp.getExpiration()));
                    Template t = sp.getPriorityFilter();
                    if (t != null) {
                        xw.typedElement("", "pf", Filter.EVENT_NAMESPACE,
                                        "template", t);
                    }
                    result = true;
                }
            } else if (Util.equals(name, "channel")) {
                if (o instanceof EventChannelParams) {
                    putTypeAttribute(namespace, name, ser);
                    EventChannelParams ecp = (EventChannelParams) o;
                    XmlWriter xw = new XmlWriter(ser);
                    xw.typedElement("", "maxsubs", XasUtil.XSD_NAMESPACE,
                                    "int",
                                    new Integer(ecp.getMaxSubscribers()));
                    xw.typedElement("", "historysize", XasUtil.XSD_NAMESPACE,
                                    "int",
                                    new Integer(ecp.getEventHistorySize()));
                    xw.typedElement("", "orderandlog", XasUtil.XSD_NAMESPACE,
                                    "boolean",
                                    Boolean.valueOf(ecp.getOrderAndLogging()));
                    result = true;
                }
                /*
                 * Commented out, since there are several alternative
                 * serializations; figure out what to do later
                  } else if (Util.equals(name, "reroute")) {
                  if (o instanceof ECRRerouteInfo) {
                  putTypeAttribute(namespace, name, ser);
                  ECRRerouteInfo eri = (ECRRerouteInfo) o;
                  XmlWriter xw = new XmlWriter(ser);
                  result = true;
                  }
                 */
            }
        } else if (chain != null) {
            result = chain.encode(o, namespace, name, ser);
        }
        return result;
    }

}
