/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.serialization;

import java.io.*;
import java.util.*;

import fuegocore.notify.core.serialization.fuego.*;
import fuegocore.util.xas.*;

/** This class serializer event objects (Notifications, Templates)
 *  Add your own handler here (for custom types)
 */
public class EventSerializationService {
    private static EventSerializationService ps;
    private Hashtable vecSerializers; // vector serializers

    public EventSerializationService() {
        vecSerializers = new Hashtable();

        TemplateSerializer ser = new TemplateSerializer();
        NotificationSerializer sen = new NotificationSerializer();
        register("fuegocore.notify.core.Template", ser);
        register("fuegocore.notify.core.Notification", sen);
        register("Template", ser);
        register("Notification", sen);
    }

    /** Add a new serializer to the framework.
     *
     * @param name Name of the serializer
     * @param im Instance of the serializer
     */
    public void register(String name, IEventMultiplexer im) {
        vecSerializers.put(name, im);
    }

    /** Remove a serializer implementation
     *
     * @param name the name of the algorithm
     */
    public void deregister(String name) {
        vecSerializers.remove(name);
    }

    /**
     * Serializes the given object as an Event Sequence
     * @param obj Object
     * @throws IOException
     * @return EventSequence
     */
    public static EventSequence toEventSequence(Object obj) throws IOException {
        if (!(obj instanceof IEventSequenceSerializable)) {
            return null;
        }
        EventSerializer evs = new EventSerializer();
        IEventSequenceSerializable nobj = (IEventSequenceSerializable) obj;
        nobj.toEventSequence(evs);
        return evs.getCurrentSequence();
    }

    /**
     * Serializes the given object as a String
     * @param obj Object
     * @throws IOException
     * @return String
     */
    public static String toString(Object obj) throws IOException {
        if (!(obj instanceof IStringSerializable)) {
            return null;
        }
        return ((IStringSerializable) obj).toString();
    }

    /**
     * Serializes the given Vector as an Event Sequence. Note the objects
     * must be of the same type.
     * @param v Vector
     * @throws IOException
     * @return EventSequence
     */
    public static EventSequence toEventSequence(Vector v) throws IOException {
        if (v == null || v.size() < 1) {
            return null;
        }
        Object obj = v.firstElement();
        if (!(obj instanceof IEventSequenceSerializable)) {
            return null;
        }
        // Get a multiplexer for this class using the class name
        IEventMultiplexer multi = get(obj.getClass().getName());
        if (multi == null) {
            System.out.println("EventSerializationService: could not find " +
                               " multiplexer for: " + obj.getClass().getName());
            System.out.println(getService().vecSerializers);
            return null;
        }
        EventSerializer evs = new EventSerializer();
        multi.multiplexEventSequence(evs, v);
        return evs.getCurrentSequence();
    }

    /**
     * Serializes the given Vector as an Event Sequence. Note the objects
     * must be of the same type.
     * @param v Vector
     * @throws IOException
     * @return String
     */
    public static String toString(Vector v) throws IOException {
        if (v == null || v.size() < 1) {
            return null;
        }
        Object obj = v.firstElement();
        if (!(obj instanceof IStringSerializable)) {
            return null;
        }
        // Get a multiplexer for this class using the class name
        IEventMultiplexer multi = get(obj.getClass().getName());
        if (multi == null) {
            System.out.println("EventSerializationService: could not find " +
                               " multiplexer for: " + obj.getClass().getName());
            return null;
        }

        return multi.multiplex(v);
    }

    /**
     * Demultiplex a Vector of objects from the event sequence.
     * @param ses EventSequence
     * @throws IOException
     * @return Vector
     */
    public static Vector fromEventSequence(EventSequence ses) throws
            IOException {
        if (ses == null) {
            return null;
        }

        Vector ret = null;

        EventSerializationService evs = EventSerializationService.getService();

        // We iterate all multiplexers to find the correct one
        for (Enumeration e = evs.vecSerializers.elements(); e.hasMoreElements(); ) {
            IEventMultiplexer multi = (IEventMultiplexer) e.nextElement();
            ret = multi.demultiplexEventSequence(ses);
            if (ret != null) {
                break;
            }
        }
        return ret;
    }

    /**
     * Demultiplex a single object from the event sequence.
     * @param ses EventSequence
     * @throws IOException
     * @return Object
     */
    public static Object objFromEventSequence(EventSequence ses) throws
            IOException {
        if (ses == null) {
            return null;
        }

        Object ret = null;

        System.out.println("== Object from event sequence");

        EventSerializationService evs = EventSerializationService.getService();
        Enumeration k = evs.vecSerializers.keys();
        // We iterate all multiplexers to find the correct one
        for (Enumeration e = evs.vecSerializers.elements(); e.hasMoreElements(); ) {
            String name = (String) k.nextElement();
            System.out.println("== Trying " + name);

            IEventMultiplexer multi = (IEventMultiplexer) e.nextElement();
            ret = multi.fromEventSequence(ses); //demultiplexEventSequence(ses);
            if (ret != null) {
                break;
            }
        }
        return ret;
    }

    /**
     * Factory method to get a serializer implementation
     * @return IEventMultiplexer
     */
    public static synchronized IEventMultiplexer get() {
        return get("default");
    }

    /**
     * Factory method to get a serializer implementation
     * @return IEventMultiplexer
     */
    public static synchronized EventSerializationService getService() {
        if (ps == null) {
            ps = new EventSerializationService();
        }
        return ps;
    }

    /**
     * Factory method to get a serializer implementation
     * @param name String
     * @return IEventMultiplexer
     */
    public static synchronized IEventMultiplexer get(String name) {
        if (ps == null) {
            ps = new EventSerializationService();
        }
        if (ps.vecSerializers != null) {
            try {
                return (IEventMultiplexer) ps.vecSerializers.get(name);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        return null;
    }
}
