/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.serialization.fuego;

import java.io.*;
import java.util.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.serialization.*;
import fuegocore.util.xas.*;

public class NotificationSerializer implements IEventMultiplexer {
    public NotificationSerializer() {
    }

    /**
     * Demultiplex an object from the given EventSequence
     * @param ses the event sequence
     * @return Object the demultiplexed object
     */
    public Object fromEventSequence(EventSequence ses) {
        Notification n = new Notification();
        if (!n.fromEventSequence(ses)) {
            return null;
        }
        return n;
    }

    /**
     * Creates a string representation of a vector of notifications
     * @param ses EventSequence
     * @throws IOException
     * @return Vector the demultiplexed vector of notifications
     */
    public Vector demultiplexEventSequence(EventSequence ses) throws
            IOException {
        boolean quit = false;
        Vector v = new Vector();

        XmlReader xr = new XmlReader(ses);

        if (xr == null || xr.isFinished()) {
            return null;
        }

        // check the root tag for a filterset
        EventSequence esf = xr.completeContent(Filter.EVENT_NAMESPACE,
                                               "eventset");
        if (esf != null) {
            xr = new XmlReader(esf);
        } else {
            return null;
        }

        while (!quit && !xr.isFinished()) {
            EventSequence es = xr.completeContent(Filter.EVENT_NAMESPACE,
                                                  "event");
            if (es == null) {
                quit = true;
            } else {
                // The constructor expects the element to be there so we add it
                // This could be done more elegantly
                EventSerializer evs = new EventSerializer();
                try {
                    XmlWriter xw = new XmlWriter(evs);
                    xw.complexElement(Filter.EVENT_NAMESPACE, "event",
                                      es);
                } catch (Exception e) {
                    e.printStackTrace();
                }

                Notification n = new Notification(evs.getCurrentSequence());
                if (n != null) {
                    v.add(n);
                }
            }
        }
        return v;
    }

    /**
     * Creates a string representation of a vector of notifications
     * @param ser TypedXmlSerializer
     * @param notifs Vector
     * @throws IOException
     */
    public void multiplexEventSequence(TypedXmlSerializer ser,
                                       Vector notifs) throws IOException {

        XmlWriter xw = new XmlWriter(ser);
        EventSerializer evs = new EventSerializer();
        if (notifs != null) {
            for (int i = 0; i < notifs.size(); i++) {
                ((Notification) notifs.elementAt(i)).toEventSequence(evs);
            }
            xw.complexElement(Filter.EVENT_NAMESPACE, "eventset",
                              evs.getCurrentSequence());
        } else {
            xw.emptyElement(Filter.EVENT_NAMESPACE, "eventset");
        }
    }

    /**
     * Creates a string representation of a vector of notifications
     * @param notifs a Vector of notifications
     * @return String the multiplexed string
     */
    public String multiplex(Vector notifs) {
        String ret = "";
        if (notifs == null) {
            return ret;
        }
        for (int i = 0; i < notifs.size(); i++) {
            Notification t = (Notification) notifs.elementAt(i);
            if (t != null) {
                ret += t.toString();
                ret += "<n_dlm/>";
            }
        }
        return ret;
    }

    /**
     *  Creates a vector representation of multiple notification in a string
     * @param str the String to parse
     * @return Vector the notifications
     */
    public Vector demultiplex(String str) {
        Vector ret = new Vector();

        int i = 0;
        int j = str.indexOf("<n_dlm/>");
        if (j == -1) {
            return null;
        } while (j != -1) {
            String s = str.substring(i, j);
            Notification t = new Notification(s);
            if (t != null) {
                ret.add(t);
            }
            i = j + "<n_dlm/>".length();
            j = str.indexOf("<n_dlm/>", i);
        }
        return ret;
    }

}
