/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.serialization.fuego;

import java.io.*;
import java.util.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.serialization.*;
import fuegocore.util.xas.*;

/**
 * Implements a serializer for Templates
 */
public class TemplateSerializer implements IEventMultiplexer {
    public TemplateSerializer() {
    }

    /**
     * Demultiplex an object from the given EventSequence
     * @param ses the event sequence
     * @return Object the demultiplexed object
     */
    public Object fromEventSequence(EventSequence ses) {
        Template t = new Template();
        if (!t.fromEventSequence(ses)) {
            return null;
        }
        return t;
    }

    /**
     * Demultiplex a Vector of Templates from the given EventSequence
     * @param ses the event sequence
     * @return Vector the demultiplexed vector of templates
     */
    public Vector demultiplexEventSequence(EventSequence ses) {
        XmlReader xr = new XmlReader(ses);
        boolean quit = false;
        Vector v = new Vector();

        if (xr == null || xr.isFinished()) {
            return null;
        }

        // check the root tag for a filterset
        EventSequence esf = xr.completeContent(Filter.EVENT_NAMESPACE,
                                               "filterset");
        if (esf == null) {
            return null;
        }

        xr = new XmlReader(esf);

        while (!quit && !xr.isFinished()) {
            EventSequence es = xr.completeContent(Filter.EVENT_NAMESPACE,
                                                  "filter");
            if (es == null) {
                quit = true;
            } else {
                // The constructor expects the element to be there so we add it
                // This could be done more elegantly
                EventSerializer evs = new EventSerializer();
                try {
                    XmlWriter xw = new XmlWriter(evs);
                    xw.complexElement(Filter.EVENT_NAMESPACE, "filter",
                                      es);
                } catch (Exception e) {
                    e.printStackTrace();
                }

                Template t = new Template(evs.getCurrentSequence());
                if (t != null) {
                    v.add(t);
                }
            }
        }
        return v;
    }

    /**
     *  Creates a string representation of a vector of templates
     * @param ser TypedXmlSerializer
     * @param templates Vector
     * @throws IOException
     */
    public void multiplexEventSequence(TypedXmlSerializer ser,
                                       Vector templates) throws
            IOException {
        if (templates == null) {
            return;
        }

        XmlWriter xw = new XmlWriter(ser);
        EventSerializer evs = new EventSerializer();

        for (int i = 0; i < templates.size(); i++) {
            ((Template) templates.elementAt(i)).toEventSequence(evs);
        }

        xw.complexElement(Filter.EVENT_NAMESPACE, "filterset",
                          evs.getCurrentSequence());
    }

    /**
     * Creates a string representation of a vector of templates
     * @param templates the input templates for multiplexing
     * @return String the multiplexed string
     */
    public String multiplex(Vector templates) {
        String ret = "";
        if (templates == null) {
            return ret;
        }
        for (int i = 0; i < templates.size(); i++) {
            Template t = (Template) templates.elementAt(i);
            if (t != null) {
                ret += t.toString();
                ret += "<template_dlm/>";
            }
        }
        return ret;
    }

    /**
     *  Creates a vector representation of multiple templates in a string
     * @param str the input string for parsing
     * @return Vector the parsed Template objects
     */
    public Vector demultiplex(String str) {
        Vector ret = new Vector();

        int i = 0;
        int j = str.indexOf("<template_dlm/>");
        if (j == -1) {
            return null;
        } while (j != -1) {
            String s = str.substring(i, j);
            Template t = new Template(s);
            if (t != null) {
                ret.add(t);
            }
            i = j + "<template_dlm/>".length();
            j = str.indexOf("<template_dlm/>", i);
        }
        return ret;
    }
}
