/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.server;

import java.util.*;

import fuegocore.notify.comms.*;
import fuegocore.notify.core.session.*;
import fuegocore.notify.util.*;

/**
 * Represents a long lasting connection with a user (a session).
 * ConnectionObject represents the session/connection and
 * provides a mapping to the transport level address. It
 * abstracts the physical address of the event client.
 */
public class ConnectionObject {
    public boolean pauseAfterHandover = false;
    public ServerChannel mtp;
    public SessionParams sp;
    public String address; // address mapping to this user
    public String origin; // address mapping to this user
    public String id;
    public String device; // device id
    public String last_server;
    public boolean handOverFailFlag = false;
    public String handoverFailReason = "";
    int last_seq = -1;
    long handover_timer = 0;
    private Object servletThread; // used for midp http push (callback)

    public void setPausedServletThread(Object obj) {
        servletThread = obj;
    }

    public Object getPausedServletThread() {
        return servletThread;
    }

    /** Create a new object.
     * @param mt the channel object that buffers notifications
     * @param a the identifier
     * @param b the device type
     * @param ls last server identifier
     */
    public ConnectionObject(ServerChannel mt, String a, String b, String ls) {
        mtp = mt;
        id = a;
        device = b;
        last_server = ls;
    }

    public String getOrigin() {
        return origin;
    }

    /**
     * Set the session parameters
     * @param p the parameter object
     */
    public void setSessionParams(SessionParams p) {
        sp = p;
        if (mtp != null) {
            if (sp.getPause()) {
                mtp.pause();
            }
            if (sp.getMaxQueueSize() > 0) {
                mtp.setQueueSize(sp.getMaxQueueSize());
            } else { // keep the old queue size
                sp.setMaxQueueSize(mtp.MAX_QUEUE);
            }
            mtp.setDeliveryMode(sp.getOrderedDelivery());
        }
    }

    /**
     * Get session parameters
     * @return SessionParameter object
     */
    public SessionParams getSessionParams() {
        return sp;
    }

    /** Pause the session
     */
    public void pauseSession() {
        mtp.pause();
    }

    /** Set the address of the client
     * @param adr the address to set
     */
    public void setAddress(String adr) {
        address = adr;
        mtp.setURL(adr, origin); // set transport mapping
    }

    /** Set the address of the client
     * @param adr the address to set
     * @param ori the origin
     */
    public void setAddress(String adr, String ori) {
        address = adr;
        origin = ori;
        mtp.setURL(adr, origin); // set transport mapping
    }

    /** Get the address of the client
     * @return String the address
     */
    public String getAddress() {
        return address;
    }

    /** Continue session (reopen) */
    public void reopenSession() {
        Log.write("********** kicking transport");
        mtp.kick();
    }

    /** Permanently remove session */
    public void destroySession() {
        mtp.destroy();
    }

    public void setLastPingSeq(int seq) {
        last_seq = seq;
    }

    /** Get the last server identifier
     * @return String the last server identifier
     */
    public String getLastServer() {
        return last_server;
    }

    /** Set the last server identifier
     * @param last the last server identifier
     */
    public void setLastServer(String last) {
        last_server = last;
    }

    /** Set the client device information
     * @param dev the device to set
     */
    public void setDevice(String dev) {
        device = dev;
    }

    /** Start handover procedure
     * @param oldserver the old server address
     * @param pah the pauseAfterHandover flag
     */
    public void activateHandover(String oldserver, boolean pah) {
        pauseAfterHandover = pah;
        handOverFailFlag = false;
        //HandoverInitiatorHandler ht = new HandoverInitiatorHandler();
        //ht.startHandoverNoThread(oldserver, id, device);
        //ht.startHandover(oldserver, id, device);
    }

    /** Callback for handover task (end handover)
     * @param oldserver the address of the old server
     */
    public void endHandover(String oldserver) {
        Log.write("Activating MTP after handover");
        if (!pauseAfterHandover) {
            mtp.kick(); // reactivate MTP
        }
        last_server = oldserver;
    }

    /** Callback for handover task for failures
     * @param reason the reason for failure
     */
    public void endHandoverFailed(String reason) {
        try {
            handOverFailFlag = true;
            handoverFailReason = reason;
            mtp.close(); // close socket, because of failure
            // This could be done more gracefully
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /** Get reference to MTP (the underlying event queue for the client)
     * @return ServerChannel reference to the MTP
     */
    public ServerChannel getMTP() {
        return mtp;
    }

    /** Returns all messages in this connection
     * @return Vector the messages in the session
     */
    public Vector getMessages() {
        if (mtp == null) {
            return null;
        }
        if (mtp.getOutbound() == null) {
            return null;
        }
        return mtp.getOutbound();
    }

    public String getID() {
        return id;
    }

    public String getDevice() {
        return device;
    }

    public String printVitalStatistics() {
        String ret = "Connection " + id + "\nDevice " + device + "\n";
        ret += "Message queue out: " + mtp.getOutbound().size() + "\n";
        return ret;
    }

}
