/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.server;

import java.util.*;

import fuegocore.message.*;
import fuegocore.message.protocol.*;
import fuegocore.message.service.*;
import fuegocore.notify.comms.*;
import fuegocore.notify.core.*;
import fuegocore.notify.core.serialization.*;
import fuegocore.notify.core.server.engine.*;
import fuegocore.notify.core.session.*;
import fuegocore.notify.util.*;
import fuegocore.util.xas.*;

/**
 * Access broker implementation.
 * This class contains connection and session management.
 * The routing protocol is abstracted to a RoutingEngine class.
 *
 */
public class EventServer implements fuegocore.message.service.MessageReceiver {

    /** Number of false positives */
    public int strayNotification = 0;
    public int lazyUpdates = 0;
    public int sequenceCount = 2;

    /** Timeout for RPC operations */
    public static long SERVER_TIMEOUT = 5000;

    /** Timeout for handovers operations */
    public static long HANDOVER_TIMEOUT = 20000;

    /** Event channel proxy update */
    public static String MSG_PROXY_UPDATE = "proxy-channel-update";

    /** Router control message */
    public static String MSG_ROUTER_CONTROL = "router-control";

    /** Router control message */
    public static String MSG_NOTIFICATION = "notification";

    /** Number of expired notifications that are received */
    int expiredNotification = 0;

    /** Singleton refererence to the event server */
    public static EventServer es = null;

    /** Buffers incoming notifications */
    Hashtable handoverbuffer = new Hashtable();

    /** mapping from connection numbers to connection objects */
    Hashtable persistent_connections = new Hashtable();

    public int GPRS_TIMEOUT = 50000; // 50 sec timeout for blocked HTTP requests

    public Vector opLog = new Vector();
    public int OPLOGSIZE = 30;

    /** Routing engine implementation */
    IRoutingEngine engine = null;
    IMobilityEngine mengine = null;

    /**
     * Get operations log
     * @return Vector
     */
    public Vector getOpLog() {
        return opLog;
    }

    /**
     * Add log entry
     * @param s String
     */
    public void addOpLogEntry(String s) {
        opLog.add(s);
        if (opLog.size() > OPLOGSIZE) {
            opLog.remove(0);
        }
    }

    /**
     * Set the routing and mobility engines
     * @param engine IRoutingEngine
     */
    public void setRoutingEngine(IRoutingEngine engine) {
        this.engine = engine;
        if (engine instanceof IMobilityEngine) {
            mengine = (IMobilityEngine) engine;
        }
    }

    /**
     * Returns the routing engine
     * @return IRoutingEngine
     */
    public IRoutingEngine getRoutingEngine() {
        return (IRoutingEngine) engine;
    }

    /**
     * Returns mobility engine
     * @return IRoutingEngine
     */
    public IMobilityEngine getMobilityEngine() {
        return mengine;
    }

    /***************************************************************************/
    /** Returns the handover buffer
     *  @return Hashtable the buffer
     */
    public synchronized Hashtable getHandoverBuffer() {
        return handoverbuffer;
    }

    /** Returns an interface to the event server (singleton)
     * Does not create the server.
     * @return the event server handle
     */
    public static synchronized EventServer get() {
        return es;
    }

    /**
     * Create an instance of the server (only if already created)
     * @param engine IRoutingEngine
     * @return EventServer
     */
    public static synchronized EventServer create(IRoutingEngine engine) {
        if (es == null) {
            es = new EventServer();
            engine.setServerReference(es);
            es.setRoutingEngine(engine);
        }
        return es;
    }

    /***************************************************************************/


    /** Returns an interface to the event server (singleton
     * @param port the port number
     * @return the event server handle
     */
    public static synchronized EventServer get(int port) {
        if (es == null) {
            es = new EventServer(port);
        }
        return es;
    }

    // Constructor
    public EventServer() {
    }

    // Constructor
    public EventServer(int port) {
    }

    /***************************************************************************/
    /* -------------------------------------------------------------- */
    /* HANDOVER */

    /** Start buffering of events during a handover for the given client
     *  @param id the interface identifier
     */
    public synchronized void startBuffering(String id) {
        Vector buffer = new Vector();
        handoverbuffer.put(id, buffer);
    }

    /** Insert a new notification to the handover buffer
     *  @param id the interface identifier
     *  @param n the notification object to buffer for the interface
     */
    public synchronized void insertBuffer(String id, Notification n) {
        Vector buffer = (Vector) handoverbuffer.get(id);
        buffer.add(n);
    }

    /** End buffering of events during a handover for the given client
     *  @param id the interface identifier
     *  @return Vector a vector containing the notifications (for merging)
     */
    public synchronized Vector endBuffering(String id) {
        return (Vector) handoverbuffer.remove(id);
    }

    /***************************************************************************/
    /***************************************************************************/
    /***************************************************************************/
    /* GENERAL  */

    /** Publish a notification
     *  @param n the notification to publish
     */
    public void publish(Notification n) {
        n.setACInterface(fuegocore.notify.profile.CommProfile.access_server_id);
        route(n);
    }

    /**
     * Route an event coming from a transport to listeners based on
     * event type and template matching.
     *
     * In merge mode the merge update is triggered by a number of
     * false positives that are detected in this method.
     * @param n the notification object to route
     */
    public void route(Notification n) {
        boolean f = false;

        // check if the event is expired
        if (n.getDeadline() > 0 &&
            (n.getDeadline() - System.currentTimeMillis()) < 0) {
            Log.write("Expired notification, dropping");
            expiredNotification++;
            return;
        }

        Vector allTemplates = new Vector();
        NotificationValue nv = n.get();
        TripletTable h = nv.getTriplets();
        int noti = 0;

        if (h != null) {

            /*
              Description of the access server matching
              We retrieve interface specific list of matched templates
              We go over all templates and format if it is necessary
             If no formatting is necessary we include the names of matched interfaces
              so the client does not need to do matching (multiplexing).
              Finally, if mobility is in progress for the given interface, we
              buffer it for later merging.
             */
            TripletTable v = null;
            v = engine.matchT(n);

            Enumeration e2 = v.getKeys();

            for (Enumeration e = v.elementsNoOrder(); e2.hasMoreElements(); ) {
                Vector v2 = (Vector) e.nextElement();
                String subid = (String) e2.nextElement();
                String no = engine.subIDtoIface(subid);  // convert to iface from subid

                if (no == null)  {
                  System.out.println("Fatal no interface for subid: " + subid);
                  System.exit(0);
                }

                Log.write("---- processing interface: " + no + " and subid: " + subid);
                if (e2.hasMoreElements()) {
                    Log.write("---- there should be more elements");
                }

                try {
                    if (v2 != null) {
                        n.setInterface(no);

                        String tlist = "";
                        boolean first = false;

                        // v2 should have size one
                        // this is legacy code
                        innerloop:for (int i = 0; i < v2.size(); i++) {
                            ITemplate tz = (ITemplate) v2.elementAt(i);
                            if (tz == null) {
                                System.out.println("Route template was null");
                                continue innerloop;
                            }

                            // skip template if we are sending to the source
                            if (!tz.getLocalSend() && no.equals(n.getAddress())) {
                                System.out.println(
                                        "Not sending to source == target");
                                System.out.println("Matched:" + no);
                                continue innerloop;
                            }

                            allTemplates.add(tz);

                            // check if notification field cropping is used, these we cannot send as a single message
                            if (tz.getNotifyFields() != null &&
                                tz.getNotifyFields().length() > 1) {
                                //System.out.println("Formatting notification to template");

                                // Format notification with template
                                Notification n2 = tz.formatNotification(n);
                                n2.setInterface(no);
                                n2.setSubId(tz.getSubId()); // one template at a time if they are prune

                                noti++;

                                if (handoverbuffer.get(no) != null) {
                                    Log.write("Handover buffer active:" + no);
                                    insertBuffer(no, n2);
                                } else {
                                    send(n2);
                                }
                                continue innerloop;
                            } else {
                                Log.write("Doing standard delivery:" + no);
                                first = true;
                            }
                        } // for

                        if (first) {
                            System.out.println("Matched for:" + no);
                            Log.write("Matched for: " + no);
                            // Set the matched identifiers for the client
                            n.setSubId(subid);
                            Notification n3 = new Notification(n);
                            n3.setInterface(no);

                            // If the client is in a handover phase, we buffer the incoming notification
                            // The notification is later merged
                            noti++;
                            if (handoverbuffer.get(no) != null) {
                                Log.write("Handover buffer active:" + no);
                                insertBuffer(no, n3);
                            } else { // send event to proper interface
                                send(n3);
                            }
                        } // first
                    } else { // if v2
                        Log.write("Route: v2 was null for " + no);
                    }
                } catch (Exception ex) {
                    ex.printStackTrace();
                }
                Log.write("end route for loop");
            } // for
            Log.write("end route matching");
        } // if

        /** Callback for demo GUI **/
        if (fuegocore.notify.profile.UserProfile.demoRef != null) {
            fuegocore.notify.profile.UserProfile.demoRef.matchedTemplates(
                    allTemplates, n);

        }
        if (noti == 0) {
            strayNotification++;
        }
    }

    /** Send a message to a client (session)
     * @param n the notification to send
     */
    public void send(Notification n) {
        Log.write("*** Sending message for " + n.getInterface());

        ConnectionObject co = (ConnectionObject) getConnection(n.getInterface());
        if (co == null) {
            co = matchConnectionsList(n.getInterface());
            if (co != null) {
                System.out.println("Connection found using backup mechanism.");
            }
        }
        ServerChannel mt = null;

        if (co != null) {
            mt = co.getMTP();
        }
        if (mt != null) {
            // Replace obsolete notifications/messages from the session queue
            if (n.getReplaceActive() && n.getReplaceId() != null) {
                mt.replaceObsoleteMessages(n.getReplaceId());
            }

            // J2ME special push mode
            System.out.println("Connection for: " + n.getInterface() +
                               " is paused?:" +
                               co.getSessionParams().getBlockedPull());
            System.out.println("Device: " + co.getDevice());

            if (co.getSessionParams().getBlockedPull() &&
                co.getDevice() != null && (co.getDevice().indexOf("cldc") != -1 ||
                                           co.getDevice().indexOf("phone") !=
                                           -1)) {
                System.out.println("------------------ Sending");
                // Add msg to queue, queue is paused so it won't be sent
                mt.send(n);

                System.out.println("Kicking thread: " +
                                   co.getPausedServletThread());

                Object waitingThread = co.getPausedServletThread();
                if (waitingThread != null) {

                    // Kick the servlet thread into action
                    synchronized (waitingThread) {
                        waitingThread.notify();
                    }
                }
            } else {
                System.out.println("************ Attempting to send:" +
                                   mt.getSender() +
                                   "  Origin: " + mt.getOrigin() + " " +
                                   " for: " + n.getInterface());

                mt.send(n);
            }
        } else {
            System.out.println(
                    "Send notification: Message for non-existent interface:[" +
                    n.getInterface() + "]");
            System.out.println(getConnectionsList());
            fuegocore.notify.util.Log.writeTransport(
                    "Send notification: Message for non-existent interface:" +
                    n.getInterface());
            fuegocore.notify.util.Log.writeTransport(getConnectionsList());
        }
    }

    /** Send a message in the correct channel
     * @param co the ConnectionObject
     * @param n the notification to send
     */
    public void send(Notification n, ConnectionObject co) {
        if (n == null || co == null) {
            return;
        }

        ServerChannel mt = null;
        mt = co.getMTP();
        // Replace obsolete notifications/messages from the session queue
        if (n.getReplaceActive() && n.getReplaceId() != null) {
            mt.replaceObsoleteMessages(n.getReplaceId());
        }
        mt.send(n);
    }

    /* -------------------------------------------------------------- */
    /* -------------------------------------------------------------- */
    /* CONNECTION/SESSION management */
    /** Open existing user event session or create a new session on this server
     *  @param name the name of the user/session
     *  @param device the device identifier
     *  @param last the last known access point
     *  @param sp the session parameters
     *  @param address the current communication channel identifier (token)
     *  @return the session identifier
     */
    public String openSession(String name, String device, String last,
                              SessionParams sp, String address) {
        return openSession(name, device, last, sp, address, null);
    }

    /** Open existing user event session or create a new session on this server
     *  @param name the name of the user/session
     *  @param device the device identifier
     *  @param last the last known access point
     *  @param sp the session parameters
     *  @param address the current communication channel identifier (token)
     *  @param origin the origin identifier
     *  @return the session identifier
     */
    public String openSession(String name, String device, String last,
                              SessionParams sp, String address, String origin) {
        String ret = null;
        boolean handover = false;
        boolean pauseAfterHandover = false;

        addOpLogEntry("open " + name);

        Log.write("Open session called");

        System.out.println("Open session called");

        ConnectionObject co = (ConnectionObject) getConnection(name);
        if (co == null) {
            System.out.println("Connection was not found for: [" + name + "]");
            System.out.println(getConnectionsList());

            if (last != null && last != "" &&
                !last.equals("-") &&
                fuegocore.notify.profile.CommProfile.access_server_id.indexOf(
                    "localhost") == -1 &&
                last.indexOf(fuegocore.notify.profile.CommProfile.
                             access_server_id) ==
                -1) {
                // id is ip+port
                handover = true;
            }
            System.out.println("Opening new connection for:[" + name + "]");

            // Create a persistent connection object
            ServerChannel mt = new ServerChannel(name, this, sp);

            // update lookup tables
            co = new ConnectionObject(mt, name, device, last);
            co.setSessionParams(sp);
            co.setAddress(address, origin); // set address

            if (handover) {
                if (sp.getPause()) {
                    pauseAfterHandover = true;
                }
                co.pauseSession();
            }
            synchronized (persistent_connections) {
                persistent_connections.put(name, co);
            }
            // Check that the connection was stored properly
            if (persistent_connections.get(name) == null) {
                System.out.println("Item " + name +
                                   " was not stored in hashtable");
                System.out.println("Critical fault");
                System.exit(0);
            }
            ret = co.getID();
        } else {
            Log.write("Opening an existing session");
            System.out.println("Opening an existing session: new device:" +
                               device +
                               " old :" + co.getDevice());
            // a different device
            if (!co.getAddress().equals(address) && !co.getMTP().getPause()) {
                sendSessionEndpointChange(co.getMTP(), name, address,
                                          co.getAddress());
            }
            // To prevent possible addressing problems due to changing endpoint
            // we update it only for push channels.
            // The address is relevant only for async send
            if (!sp.getPause()) {
                co.setAddress(address, origin); // set address
            }

            if (!handover && !sp.getPause()) {
                co.reopenSession();
            } else if (handover && sp.getPause()) {
                pauseAfterHandover = true;
            }
            co.setSessionParams(sp);
            co.setLastServer(last);
            co.setDevice(device);
            ret = co.getID() + "_reopen";
        }

        if (handover) {
            System.out.println("************** Roaming active");
            // USER HAS ROAMED FROM AN OLD ACCESS SERVER
            fuegocore.notify.util.Log.writeTransport(
                    "Activate handover ---------------------------");
            // blocking handover
            co.activateHandover(last, pauseAfterHandover);
            if (co.handOverFailFlag) {
                return "failure";
            }
        }
        return ret;
    }

    public void sendSessionEndpointChange(ServerChannel _mtp, String name,
                                          String newaddress, String oldaddress) {
        MessageTransportService mts = MessageTransportService.init();
        //CodecIndustry.registerContentFactory(new EventCodecFactory());
        MessageFactory mf = mts.getFactory();
        fuegocore.message.service.Message m2 = mf.createMessage();

        addOpLogEntry("end-point-change " + name);

        Notification n = new Notification();
        n.add(new EventTriplet("type", "session-endpoint-change"));
        n.add(new EventTriplet("session", name));
        n.add(new EventTriplet("old", oldaddress));
        n.add(new EventTriplet("new", newaddress));
        n.setSubId("system");
        n.setInterface(name);

        try {
            //m2.putData("saxdata", EventSerializationService.toEventSequence(n));
            m2.putData("data", n);
            MessageService ms = mts.getService();
            Url url = new Url(Url.Protocol.parse(fuegocore.notify.profile.
                                                 UserProfile.DEF_MSG_PROTOCOL),
                              "message", _mtp.getOrigin());
            m2.setDestination(url);
            ms.sendMessage(m2);
        } catch (Exception e) {
        }
    }

    /** Pull notifications from session.
     *  @param name the name of the user/session
     *  @param device the device identifier
     *  @param temps the templates to pull (Vector)
     *  @param count the number of notifications to pull
     *  @param destroy the destroy flag (if true pulled notifications are destroyed)
     *  @param address the current communication channel identifier (token)
     *  @return Vector the pulled notifications
     */
    public Vector pullSession(String name, String device, Vector temps,
                              int count, boolean destroy, String address) {
        Vector ret = null;
        boolean handover = false;

        addOpLogEntry("pull " + name);

        if (handoverbuffer.get(name) != null) {
            Log.write("Pull: failure handover-in-progress");
            return null; // "failure handover-in-progress";
        }

        if (temps == null || temps.size() == 0) {
            Log.write("Pull: failure invalid-template");
            return null; // "failure invalid-template";
        }

        ConnectionObject co = (ConnectionObject) getConnection(name);
        if (co == null) {
            Log.write("Pull: failure");
            return null; // "failure";
        } else {
            ret = co.getMTP().pull(temps, count, destroy);
        }

        return ret;
    }

    /** Pull notifications from session against subscription identifiers.
     *  @param name the name of the user/session
     *  @param device the device identifier
     *  @param subs the subscription identifiers (order) to pull
     *  @param count the number of notifications to pull
     *  @param destroy the destroy flag (if true pulled notifications are destroyed)
     *  @param address the current communication channel identifier (token)
     *  @return Vector the pulled notifications
     */
    public Vector pullSessionSubs(String name, String device, String subs,
                                  int count, boolean destroy, String address) {
        Vector ret = null;
        boolean handover = false;

        addOpLogEntry("pullSubs " + name);

        if (handoverbuffer.get(name) != null) {
            Log.write("Pull: failure handover-in-progress");
            return null;
        }

        ConnectionObject co = (ConnectionObject) getConnection(name);
        if (co == null) {
            Log.write("Connection object was not found for:" +
                      " [" + name + "]");
        }

        if (co == null) {
            Log.write("Pull: failure");
            return null; // session already exists on this server
        } else {
            Log.write("Pulling " + count + " events using: " + subs);
            ret = co.getMTP().pullSubs(subs, count, destroy);
        }
        return ret;
    }

    /**
     * Move user event session to the current access server
     * @param name the name of the user/session
     * @param last the last known access point
     * @param address the current communication channel identifier (token)
     * @param device the device identifier
     * @param origin String
     * @return String the notifications in string format
     */
    public String moveSession(String name, String last, String address,
                              String device, String origin) {

        addOpLogEntry("move " + name);

        ConnectionObject co = (ConnectionObject) getConnection(name);
        if (co != null) {
            System.out.println("------ already connected in movesession for: " +
                               name);
            return "failure"; // session already exists on this server
        } else {
            // Create a persistent connection object
            ServerChannel mt = new ServerChannel(name, this);

            // Fixme: Note that the session parameters are lost!!!

            // update lookup tables
            co = new ConnectionObject(mt, name, device, last);
            co.setAddress(address,origin); // set address
            //co.setAddress(origin); // set address

            co.setLastServer(last);

            fuegocore.notify.util.Log.writeTransport("Activate handover for " +
                    last +
                    " ---------------------------");
            synchronized (persistent_connections) {
                persistent_connections.put(name, co);
            }
            co.activateHandover(last, co.getSessionParams().getPause());

            if (co.handOverFailFlag) {
                System.out.println("HandoverFailure:" + co.handoverFailReason);
                persistent_connections.remove(name);
                co = null;
                return "failure";
            }
            co.reopenSession();
        }

        return co.getID();
    }

    /** Close user event session
     *  @param name the name of the user/session
     *  @param device the device identifier
     *  @param last the last known access point
     *  @return the session identifier
     */
    public String closeSession(String name, String device, String last) {
        String id = "failure";

        addOpLogEntry("close " + name);

        ConnectionObject co = null;

        System.out.println("Closing session:" + name);

        if (name == null) {
            return id;
        }

        synchronized (persistent_connections) {
            co = (ConnectionObject) persistent_connections.remove(name.trim());
        }

        if (co != null) {
            id = co.getID();
            // We need to unsubscribe all relevant templates locally and from the event channel
            // remove all subscriptions in this session

            Vector temps = (Vector) engine.getTemplatesForId(name).clone();
            for (int i = 0; i < temps.size(); i++) {
                ITemplate tt = (ITemplate) temps.elementAt(i);

                if (fuegocore.notify.profile.UserProfile.subCover &&
                    !tt.getECFlag()) {
                    continue;
                }

                // This could be optimized by sending the vector using a single RPC call
                boolean ret = engine.unsubscribe(tt);
                if (!ret) {
                    co.destroySession();
                    co = null;
                    System.out.println("Failed to unsubscribe: " + tt.toString());
                    return "failure event-channel subscription remove";
                }
            }
        }

        System.out.println("Destroying session");
        co.destroySession();
        co = null;

        return id;
    }

    /** Keep session but reset subscription status
     *  @param name the name of the user/session
     *  @param device the device identifier
     *  @param last the last known access point
     *  @return the session identifier
     */
    public String resetSession(String name, String device, String last) {
        String id = "failure";
        Hashtable channels = new Hashtable();

        addOpLogEntry("reset " + name);

        //System.out.println("Removing templates for connection: " + name);
        //System.out.println("Start: Templates for iface: " +
        //                   FilterService.get().getTemplates(name));
        //System.out.println("Start: " + FilterService.get().getStatistics());
        ConnectionObject co = (ConnectionObject) getConnection(name);
        if (co != null) {
            id = co.getID();
            // We need to unsubscribe all relevant templates locally and from the event channel
            // remove all subscriptions in this session
            Vector temps = (Vector) getRoutingEngine().getTemplatesForId(name).
                           clone();
            int size = temps.size();
            for (int i = 0; i < size; i++) {
                Template tt = (Template) temps.elementAt(i);
                if (tt == null) {
                    continue;
                }

                channels.put(tt.getType(), "true");

                // For standard and covering operation (if template is at the ec):
                // This could be optimized by sending the vector using a single RPC call
                boolean ret = engine.unsubscribe(tt);
                if (!ret) {
                    Log.write("There was a problem with remove sub");
                    return "failure event-channel subscription remove";
                }
            }
        }

        return id;
    }

    /** Pause user event session
     *  @param name the name of the user/session
     *  @param device the device identifier
     *  @param last the last known access point
     *  @param address the current communication channel identifier (token)
     *  @param origin the origin
     *  @return the session identifier
     */
    public String pauseSession(String name, String device, String last,
                               String address, String origin) {

        addOpLogEntry("pause " + name);

        ConnectionObject co = (ConnectionObject) getConnection(name);
        if (co == null) {
            return "failure";
        } else {
            co.setAddress(address,origin);
            //co.setAddress(origin);
            co.pauseSession();
            co.setLastServer(last);
        }
        return co.getID();
    }

    /**
     * Shuts down comms, but keeps queue management. Connection objects may be reused by
     * other sockets.
     * @param name the connection to remove
     */
    public void removeConnection(String name) {
        ConnectionObject co = (ConnectionObject) getConnection(name);
        if (co == null) {
            return;
        }
        ServerChannel mt = co.getMTP();
        if (mt != null) {
            try {
                mt.close();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

    }

    /** Get session/connection object. Checks session expiration.
     * @param name the connection to get
     * @return ConnectionObject
     */
    public ConnectionObject getConnection(String name) {
        ConnectionObject co = (ConnectionObject) persistent_connections.get(
                name);
        if (co != null) {
            co = checkSessionExpiration(co);
        }
        return co;
    }

    /** Get session/connection object. Checks session expiration.
     * @param co the ConnectionObject
     * @return ConnectionObject
     */
    public ConnectionObject checkSessionExpiration(ConnectionObject co) {
        if (co == null) {
            return null;
        }
        if (co.getSessionParams() == null) {
            return co;
        }

        int exp = co.getSessionParams().getExpiration();
        int curtime = (int) System.currentTimeMillis();

        if (exp > 0 && exp - curtime < 0) {
            Log.write("--- Session expired");

            System.out.println("--- Session expired:" + co.getID() +
                               " address:" +
                               co.getAddress());

            closeSession(co.getID(), co.getDevice(), co.getAddress());

            return null;
        }
        return co;
    }

    /**
     * Destroys connection
     * @param name the connection to destroy
     */
    public void destroyConnection(String name) {
        ConnectionObject co = null;
        synchronized (persistent_connections) {
            co = (ConnectionObject) persistent_connections.remove(name);
        }

        ServerChannel mt = co.getMTP();

        if (mt != null) {
            try {
                mt.destroy();
                mt = null;
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

    }

    /** Returns a list of connections
     * @return hashtable containing the connections
     */
    public Hashtable getConnections() {
        return persistent_connections;
    }

    /** Returns a list of connections
     * @return String the list of connections
     */
    public static String getConnectionsList() {
        String ret = "";

        for (Enumeration e = get().persistent_connections.elements();
                             e.hasMoreElements(); ) {
            ConnectionObject co = (ConnectionObject) e.nextElement();
            if (co != null) {
                ret += co.printVitalStatistics() + "\n\n";
            }
        }
        return ret;
    }

    /** Returns a list of connections, debug
     * @param name the name of the connection
     * @return ConnectionObject
     */
    public ConnectionObject matchConnectionsList(String name) {
        String ret = "";
        Enumeration e2 = get().persistent_connections.keys();
        for (Enumeration e = get().persistent_connections.elements();
                             e.hasMoreElements(); ) {
            ConnectionObject co = (ConnectionObject) e.nextElement();
            String k = (String) e2.nextElement();
            if (k.equals(name)) {
                return co;
            }
        }
        return null;
    }

    /** Returns a connection given the id and device
     * @param id the connection id to find
     * @param device the end device
     * @return ConnectionObject
     */
    public synchronized ConnectionObject getConnection(String id, String device) {
        for (Enumeration e = get().persistent_connections.elements();
                             e.hasMoreElements(); ) {
            ConnectionObject co = (ConnectionObject) e.nextElement();
            if (co != null && co.id.equals(id) && co.device.equals(device)) {
                return co;
            }
        }
        return null;
    }

    /* -------------------------------------------------------------- */
    /* -------------------------------------------------------------- */
    /* -------------------------------------------------------------- */

    /** Incoming notifications form event channels
     * @param msg the Message to process
     * @return Message */
    public fuegocore.message.service.Message invoke(fuegocore.message.service.
            Message msg) {

        try {
            EventSequence ses = (EventSequence) msg.getData("saxdata");

            String type_ = (String) msg.getData("type");

            // Process proxy update messages
            if (type_ != null && !type_.equalsIgnoreCase(MSG_NOTIFICATION)) {
                engine.router(msg);
                return null;
            }

            if (ses != null) {
                Notification n = new Notification(ses);
                route(n);
            } else {
                Object data = msg.getData("data");
                Notification n = null;
                if (data instanceof Notification) {
                    n = (Notification) data;
                } else if (data instanceof String) {
                    n = new Notification((String) data);
                }
                if (n != null) {
                    route(n);
                }
            }
        } catch (Exception e) {
            fuegocore.notify.util.Log.write("Error during notification route: " +
                                            e);
            e.printStackTrace();
        }
        return null;
    }

}
