/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.server.engine;

import java.io.*;
import java.net.*;
import java.util.*;

import fuegocore.message.*;
import fuegocore.message.protocol.*;
import fuegocore.message.service.*;
import fuegocore.notify.core.*;
import fuegocore.notify.core.channel.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.filter.poset.*;
import fuegocore.notify.core.serialization.*;
import fuegocore.notify.core.server.*;
import fuegocore.notify.core.server.manager.*;
import fuegocore.notify.core.server.ui.*;
import fuegocore.notify.util.*;
import fuegocore.util.xas.*;

/**
 * Fuego routing engine. Uses event channels in distributed operation.
 */
public class FuegoRoutingEngine implements IRoutingEngine, IMobilityEngine {
    EventServer _es;
    /** Timeout for RPC operations */
    public static long SERVER_TIMEOUT = 5000;

    /** Keeps track of subscriptions sent to event channels. */
    Hashtable ectable = new Hashtable();

    /** Keeps track of advertisements sent to event channels. */
    Hashtable advTable = new Hashtable();

    /** Keeps track of quench messages sent to event channels. */
    Hashtable quenchTable = new Hashtable();

    /** Subscriptions stored by their identifier (SessionId) */
    RoutingTable subscriptionsById = new RoutingTable();

    Hashtable ifacesbySubId = new Hashtable();

    /** Advertisements stored by their identifier (SessionId) */
    RoutingTable advertisementsById = new RoutingTable();

    /** Quench stored by their identifier (SessionId) */
    RoutingTable quenchById = new RoutingTable();

    /** Handover manager */
    HandoverManager haManager = new HandoverManager();

    public int mergeUpdateCounter = 0;

    /** Router html interface */
    FuegoUI routerUi = new FuegoUI();

    /** Event channel manager */
    RouterManager manager = new RouterManager();

    /* *************************************************************** */
    /**
     * Get the channel manager reference
     * @return ChannelManager
     */
    public RouterManager getRouterManager() {
        return manager;
    }

    public String subIDtoIface(String subid) {
        return (String)ifacesbySubId.get(subid);
    }

    /* *************************************************************** */
    /**
     * Called for messages that are not handled in the general access
     * server handler. This implements, for example, event channel specific
     * operations.
     * @param msg Message
     * @throws IOException
     * @return Object
     */
    public Object defaultBackAccessServerHandler(Message msg) throws
            IOException {

        String address = (String) msg.getData(fuegocore.message.service.
                                              PropertyNames.MESG_DATA_SENDER);
        String method = (String) msg.getData("op");

        if (method == null || method.length() < 1) {
            Log.write("DefaultAccessServerHandler: method was missing!");
            return "failure method missing";
        }

        Object a1 = (Object) msg.getData("attr1");
        Object a2 = (Object) msg.getData("attr2");
        Object a3 = (Object) msg.getData("attr3");
        Object a4 = (Object) msg.getData("attr4");
        Object a5 = (Object) msg.getData("attr5");

        if (method.equals("createEventChannel")) {
            return new Boolean(createEventChannelRPC((String) a1, (String) a2));
        } else if (method.equals("removeEventChannel")) {
            return new Boolean(removeEventChannelRPC((String) a1));
        } else if (method.equals("getEventChannelLog")) {
            return getEventChannelLogRPC();
        } else if (method.equals("pullEventChannelHistory")) {
            Vector v2 = null;
            if (a3 != null) {
                Vector v = EventSerializationService.fromEventSequence((
                        EventSequence)
                        a2);
                Integer ia3 = (Integer) a3;
                v2 = pullEventChannelHistory((String) a1,
                                             v,
                                             ia3.intValue(), address);
            }
            if (v2 != null) {
                return EventSerializationService.toEventSequence(v2);
            }
        } else if (method.equals("pullEventChannelHistoryOrder")) {
            Vector v2 = null;
            Integer ia3 = (Integer) a3;
            v2 = pullEventChannelHistoryOrder((String) a1, (String) a2,
                                              ia3.intValue(), address);
            if (v2 != null) {
                return EventSerializationService.toEventSequence(v2);
            }
        }

        // pass message to mobility handler
        Object ret = mobilityMessageHandler(msg);
        if (ret == null) {
            return "failure routing-engine-handler";
        }
        return ret;
    }

    /**
     * Mobility handler
     * @param msg Message
     */
    public Object mobilityMessageHandler(fuegocore.message.service.Message msg) {
        return null;
    }

    /* *************************************************************** */
    /**
     * Start mobility
     * @param client String
     * @param oset Vector
     * @param nset Vector
     * @return String
     */
    public String startMobility(String client, Vector oset, Vector nset) {
        return null;
    }

    /**
     *
     * @param session String
     */
    public void handoverFailure(String session) {

    }

    /**
     * Returns the handover manager
     * @return HandoverManager
     */
    public HandoverManager getHandoverManager() {
        return haManager;
    }

    /* *************************************************************** */

    /**
     * Returns the filter poset
     * @param s String event type
     * @return IFilterPoset
     */
    public IFilterPoset getPoset(String s) {
        IFilterPoset pos = (IFilterPoset) ectable.get(s);
        if (pos == null) {
            pos = PosetService.get(); // get default implementation
            ectable.put(s, pos);
        }
        return pos;
    }

    /**
     * Returns the advertisement poset
     * @param s String event type
     * @return IFilterPoset
     */
    public IFilterPoset getAdvertisementPoset(String s) {
        IFilterPoset pos = (IFilterPoset) advTable.get(s);
        if (pos == null) {
            pos = PosetService.get(); // get default implementation
            advTable.put(s, pos);
        }
        return pos;
    }

    /**
     * Returns the quench poset
     * @param s String event type
     * @return IFilterPoset
     */
    public IFilterPoset getQuenchPoset(String s) {
        IFilterPoset pos = (IFilterPoset) quenchTable.get(s);
        if (pos == null) {
            pos = PosetService.get(); // get default implementation
            quenchTable.put(s, pos);
        }
        return pos;
    }

    /**
     * Set the poset
     * @param s String event type
     * @param pos IFilterPoset
     */
    public void setPoset(String s, IFilterPoset pos) {
        ectable.put(s, pos);
    }

    public FuegoRoutingEngine() {
    }

    public FuegoRoutingEngine(EventServer es) {
        _es = es;
    }

    /**
     * Set server reference
     * @param es EventServer
     */
    public void setServerReference(EventServer es) {
        _es = es;
    }

    /**
     * Return subscriptions by interface identifier
     * @param name String
     * @return Vector
     */
    public Vector getTemplatesForId(String name) {
        return subscriptionsById.get(name);
    }

    /**
     * Return advertisements by interface identifier
     * @param name String
     * @return Vector
     */
    public Vector getAdvertisementsForId(String name) {
        return advertisementsById.get(name);
    }

    /**
     * Return quench by interface identifier
     * @param name String
     * @return Vector
     */
    public Vector getQuenchForId(String name) {
        return quenchById.get(name);
    }

    /**
     * Match the notification against the routing tables (local operation only)
     * @param n Notification
     * @return Vector
     */
    public Vector match(Notification n) {
        // get poset for the type
        IFilterPoset pos = getPoset(n.getType());
        if (pos == null) {
            return null;
        }
        return pos.match(n);
    }

    /**
     * Match the notification against the routing tables (local operation only)
     * @param n Notification
     * @return TripletTable
     */
    public TripletTable matchT(Notification n) {
        // get poset for the type
        IFilterPoset pos = getPoset(n.getType());
        if (pos == null) {
            return null;
        }
        return pos.matchT(n);
    }

    /**
     * Advertise
     * @param t Template
     * @return boolean
     */
    public boolean advertise(ITemplate t) {
        _es.addOpLogEntry("advertise " + t.getType());
        return false;
    }

    /**
     * Unadvertise
     * @param t Template
     * @return boolean
     */
    public boolean unadvertise(ITemplate t) {
        _es.addOpLogEntry("unadvertise " + t.getType());
        return false;
    }

    /**
     * Quench. Quench semantics: client receives a notification
     * when the given template has subscribers (or when the last subscriber
     * unsubscribes)
     * @param t Template
     * @return boolean
     */
    public boolean quench(ITemplate t) {
        _es.addOpLogEntry("quench " + t.getType());
        // we simply subscribe the template (it is flagged as quench)
        // The channel knows when there are subscribers
        return this.processSubAdvQuench(t);
    }

    /**
     * Unquench. Quench semantics: client receives a notification
     * when the given template has subscribers (or when the last subscriber
     * unsubscribes)
     * @param t Template
     * @return boolean
     */
    public boolean unquench(ITemplate t) {
        _es.addOpLogEntry("unquench " + t.getType());
        // we simply unsubscribe the template (it is flagged as quench)
        // The channel knows when there are subscribers
        return this.removeSubAdvQuench(t);
    }

    /**
     * Router control update message handler
     * @param msg Message
     */
    public void router(fuegocore.message.service.Message msg) {
        try {
            EventSequence ses = (EventSequence) msg.getData("saxdata");
            String type_ = (String) msg.getData("type");

            // Process proxy update messages
            if (type_ != null &&
                type_.equalsIgnoreCase(EventServer.MSG_PROXY_UPDATE)) {
                Log.write("Got proxy update: " + ses.toString());
                Vector v = EventSerializationService.fromEventSequence(ses);
                proxyChannelUpdate(v);
            } else if (type_ != null &&
                       type_.equalsIgnoreCase(EventServer.MSG_ROUTER_CONTROL)) {
                // Insert handler here
            }
        } catch (Exception e) {
            fuegocore.notify.util.Log.write("Error during notification route: " +
                                            e);
            e.printStackTrace();
        }
    }

    /** Proxy channel update message from the corresponding event channel
     *  @param t a Vector of templates
     */
    private synchronized void proxyChannelUpdate(Vector t) {
        Log.write("**** proxy channel update");
        Hashtable resetFlag = new Hashtable();

        EventChannelServer ecs = EventChannelServer.get();

        if (t == null) {
            Log.write("Template vector was null");
            // We reset all proxies
            ecs.resetAllProxies();
            return;
        }

        for (int i = 0; i < t.size(); i++) {
            Template t2 = (Template) t.elementAt(i);
            if (t2 == null) {
                continue;
            }
            String ch = t2.getType();
            String iface = t2.getInterface();
            if (ch == null || iface == null) {
                continue;
            }
            Log.write("updating: " + t.toString());
            EventChannel ec = (EventChannel) ecs.lookupProxy(ch.trim());
            if (ec == null) {
                // Create a new proxy channel
                ecs.createProxyChannel(ch.trim(), null);
                ec = (EventChannel) ecs.lookupProxy(ch.trim());
            }
            if (resetFlag.get(ch + "-" + iface) == null) {
                ec.resetSubscriptions(iface);
                resetFlag.put(ch + "-" + iface, ec);
            }
            ec.subscribe(t2);
        }
        /*
             // remove channels
             Hashtable cur = ecs.getEventChannelProxies();
             cur.clear();
         for (Enumeration j = resetFlag.elements(); j.hasMoreElements(); ) {
          Log.write("Removing channel");
          EventChannel ec2 = (EventChannel) j.nextElement();
          cur.put(ec2.getName(), ec2);
          ec2 = null;
             }
         */
    }

    /* RPC SUBSCRIPTION MANAGEMENT */

    /** This is called for incoming publish messages
     * @param n the notification to publish
     * @return true if the operation succeeded
     */
    public boolean publish(Notification n) {

        _es.addOpLogEntry("publish " + n.getType());

        n.setACInterface(fuegocore.notify.profile.CommProfile.access_server_id);

        // CHECK IF LOCAL (FAST)
        EventChannelServer ecs = EventChannelServer.get();

        // We check if the channel is temporary and the proxy is
        // available
        EventChannel proxy = ecs.lookupProxy(n.getType());
        if (proxy != null) {
            // access server dispatches
            _es.route(n); // local delivery
            return proxy.publish(n); // external delivery
        }

        // Regular event channel publish
        EventChannel ec = ecs.lookup(n.getType());
        if (ec != null) {
            return ec.publish(n);
        }

        URL url = (URL) getRouterManager().lookup((String) n.getType());
        if (url == null) {
            Log.write("Event channel type _es.lookup failed:" + n.getType());
            return false;
        }

        try {
            sendMessageToEC("publish", url.getHost(), url.getPort(), n);
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }
        return true;
    }

    /**
     * Send async message to an event channel
     * @param op String
     * @param host String
     * @param port int
     * @param ses EventSequence
     */
    private void sendMessageToEC(String op, String host, int port, Object o) {
        sendMessageToEC(op, host, port, o, null);

    }

    /**
     * Send async message to an event channel
     * @param op String
     * @param host String
     * @param port int
     * @param ses EventSequence
     * @param ses2 EventSequence
     */
    private void sendMessageToEC(String op, String host, int port,
                                 Object o1, Object o2) {
        // open one-shot connection
        try {
            Log.writeEC("Adding proxy message to outbound queue");

            MessageTransportService mts = MessageTransportService.init();
            CodecIndustry.registerContentFactory(new EventCodecFactory());
            MessageFactory mf = mts.getFactory();
            fuegocore.message.service.Message m2 = mf.createMessage();
            if (o1 != null) {
                m2.putData("saxdata", o1);
            }
            if (o2 != null) {
                m2.putData("saxdata2", o2);
            }
            m2.putData("type", "access-server-msg");
            m2.putData("operation", op);
            m2.putData("origin",
                       fuegocore.notify.profile.CommProfile.access_server_id);

            m2.setDestination(new Url(Url.Protocol.parse(fuegocore.notify.
                    profile.
                    UserProfile.DEF_MSG_PROTOCOL),
                                      "message_ec",
                                      Url.toDestination(host, port)));
            MessageService ms = mts.getService();
            ms.sendMessage(m2);
        } catch (Exception e) {
            Log.writeEC("Exception:" + e);
        }
    }

    /** This is called for incoming subscribe messages.
     *  Fast local _es.lookups are used if the event channel is resident in the same
     *  virtual machine.
     *  @param t the template to subscribe
     *  @return true if the operation succeeded
     */
    public boolean subscribe(ITemplate t) {
        _es.addOpLogEntry("subscribe " + t.getType());
        return processSubAdvQuench(t);
    }

    /**
     * Method for processing subs, advs, and quench messages
     * @param t Template
     * @return boolean
     */
    public boolean processSubAdvQuench(ITemplate t) {
        if (t.getInterface() == null) {
            Log.write("Interface was null");
            return false;
        }

        IFilterPoset pos = null;

        if (t.getAdvertisement()) {
            pos = getAdvertisementPoset(t.getType());
        } else if (t.getQuench()) {
            pos = getQuenchPoset(t.getType());
        } else {
            pos = getPoset(t.getType());

        }
        if (pos.removable((IFBase) t, t.getSubId())) {
            Log.write(
                    "A subscription with the same ID was ALREADY in the system");
            return false;
        }

        // Add filter to poset
        // Note that we add using the sub identifier in order to do
        // server-side matching for all clients, this is done only
        // for clientsm router susbscriptions are processed elsewhere
        int rcode = FilterForest.POSET_ERROR;

        try {
            rcode = pos.add((IFBase) t, t.getSubId());
        } catch (PosetException e) {
            rcode = FilterForest.POSET_ERROR;
            e.printStackTrace();
        }

        if (rcode == FilterForest.POSET_ERROR) {
            Log.write("Subscribe: filter poset error: " + rcode);
            return false;
        }

        if (t.getAdvertisement()) {
            quenchById.insert(t.getInterface(), t);
        } else if (t.getQuench()) {
            advertisementsById.insert(t.getInterface(), t);
        } else {
            subscriptionsById.insert(t.getInterface(), t);
            ifacesbySubId.put(t.getSubId(), t.getInterface());
            // Check covering
        }
        if (rcode != FilterForest.POSET_ROOT) {
            Log.write("Subscribe: template covered: " + rcode);
            /*
               try {
                    PrintWriter fout = new PrintWriter(new FileWriter(
                        "testposet.dat"), false);
                    ((FilterForest)pos).printDotFull(fout);
                    fout.close();
                  }
                  catch (Exception e2) {
                    e2.printStackTrace();
                  }
             */
            return true;
        }

        // Set flag to indicate that this template has been sent
        t.setECFlag(true);

        // CHECK IF LOCAL (FAST)

        EventChannelServer ecs = EventChannelServer.get();
        EventChannel ec = ecs.lookup(t.getType());
        if (ec != null) {
            boolean ret = false;
            Template t2 = null;
            try {
                t2 = (Template) ((Template) t).clone();
            } catch (Exception e) {}
            t2.setInterface(fuegocore.notify.profile.CommProfile.
                            access_server_id);

            if (t.getAdvertisement()) {
                ret = ecs.advertise(t2);
            } else if (t.getQuench()) {
                ret = ecs.quench(t2);
            } else {
                ret = ecs.subscribe(t2);

            }
            return ret;
        }

        System.out.println("******** remote sub/adv/quench");

        URL url = (URL) getRouterManager().lookup(t.getType());
        if (url == null) {
            Log.write("Event channel type lookup failed:" + t.getType());
            return false;
        }

        try {
            sendMessageToEC("processSubAdvQuench", url.getHost(), url.getPort(),
                            t);
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }

        return true;
    }

    /**
     * Returns the class responsible for router html UI
     * @return Class
     */
    public RouterUI getUI() {
        return routerUi;
    }

    /** This is called for incoming unsubscribe messages.
     *  Fast local lookups are used if the event channel is resident in the same
     *  virtual machine.
     *  @param t the template to subscribe
     *  @return true if the operation succeeded
     */
    public boolean unsubscribe(ITemplate t) {
        _es.addOpLogEntry("unsubscribe " + t.getType());
        return removeSubAdvQuench(t);
    }

    /**
     * Remove sub, adv, or quench from the system
     * @param t ITemplate
     * @return boolean
     */
    public boolean removeSubAdvQuench(ITemplate t) {
        Log.write("Removing subscription");

        IFilterPoset pos = null;

        if (t.getAdvertisement()) {
            pos = getAdvertisementPoset(t.getType());
        } else if (t.getQuench()) {
            pos = getQuenchPoset(t.getType());
        } else {
            pos = getPoset(t.getType());

        }
        if (!pos.removable((IFBase) t, t.getSubId())) {
            Log.write("Cannot remove template!");
            return false;
        }

        // Get reference to the original template
        Template t0 = (Template) pos.lookup((IFBase) t);

        /** remove filter */
        int rcode = FilterForest.POSET_ERROR;
        try {
            rcode = pos.remove((IFBase) t, t.getSubId());
        } catch (PosetException e) {
            rcode = FilterForest.POSET_ERROR;
            e.printStackTrace();
        }

        if (rcode == FilterForest.POSET_ERROR) {
            Log.write("Unsubscribe: POSET ERROR");
            return false;
        }

        if (t.getAdvertisement()) {
            advertisementsById.removeUsingContent(t.getInterface(), t);
        } else if (t.getQuench()) {
            quenchById.removeUsingContent(t.getInterface(), t);
        } else {
            subscriptionsById.removeUsingContent(t.getInterface(), t);
            ifacesbySubId.remove(t.getSubId());
            // Do not update ec if template has not been send and is not a
            // root node
        }
        if (rcode != FilterForest.POSET_ROOT && !t0.getECFlag()) {
            return true;
        }

        // Set to add in atomic update operation
        Vector addSet = new Vector();

        // Special case, a root filter was removed
        if (rcode == FilterForest.POSET_ROOT) {
            Log.write("A root filter was removed");
            Vector v = pos.get();
            synchronized (v) {
                for (int i = 0; i < v.size(); i++) {
                    Template _t = (Template) v.elementAt(i);
                    if (!_t.getECFlag()) {
                        addSet.add(_t);
                    }
                }
            }
        }

        t0.setECFlag(false);

        // CHECK IF LOCAL (FAST)

        // This version sends the unsubscribe directly to event channel
        EventChannelServer ecs = EventChannelServer.get();
        EventChannel ec = ecs.lookup(t.getType());
        if (ec != null) {
            Template t2 = null;
            try {
                t2 = (Template) ((Template) t).clone();
            } catch (Exception e) {
                e.printStackTrace();
            }
            t2.setInterface(fuegocore.notify.profile.CommProfile.
                            access_server_id);

            boolean ret = false;
            if (t.getAdvertisement()) {
                ret = ecs.unadvertise(t2, addSet);
            } else if (t.getQuench()) {
                ret = ecs.unquench(t2, addSet);
            } else {
                ret = ecs.unsubscribe(t2, addSet);
            }

            if (ret) {
                // mark uncovered templates as sent
                for (int i = 0; i < addSet.size(); i++) {
                    Template _t = (Template) addSet.elementAt(i);
                    _t.setECFlag(true);
                }
            }

            Log.write("EC returned: " + ret);
            return ret;
        }

        // remote

        URL url = (URL) getRouterManager().lookup(t.getType());
        if (url == null) {
            Log.write("Event channel type lookup failed");
            return false;
        }

        try {
            sendMessageToEC("removeSubAdvQuench", url.getHost(), url.getPort(),
                            t,
                            addSet);
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }
        // mark uncovered templates as sent
        for (int i = 0; i < addSet.size(); i++) {
            Template _t = (Template) addSet.elementAt(i);
            _t.setECFlag(true);
        }

        return true;
    }

    /* -------------------------------------------------------------- */
    /* RPC EVENT CHANNEL MANAGEMENT */

    /** Create Event Channel
     * @param type the name/identifier of the event channel to create
     * @param params the event channel params (EventChannelParams object in string format)
     * @return true if the operation succeeded
     */
    public boolean createEventChannelRPC(String type, String params) {
        _es.addOpLogEntry("createEC " + type);

        if (type == null) {
            return false;
        }

        // Get server
        URL url = manager.getEventChannel(type);
        if (url == null) {
            return false;
        }
        // CHECK IF LOCAL (FAST)
        EventChannelServer ecs = EventChannelServer.get();
        if (ecs.getURL().equals(url)) {
            return ecs.createChannel(type, params);
        }

        // Make SOAP RPC
        try {
            NameValuePair[] parameters = new NameValuePair[] {
                                         new NameValuePair("arg1", "String"),
                                         new NameValuePair("arg2", "String")
            };
            NameValuePair[] arguments = new NameValuePair[] {
                                        new NameValuePair("arg1", type),
                                        new NameValuePair("arg2", params)
            };

            MessageTransportService mts = MessageTransportService.init();
            MessageService ms = mts.getService();
            RpcCall call = ms.createCall("createChannel", parameters, "Boolean");

            Boolean reply = (Boolean) call.synchInvoke(new Url(Url.Protocol.
                    parse(
                            fuegocore.notify.profile.UserProfile.
                            DEF_MSG_PROTOCOL), "rpc_ec",
                    Url.toDestination(url.getHost(), url.getPort())), arguments,
                    SERVER_TIMEOUT);

            Log.write("Reply was: " + reply.booleanValue());

            if (!reply.booleanValue()) {
                return false;
            } else {
                // Create proxy channel
                ecs.createProxyChannel(type, params);
            }
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }

        return true;
    }

    /** Remove Event Channel
     * @param type the name/identifier of the event channel to remove
     * @return true if the operation succeeded
     */
    public boolean removeEventChannelRPC(String type) {
        _es.addOpLogEntry("removeEC " + type);

        // Get server
        URL url = manager.getEventChannel(type);
        if (url == null) {
            return false;
        }
        // CHECK IF LOCAL (FAST)
        EventChannelServer ecs = EventChannelServer.get();
        if (ecs.getURL().equals(url)) {
            return ecs.removeChannel(type);
        }

        // Make SOAP RPC
        try {
            NameValuePair[] parameters = new NameValuePair[] {
                                         new NameValuePair("arg1", "String")
            };
            NameValuePair[] arguments = new NameValuePair[] {
                                        new NameValuePair("arg1", type)
            };
            MessageTransportService mts = MessageTransportService.init();
            MessageService ms = mts.getService();
            RpcCall call = ms.createCall("removeChannel", parameters, "Boolean");

            Boolean reply = (Boolean) call.synchInvoke(new Url(Url.Protocol.
                    parse(
                            fuegocore.notify.profile.UserProfile.
                            DEF_MSG_PROTOCOL), "rpc_ec",
                    Url.toDestination(url.getHost(), url.getPort())), arguments,
                    SERVER_TIMEOUT);

            Log.write("Reply was: " + reply.booleanValue());
            if (!reply.booleanValue()) {
                return false;
            } else {
                // Create proxy channel
                ecs.removeProxyChannel(type);
            }
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }
        return true;
    }

    /** Get Event Channel Information from all resolution servers.
     * @return String the event channel information/log
     */
    public String getEventChannelLogRPC() {
        String ret = "";
        Hashtable ab = (Hashtable) manager.getChannels().clone();
        for (Enumeration k = ab.elements(); k.hasMoreElements(); ) {
            URL url = (URL) k.nextElement();
            if (url == null) {
                continue;
            }

            // CHECK IF LOCAL (FAST)
            EventChannelServer ecs = EventChannelServer.get();
            if (ecs.getURL().equals(url)) {
                ret += ecs.channelLog(null);
                continue;
            }

            // Make SOAP RPC
            try {
                NameValuePair[] parameters = new NameValuePair[] {
                                             new NameValuePair("arg1", "String")
                };
                NameValuePair[] arguments = new NameValuePair[] {
                                            new NameValuePair("arg1", null)
                };

                MessageTransportService mts = MessageTransportService.init();
                MessageService ms = mts.getService();
                RpcCall call = ms.createCall("channelLog", parameters, "String");

                String reply = (String) call.synchInvoke(new Url(Url.Protocol.
                        parse(
                                fuegocore.notify.profile.UserProfile.
                                DEF_MSG_PROTOCOL), "rpc_ec",
                        Url.toDestination(url.getHost(), url.getPort())),
                        arguments,
                        SERVER_TIMEOUT);

                ret += reply;
            } catch (Exception e) {
                e.printStackTrace();
            }
            ret += "\n\n";
        }
        return ret;
    }

    /* -------------------------------------------------------------- */
    /* -------------------------------------------------------------- */
    /* RPC SUBSCRIPTION MANAGEMENT: AGGREGATE RPC CALLS */

    /** Publish aggregate.
     *  Return string that contains failure code (the number of first failed
     *  notification in the given list)
     * @param v the aggregate
     * @return String the return value
     */
    public String publish(Vector v) {
        String ret = "failure";
        if (v == null) {
            return ret;
        }

        _es.addOpLogEntry("publishAll " + v.size());

        try {
            for (int i = 0; i < v.size(); i++) {
                Notification n = (Notification) v.elementAt(i);
                if (!publish(n)) {
                    ret = "failure " + i;
                    return ret;
                }
            }
            ret = "success";
        } catch (Exception e) {
            e.printStackTrace();
        }
        return ret;
    }

    /** Subscribe aggregate.
     *  Return string that contains failure code (the number of first failed
     *  notification in the given list)
     * @param v the aggregate
     * @return String the return value
     */
    public String subscribe(Vector v) {
        String ret = "failure";
        if (v == null) {
            return ret;
        }

        _es.addOpLogEntry("subscribeAll " + v.size());

        try {
            Log.write("SubscribeAll: size was " + v.size());
            for (int i = 0; i < v.size(); i++) {
                Template t = (Template) v.elementAt(i);
                if (!subscribe(t)) {
                    ret = "failure " + i;
                    // we rollback
                    for (int j = 0; j < i; j++) {
                        Template tt = (Template) v.elementAt(j);
                        unsubscribe(tt);
                    }
                    return ret;
                }
            }
            ret = "success";
        } catch (Exception e) {
            e.printStackTrace();
        }
        return ret;
    }

    /** Unsubscribe aggregate.
     *  Return string that contains failure code (the number of first failed
     *  notification in the given list). Currently this may cause multiple
     *  updates to event channels if non-lazy updates are used with the merging
     *  mode.
     * @param v the aggregate
     * @return String the return value
     */
    public String unsubscribe(Vector v) {
        String ret = "failure";
        boolean fail = false;
        int count = 0;
        if (v == null) {
            return ret;
        }

        _es.addOpLogEntry("unsubscribeAll " + v.size());

        try {
            for (int i = 0; i < v.size(); i++) {
                Template t = (Template) v.elementAt(i);
                if (!unsubscribe(t)) {
                    fail = true;
                    ret += i + " ";
                } else {
                    count++;
                }
            }
            if (!fail) {
                ret = "success";
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return ret;
    }

    /** Advertise aggregate.
     *  Return string that contains failure code (the number of first failed
     *  notification in the given list)
     * @param v the aggregate
     * @return String the return value
     */
    public String advertise(Vector v) {
        String ret = "failure";
        if (v == null) {
            return ret;
        }

        _es.addOpLogEntry("advertiseAll " + v.size());

        try {
            for (int i = 0; i < v.size(); i++) {
                Template t = (Template) v.elementAt(i);
                if (!advertise(t)) {
                    ret = "failure " + i;
                    // we rollback
                    for (int j = 0; j < i; j++) {
                        Template tt = (Template) v.elementAt(j);
                        unadvertise(tt);
                    }
                    return ret;
                }
            }
            ret = "success";
        } catch (Exception e) {
            e.printStackTrace();
        }
        return ret;
    }

    /** Unadvertise aggregate.
     *  Return string that contains failure code (the number of first failed
     *  notification in the given list).
     * @param v the aggregate
     * @return String the return value
     */
    public String unadvertise(Vector v) {
        String ret = "failure";
        boolean fail = false;
        int count = 0;
        if (v == null) {
            return ret;
        }

        _es.addOpLogEntry("unadvertiseAll " + v.size());

        try {
            for (int i = 0; i < v.size(); i++) {
                Template t = (Template) v.elementAt(i);
                if (!unadvertise(t)) {
                    fail = true;
                    ret += i + " ";
                } else {
                    count++;
                }
            }
            if (!fail) {
                ret = "success";
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return ret;
    }

    /** Quench aggregate.
     *  Return string that contains failure code (the number of first failed
     *  notification in the given list)
     * @param v the aggregate
     * @return String the return value
     */
    public String quench(Vector v) {
        String ret = "failure";
        if (v == null) {
            return ret;
        }

        _es.addOpLogEntry("quenchAll " + v.size());

        try {
            for (int i = 0; i < v.size(); i++) {
                Template t = (Template) v.elementAt(i);
                if (!quench(t)) {
                    ret = "failure " + i;
                    // we rollback
                    for (int j = 0; j < i; j++) {
                        Template tt = (Template) v.elementAt(j);
                        unquench(tt);
                    }
                    return ret;
                }
            }
            ret = "success";
        } catch (Exception e) {
            e.printStackTrace();
        }
        return ret;
    }

    /** Unquench aggregate.
     *  Return string that contains failure code (the number of first failed
     *  notification in the given list).
     * @param v the aggregate
     * @return String the return value
     */
    public String unquench(Vector v) {
        String ret = "failure";
        boolean fail = false;
        int count = 0;
        if (v == null) {
            return ret;
        }

        _es.addOpLogEntry("unquenchAll " + v.size());

        try {
            for (int i = 0; i < v.size(); i++) {
                Template t = (Template) v.elementAt(i);
                if (!unquench(t)) {
                    fail = true;
                    ret += i + " ";
                } else {
                    count++;
                }
            }
            if (!fail) {
                ret = "success";
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return ret;
    }

    /** Pull notifications from event channel history.
     *  @param channel the event channel name
     *  @param template templates
     *  @param count the number of notifications to pull
     *  @param address the current communication channel identifier (token)
     *  @return Vector the notifications
     */
    public Vector pullEventChannelHistory(String channel, Vector template,
                                          int count, String address) {
        EventSequence ret = null;
        Vector v = null;

        _es.addOpLogEntry("pullEC " + channel);

        if (template == null) {
            Log.write("pullEC: failure invalid-template");
            return null; //"failure invalid-template";
        }

        // Get server
        URL url = manager.getEventChannel(channel);
        if (url == null) {
            Log.write("pullEC: failure no-channel");
            return null; // "failure no-channel";
        }

        // CHECK IF LOCAL (FAST)
        EventChannelServer ecs = EventChannelServer.get();
        if (ecs.getURL().equals(url)) {
            return ecs.pull(channel, EventSerializationService.get("Template").
                            multiplex(template), count);
        }

        // Make SOAP RPC
        try {
            NameValuePair[] parameters = new NameValuePair[] {
                                         new NameValuePair("arg1", "String"),
                                         new NameValuePair("arg2", "String"),
                                         new NameValuePair("count", "Integer")
            };
            NameValuePair[] arguments = new NameValuePair[] {
                                        new NameValuePair("arg1", channel),
                                        new NameValuePair("arg2", template),
                                        new NameValuePair("count",
                    new Integer(count))
            };
            MessageTransportService mts = MessageTransportService.init();
            MessageService ms = mts.getService();
            RpcCall call = ms.createCall("pull", parameters, "String");

            ret = (EventSequence) call.synchInvoke(new Url(Url.Protocol.parse(
                    fuegocore.
                    notify.profile.UserProfile.DEF_MSG_PROTOCOL), "rpc_ec",
                    Url.toDestination(url.getHost(),
                                      url.getPort())), arguments,
                    SERVER_TIMEOUT);
            v = EventSerializationService.fromEventSequence(ret);
        } catch (Exception e) {
            e.printStackTrace();
            v = null;
        }
        return v;
    }

    /** Pull notifications from event channel history.
     *  @param channel the event channel name
     *  @param identifiers the notification identifiers (order)
     *  @param count the number of notifications to pull
     *  @param address the current communication channel identifier (token)
     *  @return Vector the notifications
     */
    public Vector pullEventChannelHistoryOrder(String channel,
                                               String identifiers,
                                               int count, String address) {
        EventSequence ret = null;
        Vector v = null;

        _es.addOpLogEntry("pullECO " + channel);

        if (identifiers == null) {
            Log.write("pullEC: failure invalid-identifier");
            return null;
        }

        // Get server
        URL url = manager.getEventChannel(channel);
        if (url == null) {
            Log.write("pullEC: failure no-channel");
            return null; // "failure no-channel";
        }

        // CHECK IF LOCAL (FAST)
        EventChannelServer ecs = EventChannelServer.get();
        if (ecs.getURL().equals(url)) {
            return ecs.pullOrder(channel, identifiers, count);
        }

        // Make SOAP RPC
        try {
            NameValuePair[] parameters = new NameValuePair[] {
                                         new NameValuePair("arg1", "String"),
                                         new NameValuePair("arg2", "String"),
                                         new NameValuePair("count", "Integer")
            };
            NameValuePair[] arguments = new NameValuePair[] {
                                        new NameValuePair("arg1", channel),
                                        new NameValuePair("arg2", identifiers),
                                        new NameValuePair("count",
                    new Integer(count))
            };
            MessageTransportService mts = MessageTransportService.init();
            MessageService ms = mts.getService();
            RpcCall call = ms.createCall("pullOrder", parameters, "String");

            ret = (EventSequence) call.synchInvoke(new Url(Url.Protocol.parse(
                    fuegocore.
                    notify.profile.UserProfile.DEF_MSG_PROTOCOL), "rpc_ec",
                    Url.toDestination(url.getHost(),
                                      url.getPort())), arguments,
                    SERVER_TIMEOUT);
            v = EventSerializationService.fromEventSequence(ret);
        } catch (Exception e) {
            e.printStackTrace();
            v = null;
        }
        return v;
    }

}
