/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.server.engine;

import java.io.*;
import java.util.*;

import fuegocore.message.service.*;
import fuegocore.notify.core.*;
import fuegocore.notify.core.server.*;
import fuegocore.notify.core.server.manager.*;
import fuegocore.notify.core.server.ui.*;
import fuegocore.notify.util.*;

/**
 * Interface for routing engines. Incorporates methods for both event
 * channel based communication and basic pub/sub.
 */
public interface IRoutingEngine {

    /**
     * Set server reference
     * @param es EventServer
     */
    public void setServerReference(EventServer es);

    /**
     * Match the notification against the routing tables (local operation only)
     * @param n Notification
     * @return Vector
     */
    public Vector match(Notification n);

    /**
     * Match the notification against the routing tables (local operation only)
     * @param n Notification
     * @return TripletTable
     */
    public TripletTable matchT(Notification n);

    /**
     * Return templates by interface identifier
     * @param name String
     * @return Vector
     */
    public Vector getTemplatesForId(String name);

    /**
     * Return advertisements by interface identifier
     * @param name String
     * @return Vector
     */
    public Vector getAdvertisementsForId(String name);

    /**
     * Return quench by interface identifier
     * @param name String
     * @return Vector
     */
    public Vector getQuenchForId(String name);

    /**
     * Router control update message handler
     * @param msg Message
     */
    public void router(fuegocore.message.service.Message msg);

    /**
     * Get the channel manager reference
     * @return ChannelManager
     */
    public RouterManager getRouterManager();


    public String subIDtoIface(String subid);

    /**
     * Called for messages that are not handled in the general access
     * server handler. This implements, for example, event channel specific
     * operations.
     * @param msg Message
     * @return Object
     */
    public Object defaultBackAccessServerHandler(Message msg) throws
            IOException;

    /**
     * Subscribe
     * @param t Template
     * @return boolean
     */
    public boolean subscribe(ITemplate t);

    /**
     * Unsubscribe
     * @param t Template
     * @return boolean
     */
    public boolean unsubscribe(ITemplate t);

    /**
     * Subscribe
     * @param v Vector
     * @return String
     */
    public String subscribe(Vector v);

    /**
     * Unsubscribe
     * @param v Vector
     * @return String
     */
    public String unsubscribe(Vector v);

    /**
     * Advertise
     * @param t Template
     * @return boolean
     */
    public boolean advertise(ITemplate t);

    /**
     * Unadvertise
     * @param t Template
     * @return boolean
     */
    public boolean unadvertise(ITemplate t);

    /**
     * Advertise
     * @param v Vector
     * @return String
     */
    public String advertise(Vector v);

    /**
     * Unadvertise
     * @param v Vector
     * @return String
     */
    public String unadvertise(Vector v);

    /**
     * Quench. Quench semantics: client receives a notification
     * when the given template has subscribers (or when the last subscriber
     * unsubscribes)
     * @param t Template
     * @return boolean
     */
    public boolean quench(ITemplate t);

    /**
     * Unquench. Quench semantics: client receives a notification
     * when the given template has subscribers (or when the last subscriber
     * unsubscribes)
     * @param t Template
     * @return boolean
     */
    public boolean unquench(ITemplate t);

    /**
     * Quench. Quench semantics: client receives a notification
     * when the given template has subscribers (or when the last subscriber
     * unsubscribes)
     * @param v Vector
     * @return String
     */
    public String quench(Vector v);

    /**
     * Unquench. Quench semantics: client receives a notification
     * when the given template has subscribers (or when the last subscriber
     * unsubscribes)
     * @param v Vector
     * @return String
     */
    public String unquench(Vector v);

    /**
     * Returns an instance responsible for router html UI
     * @return RouterUI
     */
    public RouterUI getUI();

    /**
     * Publish
     * @param n Notification
     * @return boolean
     */
    public boolean publish(Notification n);

    /**
     * Publish
     * @param v Vector
     * @return String
     */
    public String publish(Vector v);

}
