/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.server.handlers;

import java.util.*;

import fuegocore.message.service.*;
import fuegocore.notify.core.*;
import fuegocore.notify.core.serialization.*;
import fuegocore.notify.core.server.*;
import fuegocore.notify.core.server.engine.*;
import fuegocore.notify.core.session.*;
import fuegocore.notify.util.*;
import fuegocore.util.xas.*;

/**
 * Access Server Handler that accepts incoming SOAP messages and
 * dispatches them to the proper components of the access server.
 * This class uses the serialization service for interpreting
 * incoming data.
 */
public class AccessServerHandler implements MessageReceiver, RpcReceiver {
    EventServer es;
    IRoutingEngine ire;

    public AccessServerHandler(EventServer es) {
        this.es = es;
        ire = es.getRoutingEngine();
    }

    /**
     * SOAP RPC handler for event channel
     * @param method String
     * @param arguments NameValuePair[]
     * @return Object
     */
    public Object invoke(String method, NameValuePair[] arguments) {
        return null;
    }

    /**
     * Message handler
     * @param msg Message
     * @return Message
     */
    public Message invoke(Message msg) {

        MessageFactory mf = msg.getFactory();
        System.out.println(mf);
        Message reply = mf.createMessage();

        String address = (String) msg.getData(fuegocore.message.service.
                                              PropertyNames.MESG_DATA_SENDER);

        String origin = (String) msg.getData(fuegocore.message.service.
                                             PropertyNames.MESG_DATA_ORIGIN);

        String method = (String) msg.getData("op");

        Log.write("Method " + method + " called");
        Log.write("address: " + address + " origin: " + origin);

        if (method == null || method.length() < 1) {
            System.out.println("method was missing!");
            return reply;
        }

        Object a1 = (Object) msg.getData("attr1");
        Object a2 = (Object) msg.getData("attr2");
        Object a3 = (Object) msg.getData("attr3");
        Object a4 = (Object) msg.getData("attr4");
        Object a5 = (Object) msg.getData("attr5");

        try {
            if (method.equals("subscribe")) {
                if (a1 != null && a1 instanceof ITemplate) {
                    ITemplate t = (ITemplate) a1;
                    if (t.getAdvertisement()) {
                        reply.putData("data", new Boolean(ire.advertise(t)));
                    } else if (t.getQuench()) {
                        reply.putData("data", new Boolean(ire.quench(t)));
                    } else {
                        //System.out.println("Subscribing");
                        reply.putData("data", new Boolean(ire.subscribe(t)));
                        //System.out.println("Response: " + reply.getData("data"));
                    }
                }
            } else if (method.equals("publish")) {
                if (a1 != null && a1 instanceof Notification) {
                    Notification n = (Notification) a1;
                    reply.putData("data", new Boolean(ire.publish(n)));
                }
            } else if (method.equals("publishAll")) {
                if (a1 != null && a1 instanceof Vector) {
                    Vector v = (Vector) a1;
                    String ret = ire.publish(v);
                    if (ret != null) {
                        reply.putData("data", ret);
                    }
                }
            } else if (method.equals("subscribeAll")) {
                if (a1 != null && a1 instanceof Vector) {
                    Vector v = (Vector) a1;
                    ITemplate t0 = (ITemplate) v.firstElement();
                    if (t0.getAdvertisement()) {
                        reply.putData("data", ire.advertise(v));
                    } else if (t0.getQuench()) {
                        reply.putData("data", ire.quench(v));
                    } else {
                        reply.putData("data", ire.subscribe(v));
                    }
                }
            } else if (method.equals("unsubscribeAll")) {
                if (a1 != null && a1 instanceof Vector) {
                    Vector v = (Vector) a1;
                    ITemplate t0 = (ITemplate) v.firstElement();
                    if (t0.getAdvertisement()) {
                        reply.putData("data", new Boolean(ire.unadvertise(v)));
                    } else if (t0.getQuench()) {
                        reply.putData("data", new Boolean(ire.unquench(v)));
                    } else {
                        reply.putData("data", new Boolean(ire.unsubscribe(v)));
                    }
                }
            } else if (method.equals("unsubscribe")) {
                if (a1 != null && a1 instanceof ITemplate) {
                    ITemplate t = (ITemplate) a1;

                    if (t.getAdvertisement()) {
                        reply.putData("data", new Boolean(ire.unadvertise(t)));
                    } else if (t.getQuench()) {
                        reply.putData("data", new Boolean(ire.unquench(t)));
                    } else {
                        reply.putData("data", new Boolean(ire.unsubscribe(t)));
                    }
                }
            }
            /*
                  else if (method.equals("startHandoverPhase1")) {
                    return startHandoverPhase1(arg1, arg2, arg3);
                  }
                  else if (method.equals("startHandoverPhase2")) {
                    return startHandoverPhase2(arg1, arg2, arg3, flag);
                  }
                  else if (method.equals("startHandoverPhase3")) {
                    return startHandoverPhase3(arg1, arg2, arg3);
                  }
             */
            else if (method.equals("openSession")) {
                String ret = es.openSession((String) a1, (String) a2,
                                            (String) a3,
                                            (SessionParams) a4, address, origin);
                if (ret != null) {
                    reply.putData("data", ret);
                }
            } else if (method.equals("resetSession")) {
                String ret = es.resetSession((String) a1, (String) a2,
                                             (String) a3);
                if (ret != null) {
                    reply.putData("data", ret);
                }
            } else if (method.equals("pauseSession")) {
                String ret = es.pauseSession((String) a1, (String) a2,
                                             (String) a3,
                                             address, origin);
                if (ret != null) {
                    reply.putData("data", ret);
                }
            } else if (method.equals("closeSession")) {
                String ret = es.closeSession((String) a1, (String) a2,
                                             (String) a3);
                if (ret != null) {
                    reply.putData("data", ret);
                }
            } else if (method.equals("moveSession")) {
                String ret = es.moveSession((String) a1, (String) a2, address,
                                            (String) a3, origin);
                if (ret != null) {
                    reply.putData("data", ret);
                }
            } else if (method.equals("pullSession")) {
                Vector v2 = null;
                if (a3 != null && a3 instanceof Vector) {
                    Vector v = (Vector) a3;
                    Integer ia4 = (Integer) a4;
                    Boolean ba5 = (Boolean) a5;
                    v2 = es.pullSession((String) a1, (String) a2, v,
                                        ia4.intValue(),
                                        ba5.booleanValue(), address);
                }
                if (v2 != null) {
                    reply.putData("data", v2);
                }
            } else if (method.equals("pullSessionSubs")) {
                Vector v2 = null;
                if (a3 != null) {
                    Integer ia4 = (Integer) a4;
                    Boolean ba5 = (Boolean) a5;
                    v2 = es.pullSessionSubs((String) a1, (String) a2,
                                            (String) a3,
                                            ia4.intValue(), ba5.booleanValue(),
                                            address);
                }
                if (v2 != null) {
                    reply.putData("data", v2);
                }
            } else { // check if the routing core knows how to handle the invoke
                Object obj = ire.defaultBackAccessServerHandler(msg);
                reply.putData("data", obj);
            }
        } catch (Exception e) {
            e.printStackTrace();
            reply.putData("data", "failure access-server-handler");
        }
        return reply;
    }
}
