/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.server.handlers;

import java.net.*;
import java.util.*;

import fuegocore.message.*;
import fuegocore.message.protocol.*;
import fuegocore.message.service.*;
import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.server.*;
import fuegocore.notify.util.*;
import fuegocore.notify.core.filter.merge.*;
import fuegocore.notify.core.*;

/**
 * Handover task, coordinates handover from the new server.
 * This is from Prototype I (2003) and should be integrated with the
 * current codebase (2004).
 * <br>
 * 1. contact old server<br>
 * 2. old server relocates all subscriptions relevant for the client and updates channels<br>
 * 3. old server returns client session data <br>
 * 4  handover is complete<br>
 * 5. transport is activated<br>
 */

public class HandoverTask
    extends Thread {
  int state = INACTIVE;
  public static final int INACTIVE = 0;
  public static final int ACTIVE = 1;
  public static final int HANDOVER_1 = 2;
  public static final int HANDOVER_2 = 3;
  public static final int HANDOVER_3 = 4;
  ConnectionObject _co;
  String old_as, new_as, id, device;
  EventServer _es = EventServer.get();

  public HandoverTask(ConnectionObject co) {
    _co = co;
    state = INACTIVE;
  }

  /** Starts the handover procedure with a new thread (non-blocking)
   * @param old the old server
   * @param idd the session identifier
   * @param dev the device
   */
  public void startHandover(String old, String idd, String dev) {
    old_as = old;
    id = idd;
    device = dev;
    new_as = fuegocore.notify.profile.CommProfile.access_server_id;
    state = ACTIVE;
    start();
  }

  /** Start a handover procedure without initiating a thread for waiting
   *  the handover to complete (blocking)
   * @param old the old server
   * @param idd the session identifier
   * @param dev the device
   */
  public void startHandoverNoThread(String old, String idd, String dev) {
    old_as = old;
    id = idd;
    device = dev;
    new_as = fuegocore.notify.profile.CommProfile.access_server_id;
    state = ACTIVE;
    run();
  }

  /** Executes handover protocol */
  public void run() {
      /*
    // #1 contact old access server
    boolean success = false;
    URL url = null;
    long timer1 = System.currentTimeMillis();

    Log.writeEC("------------ start handover: " + old_as);

    // start buffering of notifications for the given id
    _es.startBuffering(id);

    // Make SOAP RPC to the access server
    try {
      NameValuePair[] parameters = new NameValuePair[] {
          new NameValuePair("arg1", "String"),
          new NameValuePair("arg2", "String"),
          new NameValuePair("arg3", "String")
      };

      NameValuePair[] arguments = new NameValuePair[] {
          new NameValuePair("arg1", new_as),
          new NameValuePair("arg2", id),
          new NameValuePair("arg3", device)
      };

      MessageTransportService mts = MessageTransportService.init();
      MessageService ms = mts.getService();
      RpcCall call = ms.createCall("startHandoverPhase1", parameters, "String");

      Log.writeEC("------------ Making call");

      MessageProtocol mp = null;

      if (fuegocore.notify.profile.UserProfile.useBaseProtocol) {
        mp = ms.getProtocol("base");
      }
      else {
        mp = mts.getProtocol();

      }
      String reply = null;
      try {
        reply = (String) call.synchInvoke(mp.constructUrl(
            fuegocore.notify.util.Auxil.getHost(old_as),
            fuegocore.notify.util.Auxil.getPort(old_as), "rpc"),
                                          arguments, EventServer.SERVER_TIMEOUT);
      }
      catch (fuegocore.message.service.TimeoutException e) {
        reply = "event channel failure";
      }

      Log.writeEC("------------ Call completed");

      if (reply == null || reply.equals("invalid connection") ||
          reply.equals("event channel failure")) {
        _es.endBuffering(id);
        Log.writeEC("Error: invalid connection: " + reply + " with session: " +
                    id);
        success = false;
      }
      else {
        // extract notifications and templates from the reply and convert them to messages
        //int i = reply.indexOf("<template_delimiter_tag/>");
        //String tmpls = reply.substring(0,i);
        Vector templates = parseTemplates(reply);
        int coverCount = 0;

        // register to filter service
        if (templates != null) {
          for (int j = 0; j < templates.size(); j++) {
            Template t = (Template) templates.elementAt(j);
            //System.out.println("registering template: " + t.toString());
            // Interface name should remain unchanged
            FilterService.get().addFilter(t.getFilters(), id, t);
            // We update the EC status also
            if (fuegocore.notify.profile.UserProfile.mergeEC) {
              if (_es.handoverSubscribeMerge(t))
                coverCount++;
            }
            // No need to do anything special for the covering mode
            // It is sufficient to just add the subscriptions locally
          }
        }
        boolean covered = false;

        // Check if the templates are covered
        if (fuegocore.notify.profile.UserProfile.mergeEC && templates.size() > 0) {
          if (coverCount >= templates.size())
            covered = true;
        }
        else if (fuegocore.notify.profile.UserProfile.subCover) {
         covered = _es.ECCover(templates);
        }

        parameters = new NameValuePair[] {
            new NameValuePair("arg1", "String"),
            new NameValuePair("arg2", "String"),
            new NameValuePair("arg3", "String"),
            new NameValuePair("flag", "Boolean")
        };

        arguments = new NameValuePair[] {
            new NameValuePair("arg1", new_as),
            new NameValuePair("arg2", id),
            new NameValuePair("arg3", device),
            new NameValuePair("flag", new Boolean(covered))
        };

        call = ms.createCall("startHandoverPhase2", parameters, "String");

        Log.writeEC("------------ Making second call");
        if (fuegocore.notify.profile.UserProfile.useBaseProtocol) {
          mp = ms.getProtocol("base");
        }
        else {
          mp = mts.getProtocol();
        }
        try {
          reply = (String) call.synchInvoke(mp.constructUrl(
              fuegocore.notify.util.Auxil.getHost(old_as),
              fuegocore.notify.util.Auxil.getPort(old_as), "rpc"),
                                            arguments,
                                            EventServer.SERVER_TIMEOUT);
        }
        catch (fuegocore.message.service.TimeoutException e) {
          reply = "event channel failure";
        }

            // Check for event channel update failure. At this point
            // if the event channel update fails, we have not yet committed the
            //     handover.
        if (reply == null || reply.equals("invalid connection") ||
            reply.equals("event channel failure")) {
          success = false;
          _es.endBuffering(id); // this may cause some lost notifications if
          // some of the event channel updates failed
          _co.endHandoverFailed("event channel failure");
          return;
        }

        Log.writeEC("------------ Call completed");
        // Second invocation end

        Log.writeEC("------------ parsing session data: " + reply);

        Vector msg = parseMessages(reply);

        // Add messages to queue (in the right order)
        if (msg != null) {
         // _co.getMTP().getOutbound().addData(msg);
          _co.getMTP().mergeNewMessages(msg);
        }
        success = true;
      }
    }
    catch (Exception e) {
      e.printStackTrace();
    }

    // #3 merge events that arrived during handover

    // stop buffering of notifications for the given id
    if (success) {
      Log.writeEC("------------ hand over success");

      Log.writeEC("-----------------------------------------");
      Vector notifs = _es.endBuffering(id);

      if (notifs != null) {
        // Allow processing of new messages, this does not affect
        // the sender thread (different from the pause method).
        _co.getMTP().setPause(false);
        // send the notifications, these are processed oldest first and
        // latest last, so the remove-notification feature works
        for (int i = 0; i < notifs.size(); i++) {
          Notification n = (Notification) notifs.elementAt(i);
          _es.send(n, _co);
        }
        Log.writeEC("Got " + notifs.size() + " buffered notifications at the new access server");
        // #4 activate transport
      }
      _co.endHandover(old_as);
      Log.writeEC("-----------------------------------------");

      // Final RPC call, remove old sessions

      // Make SOAP RPC to the access server
      NameValuePair[] parameters = new NameValuePair[] {
          new NameValuePair("arg1", "String"),
          new NameValuePair("arg2", "String"),
          new NameValuePair("arg3", "String")
      };

      NameValuePair[] arguments = new NameValuePair[] {
          new NameValuePair("arg1", new_as),
          new NameValuePair("arg2", id),
          new NameValuePair("arg3", device)
      };

      MessageTransportService mts = MessageTransportService.init();
      MessageService ms = mts.getService();
      RpcCall call = ms.createCall("startHandoverPhase3", parameters,
                                   "String");

      Log.writeEC("------------ Making call");

      MessageProtocol mp = null;

      if (fuegocore.notify.profile.UserProfile.useBaseProtocol) {
        mp = ms.getProtocol("base");
      }
      else {
        mp = mts.getProtocol();
      }
      String reply = null;
      try {
        reply = (String) call.synchInvoke(mp.constructUrl(
            fuegocore.notify.util.Auxil.getHost(old_as),
            fuegocore.notify.util.Auxil.getPort(old_as), "rpc"),
                                          arguments, EventServer.SERVER_TIMEOUT);
      }
      catch (Exception e) {
        reply = null;
        e.printStackTrace();
      }

      if (reply == null || reply.equals("invalid connection")) {
        // If remove failed it is not a critical failure for event delivery
        // the remove-operation should be retried.
      }
    }
    else {
      Log.writeEC("------------ hand over failed");
      // At this point the event channel has been updated so we cannot
      // abort the handover totally.  Events will be delivered to this
      // address
      _co.endHandoverFailed("failure after event channel update");
    }
*/
  }
}
