/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.server.handlers;

import java.util.*;

import fuegocore.message.*;
import fuegocore.message.service.*;
import fuegocore.notify.core.*;
import fuegocore.notify.core.client.*;
import fuegocore.notify.core.serialization.*;
import fuegocore.notify.core.server.*;
import fuegocore.notify.core.server.engine.*;
import fuegocore.notify.core.session.*;
import fuegocore.notify.core.serialization.EventCodecFactory;
import fuegocore.util.xas.CodecIndustry;

/**
 * Implements a handler for lightweight requests for MIDP phones
 * Uses the EventServer and the RoutingEngine. This currently
 * uses only String-based serialization.
 */
public class LightweightHandler implements ILightweightRPCHandler {
    EventServer _es;
    IRoutingEngine ire;

    public LightweightHandler() {
        _es = EventServer.get();
        ire = _es.getRoutingEngine();
    }

    /** Lightweight RPC handler
     * @param mr the MicroRequest object that encapsulates the RPC parameters
     * @param address the address of the client
     * @return String the response
     */
    public Object processRequest(MicroRequest mr, String address) {
        Object ret = "";
        boolean b = false;

        fuegocore.notify.core.serialization.fuego.TemplateSerializer tser =
                new fuegocore.notify.core.serialization.fuego.
                TemplateSerializer();

        fuegocore.notify.core.serialization.fuego.NotificationSerializer nser =
                new fuegocore.notify.core.serialization.fuego.
                NotificationSerializer();

        if (mr.op.equals("subscribe")) {
            Template t = null;
            if (mr.payload instanceof String) {
                t = (Template)new Template((String) mr.payload);
            } else {
                t = (Template) ((Object) mr.payload);
            }

            t.setInterface(mr.id);

            if (t.getAdvertisement()) {
                b = ire.advertise(t);
            } else if (t.getQuench()) {
                b = ire.quench(t);
            } else {
                b = ire.subscribe(t);
            }
            if (b) {
                return "ok";
            } else {
                return "failure";
            }
        } else if (mr.op.equals("unsubscribe")) {
            Template t = null;
            if (mr.payload instanceof String) {
                t = new Template((String) mr.payload);
            } else {
                t = (Template) ((Object) mr.payload);
            }

            t.setInterface(mr.id);

            if (t.getAdvertisement()) {
                b = ire.unadvertise(t);
            } else if (t.getQuench()) {
                b = ire.unquench(t);
            } else {
                b = ire.unsubscribe(t);
            }
            if (b) {
                return "ok";
            } else {
                return "failure";
            }
        } else if (mr.op.equals("publish")) {
            Notification n = null;
            if (mr.payload instanceof String) {
                n = (Notification)new Notification((String) mr.payload);
            } else {
                n = (Notification) ((Object) mr.payload);
            }
            if (ire.publish(n)) {
                return "ok";
            } else {
                return "failure";
            }
        } else if (mr.op.equals("subscribeAll")) {
            Vector v = null;
            if (mr.payload instanceof String) {
                v = (Vector) tser.demultiplex((String) mr.payload);
            } else {
                v = (Vector) ((Object) mr.payload);
            }

            if (v == null) {
                return "failure";
            }

            updateInterfaces(v, mr.id);

            Template t = (Template) v.firstElement();
            if (t.getAdvertisement()) {
                ret = ire.advertise(v);
            } else if (t.getQuench()) {
                ret = ire.quench(v);
            } else {
                ret = ire.subscribe(v);

            }
            return ret;
        } else if (mr.op.equals("unsubscribeAll")) {
            Vector v = null;
            if (mr.payload instanceof String) {
                v = (Vector) tser.demultiplex((String) mr.payload);
            } else {
                v = (Vector) ((Object) mr.payload);
            }
            if (v == null) {
                return "failure";
            }

            updateInterfaces(v, mr.id);

            Template t = (Template) v.firstElement();
            if (t.getAdvertisement()) {
                ret = ire.unadvertise(v);
            } else if (t.getQuench()) {
                ret = ire.unquench(v);
            } else {
                ret = ire.unsubscribe(v);

            }
            return ret;
        } else if (mr.op.equals("publishAll")) {
            Vector v = null;
            if (mr.payload instanceof String) {
                v = (Vector) nser.demultiplex((String) mr.payload);
            } else {
                v = (Vector) ((Object) mr.payload);
            }
            if (v == null) {
                return "failure";
            }
            return ire.publish(v);
        } else if (mr.op.equals("openSession")) {
            SessionParams sp = null;
            if (mr.payload instanceof String) {
                sp = new SessionParams();
                sp.parse((String) mr.payload);
            } else {
                sp = (SessionParams) ((Object) mr.payload);
            }
            return _es.openSession(mr.id, "phone@" + mr.deviceId, mr.lastserver,
                                   sp, address, address);
        } else if (mr.op.equals("pauseSession")) {
            return _es.pauseSession(mr.id, "phone@" + mr.deviceId,
                                    mr.lastserver,
                                    address, address);
        } else if (mr.op.equals("closeSession")) {
            return _es.closeSession(mr.id, "phone@" + mr.deviceId,
                                    mr.lastserver);
        } else if (mr.op.equals("createChannel")) {
            MessageFactory mf = MessageTransportService.init().getFactory();
            CodecIndustry.registerContentFactory(new EventCodecFactory());
            Message msg = mf.createMessage();
            msg.putData("op", "createChannel");
            msg.putData("attr1", mr.id);
            msg.putData("attr2", mr.payload);
            Boolean bl = null;
            try {
                bl = (Boolean) ire.defaultBackAccessServerHandler(msg);
            } catch (Exception e) {
                bl = null;
            }
            if (bl != null && bl.booleanValue()) {
                return "success";
            } else {
                return "failure";
            }
        } else if (mr.op.equals("removeChannel")) {
            MessageFactory mf = MessageTransportService.init().getFactory();
            CodecIndustry.registerContentFactory(new EventCodecFactory());
            Message msg = mf.createMessage();
            msg.putData("op", "removeChannel");
            msg.putData("attr1", mr.id);
            msg.putData("attr2", mr.payload);
            Boolean bl = null;
            try {
                bl = (Boolean) ire.defaultBackAccessServerHandler(msg);
            } catch (Exception e) {
                bl = null;
            }
            if (bl != null && bl.booleanValue()) {
                return "success";
            } else {
                return "failure";
            }
        } else if (mr.op.equals("pullSessionSubs") ||
                   mr.op.equals("pullSession")) {
            boolean flag = false;
            if (mr.flag == 1) {
                flag = true;
            }
            System.out.println("pullSessionSubs: " + mr.id + " flag: " + flag);

            ConnectionObject co = _es.getConnection(mr.id);

            if (co != null) {
                System.out.println("Connection was found");
                System.out.println("co.getSessionParams().getBlockedPull():" +
                                   co.getSessionParams().getBlockedPull() +
                                   " size: " + co.getMTP().getOutbound().size() + " flag: " +
                                   flag + " count: " + mr.count);

                // We block the pulling thread if there are no messages
                // This is done to improve the wireless performance
                synchronized (co) {
                    if (co.getSessionParams().getBlockedPull() &&
                        co.getMTP().getOutbound().size() < 1) {
                        System.out.println("-------- pausing pull thread");

                        if (co.getPausedServletThread() != null) {
                            System.out.println("ERROR: Attempted to set pause thread:" +
                                               co.getPausedServletThread());
                            System.out.println("New thread:" + this);
                            System.exit(0);
                        }
                        co.setPausedServletThread(this);
                        try {
                            synchronized (this) {
                                wait(_es.GPRS_TIMEOUT);
                            }
                        } catch (Exception e) {
                            e.printStackTrace();
                        }
                        System.out.println("-------- waking pull thread");
                        co.setPausedServletThread(null);
                    } else {
                        System.out.println("Connection object was not found for: " +
                                           mr.id);
                    }
                }
            }

            Vector vr = null;

            if (mr.op.equals("pullSession")) {
                Vector v = null;
                if (mr.payload instanceof String) {
                    v = (Vector) tser.demultiplex((String) mr.payload);
                } else {
                    v = (Vector) ((Object) mr.payload);
                }
                if (v == null) {
                    return "failure";
                }

                vr = _es.pullSession(mr.id, "phone@" + mr.deviceId, v,
                                     mr.count,
                                     flag,
                                     address);
            }
            else  {
                vr = _es.pullSessionSubs(mr.id, "phone@" + mr.deviceId,
                                         (String) mr.payload, mr.count, flag,
                                         address);
            }



            MicroRequest md = new MicroRequest();
            if (md.payload instanceof Vector) {
                md.payload = vr;
                ret = md;
            } else {
                return EventSerializationService.get("Notification").multiplex(vr);
            }
        }
        return ret;
    }

    /* -------------------------------------------------------------- */
   public void updateInterfaces(Vector v, String id) {
     if (v == null) return;
     for (int i = 0; i < v.size(); i++)  {
        Template t = (Template)v.elementAt(i);
        t.setInterface(id);
     }
   }

}
