/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.server.manager;

import java.net.*;
import java.util.*;

import fuegocore.notify.util.*;

/**
 * This class is responsible for managing event channels / routers and
 * monitoring their performance.
 */
public class RouterManager {

    /** simple hashtable of remote servers (type is the lookup keyword) */
    public Hashtable event_channels = new Hashtable();

    /* -------------------------------------------------------------- */

    /**
     * Returns a hashtable containing the event URLs
     * @return Hashtable
     */
    public Hashtable getChannels() {
        return event_channels;
    }

    /* -------------------------------------------------------------- */
    /* -------------------------------------------------------------- */
    /* EVENT CHANNEL CONTACT INFORMATION */

    /** Returns the list of current event channel servers.
     * @return a list of servers
     */
    public URL[] getServerList() {
        URL[] host = new URL[event_channels.size()];
        int i = 0;
        for (Enumeration k = event_channels.elements(); k.hasMoreElements(); ) {
            host[i++] = (URL) k.nextElement();
        }
        return host;
    }

    /** Returns the list of current event channel servers
     * @param skip skips this index
     * @return list of URLs
     */
    public URL[] getServerList(int skip) {
        URL[] host = new URL[event_channels.size()];
        int i = 0;
        for (Enumeration k = event_channels.elements(); k.hasMoreElements(); ) {
            if (i != skip) {
                host[i++] = (URL) k.nextElement();
            }
        }
        return host;
    }

    /** Hashes event types to resolution servers
     * @param type the event type to hash
     * @return the URL
     */
    public URL getEventChannel(String type) {
        int hash = type.hashCode();
        URL[] host = getServerList();
        int index = host.length;
        if (index == 0) {
            return null;
        }
        index = hash % (host.length - 1);

        return host[index];
    }

    /** Hashes event types to resolution servers.
     *  @param type the event type to hash
     *  @return index
     */
    public int getEventChannelIndex(String type) {
        int hash = type.hashCode();
        URL[] host = getServerList();
        int index = host.length;
        if (index == 0) {
            return -1;
        }
        index = (hash % host.length - 1);
        return index;
    }

    /** Hashes event types to resolution servers
     * @param type the event type to hash
     * @return URL the url to the replica
     */
    public URL getEventChannelReplica(String type) {
        int skip = getEventChannelIndex(type);
        int hash = type.hashCode();
        URL[] host = getServerList(skip);
        int index = host.length;
        if (index == 0) {
            return null;
        }
        index = hash % (host.length - 1);
        return host[index];
    }

    /** Add a new remote event channel
     * @param name the name of the event channel
     * @param host the address
     */
    public synchronized void addEventChannel(String name, URL host) {
        if (event_channels == null) {
            event_channels = new Hashtable();

        }
        String r = (String) event_channels.get(name);
        if (r == null) { // new channel
            event_channels.put(name, host);
        }
    }

    /** Set the event channel hashtable
     * @param ref the hashtable to update
     */
    public synchronized void addEventChannels(Hashtable ref) {
        event_channels = ref;
    }

    /** Lookup event channel given its type
     *  @param type the event channel name/identifier
     *  @return URL the url of the resolution server
     */
    public synchronized URL lookup(String type) {
        Log.write("Doing lookup: " + type);
        if (event_channels == null) {
            return null;
        }
        if (type == null) {
            Log.write("type was null");
            return null;
        }
        URL ret = (URL) event_channels.get(type.trim());
        if (ret == null) { // trying temporary channels
            ret = getEventChannel(type.trim());
        }
        return ret;
    }
}
