/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.session;

import java.util.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.channel.*;
import fuegocore.notify.core.client.*;

/** This class represents an event session. The session is durable
 *  by default.
 */
public class EventSession {
    private String sessionId;
    private SessionParams sp = new SessionParams(); // default params
    public static int sessionCounter = 0;
    private EventConnection ec = EventConnection.get(); // singleton pattern
    private boolean reopenedSession = false;

    // Hashtable for subscriptions (by subId)
    private Hashtable subs = new Hashtable();

    /** Returns the event connection
     * @return EventConnection the event connection object
     */
    public EventConnection getConnection() {
        return ec;
    }

    /**
     * Returns true if the session has expired
     * @return boolean
     */
    public boolean isExpired() {
        long expiration = sp.getExpiration();
        if (expiration <= 0) {
            return false;
        }
        if (System.currentTimeMillis() > expiration) {
            return true;
        }
        return false;
    }

    /**
     * Expire session
     */
    public void expire() {
        Hashtable _subs = getSubscriptions();
        String r1 = unsubscribe(new Vector(_subs.values()));
        if (r1 == null) {
            System.out.println("Session " + sessionId +
                               " expired. Unsub failed:" +
                               r1);
        }

        Hashtable advs = getAdvertisements();
        String r2 = unadvertise(new Vector(advs.values()));
        if (r2 == null) {
            System.out.println("Session " + sessionId +
                               " expired. Unsub failed:" +
                               r2);
        }
        subs.clear();
    }

    /**
     * Returns active subscriptions
     * @return Hashtable
     */
    public Hashtable getSubscriptions() {
        Hashtable ret = new Hashtable();
        Enumeration k = subs.keys();
        for (Enumeration e = subs.elements(); e.hasMoreElements(); ) {
            ITemplate t = (ITemplate) e.nextElement();
            Object subId = e.nextElement();
            if (!t.getAdvertisement()) {
                ret.put(subId, t);
            }
        }
        return subs;
    }

    /**
     * Returns active quench subscriptions
     * @return Hashtable
     */
    public Hashtable getQuenched() {
        Hashtable ret = new Hashtable();
        Enumeration k = subs.keys();
        for (Enumeration e = subs.elements(); e.hasMoreElements(); ) {
            ITemplate t = (ITemplate) e.nextElement();
            Object subId = e.nextElement();
            if (!t.getQuench()) {
                ret.put(subId, t);
            }
        }
        return subs;
    }

    /**
     * Returns active advertisements
     * @return Hashtable
     */
    public Hashtable getAdvertisements() {
        Hashtable ret = new Hashtable();
        Enumeration k = subs.keys();
        for (Enumeration e = subs.elements(); e.hasMoreElements(); ) {
            ITemplate t = (ITemplate) e.nextElement();
            Object subId = e.nextElement();
            if (t.getAdvertisement()) {
                ret.put(subId, t);
            }
        }
        return subs;
    }

    /** Creates a unique session identifier
     * @return String the session identifier
     */
    public String getSessionId() {
        return sessionId;
    }

    /** Creates a unique session identifier
     * @return String the session identifier
     */
    public static synchronized String getUniqueSessionId() {
        sessionCounter++;
        return "sid" + sessionCounter + "@" + EventConnection.get().getName();
    }

    /**
     * Create a unique session name
     * @param id String
     * @return String
     */
    public String createEventSessionId(String id) {
        id += "@" + ec.getName();
        return id;
    }

    /**
     * Constructor for the event session.
     */
    public EventSession() {
        // Get unique session identifier
        sessionId = getUniqueSessionId();
        reopenedSession = false;
    }

    /**
     * Constructor for the event session.
     * @param sid the session identifier
     */
    public EventSession(String sid) {
        // Get unique session identifier
        sessionId = sid;
        reopenedSession = false;
    }

    /**
     * Constructor for the event session.
     * @param sid the session identifier
     * @param _sp the session parameters
     */
    public EventSession(String sid, SessionParams _sp) {
        // Get unique session identifier
        sessionId = sid;
        setParams(_sp);
        reopenedSession = false;
    }

    /** Sets the session parameters of the session
     * @param _sp the session parameters
     */
    public void setParams(SessionParams _sp) {
        sp = _sp;
    }

    /************************************************************************/
    /************************************************************************/
    /************************************************************************/
    /**
     * Advertise a template.
     * @param t ITemplate
     * @return String
     */
    public String advertise(ITemplate t) {
        t.setAdvertisement(true);
        return subscribe(t);
    }

    /**
     * Advertise a vector of templates
     * @param v Vector
     * @return String
     */
    public String advertise(Vector v) {
        for (int i = 0; i < v.size(); i++) {
            ITemplate t = (ITemplate) v.elementAt(i);
            t.setAdvertisement(true);
        }
        return subscribe(v);
    }

    /**
     * Unadvertise a template.
     * @param t ITemplate
     * @return String
     */
    public boolean unadvertise(ITemplate t) {
        if (!t.getAdvertisement()) {
            return false;
        }
        return unsubscribe(t.getSubId());
    }

    /**
     * Advertise a vector of templates
     * @param v Vector
     * @return String
     */
    public String unadvertise(Vector v) {
        for (int i = 0; i < v.size(); i++) {
            ITemplate t = (ITemplate) v.elementAt(i);
            if (!t.getAdvertisement()) {
                return null;
            }
        }
        return unsubscribe(v);
    }

    /** Subscribe the given template
     * Subscription identifier may be forced by setting it using
     * setSubId(String..). This is useful if susbcription name is unique
     * and allows client to recover rapidly from state losses. Client
     * subscribes templates using setSubId, and then when state is lost
     * and session is still open at the server just resubscribes using the
     * known subscription identifiers locally.
     * Use set/get methods in ITemplate to set subId and sessionId.
     * @param t the template to subscribe
     * @return String the Subscription identifier
     **/
    public String subscribe(ITemplate t) {
        if (t.getSource() == null) {
            System.out.println("Subscribe: source missing");
            return null;
        }
        // set session identifier
        t.setSessionId(sessionId);
        String tmp = ec.subscribe(t);
        if (tmp == null || tmp.indexOf("failure") != -1) {
            System.out.println("Failure: " + tmp);
            return null;

        }
        // Store subscription into this session
        subs.put(t.getSubId(), t);
        return tmp;
    }

    /** Unsubscribe the given template
     * @param t the template to unsubscribe
     * @return boolean true is success
     */
    public boolean unsubscribe(ITemplate t) {
        // check session identifier
        if (!t.getSessionId().equals(sessionId)) {
            return false;
        }

        // Remove subscription from this session
        subs.remove(t.getSubId());

        return ec.unsubscribe(t);
    }

    /** Unsubscribe the given template
     * @param subID the subscription identifier (in any session!)
     * @return boolean true is success
     */
    public boolean unsubscribe(String subID) {
        boolean b = ec.unsubscribe(subID);
        if (!b) {
            // Remove subscription from this session
            subs.remove(subID);
        }
        return b;
    }

    /** Publish an event (does not require a session). Advertisement is not
     * enforced.
     * @param n the notification to publish
     * @return boolean true is success
     */
    public boolean publish(Notification n) {
        if (n.getAddress() == null) {
            n.setAddress(getSessionId());
        }
        return ec.publish(n);
    }

    /**********************************************************************/
    /**********************************************************************/

    /** Subscribe the given vector of templates.
     * @param vec the templates to subscribe
     * @return String the Subscription identifier, null or
     *  EventConnection.OP_FAILURE
     **/
    public String subscribe(Vector vec) {
        // set session identifier
        for (int i = 0; i < vec.size(); i++) {
            ITemplate t = (ITemplate) vec.elementAt(i);
            t.setSessionId(sessionId);
        }
        String tmp = ec.subscribeAll(vec);
        if (tmp == null || tmp.indexOf("failure") != -1) {
            return null;
        }
        for (int i = 0; i < vec.size(); i++) {
            ITemplate t = (ITemplate) vec.elementAt(i);
            // Store subscription into this session
            subs.put(t.getSubId(), t);
        }
        return tmp;
    }

    /** Unsubscribe the given vector of templates
     * @param vec the templates to unsubscribe
     * @return String the Subscription identifier, null or
     *  EventConnection.OP_FAILURE
     */
    public String unsubscribe(Vector vec) {
        // check session identifier
        for (int i = 0; i < vec.size(); i++) {
            ITemplate t = (ITemplate) vec.elementAt(i);
            if (!t.getSessionId().equals(sessionId)) {
                return null;
            }
        }
        String tmp = ec.unsubscribeAll(vec);
        if (tmp == null || tmp.indexOf("failure") != -1) {
            return null;
        }
        for (int i = 0; i < vec.size(); i++) {
            ITemplate t = (ITemplate) vec.elementAt(i);
            // Store subscription into this session
            subs.remove(t.getSubId());
        }
        return tmp;
    }

    /** Publish the given events (does not require a session)
     * @param vec the notifications to publish
     * @return String the Subscription identifier, null or
     *  EventConnection.OP_FAILURE
     */
    public String publish(Vector vec) {
        if (vec == null) {
            return null;
        }

        for (int i = 0; i < vec.size(); i++) {
            Notification n = (Notification) vec.elementAt(i);
            if (n.getAddress() == null) {
                n.setAddress(getSessionId());
            }
        }

        String tmp = ec.publishAll(vec);
        if (tmp == null || tmp.indexOf("failure") != -1) {
            return null;
        }
        return tmp;
    }

    /**********************************************************************/
    /**********************************************************************/

    /** Open a session to an access server.
     * Access server URL is determined by the connection object associated
     * with this session.
     * @return the session identifier or null if failure
     */
    public String open() {
        return open(sessionId, sp);
    }

    /**
     * Refresh the session by downloading filters from the server
     * @param sid String
     * @return boolean
     */
    public boolean refresh(String sid) {
        // TBD
        return false;
    }

    /**
     * Open a session to an access server.
     * Access server URL is determined by the connection object associated
     * with this session.
     * @param sid String
     * @param _sp SessionParams
     * @return the session identifier or null if failure
     */
    public String open(String sid, SessionParams _sp) {
        sessionId = sid;
        sp = _sp;
        String t = ec.openSession(sid, _sp);
        if (t == null || t.indexOf("failure") != -1) {
            return null;
        }
        reopenedSession = ec.reopenedSession(t);
        ec.startPullConnection(sid);
        return sid;
    }

    /**
     * Open a session to an access server.
     * Access server URL is determined by the connection object associated
     * @param _sp SessionParams
     * @return the session identifier or null if failure
     */
    public String open(SessionParams _sp) {
        sp = _sp;
        return open(sessionId, sp);
    }

    /** Helper method to check whether or not the session was been reopened.
     * @return true if session was reopened and there is no need to resubscribe
     */
    public boolean reopenedSession() {
        return reopenedSession;
    }

    /** Pause a session to an access server
     * @return the session identifier or null if failure
     */
    public String pause() {
        String tmp = ec.pauseSession(sessionId);
        if (tmp == null || tmp.indexOf("failure") != -1) {
            return null;
        }
        ec.stopPullConnection(sessionId);
        return tmp;
    }

    /** Reset session subscription status
     * @return the session identifier or null if failure
     */
    public String reset() {
        String tmp = ec.resetSession(sessionId);
        if (tmp == null || tmp.indexOf("failure") != -1) {
            return null;
        }
        subs.clear();
        return tmp;
    }

    /** Close a session to an access server
     * @return the session identifier or null if failure
     */
    public String close() {
        String tmp = ec.closeSession(sessionId);
        if (tmp == null || tmp.indexOf("failure") != -1) {
            return null;
        }
        // remove templates
        subs.clear();
        ec.stopPullConnection(sessionId);

        return tmp;
    }

    /**
     * Subscribe to receive notifications when there are subscribers
     * for the given set of filters (ITemplate).Modifies temporarily
     * the objects.
     * @param v Vector
     * @return String
     */
    public String quench(Vector v) {
        for (int i = 0; i < v.size(); i++) {
            ITemplate t = (ITemplate) v.elementAt(i);
            t.setQuench(true);
        }
        String r = subscribe(v);
        for (int i = 0; i < v.size(); i++) {
            ITemplate t = (ITemplate) v.elementAt(i);
            t.setQuench(false);
        }
        return r;
    }

    /**
     * Unsubscribe to receive notifications when there are subscribers
     * for the given set of filters (ITemplate). Modifies temporarily
     * the objects.
     * @param v Vector
     * @return String
     */
    public String unquench(Vector v) {
        for (int i = 0; i < v.size(); i++) {
            ITemplate t = (ITemplate) v.elementAt(i);
            t.setQuench(true);
        }
        String r = unsubscribe(v);
        for (int i = 0; i < v.size(); i++) {
            ITemplate t = (ITemplate) v.elementAt(i);
            t.setQuench(false);
        }
        return r;
    }

    /** Move session from another server to the current access server.
     * @param server servername
     * @param port server port
     * @return the session identifier or null if failure
     */
    public String move(String server, int port) {
        String tmp = ec.moveSession(sessionId, server, port);
        if (tmp == null || tmp.indexOf("failure") != -1) {
            return null;
        }
        return tmp;
    }

    public SessionParams getSessionParams() {
        return sp;
    }

    /**********************************************************************/

    /** Retrieves events from session that match the given set of templates.
     *  Returns a Vector containing the pulled notifications.
     *  @param t the vector of templates to filter against
     *  @param count the number of notifications to pull, 0 is all
     *  @param destroy true if pulled notifications are destroyed from the session
     *  @return Vector the pulled events
     */
    public Vector pullEvents(Vector t, int count,
                             boolean destroy) {
        return ec.pull(sessionId, t, count, destroy);
    }

    /** Retrieve events from session that match the given set of templates.
     *  @param t the vector of templates to filter against
     *  @param count the number of notifications to pull, 0 is all
     *  @param destroy true if pulled notifications are destroyed from the session
     *  @return the number of pulled and delivered events
     */
    public int pullAndDeliver(Vector t, int count,
                              boolean destroy) {
        return ec.pullAndDeliver(sessionId, t, count, destroy);
    }

    /** Retrieve events from session that match the given set of templates.
     *  Returns a Vector containing the pulled notifications.
     *  @param subs comma separated list of subscription identifers to pull
     *  @param count the number of notifications to pull, 0 is all
     *  @param destroy true if pulled notifications are destroyed from the session
     *  @return Vector the pulled events
     */
    public Vector pullEvents(String subs, int count,
                             boolean destroy) {
        return ec.pull(sessionId, subs, count, destroy);
    }

    /** Retrieve events from session that match the given set of templates.
     *  @param subs comma separated list of subscription identifers to pull
     *  @param count the number of notifications to pull, 0 is all
     *  @param destroy true if pulled notifications are destroyed from the session
     *  @return the number of pulled and delivered events
     */
    public int pullAndDeliver(String subs, int count,
                              boolean destroy) {
        return ec.pullAndDeliver(sessionId, subs, count, destroy);
    }

    /** Remove events from session that match a template
     * @param t the template to use in the purging
     * @return true if the purge succeeded
     */
    public boolean purgeEvents(ITemplate t) {
        return false;
    }

    /** Pulls notifications from the given event channel and returns them
     *  as a vector.
     *  @param channel the event channel
     *  @param templates the templates to use in matching
     *  @param count the number of notifications to pull
     *  @return Vector containing the notifications
     */
    public Vector pullEventChannelHistory(String channel, Vector templates,
                                          int count) {
        return ec.pullEventChannelHistory(channel, templates, count);
    }

    /** Pulls notifications from the given event channel and returns them
     *  as a vector.
     *  @param channel the event channel
     *  @param ids comma separated list of notification identifiers (order)
     *  @param count the number of notifications to pull
     *  @return Vector containing the notifications
     */
    public Object pullEventChannelHistoryOrder(String channel, String ids,
                                               int count) {
        return ec.pullEventChannelHistoryOrder(channel, ids, count);
    }

    /**********************************************************************/
    /**********************************************************************/

    /** Create a new event channel
     * @param channel the event channel name to create
     * @param epc the event channel parameters
     * @return true if the operation succeeded
     */
    public boolean createChannel(String channel, EventChannelParams epc) {
        return ec.createEventChannel(channel, epc.toString());
    }

    /** Remove an existing event channel
     * @param channel the event channel name to be removed
     * @return true if the operation succeeded
     */
    public boolean removeChannel(String channel) {
        return ec.removeEventChannel(channel);
    }

    /**********************************************************************/
}
