/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.core.session;

import fuegocore.notify.core.*;

/**
 * This class encapsulates the session parameters. Each session has
 * certain properties, such as the maximum queue size and persistency.
 * A session may also be paused (for pull operation). In addition,
 * we allow to define a special priority filter, which may be used
 * to prioritize control events from content events. Using this feature
 * it is sufficient to use only a single session instead of several
 * sessions making mobility management etc. easier.
 *
 * 2004 Sasu Tarkoma, HIIT
 */
public class SessionParams {
    /* Ordered delivery of asynchronous notifications */
    public boolean orderedDelivery = true;

    /* Maximum queue size */
    public int maxQueueSize = 1000;

    /* Session persistence */
    public boolean persistent = true;

    /* Paused session */
    public boolean pause = false;

    /* Expiration time for the session */
    public int expiration = 0;

    /** Filtering language/system */
    public String filterLanguage = "fuego";

    /* Priority filter for specifying possible control events that have
     * higher priority */
    public Template priorityFilter = null;

    /* Is blocked pull active (for mobile phones with http) */
    public boolean blockedPull = false;

    /* Security token for the session */
    public String securityToken = "";

    /** Returns the string representation of the object
     * @return String the string format
     */
    public String toString() {
        String ret = "";
        ret += "<od>" + orderedDelivery + "</od>";
        ret += "<mq>" + maxQueueSize + "</mq>";
        ret += "<pe>" + persistent + "</pe>";
        ret += "<pa>" + pause + "</pa>";
        ret += "<bp>" + blockedPull + "</bp>";
        ret += "<la>" + filterLanguage + "</la>";
        ret += "<sec" + securityToken + "</sec>";
        ret += "<exp>" + expiration + "</exp>";
        if (priorityFilter == null) {
            ret += "<pf></pf>";
        } else {
            ret += "<pf>" + priorityFilter.toString() + "</pf>";
        }

        return ret;
    }

    /**
     * Parse the object from a string representation.
     * @param s the string to parse
     * @return true if success
     */
    public boolean parse(String s) {
        boolean ret = true;
        String s2 = fuegocore.notify.util.Auxil.getTag(s, "od");
        try {
            orderedDelivery = false;
            if (s2.indexOf("true") != -1) {
                orderedDelivery = true;
            }
        } catch (Exception e) {
            e.printStackTrace();
            ret = false;
        }
        s2 = fuegocore.notify.util.Auxil.getTag(s, "mq");
        try {
            maxQueueSize = Integer.parseInt(s2);
        } catch (Exception e) {
            e.printStackTrace();
            ret = false;
        }
        s2 = fuegocore.notify.util.Auxil.getTag(s, "pe");
        try {
            persistent = false;
            if (s2.indexOf("true") != -1) {
                persistent = true;
            }
        } catch (Exception e) {
            e.printStackTrace();
            ret = false;
        }

        s2 = fuegocore.notify.util.Auxil.getTag(s, "pa");
        try {
            pause = false;
            if (s2.indexOf("true") != -1) {
                pause = true;
            }
        } catch (Exception e) {
            e.printStackTrace();
            ret = false;
        }

        s2 = fuegocore.notify.util.Auxil.getTag(s, "bp");
        try {
            blockedPull = false;
            if (s2.indexOf("true") != -1) {
                blockedPull = true;
            }
        } catch (Exception e) {
            e.printStackTrace();
            ret = false;
        }

        s2 = fuegocore.notify.util.Auxil.getTag(s, "exp");
        try {
            expiration = Integer.parseInt(s2);
        } catch (Exception e) {
            expiration = 0;
            e.printStackTrace();
            ret = false;
        }

        s2 = fuegocore.notify.util.Auxil.getTag(s, "pf");
        try {
            if (s2 != null && s2.length() > 0) {
                priorityFilter = new Template(s2);
            }
        } catch (Exception e) {
            priorityFilter = null;
            e.printStackTrace();
            ret = false;
        }

        filterLanguage = fuegocore.notify.util.Auxil.getTag(s, "la");
        securityToken = fuegocore.notify.util.Auxil.getTag(s, "sec");

        return ret;
    }

    /**
     * Set the pause mode.
     * @param b the pause value
     */
    public void setPause(boolean b) {
        pause = b;
    }

    /**
     * Set the expiration date.
     * @param d the expiration date
     */
    public void setExpiration(int d) {
        expiration = d;
    }

    /**
     * Get the expiration date
     * @return Date the expiration date
     */
    public int getExpiration() {
        return expiration;
    }

    /**
     * Get the pause variable.
     * @return true if the session is paused
     */
    public boolean getPause() {
        return pause;
    }

    /**
     * Set the priority filter
     * @param f the priority filter
     */
    public void setPriorityFilter(Template f) {
        priorityFilter = f;
    }

    public boolean getBlockedPull() {
        return blockedPull;
    }

    public void setBlockedPull(boolean bp) {
        blockedPull = bp;
    }

    /**
     * Get the priority filter
     * @return the priority filter
     */
    public Template getPriorityFilter() {
        return priorityFilter;
    }

    /**
     * Set the persistency of the session.
     * @param b the persistency of the session (true=persistent)
     */
    public void setPersistency(boolean b) {
        persistent = b;
    }

    /**
     * Get the persistency parameter of the session.
     * @return true if the session is persistent
     */
    public boolean getPersistency() {
        return persistent;
    }

    /** Set the maximum queue size of the session
     * @param a the maximum queue size
     */
    public void setMaxQueueSize(int a) {
        maxQueueSize = a;
    }

    /** Get the maximum queue size of the session
     *  @return int the maximum queue size
     */
    public int getMaxQueueSize() {
        return maxQueueSize;
    }

    /** Set the ordered delivery parameter.
     * @param b the ordered delivery value
     */
    public void setOrderedDelivery(boolean b) {
        orderedDelivery = b;
    }

    /**
     * Get the ordered delivery value.
     * @return true if ordered delivery is used.
     */
    public boolean getOrderedDelivery() {
        return orderedDelivery;
    }
}
