/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.demo;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.session.*;
import fuegocore.util.*;

/**
 * Fuego Core demo. Demonstrates a shared session.
 */

public class DemoClient implements Notifiable {

    private static final String TICKER =
            "Fuego Core Demonstration -- 17.09.2004 -- " +
            "Helsinki Institute for Information Technology";

    private static final String TICKER2 = " Waiting for data...  ";
    int counter = 0;
    boolean alive = true;
    SessionControlListener cl;
    EventSession controlSession;
    EventSession contentSession;

    String acquiredSession = null;
    String controlSessionId = null;

    String unsubId = null;
    String currentSubs = null;
    ScrollerFrame sc = null; // GUI
    Vector notifs = new Vector();
    Template subSession; // wait for release of a session subscription
    Template subSession2;

    public ScrollerFrame getScrollerFrame() {
        return sc;
    }

    public static void main(String[] args) {
        // set server address
        fuegocore.notify.core.client.EventConnection.get().setName("democlient");
        try {
            if (args != null && args.length > 0 && args[0] != null) {
                fuegocore.notify.core.client.EventConnection.get().setHostPort(
                        args[0],
                        Integer.parseInt(args[1]));
            } else {
                System.out.println("Incorrect arguments: host port");
                System.exit(0);
            }
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(0);
        }

        DemoClient test = new DemoClient();
        System.out.println("DemoClient\n");
    }

    public DemoClient() {
        // We create a new session for control messages
        controlSession = new EventSession("control@democlient");

        // We open and establish the session
        if (controlSession.open() == null) {
            System.out.println("controlSession open failed");
            System.exit(0);
        }

        System.out.println("Creating control template");

        Template t0 = new Template();
        t0.setSubId("system");
        t0.setSessionId("system");
        t0.setSource(this);

        Filter f0 = new Filter("type", EventTriplet.TYPE_STRING,
                               "session-endpoint-change");
        t0.addFilter(f0);
        t0.setScopeToLocal();

        // subscribe end-point-change locally
        controlSession.subscribe(t0);

        subSession = new Template();

        // The communication can be done using the control session identifier or
        // the shared session identifier. The shared session identifier is
        // better, because the clients need only to share the session name

        // Add attribute filters to the template
        Filter f1 = new Filter("type", EventTriplet.TYPE_STRING,
                               "equ session_control");
        //Filter f2 = new Filter("destination", EventTriplet.TYPE_STRING,
        //                       "equ " + controlSessionId);
        Filter f2 = new Filter("destination", EventTriplet.TYPE_STRING,
                               "exists");
        Filter f3 = new Filter("status", EventTriplet.TYPE_STRING,
                               "equ release");
        // Note that only one user may be active with the same username
        Filter f4 = new Filter("user", EventTriplet.TYPE_STRING,
                               "equ fuegodemo");
        //Filter f4 = new Filter("sessionid", EventTriplet.TYPE_STRING,
        //                          "equ shared_session");

        // Subs is required if we use PULL from a shared session
        // We could actually transfer only part of the subscriptions from a shared
        // session by just sending the sessionID, credentials, and the subscription
        // identifier. This is not needed for push, push sends all notifications
        // in the queue.
        Filter f5 = new Filter("subs", EventTriplet.TYPE_STRING,
                               "exists");

        // Define a listener class
        cl = new SessionControlListener(this);

        subSession.setSource(cl); // set the source
        unsubId = "release_" + this.hashCode();
        subSession.setSubId(unsubId);
        subSession.addFilter(f1);
        subSession.addFilter(f2);
        subSession.addFilter(f3);
        subSession.addFilter(f4);
        subSession.addFilter(f5);

        Template syncTemplate = new Template();
        Filter sf1 = new Filter("type", EventTriplet.TYPE_STRING,
                                "equ synchronization");
        Filter sf2 = new Filter("group", EventTriplet.TYPE_STRING,
                                "equ demo");

        syncTemplate.addFilter(sf1);
        syncTemplate.addFilter(sf2);

        // set handler
        SyncHandler syncl = new SyncHandler(this);
        syncTemplate.setSource(syncl);

        Template meetingTemplate = new Template();

        Filter mf1 = new Filter("type", EventTriplet.TYPE_STRING,
                                "equ demochannel");

        Filter mf2 = new Filter("subtype", EventTriplet.TYPE_STRING,
                                "equ meeting");

        Filter mf3 = new Filter("target", EventTriplet.TYPE_STRING,
                                "equ group");
        meetingTemplate.addFilter(mf1);
        meetingTemplate.addFilter(mf2);
        meetingTemplate.addFilter(mf3);
        // set handler
        MeetingHandler meet = new MeetingHandler(this);
        meetingTemplate.setSource(meet);

        contentSession = new EventSession();

        // We only subscribe once per session
        if (controlSession.reopenedSession()) {
            // Since the session is already open we only restore the
            // local subscription table
            subSession.setScopeToLocal();
            controlSession.subscribe(subSession);

            meetingTemplate.setScopeToLocal();
            controlSession.subscribe(meetingTemplate);
            syncTemplate.setScopeToLocal();
            controlSession.subscribe(syncTemplate);
        } else {
            // Do only local subscriptions. Here if we have server filtered identifiers,
            // the callback information is lost for the client.
            if (controlSession.subscribe(subSession) == null) {
                System.out.println("Subscription failure: subSession");
                System.exit(0);
            }
            if (controlSession.subscribe(meetingTemplate) == null) {
                System.out.println("Subscription failure: meetingTemplate");
                System.exit(0);
            }
            if (controlSession.subscribe(syncTemplate) == null) {
                System.out.println("Subscription failure: syncTemplate");
                System.exit(0);
            }
        }

        System.out.println("Starting GUI");
        Frame f = new Frame();
        //   f.setSize(1024/4, 760/4);
        //   sc = new ScrollerFrame(TICKER, TICKER2, 1024/4, 700/4);
        f.setSize(1024, 760);
        sc = new ScrollerFrame(TICKER, TICKER2, 1024, 700);
        f.add(sc);
        f.setVisible(true);

        System.out.println("Waiting for session");

        //closeSession(controlSessionId);
    }

    /**
     * Open shared session for push-type of communication.
     * @param sid String
     * @param subs String
     */
    public void activateSessionForPush(String sid, String subs) {
        System.out.println("Subs: [" + subs + "] sid [" + sid + "]");

        // Subscribe all subs in the incoming session
        try {
            StringTokenizer tok = new StringTokenizer(subs, ",");
            // Note that we set up local subs before opening the session
            // We can do this only for local subs
            while (tok.hasMoreElements()) {
                String tsub = (String) tok.nextElement();
                // we subscribe the subscription identifier locally
                Template t = new Template();
                t.setSource(this);
                t.setSubId(tsub); // we assume only one identifier in this demo

                // normally they are comma separated.
                t.setScopeToLocal();
                //System.out.println("local subscribe: " + t.toString());
                contentSession.subscribe(t);
                // We could store the reference for future local unsubscribe
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        acquiredSession = sid;
        currentSubs = subs;

        SessionParams sp = new SessionParams();
        sp.setPause(false); // activate shared session
        sp.setMaxQueueSize( -1); // do not update this one

        System.out.println("------ opening session");
        String id = contentSession.open(sid, sp);

        if (!contentSession.reopenedSession()) {
            System.out.println("Session did not exist! [" + sid + "] returned:" +
                               id);
        }
        System.out.println("------ Session open");
        sc.setActiveMode();

        // UPDATE SESSION MANAGEMENT IN THE BACKGROUND
        Template t2 = new Template();
        // Add attribute filters to the template
        Filter f6 = new Filter("type", EventTriplet.TYPE_STRING,
                               "equ session_control");
        Filter f7 = new Filter("status", EventTriplet.TYPE_STRING,
                               "equ acquire");
        Filter f8 = new Filter("sessionid", EventTriplet.TYPE_STRING,
                               "equ " + sid);
        t2.setSubId("acquireMsg_" + this.hashCode());
        t2.setSource(cl); // set the source
        t2.addFilter(f6);
        t2.addFilter(f7);
        t2.addFilter(f8);

        System.out.println("Unsubscribing");
        if (!controlSession.unsubscribe(subSession)) { // unsubId
            System.out.println("Unsubscribe failed");
            //System.exit(0);
        }
        if (controlSession.subscribe(t2) == null) {
            System.out.println("Acquire subscription failed");
            //System.exit(0);
        }

    }

    // after system close session event
    public void forcedSessionRelease() {
        // we unsubscribe the acquire channel
        controlSession.unsubscribe("acquireMsg_" + this.hashCode());

        // we subscribe the release channel (acquire for us)
        subSession.setScopeToExternal();

        if (controlSession.subscribe(subSession) == null) {
            System.out.println("Resubscription of subsession failed");
            //  System.exit(0);
        }
    }

    public void releaseSession(String id, String source) {
        if (id == null || !acquiredSession.equals(id)) {
            System.out.println(
                    "Request for a session that this client does not possess");
            return;
        }

        String s = contentSession.pause();
        if (s == null) {
            System.out.println("Pausing shared session failed");
        }

        // we unsubscribe the acquire channel
        controlSession.unsubscribe("acquireMsg" + this.hashCode());

        Notification n = new Notification();

        n.setType("session_control");
        n.add(new EventTriplet("destination", source));
        n.add(new EventTriplet("status", "release"));
        n.add(new EventTriplet("sessionid", id));
        n.add(new EventTriplet("subs", currentSubs));

        if (!controlSession.publish(n)) {
            System.out.println("Publish of release event failed");
        }
    }

    /**
     * Handles pushed notifications from the shared session
     * @param n Notification
     */
    public void notifyIncoming(Notification n) {
        System.out.println("Default handler. Client received notification: " +
                           n.getType() + "," +
                           n.get());

        // process system event
        if (n.getType().equals("session-endpoint-change")) {
            System.out.println("Session changed");
            // We assume it is session_control, we could check this
            sc.setWaitMode();
            forcedSessionRelease();
            return;
        }

        // process other events
        if (notifs.size() > 5) {
            notifs.remove(0); // remove oldest
        }
        notifs.add(n);

        // Notify GUI
        String ret = "";
        for (int i = 0; i < notifs.size(); i++) {
            Notification n2 = (Notification) notifs.elementAt(i);
            EventTriplet et = (EventTriplet) n2.getField(EventTriplet.formatId(
                    "data",
                    EventTriplet.TYPE_STRING));
            if (et != null) {
                ret += (String) et.getValue() + "  -  ";
            }
        }
        sc.setNextString(ret); // schedule next string to display
        sc.setNotifs(notifs);
    }
}


class SyncHandler implements Notifiable {
    DemoClient _eb;

    public SyncHandler(DemoClient eb) {
        _eb = eb;
    }

    public void notifyIncoming(Notification n) {
        System.out.println("*** Incoming sync");
        if (_eb.getScrollerFrame() != null) {
            _eb.getScrollerFrame().showSyncIcon();
        } else {
            System.out.println("Handler was null");
        }
    }

}


class MeetingHandler implements Notifiable {
    DemoClient _eb;

    public MeetingHandler(DemoClient eb) {
        _eb = eb;
    }

    public void notifyIncoming(Notification n) {
        System.out.println("*** Incoming meeting");
        if (_eb.getScrollerFrame() != null) {
            _eb.getScrollerFrame().showMeetingIcon();
        } else {
            System.out.println("Handler was null");
        }
    }
}


/** Handles the incoming session management operations */
class SessionControlListener implements Notifiable {
    int count = 0;
    DemoClient _eb;

    public SessionControlListener(DemoClient eb) {
        _eb = eb;
    }

    public int getCount() {
        return count;
    }

    /**
     * Handles the incoming session management operations
     * @param n Notification
     */
    public void notifyIncoming(Notification n) {
        System.out.println("*** Incoming session control");
        if (n == null) {
            return;
        }

        // We extract subscription identifiers
        EventTriplet et = (EventTriplet) n.getField(EventTriplet.formatId(
                "status",
                EventTriplet.TYPE_STRING));
        if (et == null || et.getValue() == null) {
            System.out.println("Failed to extract parameter: status");
            return;
        }
        String status = (String) et.getValue();

        et = (EventTriplet) n.getField(EventTriplet.formatId("sessionid",
                EventTriplet.TYPE_STRING));
        if (et == null || et.getValue() == null) {
            System.out.println("Failed to extract parameter: sessionid");
            return;
        }
        String id = (String) et.getValue();

        if (status.equals("release")) {
            et = (EventTriplet) n.getField(EventTriplet.formatId("subs",
                    EventTriplet.TYPE_STRING));
            if (et == null || et.getValue() == null) {
                System.out.println("Failed to extract parameter: subs");
                return;
            }
            String subs = (String) et.getValue();
            _eb.activateSessionForPush(id, subs);
        } else if (status.equals("acquire")) {
            et = (EventTriplet) n.getField(EventTriplet.formatId("source",
                    EventTriplet.TYPE_STRING));
            if (et == null || et.getValue() == null) {
                System.out.println("Failed to extract parameter: source");
                return;
            }
            String source = (String) et.getValue();
            _eb.releaseSession(id, source);
        }
        count++;
    }

}


class ScrollerFrame extends Canvas implements Runnable {
    private Image fuegopic;

    /** Thread used for triggering repaints */
    private Thread thread;

    /** buffer image of the screen */
    private Image image;

    /** Time of most recent paint */
    private long paintTime;

    /** Time of most recent frame rate report */
    private long statsTime;

    /** Number of frames since last frame rate report */
    int frameCount;
    private String statusMessage = null;

    /** Last reported frame rate (for re-paint) */
    int frameRate;
    int imageWidth;
    int imageHeight;
    int imageX;
    int imageY;
    int offset = 0;
    int offset2 = 0;
    int TICK_SPEED = 3;
    int FRAME_RATE = 120;
    int ticker_width = 0;
    int ticker_width2 = 0;
    String text;
    String text2;
    String next_string;
    boolean alive = true;
    boolean wait_mode = true;
    boolean uflag = false;
    double rot_angle = 0;
    Font writerfont = new Font("Arial", Font.BOLD, 20);
    Font font = new Font("Arial", Font.BOLD, 88);
    Font font2 = new Font("Arial", Font.BOLD, 72);

    Font font3 = new Font("Arial", Font.PLAIN, 72);
    FontMetrics mwriterfont = this.getFontMetrics(writerfont);
    FontMetrics mfont1 = this.getFontMetrics(font);
    FontMetrics mfont2 = this.getFontMetrics(font2);
    FontMetrics mfont3 = this.getFontMetrics(font3);

    Image myImage = null;
    Image tiimalasi = null;
    Image meetingImg = null;
    Image syncImg = null;

    private final int NONE = 0;
    private final int MEETING = 1;
    private final int SYNC = 2;

    private int OVERLAY_IMG = NONE;
    private long overlay_timer = 0;
    private long overlay_state = 0;

    int MEETING_SHOW_TIME = 30000; //30 sec
    int SYNC_SHOW_TIME = 45000; //30 sec

    Toolkit tk = null;
    AffineTransform atImageSpace = new AffineTransform();
    Vector notifs;
    int writerSlide = 0;
    int writerSlideA = 0;

    public ScrollerFrame(String tx, String tx2, int x, int y) {
        setSize(x, y);
        Graphics g;
        image = this.createImage(getWidth(), getHeight());
        offset = getWidth();
        text = tx;
        text2 = tx2;
        ticker_width = mfont1.stringWidth(text);
        ticker_width2 = mfont2.stringWidth(text2);

        fuegopic = loadImage("images/hiit.jpg");
        tiimalasi = loadImage("images/tiimalasi.gif");
        meetingImg = loadImage("images/meeting.gif");
        syncImg = loadImage("images/syncxml.gif");

        if (fuegopic != null) {
            imageWidth = fuegopic.getWidth(this);
            imageHeight = fuegopic.getHeight(this);
        } else {
            imageWidth = 0;
            imageHeight = 0;
        }

        imageX = (getWidth() - imageWidth) / 2;
        imageY = (getHeight() - imageHeight) / 2;

        FRAME_RATE = 120;

        thread = new Thread(this);
        thread.setPriority(Thread.MIN_PRIORITY);
        thread.start();
    }

    public void setNotifs(Vector v) {
        notifs = v;
    }

    public void showSyncIcon() {
        OVERLAY_IMG = SYNC;
        overlay_state = 0;

        if (!wait_mode) {
            System.out.println(
                    "Wait mode was false, probably a missing message");
            wait_mode = true;
        }
    }

    public void showMeetingIcon() {
        OVERLAY_IMG = MEETING;
        overlay_state = 0;
    }

    // Draw the notification vector
    public void drawWriter(Graphics g, int x, int y, String condition) {
        if (notifs == null || notifs.size() < 1) {
            return;
        }
        g.setFont(writerfont);
        g.setColor(new Color(0, 0, 0));
        int yh = mwriterfont.getHeight() + 4;
        if (condition == null) {
            g.drawString("All events", x, y - yh - 2);
        } else {
            g.drawString(condition, x, y - yh - 2);
        }
        String firstType = null; // first new type
        Notification nz = (Notification) notifs.lastElement();
        EventTriplet etz = (EventTriplet) nz.getField(EventTriplet.formatId(
                "domain",
                EventTriplet.TYPE_STRING));

        if (etz != null && etz.getValue() != null) {
            firstType = (String) etz.getValue();

        }
        int j = 0;
        for (int i = notifs.size() - 1; i >= 0; i--) {
            Notification n = (Notification) notifs.elementAt(i);
            double c = 200 / (double) notifs.size();
            c = c * j + 54;
            String curCond = null;
            EventTriplet etx = (EventTriplet) n.getField(EventTriplet.formatId(
                    "domain",
                    EventTriplet.TYPE_STRING));
            if (etx != null && etx.getValue() != null) {
                if (!((String) etx.getValue()).equals(condition)) {
                    if (condition != null) {
                        continue;
                    }
                }
                curCond = (String) etx.getValue();
            }
            EventTriplet et = (EventTriplet) n.getField(EventTriplet.formatId(
                    "data",
                    EventTriplet.TYPE_STRING));
            if (et != null) {
                if (j == 0 && writerSlide != 0 &&
                    curCond != null && curCond.equals(firstType)) { // slide effect for new ones
                    if (writerSlide > 0) {
                        writerSlide -= writerSlideA;
                        if (writerSlideA > 1) {
                            writerSlideA--;
                        }
                    } else {
                        writerSlide = 0;
                    }
                    c += writerSlide * 1.8;
                    if (c > 255) {
                        c = 255;
                    }
                    g.setColor(new Color((int) c, (int) c, (int) c));
                    g.drawString((String) et.getValue(), x + writerSlide,
                                 y + yh * j);
                } else {
                    g.setColor(new Color((int) c, (int) c, (int) c));
                    g.drawString((String) et.getValue(), x, y + yh * j);
                }
            }
            j++;
        }
    }

    private Image loadImage(String filename) {
        Toolkit toolkit = Toolkit.getDefaultToolkit();
        Image image = toolkit.getImage(ExtUtil.getResource(filename));
        MediaTracker mediaTracker = new MediaTracker(this);
        mediaTracker.addImage(image, 0);
        try {
            mediaTracker.waitForID(0);
        } catch (InterruptedException ie) {
            System.err.println(ie);
            return null;
        }
        return image;
    }

    public void setWaitMode() {
        wait_mode = true;
        FRAME_RATE = 120;
        System.out.println("Wait mode is true");
    }

    public void setActiveMode() {
        wait_mode = false;
        FRAME_RATE = 120;
        System.out.println("Wait mode is false");
    }

    public void updateText(String tx, String tx2) {
        text = tx;
        text2 = tx;
        ticker_width = mfont1.stringWidth(text);
        ticker_width2 = mfont2.stringWidth(text2);
    }

    public void statusMessage(String s) {
        statusMessage = s;
    }

    public void setNextString(String next) {
        next_string = next;
        writerSlide = 110;
        writerSlideA = 15;
    }

    public void destroy() {
        alive = false;
    }

    /**
     * Set the array of images to be displayed.
     * Update the width and height of the image and draw
     * the border to fit around it in the offscreen image.
     * @param im Image
     */
    void setImage(Image im) {
        this.fuegopic = im;
        if (fuegopic != null) {
            imageWidth = fuegopic.getWidth(this);
            imageHeight = fuegopic.getHeight(this);
        } else {
            imageWidth = 0;
            imageHeight = 0;
        }
        imageX = (getWidth() - imageWidth) / 2;
        imageY = (getHeight() - imageHeight) / 2;
    }

    void reset() {
        fuegopic = null;
        thread = null;
    }

    /**
     * Runs the animation and makes the repaint requests.
     * The thread will exit when it is no longer the current
     * Animation thread.
     */
    public void run() {
        Thread mythread = Thread.currentThread();
        long scheduled = System.currentTimeMillis();
        statsTime = scheduled;
        paintTime = scheduled;
        frameCount = 0;
        frameRate = 0;

        while (thread == mythread && alive) {
            synchronized (this) {
                try {
                    // Update when the next frame should be drawn
                    // and compute the delta till then
                    scheduled += FRAME_RATE; // 25 frames per second
                    long delta = scheduled - paintTime;
                    if (delta > 0) {
                        this.wait(delta);
                    }
                    // Advance and repaint the screen

                    repaint();
                } catch (InterruptedException e) {
                }
            }
        }
    }

    /**
     * Paint is called whenever the canvas should be redrawn.
     * It clears the canvas and draws the frame and the current
     * current frame from the animation.
     * @param g the Graphics context to which to draw
     */
    int wobbleCounter = 0;
    int dY = 0;
    int dK = 0;
    int mode = 0;

    public void update(Graphics g) {
        paint(g);
    }

    public void paint(Graphics g) {
        // wait screen
        if (wait_mode) {
            paint2(g);
            return;
        }

        // normal scroller
        paintTime = System.currentTimeMillis();
        if (image == null) {
            image = this.createImage(getWidth(), getHeight());

        }
        Graphics2D og = (Graphics2D) image.getGraphics();
        og.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                            RenderingHints.VALUE_ANTIALIAS_ON);

        og.setColor(new Color(255, 255, 255));
        og.fillRect(0, 0, getWidth(), getHeight());

        if (fuegopic != null) {
            og.drawImage(fuegopic, imageX, getHeight() - fuegopic.getHeight(this), this);
        }

        if (frameRate > 0) {
            og.setColor(new Color(0, 0, 0));
            og.drawString("FPS = " + frameRate, 80, getHeight() - 60);
        }

        if (wobbleCounter++ > 400) {
            wobbleCounter = 0;
            if (mode == 0) {
                dY = 0;
                mode = 1;
            } else if (mode == 1) {
                dY = -getHeight();
                mode = 0;
            } else if (mode == 2) {
                dY = 32;
                mode = 0;
            }
            dK = 0;
        }

        if (mode == 1 && dY < getHeight()) {
            dY += dK;
            dK++;
        }

        if (mode == 0 && dY < 0) {
            dY += dK;
            dK++;
            if (dY > 0) {
                dY = 0;
            }
        }

        frameCount++;

        if (statusMessage != null && statusMessage.length() > 0) {
            og.setColor(new Color(0, 0, 0));
            og.setFont(font2);
            og.drawString(statusMessage, 15,
                          getHeight() - mfont2.getHeight());
        }

        // Draw writer
        g.setColor(new Color(140, 140, 140));

        og.drawLine(0, 22, 300, 22);
        og.drawLine(341, 22, 642, 22);
        og.drawLine(682, 22, 1024, 22);

        og.setClip(0, 0, 340, 600);
        drawWriter(og, 0, 50, "home");
        og.setClip(341, 0, 681, 600);
        drawWriter(og, 341, 50, "work");
        og.setClip(682, 0, 1024, 600);
        drawWriter(og, 682, 50, null);
        og.setClip(0, 0, getWidth(), getHeight());

        paintOverlay(og);

        g.drawImage(image, 0, 0, this);

        // Update Frame rate
        int delta = (int) (paintTime - statsTime);
        if (delta > 1000 && delta < 10000) {
            frameRate = ((frameCount * 1000 + 500) / delta);
            frameCount = 0;
            statsTime = paintTime;
            //repaint();	// queue full repaint to display frame rate
        }
    }

    int ovr_tmp = 0;

    public void paintOverlay(Graphics2D g) {
        if (OVERLAY_IMG == MEETING && overlay_state >= 0) {
            if (overlay_state == 0) {
                overlay_timer = System.currentTimeMillis();
                ovr_tmp = 0;
                overlay_state = 1;
            } else if (overlay_state == 1) {
                if (ovr_tmp < 80) {
                    ovr_tmp += 5;
                } else {
                    ovr_tmp = 80;
                }
                if (System.currentTimeMillis() - overlay_timer >
                    MEETING_SHOW_TIME) {
                    overlay_state = 2;
                }
            } else if (overlay_state == 2) {
                if (ovr_tmp > 0) {
                    ovr_tmp -= 5;
                } else {
                    ovr_tmp = 0;
                    overlay_state = 3;
                }
            } else if (overlay_state == 3) {
                OVERLAY_IMG = NONE;
                overlay_state = 0;
            }

            g.setComposite(
                    AlphaComposite.getInstance(
                            AlphaComposite.SRC_OVER, ((float) ovr_tmp) / 100));

            if (meetingImg != null) {
                g.drawImage(meetingImg,
                            (getWidth() - meetingImg.getWidth(this)) / 2,
                            (getHeight() - meetingImg.getHeight(this)) / 2, this);
            }

            g.setComposite(
                    AlphaComposite.getInstance(
                            AlphaComposite.SRC_OVER, 1.0f));

        } else if (OVERLAY_IMG == SYNC && overlay_state >= 0) {
            if (overlay_state == 0) {
                overlay_timer = System.currentTimeMillis();
                ovr_tmp = 0;
                overlay_state = 1;
                // syncImg.flush();
                // ovr_tmp = 80;
            } else if (overlay_state == 1) {
                if (ovr_tmp < 80) {
                    ovr_tmp += 5;
                } else {
                    ovr_tmp = 80;
                }
                if (System.currentTimeMillis() - overlay_timer > SYNC_SHOW_TIME) {
                    overlay_state = 2;
                }
            } else if (overlay_state == 2) {
                if (ovr_tmp > 0) {
                    ovr_tmp -= 5;
                } else {
                    ovr_tmp = 0;
                    overlay_state = 3;
                }
            } else if (overlay_state == 3) {
                OVERLAY_IMG = NONE;
                overlay_state = 0;
            }
            g.setComposite(
                    AlphaComposite.getInstance(
                            AlphaComposite.SRC_OVER, ((float) ovr_tmp) / 100));

            if (syncImg != null) {
                g.drawImage(syncImg, (getWidth() - syncImg.getWidth(this)) / 2,
                            (getHeight() - syncImg.getHeight(this)) / 2, this);

            }

            g.setComposite(
                    AlphaComposite.getInstance(
                            AlphaComposite.SRC_OVER, 1.0f));

        }
    }

    public void paint2(Graphics g) {
        //   if (uflag) return;
        //   else if (!uflag) uflag = true;

        paintTime = System.currentTimeMillis();
        if (image == null) {
            image = this.createImage(getWidth(), getHeight());
        }

        Graphics2D og = (Graphics2D) image.getGraphics();
        og.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                            RenderingHints.VALUE_ANTIALIAS_ON);

        og.setColor(new Color(133, 133, 209));
        og.fillRect(0, 0, getWidth(), getHeight());

        String text = "Waiting for session";
        int sy = mfont3.getHeight();
        int sx = mfont3.stringWidth(text);

        if (tiimalasi != null) {
            og.drawImage(tiimalasi, (getWidth() - tiimalasi.getWidth(this)) / 2,
                         (getHeight() - tiimalasi.getHeight(this)) / 2, this);

        }
        og.setFont(font3);
        og.setColor(new Color(180, 180, 190));
        og.drawString(text, (getWidth() - sx) / 2, (getHeight() - sy) / 2);

        paintOverlay(og);

        g.drawImage(image, 0, 0, this);

        // Update Frame rate
        int delta = (int) (paintTime - statsTime);
        if (delta > 1000 && delta < 10000) {
            frameRate = ((frameCount * 1000 + 500) / delta);
            frameCount = 0;
            statsTime = paintTime;
            //repaint();	// queue full repaint to display frame rate
        }
    }

}
