/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.demo;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.session.*;
import fuegocore.notify.demo.DemoClient;

/**
 * Demonstrates a shared session. Test class for the demo.
 */

public class DemoClientSessionHolder implements Notifiable {
    int counter = 0;
    boolean alive = true;
    String sharedSessionId = "shared_session";
    String controlSessionId = null;
    String sub1 = "sub1";
    String sub2 = "sub2";
    String sub3 = "sub3";


    EventSession controlSession;
    EventSession contentSession;
    boolean hasContentSession = true;

    String sub = null;
    String sessionDestination = "control@democlient";

    public static void main(String[] args) {
        // set server address
        fuegocore.notify.core.client.EventConnection.get().setName(
                "demosessionholder");
        try {
            if (args != null && args.length > 0 && args[0] != null) {
                fuegocore.notify.core.client.EventConnection.get().setHostPort(
                        args[0],
                        Integer.parseInt(args[1]));
            } else {
                System.out.println("Incorrect arguments: host port");
                System.exit(0);
            }
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(0);
        }

        DemoClientSessionHolder test = new DemoClientSessionHolder();
        System.out.println("DemoClientSessionHolder\n");
    }

    public DemoClientSessionHolder() {
        // Open session
        controlSession = new EventSession();
        controlSession.open();

        System.out.println("Creating control template");

        Template t = new Template();
        // Add attribute filters to the template
        Filter f1 = new Filter("type", EventTriplet.TYPE_STRING,
                               "equ demochannel");
        Filter f2 = new Filter("subtype", EventTriplet.TYPE_STRING,
                               "equ context-change");
        //Filter f3 = new Filter("sessionid", EventTriplet.TYPE_STRING,
        //                       sharedSessionId);

        // Subs is required if we use PULL from a shared session
        // We could actually transfer only part of the subscriptions from a shared
        // session by just sending the sessionID, credentials, and the subscription
        // identifier. This is not needed for push, push sends all notifications
        // in the queue.
        //Filter f4 = new Filter("subs", EventTriplet.TYPE_STRING,
        //                       "exists");

        // Define a listener class
        t.setSource(this); // set the source
        t.setSubId("release");
        t.addFilter(f1);
        t.addFilter(f2);
        //t.addFilter(f3);
        //t.addFilter(f4);

        // We only subscribe once per session
        if (controlSession.reopenedSession()) {
            // Since the session is already open we only restore the
            // local subscription table
            t.setScopeToLocal();
            controlSession.subscribe(t);
        } else {
            // Do only local subscriptions. Here if we have server filtered identifiers,
            // the callback information is lost for the client.
            if (controlSession.subscribe(t) == null) {
                System.out.println("Subscription failure");
                System.exit(0);
            }
        }

        /** We open our own session */
        SessionParams sp = new SessionParams();
        sp.setPause(false);
        sp.setMaxQueueSize(100);

        contentSession = new EventSession();
        System.out.println("Got session id for shared session:" +
                           contentSession.getSessionId());

        if (contentSession.open(sharedSessionId, sp) == null) {
            System.out.println("Failure in opening content session");
            System.exit(0);
        }

        Template t2 = new Template();
        t2.setSubId(sub1);
        Filter f3 = new Filter("type", EventTriplet.TYPE_STRING,
                               "equ demochannel");
        Filter f4 = new Filter("domain", EventTriplet.TYPE_STRING,
                               "equ home");
        t2.addFilter(f3);
        t2.addFilter(f4);
        t2.setSource(this);

        Template t3 = new Template();
        t3.setSubId(sub2);
        Filter f5 = new Filter("type", EventTriplet.TYPE_STRING,
                               "equ demochannel");
        Filter f6 = new Filter("domain", EventTriplet.TYPE_STRING,
                               "equ work");
        t3.addFilter(f5);
        t3.addFilter(f6);
        t3.setSource(this);

        if (contentSession.reopenedSession()) {
            System.out.println("Reopened session");
            t2.setScopeToLocal();
            t3.setScopeToLocal();
            contentSession.subscribe(t2);
            contentSession.subscribe(t3);
        } else {
            contentSession.subscribe(t2);
            contentSession.subscribe(t3);
        }

        System.out.println("Waiting for notifications");

        for (; ; ) {
            try {
                Thread.sleep(1000);
            } catch (Exception e) {
                break;
            }
        }

        if (contentSession != null) {
            contentSession.close();
        }
        if (controlSession != null) {
            controlSession.close();
        }
    }

    public void releaseSession(String id, String destination) {
        if (id == null || !hasContentSession) {
            System.out.println(
                    "Request for a session that this client does not possess");
            return;
        }

        String s = contentSession.pause();
        if (s == null) {
            System.out.println("Close shared session failed");
            System.exit(0);
        }

        Notification n = new Notification();
        n.setType("session_control");

        n.add(new EventTriplet("destination", "*"));
        n.add(new EventTriplet("status", "release"));
        n.add(new EventTriplet("user", "fuegodemo"));
        n.add(new EventTriplet("sessionid", id));
        n.add(new EventTriplet("subs", sub1 + "," + sub2));
        if (!controlSession.publish(n)) {
            System.out.println("Publish of release event failed");
        }
        hasContentSession = false;
    }

    /**
     * Checks if the session was reopened
     * @return String
     */
    public String reOpenSession() {
        String str = contentSession.open();
        hasContentSession = true;
        return str;
    }


    /**
     * Handles pushed notifications from the shared session
     * @param n Notification
     */
    public void notifyIncoming(Notification n) {
        System.out.println("Received Notification: " + n.toString());

        EventTriplet et = (EventTriplet) n.getField(
                EventTriplet.formatId("subtype",
                                      EventTriplet.TYPE_STRING));

        if (et != null &&
            ((String) et.getValue()).equals("context-change")) {
            // demo control event. Note that the control event remains in the session!!!
            et = (EventTriplet) n.getField(EventTriplet.formatId(
                    "status",
                    EventTriplet.TYPE_STRING));
            if (et != null &&
                ((String) et.getValue()).indexOf("release") != -1) {
                releaseSession(sharedSessionId, sessionDestination);
            } else if (et != null &&
                       ((String) et.getValue()).indexOf("acquire") != -1) {
                reOpenSession();
            }
        }
    }
}


class ContentHandler implements Notifiable {
    DemoClient _eb;

    public ContentHandler(DemoClient eb) {
        _eb = eb;
    }

    public void notifyIncoming(Notification n) {
        System.out.println("Client received notification: " + n.getType() + "," +
                           n.get());
    }
}
