/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.demo;

import java.awt.*;
import java.util.*;

import fuegocore.notify.graphics.*;

/**
 * Graphical user interface (graph) for monitoring RTT statistics.
 */

public class RTTPlotBox {
    int p;
    int sx, sy;
    boolean flip = true;
    Vector pixelBuffer = new Vector();
    int maxsize = 100;
    int printcount = 0;
    int lastmarkercolor = -1;
    int lastmarkerx = -1;
    int lastmarker_add = -1;

    public static final int LINK_DOWN = 1; // new code
    public static final int LINK_UP = 2; // new code
    public static final int WLAN_START = 3; // same as in D211interface
    public static final int GPRS_START = 5;
    public static final int GSM_START = 7;
    public static final int QOS_CHANGE = 14;
    public static final int LINK_CHANGE = 15;

    public RTTPlotBox(int sx, int sy, int max) {
        this.sx = sx;
        this.sy = sy;
        p = 0;
        maxsize = max;
        initVector();
    }

    public void resize(int sx, int sy) {
        this.sx = sx;
        this.sy = sy;
        maxsize = sy;
    }

    public void initVector() {
        for (int i = 0; i < maxsize; i++) {
            pixelBuffer.add(new GraphObject(0, 0, 0));
        }
    }

    // add marker
    public void addPoint(int k) {
        if (lastmarker_add == -1) {
            lastmarker_add = k;
        }
    }

    public void addPoint(double val, double sd) {
        if (pixelBuffer.size() > maxsize) {
            pixelBuffer.remove(pixelBuffer.firstElement());

        }
        GraphObject d = new GraphObject(val, sd, printcount);

        if (lastmarker_add > -1) {
            d.marker = lastmarker_add;
            d.flip = flip;
            if (flip) {
                flip = false;
            } else {
                flip = true;
            }
            lastmarker_add = -1;
        }

        if (printcount++ > 10) {
            printcount = 0;

        }
        pixelBuffer.add(d);
    }

    public void addPoint(double val, double sd, double pred) {
        if (pixelBuffer.size() > maxsize) {
            pixelBuffer.remove(pixelBuffer.firstElement());

        }
        GraphObject d = new GraphObject(val, sd, printcount);

        d.setPrediction(pred);

        if (lastmarker_add > -1) {
            d.marker = lastmarker_add;
            lastmarker_add = -1;
            d.flip = flip;
            if (flip) {
                flip = false;
            } else {
                flip = true;
            }
        }

        if (printcount++ > 10) {
            printcount = 0;
        }
        pixelBuffer.add(d);
    }

    public void addPoint(double val, double sd, int marker) {
        if (pixelBuffer.size() > maxsize) {
            pixelBuffer.remove(pixelBuffer.firstElement());
        }
        GraphObject d = new GraphObject(val, sd, marker, printcount);

        if (lastmarker_add > -1) {
            d.marker = lastmarker_add;
            lastmarker_add = -1;
            d.flip = flip;
            if (flip) {
                flip = false;
            } else {
                flip = true;
            }
        }

        if (printcount++ > 10) {
            printcount = 0;
        }
        pixelBuffer.add(d);
    }

    public void addPoint(double val, double sd, int marker, double pred) {
        if (pixelBuffer.size() > maxsize) {
            pixelBuffer.remove(pixelBuffer.firstElement());

        }
        GraphObject d = new GraphObject(val, sd, marker, printcount);

        if (lastmarker_add > -1) {
            d.marker = lastmarker_add;
            lastmarker_add = -1;
            if (flip) {
                flip = false;
            } else {
                flip = true;
            }
        }

        d.setPrediction(pred);

        if (printcount++ > 10) {
            printcount = 0;
        }
        pixelBuffer.add(d);
    }

    public double getMax() {
        double ret = 0;
        for (int i = 0; i < pixelBuffer.size(); i++) {
            GraphObject go = (GraphObject) pixelBuffer.elementAt(i);
            if (go.val + go.sd > ret) {
                ret = go.val + go.sd;
            }
            if (go.prediction + go.sd > ret) {
                ret = go.prediction + go.sd;
            }
        }
        return ret;
    }

    // Overriden paint method that draws a rectange around the panel
    public void draw(Graphics og, int ox, int oy) {
        String curText = "";

        og.setColor(new Color(0, 0, 49));
        og.drawRect(0 + ox, 0 + oy, sx, sy);

        og.setColor(new Color(244, 244, 244));
        og.drawRect(1 + ox, 1 + oy, sx - 2, sy - 2);

        og.setColor(new Color(0, 0, 32));
        //og.setColor(new Color(19,69,128));
        og.fillRect(3 + ox, 3 + oy, sx - 6, sy - 6);

        og.setColor(new Color(255, 205, 0));

        double max = getMax(); // get maximum value
        max = max * 1.4; // add scale a bit

        int plot_sy = sy - 10 - 5; // plot area size y
        int plot_sx = sx - 10 - 40; // plot area size x

        double x_interval = plot_sx / maxsize;
        int print = 0;
        // vertical lines
        for (int i = 2; i < pixelBuffer.size(); i++) {
            og.setColor(new Color(50, 90, 50));
            og.drawLine(ox + (int) (5 + i * x_interval), oy + (int) (5),
                        ox + (int) (5 + i * x_interval),
                        oy + (int) (5 + plot_sy));
        }

        // if max is zero then we have zero observation
        if (max > 0) {
            // draw x-axis, three axis lines
            double y = max / 4;
            //   Hashtable used = new Hashtable();
            for (int i = 1; i < 4; i++) {
                double val = (i * y) / max; // normalize
                val = val * plot_sy; // get position within window
                val = plot_sy - val;

                og.setColor(new Color(50, 90, 50));
                og.drawLine(ox + (int) (5), oy + (int) (5 + val),
                            ox + (int) (5 + plot_sx),
                            oy + (int) (5 + val));

                og.setColor(new Color(180, 220, 180));
                String label = "" + (int) (i * y);
                //     if (used.get(label) == null)  {
                if (i == 3) {
                    og.drawString(label, ox + (int) (5 + plot_sx),
                                  oy +
                                  (int) (5 + val +
                                         og.getFontMetrics().getHeight() / 2));
                    //      used.put(label,"true");
                    //   }
                }
            }
        }

        for (int i = 2; i < pixelBuffer.size(); i++) {
            GraphObject go_old = (GraphObject) pixelBuffer.elementAt(i - 1);
            GraphObject go_old2 = (GraphObject) pixelBuffer.elementAt(i - 2);

            double valo = go_old.val;
            valo = valo / max; // normalize
            valo = valo * plot_sy; // get position within window
            valo = plot_sy - valo;
            if (valo < 0) {
                valo = 0;
            }
            if (valo > plot_sy) {
                valo = plot_sy;

            }
            double oldpred = go_old2.prediction;
            oldpred = oldpred / max; // normalize
            oldpred = oldpred * plot_sy; // get position within window
            oldpred = plot_sy - oldpred;
            if (oldpred < 0) {
                oldpred = 0;
            }
            if (oldpred > plot_sy) {
                oldpred = plot_sy;

            }

            GraphObject go = (GraphObject) pixelBuffer.elementAt(i);

            // colour change
            int marker = go.marker;
            double val = 0;
            /*
                   // Draw standard deviation
                   val = go.val - go.sd;
                   val = val / max; // normalize
                   val = val * plot_sy; // get position within window
                   val = plot_sy - val;
                   if (val < 0) {
              val = 0;
                   }
                   if (val > plot_sy) {
              val = plot_sy;

                   }
                   double val2 = go.val + go.sd;
                   val2 = val2 / max; // normalize
                   val2 = val2 * plot_sy; // get position within window
                   val2 = plot_sy - val2;

                   if (val2 < 0) {
              val2 = 0;
                   }
                   if (val2 > plot_sy) {
              val2 = plot_sy;
                   }

                   og.setColor(new Color(80, 122, 80));

             og.drawLine(ox + (int) (5 + i * x_interval - 2), oy + (int) (5 + val),
                        ox + (int) (5 + i * x_interval + 2),
                        oy + (int) (5 + val));

             og.drawLine(ox + (int) (5 + i * x_interval - 2), oy + (int) (5 + val2),
                        ox + (int) (5 + i * x_interval + 2),
                        oy + (int) (5 + val2));
             */
            val = go.val;
            val = val / max; // normalize
            val = val * plot_sy; // get position within window
            val = plot_sy - val;

            if (val < 0) {
                val = 0;
            }
            if (val > plot_sy) {
                val = plot_sy;

            }
            double pred = go_old.prediction;
            pred = pred / max; // normalize
            pred = pred * plot_sy; // get position within window
            pred = plot_sy - pred;
            if (pred < 0) {
                pred = 0;
            }
            if (pred > plot_sy) {
                pred = plot_sy;

                // Draw value
            }
            og.setColor(new Color(180, 220, 180));
            og.drawLine(ox + (int) (5 + (i - 1) * x_interval),
                        oy + (int) (5 + valo - 1),
                        ox + (int) (5 + i * x_interval),
                        oy + (int) (5 + val - 1));
            /*
                   // draw prediction
                   og.setColor(new Color(100, 255, 120));
                   og.drawLine(ox + (int) (5 + (i - 1) * x_interval - 1),
                        oy + (int) (5 + oldpred),
                        ox + (int) (5 + i * x_interval),
                        oy + (int) (5 + pred));
             */
            // Process marker
            curText = null;
            if (go.marker > 0) {
                if (go.marker == LINK_DOWN) {
                    curText = "off";
                } else if (go.marker == LINK_UP) {
                    curText = "on";
                } else if (go.marker == WLAN_START) {
                    curText = "WLAN";
                } else if (go.marker == GSM_START) {
                    curText = "GSM";
                } else if (go.marker == GPRS_START) {
                    curText = "GPRS";
                } else if (go.marker == QOS_CHANGE) {
                    curText = "qos";
                } else if (go.marker == LINK_CHANGE) {
                    curText = "link";
                }
            }

            // old style of printing value
            if (curText != null) {
                og.setColor(new Color(255, 255, 255));
                if (go.flip) {
                    og.drawString("" + curText, ox + (int) (5 + i * x_interval),
                                  oy + (int) (25));
                } else {
                    og.drawString("" + curText, ox + (int) (5 + i * x_interval),
                                  oy + (int) (65));
                }
                print = 0;
            }

        }
    }
}
