/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.demo.simple;

import java.awt.*;
import java.awt.geom.*;

import fuegocore.notify.demo.visualize.*;

/**
 * Simple example of the Fuego animation script
 */
public class SimpleCanvas extends Canvas implements Runnable, ImageResolver {

    // Create a new animation object for controlling sprites
    Animator anim = new Animator(this);

    private Image backgroundpic;

    /** Thread used for triggering repaints */
    private Thread thread;

    /** buffer image of the screen */
    private Image image;

    /** Time of most recent paint */
    private long paintTime;

    /** Time of most recent frame rate report */
    private long statsTime;

    /** Number of frames since last frame rate report */
    int frameCount;

    /** Last reported frame rate (for re-paint) */
    int frameRate;

    int imageWidth;
    int imageHeight;
    int imageX;
    int imageY;

    int FRAME_RATE = 100;
    boolean alive = true;
    double rot_angle = 0;
    Font font = new Font("Arial", Font.BOLD, 17);
    FontMetrics mfont1 = this.getFontMetrics(font);
    Image myImage = null;
    Image player1 = null;
    Image player2 = null;
    Toolkit tk = null;
    AffineTransform atImageSpace = new AffineTransform();
    public static SimpleCanvas SimpleCanvas = null;
    boolean runDemo = false;
    public static long startAnimTime = 0;

    public static void restart() {
        System.out.println("Restarting");
        SimpleCanvas.anim.replay();
        startAnimTime = 0;
    }

    public Animator getAnimator() {
        return anim;
    }

    public Image resolveImage(int i) {
        if (i == 0) {
            return player1;
        } else if (i == 1) {
            return player2;
        }
        return null;
    }

    public SimpleCanvas(int x, int y) {
        SimpleCanvas = this;
        setSize(x, y);
        Graphics g;
        image = this.createImage(getWidth(), getHeight());

        // Load the background
        backgroundpic = loadImage("fc_demo.jpg");

        // Load player characters
        player1 = loadImage("ukko.gif"); //right
        player2 = loadImage("ukkol.gif"); //left

        if (backgroundpic != null) {
            imageWidth = backgroundpic.getWidth(this);
            imageHeight = backgroundpic.getHeight(this);
        } else {
            imageWidth = 0;
            imageHeight = 0;
        }

        imageX = (getWidth() - imageWidth) / 2;
        imageY = (getHeight() - imageHeight) / 2;

        FRAME_RATE = 100;
    }

    public void kick() {
        thread = new Thread(this);
        thread.setPriority(Thread.MIN_PRIORITY);
        thread.start();
        runDemo = true;
    }

    private Image loadImage(String filename) {
        Toolkit toolkit = Toolkit.getDefaultToolkit();
        Image image = toolkit.getImage(filename);
        MediaTracker mediaTracker = new MediaTracker(this);
        mediaTracker.addImage(image, 0);
        try {
            mediaTracker.waitForID(0);
        } catch (InterruptedException ie) {
            System.err.println(ie);
            return null;
        }
        return image;
    }

    public void destroy() {
        alive = false;
    }

    /**
     * Set the array of images to be displayed.
     * Update the width and height of the image and draw
     * the border to fit around it in the offscreen image.
     * @param im Image
     */
    void setImage(Image im) {
        this.backgroundpic = im;
        if (backgroundpic != null) {
            imageWidth = backgroundpic.getWidth(this);
            imageHeight = backgroundpic.getHeight(this);
        } else {
            imageWidth = 0;
            imageHeight = 0;
        }
        imageX = (getWidth() - imageWidth) / 2;
        imageY = (getHeight() - imageHeight) / 2;
    }

    /**
     * Reset
     */
    void reset() {
        backgroundpic = null;
        thread = null;
    }

    /**
     * Runs the animation and makes the repaint requests.
     * The thread will exit when it is no longer the current
     * Animation thread.
     */
    public void run() {
        Thread mythread = Thread.currentThread();
        long scheduled = System.currentTimeMillis();
        statsTime = scheduled;
        paintTime = scheduled;
        frameCount = 0;
        frameRate = 0;

        while (thread == mythread && alive) {
            synchronized (this) {
                try {
                    // Update when the next frame should be drawn
                    // and compute the delta till then
                    scheduled += FRAME_RATE; // 25 frames per second
                    long delta = scheduled - paintTime;
                    if (delta > 0) {
                        this.wait(delta);
                    }
                    // Advance and repaint the screen

                    repaint();
                } catch (InterruptedException e) {
                }
            }
        }
    }

    public void update(Graphics g) {
        paint(g);
    }

    public void paint(Graphics g) {
        // normal scroller
        paintTime = System.currentTimeMillis();
        if (image == null) {
            image = this.createImage(getWidth(), getHeight());
        }

        Graphics2D og = (Graphics2D) image.getGraphics();
        og.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                            RenderingHints.VALUE_ANTIALIAS_ON);

        og.setColor(new Color(255, 255, 255));
        og.fillRect(0, 0, getWidth(), getHeight());

        if (backgroundpic != null) {
            og.drawImage(backgroundpic, imageX,
                         imageY, this);
        }

        if (frameRate > 0) {
            og.setColor(new Color(0, 0, 0));
            og.drawString("FPS = " + frameRate, 80, getHeight() - 10);
        }

        frameCount++;

        og.setColor(new Color(0, 0, 0));
        if (font != null) {
            og.setFont(font);
        }

        og.setClip(0, 0, getWidth(), getHeight());

        // Animate the canvas using the script
        if (runDemo) {
            if (startAnimTime == 0) {
                startAnimTime = System.currentTimeMillis();
            }
            anim.animate((int) (System.currentTimeMillis() - startAnimTime), og,
                         imageX, imageY);
        }

        // Normalize stroke
        og.setStroke(new BasicStroke(1));

        g.drawImage(image, 0, 0, this);

        // Update Frame rate
        int delta = (int) (paintTime - statsTime);
        if (delta > 1000) {
            frameRate = ((frameCount * 1000 + 500) / delta);
            frameCount = 0;
            statsTime = paintTime;
        }
    }

    /**
     * Return the time of the animation
     * @return int
     */
    public int getAnimationTime() {
        return (int) (System.currentTimeMillis() - startAnimTime);
    }
}
