/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.demo.visualize;

import java.awt.*;
import java.awt.geom.*;

/**
 * Represents one activity in the animation. May be non-graphical
 * activity, for example publishing an event
 */
public abstract class AbstractAnimEvent implements Comparable {
    // animation commands
    public static final int SETPOINT = 0;
    public static final int MOVETO = 1;
    public static final int MOVERINGTO = 2;
    public static final int EXPANDRING = 3;
    public static final int DRAWRING = 4;
    public static final int COMPRESSRING = 5;
    public static final int RESTART = 14;
    public static final int SEND = 6;
    public static final int RECEIVE = 7;

    // status codes
    public static final int NOTYET = -1;
    public static final int EXPIRED = -2;
    public static final int STILLTOGO = -3;
    public static final int DONE = 1;
    public static final int EXITLOOP = 2;

    int image_num;
    int code;
    int duration;
    int sx, sy, tx, ty; // source and target
    int time;
    int tmp = 0;
    int radius = 0;
    String text = "";

    boolean override_expiration = false;

    float transparency = 0.5f;

    /**
     * Returns the duration of the event
     * @return int
     */
    public int getDuration() {
        return duration;
    }

    /**
     * Returns the occurrence time
     * @return int
     */
    public int getTime() {
        return time;
    }

    /**
     * Returns the value of transparency
     * @return float
     */
    public float getTransparency() {
        return transparency;
    }

    /**
     * Return animation code
     * @return int
     */
    public int getCode() {
        return code;
    }

    /**
     * Reset the event
     */
    public void reset() {
        tmp = 0;
        transparency = 0.5f;
        if (code == RECEIVE) {
            tmp = 100;
        } else if (code == COMPRESSRING) {
            transparency = 0;
        }
    }

    public AbstractAnimEvent() {}

    /**
     * Create a new animation event
     * @param time int time of occurrence
     * @param code int event code
     * @param x int x-coordinate
     * @param y int y-coordinate
     * @param s String textual parameter
     */
    public AbstractAnimEvent(int time, int code, int x, int y, String s) {
        this(time, code, 0, -1, x, y, 0, 0);
        text = s;
        if (code == RECEIVE) {
            tmp = 100;
        } else if (code == COMPRESSRING) {
            transparency = 0;
        }
    }

    /**
     * Create a new animation event
     * @param time int time of occurrence
     * @param code int event code
     * @param image int image number
     * @param duration int duration of event
     * @param x int x-coordinate
     * @param y int y-coordinate
     */
    public AbstractAnimEvent(int time, int code, int image, int duration, int x,
                             int y) {
        this(time, code, image, duration, x, y, 0, 0);
    }

    /**
     * Create a new animation event
     * @param time int
     * @param code int
     * @param image int
     * @param duration int
     * @param x int
     * @param y int
     * @param tx int
     * @param ty int
     */
    public AbstractAnimEvent(int time, int code, int image, int duration, int x,
                             int y, int
                             tx, int ty) {
        this.time = time;
        this.code = code;
        image_num = image;
        this.duration = duration;
        sx = x;
        sy = y;
        this.tx = tx;
        this.ty = ty;
    }

    /**
     * Create a new animation event
     * @param time int
     * @param code int
     * @param image int
     * @param duration int
     * @param x int
     * @param y int
     * @param tx int
     * @param ty int
     * @param r int
     */
    public AbstractAnimEvent(int time, int code, int image, int duration, int x,
                             int y, int
                             tx, int ty, int r) {
        this(time, code, image, duration, x, y, tx, ty);
        radius = r;
    }

    /**
     * Draw and animate
     * @param curtime int
     * @param g Graphics2D
     * @param offsetX int
     * @param offsetY int
     * @param ir ImageResolver
     * @return int
     */
    public int draw(int curtime, Graphics2D g, int offsetX, int offsetY,
                    ImageResolver ir) {
        int ret = DONE;

        // if event is not yet valid
        if (curtime < time) {
            //System.out.println("Not yet: curtime:" + curtime + " time:" + time);
            return NOTYET;
        }

        // if the event is expired
        if (duration != -1 && curtime > time + duration && !override_expiration) {
            return EXPIRED;
        }

        Image img = ir.resolveImage(image_num);
        if (img == null) { // assert
            System.out.println("No image found in image resolver!");
            System.exit(0);
        }

        switch (code) {
        // Visuals
        case RESTART:
            System.out.println("Restarting demo");
            ret = EXITLOOP;
            break;
        case SEND:
            Font writerfont = new Font("Arial", Font.BOLD, 15);
            FontMetrics mwriterfont = g.getFontMetrics(writerfont);
            g.setFont(writerfont);
            int l = mwriterfont.stringWidth(text);
            g.setFont(writerfont);
            g.setColor(new Color(0, 0, 0));

            float f = (float) (100 - tmp);
            f = f / (float) 100;

            g.setComposite(
                    AlphaComposite.getInstance(
                            AlphaComposite.SRC_OVER, f));

            g.drawString(text, sx - l / 2 + offsetX,
                         sy - tmp - mwriterfont.getHeight() / 2 + offsetY);

            if (tmp < 100) {
                ret = STILLTOGO; // above test overrides this
                tmp += 2;
            }

            g.setComposite(
                    AlphaComposite.getInstance(
                            AlphaComposite.SRC_OVER, 1.0f));

            writerfont = null;
            mwriterfont = null;
            break;
        case RECEIVE:
            writerfont = new Font("Arial", Font.BOLD, 15);
            mwriterfont = g.getFontMetrics(writerfont);
            g.setFont(writerfont);
            l = mwriterfont.stringWidth(text);
            g.setFont(writerfont);
            g.setColor(new Color(0, 0, 0));

            f = (float) (tmp);
            f = f / (float) 100;

            g.setComposite(
                    AlphaComposite.getInstance(
                            AlphaComposite.SRC_OVER, f));

            g.drawString(text, sx - l / 2 + offsetX,
                         sy - tmp - mwriterfont.getHeight() / 2 + offsetY);

            if (tmp > 0) {
                tmp -= 2;
                ret = STILLTOGO; // above test overrides this

            }
            g.setComposite(
                    AlphaComposite.getInstance(
                            AlphaComposite.SRC_OVER, 1.0f));

            writerfont = null;
            mwriterfont = null;

            break;
        case SETPOINT:
            g.drawImage(img, sx + offsetX, sy + offsetY, null);
            ret = STILLTOGO; // above test overrides this
            break;
        case MOVETO:

            // Move a sprite from sx,sy to tx,ty in time duration
            // We also need to take the offset into account
            double delta_x = ((double) (tx - sx)) / (double) duration;
            double delta_y = ((double) (ty - sy)) / (double) duration;

            //	System.out.println("dx: " + delta_x + " dy: "+ delta_y);
            int x = sx + (int) (delta_x * (curtime - time));
            int y = sy + (int) (delta_y * (curtime - time));

            // 	System.out.println("Drawing:"  + (x+offsetX) + ","+(y+offsetY));

            g.setComposite(
                    AlphaComposite.getInstance(
                            AlphaComposite.SRC_OVER, 0.7f));

            g.drawImage(img, x + offsetX, y + offsetY, null);

            g.setComposite(
                    AlphaComposite.getInstance(
                            AlphaComposite.SRC_OVER, 1.0f));

            // Still some frames to be shown
            if (tx != x || ty != y) {
                ret = STILLTOGO;

            }
            break;
        case DRAWRING:
            g.setComposite(
                    AlphaComposite.getInstance(
                            AlphaComposite.SRC_OVER,
                            (float) Math.abs(Math.sin(time))));

            Ellipse2D e = new Ellipse2D.Double(sx + offsetX - radius,
                                               sy + offsetY - radius,
                                               radius * 2, radius * 2);

            g.setStroke(new BasicStroke(8));

            GradientPaint gp = new GradientPaint(75, 75, Color.white,
                                                 95, 95, Color.gray, true);

            g.setPaint(gp);

            g.draw(e);

            g.setComposite(
                    AlphaComposite.getInstance(
                            AlphaComposite.SRC_OVER, 1.0f));
            ret = STILLTOGO; // above test overrides this
            break;
        case MOVERINGTO:

            // Move a sprite from sx,sy to tx,ty in time duration
            // We also need to take the offset into account
            delta_x = ((double) (tx - sx)) / (double) duration;
            delta_y = ((double) (ty - sy)) / (double) duration;

            x = sx + (int) (delta_x * (curtime - time));
            y = sy + (int) (delta_y * (curtime - time));

            g.setComposite(
                    AlphaComposite.getInstance(
                            AlphaComposite.SRC_OVER, 0.5f));

            e = new Ellipse2D.Double(x + offsetX - radius, y + offsetY - radius,
                                     radius * 2, radius * 2);

            g.setStroke(new BasicStroke(8));

            gp = new GradientPaint(75, 75, Color.white,
                                   95, 95, Color.gray, true);
            g.setPaint(gp);
            g.draw(e);

            g.setComposite(
                    AlphaComposite.getInstance(
                            AlphaComposite.SRC_OVER, 1.0f));

            // Still some frames to be shown
            if (tx != x || ty != y) {
                ret = STILLTOGO;

            }
            break;
        case EXPANDRING:

            // Move a sprite from sx,sy to tx,ty in time duration
            // We also need to take the offset into account
            delta_x = ((double) (ty - tx)) / (double) duration;
            x = tx + (int) (delta_x * (curtime - time));
            if (x < tx) {
                System.out.println("Radius was below threshold");
                System.out.println("x:" + x + " tx: " + tx);
                System.exit(0);
            }
            g.setComposite(
                    AlphaComposite.getInstance(
                            AlphaComposite.SRC_OVER, transparency));

            e = new Ellipse2D.Double(sx + offsetX - x, sy + offsetY - x,
                                     x * 2, x * 2);

            g.setStroke(new BasicStroke(8));

            gp = new GradientPaint(75, 75, Color.white,
                                   95, 95, Color.gray, true);
            g.setPaint(gp);
            g.draw(e);

            g.setComposite(
                    AlphaComposite.getInstance(
                            AlphaComposite.SRC_OVER, 1.0f));

            // Still some frames to be shown
            if ((int) x < (int) ty) {
                ret = STILLTOGO;
            }
            break;
        case COMPRESSRING:

            // Move a sprite from sx,sy to tx,ty in time duration
            // We also need to take the offset into account
            delta_x = ((double) (tx - ty)) / (double) duration;
            x = tx - (int) (delta_x * (curtime - time));

            g.setComposite(
                    AlphaComposite.getInstance(
                            AlphaComposite.SRC_OVER, transparency));

            e = new Ellipse2D.Double(sx + offsetX - x, sy + offsetY - x,
                                     x * 2, x * 2);

            g.setStroke(new BasicStroke(8));

            gp = new GradientPaint(75, 75, Color.white,
                                   95, 95, Color.gray, true);
            g.setPaint(gp);
            g.draw(e);

            g.setComposite(
                    AlphaComposite.getInstance(
                            AlphaComposite.SRC_OVER, 1.0f));

            // Still some frames to be shown
            if ((int) x != (int) ty) {
                ret = STILLTOGO;

                // Do a nice fade effect on the expanding circle
            }
            if (transparency < 0.5) {
                transparency += 0.1;
            }

            break;
//        default:
//          animateExtensions(curtime, g, offsetX, offsetY, ir);
//          break;
        }

        return ret;
    }

    /**
     * Animation extensions. Add your routines here. Use unique event codes.
     * @param curtime int
     * @param g Graphics2D
     * @param offsetX int
     * @param offsetY int
     * @param ir ImageResolver
     * @return int
     */
    public int animateExtensions(int curtime, Graphics2D g, int offsetX,
                                 int offsetY,
                                 ImageResolver ir) {
        return EXPIRED;
    }

    /**
     * Comparison based on order.
     * @param other the other object to compare to
     * @return int the comparison value
     */
    public int compareTo(Object other) {
        if (other instanceof AnimEvent) {
            int a = ((AnimEvent) other).getTime();
            return getTime() - a;

        } else {
            throw new ClassCastException();
        }
    }
}
