/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.demo.visualize;

import java.awt.*;

import fuegocore.notify.core.*;
import fuegocore.notify.demo.*;

/**
 * Represents one activity in the animation. May be non-graphical
 * activity, for example publishing an event
 */
public class AnimEvent extends AbstractAnimEvent {
    public static final int PUB_CONTEXT = 8;
    public static final int PUB_ARRIVAL = 9;
    public static final int PUB_SYNC = 10;
    public static final int PUB_CONTEXT2 = 11;
    public static final int PUB_MEETING = 12;
    public static final int PUB_LUNCH = 13;
    public static final int NO_CODE = -99;

    public AnimEvent() {}

    /**
     * Create a new animation event
     * @param time int time of occurrence
     * @param code int event code
     * @param x int x-coordinate
     * @param y int y-coordinate
     * @param s String textual parameter
     */
    public AnimEvent(int time, int code, int x, int y, String s) {
        this(time, code, 0, -1, x, y, 0, 0);
        text = s;
        if (code == RECEIVE) {
            tmp = 100;
        } else if (code == COMPRESSRING) {
            transparency = 0;
        }
    }

    /**
     * Create a new animation event
     * @param time int time of occurrence
     * @param code int event code
     * @param image int image number
     * @param duration int duration of event
     * @param x int x-coordinate
     * @param y int y-coordinate
     */
    public AnimEvent(int time, int code, int image, int duration, int x, int y) {
        this(time, code, image, duration, x, y, 0, 0);
    }

    /**
     * Create a new animation event
     * @param time int
     * @param code int
     * @param image int
     * @param duration int
     * @param x int
     * @param y int
     * @param tx int
     * @param ty int
     */
    public AnimEvent(int time, int code, int image, int duration, int x, int y, int
                     tx, int ty) {
        this.time = time;
        this.code = code;
        image_num = image;
        this.duration = duration;
        sx = x;
        sy = y;
        this.tx = tx;
        this.ty = ty;
    }

    /**
     * Create a new animation event
     * @param time int
     * @param code int
     * @param image int
     * @param duration int
     * @param x int
     * @param y int
     * @param tx int
     * @param ty int
     * @param r int
     */
    public AnimEvent(int time, int code, int image, int duration, int x, int y, int
                     tx, int ty, int r) {
        this(time, code, image, duration, x, y, tx, ty);
        radius = r;
    }

    /**
     * Draw and animate
     * @param curtime int
     * @param g Graphics2D
     * @param offsetX int
     * @param offsetY int
     * @param ir ImageResolver
     * @return int
     */
    public int draw(int curtime, Graphics2D g, int offsetX, int offsetY,
                    ImageResolver ir) {

        // if event is not yet valid
        if (curtime < time) {
            //System.out.println("Not yet: curtime:" + curtime + " time:" + time);
            return NOTYET;
        }

// if the event is expired
        if (duration != -1 && curtime > time + duration && !override_expiration) {
            return EXPIRED;
        }
        /*
         Image img = ir.resolveImage(image_num);
         if (img == null) { // assert
          System.out.println("No image found in image resolver!");
          System.exit(0);
         }
         */
        int i = animateExtensions(curtime, g, offsetX, offsetY, ir);

        if (i == NO_CODE) {
            return super.draw(curtime, g, offsetX, offsetY, ir);
        } else {
            return i;
        }
    }

    /**
     * Animation extensions. Add your routines here. Use unique event codes.
     * @param curtime int
     * @param g Graphics2D
     * @param offsetX int
     * @param offsetY int
     * @param ir ImageResolver
     * @return int
     */
    public int animateExtensions(int curtime, Graphics2D g, int offsetX,
                                 int offsetY,
                                 ImageResolver ir) {
        int ret = NO_CODE;

        switch (code) {
// Publish events (no visuals)
        case RESTART:
            System.out.println("Restarting demo");
            ret = EXITLOOP;
            DemoFrame.restart();
            break;
        case PUB_CONTEXT:
            Notification n = new Notification();
            n.setType("demochannel");
            n.add(new EventTriplet("subtype", "context-change"));
            n.add(new EventTriplet("status", "release endpoint"));
            n.add(new EventTriplet("user", "demo"));
            if (!MainDemo.mainref.publish(n)) {
                System.out.println("Publish of context event failed");
            }
            ret = DONE;
            break;
        case PUB_CONTEXT2:
            n = new Notification();
            n.setType("demochannel");
            n.add(new EventTriplet("subtype", "context-change"));
            n.add(new EventTriplet("status", "acquire endpoint"));
            n.add(new EventTriplet("user", "demo"));
            if (!MainDemo.mainref.publish(n)) {
                System.out.println("Publish of context event failed");
            }
            ret = DONE;
            break;
        case PUB_ARRIVAL:
            n = new Notification();
            n.setType("demochannel");
            n.add(new EventTriplet("user", "demo"));
            n.add(new EventTriplet("subtype", "arrival"));

            if (!MainDemo.mainref.publish(n)) {
                System.out.println("Publish of arrival event failed");
            }
            ret = DONE;
            break;
        case PUB_SYNC:
            n = new Notification();
            n.setType("synchronization");
            n.add(new EventTriplet("target", "meeting.txt"));
            n.add(new EventTriplet("context", "meeting 10.10.2004"));
            n.add(new EventTriplet("group", "demo"));
            if (!MainDemo.mainref.publish(n)) {
                System.out.println("Publish of sync event failed");
            }
            ret = DONE;
            break;
        case PUB_MEETING:
            n = new Notification();
            n.setType("demochannel");
            n.add(new EventTriplet("subtype", "meeting"));
            n.add(new EventTriplet("target", "group"));
            n.add(new EventTriplet("context", "meeting 10.10.2004"));
            if (!MainDemo.mainref.publish(n)) {
                System.out.println("Publish of meeting event failed");
            }
            ret = DONE;
            break;
        case PUB_LUNCH:
            n = new Notification();
            n.setType("demochannel");
            n.add(new EventTriplet("subtype", "lunch"));
            n.add(new EventTriplet("target", "coworkers"));
            n.add(new EventTriplet("context", "How about pizza?"));
            if (!MainDemo.mainref.publish(n)) {
                System.out.println("Publish of lunch event failed");
            }
            ret = DONE;
            break;
        }
        return ret;
    }
}
