/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */
package fuegocore.notify.sync.service;

import java.util.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.Filter;
import fuegocore.notify.core.session.EventSession;

/**
 * SyncService implements the server-side component of the SyncSpace
 * service. Idea is to support continuous collection tracking using
 * filters and events.
 *
 * This is a prototype.
 *
 * The SyncService listens for profile publications
 * from clients and stores those profiles. Profile format is the Template
 * class using String serialization and Fuego filtering language. You could
 * also implement your own custom templates (see test.custom for example).
 * In this case, the benefits of using the event service are:
 * - anonymous and decoupled communication
 * - abstracts the physical address and devices of the clients
 * - allows to match and cover profiles automatically
 *
 * Add and del operations are provided. Modify may be performed using these
 * primitives.
 *
 * Upon change the profile is re-matched against the descriptions.
 * The service could also keep a record of delivered descriptions to
 * prevent flooding the client.
 *
 * Currently, services are points in the content space (Notifications),
 * and profiles are subsets of the content space.
 *
 */
public class SyncService {
    int counter = 0;

    // Vector for profiles
    Vector profiles = new Vector();
    // Vector for service announcements
    Vector services = new Vector();
    EventSession session;

    /** mapping from profiles to receivers */
    HashMap profileMapping = new HashMap();

    public Vector getProfiles() {
        return profiles;
    }

    public Vector getServices() {
        return services;
    }

    public static void main(String[] args) {
        // set server address
        fuegocore.notify.core.client.EventConnection.get().setName(
                "SyncService");
        try {
            if (args != null && args.length > 0 && args[0] != null) {
                fuegocore.notify.core.client.EventConnection.get().setHostPort(
                        args[0],
                        Integer.parseInt(args[1]));
            } else {
                System.out.println("Incorrect arguments: host port");
                System.exit(0);
            }
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(0);
        }
        SyncService test = new SyncService();
        System.out.println("SyncService\n");
    }

    public SyncService() {
        session = new EventSession();
        // Open session
        String sid = session.open();

        if (sid == null) {
            System.out.println("Session open failed");
        } else {
            System.out.println("Session opened. Id: " + sid);
        }

        System.out.println("Creating template 1");

        Template t1 = new Template();

        // Add attribute filters to the template
        // I'm using synchronization channel, but this could be
        // for example a dynamos channel
        Filter f1 = new Filter("type", EventTriplet.TYPE_STRING,
                               "equ synchronization");
        Filter f2 = new Filter("operation", EventTriplet.TYPE_STRING,
                               "equ announce-add|equ announce-del");
        Filter f3 = new Filter("data", EventTriplet.TYPE_SERIALIZED,
                               "exists");

        // Define a listener class
        ProfileChangeListener cl = new ProfileChangeListener(this);
        t1.setSource(cl); // set the source
        t1.addFilter(f1);
        t1.addFilter(f2);
        t1.addFilter(f3);

        Template t2 = new Template();

        // Add attribute filters to the template
        Filter f4 = new Filter("type", EventTriplet.TYPE_STRING,
                               "equ synchronization");
        Filter f5 = new Filter("operation", EventTriplet.TYPE_STRING,
                               "equ profile-change-add|equ profile-change-del");
        Filter f6 = new Filter("data", EventTriplet.TYPE_SERIALIZED,
                               "exists");
        Filter f7 = new Filter("return-address", EventTriplet.TYPE_STRING,
                               "exists");

        // Define a listener class
        QueryChangeListener cl2 = new QueryChangeListener(this);
        t2.setSource(cl2); // set the source
        t2.addFilter(f4);
        t2.addFilter(f5);
        t2.addFilter(f6);
        t2.addFilter(f7);

        // We only subscribe once per session
        if (session.reopenedSession()) {
            // Since the session is already open we only restore the
            // local subscription table
            t1.setScopeToLocal();
            t2.setScopeToLocal();
            session.subscribe(t1);
            session.subscribe(t2);
        } else {
            // Do only local subscriptions. Here if we have server filtered identifiers,
            // the callback information is lost for the client.
            String sret = session.subscribe(t1);
            if (sret == null) {
                System.out.println("Subscription failure 1: " + sret);
                System.exit(0);
            }
            sret = session.subscribe(t2);
            if (sret == null) {
                System.out.println("Subscription failure 2: " + sret);
                System.exit(0);
            }
        }
        System.out.println("Waiting for notifications");

        for (; ; ) {
            try {
                Thread.sleep(1000);
                System.out.println("---- Status ---- ");
                System.out.println("---- Profiles: " + getProfiles().size());
                System.out.println("---- Descriptions: " + getServices().size());
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * Notifies client collection about profile deletion
     * @param n Notification
     */
    public void notifyProfileChange(Notification n) {
        sendNotificationDel(n);
    }

    /**
     * Match profiles against the service description
     * @param desc Notification
     */
    public void matchProfilesAgainstService(Notification desc) {
        System.out.println("*********************** matching profiles " +
                           getProfiles().size() + " ******");
        int match = 0;
        for (int i = 0; i < getProfiles().size(); i++) {
            Template t = (Template) getProfiles().elementAt(i);
            if (t.match(null, desc)) {
                System.out.println("********** match ******");
                sendNotification(t, desc);
                match++;
            }
        }
        System.out.println("*** matched " + match);
    }

    /**
     * Match service descriptions against the given profile
     * @param profile Template
     */
    public void matchServicesAgainstProfile(Template profile) {
        System.out.println("*********************** matching " +
                           getServices().size() + " services ******");
        int match = 0;
        for (int i = 0; i < getServices().size(); i++) {
            Notification n = (Notification) getServices().elementAt(i);
            if (profile.match(null, n)) {
                System.out.println("********** match ******");
                sendNotification(profile, n);
                match++;
            }
        }
        System.out.println("*** matched " + match);
    }

    /** Sends a profile-add notification to the source address of the profile */
    public void sendNotification(Template t, Notification n) {
        System.out.println("#################### sendNotification");
        String address = t.getInterface();
        if (address == null) {
            System.out.println("sendNotification: Address was null");
        } else {
            Notification n2 = new Notification();
            n2.setType("synchronization");
            n2.add(new EventTriplet("op", "add"));
            n2.add(new EventTriplet("receiver-address", address));
            n2.add(new EventTriplet("data", EventTriplet.TYPE_SERIALIZED, n));
            System.out.println("---- sending to: " + address);
            System.out.println("sending: " + n.toString());
            session.publish(n2);

            // Store a record of this update
            HashSet pf = (HashSet) profileMapping.get(n);
            if (pf == null) {
                pf = new HashSet();
                profileMapping.put(n, pf);
            }
            pf.add(address);
        }
    }

    /** Sends a profile-del notification to the source addresses of the profile */
    public void sendNotificationDel(Notification n) {
        System.out.println("#################### sendNotification del");

        // Store a record of this update
        HashSet pf = (HashSet) profileMapping.get(n);
        if (pf == null) {
            System.out.println("Addresses not found");
            return;
        }
        Iterator i = pf.iterator();
        while (i.hasNext()) {
            String address = (String) i.next();
            if (address == null) {
                System.out.println("sendNotificationDel: Address was null");
            } else {
                Notification n2 = new Notification();
                n2.setType("synchronization");
                n2.add(new EventTriplet("op", "del"));
                n2.add(new EventTriplet("receiver-address", address));
                n2.add(new EventTriplet("data", EventTriplet.TYPE_SERIALIZED, n));
                System.out.println("---- sending to: " + address);
                System.out.println("sending: " + n.toString());
                session.publish(n2);
            }
        }
        pf.clear();
        profileMapping.remove(n);
    }
}
