/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.sync.service;

import java.util.Vector;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.Filter;
import fuegocore.notify.core.session.EventSession;
import fuegocore.notify.sync.client.ISyncCallback;
import fuegocore.notify.sync.client.ISyncSpace;

/**
 * SyncSpace implements the client-side component of the SyncSpace
 * service. Idea is to support continuous collection tracking using
 * filters and events.
 *
 * This is a prototype.
 */
public class SyncSpace implements ISyncSpace {
    int counter = 0;
    EventBase evref;
    String sid;
    EventSession session;

    /** The advertised objects */
    Vector objects = new Vector();

    /** The queries objects */
    Vector queries = new Vector();

    /** The received profiles */
    Vector profiles = new Vector();

    private Vector listeners = new Vector();

    public SyncSpace(EventBase ef) {
        sid = "" + this.hashCode();
        evref = ef;

        session = new EventSession(sid);
        String sid2 = session.open();
        if (sid2 == null) {
            System.out.println("Session open failed");
        } else {
            System.out.println("Session opened. Id: " + sid2);
        }

        System.out.println("Creating template 1");

        Template t1 = new Template();
        // Add attribute filters to the template
        Filter f1 = new Filter("type", EventTriplet.TYPE_STRING,
                               "equ synchronization");
        Filter f2 = new Filter("receiver-address", EventTriplet.TYPE_STRING,
                               "equ " + sid);

        // Define a listener class
        SubListener cl = new SubListener(this);
        t1.setSource(cl); // set the source
        t1.addFilter(f1);
        t1.addFilter(f2);

        // We only subscribe once per session
        if (session.reopenedSession()) {
            // Since the session is already open we only restore the
            // local subscription table
            t1.setScopeToLocal();
            session.subscribe(t1);
        } else {
            // Do only local subscriptions. Here if we have server filtered identifiers,
            // the callback information is lost for the client.
            String sret = session.subscribe(t1);
            if (sret == null) {
                System.out.println("Subscription failure 1: " + sret);
                System.exit(0);
            }
        }
    }

    /**
     * Add an object into the synced space
     * @param obj Object the filter (context profile) must contain URI!
     * @return boolean
     */
    public boolean add(Object obj) {
        if (objects.contains(obj)) {
            return false;
        }

        Notification n = new Notification();
        n.setType("synchronization");
        n.add(new EventTriplet("operation", "announce-add"));
        n.add(new EventTriplet("data", EventTriplet.TYPE_SERIALIZED,
                               obj));
        System.out.println("----------------\nDesc: " + obj);
        System.out.println("Publishing event");
        boolean ret = session.publish(n);
        if (ret) {
            objects.add(obj);
        }
        return ret;
    }

    /**
     * Remove an object from the synced space
     * @param obj Object the filter (context profile) must contain URI!
     * @return boolean
     */
    public boolean remove(Object obj) {
        if (!objects.contains(obj)) {
            return false;
        }
        Notification n = new Notification();
        n.setType("synchronization");
        n.add(new EventTriplet("operation", "announce-del"));
        n.add(new EventTriplet("data", EventTriplet.TYPE_SERIALIZED,
                               obj));
        System.out.println("----------------\nDesc: " + obj);
        System.out.println("Publishing event");
        boolean ret = session.publish(n);
        if (ret) {
            objects.remove(obj);
        }
        return ret;
    }

    /**
     * Delete the collection. Closes session.
     * @return boolean
     */
    public boolean delete() {
        for (int i = 0; i < objects.size(); i++) {
            Object obj = objects.elementAt(i);
            remove(obj);
        }
        for (int i = 0; i < queries.size(); i++) {
            Object obj = queries.elementAt(i);
            removeQuery(obj);
        }
        profiles.clear();
        session.close();
        return true;
    }

    /**
     * Adds the given context query
     * @param obj Object
     * @return boolean
     */
    public boolean addQuery(Object obj) {
        if (queries.contains(obj)) {
            return false;
        }

        Notification n = new Notification();
        n.setType("synchronization");
        n.add(new EventTriplet("operation", "profile-change-add"));
        n.add(new EventTriplet("return-address", sid));
        n.add(new EventTriplet("data", EventTriplet.TYPE_SERIALIZED,
                               obj));
        boolean ret = session.publish(n);
        if (ret) {
            queries.add(obj);
        }
        return ret;

    }

    /**
     * Remove the given context query
     * @param obj Object
     * @return boolean
     */
    public boolean removeQuery(Object obj) {
        if (!queries.contains(obj)) {
            return false;
        }
        Notification n = new Notification();
        n.setType("synchronization");
        n.add(new EventTriplet("operation", "profile-change-del"));
        n.add(new EventTriplet("return-address", sid));
        n.add(new EventTriplet("data", EventTriplet.TYPE_SERIALIZED,
                               obj));
        boolean ret = session.publish(n);
        if (ret) {
            queries.remove(obj);
        }

        // remove all objects in the collection that match only
        // this template
        Vector delList = new Vector();

        for (int j = 0; j < profiles.size(); j++) {
            boolean others = false;
            Notification e = (Notification) profiles.elementAt(j);
            for (int i = 0; i < queries.size(); i++) {
                Template t = (Template) queries.elementAt(i);
                if (t.match(null, e)) {
                    others = true;
                    break;
                }
            }
            if (!others) {
                delList.add(e);
            }
        }
        profiles.removeAll(delList);
        return ret;
    }

    /**
     * Returns the sync collection as ITemplates (must contain URIs)
     * @return Vector
     */
    public Vector getCollection() {
        return profiles;
    }

    /**
     * Add a profile to the collection
     * @param obj Object
     * @return boolean
     */
    boolean addCollectionProfile(Object obj) {
        if (!profiles.contains(obj)) {
            profiles.add(obj);
        }
        fireEventToListeners(this);
        return true;
    }

    /**
     * Remove a profile from the collection
     * @param obj Object
     * @return boolean
     */
    boolean removeCollectionProfile(Object obj) {
        Notification n2 = (Notification) obj;
        // Have to do a string or identifier match
        for (int i = 0; i < profiles.size(); i++) {
            Notification n = (Notification) profiles.elementAt(i);
            if (n.toString().equals(obj.toString())) {
                profiles.remove(n);
                System.out.println("Profile removed");
                break;
            }
        }
        fireEventToListeners(this);
        return true;
    }

    /**
     * Adds a change listener to the data structure
     * @param cl ISyncCallback
     * @return boolean
     */
    public boolean addListener(ISyncCallback cl) {
        return listeners.add(cl);
    }

    /**
     * Adds a change listener to the data structure
     * @param cl ISyncCallback
     * @return boolean
     */
    public boolean removeListener(ISyncCallback cl) {
        return listeners.remove(cl);
    }

    /**
     * Fires an internal event to the listeners
     * @param event String
     * @param x IFBase
     */
    private void fireEventToListeners(Object event) {
        for (int i = 0; i < listeners.size(); i++) {
            ISyncCallback pl = (ISyncCallback) listeners.elementAt(i);
            pl.collectionChanged(event);
        }
    }
}


/**
 * Listener class for context updates
 */
class SubListener implements Notifiable {
    int count = 0;
    SyncSpace _eb;

    public SubListener(SyncSpace eb) {
        _eb = eb;
    }

    public int getCount() {
        return count;
    }

    /**
     * Process incoming collection updates
     * @param n Notification
     */
    public void notifyIncoming(Notification n) {
        // Retrieve the tuple value
        EventTriplet et = (EventTriplet) n.getField(EventTriplet.formatId(
                "op",
                EventTriplet.TYPE_STRING));

        String op = (String) et.getValue();

        // Retrieve the tuple value
        et = (EventTriplet) n.getField(EventTriplet.formatId(
                "data",
                EventTriplet.TYPE_SERIALIZED));

        if (et == null) {
            System.out.println("Data was null");
            return;
        }

        Notification data = (Notification) et.getValue();
        System.out.println("Received op: " + op + " update: " + data);

        if (op != null && op.equals("add")) {
            _eb.addCollectionProfile(data);
        } else if (op != null && op.equals("del")) {
            _eb.removeCollectionProfile(data);
        }

        count++;
    }

}
