/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.test.custom.core;

import java.io.*;
import java.util.*;

import fuegocore.message.service.*;
import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.filter.poset.*;
import fuegocore.notify.core.server.*;
import fuegocore.notify.core.server.engine.*;
import fuegocore.notify.core.server.manager.*;
import fuegocore.notify.core.server.ui.*;
import fuegocore.notify.util.*;

/**
 * Example of a simple router
 */
public class SimpleRouter implements IRoutingEngine {
    /** Router html interface */
    SimpleUI simpleUi = new SimpleUI();

    EventServer _es;

    Hashtable ectable = new Hashtable();

    IFilterPoset pos; // for subscriptions

    /** Subscriptions stored by their identifier (SessionId) */
    RoutingTable subscriptionsById = new RoutingTable();

    public SimpleRouter() {}

    /**
     * Set server reference
     * @param es EventServer
     */
    public void setServerReference(EventServer es) {
        _es = es;
    }

    /**
     * Returns the filter poset
     * @param s String event type
     * @return IFilterPoset
     */
    public IFilterPoset getPoset(String s) {
        if (pos == null) {
            pos = PosetService.get(); // get default implementation
        }
        return pos;
    }

    public SimpleRouter(EventServer es) {
        _es = es;
    }

    /**
     * Return subscriptions by interface identifier
     * @param name String
     * @return Vector
     */
    public Vector getTemplatesForId(String name) {
        return subscriptionsById.get(name);
    }

    public RoutingTable getTemplates() {
        return subscriptionsById;
    }

    /**
     * Match the notification against the routing tables (local operation only)
     * @param n Notification
     * @return Vector
     */
    public Vector match(Notification n) {
        // get poset for the type
        IFilterPoset pos = getPoset(n.getType());
        if (pos == null) {
            return null;
        }
        return pos.match(n);
    }

    /**
     * Match the notification against the routing tables (local operation only)
     * @param n Notification
     * @return TripletTable
     */
    public TripletTable matchT(Notification n) {
        // get poset for the type
        IFilterPoset pos = getPoset(n.getType());
        if (pos == null) {
            return null;
        }
        return pos.matchT(n);
    }

    /**
     * Return advertisements by interface identifier
     * @param name String
     * @return Vector
     */
    public Vector getAdvertisementsForId(String name) {
        return null;
    }

    /**
     * Return quench by interface identifier
     * @param name String
     * @return Vector
     */
    public Vector getQuenchForId(String name) {
        return null;
    }

    /**
     * Router control update message handler
     * @param msg Message
     */
    public void router(fuegocore.message.service.Message msg) {

    }

    /**
     * Get the channel manager reference
     * @return ChannelManager
     */
    public RouterManager getRouterManager() {
        return null;
    }

    /**
     * Called for messages that are not handled in the general access
     * server handler. This implements, for example, event channel specific
     * operations.
     * @param msg Message
     * @return Object
     */
    public Object defaultBackAccessServerHandler(Message msg) throws
            IOException {
        return null;
    }

    /**
     * Subscribe
     * @param t Template
     * @return boolean
     */
    public boolean subscribe(ITemplate t) {

        if (t.getInterface() == null) {
            Log.write("Interface was null");
            Log.write("Template: " + t.toString());
            return false;
        }

        IFilterPoset pos = null;

        pos = getPoset(t.getType());

        if (pos.removable((IFBase) t, t.getSubId())) {
            Log.write(
                    "A subscription with the same ID was ALREADY in the system");
            return false;
        }

        // Add filter to poset
        // Note that we add using the sub identifier in order to do
        // server-side matching for all clients, this is done only
        // for clients, router susbscriptions are processed elsewhere
        int rcode = FilterForest.POSET_ERROR;

        try {
            rcode = pos.add((IFBase) t, t.getSubId());
        } catch (PosetException e) {
            rcode = FilterForest.POSET_ERROR;
            e.printStackTrace();
        }

        if (rcode == FilterForest.POSET_ERROR) {
            Log.write("Subscribe: filter poset error: " + rcode);
            return false;
        }

        subscriptionsById.insert(t.getInterface(), t);
        return true;
    }

    /**
     * Unsubscribe
     * @param t Template
     * @return boolean
     */
    public boolean unsubscribe(ITemplate t) {

        Log.write("Removing subscription");

        IFilterPoset pos = null;

        pos = getPoset(t.getType());

        if (!pos.removable((IFBase) t, t.getSubId())) {
            Log.write("Cannot remove template!");
            Log.write("Template:" + t.toString());
            return false;
        }

        // Get reference to the original template
        ITemplate t0 = (ITemplate) pos.lookup((IFBase) t);

        /** remove filter */
        int rcode = FilterForest.POSET_ERROR;
        try {
            rcode = pos.remove((IFBase) t, t.getSubId());
        } catch (PosetException e) {
            rcode = FilterForest.POSET_ERROR;
            e.printStackTrace();
        }

        if (rcode == FilterForest.POSET_ERROR) {
            Log.write("Unsubscribe: POSET ERROR");
            return false;
        }

        subscriptionsById.removeUsingContent(t.getInterface(), t);
        return true;
    }

    /** Subscribe aggregate.
     *  Return string that contains failure code (the number of first failed
     *  notification in the given list)
     * @param v the aggregate
     * @return String the return value
     */
    public String subscribe(Vector v) {
        String ret = "failure";
        if (v == null) {
            return ret;
        }

        try {
            Log.write("SubscribeAll: size was " + v.size());
            for (int i = 0; i < v.size(); i++) {
                ITemplate t = (ITemplate) v.elementAt(i);
                if (!subscribe(t)) {
                    ret = "failure " + i;
                    // we rollback
                    for (int j = 0; j < i; j++) {
                        ITemplate tt = (ITemplate) v.elementAt(j);
                        unsubscribe(tt);
                    }
                    return ret;
                }
            }
            ret = "success";
        } catch (Exception e) {
            e.printStackTrace();
        }
        return ret;
    }

    /** Unsubscribe aggregate.
     *  Return string that contains failure code (the number of first failed
     *  notification in the given list). Currently this may cause multiple
     *  updates to event channels if non-lazy updates are used with the merging
     *  mode.
     * @param v the aggregate
     * @return String the return value
     */
    public String unsubscribe(Vector v) {
        String ret = "failure";
        boolean fail = false;
        int count = 0;
        if (v == null) {
            return ret;
        }

        try {
            for (int i = 0; i < v.size(); i++) {
                ITemplate t = (ITemplate) v.elementAt(i);
                if (!unsubscribe(t)) {
                    fail = true;
                    ret += i + " ";
                } else {
                    count++;
                }
            }
            if (!fail) {
                ret = "success";
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return ret;
    }

    /**
     * Advertise
     * @param t Template
     * @return boolean
     */
    public boolean advertise(ITemplate t) {
        return false;
    }

    /**
     * Unadvertise
     * @param t Template
     * @return boolean
     */
    public boolean unadvertise(ITemplate t) {
        return false;
    }

    /**
     * Advertise
     * @param v Vector
     * @return String
     */
    public String advertise(Vector v) {
        return null;
    }

    /**
     * Unadvertise
     * @param v Vector
     * @return String
     */
    public String unadvertise(Vector v) {
        return null;
    }

    /**
     * Quench. Quench semantics: client receives a notification
     * when the given template has subscribers (or when the last subscriber
     * unsubscribes)
     * @param t Template
     * @return boolean
     */
    public boolean quench(ITemplate t) {
        return false;
    }

    /**
     * Unquench. Quench semantics: client receives a notification
     * when the given template has subscribers (or when the last subscriber
     * unsubscribes)
     * @param t Template
     * @return boolean
     */
    public boolean unquench(ITemplate t) {
        return false;
    }

    /**
     * Quench. Quench semantics: client receives a notification
     * when the given template has subscribers (or when the last subscriber
     * unsubscribes)
     * @param v Vector
     * @return String
     */
    public String quench(Vector v) {
        return null;
    }

    /**
     * Unquench. Quench semantics: client receives a notification
     * when the given template has subscribers (or when the last subscriber
     * unsubscribes)
     * @param v Vector
     * @return String
     */
    public String unquench(Vector v) {
        return null;
    }

    /**
     * Returns an instance responsible for router html UI
     * @return RouterUI
     */
    public RouterUI getUI() {
        return simpleUi;
    }

    /**
     * Publish
     * @param n Notification
     * @return boolean
     */
    public boolean publish(Notification n) {
        _es.route(n);
        return true;
    }

    /** Publish aggregate.
     *  Return string that contains failure code (the number of first failed
     *  notification in the given list)
     * @param v the aggregate
     * @return String the return value
     */
    public String publish(Vector v) {
        String ret = "failure";
        if (v == null) {
            return ret;
        }

        try {
            for (int i = 0; i < v.size(); i++) {
                Notification n = (Notification) v.elementAt(i);
                if (!publish(n)) {
                    ret = "failure " + i;
                    return ret;
                }
            }
            ret = "success";
        } catch (Exception e) {
            e.printStackTrace();
        }
        return ret;
    }

    /* ************************************************************************ */

    /**
     * Debug method to retrieve the current subscriptions at the event channel
     * @return Vector
     */
    public synchronized Vector getSubscriptions() {
        Vector all = getPoset(null).getAllNodes();
        Vector temps = new Vector();
        for (int i = 0; i < all.size(); i++) {
            Object o = all.elementAt(i);
            if (o instanceof ITemplate) {
                temps.add((ITemplate) o);
            }
        }
        return temps;
    }

    /**
     * Returns subscriptions for the given interface
     * @param iface the interface
     * @return Vector
     */
    public synchronized Vector getSubscriptions(String iface) {
        Vector all = getPoset(null).getAllNodes();
        Vector temps = new Vector();
        for (int i = 0; i < all.size(); i++) {
            Object o = all.elementAt(i);
            if (o instanceof ITemplate) {
                ITemplate t = (ITemplate) o;
                if (t.getInterface().equals(iface)) {
                    temps.add((ITemplate) o);
                }
            }
        }
        return temps;
    }

    public String subIDtoIface (String subid) {
        // XXX Probably incorrect but we're not using this now
        return subid;
    }
}
