/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.test.custom.core;

import java.io.*;
import java.util.*;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.core.serialization.*;
import fuegocore.notify.util.*;
import fuegocore.util.xas.*;

/** Simple example of a custom template. Implements topic-based matching.
 *  Each notification must have a topic field.
 */
public class SimpleTemplate implements ITemplate, IFBase,
        IEventSequenceSerializable,
        IStringSerializable {
    private String subId;
    private String sessionId;
    private String iface;
    private Notifiable source;
    private Vector filterSet;
    private Vector tid;
    private boolean scope = false;

    public SimpleTemplate() {
        filterSet = new Vector();
    }

    /**
     * Set quench value
     * @param b boolean
     */
    public void setQuench(boolean b) {}

    /**
     * Get quench value
     * @return boolean
     */
    public boolean getQuench() {
        return false;
    }

    /** Merges the current object with the given object
     * @param obj the object to be merged with the current object
     * @return Object the merged result
     */
    public Object mergeObject(Object obj) {
        return null;
    }


    /**
     * Flag the template as an advertisement
     * @param b boolean
     */
    public void setAdvertisement(boolean b) {}

    /**
     * Returns the advertisement flag
     * @return boolean
     */
    public boolean getAdvertisement() {
        return false;
    }

    /** Set the subscription ID
     * @param s the id
     */
    public void setSubId(String s) {
        subId = s;
    }

    /** Get the subscription ID
     * @return the subscription identifier
     * */
    public String getSubId() {
        return subId;
    }

    /** Set the session ID
     * @param s the session ID to set
     */
    public void setSessionId(String s) {
        sessionId = s;
    }

    /** Get the session ID
     * @return the session identifier
     */
    public String getSessionId() {
        return sessionId;
    }

    /** Formats a notification according to the required notification fields
     * specified in this template
     * @param n the notification
     * @return the formatted notification
     */
    public Notification formatNotification(Notification n) {
        return n;
    }

    /** Sets the local send property. If this is true notifications from the client are forwarded back to the client
     *  if they match one of the client's active subscriptions.
     * @param b the local send option
     */
    public void setLocalSend(boolean b) {}

    /** Returns the local send property.  If this is true notifications from the client are forwarded back to the client
     *  if they match one of the client's active subscriptions.
     * @return true if local send is active
     */
    public boolean getLocalSend() {
        return false;
    }

    /** Sets the scope of the subscription (template) to local */
    public void setScopeToLocal() {
        scope = true;
    }

    /** Sets the scope of the subscription (template) to external */
    public void setScopeToExternal() {
        scope = false;
    }

    /** Returns the current scope local (true) / external (false)
     * @return true if scope is local */
    public boolean getScope() {
        return scope;
    }

    /** Adds the given filter to the template.
     * @param f the attribute filter to add
     */
    public void addFilter(Object f) {
        filterSet.add(f);
    }

    /** Removes the given filter.
     * @param f the filter to remove
     */
    public void removeFilter(Object f) {
        filterSet.remove(f);
    }

    /** Returns the set of filters.
     * @return returns a Vector containing the filter objects
     */
    public Vector getFilters() {
        return filterSet;
    }

    /** Set the attribute filters
     *  @param v the Vector
     */
    public void setFilters(Vector v) {
        filterSet = v;
    }

    /* *************************************************************************/
    /* *************************************************************************/

    /**
     * Set the subscribing interface
     * @param s String
     */
    public void setInterface(String s) {
        iface = s;
    }

    /**
     * Get the subscribing interface
     * @return String
     */
    public String getInterface() {
        return iface;
    }

    /** Add the type of the subscription to the template (event channel name)
     * @param type the type to add
     */
    public void addType(String type) {}

    /** Returns the type of the subscription (event channel name)
     * @return String the type of the notification */
    public String getType() {
        return "custom";
    }

    /** Get the source of the subscription / template
     * @return Notifiable the source of the template
     */
    public Notifiable getSource() {
        return source;
    }

    /** Set the source of the subscription / template
     * @param eb the Notifiable interface
     */
    public void setSource(Notifiable eb) {
        source = eb;
    }

    /**
     * Set event channel flag (internal use)
     * @param b boolean
     */
    public void setECFlag(boolean b) {}

    /**
     * Get event channel flag (internal use)
     * @return boolean
     */
    public boolean getECFlag() {
        return false;
    }

    /** Returns the notify field names (separated by commas) as a String.
     *  The notify fields determine which fields are forwarded in the
     *  notification and which fields are omitted. This is an optional feature.
     *  @return the notify field names
     */
    public String getNotifyFields() {
        return null;
    }

    /** Sets the notify field names (separated by commas) as a String.
     *  @param s the notify fields
     */
    public void setNotifyFields(String s) {}

    /** Matches the object with the given object
     * @param obj the EventTriplet object to match, may be null
     * @param noti the TripletTable, may not be null
     * @return boolean true if the match succeeded
     */
    public boolean match(Object obj, Object noti) {
        if (noti instanceof Notification) {
            Notification n = (Notification) noti;

            // Retrieve the tuple value
            EventTriplet et = (EventTriplet) n.getField(
                    EventTriplet.formatId("topic", EventTriplet.TYPE_STRING));
            if (et != null) {
                String topic = (String) et.getValue();
                for (int i = 0; i < filterSet.size(); i++) {
                    if (topic.equals((String) filterSet.elementAt(i))) {
                        return true;
                    }
                }
            }
        }
        return false;
    }

    /** Merges the object with the given object
     * @param obj the object to be merged
     * @return boolean true if the merge succeeded
     */
    public boolean merge(Object obj) {
        return false;
    }

    /**
     * Covering test: is the given object covered
     * @param obj the object that covers
     * @return true if the current object was covered
     */
    public boolean isCovered(Object obj) {
        return false;
    }

    /**
     * Covering test
     * @param obj the object that is covered
     * @return true if the current object covers the given object
     */
    public boolean covers(Object obj) {
        return false;
    }

    /**
     * Overlapping test
     * @param obj the input object
     * @return true if the current object overlaps with the given object
     */
    public boolean overlaps(Object obj) {
        return false;
    }

    /**
     * Get id for hash lookups.
     * @return String the identifier
     */
    public String getCoverId() {
        return toString();
    }

    /**
     * Add an interface identifier (used by posets)
     * @param id Object
     */
    public synchronized void addTid(Object id) {
        if (tid == null) {
            tid = new Vector();
        }
        tid.add(id);
    }

    /**
     * Returns the interface vector (multiset)
     * @return Vector
     */
    public Vector getTid() {
        return tid;
    }

    /**
     * Set the depth of this object in the poset tree.
     * @param i the depth to set
     */
    public void setDepth(int i) {}

    /**
     * Returns the depth of this object in the poset tree.
     * @return the depth as int
     */
    public int getDepth() {
        return 0;
    }

    /** Returns the cost of the filter (generality).
     *  Higher value denotes more general filters.
     * @return int the cost
     */
    public int getCost() {
        return 0;
    }

    /**
     * Returns the size of the object
     * If filter the # disjuncts, if disjunct the # predicates
     * @return the size
     */
    public int getSize() {
        return filterSet.size();
    }

    public boolean noSelectivity() {
        return false;
    }

    /* *** SERIALIZATION **************************************************/

    public SimpleTemplate(String str) {
        parse(str);
    }

    public String toString() {
        String ret = "";
        if (filterSet == null) {
            return "";
        }
        for (int i = 0; i < filterSet.size(); i++) {
            String s = (String) filterSet.elementAt(i);
            ret += s + "\n";
            ret += "<dlm/>\n";
        }
        ret += "<id>" + subId + "</id>";
        ret += "<interface>" + iface + "</interface>";
        ret += "<sessionid>" + sessionId + "</sessionid>";
        return ret;
    }

    /**
     * Parse object from string
     * @param str String
     * @return boolean
     */
    public boolean parse(String str) {
        int i = 0;
        int j = str.indexOf("<dlm/>");
        if (j == -1) {
            return false;
        }

        if (filterSet == null) {
            filterSet = new Vector();
        } else {
            filterSet.clear();

        } while (j != -1) {
            String s = str.substring(i, j);
            filterSet.add(s);
            i = j + "<dlm/>".length();
            j = str.indexOf("<dlm/>", i);
        }

        try {
            subId = Auxil.getTag(str, "id");
        } catch (Exception e) {
            subId = null;
            e.printStackTrace();
        }
        try {
            iface = Auxil.getTag(str, "interface");
        } catch (Exception e) {
            e.printStackTrace();
        }
        try {
            sessionId = Auxil.getTag(str, "sessionid");
        } catch (Exception e) {
            e.printStackTrace();
        }
        return true;
    }

    /* *************************************************************************/
    /* *************************************************************************/
    /** XAS Serialization */

    /** Default constructor
     * @param es the EventSequence representation for parsing the notification
     */
    public SimpleTemplate(EventSequence es) {
        fromEventSequence(es);
    }

    /**
     * Parse object from the given event sequence
     * @param es EventSequence
     * @return boolean
     */
    public boolean fromEventSequence(EventSequence es) {
        XmlReader xr = new XmlReader(es);

        boolean quit = false;

        // check the root tag for a filter
        EventSequence esf = xr.completeContent(Filter.EVENT_NAMESPACE,
                                               "simplefilter");

        // Check that the element is correct
        if (esf == null) {
            System.out.println("SimpleTemplate: invalid root element");
            return false;
        }

        xr = new XmlReader(esf);

        // now parse the rest of the filter
        String tmp = (String) xr.typedContent("", "topics",
                                              XasUtil.XSD_NAMESPACE, "string");
        if (!decodeTopics(tmp)) {
            return false;
        }

        subId = (String) xr.typedContent("", "id", XasUtil.XSD_NAMESPACE,
                                         "string");
        iface = (String) xr.typedContent("", "interface", XasUtil.XSD_NAMESPACE,
                                         "string");
        sessionId = (String) xr.typedContent("", "sessionid",
                                             XasUtil.XSD_NAMESPACE,
                                             "string");

        return true;
    }

    private boolean decodeTopics(String str) {
        if (str == null) {
            return false;
        }
        StringTokenizer tok = new StringTokenizer(str, ",");
        if (filterSet == null) {
            filterSet = new Vector();
        } else {
            filterSet.clear();
        } while (tok.hasMoreElements()) {
            String s = (String) tok.nextElement();
            filterSet.add(s);
        }
        return true;
    }

    private String encodeTopics() {
        String str = "";
        for (int i = 0; i < filterSet.size(); i++) {
            str += filterSet.elementAt(i);
            if (i < filterSet.size() - 1) {
                str += ",";
            }
        }
        return str;
    }

    /**
     * Insert the filters to a SAX event sequence.
     * @param ser TypedXmlSerializer
     * @throws IOException
     */
    public void toEventSequence(TypedXmlSerializer ser) throws IOException {
        XmlWriter xw = new XmlWriter(ser);

        EventSerializer evs = new EventSerializer();
        XmlWriter xw2 = new XmlWriter(evs);

        xw2.typedElement("", "topics", XasUtil.XSD_NAMESPACE, "string",
                         encodeTopics());
        xw2.typedElement("", "id", XasUtil.XSD_NAMESPACE, "string", subId);
        xw2.typedElement("", "interface", XasUtil.XSD_NAMESPACE, "string",
                         iface);
        xw2.typedElement("", "sessionid", XasUtil.XSD_NAMESPACE, "string",
                         sessionId);
        xw.complexElement(Filter.EVENT_NAMESPACE, "simplefilter",
                          evs.getCurrentSequence());
    }
}
