/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.test.pubsub;

import fuegocore.notify.core.*;
import fuegocore.notify.core.filter.*;
import fuegocore.notify.profile.*;

/**
 * Demonstrates local event passing and listener classes.
 */
/*
 Template (subscription) looks like this, these could be
 loaded from an XML document:
 <name>subtype</name><type>3</type>
  <filter>in change</filter>
 <dlm/>
 <name>month</name><type>0</type>
  <filter>equ 5.0|equ 6.0</filter>
 <dlm/>
 <name>day</name><type>0</type>
  <filter>lt 10.0&gr 5.0</filter><dlm/>
 <nfields>*</nfields>
 <localsend>true</localsend>
 <expiration></expiration>
 <id>filled by system</id>
 <interface>filled by system</interface></textblock>
 */

public class Tester extends EventBase {

    int count1_ref = 0; // test variable

    String month[] = new String[] {
                     "January",
                     "February", "March", "April", "May",
                     "June", "July", "August", "September", "October",
                     "November", "December"};

    String day[] = new String[] {
                   "Monday",
                   "Tuesday", "Wednesday", "Thursday", "Friday",
                   "Saturday", "Sunday"};

    public static void main(String[] args) {
        UserProfile.setName("user");

        String arg = "test";

        if (args != null && args.length > 0 && args[0] != null) {
            arg = args[0];

        }
        Tester test = new Tester(arg);
    }

    public Tester(String gateway) {
        // Create a subscription
        // Note that local events do not need a session

        // All events pertaining to May or June between 5 and 10
        // and must not be on the weekend
        Template t = new Template();
        Filter f0 = new Filter("type", EventTriplet.TYPE_STRING,
                               "equ calendar");
        Filter f1 = new Filter("subtype", EventTriplet.TYPE_STRING,
                               "in change");
        Filter f2 = new Filter("month", EventTriplet.TYPE_INT,
                               "equ 5|equ 6"); // equivalent to "gte 2&lte 6"
        Filter f3 = new Filter("day", EventTriplet.TYPE_INT,
                               "gr 5&lt 10");
        Filter f4 = new Filter("date", EventTriplet.TYPE_STRING,
                               "nin Saturday&nin Sunday");
        // Filter 4 is equivalent to (we support duplicate names with different types
        //	Filter f4 = new Filter("date",EventTriplet.TYPE_INT,
        //	"neq 5&neq 6");

        // Make it local (otherwise external)
        t.setScopeToLocal();

        // Add filters to subscription
        t.addFilter(f0);
        t.addFilter(f1);
        t.addFilter(f2);
        t.addFilter(f3);
        t.addFilter(f4);
        t.setSource(this);

        String id = subscribe(t);
        if (id == null) {
            System.out.println("Local subscribe failed");
            System.exit(0);
        }

        /* SECOND SUBSCRIPTION */
        /* This demonstrates the use of external listener class */
        Template t2 = new Template();
        Filter f5 = new Filter("type", EventTriplet.TYPE_STRING,
                               "equ calendar");
        Filter f6 = new Filter("subtype", EventTriplet.TYPE_STRING,
                               "in change");
        Filter f7 = new Filter("month", EventTriplet.TYPE_INT,
                               "equ 5|equ 6");

        // Make it local (otherwise external)
        t2.setScopeToLocal();

        // Add filters to subscription
        t2.addFilter(f5);
        t2.addFilter(f6);
        t2.addFilter(f7);

        CalendarListener cl = new CalendarListener();

        t2.setSource(cl); // set the source
        id = subscribe(t2);

        if (id == null) {
            System.out.println("Local subscribe failed");
            System.exit(0);
        }

        /* PLACEHOLDER TEST */
        Template t3 = new Template("calendar");
        Filter f8 = new Filter("type", EventTriplet.TYPE_STRING,
                               "equ calendar");
        Filter f9 = new Filter("subtype", EventTriplet.TYPE_STRING,
                               "in $date&nin Monday");
        Filter f10 = new Filter("month", EventTriplet.TYPE_INT,
                                "equ 5|equ 6");

        // Make it local (otherwise external)
        t3.setScopeToLocal();

        // Add filters to subscription
        t3.addFilter(f8);
        t3.addFilter(f9);
        t3.addFilter(f10);
        CalendarListener cl2 = new CalendarListener();

        t3.setSource(cl2); // set the source
        id = subscribe(t3);
        if (id == null) {
            System.out.println("Local subscribe failed");
            System.exit(0);
        }

        /* Now, we test the subscription */
        int count1 = 0, count2 = 0, count3 = 0;
        for (int j = 0; j < 6; j++) {
            for (int i = 0; i < 30; i++) {
                try {
                    int d = i % 7;
                    Notification n = new Notification();
                    n.setType("calendar");
                    // Make it local (otherwise external)
                    n.setScopeToLocal();
                    n.add(new EventTriplet("subtype", "change " + day[d]));
                    n.add(new EventTriplet("month", j));
                    n.add(new EventTriplet("day", i));
                    n.add(new EventTriplet("date", day[d]));
                    n.add(new EventTriplet("date", d));

                    if (j == 5 || j == 6) {
                        count2++;
                        if (i > 5 && i < 10 && d != 5 && d != 6) {
                            count1++;
                        }
                        if (d != 0) {
                            count3++;
                        }
                    }

                    // Publish the event
                    publish(n);
                } catch (Exception e) {
                    e.printStackTrace();
                    System.out.println("Exception: " + e);
                }
            }
        }
        System.out.println("Template 1: " + count1 + " test: " + count1_ref);
        System.out.println("Template 2: " + count2 + " test: " + cl.getCount());
        System.out.println("Template 3: " + count3 + " test: " + cl2.getCount());
    }

    public void notifyIncoming(Notification n) {
        System.out.println("Received notification " + (count1_ref++) + ": " +
                           n.getType() + "," + n.get());
    }

}


class CalendarListener implements Notifiable {
    int count = 0;

    public int getCount() {
        return count;
    }

    public void notifyIncoming(Notification n) {
        // Retrieve the tuple value
        EventTriplet et = (EventTriplet) n.getField(EventTriplet.formatId("day",
                EventTriplet.TYPE_INT));
        if (et != null) {
            System.out.println("Got value: " +
                               ((Integer) et.getValue()).intValue());
        }
        count++;
    }

}
