/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.util;

import java.util.*;

import fuegocore.notify.core.*;

/**
 * Implements a simple filter table based on a hashtable and vectors.
 * Preserves tuple order.
 */

public class TripletTable {
    private Vector entries = new Vector();
    private Hashtable tbl = new Hashtable();

    public void put(String obj, Object obj2) {
        insert(obj, obj2);
    }

    /**
     * Get triplets, order is not guaranteed
     * @return Enumeration
     */
    public Enumeration elementsNoOrder() {
        return tbl.elements();
    }

    /**
     * Get triplets in order
     * @return Enumeration
     */
    public Enumeration elements() {
        return entries.elements();
    }

    /**
     *  Get triplets in order
     * @return Vector
     */
    public Vector getAllInOrder() {
        return entries;
    }

    /**
     * Insert an object
     * @param obj String
     * @param obj2 Object
     */
    public void insert(String obj, Object obj2) {
        if (obj == null || obj2 == null) {
            return;
        }

        Vector vec = (Vector) tbl.get(obj);
        if (vec == null) {
            vec = new Vector();
            tbl.put(obj, vec);
        }

        vec.addElement(obj2);
        entries.addElement(obj2);
    }

    /* Remove an object */
    public boolean remove(String obj, Object obj2) {
        if (tbl == null) {
            return false;
        }

        Vector vec = (Vector) tbl.get(obj);
        if (vec != null) {
            vec.remove(obj2);
            entries.remove(obj2);
            if (vec.size() == 0) { // garbage collect
                tbl.remove(obj);
            }
            return true;
        }

        return false;
    }

    /**
     * Remove a category (name)
     * @param obj String
     * @return Vector
     */
    public Vector remove(String obj) {
        if (tbl == null) {
            return null;
        }

        Vector vec = (Vector) tbl.remove(obj);
        if (vec == null) {
            return null;
        }
        for (int i = 0; i < vec.size(); i++) {
            Object o = vec.elementAt(i);
            entries.remove(o);
        }
        return vec;
    }

    /* Return size of the table */
    public int size() {
        if (tbl == null) {
            return 0;
        }
        return tbl.size();
    }

    /**
     * Extract an object category with name
     * @param obj String
     * @return Vector
     */
    public Vector get(String obj) {
        if (tbl == null || obj == null) {
            return null;
        }

        Vector vec = (Vector) tbl.get(obj);
        return vec;
    }

    /**
     * Extract the given object from the given category
     * @param obj String
     * @param i int
     * @return Object
     */
    public Object get(String obj, int i) {
        if (tbl == null || obj == null) {
            return null;
        }

        Vector vec = (Vector) tbl.get(obj);
        if (vec != null && vec.size() > i) {
            return vec.elementAt(i);
        }
        return null;
    }

    /**
     * Get the first object in the given category
     * @param obj String
     * @return Object
     */
    public Object getFirst(String obj) {
        if (tbl == null || obj == null) {
            return null;
        }
        Vector vec = (Vector) tbl.get(obj);
        if (vec != null && vec.size() > 0) {
            return vec.firstElement();
        }
        return null;
    }

    /**
     * Get the keys (names of categories)
     * @return Enumeration
     */
    public Enumeration getKeys() {
        if (tbl == null) {
            return null;
        }
        return tbl.keys();
    }

    public void printTemplates() {
        int j = 0;
        fuegocore.notify.util.Log.write("Printing triplet table");
        for (Enumeration e = tbl.elements(); e.hasMoreElements(); ) {
            Vector vec = (Vector) e.nextElement();

            fuegocore.notify.util.Log.write((j++) + " --------------");

            if (vec != null) {
                for (int i = 0; i < vec.size(); i++) {
                    Template ev = (Template) vec.elementAt(i);
                    fuegocore.notify.util.Log.writeEC(ev.toString());
                }
            }

        }
    }

    public void printTriplets() {
        int j = 0;
        System.out.println("Printing triplet table");
        for (Enumeration e = tbl.elements(); e.hasMoreElements(); ) {
            Vector vec = (Vector) e.nextElement();
            if (vec != null) {
                for (int i = 0; i < vec.size(); i++) {
                    EventTriplet ev = (EventTriplet) vec.elementAt(i);
                    System.out.println(ev.toString());
                }
            }

        }
    }

    public String printKeys() {
        String ret = "keys:";
        for (Enumeration e = tbl.keys(); e.hasMoreElements(); ) {
            ret += " " + (String) e.nextElement();
        }
        return ret;
    }

    public String statistics() {
        int j = 0, k = 0;
        for (Enumeration e = tbl.elements(); e.hasMoreElements(); ) {
            Vector vec = (Vector) e.nextElement();
            k += vec.size();
            j++;
        }
        return " 1-level count:" + j + " 2-level count:" + k;
    }

}
