/*
 * Copyright 2006 Helsinki Institute for Information Technology
 *
 * This file is a part of Fuego middleware.  Fuego middleware is free
 * software; you can redistribute it and/or modify it under the terms
 * of the MIT license, included as the file MIT-LICENSE in the Fuego
 * middleware source distribution.  If you did not receive the MIT
 * license with the distribution, write to the Fuego Core project at
 * fuego-core-users@hoslab.cs.helsinki.fi.
 */

package fuegocore.notify.util;

import java.util.*;

/**
 * Builds a history of values and gives an estimate of
 * the history. Used in QoS models.
 */
public class ValueHistory {
    int size = 3;
    Vector buffer; // actual observed data
    int count = 0;
    double lastforecast; // last forecast
    double forecast; // current forecast
    double forecast_alpha = 0.7; // forecast parameter

    public ValueHistory(int si) {
        size = si;
        buffer = new Vector();
        forecast = 0;
        lastforecast = 0;
    }

    /**
     * Add a new value and update the forecast
     * @param t long
     */
    public synchronized void add(long t) {
        buffer.add(new Long(t));

        if (buffer.size() > size) {
            // remove the first entry in the buffer
            buffer.remove(buffer.firstElement());
        }
        calculateForecast();
    }

    /**
     * Set forecast value
     * @param a double
     */
    public void setForecastAlpha(double a) {
        forecast_alpha = a;
    }

    /**
     * Get the current value of the forecast
     * @return double
     */
    public double getForecast() {
        return forecast;
    }

    /**
     * Uses exponential smoothing on the history of values.
     * Lambda is between 0 - 1.
     * The difference equation is:<br>
     * Ft+1 = lambda*xi + (1-lambda)*Ft
     *
     * @param alpha double
     * @return double
     */
    public double getValue(double alpha) {
        return getValue(buffer.size() - 1, alpha);
    }

    /**
     * Retrieve the current value of the history
     * @param i int
     * @param alpha double
     * @return double
     */
    public double getValue(int i, double alpha) {
        if (i < 0) {
            return 0;
        }
        long v = 0;

        try {
            Long l = (Long) buffer.elementAt(i);
            v = l.longValue();
        } catch (Exception e) {
            e.printStackTrace();
        }
        double val = alpha * v;
        val += (1 - alpha) * getValue(i - 1, alpha);

        return val;
    }

    /**
     * Uses exponential smoothing to forecast values
     * Lambda is between 0 - 1.
     * New Forecast = last period's forecast + alpha(last periods actual demand - last periods forecast.
     */
    public void calculateForecast() {
        Long l = (Long) buffer.lastElement();

        lastforecast = forecast;
        forecast = lastforecast +
                   forecast_alpha * (l.longValue() - lastforecast);
    }

    /**
     * Calculate the forecast
     * @param lambda double
     */
    public void calculateForecast(double lambda) {
        Long l = (Long) buffer.lastElement();

        lastforecast = forecast;
        forecast = lastforecast + lambda * (l.longValue() - lastforecast);
    }
}
